   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    BLOAD MODULE                     */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
#include "setup.h"

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "clipsmem.h"
#include "utility.h"
#include "expressn.h"
#include "evaluatn.h"
#include "router.h"
#include "constrct.h"
#include "bload.h"
#include "bsave.h"

#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)

#if DEFGENERIC_CONSTRUCT
#include "genrcbin.h"
#endif

#if DEFFUNCTION_CONSTRUCT
#include "deffnctn.h"
#endif

#if MAC_MPW || MAC_TC5
#include <Types.h>
#include <Files.h>
#if MAC_MPW
#include <strings.h>
#else
#include <pascal.h>
#endif
#endif

#if MAC_TC4
#undef TRUE
#undef FALSE
#include <MacTypes.h>
#include <FileMgr.h>
#include <pascal.h>
#endif

#if IBM_TBC || IBM_MSC
#include <io.h>
#include <fcntl.h>
#include <limits.h>
#endif  

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/  

#if ANSI_COMPILER         
   static struct FunctionDefinition * HUGE_ADDR *ReadNeededFunctions(unsigned long int *,int *);
   static struct FunctionDefinition  *FastFindFunction(char *,struct FunctionDefinition *); 
   static SYMBOL_HN * HUGE_ADDR      *ReadNeededSymbols(unsigned long int *);
   static FLOAT_HN * HUGE_ADDR       *ReadNeededFloats(unsigned long int *);
   static INTEGER_HN * HUGE_ADDR     *ReadNeededIntegers(unsigned long int *);
   static struct expr HUGE_ADDR      *BloadExpressions(void);
   static VOID                        RefreshExpressions(void);
   static int                         GenOpen(char *);
   static VOID                        GenClose(void);
   static int                         ClearBload(void);
   static VOID                        AbortBload(void);
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static struct BinaryFunction      *AddBloadFunctionToList(char *,int,VOID (*)(void),
                                                             struct BinaryFunction *);
#endif
#else   
   static struct FunctionDefinition * HUGE_ADDR *ReadNeededFunctions();
   static struct FunctionDefinition  *FastFindFunction(); 
   static SYMBOL_HN * HUGE_ADDR      *ReadNeededSymbols();
   static FLOAT_HN * HUGE_ADDR       *ReadNeededFloats();
   static INTEGER_HN * HUGE_ADDR     *ReadNeededIntegers();
   static struct expr HUGE_ADDR      *BloadExpressions();
   static VOID                        RefreshExpressions();
   static int                         GenOpen();
   static VOID                        GenClose();
   static int                         ClearBload();
   static VOID                        AbortBload();
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static struct BinaryFunction      *AddBloadFunctionToList();
#endif
#endif           
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static int                      BloadActive = CLIPS_FALSE;
   static unsigned long            NumberOfExpressions;
   static struct BinaryFunction   *BeforeBloadFunctions = NULL;
   static struct BinaryFunction   *AfterBloadFunctions = NULL;
   static struct BinaryFunction   *BloadReadyFunctions = NULL;
   static struct BinaryFunction   *ClearBloadReadyFunctions = NULL;
   static struct BinaryFunction   *AbortBloadFunctions = NULL;
   
#if  MAC_TC4 || MAC_TC5 || MAC_MPW
   static short int        BinaryRefNum;
#endif

#if IBM_TBC || IBM_MSC
   static int              BinaryFileHandle;
#endif

#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   static FILE            *BinaryFP;
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle char                                  *BinaryPrefixID = "\1\2\3\4CLIPS";
   globle char                                  *BinaryVersionID = "V5.10";
   globle struct expr HUGE_ADDR                 *ExpressionArray;
   globle SYMBOL_HN * HUGE_ADDR                 *SymbolArray;
   globle struct floatHashNode * HUGE_ADDR      *FloatArray;
   globle INTEGER_HN * HUGE_ADDR                *IntegerArray;
   globle struct FunctionDefinition * HUGE_ADDR *FunctionArray;

/*******************************************************/
/* Bload: Loads the binary representation of the CLIPS */
/*   environment from a file.                          */
/*******************************************************/
globle int Bload(fileName)
  char *fileName;
  {
   unsigned long int numberOfFunctions;
   int error,ready;
   unsigned long int numberOfSymbols, numberOfFloats, numberOfIntegers;
   char IDbuffer[20];
   char constructBuffer[CONSTRUCT_HEADER_SIZE];
   struct BinaryItem *biPtr;
   struct BinaryFunction *bfPtr;
   
   /*================*/
   /* Open the file. */
   /*================*/
   
   if (GenOpen(fileName) == 0) return(CLIPS_FALSE);
   
   /*=====================================*/
   /* Determine if this is a binary file. */
   /*=====================================*/
   
   GenRead(IDbuffer,(unsigned long) strlen(BinaryPrefixID) + 1);
   if (strcmp(IDbuffer,BinaryPrefixID) != 0)
     {
      PrintCLIPS(WERROR,"File ");
      PrintCLIPS(WERROR,fileName);
      PrintCLIPS(WERROR," is not a binary construct file\n");
      GenClose();
      return(CLIPS_FALSE);
     }

   GenRead(IDbuffer,(unsigned long) strlen(BinaryVersionID) + 1);
   if (strcmp(IDbuffer,BinaryVersionID) != 0)
     {
      PrintCLIPS(WERROR,"File ");
      PrintCLIPS(WERROR,fileName);
      PrintCLIPS(WERROR," is an incompatible binary construct file\n");
      GenClose();
      return(CLIPS_FALSE);
     }
       
   /*====================*/
   /* Clear environment. */
   /*====================*/
   
   if (BloadActive) 
     {
      if (ClearBload() == CLIPS_FALSE)
        {
         GenClose();
         return(CLIPS_FALSE); 
        } 
     }
   
   bfPtr = BeforeBloadFunctions;
   while (bfPtr != NULL)
     {
      (*bfPtr->func)();
      bfPtr = bfPtr->next;
     }
   
   /*========================================*/
   /* Make sure it's safe to do a bload now. */
   /*========================================*/
   
   bfPtr = BloadReadyFunctions;
   error = FALSE;
   while (bfPtr != NULL)
     {
      ready = (* ((int (*)(VOID_ARG)) bfPtr->func))();
      if (ready == FALSE)
        {
         if (! error)
           { PrintCLIPS(WERROR,"Some constructs are still in use by the current binary image:\n"); }
         PrintCLIPS(WERROR,"   ");
         PrintCLIPS(WERROR,bfPtr->name);
         PrintCLIPS(WERROR,"\n");
         error = TRUE;
        }
      bfPtr = bfPtr->next;
     }
     
   if (error == TRUE)
     {
      PrintCLIPS(WERROR,"Binary load cannot continue.\n");
      GenClose();
      AbortBload();
      return(CLIPS_FALSE);  
     }
   
   /*====================================================*/
   /* Read in the functions needed by this binary image. */
   /*====================================================*/

   FunctionArray = ReadNeededFunctions(&numberOfFunctions,&error);
   if (error)
     {
      GenClose();
      AbortBload();
      return(CLIPS_FALSE);
     }

   /*================================================*/
   /* Read in the atoms needed by this binary image. */
   /*================================================*/
      
   SymbolArray = ReadNeededSymbols(&numberOfSymbols);
   FloatArray = ReadNeededFloats(&numberOfFloats);
   IntegerArray = ReadNeededIntegers(&numberOfIntegers);
     
   /*======================================================*/
   /* Read in the expressions needed by this binary image. */
   /*======================================================*/
     
   ExpressionArray = BloadExpressions();
      
   /*======================================================*/
   /* Read in the constructs stored in this binary image.  */
   /* Any load functions which automatically handle the    */
   /* refresh may not rely on expressions being refreshed. */
   /*======================================================*/
     
   GenRead(constructBuffer,(unsigned long) CONSTRUCT_HEADER_SIZE);
   while (strncmp(constructBuffer,BinaryPrefixID,CONSTRUCT_HEADER_SIZE) != 0)
     {
      BOOLEAN found;
      
      found = CLIPS_FALSE;
      biPtr = ListOfBinaryItems;
      while (biPtr != NULL)
        {
         if (strncmp(biPtr->name,constructBuffer,CONSTRUCT_HEADER_SIZE) == 0)
           {
            if (biPtr->bloadFunction != NULL) 
              {
               (*biPtr->bloadFunction)();
               found = CLIPS_TRUE;
              }
            biPtr = NULL;
           }
         else biPtr = biPtr->next;
        }
        
      if (! found)
        {
         unsigned long int space;
         char *tempArray;
         
         GenRead(&space,(unsigned long) sizeof(unsigned long));
         tempArray = (char *) genlongalloc(space);
         GenRead(tempArray,space);
         PrintCLIPS(WDIALOG,"\nSkipping ");
         PrintCLIPS(WDIALOG,constructBuffer);
         PrintCLIPS(WDIALOG," constructs because of unavailibility\n");
         genlongfree(tempArray,space);
        }
        
      GenRead(constructBuffer,(unsigned long) CONSTRUCT_HEADER_SIZE);
     }
    
   /*=================*/
   /* Close the file. */
   /*=================*/
   
   GenClose();

   /*===============================================================================*/
   /* Refresh expressions and constructs by replacing array indices with addresses. */
   /*===============================================================================*/
   
   RefreshExpressions();
   
   biPtr = ListOfBinaryItems;
   while (biPtr != NULL)
     {
      if (biPtr->refreshFunction != NULL) (*biPtr->refreshFunction)();
      biPtr = biPtr->next;
     }
   
   /*==================*/
   /* Free up storage. */
   /*==================*/
   
   if (FunctionArray != NULL)
     genlongfree((VOID *) FunctionArray,(long) sizeof(struct FunctionDefinition *) * numberOfFunctions);
   if (SymbolArray != NULL)
     genlongfree((VOID *) SymbolArray,(long) sizeof(SYMBOL_HN *) * numberOfSymbols);
   if (FloatArray != NULL)
     genlongfree((VOID *) FloatArray,(long) sizeof(FLOAT_HN *) * numberOfFloats);
   if (IntegerArray != NULL)
     genlongfree((VOID *) IntegerArray,(long) sizeof(INTEGER_HN *) * numberOfIntegers);
     
   /*=============================*/
   /* Do any reset functionality. */
   /*=============================*/
  
   bfPtr = AfterBloadFunctions;
   while (bfPtr != NULL)
     {
      (*bfPtr->func)();
      bfPtr = bfPtr->next;
     }

   /*============================================*/
   /* Add a clear function to remove binary load */
   /* when a clear command is issued.            */
   /*============================================*/
   
   BloadActive = CLIPS_TRUE;
   AddClearFunction("bload",(VOID (*)()) ClearBload,100);
   
   /*================================================*/
   /* Return CLIPS_TRUE to indicate load successful. */
   /*================================================*/
   
   return(CLIPS_TRUE);
  }

/********************************************************/
/* ReadNeededFunctions: Reads in the names of functions */
/*   needed by the binary image.                        */
/********************************************************/
static struct FunctionDefinition * HUGE_ADDR *ReadNeededFunctions(numberOfFunctions,error)
  unsigned long int *numberOfFunctions;
  int *error;
  {
   char *functionNames, *namePtr;
   unsigned long int space, i, temp;
   struct FunctionDefinition * HUGE_ADDR *newFunctionArray, *functionPtr;
   int functionsNotFound = 0;
   
   /*===================================================*/
   /* Determine the number of function names to be read */
   /* and the space required for them.                  */
   /*===================================================*/
   
   GenRead(numberOfFunctions,(unsigned long) sizeof(unsigned long int));
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   if (*numberOfFunctions == 0) 
     {
      *error = CLIPS_FALSE;
      return(NULL);
     }
 
   /*=======================================*/
   /* Allocate area for strings to be read. */
   /*=======================================*/
   
   functionNames = genlongalloc(space);
   GenRead(functionNames,(unsigned long) space);
   
   /*====================================================*/
   /* Store the function pointers in the function array. */
   /*====================================================*/
   
   temp = (long) sizeof(struct FunctionDefinition *) * *numberOfFunctions;
   newFunctionArray = (struct FunctionDefinition **) genlongalloc(temp);
   namePtr = functionNames;
   functionPtr = NULL;
   for (i = 0; i < *numberOfFunctions; i++)
     {
      if ((functionPtr = FastFindFunction(namePtr,functionPtr)) == NULL)
        {
         if (! functionsNotFound) 
           { 
            PrintCLIPS(WERROR,"The following undefined functions are ");
            PrintCLIPS(WERROR,"referenced by this binary image:\n");
           }
           
         PrintCLIPS(WERROR,"   ");
         PrintCLIPS(WERROR,namePtr);
         PrintCLIPS(WERROR,"\n");
         functionsNotFound = 1;
        }
      
      newFunctionArray[i] = functionPtr;
      namePtr += strlen(namePtr) + 1;
     }

   /*==========================================*/
   /* Free the memory used by the name buffer. */
   /*==========================================*/
   
   genlongfree(functionNames,space);
   
   /*==================================================*/
   /* If any of the required functions were not found, */
   /* then free the memory used by the function array. */
   /*==================================================*/
   
   if (functionsNotFound)
     {
      genlongfree((VOID *) newFunctionArray,temp);
      newFunctionArray = NULL;
     }
     
   /*===================================*/
   /* Set globals to appropriate values */
   /* and return the function array.    */
   /*===================================*/
   
   *error = functionsNotFound;
   return(newFunctionArray);
  }
  
/*************************************************/
/* FastFindFunction: Search the CLIPS function */
/*   list for a specific function.               */
/*************************************************/
static struct FunctionDefinition *FastFindFunction(fun_name,end_function)
  char *fun_name;
  struct FunctionDefinition *end_function;
  {
   struct FunctionDefinition *fun_list, *fun_ptr;

   fun_list = GetFunctionList();
   if (fun_list == NULL) { return(NULL); }

   if (end_function != NULL)
     { fun_ptr = end_function->next; }
   else
     { fun_ptr = fun_list; }
     
   while (strcmp(fun_name,ValueToString(fun_ptr->callFunctionName)) != 0) 
     { 
      fun_ptr = fun_ptr->next;
      if (fun_ptr == end_function) return(NULL);
      if (fun_ptr == NULL) fun_ptr = fun_list;
     }

   return(fun_ptr);
  }
  
/************************************************/
/* ReadNeededSymbols: Reads in the symbols used */
/*   by the binary image.                       */
/************************************************/
static SYMBOL_HN * HUGE_ADDR *ReadNeededSymbols(numberOfSymbols)
  unsigned long int *numberOfSymbols;
  {
   char *symbolNames, *namePtr;
   unsigned long space;
   unsigned long i;
   
   /*=================================================*/
   /* Determine the number of symbol names to be read */
   /* and space required for them.                    */
   /*=================================================*/
   
   GenRead(numberOfSymbols,(unsigned long) sizeof(unsigned long int));
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   if (*numberOfSymbols == 0) return(NULL);
   
   /*=======================================================*/
   /* Determine the amount of space needed for the symbols. */
   /*=======================================================*/
    
  
   /*=======================================*/
   /* Allocate area for strings to be read. */
   /*=======================================*/
   
   symbolNames = genlongalloc(space);
   GenRead(symbolNames,space);
   
   /*================================================*/
   /* Store the symbol pointers in the symbol array. */
   /*================================================*/
   
   SymbolArray = (SYMBOL_HN **) 
                 genlongalloc((long) sizeof(SYMBOL_HN *) * (long) *numberOfSymbols);
   namePtr = symbolNames;
   for (i = 0; i < *numberOfSymbols; i++)
     {
      SymbolArray[i] = (SYMBOL_HN *) AddSymbol(namePtr);
      namePtr += strlen(namePtr) + 1;
     }
     
   /*=======================*/
   /* Free the name buffer. */
   /*=======================*/
   
   genlongfree(symbolNames,space);
   
   /*==========================*/
   /* Return the symbol array. */
   /*==========================*/

   return(SymbolArray);
  }
  
/**********************************************/
/* ReadNeededFloats: Reads in the floats used */
/*   by the binary image.                     */
/**********************************************/
static FLOAT_HN * HUGE_ADDR *ReadNeededFloats(numberOfFloats)
  unsigned long int *numberOfFloats;
  {
   double HUGE_ADDR *floatValues;
   unsigned long i;
   
   /*============================================*/
   /* Determine the number of floats to be read. */
   /*============================================*/
   
   GenRead(numberOfFloats,(unsigned long) sizeof(unsigned long int));
   if (*numberOfFloats == 0) return(NULL);
  
   /*===============================*/
   /* Allocate area for the floats. */
   /*===============================*/
   
   floatValues = (double *) genlongalloc(sizeof(double) * *numberOfFloats);
   GenRead((VOID *) floatValues,sizeof(double) * *numberOfFloats);
   
   /*======================================*/
   /* Store the floats in the float array. */
   /*======================================*/
   
   FloatArray = (FLOAT_HN **) 
               genlongalloc((long) sizeof(FLOAT_HN *) * (long) *numberOfFloats);
   for (i = 0; i < *numberOfFloats; i++)
     { FloatArray[i] = (FLOAT_HN *) AddDouble(floatValues[i]); }
     
   /*========================*/
   /* Free the float buffer. */
   /*========================*/
   
   genlongfree((VOID *) floatValues,sizeof(double) * *numberOfFloats);
   
   /*=========================*/
   /* Return the float array. */
   /*=========================*/

   return(FloatArray);
  }

/**********************************************/
/* ReadNeededIntegers: Reads in the integers used */
/*   by the binary image.                     */
/**********************************************/
static INTEGER_HN * HUGE_ADDR *ReadNeededIntegers(numberOfIntegers)
  unsigned long int *numberOfIntegers;
  {
   long int HUGE_ADDR *integerValues;
   unsigned long i;
   
   /*==============================================*/
   /* Determine the number of integers to be read. */
   /*==============================================*/
   
   GenRead(numberOfIntegers,(unsigned long) sizeof(unsigned long int));
   if (*numberOfIntegers == 0) return(NULL);
  
   /*=================================*/
   /* Allocate area for the integers. */
   /*=================================*/
   
   integerValues = (long *) genlongalloc(sizeof(long) * *numberOfIntegers);
   GenRead((VOID *) integerValues,sizeof(long) * *numberOfIntegers);
   
   /*==========================================*/
   /* Store the integers in the integer array. */
   /*==========================================*/
   
   IntegerArray = (INTEGER_HN **) 
           genlongalloc((long) sizeof(INTEGER_HN *) * (long) *numberOfIntegers);
   for (i = 0; i < *numberOfIntegers; i++)
     { IntegerArray[i] = (INTEGER_HN *) AddLong(integerValues[i]); }
     
   /*==========================*/
   /* Free the integer buffer. */
   /*==========================*/
   
   genlongfree((VOID *) integerValues,sizeof(long int) * *numberOfIntegers);
   
   /*===========================*/
   /* Return the integer array. */
   /*===========================*/

   return(IntegerArray);
  }

/**********************************************************************/
/* Bload Expressions: Loads in the expressions (i.e. test structures) */
/*   used by the binary image.                                        */
/**********************************************************************/
static struct expr HUGE_ADDR *BloadExpressions()
  {
   unsigned long space;
   struct expr HUGE_ADDR *testArray;
   
   /*=================================================*/
   /* Determine the number of expressions to be read. */
   /*=================================================*/
   
   GenRead(&NumberOfExpressions,(unsigned long) sizeof(long int));
   if (NumberOfExpressions == 0) return(NULL);
     
   /*==========================*/
   /* Read in the expressions. */
   /*==========================*/
   
   space = NumberOfExpressions * sizeof(struct expr);
   testArray = (struct expr *) genlongalloc(space);
   GenRead((VOID *) testArray,space);
   
   /*==============================*/
   /* Return the expression array. */
   /*==============================*/
   
   return(testArray);
  }
  
  
/**********************************************************/
/* RefreshExpressions: Resets the pointers in expressions */
/**********************************************************/
static VOID RefreshExpressions()
  {
   register unsigned long i;
   long index;
   
   for (i = 0; i < NumberOfExpressions; i++)
     {
      switch(ExpressionArray[i].type)
        {
         case FCALL:
           ExpressionArray[i].value = 
              (VOID *) FunctionArray[(long int) ExpressionArray[i].value];
           break;
         
         case GCALL:
#if DEFGENERIC_CONSTRUCT
           ExpressionArray[i].value = (VOID *) GenericPointer(ExpressionArray[i].value);
#else
           ExpressionArray[i].value = NULL;
#endif
           break;

         case PCALL:
#if DEFFUNCTION_CONSTRUCT
           ExpressionArray[i].value = (VOID *) DFPointer(ExpressionArray[i].value);
#else
           ExpressionArray[i].value = NULL;
#endif
           break;

         case INTEGER:
           ExpressionArray[i].value = 
              (VOID *) IntegerArray[(long int) ExpressionArray[i].value];
           IncrementCount(ExpressionArray[i].value);
           break;
           
         case FLOAT:
           ExpressionArray[i].value = 
             (VOID *) FloatArray[(long int) ExpressionArray[i].value];
           IncrementCount(ExpressionArray[i].value);
           break;

         case INSTANCE_NAME:
#if ! OBJECT_SYSTEM
           ExpressionArray[i].type = SYMBOL;
#endif
         case GBWORD:
         case SYMBOL:
         case STRING:
           ExpressionArray[i].value = 
                 (VOID *) SymbolArray[(long int) ExpressionArray[i].value];
           IncrementCount(ExpressionArray[i].value);
           break;
           
         case SCALL_COMP_VARS:
         case SCALL_GET_VAR:
           /* Do nothing for these cases. */
           break;
         
         default:
           /* This should be an error. */
           break;
        }
     
      index = (long int) ExpressionArray[i].next_arg;
      if (index == -1L)
        { ExpressionArray[i].next_arg = NULL; }
      else
        { ExpressionArray[i].next_arg = (struct expr *) &ExpressionArray[index]; }
        
      index = (long int) ExpressionArray[i].arg_list;
      if (index == -1L)
        { ExpressionArray[i].arg_list = NULL; }
      else
        { ExpressionArray[i].arg_list = (struct expr *) &ExpressionArray[index]; }
     
     }
  }
  
/*****************************************/
/* GenOpen: Generic and machine specific */
/*   code for opening a file.            */
/*****************************************/
static int GenOpen(fileName)
  char *fileName;
  {
#if  MAC_TC4 || MAC_TC5 || MAC_MPW
   Str255 tempName;
   OSErr resultCode;
   Str255 volName;
   short int vRefNum;
   
   resultCode = GetVol(volName,&vRefNum);
   if (resultCode != noErr) 
     {
      OpenErrorMessage("bload",fileName);
      return(0);
     }
   strcpy((char *) tempName,fileName);
#if MAC_MPW || MAC_TC5
   c2pstr((char *) tempName);
#endif
#if  MAC_TC4
   CtoPstr((char *) tempName);
#endif
   resultCode = FSOpen(tempName,vRefNum,&BinaryRefNum);
   if (resultCode != noErr) 
     {
      OpenErrorMessage("bload",fileName);
      return(0);
     }

#endif

#if IBM_TBC || IBM_MSC
   BinaryFileHandle = open(fileName,O_RDONLY | O_BINARY);
   if (BinaryFileHandle == -1)
     {
      OpenErrorMessage("bload",fileName);
      return(0);
     }
#endif

#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   if ((BinaryFP = fopen(fileName,"rb")) == NULL)
     {
      OpenErrorMessage("bload",fileName);
      return(0);
     }
#endif

   return(1);
  }
  
/*****************************************/
/* GenRead: Generic and machine specific */
/*   code for reading from a file.       */
/*****************************************/
globle VOID GenRead(dataPtr,size)
  VOID *dataPtr;
  unsigned long size;
  {
#if  MAC_TC4 || MAC_TC5 || MAC_MPW
   long dataSize;
   OSErr resultCode;
  
   dataSize = size;
   resultCode = FSRead(BinaryRefNum,&dataSize,dataPtr);
#endif

#if IBM_TBC || IBM_MSC
   char HUGE_ADDR *tempPtr;
   
   tempPtr = dataPtr;
   while (size > INT_MAX)
     {
      read(BinaryFileHandle,(VOID *) tempPtr,(unsigned int) INT_MAX);
      size -= INT_MAX;
      tempPtr = tempPtr + INT_MAX;
     }

   if (size > 0) read(BinaryFileHandle,(VOID *) tempPtr,(unsigned int) size);
#endif

#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC) 
   unsigned int temp, number_of_reads, read_size;
   
   if (sizeof(int) == sizeof(long))
     { read_size = size; }
   else
     { read_size = (1L << (sizeof(int) * 8L)) - 1L ; }
   number_of_reads = size / read_size;
   temp = size - ((long) number_of_reads * (long) read_size);
   
   while (number_of_reads > 0)
     {
      fread(dataPtr,read_size,1,BinaryFP);
      dataPtr = ((char *) dataPtr) + read_size;
      number_of_reads--;
     }
     
   fread(dataPtr,temp,1,BinaryFP);
#endif
  }
  
/*******************************************/
/* GenClose:  Generic and machine specific */
/*   code for closing a file.              */
/*******************************************/
static VOID GenClose()
  {
#if  MAC_TC4 || MAC_TC5 || MAC_MPW
   FSClose(BinaryRefNum);
#endif

#if IBM_TBC || IBM_MSC
   close(BinaryFileHandle);
#endif

#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   fclose(BinaryFP);
#endif
  }
    
/***********************************************************/
/* BLOADED: Returns CLIPS_TRUE if the current environment is the */
/*   result of a bload command, otherwise returns CLIPS_FALSE.   */
/***********************************************************/
globle BOOLEAN Bloaded()
  {
   return(BloadActive);
  }

/*****************************************************************/
/* ClearBload: Clears a binary image from the CLIPS environment. */
/*****************************************************************/
static int ClearBload()
  {
   unsigned long int i, space;
   struct BinaryItem *biPtr;
   struct BinaryFunction *bfPtr;
   int ready,error;

   /* ===========================================
      Make sure it's safe to do a clear-bload now
      =========================================== */
   bfPtr = ClearBloadReadyFunctions;
   error = FALSE;
   while (bfPtr != NULL)
     {
      ready = (* ((int (*)(VOID_ARG)) bfPtr->func))();
      if (ready == FALSE)
        {
         if (! error)
           { PrintCLIPS(WERROR,"Some constructs are still in use by the current binary image:\n"); }
         PrintCLIPS(WERROR,"   "); 
         PrintCLIPS(WERROR,bfPtr->name);
         PrintCLIPS(WERROR,"\n");
         error = TRUE;
        }
      bfPtr = bfPtr->next;
     }
     
   if (error == TRUE)
     {
      PrintCLIPS(WERROR,"Binary clear cannot continue.\n");
      return(CLIPS_FALSE);
     }
     
   
   /*==========================*/
   /* Update the symbol table. */
   /*==========================*/
   
   for (i = 0; i < NumberOfExpressions; i++)
     {
      if ((ExpressionArray[i].type == SYMBOL) || 
          (ExpressionArray[i].type == INSTANCE_NAME) ||
          (ExpressionArray[i].type == GBWORD) ||
          (ExpressionArray[i].type == STRING))
        { DecrementSymbolCount(ExpressionArray[i].value); }
      else if (ExpressionArray[i].type == FLOAT)
        { DecrementFloatCount(ExpressionArray[i].value); }
      else if (ExpressionArray[i].type == INTEGER)
        { DecrementIntegerCount(ExpressionArray[i].value); }
     }
     
   /*=============================*/
   /* Call bload clear functions. */
   /*=============================*/
   
   biPtr = ListOfBinaryItems;
   while (biPtr != NULL)
     {
      if (biPtr->clearFunction != NULL) (*biPtr->clearFunction)();
      biPtr = biPtr->next;
     }
   
   /*========================*/
   /* Free expression array. */
   /*========================*/ 
  
   space = NumberOfExpressions * sizeof(struct expr);
   if (space != 0) genlongfree((VOID *) ExpressionArray,space);
   
   /*==================================*/
   /* Remove the bload clear function. */
   /*==================================*/
   
   BloadActive = CLIPS_FALSE;
   RemoveClearFunction("bload");
   return(CLIPS_TRUE);
  }

/*******************************************************************************/
/* AbortBload: Cleans up effects of before-bload functions in event of failure */
/*******************************************************************************/
static VOID AbortBload()
  {
   struct BinaryFunction *bfPtr;

   bfPtr = AbortBloadFunctions;
   while (bfPtr != NULL)
     {
      (*bfPtr->func)();
      bfPtr = bfPtr->next;
     }
  }

/*****************************/
/* AddBeforeBloadFunction:   */
/*****************************/
globle VOID AddBeforeBloadFunction(name,func,priority)
  char *name;
  VOID (*func)();
  int priority;
  {
   BeforeBloadFunctions = AddBloadFunctionToList(name,priority,func,BeforeBloadFunctions);
  }
  
/*****************************/
/* AddAfterBloadFunction:   */
/*****************************/
globle VOID AddAfterBloadFunction(name,func,priority)
  char *name;
  VOID (*func)();
  int priority;
  {
   AfterBloadFunctions = AddBloadFunctionToList(name,priority,func,AfterBloadFunctions);
  }
  
/*****************************/
/* AddBloadReadyFunction:   */
/*****************************/
globle VOID AddBloadReadyFunction(name,func,priority)
  char *name;
  int (*func)();
  int priority;
  {
   BloadReadyFunctions = AddBloadFunctionToList(name,priority,
                         (VOID (*)(VOID_ARG)) func,BloadReadyFunctions);
  }
  
/*********************************/
/* AddClearBloadReadyFunction:   */
/*********************************/
globle VOID AddClearBloadReadyFunction(name,func,priority)
  char *name;
  int (*func)();
  int priority;
  {
   ClearBloadReadyFunctions = AddBloadFunctionToList(name,priority,
                            (VOID (*)(VOID_ARG)) func,ClearBloadReadyFunctions);
  }
  
/*****************************/
/* AddBeforeBloadFunction:   */
/*****************************/
globle VOID AddAbortBloadFunction(name,func,priority)
  char *name;
  VOID (*func)();
  int priority;
  {
   AbortBloadFunctions = AddBloadFunctionToList(name,priority,func,AbortBloadFunctions);
  }
  
/*****************************/
/* AddBloadFunctionToList:   */
/*****************************/
static struct BinaryFunction *AddBloadFunctionToList(name,priority,func,head)
  char *name;
  int priority;
  VOID (*func)();
  struct BinaryFunction *head;
  {
   struct BinaryFunction *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(BinaryFunction);

   newPtr->name = name;
   newPtr->func = func;
   newPtr->priority = priority;
   
   if (head == NULL)
     {
      newPtr->next = NULL;
      return(newPtr);
     }
     
   currentPtr = head;
   while ((currentPtr != NULL) ? (priority < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = head;
      head = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   return(head);
  }

#endif

/********************************************************/
/* BloadCommand: Performs the top level bload command. */
/********************************************************/
globle VOID BloadCommand()
  {
#if (! RUN_TIME) && (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   char *fileName;
   
   if (ArgCountCheck("bload",EXACTLY,1) == -1) return;
   fileName = GetFileName("bload",1);
   if (fileName != NULL) Bload(fileName);
#endif
  }
