   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*              CONSTRUCT COMPILER MODULE              */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*      Barry Cameron                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _CONSCOMP_SOURCE_

#include "setup.h"

#if CONSTRUCT_COMPILER && (! RUN_TIME)

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "symbol.h"
#include "clipsmem.h"
#include "constant.h"
#include "expressn.h"
#include "evaluatn.h"
#include "utility.h"
#include "router.h"
#include "conscomp.h"

#define FSIZE 80

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   VOID                               ConstructsToCCommand(void);
   static int                         GenerateCode(char *,int,int);
   static int                         SymbolsToCode(char *,int);
   static int                         FloatsToCode(char *,int);
   static int                         IntegersToCode(char *,int);
   static int                         HashTablesToCode(char *);
   static VOID                        MarkBuckets(void);
   static VOID                        UnmarkBuckets(void);
   static VOID                        ListUserFunctions(FILE *);
   static int                         FunctionsToCode(char *);
   static int                         SetUpInitFile(char *);
   static VOID                        PrintCString(FILE *,char *);
   VOID                               ConstructsToCCommandDefinition(void);
   static VOID                        DumpExpression(struct expr *);
#else
   VOID                               ConstructsToCCommand();
   static int                         GenerateCode();
   static int                         SymbolsToCode();
   static int                         FloatsToCode();
   static int                         IntegersToCode();
   static int                         HashTablesToCode();
   static VOID                        MarkBuckets();
   static VOID                        UnmarkBuckets();
   static VOID                        ListUserFunctions();
   static int                         FunctionsToCode();
   static int                         SetUpInitFile();
   static VOID                        PrintCString();
   VOID                               ConstructsToCCommandDefinition();
   static VOID                        DumpExpression();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static int                       ImageID;
   static FILE                     *HeaderFP;
   static char                     *FilePrefix;
   static int                       MaxIndices = 2000;
   static FILE                     *ExpressionFP;
   static BOOLEAN                   ExpressionHeader;
   static int                       ExpressionCount;
   static int                       ExpressionVersion;
   static struct CodeGeneratorItem *ListOfCodeGeneratorItems = NULL;

/*********************************************************/
/* ConstructsToCCommand:  Converts the rule network into the */
/*   corresponding C data structures.                    */
/*********************************************************/
globle VOID ConstructsToCCommand()
  {
   char *fileName;
   DATA_OBJECT arg_ptr;
   int arg_count;
   int id, max;
#if VMS || IBM_MSC || IBM_TBC
   int i;
#endif

   /*============================================*/
   /* Check for appropriate number of arguments. */
   /*============================================*/
   
   if ((arg_count = ArgRangeCheck("constructs-to-c",2,3)) == -1) return;

   /*====================================*/
   /* Get the file name to place C code. */
   /*====================================*/
  
   if (ArgTypeCheck("constructs-to-c",1,SYMBOL_OR_STRING,&arg_ptr) == CLIPS_FALSE)
     { return; }
     
   fileName = DOToString(arg_ptr);

#if VMS || IBM_MSC || IBM_TBC
   /*Check for '.' VAX or IBM PC fileName */
   for(i = 0 ; *(fileName+i) ; i++)
     {
      if(*(fileName+i) == '.')
        {
         PrintCLIPS(WERROR,"Invalid file name ");
         PrintCLIPS(WERROR,fileName);
         PrintCLIPS(WERROR," contains \'.\'\n");
         return;
        }
      }
#endif
   
   /*==============================*/
   /* Get the remaining arguments. */
   /*==============================*/
   
   if (ArgTypeCheck("constructs-to-c",2,INTEGER,&arg_ptr) == CLIPS_FALSE)
     { return; }
     
   id = DOToInteger(arg_ptr);
   if (id < 0)
     {
      ExpectedTypeError("constructs-to-c",2,"positive integer");
      return;
     }
   
   if (arg_count == 3)
     {
      if (ArgTypeCheck("constructs-to-c",3,INTEGER,&arg_ptr) == CLIPS_FALSE)
        { return; }
     
      max = DOToInteger(arg_ptr);
   
      if (max < 0)
        {
         ExpectedTypeError("constructs-to-c",3,"positive integer");
         return;
        }
     }
   else
     { max = 10000; }
   
   GenerateCode(fileName,id,max);
  }
 
/********************************************************/
/* GenerateCode:                                       */
/********************************************************/
static int GenerateCode(fileName,rule_id,max)
  char *fileName;
  int rule_id, max;
  {
   char fname[FSIZE];
   int version;
   int fileVersion;
   struct CodeGeneratorItem *cgPtr;
   
   /* =====================
      Call before functions
      ===================== */
   cgPtr = ListOfCodeGeneratorItems;
   while (cgPtr != NULL)
     {
      if (cgPtr->beforeFunction != NULL)
        (*cgPtr->beforeFunction)();
      cgPtr = cgPtr->next;
     }
     
   /*=================================================*/
   /* Do a periodic Cleanup without using heuristics. */
   /*=================================================*/
   
   PeriodicCleanup(CLIPS_FALSE,CLIPS_FALSE);
   
   FilePrefix = fileName;
   ImageID = rule_id;
   MaxIndices = max;
   ExpressionFP = NULL;
   ExpressionVersion = 1;
   ExpressionHeader = CLIPS_TRUE;
   ExpressionCount = 0;
   
   /*=====================================================*/
   /* Open a header file for dumping general information. */
   /*=====================================================*/
   
   sprintf(fname,"%s.h",fileName);
   if ((HeaderFP = fopen(fname,"w")) == NULL) 
     {
      OpenErrorMessage("constructs-to-c",fname);
      return(0);
     }
   fprintf(HeaderFP,"#include <stdio.h>\n");
   fprintf(HeaderFP,"#include \"setup.h\"\n"); 
   fprintf(HeaderFP,"#include \"expressn.h\"\n"); 
   fprintf(HeaderFP,"#include \"clips.h\"\n");  
   fprintf(HeaderFP,"\n#define VS (VOID *)\n"); 
   fprintf(HeaderFP,"\n");
   
   /*=========================================================*/
   /* Give extern declarations for user and system functions. */
   /*=========================================================*/
   
   ListUserFunctions(HeaderFP);
   
   fprintf(HeaderFP,"\n\n");
   fprintf(HeaderFP,"/****************************/\n");
   fprintf(HeaderFP,"/* EXTERN ARRAY DEFINITIONS */\n");
   fprintf(HeaderFP,"/****************************/\n\n");
   
   /*===============================================*/
   /* Give definitions of the external and internal */
   /* functions to be accessed by test structures   */
   /* and function definition structures.           */
   /*===============================================*/

   MarkBuckets();
   
   HashTablesToCode(fileName);
   
   version = SymbolsToCode(fileName,4);
   version = FloatsToCode(fileName,version);
   version = IntegersToCode(fileName,version);
   
   FunctionsToCode(fileName);
   
   fileVersion = 4;
   cgPtr = ListOfCodeGeneratorItems;
   while (cgPtr != NULL)
     {
      if (cgPtr->generateFunction != NULL)
        {
         (*cgPtr->generateFunction)(fileName,fileVersion,HeaderFP,ImageID,MaxIndices);
         fileVersion++;
        }
      cgPtr = cgPtr->next;
     }
     
   UnmarkBuckets();
   
   if (ExpressionFP != NULL) 
     { 
      fprintf(ExpressionFP,"};\n");
      fclose(ExpressionFP);
     }
     
   fclose(HeaderFP);
   
   /*====================================*/
   /* Write the initialization function. */
   /*====================================*/
   
   SetUpInitFile(fileName);
   
   return(1);
  }

/**************************************/
/**************************************/
/******                          ******/
/****** DEFINE THE FUNCTION LIST ******/
/******                          ******/
/**************************************/
/**************************************/

/**********************************************************/
/* ListUserFunctions:  Loop through the list of function  */
/*    definitions, and define them as external functions. */
/**********************************************************/
static VOID ListUserFunctions(fp)
  FILE *fp;
  {
   struct FunctionDefinition *fctn_ptr;

   fprintf(fp,"\n");
   fprintf(fp,"/************************************/\n");
   fprintf(fp,"/* EXTERNAL FUNCTION DEFINITIONS    */\n");
   fprintf(fp,"/************************************/\n\n");
   
   fctn_ptr = GetFunctionList();
   while (fctn_ptr != NULL)
     {
      fprintf(fp,"extern ");
      switch(fctn_ptr->returnValueType)
        {
         case 'i':
         case 'b':
           fprintf(fp,"int ");
           break;
           
         case 'l':
           fprintf(fp,"long ");
           break;
           
         case 'f':
           fprintf(fp,"float ");
           break;
           
         case 'd':
           fprintf(fp,"double ");
           break;
           
         case 'w':
         case 's':
           fprintf(fp,"SYMBOL_HN *");
           break;
           
         case 'c':
           fprintf(fp,"char ");
           break;
           
         case 'a':
           fprintf(fp,"VOID * ");
           break;
                      
         case 'v':           
         case 'm':
         case 'u':
           fprintf(fp,"VOID ");
           break;
        }
        
      fprintf(fp,"%s();\n",fctn_ptr->actualFunctionName);
      fctn_ptr = fctn_ptr->next;
     }
  }

/*****************************************************/
/* FunctionsToCode: Produce the structure definitions */
/*   for the function list.                          */
/*****************************************************/
static int FunctionsToCode(fileName)
  char *fileName;
  {
   int i = 0;
   FILE *fp;
   int version = 1;
   int newHeader = CLIPS_TRUE;
   struct FunctionDefinition *fctnPtr;
   
   fctnPtr = GetFunctionList();
   while (fctnPtr != NULL)
     {
      fctnPtr->bsaveIndex = i++;
      fctnPtr = fctnPtr->next;
     }
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   if ((fp = NewCFile(fileName,2,version)) == NULL) return(0);

   /*===============================================*/
   /* Construct the definition of the function list */
   /* from the definitions of the functions.        */
   /*===============================================*/

   fprintf(fp,"\n\n");
   fprintf(fp,"/************************************/\n");
   fprintf(fp,"/* FUNCTION LIST DEFINITION         */\n");
   fprintf(fp,"/************************************/\n\n");

   i = 1;
   fctnPtr = GetFunctionList();
   while (fctnPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct FunctionDefinition ftd%d_%d[] = {\n",ImageID,version);
         fprintf(HeaderFP,"extern struct FunctionDefinition ftd%d_%d[];\n",ImageID,version);
         newHeader = CLIPS_FALSE;
        }
     
      fprintf(fp,"{"); 
      PrintSymbolReference(fp,fctnPtr->callFunctionName);
      fprintf(fp,",\"%s\",",fctnPtr->actualFunctionName);
      fprintf(fp,"'%c',",fctnPtr->returnValueType);
      fprintf(fp,"PTIF %s,",fctnPtr->actualFunctionName);
      fprintf(fp,"NULL,");
      fprintf(fp,"0,");

      PrintFunctionReference(fp,fctnPtr->next);
        
      i++;
      fctnPtr = fctnPtr->next;
      if ((i > MaxIndices) || (fctnPtr == NULL))
        {
         fprintf(fp,"}};\n");
         fclose(fp);
         i = 1;
         version++;
         if (fctnPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,2,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,"},\n"); }
     }
     
   return(1);
  }
  
/**********************************************************************/
/* SymbolsToCode: Loops through the hash table calling enum_entries to */
/*   produce the definitions for each of the hash table entries.      */
/**********************************************************************/
static int SymbolsToCode(fileName,version)
  char *fileName;
  int version;
  {
   int i, j;
   struct symbolHashNode *hash_ptr;
   int count;
   int number_of_entries;
   struct symbolHashNode **symbolTable;
   int newHeader = CLIPS_TRUE;
   int arrayVersion = 1;
   FILE *fp;
   
   /*====================================*/
   /* Count the total number of entries. */
   /*====================================*/
   
   symbolTable = GetSymbolTable();
   count = number_of_entries = 0;
   
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      hash_ptr = symbolTable[i];
      while (hash_ptr != NULL)
        { 
         number_of_entries++;
         hash_ptr = hash_ptr->next;
        }
     }
     
   if (number_of_entries == 0) return(version);
   
   for (i = 1; i <= (number_of_entries / MaxIndices) + 1 ; i++)
     { fprintf(HeaderFP,"extern struct symbolHashNode shn%d_%d[];\n",ImageID,i); }
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   if ((fp = NewCFile(fileName,1,version)) == NULL) return(-1);
   
   /*===================*/
   /* List the entries. */
   /*===================*/
   
   j = 1;
   
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      hash_ptr = symbolTable[i];
      while (hash_ptr != NULL)
        { 
         if (newHeader)
           {
            fprintf(fp,"struct symbolHashNode shn%d_%d[] = {\n",ImageID,arrayVersion);
            newHeader = CLIPS_FALSE;
           }
        
         if (hash_ptr->next == NULL)
           { fprintf(fp,"{NULL,"); }
         else
           { 
            if ((count + 1) == MaxIndices)
              { fprintf(fp,"{&shn%d_%d[%d],",ImageID,arrayVersion + 1,0); }
            else
              { fprintf(fp,"{&shn%d_%d[%d],",ImageID,arrayVersion,j); }
           }
           
         fprintf(fp,"%d,0,0,0,%d,",hash_ptr->count + 1,i); 
         PrintCString(fp,hash_ptr->contents);
         
         fprintf(fp,",");
         PrintSymbolReference(fp,hash_ptr->relatedSymbol);
         
         count++;
         j++;
         
         if ((count == number_of_entries) || (j > MaxIndices))
           { 
            fprintf(fp,"}};\n");
            fclose(fp);
            j = 1;
            arrayVersion++;
            version++;
            if (count < number_of_entries)
              {
               if ((fp = NewCFile(fileName,1,version)) == NULL) return(0);
               newHeader = CLIPS_TRUE;
              } 
           }
         else
           { fprintf(fp,"},\n"); }
           
         hash_ptr = hash_ptr->next;
        }
     }
     
   return(version);
  }
  
/**********************************************************************/
/* FloatsToCode: Loops through the hash table calling enum_entries to */
/*   produce the definitions for each of the hash table entries.      */
/**********************************************************************/
static int FloatsToCode(fileName,version)
  char *fileName;
  int version;
  {
   int i, j;
   struct floatHashNode *hash_ptr;
   int count;
   int number_of_entries;
   struct floatHashNode **floatTable;
   int newHeader = CLIPS_TRUE;
   FILE *fp;
   int arrayVersion = 1;
   
   
   /*====================================*/
   /* Count the total number of entries. */
   /*====================================*/
   
   floatTable = GetFloatTable();
   count = number_of_entries = 0;
   
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      hash_ptr = floatTable[i];
      while (hash_ptr != NULL)
        { 
         number_of_entries++;
         hash_ptr = hash_ptr->next;
        }
     }
     
   if (number_of_entries == 0) return(version);
   
   for (i = 1; i <= (number_of_entries / MaxIndices) + 1 ; i++)
     { fprintf(HeaderFP,"extern struct floatHashNode fhn%d_%d[];\n",ImageID,i); }
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   if ((fp = NewCFile(fileName,1,version)) == NULL) return(-1);
   
   /*===================*/
   /* List the entries. */
   /*===================*/
   
   j = 1;
   
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      hash_ptr = floatTable[i];
      while (hash_ptr != NULL)
        { 
         if (newHeader)
           {
            fprintf(fp,"struct floatHashNode fhn%d_%d[] = {\n",ImageID,arrayVersion);
            newHeader = CLIPS_FALSE;
           }
        
         if (hash_ptr->next == NULL)
           { fprintf(fp,"{NULL,"); }
         else
           { 
            if ((count + 1) == MaxIndices)
              { fprintf(fp,"{&fhn%d_%d[%d],",ImageID,arrayVersion + 1,0); }
            else
              { fprintf(fp,"{&fhn%d_%d[%d],",ImageID,arrayVersion,j); }
           }
           
         fprintf(fp,"%d,0,0,0,%d,",hash_ptr->count + 1,i); 
         fprintf(fp,"%s",FloatToString(hash_ptr->contents));
         
         count++;
         j++;
         
         if ((count == number_of_entries) || (j > MaxIndices))
           { 
            fprintf(fp,"}};\n");
            fclose(fp);
            j = 1;
            version++;
            arrayVersion++;
            if (count < number_of_entries)
              {
               if ((fp = NewCFile(fileName,1,version)) == NULL) return(0);
               newHeader = CLIPS_TRUE;
              } 
           }
         else
           { fprintf(fp,"},\n"); }
           
         hash_ptr = hash_ptr->next;
        }
     }
     
   return(version);
  }
  
/**********************************************************************/
/* IntegersToCode: Loops through the hash table calling enum_entries to */
/*   produce the definitions for each of the hash table entries.      */
/**********************************************************************/
static int IntegersToCode(fileName,version)
  char *fileName;
  int version;
  {
   int i, j;
   struct integerHashNode *hash_ptr;
   int count;
   int number_of_entries;
   struct integerHashNode **integerTable;
   int newHeader = CLIPS_TRUE;
   FILE *fp;
   int arrayVersion = 1;
   
   
   /*====================================*/
   /* Count the total number of entries. */
   /*====================================*/
   
   integerTable = GetIntegerTable();
   count = number_of_entries = 0;
   
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      hash_ptr = integerTable[i];
      while (hash_ptr != NULL)
        { 
         number_of_entries++;
         hash_ptr = hash_ptr->next;
        }
     }
     
   if (number_of_entries == 0) return(version);
   
   for (i = 1; i <= (number_of_entries / MaxIndices) + 1 ; i++)
     { fprintf(HeaderFP,"extern struct integerHashNode ihn%d_%d[];\n",ImageID,i); }
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   if ((fp = NewCFile(fileName,1,version)) == NULL) return(-1);
   
   /*===================*/
   /* List the entries. */
   /*===================*/
   
   j = 1;
   
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      hash_ptr = integerTable[i];
      while (hash_ptr != NULL)
        { 
         if (newHeader)
           {
            fprintf(fp,"struct integerHashNode ihn%d_%d[] = {\n",ImageID,arrayVersion);
            newHeader = CLIPS_FALSE;
           }
        
         if (hash_ptr->next == NULL)
           { fprintf(fp,"{NULL,"); }
         else
           { 
            if ((count + 1) == MaxIndices)
              { fprintf(fp,"{&ihn%d_%d[%d],",ImageID,arrayVersion + 1,0); }
            else
              { fprintf(fp,"{&ihn%d_%d[%d],",ImageID,arrayVersion,j); }
           }
           
         fprintf(fp,"%d,0,0,0,%d,",hash_ptr->count + 1,i); 
         fprintf(fp,"%ld",hash_ptr->contents);
         
         count++;
         j++;
         
         if ((count == number_of_entries) || (j > MaxIndices))
           { 
            fprintf(fp,"}};\n");
            fclose(fp);
            j = 1;
            version++;
            arrayVersion++;
            if (count < number_of_entries)
              {
               if ((fp = NewCFile(fileName,1,version)) == NULL) return(0);
               newHeader = CLIPS_TRUE;
              } 
           }
         else
           { fprintf(fp,"},\n"); }
           
         hash_ptr = hash_ptr->next;
        }
     }
     
   return(version);
  }

/****************************************************************/
/* HashTablesToCode: Produces the definition of the hash tables */
/*   for symbols, floats, and integers.                         */
/****************************************************************/
static int HashTablesToCode(fileName)
  char *fileName;
  {
   int i;
   FILE *fp;
   struct symbolHashNode **symbolTable;
   struct floatHashNode **floatTable;
   struct integerHashNode **integerTable;

   /*========================*/
   /* Dump the symbol table. */
   /*========================*/
   
   symbolTable = GetSymbolTable();
   
   if ((fp = NewCFile(fileName,1,1)) == NULL) return(0);
   
   fprintf(HeaderFP,"extern struct symbolHashNode *sht%d[];\n",ImageID);
   fprintf(fp,"struct symbolHashNode *sht%d[%d] = {\n",ImageID,SYMBOL_HASH_SIZE);
   
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
      {
       PrintSymbolReference(fp,symbolTable[i]);

       if (i + 1 != SYMBOL_HASH_SIZE) fprintf(fp,",\n");
      }

    fprintf(fp,"};\n");
    
    fclose(fp);
    
   /*========================*/
   /* Dump the float table. */
   /*========================*/
   
   floatTable = GetFloatTable();
   
   if ((fp = NewCFile(fileName,1,2)) == NULL) return(0);
   
   fprintf(HeaderFP,"extern struct floatHashNode *fht%d[];\n",ImageID);
   fprintf(fp,"struct floatHashNode *fht%d[%d] = {\n",ImageID,FLOAT_HASH_SIZE);
   
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
      {
       if (floatTable[i] == NULL) { fprintf(fp,"NULL"); }
       else PrintFloatReference(fp,floatTable[i]);

       if (i + 1 != FLOAT_HASH_SIZE) fprintf(fp,",\n");
      }

    fprintf(fp,"};\n");
    
    fclose(fp);
    
   /*=========================*/
   /* Dump the integer table. */
   /*=========================*/
   
   integerTable = GetIntegerTable();
   
   if ((fp = NewCFile(fileName,1,3)) == NULL) return(0);
   
   fprintf(HeaderFP,"extern struct integerHashNode *iht%d[];\n",ImageID);
   fprintf(fp,"struct integerHashNode *iht%d[%d] = {\n",ImageID,INTEGER_HASH_SIZE);
   
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
      {
       if (integerTable[i] == NULL) { fprintf(fp,"NULL"); }
       else PrintIntegerReference(fp,integerTable[i]);

       if (i + 1 != INTEGER_HASH_SIZE) fprintf(fp,",\n");
      }

    fprintf(fp,"};\n");
    
    fclose(fp);
    
    return(1);
   }
   
/************************************************************/
/* PrintSymbolReference:                          */
/************************************************************/
globle VOID PrintSymbolReference(fp,sPtr)
  FILE *fp;
  struct symbolHashNode *sPtr;
  {
   if (sPtr == NULL) fprintf(fp,"NULL");
   else fprintf(fp,"&shn%d_%d[%d]",ImageID,
                                    (sPtr->bucket / MaxIndices) + 1,
                                    sPtr->bucket % MaxIndices); 
  }
  
/************************************************************/
/* PrintFloatReference:                          */
/************************************************************/
globle VOID PrintFloatReference(fp,sPtr)
  FILE *fp;
  struct floatHashNode *sPtr;
  {
   fprintf(fp,"&fhn%d_%d[%d]",ImageID,
                                 (sPtr->bucket / MaxIndices) + 1,
                                 sPtr->bucket % MaxIndices); 
  }
  
/************************************************************/
/* PrintIntegerReference:                          */
/************************************************************/
globle VOID PrintIntegerReference(fp,sPtr)
  FILE *fp;
  struct integerHashNode *sPtr;
  {
   fprintf(fp,"&ihn%d_%d[%d]",ImageID,
                                 (sPtr->bucket / MaxIndices) + 1,
                                 sPtr->bucket % MaxIndices); 
  }

/************************************************************/
/* PrintFunctionReference:                          */
/************************************************************/
globle VOID PrintFunctionReference(fp,funcPtr)
  FILE *fp;
  struct FunctionDefinition *funcPtr;
  {
   if (funcPtr == NULL) fprintf(fp,"NULL");
   else
      fprintf(fp,"&ftd%d_%d[%d]",ImageID,
                                  (funcPtr->bsaveIndex / MaxIndices) + 1,
                                   funcPtr->bsaveIndex % MaxIndices); 
  }     
           
#if DEFFUNCTION_CONSTRUCT
/************************************************************/
/* PrintDeffunctionReference:                          */
/************************************************************/
globle VOID PrintDeffunctionReference(fp,dfPtr)
  FILE *fp;
  struct dfunc *dfPtr;
  {
   if (dfPtr == NULL) fprintf(fp,"NULL");
   else
      fprintf(fp,"&dfn%d_%d[%d]",ImageID,
                                   (int) ((dfPtr->bsaveIndex / MaxIndices) + 1),
                                   (int) (dfPtr->bsaveIndex % MaxIndices)); 
  }     
#endif
           
#if DEFGENERIC_CONSTRUCT
/************************************************************/
/* PrintGenericFunctionReference:                          */
/************************************************************/
globle VOID PrintGenericFunctionReference(fp,gfunc)
  FILE *fp;
  GENERIC_FUNC *gfunc;
  {
   if (gfunc == NULL) fprintf(fp,"NULL");
   else
      fprintf(fp,"&gfn%d_%d[%d]",ImageID,
                                  (int) ((gfunc->bsaveIndex / MaxIndices) + 1),
                                  (int) (gfunc->bsaveIndex % MaxIndices)); 
  }     
#endif
           
/************************************************************/
/* SetUpInitFile: Produces the initialization function */
/*   for installing this rule set.                          */
/************************************************************/
static int SetUpInitFile(fileName)
  char *fileName;
  {
   char fname[FSIZE];
   FILE *fp;
   struct CodeGeneratorItem *cgPtr;
   
   sprintf(fname,"%s.c",fileName);
   if ((fp = fopen(fname,"w")) == NULL) 
     {
      OpenErrorMessage("constructs-to-c",fname);
      return(0);
     }
     
   fprintf(fp,"#include \"%s.h\"\n",fileName); 
   fprintf(fp,"\n");
   fprintf(fp,"#include \"utility.h\"\n");
   fprintf(fp,"#include \"generate.h\"\n");
   fprintf(fp,"#include \"build.h\"\n\n");
      
   fprintf(fp,"#if ANSI_COMPILER\n");
   fprintf(fp,"   VOID InitCImage_%d(void);\n",ImageID);
   fprintf(fp,"#else\n");
   fprintf(fp,"   VOID InitCImage_%d();\n",ImageID);
   fprintf(fp,"#endif\n");
   
   fprintf(fp,"\n");
   fprintf(fp,"/*******************************************/\n");
   fprintf(fp,"/* CONSTRUCT IMAGE INITIALIZATION FUNCTION */\n");
   fprintf(fp,"/*******************************************/\n");

   fprintf(fp,"\nVOID InitCImage_%d()\n",ImageID);
   fprintf(fp,"  {\n");

   fprintf(fp,"   ClearCLIPS();\n");
   fprintf(fp,"   PeriodicCleanup(CLIPS_TRUE,CLIPS_FALSE);\n");
   fprintf(fp,"   SetSymbolTable(sht%d);\n",ImageID);
   fprintf(fp,"   SetFloatTable(fht%d);\n",ImageID);
   fprintf(fp,"   SetIntegerTable(iht%d);\n",ImageID);
   fprintf(fp,"   RefreshBooleanSymbols();\n");
   fprintf(fp,"   InstallFunctionList(ftd%d_1);\n\n",ImageID);
   fprintf(fp,"   InitGenModule();\n\n");

   /*===============================*/
   /* User specific initialization. */
   /*===============================*/
  
   cgPtr = ListOfCodeGeneratorItems;
   while (cgPtr != NULL)
     {
      if (cgPtr->initFunction != NULL) 
        {
         (*cgPtr->initFunction)(fp,ImageID,MaxIndices);
         fprintf(fp,"\n");
        }
      cgPtr = cgPtr->next;
     }
     
   /*================================*/
   /* Close the initialization file. */
   /*================================*/
   
   fprintf(fp,"  }\n");
   
   fclose(fp);
   
   return(1);
  }

/************************************************************/
/* PrintCString:        */
/************************************************************/
static VOID PrintCString(file_ptr,str)
  FILE *file_ptr;
  char *str;
  {
   int i, slen;
   
   fprintf(file_ptr,"\"");
   slen = strlen(str);
   for (i = 0 ; i < slen ; i++)
     {
      if ((str[i] == '"') || (str[i] == '\\'))
        { 
         fputc('\\',file_ptr);
         fputc(str[i],file_ptr);
        }
      else if (str[i] == '\n')
        {
         fputc('\\',file_ptr);
         fputc('n',file_ptr);
        }
      else
        { fputc(str[i],file_ptr); }
     }
   
   fprintf(file_ptr,"\"");
  }
  
/*******************************************/
/* NewCFile:              */
/*******************************************/
globle FILE *NewCFile(fileName,id,version)
  char *fileName;
  int id, version;
  {
   char fname[FSIZE];
   FILE *newFP;
   
   sprintf(fname,"%s%d_%d.c",fileName,id,version);
   
   if ((newFP = fopen(fname,"w")) == NULL) 
     {
      OpenErrorMessage("constructs-to-c",fname);
      return(NULL);
     }
     
   fprintf(newFP,"#include \"%s.h\"\n",fileName); 
   fprintf(newFP,"\n");
   return(newFP);
  }

/*****************************************************/
/* MarkBuckets: Pretty much duplicated from bsave. */
/*****************************************************/
static VOID MarkBuckets()
  {
   unsigned int count;
   int i;
   SYMBOL_HN *symbolPtr, **symbolArray;
   FLOAT_HN *floatPtr, **floatArray;
   INTEGER_HN *integerPtr, **integerArray;
   
   symbolArray = GetSymbolTable();
   count = 0;
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         symbolPtr->bucket = count++;
         symbolPtr = symbolPtr->next;
        }
     }
     
   floatArray = GetFloatTable();
   count = 0;
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         floatPtr->bucket = count++;
         floatPtr = floatPtr->next;
        }
     }
     
   integerArray = GetIntegerTable();
   count = 0;
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         integerPtr->bucket = count++;
         integerPtr = integerPtr->next;
        }
     }
  }
  
/*****************************************************/
/* UnmarkBuckets: Pretty much duplicated from bsave. */
/*****************************************************/
static VOID UnmarkBuckets()
  {
   int i;
   SYMBOL_HN *symbolPtr, **symbolArray;
   FLOAT_HN *floatPtr, **floatArray;
   INTEGER_HN *integerPtr, **integerArray;
   
   symbolArray = GetSymbolTable();
      
   for (i = 0; i < SYMBOL_HASH_SIZE; i++)
     {
      symbolPtr = symbolArray[i];
      while (symbolPtr != NULL)
        {
         symbolPtr->bucket = i;
         symbolPtr = symbolPtr->next;
        }
     }
     
   floatArray = GetFloatTable();
      
   for (i = 0; i < FLOAT_HASH_SIZE; i++)
     {
      floatPtr = floatArray[i];
      while (floatPtr != NULL)
        {
         floatPtr->bucket = i;
         floatPtr = floatPtr->next;
        }
     }
     
   integerArray = GetIntegerTable();
      
   for (i = 0; i < INTEGER_HASH_SIZE; i++)
     {
      integerPtr = integerArray[i];
      while (integerPtr != NULL)
        {
         integerPtr->bucket = i;
         integerPtr = integerPtr->next;
        }
     }
  }

/**********************************************************/
/* ExpressionToCode:   */
/**********************************************************/
globle int ExpressionToCode(fp,exprPtr)
  FILE *fp;
  struct expr *exprPtr;
  {
   /*========================================*/
   /* Print the reference to the expression. */
   /*========================================*/
   
   if (exprPtr == NULL) 
     {
      fprintf(fp,"NULL");
      return(0);
     }
   else
     { fprintf(fp,"&exp%d_%d[%d]",ImageID,ExpressionVersion,ExpressionCount); }
     
   /*==================================================*/
   /* Create a new expression code file, if necessary. */
   /*==================================================*/
   
   if (ExpressionHeader == CLIPS_TRUE)
     {
      if ((ExpressionFP = NewCFile(FilePrefix,3,ExpressionVersion)) == NULL) 
        { return(-1); }
     
      fprintf(ExpressionFP,"struct expr exp%d_%d[] = {\n",ImageID,ExpressionVersion);
      fprintf(HeaderFP,"extern struct expr exp%d_%d[];\n",ImageID,ExpressionVersion);
      ExpressionHeader = CLIPS_FALSE;
     }
   else
     { fprintf(ExpressionFP,",\n"); }
     
   /*===========================*/
   /* Dump the expression code. */
   /*===========================*/
     
   DumpExpression(exprPtr);
   
   /*=========================================*/
   /* Close the expression file if necessary. */
   /*=========================================*/
   
   if (ExpressionCount >= MaxIndices)
     {
      ExpressionCount = 0;
      ExpressionVersion++;
      fprintf(ExpressionFP,"};\n");
      fclose(ExpressionFP);
      ExpressionFP = NULL;
      ExpressionHeader = CLIPS_TRUE;
     }
     
   return(1);
  }
     
/**********************************************************/
/* DumpExpression:   */
/**********************************************************/
static VOID DumpExpression(exprPtr)
  struct expr *exprPtr;
  {
     
   while (exprPtr != NULL)
     {
      fprintf(ExpressionFP,"{");
      fprintf(ExpressionFP,"%d,",exprPtr->type);
      fprintf(ExpressionFP,"VS ");
      switch (exprPtr->type)
        {
         case FCALL:
           PrintFunctionReference(ExpressionFP,exprPtr->value);
           break;
           
         case INTEGER:
           PrintIntegerReference(ExpressionFP,exprPtr->value);
           break;
           
         case FLOAT:
           PrintFloatReference(ExpressionFP,exprPtr->value);
           break;
           
         case PCALL:
#if DEFFUNCTION_CONSTRUCT
           PrintDeffunctionReference(ExpressionFP,(struct dfunc *) exprPtr->value);
#else
           fprintf(ExpressionFP,"NULL");
#endif         
           break;
  
         case GCALL:
#if DEFGENERIC_CONSTRUCT
           PrintGenericFunctionReference(ExpressionFP,(GENERIC_FUNC *) exprPtr->value);
#else
           fprintf(ExpressionFP,"NULL");
#endif         
           break;
  
         case STRING:
         case SYMBOL:
         case INSTANCE_NAME:
         case GBWORD:
           PrintSymbolReference(ExpressionFP,exprPtr->value);
           break;
           
         case SCALL_COMP_VARS:
         case SCALL_GET_VAR:
           fprintf(ExpressionFP,"%lu",(unsigned long) exprPtr->value);
           break;
        }
      fprintf(ExpressionFP,",");
      
      ExpressionCount++;
      if (exprPtr->arg_list == NULL) 
        { fprintf(ExpressionFP,"NULL,"); }
      else
        { 
         fprintf(ExpressionFP,"&exp%d_%d[%d],",ImageID,ExpressionVersion,
                                                       ExpressionCount); 
        }
        
      if (exprPtr->next_arg == NULL) 
        { fprintf(ExpressionFP,"NULL}"); }
      else
        { 
         fprintf(ExpressionFP,"&exp%d_%d[%d]}",ImageID,ExpressionVersion,
                              ExpressionCount + ExpressionSize(exprPtr->arg_list)); 
        }
        
      if (exprPtr->arg_list != NULL) 
        {
         fprintf(ExpressionFP,",\n");
         DumpExpression(exprPtr->arg_list);
        }
      
      exprPtr = exprPtr->next_arg;
      if (exprPtr != NULL) fprintf(ExpressionFP,",\n"); 
     }
  }

/*******************************************/
/* ConstructsToCCommandDefinition:              */
/*******************************************/
globle VOID ConstructsToCCommandDefinition()
  {
   DefineFunction("constructs-to-c", 'v', PTIF ConstructsToCCommand, "ConstructsToCCommand");
  }

/***************************/
/* AddCodeGeneratorItem:   */
/***************************/
globle BOOLEAN AddCodeGeneratorItem(name,p_num,beforeFunction,initFunction,generateFunction)
  char *name;
  int p_num;
  VOID (*beforeFunction)();
  VOID (*initFunction)();
  int (*generateFunction)();
  {
   struct CodeGeneratorItem *newPtr, *currentPtr, *lastPtr = NULL;

   newPtr = get_struct(CodeGeneratorItem);

   newPtr->name = name;
   newPtr->beforeFunction = beforeFunction;
   newPtr->initFunction = initFunction;
   newPtr->generateFunction = generateFunction;
   newPtr->priority = p_num;
   
   if (ListOfCodeGeneratorItems == NULL)
     {
      newPtr->next = NULL;
      ListOfCodeGeneratorItems = newPtr;
      return(1);
     }
     
   currentPtr = ListOfCodeGeneratorItems;
   while ((currentPtr != NULL) ? (p_num < currentPtr->priority) : CLIPS_FALSE)
     {
      lastPtr = currentPtr; 
      currentPtr = currentPtr->next;
     }

   if (lastPtr == NULL)
     { 
      newPtr->next = ListOfCodeGeneratorItems;
      ListOfCodeGeneratorItems = newPtr;
     }
   else
     {
      newPtr->next = currentPtr;
      lastPtr->next = newPtr;
     }
   
   return(1);
  }  
#else

#if ANSI_COMPILER  
   VOID                               ConstructsToCCommand(void);
#else
   VOID                               ConstructsToCCommand();
#endif

      /*====================================*/
      /* Definition for rule compiler stub. */
      /*====================================*/

   VOID ConstructsToCCommand() {}

#endif
