   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 DEFFUNCTION MODULE                  */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if DEFFUNCTION_CONSTRUCT

#define _DEFFNCTN_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#if ANSI_COMPILER
#include <string.h>
#endif

#include "constant.h"
#include "clipsmem.h"
#include "scanner.h"
#include "constrct.h"
#include "spclform.h"
#include "generate.h"
#include "router.h"
#include "utility.h"
#include "deffnctn.h"
#include "facts.h"
#include "parsutil.h"
#include "sysprime.h"

#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif

#if DEFGENERIC_CONSTRUCT
#include "genrcfun.h"
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
#include "conscomp.h"
#endif

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define BEGIN_TRACE ">> "
#define END_TRACE   "<< "

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER 
   static int                     CheckDeffunctionCall(struct dfunc *,int);
   static DATA_OBJECT            *EvaluateDFParameters(EXPRESSION *,int);
   static VOID                    GrabWildargs(DATA_OBJECT *,int);

#if (! BLOAD_ONLY) && (! RUN_TIME)
   static BOOLEAN                 ParseDeffunction(char *);
   static struct dfunc           *AddDeffunction(SYMBOL_HN *,EXPRESSION *,int,int,int);
   static EXPRESSION             *ParseParameters(char *,SYMBOL_HN **,int *);
   static int                     FindParameter(SYMBOL_HN *,EXPRESSION *);
   static BOOLEAN                 ReplaceParameters(EXPRESSION *,EXPRESSION *,
                                                    SYMBOL_HN *);
   static int                     ClearDeffunctions(void);
   static VOID                    RemoveDeffunction(struct dfunc *);
   static VOID                    SaveDeffunctionHeaders(char *);
   static VOID                    SaveDeffunctions(char *);
#endif   

#if DEBUGGING_FUNCTIONS
   static VOID                    TraceDeffunction(char *);
#endif

#if BLOAD_AND_BSAVE
   static VOID                    BsaveDFFind(void);
   static VOID                    BsaveDFExpressions(FILE *);
   static VOID                    BsaveDeffunctions(FILE *);
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static int                     DFBloadReady(void);
   static VOID                    BloadDeffunctions(void);
   static VOID                    RefreshDeffunctions(void);
   static VOID                    ClearDFBload(void);
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeffunctionsToCode(char *,int,FILE *,int,int);
   static VOID                    InitDeffunctionsCode(FILE *,int,int);
   static VOID                    ReadyDeffunctionsForCode(void);
#endif

#else

   static int                     CheckDeffunctionCall();
   static DATA_OBJECT            *EvaluateDFParameters();
   static VOID                    GrabWildargs();
   
#if (! BLOAD_ONLY) && (! RUN_TIME)
   static BOOLEAN                 ParseDeffunction();
   static struct dfunc           *AddDeffunction();
   static EXPRESSION             *ParseParameters();
   static int                     FindParameter();
   static BOOLEAN                 ReplaceParameters();
   static int                     ClearDeffunctions();
   static VOID                    RemoveDeffunction();
   static VOID                    SaveDeffunctionHeaders();
   static VOID                    SaveDeffunctions();
#endif

#if DEBUGGING_FUNCTIONS
   static VOID                    TraceDeffunction();
#endif

#if BLOAD_AND_BSAVE
   static VOID                    BsaveDFFind();
   static VOID                    BsaveDFExpressions();
   static VOID                    BsaveDeffunctions();
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   static int                     DFBloadReady();
   static VOID                    BloadDeffunctions();
   static VOID                    RefreshDeffunctions();
   static VOID                    ClearDFBload();
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeffunctionsToCode();
   static VOID                    InitDeffunctionsCode();
   static VOID                    ReadyDeffunctionsForCode();
#endif

#endif

/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
#if BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY
   globle struct dfunc HUGE_ADDR *deffunctionArray = NULL;
#endif

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */

#if (! BLOAD_ONLY) && (! RUN_TIME)  
   static struct token       DFInputToken;
   static int                DeffunctionError;
#endif
   static struct dfunc      *DFList = NULL,*DFBot = NULL;
   static struct dataObject *DFParamArray = NULL;
   static int                DFParamSize = -1;  
   static SYMBOL_HN         *CurrentDeffunctionName = NULL;
   
#if DEBUGGING_FUNCTIONS
   static int                WatchDeffunctions = OFF;
#endif

#if BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY
   static long               DFCount = 0L;
#endif
   
/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/******************************/
/* InitializeDeffunctions:    */
/******************************/
globle VOID InitializeDeffunctions()
  {
#if ! RUN_TIME

#if ! BLOAD_ONLY
   AddConstruct("deffunction",ParseDeffunction);
   AddClearFunction("deffunctions",(VOID (*)(VOID_ARG)) ClearDeffunctions,-1000);
   AddSaveFunction("deffunction-headers",PTIF SaveDeffunctionHeaders,1000);
   AddSaveFunction("deffunctions",PTIF SaveDeffunctions,0);
   DefineFunction("undeffunction",'v',PTIF CmdUndeffunction,"CmdUndeffunction");
#endif   
   
#if DEBUGGING_FUNCTIONS
   DefineFunction("list-deffunctions",'v',PTIF CmdListDeffunctions,"CmdListDeffunctions");
   DefineFunction("ppdeffunction",'v',PTIF CmdPPDeffunction,"CmdPPDeffunction");
#endif
   DefineFunction("(df-runknown)",'u',PTIF DFRtnUnknown,"DFRtnUnknown");
   DefineFunction("(df-getbind)",'u',PTIF DeffunctionGetBind,"DeffunctionGetBind");
   DefineFunction("(df-wildargs)",'u',PTIF DFWildargs,"DFWildargs");
   
#if BLOAD_AND_BSAVE || BLOAD
   AddBeforeBloadFunction("deffunctions",(VOID (*)(VOID_ARG)) ClearDeffunctions,-1000);
#endif
#if BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY
   AddBloadReadyFunction("deffunctions",DFBloadReady,0);
   AddClearBloadReadyFunction("deffunctions",DFBloadReady,0);
#endif
#if BLOAD_AND_BSAVE
   AddBinaryItem("deffunctions",0,BsaveDFFind,BsaveDFExpressions,BsaveDeffunctions,
                                  BloadDeffunctions,RefreshDeffunctions,ClearDFBload);
#endif
#if BLOAD || BLOAD_ONLY
   AddBinaryItem("deffunctions",0,NULL,NULL,NULL,BloadDeffunctions,
                                  RefreshDeffunctions,ClearDFBload);
#endif

#if CONSTRUCT_COMPILER
   AddCodeGeneratorItem("deffunctions",0,ReadyDeffunctionsForCode,
                        InitDeffunctionsCode,DeffunctionsToCode);
#endif

#endif

#if DEBUGGING_FUNCTIONS
   AddWatchItem("deffunctions",&WatchDeffunctions,32);
#endif

  }

/*************************************************************/
/* FindDeffunction:  Searches for a deffunction in the list  */
/*  of deffunctions. Returns a pointer to the deffunction if */
/*  found, otherwise NULL.                                   */
/*************************************************************/
globle VOID *FindDeffunction(df_name)
  char *df_name;
  {
   struct dfunc *df_ptr;

   df_ptr = DFList;
   while (df_ptr != NULL)
     {
      if (strcmp(ValueToString(df_ptr->name),df_name) == 0)
        { return((VOID *) df_ptr); }
      
      df_ptr = df_ptr->next; 
     }

   return(NULL);
  }
  
/*********************************************************************/
/* FindDeffunctionBySymbol:  Searches for a deffunction in the list  */
/*  of deffunctions. Returns a pointer to the deffunction if         */
/*  found, otherwise NULL.                                           */
/*********************************************************************/
globle VOID *FindDeffunctionBySymbol(df_sym)
  SYMBOL_HN *df_sym;
  {
   struct dfunc *df_ptr;

   df_ptr = DFList;
   while (df_ptr != NULL)
     {
      if (df_ptr->name == df_sym)
        { return((VOID *) df_ptr); }
      
      df_ptr = df_ptr->next; 
     }

   return(NULL);
  }
  
/***************************************************************/
/* CallDeffunction:                                            */
/***************************************************************/
globle VOID CallDeffunction(dptr,args,result)
  struct dfunc *dptr;
  EXPRESSION *args;
  DATA_OBJECT *result;
  {
   int oldce,itmp,oldStackSize,svReturnContext,svBreakContext;
   SYMBOL_HN *previousDeffunctionName;
   DATA_OBJECT *oldParamArray,*oldBindList;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   EvaluationError = FALSE;
   if (HaltExecution)
     return;
   itmp = CountArguments(args);
   if (CheckDeffunctionCall(dptr,itmp) == FALSE)
     {
      SetEvaluationError(TRUE);
      return;
     }
   oldce = ExecutingConstruct();
   SetExecutingConstruct(TRUE);
   previousDeffunctionName = CurrentDeffunctionName;
   CurrentDeffunctionName = dptr->name;
   CurrentEvaluationDepth++;
   oldParamArray = DFParamArray;
   DFParamArray = EvaluateDFParameters(args,itmp);
   if (EvaluationError)
     {
      DFParamArray = oldParamArray;
      CurrentDeffunctionName = previousDeffunctionName;
      CurrentEvaluationDepth--;
      PeriodicCleanup(FALSE,TRUE);
      SetExecutingConstruct(oldce);
      return;
     }
   oldStackSize = DFParamSize;
   DFParamSize = itmp;
   oldBindList = BindList;
   BindList = NULL;
   dptr->executing++;
   
   svBreakContext = BreakContext;
   BreakContext = CLIPS_FALSE;
   svReturnContext = ReturnContext;
   ReturnContext = CLIPS_TRUE;
#if DEBUGGING_FUNCTIONS
   if (WatchDeffunctions)
     {
      TraceDeffunction(BEGIN_TRACE);
      EvaluateExpression(dptr->code,result);     
      TraceDeffunction(END_TRACE);
     }
   else
#endif
     EvaluateExpression(dptr->code,result);
   BreakContext = svBreakContext;
   ReturnContext = svReturnContext;
   ReturnFlag = CLIPS_FALSE;
   
   dptr->executing--;
   if (DFParamArray != NULL)
     rm((VOID *) DFParamArray,(int) (sizeof(DATA_OBJECT) * DFParamSize));
   DFParamArray = oldParamArray;
   DFParamSize = oldStackSize;
   if (BindList != NULL)
     ReturnValues(BindList);
   BindList = oldBindList;
   CurrentDeffunctionName = previousDeffunctionName;
   CurrentEvaluationDepth--;
   PropogateReturnValue(result);
   PeriodicCleanup(FALSE,TRUE);
   SetExecutingConstruct(oldce);
  }

globle int DeleteDeffunction(vptr)
  VOID *vptr;
  {
   struct dfunc *dptr;
   
   dptr = (struct dfunc *) vptr;
#if BLOAD_ONLY || RUN_TIME
   PrintCLIPS(WERROR,"Unable to delete deffunction ");
   PrintCLIPS(WERROR,(dptr != NULL) ? ValueToString(dptr->name) : "*");
   PrintCLIPS(WERROR,".\n");
   return(0);
#else

#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) 
     {
      PrintCLIPS(WERROR,"Unable to delete deffunction ");
      PrintCLIPS(WERROR,(dptr != NULL) ? ValueToString(dptr->name) : "*");
      PrintCLIPS(WERROR,".\n");
      return(0);
     }
#endif
   if (dptr == NULL)
      return(ClearDeffunctions());
   if ((dptr->busy > 0) || (dptr->executing > 0))
     {
      PrintCLIPS(WERROR,"Unable to delete deffunction ");
      PrintCLIPS(WERROR,ValueToString(dptr->name));
      PrintCLIPS(WERROR,".\n");
      return(0);
     }
   RemoveDeffunction(dptr);
   return(1);
#endif
  }

globle char *GetDeffunctionName(ptr)
  VOID *ptr;
  {
   return(ValueToString(((struct dfunc *) ptr)->name));
  }
  
globle VOID *GetNextDeffunction(ptr)
  VOID *ptr;
  {
   if (ptr == NULL)
     return((VOID *) DFList);
   return((VOID *) ((struct dfunc *) ptr)->next);
  }
  
globle int IsDeffunctionDeletable(ptr)
  VOID *ptr;
  {
#if BLOAD_ONLY || RUN_TIME
   return(FALSE);
#else
   struct dfunc *dptr;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     return(FALSE);
#endif
   dptr = (struct dfunc *) ptr;
   return(((dptr->busy == 0) && (dptr->executing == 0)) ? TRUE : FALSE);
#endif
  }

/*********************************************/
/* CmdUndeffunction: removes a deffunction. */
/*   Syntax: (undeffunction <name> | *)      */
/*********************************************/
globle VOID CmdUndeffunction()
  {
   DATA_OBJECT ptr;
   struct dfunc *dptr;

   if (ArgCountCheck("undeffunction",EXACTLY,1) == -1)
     return;
   if (ArgTypeCheck("undeffunction",1,SYMBOL,&ptr) == FALSE)
     return;
   dptr = (struct dfunc *) FindDeffunctionBySymbol((SYMBOL_HN *) ptr.value);
   if ((dptr == NULL) ? (strcmp(DOToString(ptr),"*") != 0) : FALSE)
     {
      PrintCLIPS(WERROR,"Unable to find deffunction ");
      PrintCLIPS(WERROR,DOToString(ptr));
      PrintCLIPS(WERROR," in function undeffunction.\n");
      return;
     }
   DeleteDeffunction((VOID *) dptr);
  }

/**********************************************************************/
/* DeffunctionGetBind:  Performs bind operation within a deffunction. */
/**********************************************************************/
globle VOID DeffunctionGetBind(bind_result)
  DATA_OBJECT_PTR bind_result;
  {
   EXPRESSION *expr_ptr;
   int index;
   
   /*============================================*/
   /* Get the name of the variable being sought. */
   /*============================================*/

   expr_ptr = GetFirstArgument();

   /*===============================================*/
   /* Search for the variable in the list of binds. */
   /*===============================================*/

   if (GetBoundVariable(bind_result,(SYMBOL_HN *) expr_ptr->value))
     return;

   index = ValueToInteger(expr_ptr->next_arg->value);
   if (index == 0)
     {    
      SetEvaluationError(TRUE);
      PrintCLIPS(WERROR,"Variable ");
      PrintCLIPS(WERROR,ValueToString(expr_ptr->value));
      PrintCLIPS(WERROR," unbound in deffunction ");
      PrintCLIPS(WERROR,ValueToString(CurrentDeffunctionName)); 
      PrintCLIPS(WERROR,".\n");
      bind_result->type = SYMBOL;
      bind_result->value = (VOID *) CLIPSFalseSymbol;
      return;
     }

   /*=============================================================*/
   /* Variable was bound in the argument list of the deffunction. */
   /*=============================================================*/
   if (index > 0)
     CopyMemory(DATA_OBJECT,1,bind_result,&DFParamArray[index - 1]);
   else
     GrabWildargs(bind_result,-index);
  }

/*****************************************************************/
/* DFRtnUnknown: Used by interpretive function calls.  */
/*****************************************************************/
globle VOID DFRtnUnknown(rvalue)
  DATA_OBJECT_PTR rvalue;
  {
   CopyMemory(DATA_OBJECT,1,rvalue,&DFParamArray[DOPToInteger(GetFirstArgument()) - 1]);
  }
  
/**********************************************************************
 DFWildargs: Copies from i..DFParamSize of DFParamArray into
              multi-field variable - expands multi-fields on stack
 **********************************************************************/  
globle VOID DFWildargs(result)
  DATA_OBJECT *result;
  {
   GrabWildargs(result,DOPToInteger(GetFirstArgument()));
  }
  
#if DEBUGGING_FUNCTIONS

/***********************************************************/
/* CmdListDeffunctions: displays the list of deffunctions. */
/*   Syntax: (list-deffunctions)                           */
/***********************************************************/
globle VOID CmdListDeffunctions()
  {
   if (ArgCountCheck("list-deffunctions",EXACTLY,0) == -1) return;
   ListDeffunctions();
  }
  
/**************************************************/
/* CmdPPDeffunction: pretty prints a deffunction. */
/*   Syntax: (ppdeffunction <deffunction name>)   */
/**************************************************/
globle VOID CmdPPDeffunction()
  {
   char *df_name;
   struct dfunc *df_ptr;

   df_name = GetConstructName(1,1,"ppdeffunction","deffunction name");
   if (df_name == NULL) return;
   df_ptr = (struct dfunc *) FindDeffunction(df_name);
   if (df_ptr == NULL)        
     {
      PrintCLIPS(WERROR,"Unable to find deffunction ");
      PrintCLIPS(WERROR,df_name);
      PrintCLIPS(WERROR," in function ppdeffunction.\n");
     }
   else if (df_ptr->ppForm != NULL)
     PrintCLIPS(WDISPLAY,df_ptr->ppForm);
  }

globle char *GetDeffunctionPPForm(ptr)
  VOID *ptr;
  {
   return(((struct dfunc *) ptr)->ppForm);
  }

globle VOID ListDeffunctions()
  {
   struct dfunc *dfuncPtr;
   long count = 0;

   dfuncPtr = DFList;
   while (dfuncPtr != NULL)
     {         
      if (GetHaltExecution() == TRUE) return;
      PrintCLIPS(WDISPLAY,ValueToString(dfuncPtr->name));
      PrintCLIPS(WDISPLAY,"\n");
      dfuncPtr = dfuncPtr->next;
      count++;
     }
   PrintTally(WDISPLAY,count,"deffunction","deffunctions");
  }

#endif

#if RUN_TIME

/***************************/
/* SetListOfDeffunctions:  */
/***************************/
globle VOID SetListOfDeffunctions(head)
  VOID *head;
  {
   struct dfunc *dhead;
   
   dhead = (struct dfunc *) head;
   DFList = dhead;
   if (dhead == NULL)
     DFBot = NULL;
   else
     {
      while (dhead->next != NULL)
        dhead = dhead->next;
      DFBot = dhead;
     }
  }

#endif
  
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
static int CheckDeffunctionCall(dptr,args)
  struct dfunc *dptr;
  int args;
  {
   if (args < dptr->minNumberOfParameters)
     {
      if (dptr->maxNumberOfParameters == -1)
        ExpectedCountError(ValueToString(dptr->name),AT_LEAST,dptr->minNumberOfParameters);
      else
        ExpectedCountError(ValueToString(dptr->name),EXACTLY,dptr->minNumberOfParameters);
      return(FALSE);
     }
   else if ((args > dptr->minNumberOfParameters) && 
            (dptr->maxNumberOfParameters != -1))
     {
      ExpectedCountError(ValueToString(dptr->name),EXACTLY,dptr->minNumberOfParameters);
      return(FALSE);
     }
   return(TRUE);
  }


/****************************************/
/* EvaluateDFParameters:                */
/****************************************/
static DATA_OBJECT *EvaluateDFParameters(parameterList,numberOfParameters)
  EXPRESSION *parameterList;
  int numberOfParameters;
  {
   DATA_OBJECT *rva;
   int i = 0;
   
   if (numberOfParameters == 0) return(NULL);
   
   rva = (DATA_OBJECT *) gm2(sizeof(DATA_OBJECT) * numberOfParameters);
   
   while (parameterList != NULL)
     {
      if ((EvaluateExpression(parameterList,&rva[i]) == TRUE) ? TRUE :
          (rva[i].type == RVOID))
        {
         if (rva[i].type == RVOID)
           {
            PrintCLIPS(WERROR,"Functions without a return value are illegal as deffunction arguments.\n");
            SetEvaluationError(TRUE);
           }
         PrintCLIPS(WERROR,"\nThis error occurred while evaluating arguments ");
         PrintCLIPS(WERROR,"for the deffunction ");
         PrintCLIPS(WERROR,ValueToString(CurrentDeffunctionName));
         PrintCLIPS(WERROR,".\n");
         rm(rva,sizeof(DATA_OBJECT) * numberOfParameters);
         return(NULL);
        }
      parameterList = parameterList->next_arg;
      if (parameterList != NULL)
        { rva[i].next = &rva[i+1]; }
      else
        { rva[i].next = NULL; }
      i++;
     }
     
   return(rva);
  }

static VOID GrabWildargs(result,index)
  DATA_OBJECT *result;
  int index;
  {
   register int i,j,k;
   int size;
   struct fact *fptr;

   result->type = MULTIFIELD;
   result->begin = 0;
   result->end = -1;
   if (DFParamSize-index < 0)
     {
      result->value = (VOID *) CreateMultifield(0);
      return;
     }
   for (i = index-1 , size = 0 ; i < DFParamSize ; i++)
     {
      if (DFParamArray[i].type != MULTIFIELD)
        size++;
      else
        size += DFParamArray[i].end - DFParamArray[i].begin + 1;
     }
   result->end = size-1;
   result->value = (VOID *) CreateMultifield(size);
   fptr = (struct fact *) result->value;
   for (i = index-1 , j = 0 ; i < DFParamSize ; i++)
     {
      if (DFParamArray[i].type != MULTIFIELD)
        {
         fptr->atoms[j].type = DFParamArray[i].type; 
         fptr->atoms[j].value = DFParamArray[i].value; 
         j++;
        }
      else
        {
         for (k = DFParamArray[i].begin ;
              k <= DFParamArray[i].end ; k++ , j++)
           {
            fptr->atoms[j].type = 
                     ((struct fact *) DFParamArray[i].value)->atoms[k].type;
            fptr->atoms[j].value = 
                     ((struct fact *) DFParamArray[i].value)->atoms[k].value;
           }         
        }
     }
  }

#if (! BLOAD_ONLY) && (! RUN_TIME)

/***************************************************************/
/* ParseDeffunction:                                           */
/***************************************************************/
static BOOLEAN ParseDeffunction(readSource)
  char *readSource;
  {
   SYMBOL_HN *deffunctionName;
   EXPRESSION *actions;
   EXPRESSION *parameterList;
   SYMBOL_HN *wildcard;
   int min,max;
   struct dfunc *dptr;

   DeffunctionError = FALSE;
   SetPPBufferStatus(ON);

   FlushPPBuffer();          
   SetIndentDepth(3);    
   SavePPBuffer("(deffunction ");

#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded() == TRUE) 
     {
      PrintCLIPS(WERROR,"\nCannot load deffunctions with binary load in effect.\n");
      return(TRUE);
     }
#endif

   /*=======================================================*/
   /* Parse the name and comment fields of the deffunction. */
   /*=======================================================*/  

   deffunctionName = GetConstructNameAndComment(readSource,&DFInputToken,"deffunction",
                                                FindDeffunction,NULL,"!",TRUE);
   if (deffunctionName == NULL) { return(TRUE); } 
   if (ValidConstruct(ValueToString(deffunctionName)) == CLIPS_TRUE)
     {
      PrintCLIPS(WERROR,"Deffunctions are not allowed to replace constructs.\n");
      return(TRUE);
     }
   if (FindFunction(ValueToString(deffunctionName)) != NULL)
     {
      PrintCLIPS(WERROR,"Deffunctions are not allowed to replace external functions.\n");
      return(TRUE);
     }
   
#if DEFGENERIC_CONSTRUCT
   if (FindDefgenericBySymbol(deffunctionName) != NULL)
     {
      PrintCLIPS(WERROR,"Deffunctions are not allowed to replace generic functions.\n");
      return(TRUE);
     }
#endif

   /*==========================*/
   /* Parse the argument list. */
   /*==========================*/
   
   parameterList = ParseParameters(readSource,&wildcard,&DeffunctionError);
   if (DeffunctionError == TRUE) { return(DeffunctionError); } 

   /*===================================================================*/
   /* Go ahead and add the deffunction so it can be recursively called. */
   /*===================================================================*/
   
   min = CountArguments(parameterList);
   if (wildcard != NULL)
     {
      min--;
      max = -1;
     }
   else
     max = min;
   dptr = AddDeffunction(deffunctionName,NULL,min,max,TRUE);
   if (dptr == NULL)
     {
      ReturnExpression(parameterList);
      return(TRUE);
     }
     
   /*==================================================*/
   /* Parse the actions contained within the function. */
   /*==================================================*/
   
   SavePPBuffer("\n   ");
   SetIndentDepth(3);

   actions = GroupActions(readSource,&DFInputToken,TRUE,NULL);

   if (actions == NULL) 
     {
      ReturnExpression(parameterList);
      if (dptr->busy == 0)
        RemoveDeffunction(dptr);
      return(TRUE);
     }
   
   /*=============================*/
   /* Reformat the closing token. */
   /*=============================*/
   
   PPBackup();
   PPBackup();
   SavePPBuffer(DFInputToken.print_rep);
   SavePPBuffer("\n");
   
   if (actions->arg_list == NULL)
     {
      ReturnExpression(actions);
      ReturnExpression(parameterList);
#if DEBUGGING_FUNCTIONS
      if (GetConserveMemory() == TRUE)
        dptr->ppForm = NULL;
      else
        dptr->ppForm = CopyPPBuffer();
#else
      dptr->ppForm = NULL;
#endif
      return(FALSE); 
     }
   actions = CompactActions(actions);
   
   /*========================================================*/
   /* Replace the variables with appropriate function calls. */
   /*========================================================*/

   DeffunctionError = ReplaceParameters(actions,parameterList,wildcard);
   if (DeffunctionError) 
     { 
      ReturnExpression(actions);
      ReturnExpression(parameterList);
      if (dptr->busy == 0)
        RemoveDeffunction(dptr);
      return(DeffunctionError); 
     }
     
   /*======================*/
   /* Add the deffunction. */
   /*======================*/
   
   AddDeffunction(deffunctionName,actions,min,max,FALSE);

   ReturnExpression(parameterList);
   
   return(DeffunctionError);       
  }

/*************************************************************/
/* AddDeffunction:                                  */
/*************************************************************/
static struct dfunc *AddDeffunction(name,actions,min,max,headerp)
  SYMBOL_HN *name;
  EXPRESSION *actions;
  int min,max,headerp;
  {
   struct dfunc *dfuncPtr;
   int oldbusy;
   
   /*===============================================================*/
   /* If the deffunction doesn't exist, create a new structure to   */
   /* contain it and add it to the List of deffunctions. Otherwise, */
   /* use the existing structure and remove the pretty print form   */
   /* and interpretive code.                                        */
   /*===============================================================*/
   dfuncPtr = (struct dfunc *) FindDeffunctionBySymbol(name);
   if (dfuncPtr == NULL)
     {
      dfuncPtr = get_struct(dfunc);
      dfuncPtr->name = name;
      dfuncPtr->code = NULL;
      dfuncPtr->ppForm = NULL;
      dfuncPtr->minNumberOfParameters = min;
      dfuncPtr->maxNumberOfParameters = max;
      IncrementSymbolCount(name);
      dfuncPtr->busy = 0;
      dfuncPtr->executing = 0;
     }
   else if (dfuncPtr->executing == 0)
     {
      dfuncPtr->minNumberOfParameters = min;
      dfuncPtr->maxNumberOfParameters = max;
      oldbusy = dfuncPtr->busy;
      ExpressionDeinstall(dfuncPtr->code);
      dfuncPtr->busy = oldbusy;
      ReturnPackedExpression(dfuncPtr->code);
      dfuncPtr->code = NULL; 
      if (dfuncPtr->ppForm != NULL)
        {
         rm(dfuncPtr->ppForm,
                     (int) sizeof(char) * ((int) strlen(dfuncPtr->ppForm) + 1));
         dfuncPtr->ppForm = NULL;
        }
      if (dfuncPtr->prev == NULL)
        DFList = dfuncPtr->next;
      else
        dfuncPtr->prev->next = dfuncPtr->next;
      if (dfuncPtr->next == NULL)
        DFBot = dfuncPtr->prev;
      else
        dfuncPtr->next->prev = dfuncPtr->prev;
     }
   else
     {
      PrintCLIPS(WERROR,"Deffunction ");
      PrintCLIPS(WERROR,ValueToString(name));
      PrintCLIPS(WERROR," may not be redefined while it is executing.\n");
      return(NULL);
     }
         
   dfuncPtr->next = NULL;
   dfuncPtr->prev = DFBot;
   if (DFList == NULL)
     DFList = dfuncPtr;
    else
      DFBot->next = dfuncPtr;
    DFBot = dfuncPtr;
      
   /*====================================*/
   /* Install the new interpretive code. */
   /*====================================*/
   
   if (actions != NULL)
     {
      oldbusy = dfuncPtr->busy;
      ExpressionInstall(actions);
      dfuncPtr->busy = oldbusy;
      dfuncPtr->code = PackExpression(actions);
      ReturnExpression(actions);
     }

   /*=================================================================*/
   /* Install the pretty print form if memory is not being conserved. */
   /*=================================================================*/
   
#if DEBUGGING_FUNCTIONS
   if ((GetConserveMemory() == TRUE) || (headerp == TRUE))
     dfuncPtr->ppForm = NULL;
   else
     dfuncPtr->ppForm = CopyPPBuffer();
#else
   dfuncPtr->ppForm = NULL;
#endif
   return(dfuncPtr);
  }

/*****************************************************************/
/* ParseParameters:  Parses the parameter list of a deffunction. */
/*****************************************************************/ 
static EXPRESSION *ParseParameters(readSource,wildcard,error)
  char *readSource;
  SYMBOL_HN **wildcard;
  int *error;
  {
   EXPRESSION *parameterList = NULL, *nextOne, *lastOne = NULL,*check;
   
   *wildcard = NULL;
   if (DFInputToken.type != LPAREN)
     {
      PrintCLIPS(WERROR,"Expected '(' to open parameter list.\n");
      *error = TRUE;
      return(NULL);
     }    
     
   GetToken(readSource,&DFInputToken);
   
   while ((DFInputToken.type == BWORD) || (DFInputToken.type == BWORDS))
     {
      for (check = parameterList ; check != NULL ; check = check->next_arg)
        if (check->value == DFInputToken.value)
         {
          PrintCLIPS(WERROR,"Duplicate parameter names not allowed.\n");
          ReturnExpression(parameterList);
          *error = TRUE;
          return(NULL);
         }
      if (*wildcard != NULL)
        {
         PrintCLIPS(WERROR,"No parameters allowed after wildcard parameter.\n");
         ReturnExpression(parameterList);
         *error = TRUE;
         return(NULL);
        }
        
      nextOne = GenConstant(DFInputToken.type,DFInputToken.value);
      
      if (DFInputToken.type == BWORDS)
        *wildcard = (SYMBOL_HN *) DFInputToken.value;
      if (lastOne == NULL)
        { parameterList = nextOne; }
      else
        { lastOne->next_arg = nextOne; }
      lastOne = nextOne;
      SavePPBuffer(" ");
      GetToken(readSource,&DFInputToken);
     }  
   
   if (DFInputToken.type != RPAREN)
     {
      PrintCLIPS(WERROR,"Expected ')' to close parameter list.\n");
      *error = TRUE;
      ReturnExpression(parameterList);
      return(NULL);
     }
     
   if (parameterList != NULL)
     {
      PPBackup();
      PPBackup();
      SavePPBuffer(")");
     }
   
   return(parameterList);
  }
  
/*****************************************/
/* FindParameter:                        */
/*****************************************/
static int FindParameter(name,parameterList)
  SYMBOL_HN *name;
  EXPRESSION *parameterList;
  {
   int i = 1;
   
   while (parameterList != NULL)
     {
      if (parameterList->value == (VOID *) name) return(i);
      i++;
      parameterList = parameterList->next_arg;
     }
     
   return(0);
  }
  
/*****************************************/
/* ReplaceParameters:                    */
/*****************************************/
static BOOLEAN ReplaceParameters(actions,parameterList,wildcard)
  EXPRESSION *actions;
  EXPRESSION *parameterList;
  SYMBOL_HN *wildcard;
  {
   int position;
   BOOLEAN boundAlso;
   EXPRESSION *arg_lvl;
   SYMBOL_HN *bindName;
   
   while (actions != NULL)
     {
      if ((actions->type == BWORD) || (actions->type == BWORDS))
        {
         /*===============================================*/
         /* See if the variable is in the parameter list. */
         /*===============================================*/
          
         bindName = (SYMBOL_HN *) actions->value;
         position = FindParameter(bindName,parameterList);
           
         /*=================================================================*/
         /* Check to see if the variable is bound within the deffunction.   */
         /*=================================================================*/
   
         boundAlso = SearchParsedBindNames(bindName);
         
         /*=============================================*/
         /* If variable is not defined in the parameter */
         /* list or as part of a bind action then...    */
         /*=============================================*/
         
         if ((position == 0) && (boundAlso == FALSE))
           {
            PrintCLIPS(WERROR,"\nUndefined variable ");
            PrintCLIPS(WERROR,ValueToString(bindName));        
            PrintCLIPS(WERROR," referenced in deffunction.\n");
            return(TRUE);
           }
           
         /*===================================================*/
         /* Else if variable is defined in the parameter list */
         /* and not rebound within the deffunction then...    */
         /*===================================================*/
         
         else if ((position > 0) && (boundAlso == FALSE))
           {
            actions->type = FCALL;
            if (bindName != wildcard)
              { actions->value = (VOID *) FindFunction("(df-runknown)"); }
            else
              { actions->value = (VOID *) FindFunction("(df-wildargs)"); }
            actions->arg_list = GenConstant(INTEGER,AddLong((long) position));
           }
           
         /*===========================================================*/
         /* Else the variable is rebound within the deffunction so... */
         /*===========================================================*/
         
         else 
           {
            actions->type = FCALL;
            actions->value = (VOID *) FindFunction("(df-getbind)");
            
            actions->arg_list = GenConstant(SYMBOL,bindName);
            arg_lvl = actions->arg_list;

            if (position == 0)
              { arg_lvl->next_arg = GenConstant(INTEGER,AddLong(0L)); }
            else if (bindName != wildcard)
              { arg_lvl->next_arg = GenConstant(INTEGER,AddLong((long) position)); }
            else
              { arg_lvl->next_arg = GenConstant(INTEGER,AddLong((long) -position)); }
           }
        }
#if DEFGLOBAL_CONSTRUCT
      else if (actions->type == GBWORD)
        { if (ReplaceGlobalVariable(actions) == FALSE) return(TRUE); }
#endif
        
      else if (actions->arg_list != NULL)
        {
         if (ReplaceParameters(actions->arg_list,parameterList,wildcard) == TRUE)
           return(TRUE);
        }
           
      actions = actions->next_arg;
     }
     
   return(FALSE);
  }
       
/*************************************************************/
/* ClearDeffunctions:                                  */
/*************************************************************/
static int ClearDeffunctions()
  {
   struct dfunc *dptr,*dtmp;
   int oldbusy,success = 1;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(0);
#endif
   
   dptr = DFList;
   while (dptr != NULL)
     {
      if (dptr->executing > 0)
        {
         PrintCLIPS(WERROR,"Unable to delete deffunction ");
         PrintCLIPS(WERROR,ValueToString(dptr->name));
         PrintCLIPS(WERROR,".\n");
         success = 0;
        }
      else
        {
         oldbusy = dptr->busy;
         ExpressionDeinstall(dptr->code);
         dptr->busy = oldbusy;
         ReturnPackedExpression(dptr->code);
         dptr->code = NULL;
        }
      dptr = dptr->next;
     }
     
   dptr = DFList;
   while (dptr != NULL)
     {
      dtmp = dptr;
      dptr = dptr->next;
      if (dtmp->executing == 0)
        {
         if (dtmp->busy > 0)
           {
            PrintCLIPS(WERROR,"WARNING: Deffunction ");
            PrintCLIPS(WERROR,ValueToString(dtmp->name));
            PrintCLIPS(WERROR," only partially deleted due to usage by other constructs.\n");
            success = 0;
           }
         else
           RemoveDeffunction(dtmp);
        }
     }
   return(success);
  }

/******************************/
/* RemoveDeffunction:         */
/******************************/
static VOID RemoveDeffunction(dptr)
  struct dfunc *dptr;
  {
   if (GetCompilationsWatch())
     {
      PrintCLIPS(WDIALOG,"Removing deffunction ");
      PrintCLIPS(WDIALOG,ValueToString(dptr->name));
      PrintCLIPS(WDIALOG,".\n");
     }
   DecrementSymbolCount(dptr->name);
   ExpressionDeinstall(dptr->code);
   ReturnPackedExpression(dptr->code);
   if (dptr->ppForm != NULL)
     rm(dptr->ppForm,(int) sizeof(char) * ((int) strlen(dptr->ppForm) + 1));
   if (dptr->prev == NULL)
     DFList = dptr->next;
   else
     dptr->prev->next = dptr->next;
   if (dptr->next == NULL)
     DFBot = dptr->prev;
   else
     dptr->next->prev = dptr->prev;
   rtn_struct(dfunc,dptr);
  }

/***********************************/
/* SaveDeffunctionHeaders:         */
/***********************************/
static VOID SaveDeffunctionHeaders(log_name)
  char *log_name;
  {
   struct dfunc *def_ptr;
   register int i;
   
   def_ptr = DFList;

   while (def_ptr != NULL)
     {
      if (def_ptr->ppForm != NULL)
        {
         PrintCLIPS(log_name,"(deffunction ");
         PrintCLIPS(log_name,ValueToString(def_ptr->name));
         PrintCLIPS(log_name," (");
         for (i = 0 ; i < def_ptr->minNumberOfParameters ; i++)
           {
            PrintCLIPS(log_name,"?p");
            PrintLongInteger(log_name,(long) i);
            if (i != def_ptr->minNumberOfParameters-1)
              PrintCLIPS(log_name," ");
           }
         if (def_ptr->maxNumberOfParameters == -1)
           {
            if (def_ptr->minNumberOfParameters != 0)
              PrintCLIPS(log_name," ");
            PrintCLIPS(log_name,"$?wildargs))\n\n");
           }
         else
           PrintCLIPS(log_name,"))\n\n");
        }
      def_ptr = def_ptr->next;
     }
  }

/*****************************/
/* SaveDeffunctions:         */
/*****************************/
static VOID SaveDeffunctions(log_name)
  char *log_name;
  {
   struct dfunc *def_ptr;
   
   def_ptr = DFList;
   while (def_ptr != NULL)
     {
      if (def_ptr->ppForm != NULL)
        {
         PrintInChunks(log_name,def_ptr->ppForm);
         PrintCLIPS(log_name,"\n");
        }
      def_ptr = def_ptr->next;
     }
  }

#endif 

#if DEBUGGING_FUNCTIONS

/********************************************************************
  TraceDeffunction
 ********************************************************************/
static VOID TraceDeffunction(tstring)
  char *tstring;
  {
   register int i;

   PrintCLIPS(WTRACE,"DFN ");
   PrintCLIPS(WTRACE,tstring);
   PrintCLIPS(WTRACE,ValueToString(CurrentDeffunctionName));
   PrintCLIPS(WTRACE," ED:");
   PrintLongInteger(WTRACE,(long) CurrentEvaluationDepth);
   PrintCLIPS(WTRACE," (");
   for (i = 0 ; i < DFParamSize ; i++)
     {
      PrintDataObject(WTRACE,&DFParamArray[i]);
      if (i != DFParamSize-1)
        PrintCLIPS(WTRACE," ");
     }
   PrintCLIPS(WTRACE,")\n");
  }

#endif

/* ======================================
   **************************************
        BINARY LOAD AND SAVE FUNCTIONS
   ====================================== 
   **************************************/
#if BLOAD_AND_BSAVE

static VOID BsaveDFFind()
  {
   struct dfunc *dptr;
   
   DFCount = 0L;
   for (dptr = DFList ; dptr != NULL ; dptr = dptr->next)
     {
      dptr->bsaveIndex = DFCount++;
      dptr->name->neededSymbol = CLIPS_TRUE;
      ExpressionCount += ExpressionSize(dptr->code);
      MarkNeededItems(dptr->code);
     }
  }
  
static VOID BsaveDFExpressions(fp)
  FILE *fp;
  {
   struct dfunc *dptr;
   
   for (dptr = DFList ; dptr != NULL ; dptr = dptr->next)
     BsaveExpression(dptr->code,fp);
  }
  
static VOID BsaveDeffunctions(fp)
  FILE *fp;
  {
   unsigned long space;
   struct dfunc *dptr;
   struct dfunc dummy_df;
   
   if (DFCount == 0L)
     {
      space = (unsigned long) DFCount;
      GenWrite((VOID *) &space,(unsigned long) sizeof(unsigned long),fp);
      return;
     }
   space = (unsigned long) (sizeof(long) + (sizeof(struct dfunc) * DFCount));
   GenWrite((VOID *) &space,(unsigned long) sizeof(unsigned long),fp);
   GenWrite((VOID *) &DFCount,(unsigned long) sizeof(long),fp);
   for (dptr = DFList ; dptr != NULL ; dptr = dptr->next)
     {
      CopyMemory(struct dfunc,1,&dummy_df,dptr);
      dummy_df.busy = 0;
      dummy_df.executing = 0;
      dummy_df.name = (SYMBOL_HN *) dptr->name->bucket;
      if (dptr->code != NULL)
        {
         dummy_df.code = (EXPRESSION *) ExpressionCount;
         ExpressionCount += ExpressionSize(dptr->code);
        }
      else
        dummy_df.code = (EXPRESSION *) -1;
      if (dptr->prev != NULL)
        dummy_df.prev = (struct dfunc *) dptr->prev->bsaveIndex;
      else
        dummy_df.prev = (struct dfunc *) -1;
      if (dptr->next != NULL)
        dummy_df.next = (struct dfunc *) dptr->next->bsaveIndex;
      else
        dummy_df.next = (struct dfunc *) -1;
      GenWrite((VOID *) &dummy_df,(unsigned long) sizeof(struct dfunc),fp);
     }
   DFCount = 0L;
  }
  
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE

static int DFBloadReady()
  {
   struct dfunc *dptr;
   
   for (dptr = DFList ; dptr != NULL ; dptr = dptr->next)
     if ((dptr->busy > 0) || (dptr->executing > 0))
       return(FALSE);
   return(TRUE);
  }

static VOID BloadDeffunctions()
  {
   unsigned long space;
   
   GenRead((VOID *) &space,(unsigned long) sizeof(unsigned long));
   if (space == 0L)
     return;
   GenRead((VOID *) &DFCount,(unsigned long) sizeof(long));
   if (DFCount == 0L)
     return;
   space -= sizeof(long);
   deffunctionArray = (struct dfunc HUGE_ADDR *) genlongalloc(space);
   GenRead((VOID *) deffunctionArray,space); 
  }
  
static VOID RefreshDeffunctions()
  {
   register long i;
   
   if (DFCount == 0L)
     return;
   for (i = 0L ; i < DFCount ; i++)
     {
      deffunctionArray[i].name = SymbolPointer(deffunctionArray[i].name);
      IncrementCount(deffunctionArray[i].name);
      deffunctionArray[i].ppForm = NULL;
      deffunctionArray[i].code = ExpressionPointer(deffunctionArray[i].code);
      deffunctionArray[i].prev = DFPointer(deffunctionArray[i].prev);
      deffunctionArray[i].next = DFPointer(deffunctionArray[i].next);
     }
   DFList = (struct dfunc *) &deffunctionArray[0];
   DFBot = (struct dfunc *) &deffunctionArray[DFCount-1];
  }
  
static VOID ClearDFBload()
  {
   register long i;
   unsigned long space;
      
   DFList = NULL;
   DFBot = NULL;
   for (i = 0L ; i < DFCount ; i++)
     DecrementSymbolCount(deffunctionArray[i].name);
   space = (unsigned long) (sizeof(struct dfunc) * DFCount);
   if (space == 0L)
     return;
   genlongfree((VOID *) deffunctionArray,space);
   deffunctionArray = NULL;
   DFCount = 0L;
  }
  
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)

/*******************************************************************/
/* DeffunctionsToCode:  Produces all the deffunctions definitions. */
/*******************************************************************/
static int DeffunctionsToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   struct dfunc *dfPtr;
   FILE *fp;
   int version = 1;
   int i = 1;
   int newHeader = CLIPS_TRUE;
   
   fprintf(headerFP,"#include \"deffnctn.h\"\n");
   if (DFList == NULL)
     return(-1);
   if ((fp = NewCFile(fileName,fileID,version)) == NULL)
     return(0);
   dfPtr = DFList;
   while (dfPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct dfunc dfn%d_%d[] = {\n",imageID,version);
         fprintf(headerFP,"extern struct dfunc dfn%d_%d[];\n",
                 imageID,version);
         newHeader = CLIPS_FALSE;
        }
     
      fprintf(fp,"{0,0,"); 
      PrintSymbolReference(fp,dfPtr->name);
      fprintf(fp,",NULL,");
      ExpressionToCode(fp,dfPtr->code);
      fprintf(fp,",%d,%d,0L,",dfPtr->minNumberOfParameters,
                              dfPtr->maxNumberOfParameters);
      PrintDeffunctionReference(fp,dfPtr->prev);
      fprintf(fp,",");
      PrintDeffunctionReference(fp,dfPtr->next);
      i++;
      dfPtr = dfPtr->next;
      if ((i > maxIndices) || (dfPtr == NULL))
        {
         fprintf(fp,"}\n};\n");
         fclose(fp);
         i = 1;
         version++;
         if (dfPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL)
              return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        fprintf(fp,"},\n");
     }
   return(1);
  }
  
/**************************/
/* InitDeffunctionsCode:  */
/**************************/
static VOID InitDeffunctionsCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   if (DFList != NULL)
     fprintf(initFP,"   SetListOfDeffunctions(dfn%d_1);\n",imageID);
   else
     fprintf(initFP,"   SetListOfDeffunctions(NULL);\n");
  }

/******************************/
/* ReadyDeffunctionsForCode:  */
/******************************/
static VOID ReadyDeffunctionsForCode()
  {
   struct dfunc *dfPtr;
   register long i;
   
   for (dfPtr = DFList , i = 0L ; dfPtr != NULL ; 
        dfPtr = dfPtr->next , i++)
     dfPtr->bsaveIndex = i;
  }
  
#endif

#endif

    
