   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                  DEFGLOBAL MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
#define _DEFGLOBL_SOURCE_

#include "setup.h"

#if DEFGLOBAL_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "expressn.h"
#include "evaluatn.h"
#include "constrct.h"
#include "router.h"
#include "utility.h"
#include "facts.h"
#include "generate.h"
#include "defglobl.h"
   
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#endif

#if CONSTRUCT_COMPILER
#include "conscomp.h"
#endif

#define ARRAY_INCREMENT 20
  
#if ANSI_COMPILER  
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static BOOLEAN                 ParseDefglobal(char *);
   static BOOLEAN                 GetVariableDefinition(char *,int *);
   static VOID                    AddDefglobal(SYMBOL_HN *,DATA_OBJECT_PTR,struct expr *);
   static VOID                    SaveDefglobals(char *);
#endif   
   VOID                           PpdefglobalCommand(void);
   VOID                           ListDefglobalsCommand(void);
   int                            SetResetGlobalsCommand(void);
   int                            GetResetGlobalsCommand(void);
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind(void);
   static VOID                    BsaveExpressions(FILE *);
   static VOID                    BsaveDefglobals(FILE *);
#endif
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static VOID                    BloadDefglobals(void);
   static VOID                    ClearBload(void);
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DefglobalsToCode(char *,int,FILE *,int,int);
   static VOID                    InitDefglobalsCode(FILE *,int,int);
#endif
#else
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static BOOLEAN                 ParseDefglobal();
   static BOOLEAN                 GetVariableDefinition();
   static VOID                    AddDefglobal();
   static VOID                    SaveDefglobals();
#endif
   VOID                           PpdefglobalCommand();
   VOID                           ListDefglobalsCommand();
   int                            SetResetGlobalsCommand();
   int                            GetResetGlobalsCommand();
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind();
   static VOID                    BsaveExpressions();
   static VOID                    BsaveDefglobals();
#endif
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static VOID                    BloadDefglobals();
   static VOID                    ClearBload();
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DefglobalsToCode();
   static VOID                    InitDefglobalsCode();
#endif
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle int                   ChangeToGlobals = CLIPS_FALSE; 
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
  
   static struct defglobal     *DefglobalArray = NULL;
   static int                   SizeOfDefglobalArray = 0;
   static int                   NumberOfDefglobals = 0;
   static struct defglobalPtr  *ListOfDefglobals = NULL;
   static BOOLEAN               ResetGlobals = CLIPS_TRUE;
   
#if DEBUGGING_FUNCTIONS
   static BOOLEAN               WatchGlobals = OFF;
#endif
   
#if (BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY)
   static struct defglobal HUGE_ADDR     *BDefglobalArray;
   static struct defglobalPtr HUGE_ADDR  *BDefglobalPointersArray;
#endif

/***************************/
/* InitializeDefglobal:    */
/***************************/
globle VOID InitializeDefglobal()
  {
#if (! BLOAD_ONLY) && (! RUN_TIME)
   AddConstruct("defglobal",ParseDefglobal);
   AddSaveFunction("defglobal", PTIF SaveDefglobals,40);
   AddClearFunction("defglobal",ClearDefglobals,0);
#endif   

#if (BLOAD_AND_BSAVE || BLOAD) 
   AddBeforeBloadFunction("defglobal",ClearDefglobals,0);
   AddAfterBloadFunction("defglobal",ResetDefglobals,50);
#endif

#if BLOAD_AND_BSAVE 
   AddBinaryItem("defglobal",0,BsaveFind,BsaveExpressions,BsaveDefglobals,
                               BloadDefglobals,NULL,ClearBload);
#endif

#if (BLOAD || BLOAD_ONLY)
   AddBinaryItem("defglobal",0,NULL,NULL,NULL,BloadDefglobals,NULL,ClearBload);
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
   AddCodeGeneratorItem("defglobal",40,NULL,InitDefglobalsCode,DefglobalsToCode);
#endif

   AddResetFunction("defglobal",ResetDefglobals,50);
   
#if DEBUGGING_FUNCTIONS
#if ! RUN_TIME
   DefineFunction("list-defglobals", 'v', PTIF ListDefglobalsCommand, "ListDefglobalsCommand");
   DefineFunction("ppdefglobal",     'v', PTIF PpdefglobalCommand,   "PpdefglobalCommand"); 
#endif
   AddWatchItem("globals",&WatchGlobals,0);
#endif

#if ! RUN_TIME
   DefineFunction("(rglobal)",'u', PTIF GlobalRtnUnknown,"GlobalRtnUnknown"); 
   DefineFunction("set-reset-globals",'b',
                  SetResetGlobalsCommand,"SetResetGlobalsCommand");
   DefineFunction("get-reset-globals",'b',
                  GetResetGlobalsCommand,"GetResetGlobalsCommand");
#endif
  }

#if (! BLOAD_ONLY) && (! RUN_TIME)
/***************************************************************/
/* ParseDefglobal:                                  */
/***************************************************************/
static BOOLEAN ParseDefglobal(readSource)
  char *readSource;
  {
   int defglobalError;
   
   defglobalError = CLIPS_FALSE;
   SetPPBufferStatus(ON);

   FlushPPBuffer();          
   SetIndentDepth(3);    
   SavePPBuffer("(defglobal\n   ");

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) 
     {
      PrintCLIPS(WERROR,"\nCannot load defglobals with binary load in effect.\n");
      return(CLIPS_TRUE);
     }
#endif
     
   /*======================*/
   /* Parse the variables. */
   /*======================*/
   
   while (GetVariableDefinition(readSource,&defglobalError))
     { PPCRAndIndent(); }
   
   return(defglobalError);       
  }

/*********************************************************************/
/* GetVariableDefinition:                                            */
/*********************************************************************/
static BOOLEAN GetVariableDefinition(readSource,defglobalError)
  char *readSource;
  int *defglobalError;
  {
   struct token InputToken;
   SYMBOL_HN *variableName;
   struct expr *assignPtr;
   DATA_OBJECT assignValue;

   /*========================================*/
   /* Get next token, which should either be */
   /* a closing parenthesis or a variable.   */
   /*========================================*/

   GetToken(readSource,&InputToken);
   if (InputToken.type == RPAREN) return(CLIPS_FALSE);
   
   if (InputToken.type == BWORD)
     {
      PrintCLIPS(WERROR,"\nGlobal variable names must begin and end with *'s\n");
      *defglobalError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     } 
   else if (InputToken.type != GBWORD)
     { 
      SyntaxErrorMessage("defglobal");
      *defglobalError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
   
   variableName = (SYMBOL_HN *) InputToken.value;
     
   SavePPBuffer(" ");
   
   /*================================*/
   /* Print out compilation message. */
   /*================================*/
   
   if ((GetWatchItem("compilations") == ON) && GetPrintWhileLoading())
     { 
      if (QFindDefglobal(variableName) >= 0) PrintCLIPS(WDIALOG,"Redefining defglobal: ?");
      else PrintCLIPS(WDIALOG,"Defining defglobal: "); 
      PrintCLIPS(WDIALOG,ValueToString(variableName));
      PrintCLIPS(WDIALOG,"\n");
     }
   else if (GetPrintWhileLoading())
     { PrintCLIPS(WDIALOG,":"); }
   
   /*==============================*/
   /* The next token must be an =. */
   /*==============================*/
     
   GetToken(readSource,&InputToken);
   if (strcmp(InputToken.print_rep,"=") != 0)
     {
      SyntaxErrorMessage("defglobal");
      *defglobalError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }

   SavePPBuffer(" ");

   /*======================================================*/
   /* Parse the expression to be assigned to the variable. */
   /*======================================================*/

   assignPtr = ParseAtomOrExpression(readSource);
   if (assignPtr == NULL)
     {
      *defglobalError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
     
   /*==========================*/
   /* Evaluate the expression. */
   /*==========================*/
   
   SetEvaluationError(CLIPS_FALSE);
   if (EvaluateExpression(assignPtr,&assignValue))
     {
      ReturnExpression(assignPtr);
      *defglobalError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
     
   /*======================================*/
   /* Add the variable to the global list. */
   /*======================================*/

   AddDefglobal(variableName,&assignValue,assignPtr);
   return(CLIPS_TRUE);
  }
 
/*************************************************************/
/* AddDefglobal:                                  */
/*************************************************************/
static VOID AddDefglobal(name,vPtr,ePtr)
  SYMBOL_HN *name;
  DATA_OBJECT_PTR vPtr;
  struct expr *ePtr;
  {
   struct defglobal *defglobalPtr;
   BOOLEAN new = CLIPS_FALSE;
   int i;
   struct defglobalPtr *newDefglobal = NULL, *dgPtr, *lastPtr, *endPtr;
   
   /*===================================================================*/
   /* If the defglobal is already defined, used the existing structure. */
   /* Otherwise, use an empty slot in the Defglobal Array. If no such   */
   /* slot exists, then expand the array and use the added slot.        */
   /*===================================================================*/
   
   defglobalPtr = (struct defglobal *) FindDefglobal(ValueToString(name));
   
   if (defglobalPtr == NULL)
     { 
      if (NumberOfDefglobals < SizeOfDefglobalArray)
        { defglobalPtr = &DefglobalArray[NumberOfDefglobals]; }
      else
        {
         DefglobalArray = (struct defglobal *) 
                          genrealloc(DefglobalArray,
                                     SizeOfDefglobalArray * sizeof(struct defglobal),
                                     (SizeOfDefglobalArray + ARRAY_INCREMENT) * sizeof(struct defglobal));
         for (i = SizeOfDefglobalArray ; i < SizeOfDefglobalArray + ARRAY_INCREMENT ; i++)
           { DefglobalArray[i].name = NULL; }
         defglobalPtr = &DefglobalArray[SizeOfDefglobalArray];
         SizeOfDefglobalArray += ARRAY_INCREMENT;
        }
      new = CLIPS_TRUE;
      newDefglobal = get_struct(defglobalPtr);
      newDefglobal->next = NULL;
      newDefglobal->location = NumberOfDefglobals;
      NumberOfDefglobals++;
     }
      
   /*===========================================*/
   /* Remove the old values from the defglobal. */
   /*===========================================*/
   
   if (new == CLIPS_FALSE)
     {
      ValueDeinstall(&defglobalPtr->current);
      if (defglobalPtr->current.type == MULTIFIELD) 
        { ReturnElements(defglobalPtr->current.value); }
     
      ExpressionDeinstall(defglobalPtr->initial);
      ReturnExpression(defglobalPtr->initial);
     }
   
   /*=======================================*/
   /* Copy the new values to the defglobal. */
   /*=======================================*/
   
   defglobalPtr->current.type = vPtr->type;
   if (vPtr->type != MULTIFIELD) defglobalPtr->current.value = vPtr->value; 
   else DuplicateSegment(&defglobalPtr->current,vPtr);
   ValueInstall(&defglobalPtr->current);

   defglobalPtr->initial= ePtr;
   ExpressionInstall(ePtr);
   ChangeToGlobals = CLIPS_TRUE;
   
   /*=============================================*/
   /* If the defglobal was redefined, we're done. */
   /*=============================================*/
   
   if (new == CLIPS_FALSE) return;
   
   /*===================================*/
   /* Copy the defglobal variable name. */
   /*===================================*/
   
   defglobalPtr->name = name;
   IncrementSymbolCount(name);
   
   /*=============================================*/
   /* Add the defglobal to the list of defglobals */
   /*=============================================*/
   
   dgPtr = ListOfDefglobals;
   lastPtr = NULL;
   endPtr = dgPtr;
   
   while (dgPtr != NULL)
     {
      if (&DefglobalArray[dgPtr->location] == defglobalPtr)
        {
         if (lastPtr == NULL) ListOfDefglobals = dgPtr->next;
         else lastPtr->next = dgPtr->next;
         
         newDefglobal = dgPtr;
         dgPtr = dgPtr->next;
         newDefglobal->next = NULL;
         if (dgPtr == NULL) endPtr = lastPtr;
         else endPtr = dgPtr;
        }
      else
        {
         lastPtr = dgPtr;
         dgPtr = dgPtr->next;
         if (dgPtr == NULL) endPtr = lastPtr;
         else endPtr = dgPtr;
        }
     }
     
   if (endPtr == NULL) ListOfDefglobals = newDefglobal;
   else endPtr->next = newDefglobal;
   
  }
#endif

/**********************************************************/
/* FindDefglobal:  Searches for a defglobal in the array  */
/*  of defglobals. Returns a pointer to the defglobal if  */
/*  found, otherwise NULL.                                */
/**********************************************************/
globle VOID *FindDefglobal(defglobalName)
  char *defglobalName;
  {
   int i = 0;

   while (i < SizeOfDefglobalArray)
     {
      if (DefglobalArray[i].name == NULL)
        { i++; }
      else if (strcmp(ValueToString(DefglobalArray[i].name),defglobalName) == 0)
        { return((VOID *) &DefglobalArray[i]); }
      else
        { i++; }
     }

   return(NULL);
  }

/************************************************************/
/* GetNextDefglobal:                            */
/************************************************************/
globle VOID *GetNextDefglobal(gPtr)
  VOID *gPtr;
  {
   if (gPtr == NULL)
     { return((VOID *) ListOfDefglobals); }
   else
     { return((VOID *) (((struct defglobalPtr *) gPtr)->next)); }
  }

/***************************************************/
/* GetDefglobalValueForm:                          */
/***************************************************/
globle VOID GetDefglobalValueForm(buffer,bufferLength,theGlobal)
  char *buffer;
  int bufferLength;
  VOID *theGlobal;
  {
   struct defglobal *defglobalPtr;
   
   defglobalPtr = &DefglobalArray[((struct defglobalPtr *) theGlobal)->location];
   
   OpenStringDestination("GlobalValueForm",buffer,bufferLength);
   PrintCLIPS("GlobalValueForm","?");
   PrintCLIPS("GlobalValueForm",ValueToString(defglobalPtr->name));
   PrintCLIPS("GlobalValueForm"," = ");
   PrintDataObject("GlobalValueForm",&defglobalPtr->current);
   CloseStringDestination("GlobalValueForm");
  }
  
/***************************************************/
/* GetDefglobalPPForm:                          */
/***************************************************/
globle VOID GetDefglobalPPForm(buffer,bufferLength,theGlobal)
  char *buffer;
  int bufferLength;
  VOID *theGlobal;
  {
   struct defglobal *defglobalPtr;
   
   defglobalPtr = &DefglobalArray[((struct defglobalPtr *) theGlobal)->location];
   
   OpenStringDestination("GlobalValueForm",buffer,bufferLength);
   PrintCLIPS("GlobalValueForm","?");
   PrintCLIPS("GlobalValueForm",ValueToString(defglobalPtr->name));
   PrintCLIPS("GlobalValueForm"," = ");
   PrintExpression("GlobalValueForm",defglobalPtr->initial);
   CloseStringDestination("GlobalValueForm");
  }
  
/***************************************************/
/* GetDefglobalName:                          */
/***************************************************/
globle char *GetDefglobalName(theGlobal)
  VOID *theGlobal;
  {
   return(ValueToString(DefglobalArray[((struct defglobalPtr *) theGlobal)->location].name));
  }

/************************************************************/
/* GetActualDefglobal:                                     */
/************************************************************/
globle VOID *GetActualDefglobal(gPtr)
  VOID *gPtr;
  {
   return((VOID *) &DefglobalArray[((struct defglobalPtr *) gPtr)->location]);
  }

/************************************************************/
/* GetIndexedDefglobal:                                     */
/************************************************************/
globle VOID *GetIndexedDefglobal(index)
  int index;
  {
   return((VOID *) &DefglobalArray[index]);
  }
  
/**********************************************************/
/* QFindDefglobal:  Searches for a defglobal in the array */
/*  of defglobals. Returns the index to the defglobal if  */
/*  found, otherwise -1.                                  */
/**********************************************************/
globle int QFindDefglobal(defglobalName)
  SYMBOL_HN *defglobalName;
  {
   int i = 0;

   while (i < SizeOfDefglobalArray)
     {
      if (DefglobalArray[i].name == NULL)
        { i++; }
      else if (DefglobalArray[i].name == defglobalName)
        { return(i); }
      else
        { i++; }
     }

   return(-1);
  }
  
/**********************************************************/
/* GetDefglobalValue:                                 */
/**********************************************************/
globle BOOLEAN GetDefglobalValue(variableName,vPtr)
  char *variableName;
  DATA_OBJECT_PTR vPtr;
  {
   struct defglobal *dgPtr;
   
   if ((dgPtr = (struct defglobal *) FindDefglobal(variableName)) != NULL)
     { 
      vPtr->type = dgPtr->current.type;
      vPtr->value = dgPtr->current.value;
      vPtr->begin = dgPtr->current.begin;
      vPtr->end = dgPtr->current.end;
      if (vPtr->type == MULTIFIELD)
        { 
         vPtr->value = CreateMultifield(vPtr->end + 1);
         CopyMemory(struct element,vPtr->end + 1,
                                   &((SEGMENT) vPtr->value)->atoms[0],
                                   &((SEGMENT) dgPtr->current.value)->atoms[dgPtr->current.begin]);
        }
      return(CLIPS_TRUE);
     }
     
   return(CLIPS_FALSE);
  }

/***************************************************************/
/* QGetDefglobalValue: Returns the value of a global variable. */
/*   This function uses an index to reference the location of  */
/*   the global variable within the defglobal array.           */
/***************************************************************/
globle VOID QGetDefglobalValue(position,vPtr)
  int position;
  DATA_OBJECT_PTR vPtr;
  {
   struct defglobal *dgPtr;
   
   dgPtr = &DefglobalArray[position];
   
   vPtr->type = dgPtr->current.type;
   vPtr->value = dgPtr->current.value;
   vPtr->begin = dgPtr->current.begin;
   vPtr->end = dgPtr->current.end;
   
   /*===========================================================*/
   /* If the global contains a multifield value, return a copy  */
   /* of the value so that routines which use this value are    */
   /* not affected if the value of the global is later changed. */ 
   /*===========================================================*/
           
   if (vPtr->type == MULTIFIELD)
     { 
      vPtr->value = CreateMultifield(vPtr->end + 1);
      CopyMemory(struct element,vPtr->end + 1,
                                &((SEGMENT) vPtr->value)->atoms[0],
                                &((SEGMENT) dgPtr->current.value)->atoms[dgPtr->current.begin]);
     }
  }

/**********************************************************/
/* SetDefglobalValue:                                 */
/**********************************************************/
globle BOOLEAN SetDefglobalValue(variableName,vPtr)
  char *variableName;
  DATA_OBJECT_PTR vPtr;
  {
   struct defglobal *dgPtr;
   
   if ((dgPtr = (struct defglobal *) FindDefglobal(variableName)) != NULL)
     {   
#if DEBUGGING_FUNCTIONS 
      if (WatchGlobals == ON)
        { 
         PrintCLIPS(WTRACE,":== ?");
         PrintCLIPS(WTRACE,variableName);
         PrintCLIPS(WTRACE," ==> ");
         PrintDataObject(WTRACE,vPtr);
         PrintCLIPS(WTRACE," <== ");
         PrintDataObject(WTRACE,&dgPtr->current);
         PrintCLIPS(WTRACE,"\n");
        }
#endif
        
      ValueDeinstall(&dgPtr->current);
      if (dgPtr->current.type == MULTIFIELD) 
        { ReturnElements(dgPtr->current.value); }
         
      dgPtr->current.type = vPtr->type;
      if (vPtr->type != MULTIFIELD) dgPtr->current.value = vPtr->value; 
      else DuplicateSegment(&dgPtr->current,vPtr);
      ValueInstall(&dgPtr->current);
      
      ChangeToGlobals = CLIPS_TRUE;
      return(CLIPS_TRUE);
     }
     
   return(CLIPS_FALSE);
  }
  
/**********************************************************/
/* QSetDefglobalValue:                                    */
/**********************************************************/
globle VOID QSetDefglobalValue(position,vPtr)
  int position;
  DATA_OBJECT_PTR vPtr;
  {
   struct defglobal *dgPtr;
   
   dgPtr = &DefglobalArray[position];

#if DEBUGGING_FUNCTIONS   
   if (WatchGlobals == ON)
     { 
      PrintCLIPS(WTRACE,":== ?");
      PrintCLIPS(WTRACE,ValueToString(dgPtr->name));
      PrintCLIPS(WTRACE," ==> ");
      PrintDataObject(WTRACE,vPtr);
      PrintCLIPS(WTRACE," <== ");
      PrintDataObject(WTRACE,&dgPtr->current);
      PrintCLIPS(WTRACE,"\n");
     }
#endif
      
   ValueDeinstall(&dgPtr->current);
   if (dgPtr->current.type == MULTIFIELD) 
     { ReturnElements(dgPtr->current.value); }
         
   dgPtr->current.type = vPtr->type;
   if (vPtr->type != MULTIFIELD) dgPtr->current.value = vPtr->value; 
   else DuplicateSegment(&dgPtr->current,vPtr);
   ValueInstall(&dgPtr->current);
   
   ChangeToGlobals = CLIPS_TRUE;
  }
  
/*************************************************************/
/* GetNumberOfDefglobals:                                  */
/*************************************************************/
globle int GetNumberOfDefglobals()
  {
   return(NumberOfDefglobals);
  }

/************************************************************/
/* ResetDefglobals: Handles (reset) command for defglobals. */
/************************************************************/
globle VOID ResetDefglobals()
  {
   struct defglobal *defglobalPtr;
   DATA_OBJECT assignValue;
   int i = 0;
     
   if (ResetGlobals == CLIPS_FALSE) return;
   
   while (i < NumberOfDefglobals)
     {
      if (GetHaltExecution() == CLIPS_TRUE) return;
      
      defglobalPtr = &DefglobalArray[i];
     
      if (EvaluateExpression(defglobalPtr->initial,&assignValue))
        {
         assignValue.type = SYMBOL;
         assignValue.value = (VOID *) CLIPSFalseSymbol;
        }
     
      QSetDefglobalValue(i,&assignValue);
      i++;
     }
  }

#if (! RUN_TIME) && (! BLOAD_ONLY)
/*****************************/
/* SaveDefglobals:          */
/*****************************/
static VOID SaveDefglobals(logicalName)
  char *logicalName;
  {
   struct defglobal *defglobalPtr;
   struct defglobalPtr *dgPtr;

   if (ListOfDefglobals != NULL) PrintCLIPS(logicalName,"(defglobal");
   
   dgPtr = ListOfDefglobals;
   while (dgPtr != NULL)
     {
      defglobalPtr = &DefglobalArray[dgPtr->location];
      PrintCLIPS(logicalName,"\n   ");
      PrintCLIPS(logicalName,"?");
      PrintCLIPS(logicalName,ValueToString(defglobalPtr->name));
      PrintCLIPS(logicalName," = ");
      PrintExpression(logicalName,defglobalPtr->initial);
      
      dgPtr = dgPtr->next;
     }
     
   if (ListOfDefglobals != NULL) PrintCLIPS(logicalName,")\n");
  }
  
/*************************************************************/
/* ClearDefglobals:                                  */
/*************************************************************/
globle VOID ClearDefglobals()
  {
   struct defglobal *defglobalPtr;
   struct defglobalPtr *dgPtr;
   int i = 0;

#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return;
#endif
   
   while (ListOfDefglobals != NULL)
     {
      dgPtr = ListOfDefglobals->next;
      rtn_struct(defglobalPtr,ListOfDefglobals);
      ListOfDefglobals = dgPtr;
     }
     
   while (i < NumberOfDefglobals)
     {
      defglobalPtr = &DefglobalArray[i];
      ValueDeinstall(&defglobalPtr->current);
      if (defglobalPtr->current.type == MULTIFIELD) 
        { ReturnElements(defglobalPtr->current.value); }
        
      ExpressionDeinstall(defglobalPtr->initial);
      ReturnExpression(defglobalPtr->initial);
      
      DecrementSymbolCount(defglobalPtr->name);
      i++;
     }
  
   if (DefglobalArray != NULL)
     {
      rm(DefglobalArray,SizeOfDefglobalArray * sizeof(struct defglobal));
      DefglobalArray = NULL;
     }
     
   SizeOfDefglobalArray = 0;
   NumberOfDefglobals = 0;
   ChangeToGlobals = CLIPS_TRUE;
  }
#endif
  
/***************************************************/
/* GlobalRtnUnknown: Used by interpretive function */
/*   calls to access the value of a global.        */
/***************************************************/
globle VOID GlobalRtnUnknown(rvalue)
  DATA_OBJECT_PTR rvalue;
  {
   int index;
   
   index = DOPToInteger(GetFirstArgument());
   if (index >= NumberOfDefglobals)
     {
      rvalue->type = SYMBOL;
      rvalue->value = (VOID *) CLIPSFalseSymbol;
      return;
     }
   QGetDefglobalValue(index,rvalue);
  }
  
/*****************************************************************/
/* ReplaceGlobalVariable:                                          */
/*****************************************************************/
globle BOOLEAN ReplaceGlobalVariable(ePtr)
  struct expr *ePtr;
  {
   int pos;
   
   pos = QFindDefglobal((SYMBOL_HN *) ePtr->value);
   
   if (pos < 0) 
     {
      PrintCLIPS(WERROR,"\nGlobal variable ?");
      PrintCLIPS(WERROR,ValueToString(ePtr->value));
      PrintCLIPS(WERROR," was referenced, but is not defined\n");  
      return(CLIPS_FALSE);
     }
     
   ePtr->type = FCALL;
   ePtr->value = (VOID *) FindFunction("(rglobal)");
   ePtr->arg_list = GenConstant(INTEGER,AddLong((long int) pos));
   
   return(CLIPS_TRUE);
  }

/************************************************************/
/* SetListOfDefglobals:                            */
/************************************************************/
globle VOID SetListOfDefglobals(listPtr,arrayPtr,count)
  VOID *listPtr, *arrayPtr;
  int count;
  {
   ListOfDefglobals = (struct defglobalPtr *) listPtr; 
   DefglobalArray = (struct defglobal *) arrayPtr; 
   NumberOfDefglobals = count;
   SizeOfDefglobalArray = count;
  }

#if DEBUGGING_FUNCTIONS
/***********************************************************/
/* ListDefglobalsCommand: displays the list of defglobals. */
/*   Syntax: (list-defglobals)                             */
/***********************************************************/
globle VOID ListDefglobalsCommand()
  {  
   if (ArgCountCheck("list-defglobals",EXACTLY,0) == -1) return;
   ListDefglobals();
  }
  
/********************************************************/
/* ListDefglobals: Displays the list of defglobals. */
/********************************************************/
globle VOID ListDefglobals()
  {
   struct defglobal *defglobalPtr;
   struct defglobalPtr *dgPtr;

   dgPtr = ListOfDefglobals;
   while (dgPtr != NULL)
     {
      if (GetHaltExecution() == CLIPS_TRUE) return;
      
      defglobalPtr = &DefglobalArray[dgPtr->location];
      
      PrintCLIPS(WDISPLAY,"?");
      PrintCLIPS(WDISPLAY,ValueToString(defglobalPtr->name));
      PrintCLIPS(WDISPLAY," = ");
      PrintDataObject(WDISPLAY,&defglobalPtr->current);
      PrintCLIPS(WDISPLAY,"\n");
      
      dgPtr = dgPtr->next;
     }
   
   PrintTally(WDISPLAY,(long) NumberOfDefglobals,"defglobal","defglobals");
  }
  
/**************************************************/
/* PpdefglobalCommand: Pretty prints a defglobal. */
/*   Syntax: (ppdefglobal <defglobal-name>)       */
/**************************************************/
globle VOID PpdefglobalCommand()
  {
   char *defglobalName;
   VOID *defglobalPtr;

   defglobalName = GetConstructName(1,1,"ppdefglobal","defglobal name");
   if (defglobalName == NULL) return;   

   defglobalPtr = FindDefglobal(defglobalName);
   if (defglobalPtr == NULL) 
     { 
      PrintCLIPS(WERROR,"Unable to find defglobal ");
      PrintCLIPS(WERROR,defglobalName);
      PrintCLIPS(WERROR,"\n");
      return;
     }
     
   PrintCLIPS(WDISPLAY,"(defglobal ?");
   PrintCLIPS(WDISPLAY,defglobalName);
   PrintCLIPS(WDISPLAY," = ");
   PrintExpression(WDISPLAY,((struct defglobal *) defglobalPtr)->initial);
   PrintCLIPS(WDISPLAY,")\n");

   return;
  }
  
#endif

/************************************************************/
/* SetResetGlobalsCommand:                        */
/************************************************************/
globle int SetResetGlobalsCommand()
  {
   int oldValue;
   DATA_OBJECT arg_ptr;
   
   oldValue = GetResetGlobals();
   
   if (ArgCountCheck("set-reset-globals",EXACTLY,1) == -1) 
     { return(oldValue); }
     
   RtnUnknown(1,&arg_ptr);
     
   if ((arg_ptr.value == (VOID *) CLIPSFalseSymbol) && (arg_ptr.type == SYMBOL))
     { SetResetGlobals(CLIPS_FALSE); }
   else
     { SetResetGlobals(CLIPS_TRUE); }
     
   return(oldValue);
  }
  
/************************************************************/
/* GetResetGlobalsCommand:                        */
/************************************************************/
globle int GetResetGlobalsCommand()
  {
   int oldValue;
   
   oldValue = GetResetGlobals();
   
   if (ArgCountCheck("get-reset-globals",EXACTLY,0) == -1) 
     { return(oldValue); }
     
   return(oldValue);
  }
  
/***************************************************/
/* GetResetGlobals:                              */
/***************************************************/
globle BOOLEAN GetResetGlobals()
  { return(ResetGlobals); }
  
/***************************************************/
/* SetResetGlobals:                              */
/***************************************************/
globle BOOLEAN SetResetGlobals(value)
  int value;
  { 
   int ov;
   
   ov = ResetGlobals;
   ResetGlobals = value;
   return(ov);
  }
  
/***************************************************/
/* GetGlobalsChanged:                              */
/***************************************************/
globle int GetGlobalsChanged()
  { return(ChangeToGlobals); }
  
/***************************************************/
/* SetGlobalsChanged:                              */
/***************************************************/
globle VOID SetGlobalsChanged(value)
  int value;
  { 
   ChangeToGlobals = value;
  }

#if BLOAD_AND_BSAVE
/*******************************/
/* BsaveDefglobalFind:         */
/*******************************/
static VOID BsaveFind()
   {
    struct defglobalPtr *dglist;
    struct defglobal *gPtr;
   
    dglist = (struct defglobalPtr *) GetNextDefglobal(NULL);
    while (dglist != NULL)
      {
       gPtr = (struct defglobal *) GetActualDefglobal(dglist);
       gPtr->name->neededSymbol = CLIPS_TRUE;
       ExpressionCount += ExpressionSize(gPtr->initial);
       MarkNeededItems(gPtr->initial);
       dglist = dglist->next;
      }
    }
     
/*******************************/
/* BsaveExpressions:  */
/*******************************/
static VOID BsaveExpressions(fp) 
  FILE *fp; 
  {
   struct defglobalPtr *dglist;
   struct defglobal *gPtr;

   dglist = (struct defglobalPtr *) GetNextDefglobal(NULL);
   while (dglist != NULL)
     {
      gPtr = (struct defglobal *) GetActualDefglobal(dglist);
      BsaveExpression(gPtr->initial,fp);
      dglist = dglist->next;
     }
  }
  
/********************************************************/
/* BsaveDefglobals: Writes out all defglobal structures */ 
/*   to the binary file                                 */
/********************************************************/
static VOID BsaveDefglobals(fp)
  FILE *fp;
  {
   unsigned long int count = 0, space = 0,i;
   struct defglobal newGlobal, *tempGPtr;
   struct defglobalPtr newGlobalPtr, *tempGPPtr;
   
   /*=====================================*/
   /* Write out the number of defglobals. */
   /*=====================================*/
  
   count = GetNumberOfDefglobals();
   space = count * (sizeof(struct defglobal) + sizeof(struct defglobalPtr))
           + sizeof(unsigned long);
   GenWrite(&space,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
   if (count == 0) return;
     
   /*=====================================*/
   /* Write out each defglobal structure. */
   /*=====================================*/
   
   for (i = 0 ; i < count ; i++)
     {
      tempGPtr = (struct defglobal *) GetIndexedDefglobal((int) i);
      newGlobal.name = (SYMBOL_HN *) tempGPtr->name->bucket;
      newGlobal.current.type = RVOID;
      
      if (tempGPtr->initial != NULL)
        { 
         newGlobal.initial = (struct expr *) ExpressionCount;
         ExpressionCount += ExpressionSize(tempGPtr->initial); 
        }
      else
        { newGlobal.initial = (struct expr *) -1L; }
        
      GenWrite(&newGlobal,(unsigned long) sizeof(struct defglobal),fp);
     }
      
   /*=============================================*/
   /* Write out each defglobal pointer structure. */
   /*=============================================*/
   
   i = 0;
   tempGPPtr = (struct defglobalPtr *) GetNextDefglobal(NULL);
   while (tempGPPtr != NULL)
     {
      newGlobalPtr.location = tempGPPtr->location;
        
      i++;
      if (tempGPPtr->next != NULL)
        { newGlobalPtr.next = (struct defglobalPtr *) i; }
      else
        { newGlobalPtr.next = (struct defglobalPtr *) -1L; }
        
      GenWrite(&newGlobalPtr,(unsigned long) sizeof(struct defglobalPtr),fp);
      tempGPPtr = (struct defglobalPtr *) GetNextDefglobal(tempGPPtr); 
     }
  }
#endif

#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
/********************************************************************/
/* BloadDefglobals: Loads the defglobals used by this binary image. */
/********************************************************************/
static VOID BloadDefglobals()
  {
   unsigned long int space, i;
   long int index;
   unsigned long int defglobalCount;
   
   /*===============================*/
   /* Get the number of defglobals. */
   /*===============================*/
  
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   GenRead(&defglobalCount,(unsigned long) sizeof(unsigned long int));
   if (defglobalCount == 0)
     {
      DefglobalArray = NULL;
      return;
     }
  
   space = defglobalCount * sizeof(struct defglobal);
   BDefglobalArray = (struct defglobal *) genlongalloc(space);
   GenRead((VOID *) BDefglobalArray,space);
   
   space = defglobalCount * sizeof(struct defglobalPtr);
   BDefglobalPointersArray = (struct defglobalPtr *) genlongalloc(space);
   GenRead((VOID *) BDefglobalPointersArray,space);
           
   /*=============================*/
   /* Reset defglobal structures. */
   /*=============================*/
 
   for (i = 0; i < defglobalCount; i++)
     {
      BDefglobalArray[i].name = SymbolPointer(BDefglobalArray[i].name);
      IncrementCount(BDefglobalArray[i].name);
      BDefglobalArray[i].initial = (struct expr *)
                         ExpressionPointer(BDefglobalArray[i].initial);
     }
   
   for (i = 0; i < defglobalCount; i++)
     {
      index = (long int) BDefglobalPointersArray[i].next;
      if (index != -1)
        { BDefglobalPointersArray[i].next = 
             (struct defglobalPtr *) &BDefglobalPointersArray[index]; }
      else
        { BDefglobalPointersArray[i].next = NULL; }
     }
     
   /*============================================*/
   /* Set the deffacts to the new deffacts list. */
   /*============================================*/
   
   SetListOfDefglobals((VOID *) BDefglobalPointersArray,
                       (VOID *) BDefglobalArray,(int) defglobalCount);
  }

/****************************************************************/
/* ClearBload:                                                  */
/****************************************************************/
static VOID ClearBload()
  {
   unsigned long int i, space;
   
   for (i = 0; i < NumberOfDefglobals; i++)
     { 
      DecrementSymbolCount(BDefglobalArray[i].name); 
      
      ValueDeinstall(&(BDefglobalArray[i].current));
      if (BDefglobalArray[i].current.type == MULTIFIELD) 
        { ReturnElements(BDefglobalArray[i].current.value); }
     }
   
   space = NumberOfDefglobals * sizeof(struct defglobal);
   if (space != 0) genlongfree((VOID *) BDefglobalArray,space);  
   
   space = NumberOfDefglobals * sizeof(struct defglobalPtr);
   if (space != 0) genlongfree((VOID *) BDefglobalPointersArray,space);
   
   SetListOfDefglobals(NULL,NULL,0);
  }
#endif 
  
#if CONSTRUCT_COMPILER && (! RUN_TIME)

/**************************************************************/
/* InitDefglobalsCode:   */
/**************************************************************/
static VOID InitDefglobalsCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   if (GetNextDefglobal(NULL) != NULL)
     { 
      fprintf(initFP,"   SetListOfDefglobals(dgp%d_1,dgb%d_1,%d);\n",
                 imageID,imageID,GetNumberOfDefglobals()); 
      fprintf(initFP,"   ResetDefglobals();\n"); 
     }
   else
     { fprintf(initFP,"   SetListOfDefglobals(NULL,NULL,0);\n"); }
  }

/**************************************************************/
/* DefglobalsToCode:  Produces all the defglobal definitions. */
/**************************************************************/
static int DefglobalsToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   struct defglobal *dfPtr;
   struct defglobalPtr *dpPtr;
   FILE *fp;
   int version = 1;
   int i = 1;
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   
   /*===========================*/
   /* Dump the defglobal array. */
   /*===========================*/
   
   fprintf(headerFP,"#include \"defglobl.h\"\n");
   if (GetNumberOfDefglobals() == 0) return(-1);
   
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
   fprintf(fp,"#include \"constant.h\"\n"); 
   fprintf(fp,"struct defglobal dgb%d_%d[] = {\n",imageID,version);
   fprintf(headerFP,"extern struct defglobal dgb%d_%d[];\n",imageID,version);
   
   for (i = 1; i <= GetNumberOfDefglobals(); i++)
     {
      if (i != 1) fprintf(fp,",\n"); 
      fprintf(fp,"{"); 
      dfPtr = (struct defglobal *) GetIndexedDefglobal(i-1);
      PrintSymbolReference(fp,dfPtr->name); /* name */
      fprintf(fp,",{NULL,RVOID},");         /* current */
      ExpressionToCode(fp,dfPtr->initial);
      fprintf(fp,"}");
     }
   
   fprintf(fp,"};\n");
   fclose(fp);
   
   /*==================================*/
   /* Dump the defglobal pointer list. */
   /*==================================*/
   
   if ((fp = NewCFile(fileName,fileID,2)) == NULL) return(0);
   fprintf(fp,"struct defglobalPtr dgp%d_1[] = {\n",imageID);
   fprintf(headerFP,"extern struct defglobalPtr dgp%d_1[];\n",imageID);
   i = 1;
   
   dpPtr = (struct defglobalPtr *) GetNextDefglobal(NULL);
   while (dpPtr != NULL)
     {
      fprintf(fp,"{%d,",dpPtr->location);
      if (dpPtr->next != NULL)
        { fprintf(fp,"&dgp%d_1[%d]}",imageID,i); }
      else
        { fprintf(fp,"NULL}"); }
        
      i++;
      dpPtr = (struct defglobalPtr *) GetNextDefglobal(dpPtr);
      if (dpPtr == NULL) fprintf(fp,"};\n");
      else fprintf(fp,",\n");
     }
     
   fclose(fp);
   return(1);
  }
#endif

#endif /* DEFGLOBAL_CONSTRUCT */



