   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                  RULE HEADER FILE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_defrule

#define _H_defrule


struct patptr;
struct defrule;


#include "symbol.h"
#include "expressn.h"
#include "network.h"

struct patptr
  {
   struct patternNode *pptr;
   struct joinNode *jptr;
   struct patptr *next;
  };
  
struct defrule
  {
   struct symbolHashNode *name;
   char *pp_form;  
   int salience;
   unsigned int complexity      : 15;
   unsigned int afterBreakpoint :  1;
   struct expr *dynamicSalience;
   struct expr *actions;
   struct joinNode *logicalJoin;
   struct patptr *pats;
   struct defrule *disjunct;
   struct defrule *next;               
  };

#define RULEPTR VOID *

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _DEFRULE_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
   LOCALE VOID                           InitializeDefrules(void);
   LOCALE VOID                          *FindDefrule(char *);
   LOCALE BOOLEAN                        DeleteDefrule(VOID *);
   LOCALE RULEPTR                        GetNextDefrule(RULEPTR);
   LOCALE char                          *GetDefrulePPForm(RULEPTR);
   LOCALE char                          *GetDefruleName(RULEPTR);
   LOCALE RULEPTR                        GetIndexedDefrule(int);
   LOCALE int                            GetDisjunctIndex(RULEPTR);
   LOCALE BOOLEAN                        ListMatches(VOID *);
   LOCALE BOOLEAN                        RefreshDefrule(VOID *);
   LOCALE BOOLEAN                        GetRulesWatch(void);
   LOCALE int                            EvaluateSalience(RULEPTR);
   LOCALE VOID                           SetListOfDefrules(VOID *);
   LOCALE VOID                           ClearDefrules(void);
   LOCALE BOOLEAN                        SetRuleDeletions(int);
   LOCALE BOOLEAN                        GetRuleDeletions(void);
   LOCALE VOID                           SalienceInformationError(char *);
   LOCALE BOOLEAN                        IsDefruleDeletable(VOID *);
   LOCALE BOOLEAN                        DeleteNamedDefrule(char *);
   LOCALE int                            ReturnDefrule(struct defrule *);
#else
   LOCALE VOID                           InitializeDefrules();
   LOCALE VOID                          *FindDefrule();
   LOCALE BOOLEAN                        DeleteDefrule();
   LOCALE RULEPTR                        GetNextDefrule();
   LOCALE char                          *GetDefrulePPForm();
   LOCALE char                          *GetDefruleName();
   LOCALE RULEPTR                        GetIndexedDefrule();
   LOCALE int                            GetDisjunctIndex();
   LOCALE BOOLEAN                        ListMatches();
   LOCALE BOOLEAN                        RefreshDefrule();
   LOCALE BOOLEAN                        GetRulesWatch();
   LOCALE int                            EvaluateSalience();
   LOCALE VOID                           SetListOfDefrules();
   LOCALE VOID                           ClearDefrules();
   LOCALE BOOLEAN                        SetRuleDeletions();
   LOCALE BOOLEAN                        GetRuleDeletions();
   LOCALE VOID                           SalienceInformationError();
   LOCALE BOOLEAN                        IsDefruleDeletable();
   LOCALE BOOLEAN                        DeleteNamedDefrule();
   LOCALE int                            ReturnDefrule();
#endif

#endif









