   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*               DEFTEMPLATE HEADER FILE               */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_deftempl

#define _H_deftempl

struct templateSlot;
struct deftemplate;
struct slp;
struct sap;
struct dhash;

#include "expressn.h"
#include "facts.h"
#include "reorder.h"
#include "symbol.h"

struct templateSlot
  {
   struct symbolHashNode *slotName;
   unsigned int multiSlot : 1;        
   unsigned int anyAllowed : 1;
   unsigned int symbolsAllowed : 1;
   unsigned int stringsAllowed : 1;
   unsigned int floatsAllowed : 1;
   unsigned int integersAllowed : 1;
   unsigned int instancesAllowed : 1;
   unsigned int addressesAllowed : 1;
   unsigned int anyRestriction : 1;
   unsigned int symbolRestriction : 1;
   unsigned int stringRestriction : 1;
   unsigned int numberRestriction : 1;
   unsigned int floatRestriction : 1;
   unsigned int integerRestriction : 1;
   struct expr *defaultList;
   struct expr *restrictionList;
   struct expr *minValue;
   struct expr *maxValue;    
   struct templateSlot *next;
  };
  
struct deftemplate
  {
   struct symbolHashNode *name;
   char *pp_form; 
   struct templateSlot *slotList;  
   unsigned int hasMultiSlot : 1;
   unsigned int numberOfSlots : 15;          
   struct deftemplate *next; 
  };

struct slp
  {
   struct symbolHashNode *name;
   struct node *expr;
   struct slp *next;
  };
  
struct sap
  {
   struct symbolHashNode *name;
   struct expr *expr;
   struct sap *next;
  };
  
struct dhash 
  {
   struct deftemplate *dptr;
   struct dhash *next;
  };

#define SIZE_DEFTEMPLATE_HASH 131

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _DEFTEMPL_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
   LOCALE VOID                           InitializeDeftemplates(VOID);
   LOCALE BOOLEAN                        DeleteDeftemplate(VOID *);
   LOCALE VOID                           ReturnSlots(struct templateSlot *);
   LOCALE struct templateSlot           *FindSlot(struct deftemplate *,struct symbolHashNode *);
   LOCALE int                            FindSlotPosition(struct deftemplate *,struct symbolHashNode *);
   LOCALE VOID                           AddDeftemplate(struct deftemplate *);
   LOCALE VOID                          *FindDeftemplate(char *);
   LOCALE VOID                          *QFindDeftemplate(struct symbolHashNode *);
   LOCALE VOID                          *GetNextDeftemplate(VOID *);
   LOCALE char                          *GetDeftemplateName(VOID *);
   LOCALE char                          *GetDeftemplatePPForm(VOID *);
   LOCALE int                            PPDeftemplate(char *,char *);
   LOCALE VOID                           ListDeftemplates(void);
   LOCALE BOOLEAN                        PrintTemplateFact(char *,struct fact *);
   LOCALE struct slp                    *FindSlotItem(struct templateSlot *,struct slp *);
   LOCALE BOOLEAN                        CheckSlotType(int,struct templateSlot *);
   LOCALE BOOLEAN                        CheckSlotAllowedValues(int,VOID *,struct templateSlot *);
   LOCALE BOOLEAN                        CheckSlotRange(int,VOID *,struct templateSlot *);
   LOCALE int                            GDDCommand(VOID);
   LOCALE int                            SDDCommand(VOID);
   LOCALE BOOLEAN                        SetDynamicDeftemplateChecking(int);
   LOCALE BOOLEAN                        GetDynamicDeftemplateChecking(VOID);
   LOCALE VOID                           SetListOfDeftemplates(VOID *);
   LOCALE VOID                           QSetListOfDeftemplates(VOID *);
   LOCALE VOID                           ClearDeftemplates(VOID);
   LOCALE BOOLEAN                        IsDeftemplateDeletable(VOID *);
   LOCALE BOOLEAN                        DeleteNamedDeftemplate(char *);
#else
   LOCALE VOID                           InitializeDeftemplates();
   LOCALE BOOLEAN                        DeleteDeftemplate();
   LOCALE VOID                           ReturnSlots();
   LOCALE struct templateSlot           *FindSlot();
   LOCALE int                            FindSlotPosition();
   LOCALE VOID                           AddDeftemplate();
   LOCALE BOOLEAN                        DeleteDeftemplate();
   LOCALE VOID                          *FindDeftemplate();
   LOCALE VOID                          *QFindDeftemplate();
   LOCALE VOID                          *GetNextDeftemplate();
   LOCALE char                          *GetDeftemplateName();
   LOCALE char                          *GetDeftemplatePPForm();
   LOCALE int                            PPDeftemplate();
   LOCALE VOID                           ListDeftemplates();
   LOCALE BOOLEAN                        PrintTemplateFact();
   LOCALE struct slp                    *FindSlotItem();
   LOCALE BOOLEAN                        CheckSlotType();
   LOCALE BOOLEAN                        CheckSlotAllowedValues();
   LOCALE BOOLEAN                        CheckSlotRange();
   LOCALE int                            GDDCommand();
   LOCALE int                            SDDCommand();
   LOCALE BOOLEAN                        SetDynamicDeftemplateChecking();
   LOCALE BOOLEAN                        GetDynamicDeftemplateChecking();
   LOCALE VOID                           SetListOfDeftemplates();
   LOCALE VOID                           QSetListOfDeftemplates();
   LOCALE VOID                           ClearDeftemplates();
   LOCALE BOOLEAN                        IsDeftemplateDeletable();
   LOCALE BOOLEAN                        DeleteNamedDeftemplate();
#endif

#endif





