   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10 07/17/91             */
   /*                                                     */
   /*                DEFTEMPLATE LHS MODULE               */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
  
#define _DEFTMLHS_SOURCE_

#include "setup.h"

#if DEFTEMPLATE_CONSTRUCT && DEFRULE_CONSTRUCT && (! RUN_TIME) && (! BLOAD_ONLY)

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "expressn.h"
#include "router.h"
#include "utility.h"
#include "constrct.h"
#include "reorder.h"
#include "ruleprsr.h"
#include "parsutil.h"
#include "generate.h"

#include "deftempl.h"
#include "deftmlhs.h"
  
#if ANSI_COMPILER  
   static struct slp             *GetLHSSlots(char *,struct token *,struct deftemplate *,int *);
   static struct slp             *GetSingleLHSSlot(char *,struct token *,struct templateSlot *,int *);
   static struct node            *ReorderLHSSlotValues(struct templateSlot *,struct slp *,struct node *);
   static BOOLEAN                 MultiplyDefinedLHSSlots(struct slp *);
   static VOID                    ReturnSLPs(struct slp *);
   static BOOLEAN                 CheckLHSSlotTypes(struct slp *,struct templateSlot *);
#else
   static struct slp             *GetLHSSlots();
   static struct slp             *GetSingleLHSSlot();
   static struct node            *ReorderLHSSlotValues();
   static BOOLEAN                 MultiplyDefinedLHSSlots();
   static VOID                    ReturnSLPs();
   static BOOLEAN                 CheckLHSSlotTypes();
#endif
   
/**************************************************/
/* DeftemplateLHSParse: Parses a LHS pattern that */
/*   uses the deftemplate format.                 */
/**************************************************/
globle struct node *DeftemplateLHSParse(readSource,hname)
  char *readSource;
  SYMBOL_HN *hname;
  {
   struct node *head;
   struct slp *first_slot;
   struct token temp_tkn;
   struct deftemplate *dtmpl_ptr;
   int error;
   
   head = NULL;
   
   /*===============================================*/
   /* Get a pointer to the deftemplate information. */
   /*===============================================*/
   
   dtmpl_ptr = (struct deftemplate *) FindDeftemplate(hname->contents);
   if (dtmpl_ptr == NULL)
     { CLIPSSystemError("DEFTMFUN",2); }

   /*===============================================================*/
   /* Make sure the deftemplate name is not connected to subfields. */
   /*===============================================================*/
   
   GetToken(readSource,&temp_tkn);
   if ((temp_tkn.type == LOR) || (temp_tkn.type == LAND))
     {
      PrintCLIPS(WERROR,"\nDeftemplate name used with connective constraint.\n");
      return(NULL);
     }
   
   /*===============================================*/
   /* Create pattern node for the deftemplate name. */
   /*===============================================*/
       
   head = GetNode();
   head->type = SINGLE;
   head->negated = CLIPS_FALSE;
   head->bottom = GetNode();
   head->bottom->type = SYMBOL;
   head->bottom->negated = CLIPS_FALSE;
   head->bottom->value = (VOID *) hname;

   /*======================================*/
   /* Get other fields in the deftemplate. */
   /*======================================*/
   
   error = CLIPS_FALSE;
   first_slot = GetLHSSlots(readSource,&temp_tkn,dtmpl_ptr,&error);
   if (error)
     {
      ReturnNodes(head);
      return(NULL);
     }

   /*==============================================================*/
   /* Determine if any of the slot names were used more than once. */
   /*==============================================================*/
   
   if (MultiplyDefinedLHSSlots(first_slot))
     {
      ReturnNodes(head);
      ReturnSLPs(first_slot);
      return(NULL);
     }
     
   /*===========================*/
   /* Reorder slot definitions. */
   /*===========================*/
     
   head = ReorderLHSSlotValues(dtmpl_ptr->slotList,first_slot,head);
   ReturnSLPs(first_slot);
   
   /*=========================*/
   /* Return the LHS pattern. */
   /*=========================*/
   
   return(head);   
  }

/******************************************/
/* GetLHSSlots: Retrieves all of the slot */
/*   values used in a LHS pattern.        */
/******************************************/
static struct slp *GetLHSSlots(readSource,tempToken,dtmpl_ptr,error)
  char *readSource;
  struct token *tempToken;
  struct deftemplate *dtmpl_ptr;
  int *error;
  {
   struct slp *first_slot = NULL, *next_slot, *last_slot = NULL;
   struct templateSlot *slotPtr;
   
   /*=======================================================*/
   /* Continue parsing slot definitions until the pattern's */
   /* closing right parenthesis is encountered.             */
   /*=======================================================*/
   
   while (tempToken->type != RPAREN)
     {
      PPBackup();
      SavePPBuffer(" ");
      SavePPBuffer(tempToken->print_rep);
      
      /*=================================================*/
      /* Slot definitions begin with a left parenthesis. */
      /*=================================================*/
      
      if (tempToken->type != LPAREN)
        {
         *error = CLIPS_TRUE;
         SyntaxErrorMessage("deftemplate patterns");
         return(NULL);
        }
        
      /*====================*/
      /* Get the slot name. */
      /*====================*/
      
      GetToken(readSource,tempToken);
      if (tempToken->type != SYMBOL)
        {
         *error = CLIPS_TRUE;
         SyntaxErrorMessage("deftemplate patterns");
         return(NULL);
        }
        
      /*==========================================================*/
      /* Determine if the slot name is valid for the deftemplate. */
      /*==========================================================*/
      
      if ((slotPtr = FindSlot(dtmpl_ptr,tempToken->value)) == NULL)
        {
         *error = CLIPS_TRUE;
         PrintCLIPS(WERROR,"\nInvalid slot ");
         PrintCLIPS(WERROR,ValueToString(tempToken->value));
         PrintCLIPS(WERROR," not defined in corresponding deftemplate ");
         PrintCLIPS(WERROR,ValueToString(dtmpl_ptr->name));
         PrintCLIPS(WERROR,"\n");
         ReturnSLPs(first_slot);
         return(NULL);
        }

      /*==============================================================*/
      /* Get the pattern matching values used in the slot definition. */
      /*==============================================================*/
      
      next_slot = GetSingleLHSSlot(readSource,tempToken,slotPtr,error);
      if (*error)
        {
         ReturnSLPs(first_slot);
         ReturnSLPs(next_slot);
         return(NULL);
        }
        
      /*===================================================*/
      /* Check the slot types to make sure they are valid. */
      /*===================================================*/
      
      if (CheckLHSSlotTypes(next_slot,slotPtr) == 0)
        {
         *error = CLIPS_TRUE;
         ReturnSLPs(first_slot);
         ReturnSLPs(next_slot);
         return(NULL);
        }
        
      /*=====================================*/
      /* Add the slot definition to the list */ 
      /* of slot definitions already parsed. */
      /*=====================================*/
      
      if (last_slot == NULL)
        { first_slot = next_slot; }
      else
        { last_slot->next = next_slot; }
      last_slot = next_slot;
         
      /*==============================*/
      /* Begin parsing the next slot. */
      /*==============================*/
      
      GetToken(readSource,tempToken);
     }
   
   /*===========================================================*/
   /* Return all the slot definitions found in the lhs pattern. */
   /*===========================================================*/
     
   return(first_slot);
  }
  
/*****************************************************/
/* GetSingleLHSSlot: Get the pattern matching values */
/*   to be associated with a slot name.              */
/*****************************************************/
static struct slp *GetSingleLHSSlot(readSource,tempToken,slotPtr,error)
  char *readSource;
  struct token *tempToken;
  struct templateSlot *slotPtr;
  int *error;
  {
   struct slp *next_slot;
   struct node *multi_node, *new_node, *last_node;
   
   next_slot = get_struct(slp);
   next_slot->name = (SYMBOL_HN *) tempToken->value;
   next_slot->expr = NULL;
   next_slot->next = NULL;

   SavePPBuffer(" ");
   GetToken(readSource,tempToken);

   /*====================================*/
   /* Get value for a single field slot. */
   /*====================================*/
   
   if (slotPtr->multiSlot == CLIPS_FALSE)
     {  
      /*=======================*/
      /* Get the single value. */
      /*=======================*/
      
      new_node = RestrictionParse(readSource,tempToken); 
      if (new_node == NULL)
        {
         *error = CLIPS_TRUE;
         ReturnSLPs(next_slot);
         return(NULL);
        }
      
      /*======================================*/
      /* Multi field wildcards and variables  */
      /* not allowed in a single field slot.  */
      /*======================================*/
      
      if ((new_node->type == BWORDS) || 
          (new_node->type == MULTIFIELD))
        {
         PrintCLIPS(WERROR,"\nA multifield value cannot be used ");
         PrintCLIPS(WERROR,"in the single field slot ");
         PrintCLIPS(WERROR,slotPtr->slotName->contents);
         PrintCLIPS(WERROR,"\n");
         *error = CLIPS_TRUE;
         ReturnNodes(new_node);
         ReturnSLPs(next_slot);
         return(NULL);
        }
     }
     
   /*====================================*/
   /* Get values for a multi field slot. */
   /*====================================*/
   
   else
     {
      multi_node = NULL;
      last_node = NULL;
      
      /*===============================================*/
      /* Continue parsing the multi field values until */
      /* a right parenthesis is encountered.           */
      /*===============================================*/
      
      while (tempToken->type != RPAREN)
        {
         /*==============*/
         /* Get a value. */
         /*==============*/
         
         new_node = RestrictionParse(readSource,tempToken); 
         if (new_node == NULL)
           {
            *error = CLIPS_TRUE;
            ReturnNodes(multi_node);
            ReturnSLPs(next_slot);
            return(NULL);
           }
           
         /*====================================*/
         /* Fix pretty print representation to */
         /* include spaces between values.     */
         /*====================================*/
                       
         if (tempToken->type != RPAREN)
           {
            PPBackup();
            SavePPBuffer(" ");
            SavePPBuffer(tempToken->print_rep);
           }
           
         /*=====================================*/
         /* Add the value to the list of values */
         /* for this multi field slot.          */
         /*=====================================*/
         
         if (last_node == NULL)
           { multi_node = new_node; }
         else
           { last_node->right = new_node; }
         last_node = new_node;
        }
      new_node = multi_node;
     }
   
   /*========================================================*/
   /* The slot definition must end with a right parenthesis. */
   /*========================================================*/
   
   if (tempToken->type != RPAREN)   
     {
      PPBackup();                         /* new */
      SavePPBuffer(" ");                  /* new */
      SavePPBuffer(tempToken->print_rep); /* new */
      SyntaxErrorMessage("deftemplate patterns");
      *error = CLIPS_TRUE;
      ReturnNodes(new_node);
      ReturnSLPs(next_slot);
      return(NULL);
     }        

   /*=================================*/
   /* Add the slot values to the slot */
   /* structure and return it.        */
   /*=================================*/
   
   next_slot->expr = new_node;
   return(next_slot);
  }

/****************************************************/ 
/* CheckLHSSlotTypes:                               */
/****************************************************/
static BOOLEAN CheckLHSSlotTypes(lhsSlots,slotPtr)
  struct slp *lhsSlots;
  struct templateSlot *slotPtr;
  {
   struct node *orNode, *andNode, *multiNode;
   char returnType;
      
   while (lhsSlots != NULL)
     {
      multiNode = lhsSlots->expr;
      while (multiNode != NULL)
        {
         orNode = multiNode->bottom;
         while (orNode != NULL)
           {
            andNode = orNode;
            while (andNode != NULL)
              {
               if ((slotPtr->anyAllowed == CLIPS_FALSE) &&
                   (((andNode->type == INTEGER) && (slotPtr->integersAllowed != CLIPS_TRUE)) ||
                    ((andNode->type == FLOAT) && (slotPtr->floatsAllowed != CLIPS_TRUE)) ||
                    ((andNode->type == STRING) && (slotPtr->stringsAllowed != CLIPS_TRUE)) ||
#if OBJECT_SYSTEM
                    ((andNode->type == INSTANCE_NAME) && (slotPtr->instancesAllowed != CLIPS_TRUE)) ||
#endif
                    ((andNode->type == SYMBOL) && (slotPtr->symbolsAllowed != CLIPS_TRUE))))
                 {
                  PrintCLIPS(WERROR,"\nLiteral restriction value(s) do not match\n");
                  PrintCLIPS(WERROR,"the allowed types for slot ");
                  PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
                  PrintCLIPS(WERROR,"\n");
                  return(0);
                 }
               else if ((andNode->type == FCALL) && (slotPtr->anyAllowed == CLIPS_FALSE))
                 {
                  returnType = ExpressionFunctionType(andNode->expression);
                  if ((((returnType == 'i') || (returnType == 'l')) && (slotPtr->integersAllowed != CLIPS_TRUE)) ||
                      (((returnType == 'f') || (returnType == 'd')) && (slotPtr->floatsAllowed != CLIPS_TRUE)) ||
                      ((returnType == 's') && (slotPtr->stringsAllowed != CLIPS_TRUE)) ||
                      ((returnType == 'a') && (slotPtr->addressesAllowed != CLIPS_TRUE)) ||
                      (((returnType == 'w') || (returnType == 'c')) && (slotPtr->symbolsAllowed != CLIPS_TRUE)) ||
                      ((returnType == 'b') || (returnType == 'v')))
                    {
                     PrintCLIPS(WERROR,"\nThe function return value does not match\n");
                     PrintCLIPS(WERROR,"the allowed types for slot ");
                     PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
                     PrintCLIPS(WERROR,"\n");
                     return(0);
                    }
                 }
           
               if (CheckSlotAllowedValues(andNode->type,andNode->value,slotPtr) == CLIPS_FALSE)
                 {
                  PrintCLIPS(WERROR,"\nLiteral restriction value(s) do not match\n");
                  PrintCLIPS(WERROR,"the allowed values for slot ");
                  PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
                  PrintCLIPS(WERROR,"\n");
                  return(0);
                 }
               else if (CheckSlotRange(andNode->type,andNode->value,slotPtr) == CLIPS_FALSE)
                 {
                  PrintCLIPS(WERROR,"\nLiteral restriction value(s) do not fall\n");
                  PrintCLIPS(WERROR,"in the allowed range for slot ");
                  PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
                  PrintCLIPS(WERROR,"\n");
                  return(0);
                 }
              
               andNode = andNode->right;
              }
            orNode = orNode->bottom;
           }
         multiNode = multiNode->right;
        }
        
      lhsSlots = lhsSlots->next;
     }
     
   return(1);
  }

/****************************************************/ 
/* ReorderLHSSlotValues: Reorders LHS pattern slots */
/*   to match the order used by the deftemplate.    */
/****************************************************/ 
static struct node *ReorderLHSSlotValues(slotPtr,first_slot,head)
  struct templateSlot *slotPtr;
  struct slp *first_slot;
  struct node *head;
  {
   struct node *last_node, *new_node;
   struct slp *slotItem;
   
   last_node = head;
   while (slotPtr != NULL)
     {
      /*==========================================================*/
      /* If the slot wasn't defined in the pattern, use a single  */
      /* field wildcard for a single field slot and a multi field */
      /* wildcard for a multi field slot.                         */
      /*==========================================================*/
      
      if ((slotItem = FindSlotItem(slotPtr,first_slot)) == NULL)
        {
         new_node = GetNode();
         new_node->negated = CLIPS_FALSE;
         if (slotPtr->multiSlot == CLIPS_FALSE)
           { new_node->type = SINGLE; }
         else
           { new_node->type = MULTIFIELD; }
        }
      else
       
      /*==========================================*/
      /* Otherwise, use the field associated with */
      /* the slot in the LHS pattern.             */ 
      /*==========================================*/
                    
        { 
         new_node = slotItem->expr;
         slotItem->expr = NULL; 
        }
        
      /*=============================================*/
      /* Add the field to the new reordered pattern. */
      /*=============================================*/
      
      if (new_node != NULL)
        {
         last_node->right = new_node;
         last_node = new_node;
        }
        
      slotPtr = slotPtr->next;
     }
    
   /*======================================================*/
   /* Return the reordered pattern with unspecified fields */
   /* replaced with match anything (? or $?) values.       */ 
   /*======================================================*/
   
   return(head);
  }

/******************************************************/
/* MultiplyDefinedLHSSlots: Determines if a slot name */
/*   was used more than once in a LHS pattern.        */
/******************************************************/
static BOOLEAN MultiplyDefinedLHSSlots(first_slot)
  struct slp *first_slot;
  {
   struct slp *next_slot;
   int rv = CLIPS_FALSE;
   
   while (first_slot != NULL)
     {
      next_slot = first_slot->next;
      while (next_slot != NULL)
        {
         if (next_slot->name == first_slot->name)
           {
            PrintCLIPS(WERROR,"\nMultiple occurences of slot ");
            PrintCLIPS(WERROR,first_slot->name->contents);
            PrintCLIPS(WERROR,"\n");
            rv = CLIPS_TRUE;
           }
         next_slot = next_slot->next;
        }
      first_slot = first_slot->next;
     }
  
   return(rv);
  }

/********************************************/
/* ReturnSLPs: Returns data structures used */ 
/*   for parsing deftemplates on the LHS.   */
/********************************************/
static VOID ReturnSLPs(slps)
  struct slp *slps;
  {
   struct slp *tempSlps;

   while (slps != NULL)
     {
      tempSlps = slps->next;
      if (slps->expr != NULL) ReturnNodes(slps->expr);
      rtn_struct(slp,slps);
      slps = tempSlps;
     }
  }
  
#endif


