   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 ENGINE HEADER FILE                  */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_engine

#define _H_engine

struct activation;

#include "symbol.h"
#include "match.h"
#include "defrule.h"
#include "network.h"

struct activation
  { 
   struct defrule *theRule;
   struct partialMatch *basis;  
   struct partialMatch *sortedBasis;
   int salience;
   unsigned long int timetag;
   int randomID;      
   struct activation *prev;             
   struct activation *next;       
  };
  
#define DEPTH_STRATEGY 0
#define BREADTH_STRATEGY 1
#define LEX_STRATEGY 2
#define MEA_STRATEGY 3
#define COMPLEXITY_STRATEGY 4
#define SIMPLICITY_STRATEGY 5
#define RANDOM_STRATEGY 6

#define DEFAULT_STRATEGY DEPTH_STRATEGY

#define MAX_PATTERNS_CHECKED 64

#define LESS_THAN -1
#define EQUAL 0
#define GREATER_THAN 1

#define WHEN_DEFINED 0
#define WHEN_ACTIVATED 1
#define EVERY_CYCLE 2

#if ANSI_COMPILER
   long                           RunCLIPS(long);
   VOID                           AddActivation(struct joinNode *,struct partialMatch *);
   VOID                           ClearRuleFromAgenda(struct symbolHashNode *);
   VOID                          *GetNextActivation(VOID *);
   char                          *GetActivationName(VOID *);
   struct partialMatch           *ActivationBasis(VOID *);
   int                            GetActivationSalience(VOID *);
   int                            SetActivationSalience(VOID *,int);
   VOID                           GetActivationPPForm(char *,int,VOID *);
   BOOLEAN                        MoveActivationToTop(VOID *);
   BOOLEAN                        DeleteActivation(VOID *);
   VOID                           PrintActivation(char *,struct activation *);
   VOID                           PrintCRSVActivation(char *,struct activation *);
   VOID                           ListAgenda(void);
   BOOLEAN                        AddRunFunction(char *,VOID (*)(),int);
   BOOLEAN                        RemoveRunFunction(char *);
   VOID                           RemoveActivation(struct activation *,int,int);
   VOID                           RemoveAllActivations(void);
   int                            GetAgendaChanged(void);
   VOID                           SetAgendaChanged(int);
   int                            SetStrategy(int);
   int                            GetStrategy(void);
   char                          *GetRuleFiring(void);
   long int                       GetNumberOfActivations(void);
   VOID                           InitializeEngine(void);
   VOID                           AddBreakpoint(void *);
   BOOLEAN                        RemoveBreakpoint(void *);
   VOID                           RemoveAllBreakpoints(void);
   VOID                           ListBreakpoints(void);
   BOOLEAN                        DefruleHasBreakpoint(void *);
   BOOLEAN                        GetSalienceEvaluation(void);
   BOOLEAN                        SetSalienceEvaluation(BOOLEAN);
   VOID                           RefreshAgenda(void);
   VOID                           ReorderAgenda(void);
#else
   long                           RunCLIPS();
   VOID                           AddActivation();
   VOID                           ClearRuleFromAgenda();
   VOID                          *GetNextActivation();
   char                          *GetActivationName();
   struct partialMatch           *ActivationBasis();
   int                            GetActivationSalience();
   int                            SetActivationSalience();
   VOID                           GetActivationPPForm();
   BOOLEAN                        MoveActivationToTop();
   BOOLEAN                        DeleteActivation();
   VOID                           PrintActivation();
   VOID                           PrintCRSVActivation();
   VOID                           ListAgenda();
   BOOLEAN                        AddRunFunction();
   BOOLEAN                        RemoveRunFunction();
   VOID                           RemoveActivation();
   VOID                           RemoveAllActivations();
   int                            GetAgendaChanged();
   VOID                           SetAgendaChanged();
   int                            SetStrategy();
   int                            GetStrategy();
   char                          *GetRuleFiring();
   long int                       GetNumberOfActivations();
   VOID                           InitializeEngine();
   VOID                           AddBreakpoint();
   BOOLEAN                        RemoveBreakpoint();
   VOID                           RemoveAllBreakpoints();
   VOID                           ListBreakpoints();
   BOOLEAN                        DefruleHasBreakpoint();
   BOOLEAN                        GetSalienceEvaluation();
   BOOLEAN                        SetSalienceEvaluation();
   VOID                           RefreshAgenda();
   VOID                           ReorderAgenda();
#endif

#ifndef _ENGINE_SOURCE_
   extern struct defrule      *ExecutingRule;
   extern BOOLEAN              HaltRules;
   extern BOOLEAN              DeletedFiringRule;
   extern struct joinNode     *TheLogicalJoin;
   extern int                  AgendaChanged;
#endif

#endif






