   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*               EVALUATION HEADER FILE                */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_evaluatn

#define _H_evaluatn

#include "constant.h"
#include "symbol.h"
#include "expressn.h"

struct dataObject
  {
   struct symbolHashNode *name;
   int type;
   VOID *value;
   int begin;
   int end;
   struct dataObject *next;
  };
  
typedef struct dataObject DATA_OBJECT;
typedef struct dataObject * DATA_OBJECT_PTR;

#define GetDOLength(target)       (((target).end - (target).begin) + 1)
#define GetpDOLength(target)      (((target)->end - (target)->begin) + 1)
#define GetDOBegin(target)        ((target).begin + 1)
#define GetDOEnd(target)          ((target).end + 1)
#define GetpDOBegin(target)       ((target)->begin + 1)
#define GetpDOEnd(target)         ((target)->end + 1)
#define SetDOBegin(target,val)   ((target).begin = (val) - 1) 
#define SetDOEnd(target,val)     ((target).end = (val) - 1)
#define SetpDOBegin(target,val)   ((target)->begin = (val) - 1) 
#define SetpDOEnd(target,val)     ((target)->end = (val) - 1)

#define DOPToString(target) (((struct symbolHashNode *) (target->value))->contents)
#define DOPToDouble(target) (((struct floatHashNode *) (target->value))->contents)
#define DOPToFloat(target) ((float) (((struct floatHashNode *) (target->value))->contents))
#define DOPToLong(target) (((struct integerHashNode *) (target->value))->contents)
#define DOPToInteger(target) ((int) (((struct integerHashNode *) (target->value))->contents))
#define DOPToPointer(target)       ((target)->value) 

#define DOToString(target) (((struct symbolHashNode *) (target.value))->contents)
#define DOToDouble(target) (((struct floatHashNode *) (target.value))->contents)
#define DOToFloat(target) ((float) (((struct floatHashNode *) (target.value))->contents))
#define DOToLong(target) (((struct integerHashNode *) (target.value))->contents)
#define DOToInteger(target) ((int) (((struct integerHashNode *) (target.value))->contents))
#define DOToPointer(target)        ((target).value) 

#define GetMFLength(target)     (((struct fact *) (target))->factLength) 
#define GetMFPtr(target,index)  (&(((struct fact *) (target))->atoms[index-1])) 
#define SetMFType(target,index,value)  (((struct fact *) (target))->atoms[index-1].type = (value)) 
#define SetMFValue(target,index,val)  (((struct fact *) (target))->atoms[index-1].value = (VOID *) (val))  
#define GetMFType(target,index)  (((struct fact *) (target))->atoms[index-1].type) 
#define GetMFValue(target,index)  (((struct fact *) (target))->atoms[index-1].value) 

#define CoerceToLongInteger(t,v) ((t == INTEGER) ? ValueToLong(v) : (long int) ValueToDouble(v))
#define CoerceToInteger(t,v) ((t == INTEGER) ? (int) ValueToLong(v) : (int) ValueToDouble(v))
#define CoerceToDouble(t,v) ((t == INTEGER) ? (double) ValueToLong(v) : ValueToDouble(v))
  
#define GetFirstArgument()           (CurrentExpression->arg_list)
#define GetNextArgument(ep)          (ep->next_arg)

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _EVALUATN_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
   LOCALE int                            EvaluateExpression(struct expr *,struct dataObject *);
   LOCALE int                            DefineFunction(char *,int,int (*)(),char *);
   LOCALE char                          *RtnLexeme(int);
   LOCALE double                         RtnDouble(int);
   LOCALE long                           RtnLong(int);
   LOCALE struct dataObject             *RtnUnknown(int,struct dataObject *);
   LOCALE int                            rmultype(struct dataObject *,int);
   LOCALE VOID                          *rmulvalue(struct dataObject *,int);
   LOCALE char                          *rmulstring(struct dataObject *,int);
   LOCALE struct symbolHashNode         *rmulhash(struct dataObject *,int);
   LOCALE double                         rmulfloat(struct dataObject *,int);
   LOCALE VOID                           AssignUnknown(int,struct symbolHashNode *,double,struct dataObject *);
   LOCALE int                            RtnArgCount(void);
   LOCALE VOID                           SetEvaluationError(BOOLEAN);
   LOCALE int                            GetEvaluationError(void);
   LOCALE VOID                           SetHaltExecution(int);
   LOCALE int                            GetHaltExecution(void);
   LOCALE int                            ArgCountCheck(char *,int,int);
   LOCALE int                            ArgRangeCheck(char *,int,int);
   LOCALE int                            ArgTypeCheck(char *,int,int,struct dataObject *);
   LOCALE VOID                           ReturnValues(struct dataObject *);
   LOCALE VOID                           PrintDataObject(char *,struct dataObject *);
   LOCALE VOID                           SetMultifieldErrorValue(struct dataObject *);
   LOCALE VOID                           ValueInstall(struct dataObject *);
   LOCALE VOID                           ValueDeinstall(struct dataObject *);
   LOCALE VOID                           PropogateReturnValue(struct dataObject *);
   LOCALE BOOLEAN                        GetBoundVariable(struct dataObject *,struct symbolHashNode *);
#if DEFFUNCTION_CONSTRUCT || DEFGENERIC_CONSTRUCT
   LOCALE int                            CLIPSFunctionCall(char *,char *,DATA_OBJECT *);
#endif
   LOCALE VOID                           CopyDataObject(DATA_OBJECT *,DATA_OBJECT *);
#else
   LOCALE int                            EvaluateExpression();
   LOCALE int                            DefineFunction();
   LOCALE char                          *RtnLexeme();
   LOCALE double                         RtnDouble();
   LOCALE long                           RtnLong();
   LOCALE struct dataObject             *RtnUnknown();
   LOCALE int                            rmultype();
   LOCALE VOID                          *rmulvalue();
   LOCALE char                          *rmulstring();
   LOCALE struct symbolHashNode         *rmulhash();
   LOCALE double                         rmulfloat();
   LOCALE VOID                           AssignUnknown();
   LOCALE int                            RtnArgCount();
   LOCALE VOID                           SetEvaluationError();
   LOCALE int                            GetEvaluationError();
   LOCALE VOID                           SetHaltExecution();
   LOCALE int                            GetHaltExecution();
   LOCALE int                            ArgCountCheck();
   LOCALE int                            ArgRangeCheck();
   LOCALE int                            ArgTypeCheck();
   LOCALE VOID                           ReturnValues();
   LOCALE VOID                           PrintDataObject();
   LOCALE VOID                           SetMultifieldErrorValue();
   LOCALE VOID                           ValueInstall();
   LOCALE VOID                           ValueDeinstall();
   LOCALE VOID                           PropogateReturnValue();
   LOCALE BOOLEAN                        GetBoundVariable();
#if DEFFUNCTION_CONSTRUCT || DEFGENERIC_CONSTRUCT
   LOCALE int                            CLIPSFunctionCall();
#endif
   LOCALE VOID                           CopyDataObject();
#endif

#ifndef _EVALUATN_SOURCE_
   extern struct expr            *CurrentExpression;
   extern int                     EvaluationError;
   extern int                     HaltExecution;
   extern int                     CurrentEvaluationDepth;
   extern struct dataObject      *BindList;
#endif

#endif






