   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                   GENERATE MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
#define _GENERATE_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"
#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "expressn.h"
#include "reorder.h"
#include "variable.h"
#include "evaluatn.h"
#include "utility.h"
#include "router.h"
#include "generate.h"
  
#if DEFGLOBAL
#include "defglobl.h"
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/
    
   globle VOID              *PTR_NOP;
   globle VOID              *PTR_CONSTANT;
   globle VOID              *PTR_NOTCONSTANT;
   globle VOID              *PTR_GET_FIELD;
   globle VOID              *PTR_AND; 
   globle VOID              *PTR_OR; 
   globle VOID              *PTR_EQ;
   globle VOID              *PTR_NEQ; 
   globle VOID              *PTR_NOT;
   globle VOID              *PTR_EQ_FIELD;
   globle VOID              *PTR_NEQ_FIELD;
   
/****************************************************/
/* InitGenModule: Initializes the function pointers */
/*   used by the generate module.                   */
/****************************************************/
globle VOID InitGenModule()
  {
   PTR_AND          = (VOID *) FindFunction("and");
   PTR_OR           = (VOID *) FindFunction("or"); 
   PTR_EQ           = (VOID *) FindFunction("eq");
   PTR_NEQ          = (VOID *) FindFunction("neq");
   PTR_NOT          = (VOID *) FindFunction("not");
#if DEFRULE_CONSTRUCT
   PTR_NOP          = (VOID *) FindFunction("(nop)");
   PTR_CONSTANT     = (VOID *) FindFunction("(constant)");
   PTR_NOTCONSTANT  = (VOID *) FindFunction("(notconstant)");
   PTR_EQ_FIELD     = (VOID *) FindFunction("(eq_field)");
   PTR_NEQ_FIELD    = (VOID *) FindFunction("(neq_field)");
   PTR_GET_FIELD    = (VOID *) FindFunction("(get_field)");
#endif

   if ((PTR_AND == NULL) || (PTR_OR == NULL) ||
#if DEFRULE_CONSTRUCT       
       (PTR_NOP == NULL) || 
       (PTR_CONSTANT == NULL) || (PTR_NOTCONSTANT == NULL) ||
       (PTR_GET_FIELD == NULL) ||
       (PTR_EQ_FIELD == NULL) || (PTR_NEQ_FIELD == NULL) ||
#endif
       (PTR_EQ == NULL) || (PTR_NEQ == NULL) || (PTR_NOT == NULL))
     {
      CLIPSSystemError("GENERATE",1);
      ExitCLIPS(4);
     }
 
  }
  
/******************************************************/
/* GenConstant: Generates a constant expression value */
/*   of type string, symbol, or number.                 */
/******************************************************/
globle struct expr *GenConstant(type,value)
  int type;
  VOID *value;
  {
   struct expr *top;

   top = get_struct(expr);
   top->next_arg = NULL;
   top->arg_list = NULL;
   top->type = type;
   top->value = value;
     
   return(top);
  }
  
/***************************************************************/
/* GenTwoIntegers: */
/***************************************************************/
globle struct expr *GenTwoIntegers(e1,e2)
  int e1, e2;
  {
   struct expr *top;

   top = GenConstant(INTEGER,AddLong((long int) e1));
   top->next_arg = GenConstant(INTEGER,AddLong((long int) e2));
   
   return(top);
  }
  
#if (! RUN_TIME) && (! BLOAD_ONLY) && DEFRULE_CONSTRUCT

/****************************************************/
/* GenGetvar: Produces an expression of the format */   
/*   (getvar <pattern> <field>)                     */
/****************************************************/
globle struct expr *GenGetvar(pattern,element)
  int pattern, element;
  {
   struct expr *top;
   
   top = GenConstant(SCALL_GET_VAR,GenGetvarValue(pattern,element));
   
   return(top);
  }
  
/****************************************************/
/* GenGetvarValue:               */
/****************************************************/
globle VOID *GenGetvarValue(pattern,element)
  int pattern, element;
  {
   union
     {
      void *theValue;
      struct 
         {
          unsigned int flag : 1;
          unsigned int int1 : 8;
          unsigned int int2 : 8;
         } theInts;
     } hack;

   hack.theValue = NULL;
   if (element < 0) 
     {
      hack.theInts.flag = 1;
      element = -element;
     }
   else hack.theInts.flag = 0;
   hack.theInts.int1 = pattern;
   hack.theInts.int2 = element;
   
   return(hack.theValue);
  }

/****************************************************/
/* GenGetfield: Produces an expression of the format */   
/*   (getfield <field>)                     */
/****************************************************/
globle struct expr *GenGetfield(element)
  int element;
  {
   struct expr *top;

   top = GenConstant(FCALL,PTR_GET_FIELD);
   top->arg_list = GenConstant(INTEGER,AddLong((long) element));
   
   return(top);
  }

/******************************************************/
/* GenFourIntegers:  Produces an argument list to be  */
/*   used with other expressions (primarily eq_vars   */
/*   or neq_vars expressions).                        */
/*   Example:                                         */
/*     If the third element of the first pattern is   */
/*     to be compared with the second element of      */
/*     the third pattern then the argument list       */
/*     (1 3 3 2) would be produced.                   */
/*     If the expression to be evaluated were eq_var, */
/*     then the arguments would be combined with the  */
/*     expression to produce (eq_vars 1 3 3 2).       */
/******************************************************/
globle struct expr *GenFourIntegers(p1,e1,p2,e2)
  int p1, e1, p2, e2;
  {
   struct expr *top;

   top = GenTwoIntegers(p1,e1);
   top->next_arg->next_arg = GenTwoIntegers(p2,e2);
   
   return(top);
  }

  
/**************************************************************/
/* GenAnd:  Generates an and function call with no arguments. */
/**************************************************************/
globle struct expr *GenAnd()
  { return(GenConstant(FCALL,PTR_AND)); }
  
/************************************************************/
/* GenOr:  Generates an or function call with no arguments. */
/************************************************************/
globle struct expr *GenOr()
  { return(GenConstant(FCALL,PTR_OR)); }
  
/*******************************************************/
/* GenPNConstant: Generates a constant test expression */
/*   for use in the pattern network.                   */
/*   Convert value   To   (constant value)             */
/*   Convert ~value  To   (notconstant value)          */
/*******************************************************/
globle struct expr *GenPNConstant(logic,type,value)
  int logic, type;
  VOID *value;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = GenConstant(FCALL,PTR_CONSTANT); }
   else
     { top = GenConstant(FCALL,PTR_NOTCONSTANT); }

   top->arg_list = GenConstant(type,value);
     
   return(top);
  }
  
/*******************************************************/
/* GenJNConstant: Generates a constant test expression */
/*   for use in the join network.                      */
/*   Convert                                           */
/*      value                                          */
/*   To                                                */
/*      (eq (getvar <pattern> <field>) value)          */
/*   Convert                                           */
/*      ~value                                         */
/*   To                                                */
/*      (neq (getvar <pattern> <field>) value)         */
/*******************************************************/
globle struct expr *GenJNConstant(logic,type,value,pattern,element)
  int logic, type;
  VOID *value;
  int pattern, element;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = GenConstant(FCALL,PTR_EQ); }
   else
     { top = GenConstant(FCALL,PTR_NEQ); }

   top->arg_list = GenGetvar(pattern,element);
   top->arg_list->next_arg = GenConstant(type,value);
     
   return(top);
  }
  
/********************************************************/
/* GenJNColon: Generates a test expression for use in   */
/*   the join network for the such that operator.       */
/*   Convert                                            */
/*      :(expression)                                   */
/*   To                                                 */
/*      (expression)                                    */
/*   Convert                                            */
/*      ~:(expression)                                  */
/*   To                                                 */
/*      (! (expresion))                                 */
/*   The variables in the expressions are replaced with */
/*   getvar calls.                                      */
/********************************************************/
globle struct expr *GenJNColon(logic,ePtr,pattern,element,anywhere)
  int logic;
  struct expr *ePtr;
  int pattern, element;
  int anywhere;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = ePtr; }
   else
     { 
      top = GenConstant(FCALL,PTR_NOT);
      top->arg_list = ePtr;
     }
        
   GetvarReplace(top,pattern,element,anywhere);
   return(top);
  }
  
/********************************************************/
/* GenPNColon: Generates a test expression for use in   */
/*   the pattern network for the such that operator.    */
/*   Convert                                            */
/*      :(expression)                                   */
/*   To                                                 */
/*      (expression)                                    */
/*   Convert                                            */
/*      ~:(expression)                                  */
/*   To                                                 */
/*      (! (expresion))                                 */
/*   The variables in the expressions are replaced with */
/*   getfield calls.                                    */
/********************************************************/
globle struct expr *GenPNColon(logic,ePtr,pattern,element)
  int logic;
  struct expr *ePtr;
  int pattern, element;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = ePtr; }
   else
     { 
      top = GenConstant(FCALL,PTR_NOT);
      top->arg_list = ePtr;
     }
        
   GetfieldReplace(top,pattern,element);
   return(top);
  }
  
/********************************************************/
/* GenJNEq: Generates a test expression for use in      */
/*   the join network for the = operator.               */
/*   Convert                                            */
/*      =(expression)                                   */
/*   To                                                 */
/*      (eq (getvar <pattern> <field>) (expression))    */
/*   Convert                                            */
/*      ~=(expression)                                  */
/*   To                                                 */
/*      (neq (getvar <pattern> <field>) (expression))   */
/*   The variables in the expressions are replaced with */
/*   getvar calls.                                      */
/********************************************************/
globle struct expr *GenJNEq(logic,ePtr,pattern,element,anywhere)
  int logic;
  struct expr *ePtr;
  int pattern, element, anywhere;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = GenConstant(FCALL,PTR_EQ); }
   else
     { top = GenConstant(FCALL,PTR_NEQ); }
     
   top->arg_list = GenGetvar(pattern,element);
   top->arg_list->next_arg = ePtr;
      
   GetvarReplace(top->arg_list->next_arg,pattern,element,anywhere);
   return(top);
  }
  
/********************************************************/
/* GenPNEq: Generates a test expression for use in      */
/*   the join network for the = operator.               */
/*   Convert                                            */
/*      =(expression)                                   */
/*   To                                                 */
/*      (eq (getfield <field>) (expression))            */
/*   Convert                                            */
/*      ~=(expression)                                  */
/*   To                                                 */
/*      (neq (getfield <field>) (expression))           */
/*   The variables in the expressions are replaced with */
/*   getfield calls.                                    */
/********************************************************/
globle struct expr *GenPNEq(logic,ePtr,pattern,element)
  int logic;
  struct expr *ePtr;
  int pattern, element;
  {
   struct expr *top;
   
   if (logic == POSITIVE)
     { top = GenConstant(FCALL,PTR_EQ); }
   else
     { top = GenConstant(FCALL,PTR_NEQ); }
     
   top->arg_list = GenGetfield(element);
   top->arg_list->next_arg = ePtr;
      
   GetfieldReplace(top->arg_list->next_arg,pattern,element);
   return(top);
  }
  
/***************************************************/
/* GetvarReplace: Replaces occurences of variables */
/*   in expressions with get_var calls.            */
/***************************************************/
globle VOID GetvarReplace(expr,pattern,element,anywhere)
  struct expr *expr;
  int pattern, element, anywhere;
  {
   struct var_info *var_ptr;
   int fap;
   
   while (expr != NULL)
     {
      if ((expr->type == BWORD) || 
          (expr->type == BWORDS))
        {
         if (anywhere)
           { var_ptr = FindVariable(expr->value,1,pattern,-1,INSIDE); }
         else
           { var_ptr = FindVariable(expr->value,1,pattern,element+1,INSIDE); }
           
         if (var_ptr == NULL)
           {  
            fap = GetFactAddressPosition(expr->value);
            
            if (fap == 0)
              {
               CLIPSSystemError("GENERATE",2);
               ExitCLIPS(4);
              }
             
            expr->type = FCALL;
            expr->value = (VOID *) FindFunction("(pointer)");
            expr->arg_list = GenConstant(INTEGER,AddLong((long) fap));
           }
         else
           {
            expr->type = SCALL_GET_VAR;
            expr->value = GenGetvarValue(var_ptr->pattern,var_ptr->element);
           }
        }
#if DEFGLOBAL
      else if (expr->type == GBWORD)
        { ReplaceGlobalVariable(expr); }
#endif
      else if ((expr->type == FCALL)
#if DEFGENERIC_CONSTRUCT
               || (expr->type == GCALL)
#endif
#if DEFFUNCTION_CONSTRUCT
               || (expr->type == PCALL)
#endif
         )
        { GetvarReplace(expr->arg_list,pattern,element,anywhere); }
        
      expr = expr->next_arg;
     }
  }
  
/***************************************************/
/* GetfieldReplace: Replaces occurences of variables */
/*   in expressions with get_field calls.            */
/***************************************************/
globle VOID GetfieldReplace(expr,pattern,element)
  struct expr *expr;
  int pattern, element;
  {
   struct var_info *var_ptr;
   
   while (expr != NULL)
     {
      if ((expr->type == BWORD) || (expr->type == BWORDS))
        {
         var_ptr = FindVariable(expr->value,pattern,pattern,element+1,INSIDE);
         if (var_ptr == NULL)
           { 
            CLIPSSystemError("GENERATE",3);
            ExitCLIPS(4);
           }
           
         expr->type = FCALL;
         expr->value = PTR_GET_FIELD;
         expr->arg_list = GenConstant(INTEGER,AddLong((long) var_ptr->element));
        }
#if DEFGLOBAL
      else if (expr->type == GBWORD)
        { ReplaceGlobalVariable(expr); }
#endif
      else if ((expr->type == FCALL)
#if DEFGENERIC_CONSTRUCT
               || (expr->type == GCALL)
#endif
#if DEFFUNCTION_CONSTRUCT
               || (expr->type == PCALL)
#endif
         )
        { GetfieldReplace(expr->arg_list,pattern,element); }
        
      expr = expr->next_arg;
     }
  }
   
/****************************************************/
/* GenJNVariableComparison               */
/****************************************************/
globle struct expr *GenJNVariableComparison(logic,value,pattern,element,anywhere)
  int logic;
  VOID *value;
  int pattern, element, anywhere;
  {
   struct var_info *var_ptr;
   struct expr *top;
   union
     {
      void *theValue;
      struct 
         {
          unsigned int pass : 1;
          unsigned int fail : 1;
          unsigned int int1 : 7;
          unsigned int int2 : 8;
          unsigned int int3 : 7;
         } theInts;
     } hack;
   
   if (anywhere)
     {
      var_ptr = FindVariable(value,1,pattern,-1,INSIDE);
     }
   else
     { 
      var_ptr = FindVariable(value,1,pattern,element+1,INSIDE);
     }
                                 
   if (var_ptr == NULL)
     {
      CLIPSSystemError("GENERATE",4);
      ExitCLIPS(4);
     }
          
   hack.theValue = NULL;
   hack.theInts.pass = 0;
   hack.theInts.fail = 0;
   hack.theInts.int1 = (unsigned int) element;
   hack.theInts.int2 = (unsigned int) var_ptr->pattern;
   hack.theInts.int3 = (unsigned int) var_ptr->element;  
   
   if (logic == POSITIVE)
     { 
      hack.theInts.pass = 1;
      top = GenConstant(SCALL_COMP_VARS,hack.theValue); 
     }
   else
     {
      hack.theInts.fail = 1;
      top = GenConstant(SCALL_COMP_VARS,hack.theValue); 
     }
   
   return(top);
  }

/****************************************************/
/* GenPNVariableComparison               */
/****************************************************/
globle struct expr *GenPNVariableComparison(logic,value,pattern,element)
  int logic;
  VOID *value;
  int pattern, element;
  {
   struct var_info *var_ptr;
   struct expr *top;
   
   var_ptr = FindVariable(value,pattern,pattern,element+1,INSIDE);
                                 
   if (var_ptr == NULL)
     {
      CLIPSSystemError("GENERATE",5);
      ExitCLIPS(4);
     }
                
   if (logic == POSITIVE)
     { top = GenConstant(FCALL,PTR_EQ_FIELD); }
   else
     { top = GenConstant(FCALL,PTR_NEQ_FIELD); } 
     
   top->arg_list = GenTwoIntegers(element,var_ptr->element);
 
   return(top);
  }
  
/****************************************************/
/* CombineExpressions       */
/****************************************************/
globle struct expr *CombineExpressions(exp1,exp2)
  struct expr *exp1, *exp2;
  {
   struct expr *chase_ptr;
   
   /*========*/
   /* CASE 1 */
   /*========*/
   
   if (exp1 == NULL)
     { return(exp2); }
     
   /*========*/
   /* CASE 2 */
   /*========*/
   
   if (exp2 == NULL)
     { return(exp1); }
     
   /*========*/
   /* CASE 3 */
   /*========*/
   
   if ((exp1->value == PTR_AND) && 
       (exp2->value != PTR_AND))
     {
      chase_ptr = exp1->arg_list;
      if (chase_ptr == NULL)
        { 
         rtn_struct(expr,exp1);
         return(exp2);
        }
        
      while (chase_ptr->next_arg != NULL) 
        { chase_ptr = chase_ptr->next_arg; }
      
      chase_ptr->next_arg = exp2;
      return(exp1);
     }
     
   /*========*/
   /* CASE 4 */
   /*========*/
   
   if ((exp1->value != PTR_AND) && 
       (exp2->value == PTR_AND))
     {
      chase_ptr = exp2->arg_list;
      if (chase_ptr == NULL)
        { 
         rtn_struct(expr,exp2);
         return(exp1);
        }
        
      exp2->arg_list = exp1;
      exp1->next_arg = chase_ptr;
      
      return(exp2);
     }
     
   /*========*/
   /* CASE 5 */
   /*========*/
     
   if ((exp1->value == PTR_AND) && 
       (exp2->value == PTR_AND))
     {
      chase_ptr = exp1->arg_list;
      if (chase_ptr == NULL)
        { 
         rtn_struct(expr,exp1);
         return(exp2);
        }
        
      while (chase_ptr->next_arg != NULL) 
        { chase_ptr = chase_ptr->next_arg; }
      
      chase_ptr->next_arg = exp2->arg_list;
      rtn_struct(expr,exp2);
      
      return(exp1);
     }
  
   /*========*/
   /* CASE 6 */
   /*========*/
   
   chase_ptr = GenAnd();
   chase_ptr->arg_list = exp1;
   exp1->next_arg = exp2;
   return(chase_ptr);
  }


#endif
