   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                INTERFACE FILE MODULE                */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Bebe Ly                                              */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _INTRFILE_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#include "clipsmem.h"
#include "evaluatn.h"
#include "router.h"
#include "utility.h"
#include "constrct.h"
#include "intrfile.h"

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bsave.h"
#include "bload.h"
#endif

#if ANSI_COMPILER  
#if DEBUGGING_FUNCTIONS
   static int                     FindCRSVTrace(char *);
   static int                     FindDribble(char *);
   static int                     GetcDribble(char *);
   static int                     UngetcDribble(int,char *);
   static VOID                    ExitDribble(int);
   static VOID                    PrintDribble(char *,char *);
   static VOID                    PutcDribbleBuffer(int);
   static VOID                    PrintCRSVTrace(char *,char *);
   static VOID                    ExitCRSVTrace(int);
#endif
   static int                     FindBatch(char *);
   static int                     GetcBatch(char *);
   static int                     UngetcBatch(int,char *);
   static VOID                    ExitBatch(int);
   static VOID                    AddBatch(int,VOID *,int,char *);
   VOID                           BatchCommand(void);
   VOID                           CRSVTraceOnCommand(void);
   VOID                           CRSVTraceOffCommand(void);
   VOID                           LoadCommand(void);
   VOID                           SaveCommand(void);
   VOID                           DribbleOnCommand(void);
   VOID                           DribbleOffCommand(void);
#else
#if DEBUGGING_FUNCTIONS
   static VOID                    PrintDribble();
   static VOID                    PutcDribbleBuffer();
   static VOID                    PrintCRSVTrace();
   static int                     FindDribble();
   static int                     GetcDribble();
   static int                     UngetcDribble();
   static VOID                    ExitDribble();
   static int                     FindCRSVTrace();
   static VOID                    ExitCRSVTrace();
#endif
   static int                     FindBatch();
   static int                     GetcBatch();
   static int                     UngetcBatch();
   static VOID                    ExitBatch();
   static VOID                    AddBatch();
   VOID                           BatchCommand();
   VOID                           CRSVTraceOnCommand();
   VOID                           CRSVTraceOffCommand();
   VOID                           LoadCommand();
   VOID                           SaveCommand();
   VOID                           DribbleOnCommand();
   VOID                           DribbleOffCommand();
#endif
  
/***********************************************************/
/* batchEntry structure:                                   */
/***********************************************************/

#define FILE_BATCH   0
#define STRING_BATCH 1

struct batchEntry
  {
   int batchType;
   VOID *inputSource;
   char *theString;
   struct batchEntry *next;
  };

/*********************************/
/* DRIBBLE AND BATCH DEFINITIONS */
/*********************************/

#define BUFFSIZE       120          /* Buffer allocation increments  */
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

#if DEBUGGING_FUNCTIONS
   static FILE               *DribbleFP = NULL;
   static char               *DribbleBuffer;
   static int                 DribbleCurrentPosition = 0;
   static int                 DribbleMaximumPosition = 0;
   static FILE               *CRSVTraceFP = NULL;
#if ANSI_COMPILER
   static int               (*DribbleStatusFunction)(int) = NULL;
#else
   static int               (*DribbleStatusFunction)() = NULL;
#endif
#endif
   static int                 BatchType;
   static VOID               *BatchSource = NULL;
   static char               *BatchBuffer;
   static int                 bat_loc = 0;
   static int                 bat_max = 0;
   static struct batchEntry  *batch_top = NULL;
   static struct batchEntry  *batch_bot = NULL;

/****************************/
/* FileFunctionDefinitions: */
/****************************/
#if ! RUN_TIME
globle VOID FileFunctionDefinitions()
  {
#if DEBUGGING_FUNCTIONS
   DefineFunction("batch",         'v', PTIF BatchCommand,        "BatchCommand");
   DefineFunction("dribble-on",    'v', PTIF DribbleOnCommand,    "DribbleOnCommand");
   DefineFunction("dribble-off",   'v', PTIF DribbleOffCommand,   "DribbleOffCommand");
   DefineFunction("crsv-trace-on", 'v', PTIF CRSVTraceOnCommand,  "CRSVTraceOnCommand");
   DefineFunction("crsv-trace-off",'v', PTIF CRSVTraceOffCommand, "CRSVTraceOffCommand");
   DefineFunction("save",          'v', PTIF SaveCommand,         "SaveCommand");
#endif
   DefineFunction("load",          'v', PTIF LoadCommand,         "LoadCommand");
#if BLOAD_AND_BSAVE
   DefineFunction("bsave",'v', PTIF BsaveCommand,"BsaveCommand");
#endif
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   DefineFunction("bload",'v', PTIF BloadCommand,"BloadCommand");
#endif
  }
#endif
  
/*********************/
/* TRACING FUNCTIONS */
/*********************/

#if DEBUGGING_FUNCTIONS
/*************************************************/
/* FindDribble:                                   */
/*************************************************/
static int FindDribble(logicalName)
  char *logicalName;
  {
   if ( (strcmp(logicalName,"stdout") == 0) ||
        (strcmp(logicalName,"stdin") == 0) ||
        (strcmp(logicalName,WCLIPS) == 0) ||
        (strcmp(logicalName,WTRACE) == 0) ||
        (strcmp(logicalName,WERROR) == 0) ||
        (strcmp(logicalName,WDISPLAY) == 0) ||
        (strcmp(logicalName,WDIALOG) == 0) )
     { return(CLIPS_TRUE); }

    return(CLIPS_FALSE);
  }

/*************************************************/
/* PrintDribble:                                  */
/*************************************************/
static VOID PrintDribble(logicalName,str)
  char *logicalName, *str;
  {
   int i = 0;
  
   /*======================================*/
   /* Send the output to the dribble file. */
   /*======================================*/
   
   for (i = 0 ; str[i] != EOS ; i++)
     { PutcDribbleBuffer(str[i]); }

   /*===========================================================*/
   /* Send the output to any routers interested in printing it. */
   /*===========================================================*/
   
   DeactivateRouter("dribble");
   PrintCLIPS(logicalName,str);
   ActivateRouter("dribble");
  }

/*************************************************/
/* GetcDribble:                                   */
/*************************************************/
static int GetcDribble(logicalName)
  char *logicalName;
  {
   int rv;
 
   DeactivateRouter("dribble");
   rv = GetcCLIPS(logicalName);
 
   ActivateRouter("dribble");
 
   PutcDribbleBuffer(rv);
 
   return(rv);         
  }

/*************************************************/
/* PutcDribbleBuffer:                            */
/*************************************************/
static VOID PutcDribbleBuffer(rv)
  int rv;
  {
   if (rv == EOF)
     {
      if (DribbleCurrentPosition > 0)
        {
         fprintf(DribbleFP,"%s",DribbleBuffer);
         DribbleCurrentPosition = 0;
         DribbleBuffer[0] = EOS;
        }
     }
   else if (CLIPSInputCount < 0)
     {
      if (DribbleCurrentPosition > 0)
        {
         fprintf(DribbleFP,"%s",DribbleBuffer);
         DribbleCurrentPosition = 0;
         DribbleBuffer[0] = EOS;
        }
      
      fputc(rv,DribbleFP);
     }
   else
     {
      DribbleBuffer = ExpandStringWithChar(rv,DribbleBuffer,
                                           &DribbleCurrentPosition,
                                           &DribbleMaximumPosition,
                                           DribbleMaximumPosition+BUFFSIZE);
     }
  }

/*************************************************/
/* UngetcDribble:                                 */
/*************************************************/    
static int UngetcDribble(ch,logicalName)
  int ch;
  char *logicalName;
  {
   int rv;

   DeactivateRouter("dribble");
   if (DribbleCurrentPosition > 0) DribbleCurrentPosition--;
   DribbleBuffer[DribbleCurrentPosition] = EOS;
   rv = UngetcCLIPS(ch,logicalName);
   ActivateRouter("dribble");

   return(rv);
  }

/*************************************************/
/* ExitDribble:                                   */
/*************************************************/
static VOID ExitDribble(num)
  int num;
  {
#if MAC_MPW
#pragma unused(num)
#endif   
   if (DribbleCurrentPosition > 0) 
     { fprintf(DribbleFP,"%s",DribbleBuffer); }
     
   if (DribbleFP != NULL) fclose(DribbleFP);
  }

/*************************************************/
/* DribbleOnCommand:                             */
/*************************************************/
globle VOID DribbleOnCommand()
  {
   char *fileName;

   if (ArgCountCheck("dribble-on",EXACTLY,1) == -1) return;
   if ((fileName = GetFileName("dribble-on",1)) == NULL) return;
  
   OpenDribble(fileName);
  }
  
/*************************************************/
/* OpenDribble:                                 */
/*************************************************/
globle BOOLEAN OpenDribble(fileName)
  char *fileName;
  {
   if (DribbleFP != NULL)
     { CloseDribble(); }
 
   DribbleFP = fopen(fileName,"w");
   if (DribbleFP == NULL)
     {
      OpenErrorMessage("dribble-on",fileName);     
      return(0);
     }

   AddRouter("dribble",         /* Router name     */
              40,                /* Priority        */
              FindDribble,       /* Query function  */
              (int (*)()) PrintDribble,      /* Print function  */
              GetcDribble,       /* Getc function   */
              UngetcDribble,     /* Ungetc function */
              (int (*)()) ExitDribble);      /* Exit function   */

   DribbleCurrentPosition = 0;

   if (DribbleStatusFunction != NULL)
     { (*DribbleStatusFunction)(CLIPS_TRUE); }
     
   return(1);
  }
  
/*************************************************/
/* DribbleActive:                               */
/*************************************************/
globle BOOLEAN DribbleActive()
  {
   if (DribbleFP != NULL) return(CLIPS_TRUE);
   
   return(CLIPS_FALSE);
  }

/*************************************************/
/* DribbleOffCommand:                            */
/*************************************************/
globle VOID DribbleOffCommand()
  {
   if (ArgCountCheck("dribble-off",EXACTLY,0) == -1) return;
   CloseDribble();
  }

/*************************************************/
/* CloseDribble:                                */
/*************************************************/
globle BOOLEAN CloseDribble()
  {   
   int rv = 0;
   
   if (DribbleStatusFunction != NULL)
     { (*DribbleStatusFunction)(CLIPS_FALSE); }
     
   if (DribbleFP != NULL)
     {    
      if (DribbleCurrentPosition > 0) 
        { fprintf(DribbleFP,"%s",DribbleBuffer); }
      DeleteRouter("dribble");
      if (fclose(DribbleFP) == 0) rv = 1;
     }
   else
     { rv = 1; }
 
   DribbleFP = NULL;
   
   if (DribbleBuffer != NULL)
     {
      rm(DribbleBuffer,DribbleMaximumPosition);
      DribbleBuffer = NULL;
     }
     
   DribbleCurrentPosition = 0;
   DribbleMaximumPosition = 0;
      
   return(rv);
  }
  
/*******************************/
/* SetDribbleStatusFunction */
/*******************************/
globle VOID SetDribbleStatusFunction(fnptr)
  int (*fnptr)();
  {
   DribbleStatusFunction = fnptr;
  }
#endif

/*******************/
/* BATCH FUNCTIONS */
/*******************/


/*************************************************/
/* FindBatch:                                   */
/*************************************************/
static int FindBatch(logicalName)
  char *logicalName;
  {
   if (strcmp(logicalName,"stdin") == 0)
     { return(CLIPS_TRUE); }

   return(CLIPS_FALSE);
  }

/*************************************************/
/* GetcBatch:                                   */
/*************************************************/
static int GetcBatch(logicalName)
  char *logicalName;
  {
   return(LLGetcBatch(logicalName,CLIPS_FALSE));
  }
  
/*************************************************/
/* LLGetcBatch:                                   */
/*************************************************/
globle int LLGetcBatch(logicalName,return_on_eof)
  char *logicalName;
  int return_on_eof;
  {
   int rv = EOF, flag = 1;
          
   /*=================================================*/
   /* Get a character until a valid character appears */
   /* or no more batch files are left.                */
   /*=================================================*/
   
   while ((rv == EOF) && (flag == 1))
     {
      if (BatchType == FILE_BATCH)
        { rv = getc((FILE *) BatchSource); }
      else
        { rv = GetcCLIPS((char *) BatchSource); }
        
      if (rv == EOF)
        {
         if (bat_loc > 0) PrintCLIPS("stdout",(char *) BatchBuffer);
         flag = RemoveBatch();
        }
     }

   if (rv == EOF)
     {
      if (bat_loc > 0) PrintCLIPS("stdout",(char *) BatchBuffer);
      DeleteRouter("batch");
      RemoveBatch();
      if (return_on_eof == CLIPS_TRUE)
        { return (EOF); }
      else
        { return(GetcCLIPS(logicalName)); }
     }

   BatchBuffer = ExpandStringWithChar((char) rv,BatchBuffer,&bat_loc,
                                      &bat_max,bat_max+BUFFSIZE);

   if (((char) rv == '\n') || ((char) rv == '\r'))
     {
      PrintCLIPS("stdout",(char *) BatchBuffer);
      bat_loc = 0;
      if ((BatchBuffer != NULL) && (bat_max > BUFFSIZE))
        {
         rm(BatchBuffer,bat_max);
         bat_max = 0;
         BatchBuffer = NULL;
        }
     }

   return(rv);
  }
    
/*************************************************/
/* UngetcBatch:                                  */
/*************************************************/
static int UngetcBatch(ch,logicalName)
  int ch;
  char *logicalName;
  {
#if MAC_MPW
#pragma unused(logicalName);
#endif
   if (bat_loc > 0) bat_loc--;
   if (BatchBuffer != NULL) BatchBuffer[bat_loc] = EOS;
   if (BatchType == FILE_BATCH)
     { return(ungetc(ch,(FILE *) BatchSource)); }
     
   return(UngetcCLIPS(ch,(char *) BatchSource));
  }

/*************************************************/
/* ExitBatch:                                    */
/*************************************************/
static VOID ExitBatch(num)
  int num;
  { 
#if MAC_MPW
#pragma unused(num);
#endif
   CloseAllBatchSources();
  }

/*************************************************/
/* BatchCommand:                                 */
/*************************************************/
globle VOID BatchCommand()
  {
   char *fileName;

   if (ArgCountCheck("batch",EXACTLY,1) == -1) return;
   if ((fileName = GetFileName("batch",1)) == NULL) return;
   
   OpenBatch(fileName,CLIPS_FALSE);
  }
  
/*************************************************/
/* OpenBatch:                                   */
/*************************************************/
globle int OpenBatch(fileName,placeAtEnd)
  char *fileName;
  int placeAtEnd;
  {
   FILE *temp_fp;
   
   temp_fp = fopen(fileName,"r");

   if (temp_fp == NULL)
     { 
      OpenErrorMessage("batch",fileName);
      return(0);
     }

   if (batch_top == NULL)
     {
      AddRouter("batch",           /* Router name     */
                 20,                /* Priority        */
                 FindBatch,        /* Query function  */
                 NULL,              /* Print function  */
                 GetcBatch,         /* Getc function   */
                 UngetcBatch,       /* Ungetc function */
                 (int (*)()) ExitBatch);       /* Exit function   */
     }

   AddBatch(placeAtEnd,(VOID *) temp_fp,FILE_BATCH,NULL);

   return(1);
  }

/*****************************************************************/
/* OpenStringBatch: Opens a string source for batch processing.  */
/*   The memory allocated for the argument stringName must be    */
/*   deallocated by the user. The memory allocated for theString */
/*   will be deallocated by the batch routines when batch        */
/*   processing for the  string is completed.                    */
/*****************************************************************/
globle int OpenStringBatch(stringName,theString,placeAtEnd)
  char *stringName;
  char *theString;
  int placeAtEnd;
  {   
   if (OpenStringSource(stringName,theString,0) == 0)
     { return(0); }

   if (batch_top == NULL)
     {
      AddRouter("batch",           /* Router name     */
                 20,                /* Priority        */
                 FindBatch,        /* Query function  */
                 NULL,              /* Print function  */
                 GetcBatch,         /* Getc function   */
                 UngetcBatch,       /* Ungetc function */
                 (int (*)()) ExitBatch);       /* Exit function   */
     }

   AddBatch(placeAtEnd,(VOID *) stringName,STRING_BATCH,theString);

   return(1);
  }
   
/***************************************************************/
/* AddBatch                                                   */
/***************************************************************/
static VOID AddBatch(placeAtEnd,theSource,type,theString)
  int placeAtEnd;
  VOID *theSource;
  int type;
  char *theString;
  {
   struct batchEntry *bptr;

   /*=========================*/
   /* Create the batch entry, */
   /*=========================*/
   
   bptr = get_struct(batchEntry);
   bptr->batchType = type;
   bptr->inputSource = theSource;
   bptr->theString = theString;
   bptr->next = NULL;
   
   if (batch_top == NULL)
     {
      batch_top = bptr;
      batch_bot = bptr;
      BatchType = type;
      BatchSource = theSource;
      bat_loc = 0;
     }
   else if (placeAtEnd == CLIPS_FALSE)
     {
      bptr->next = batch_top;
      batch_top = bptr;
      BatchType = type;
      BatchSource = theSource;
      bat_loc = 0;
     }
   else
     {
      batch_bot->next = bptr;
      batch_bot = bptr;
     }
  }

/***************************************************************/
/* RemoveBatch                                                */
/***************************************************************/
globle int RemoveBatch()
  {
   struct batchEntry *bptr;
   int rv;

   if (batch_top == NULL) return(0);
   
   if (batch_top->batchType == FILE_BATCH)
     { fclose((FILE *) batch_top->inputSource); }
   else
     { 
      CloseStringSource((char *) batch_top->inputSource); 
      rm(batch_top->theString,(int) strlen(batch_top->theString) + 1);
     }
   
   bptr = batch_top;
   batch_top = batch_top->next;
   
   rtn_struct(batchEntry,bptr);
   
   if (batch_top == NULL)
     { 
      batch_bot = NULL;
      BatchSource = NULL;
      if (BatchBuffer != NULL)
        {
         rm(BatchBuffer,bat_max);
         BatchBuffer = NULL;
        }
      bat_loc = 0;
      bat_max = 0;
      rv = 0;
     }
   else
     { 
      BatchType = batch_top->batchType;
      BatchSource = batch_top->inputSource; 
      bat_loc = 0;
      rv = 1;
     }
     
   return(rv);
  }

/*************************************************/
/* BatchActive: */
/*************************************************/
globle BOOLEAN BatchActive()
  {
   if (batch_top != NULL) return(CLIPS_TRUE);
   
   return(CLIPS_FALSE);
  }

/*************************************************/
/* CloseAllBatchSources: */
/*************************************************/
globle VOID CloseAllBatchSources()
  {  
   if (BatchBuffer != NULL)
     {
      if (bat_loc > 0) PrintCLIPS("stdout",(char *) BatchBuffer);
      rm(BatchBuffer,bat_max);
      BatchBuffer = NULL;
      bat_loc = 0;
      bat_max = 0;
     }
   
   DeleteRouter("batch");  
   while (RemoveBatch());
  }
  
/************************/
/* CRSV TRACE FUNCTIONS */
/************************/

#if DEBUGGING_FUNCTIONS
/*************************************************/
/* FindCRSVTrace:                                 */
/*************************************************/
static int FindCRSVTrace(logicalName)
  char *logicalName;
  {
   if (strcmp(logicalName,"wcrsv_tr") == 0) return(CLIPS_TRUE);
    
   return(CLIPS_FALSE);
  }

/*************************************************/
/* PrintCRSVTrace:                                */
/*************************************************/
static VOID PrintCRSVTrace(logicalName,str)
  char *logicalName, *str;
  {
#if MAC_MPW
#pragma unused(logicalName);
#endif
   fprintf(CRSVTraceFP,"%s",str);
  }

/*************************************************/
/* ExitCRSVTrace:                                 */
/*************************************************/
static VOID ExitCRSVTrace(num)
  int num;
  {
#if MAC_MPW
#pragma unused(num);
#endif
   if (CRSVTraceFP != NULL) fclose(CRSVTraceFP);
  }
  
/*************************************************/
/* CRSVTraceOnCommand:                                */
/*************************************************/
globle VOID CRSVTraceOnCommand()
  {
   char *fileName;

   if (ArgCountCheck("crsv-trace-on",EXACTLY,1) == -1) return;
   if ((fileName = GetFileName("crsv-trace-on",1)) == NULL)
     { return; }
   
   OpenCRSVTrace(fileName);
  }

/*************************************************/
/* OpenCRSVTrace:                                 */
/*************************************************/
globle BOOLEAN OpenCRSVTrace(fileName)
  char *fileName;
  {
   if (CRSVTraceFP != NULL)
     { CloseCRSVTrace(); }
 
   CRSVTraceFP = fopen(fileName,"w");
   if (CRSVTraceFP == NULL)
     {
      OpenErrorMessage("crsv_tr",fileName);     
      return(0);
     }

   AddRouter("crsv_tr",           /* Router name     */
              40,                  /* Priority        */
              FindCRSVTrace,       /* Query function  */
              (int (*)()) PrintCRSVTrace,      /* Print function  */
              NULL,                /* Getc function   */
              NULL,                /* Ungetc function */
              (int (*)()) ExitCRSVTrace        /* exit function   */
             );

   PrintCLIPS("wcrsv_tr","CRSV>");
   PrintCLIPS("wcrsv_tr","\n");
   
   return(1);
  }
  
/*************************************************/
/* CRSVTraceActive:                               */
/*************************************************/
globle BOOLEAN CRSVTraceActive()
  {
   if (CRSVTraceFP != NULL) return(CLIPS_TRUE);

   return(CLIPS_FALSE);
  }

/*************************************************/
/* CRSVTraceOffCommand:                                  */
/*************************************************/
globle VOID CRSVTraceOffCommand()
  {
   if (ArgCountCheck("crsv-trace-off",EXACTLY,0) == -1) 
     { return; }
  
   CloseCRSVTrace();
  }

/*************************************************/
/* CloseCRSVTrace:                                */
/*************************************************/
globle int CloseCRSVTrace()
  {
   if (CRSVTraceFP != NULL)
     { 
      DeleteRouter("crsv_tr");
      fclose(CRSVTraceFP);
     }
     
   CRSVTraceFP = NULL;
   
   return(1);
  }
#endif

/**************************/
/* LOAD AND SAVE COMMANDS */
/**************************/

/******************************************************************/
/* LoadCommand: parses the load command and calls LoadConstructs */
/*   which will load a set of constructs from a file.             */
/*   Syntax:  (load <file-name>)                                  */
/******************************************************************/
globle VOID LoadCommand()
  {
#if (! BLOAD_ONLY) && (! RUN_TIME)
   char *file_found;

   if (ArgCountCheck("load",EXACTLY,1) == -1) return;
   if ((file_found = GetFileName("load",1)) == NULL) return;
   
   SetPrintWhileLoading(CLIPS_TRUE);

   if (LoadConstructs(file_found) == -1) 
     {
      SetPrintWhileLoading(CLIPS_FALSE); 
      OpenErrorMessage("load",file_found);
      return; 
     }
   
   SetPrintWhileLoading(CLIPS_FALSE); 
   return;
#else
   PrintCLIPS(WDIALOG,"Load is not available in this environment\n");
   return;
#endif
  }

#if DEBUGGING_FUNCTIONS
/**********************************************/
/* SaveCommand:  Executes the save commands. */
/*   Syntax:  (save <file-name>)              */ 
/**********************************************/ 
globle VOID SaveCommand()
  {
#if (! BLOAD_ONLY) && (! RUN_TIME)
   char *file_found;

   if (ArgCountCheck("save",EXACTLY,1) == -1) return;
   if ((file_found = GetFileName("save",1)) == NULL) return;

   if (SaveConstructs(file_found) == CLIPS_FALSE) 
     {
      OpenErrorMessage("save",file_found);
      return;
     }
   
   return;
#else
   PrintCLIPS(WDIALOG,"Save is not available in this environment\n");
   return;
#endif
  }
#endif



