   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                  MATCH HEADER FILE                  */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_match

#define _H_match

struct patternMatch;
struct partialMatch;
struct alphaMatch;
struct multifieldMarker;

#include "engine.h"
#include "evaluatn.h"
#include "facts.h"
#include "network.h"

/************************************************************/
/* PATTERNMATCH STRUCTURE:                                  */
/************************************************************/
struct patternMatch
  {
   struct patternMatch *next;
   struct patternNode *matchingPattern;
  };

/************************************************************/
/* PARTIALMATCH STRUCTURE:                                  */
/************************************************************/
struct partialMatch
  {
   unsigned int betaMemory  :  1;
   unsigned int busy        :  1;
   unsigned int activationf :  1;
   unsigned int dependentsf :  1;
   unsigned int notOriginf  :  1;
   unsigned int counterf    :  1;
   unsigned int bcount      : 10;
   struct partialMatch *next;
   VOID *dependents;
   struct alphaMatch *binds[1];
  }; 

/************************************************************/
/* ALPHAMATCH STRUCTURE:                                    */
/************************************************************/
struct alphaMatch
  {
   struct alphaMatch *aorigin;
   long int whoset;
   struct fact *origin;
   struct multifieldMarker *marker;
  };

/************************************************************/
/* MULTIFIELDMARKER STRUCTURE: Used in the pattern matching */
/* process to mark the range of fields that the $? and      */
/* $?variables match because a single pattern restriction   */
/* may span zero or more fields of a fact.                  */
/************************************************************/
struct multifieldMarker
   {
    int element;
    int start;                                   
    int end;
    struct multifieldMarker *next;
   };
   
#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _MATCH_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
   VOID                           PatternMatch(struct fact *,struct element *,
                                               struct patternNode *,int,int,
                                               struct multifieldMarker *,
                                               struct multifieldMarker *);
   VOID                           GetFieldSysFunction(struct dataObject *);
   VOID                           PatternNetErrorMessage(struct patternNode *);
   VOID                           TraceErrorToRule(struct joinNode *,char *); 
#else
   VOID                           PatternMatch();
   VOID                           GetFieldSysFunction();
   VOID                           PatternNetErrorMessage();
   VOID                           TraceErrorToRule();  
#endif 

#endif






