   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                OBJECT MESSAGE COMMANDS              */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#endif
#include "classcom.h"
#include "classfun.h"
#include "clipsmem.h"
#include "constant.h"
#include "constrct.h"
#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif
#include "evaluatn.h"
#include "facts.h"
#include "generate.h"
#include "insfun.h"
#include "msgfun.h"
#include "multivar.h"
#include "router.h"
#include "scanner.h"
#include "parsutil.h"
#include "spclform.h"
#include "utility.h"

#define _MSGCOM_SOURCE_
#include "msgcom.h"

#if (! BLOAD_ONLY) && (! RUN_TIME)
extern struct token ObjectParseToken;
#endif

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
#define SELF_STRING     "self"
#define SELF_LEN         4
#define SELF_SLOT_REF   ':'
#define PRINT_STRING    "print"

/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER

#if (! BLOAD_ONLY) && (! RUN_TIME)
static int ParseDefmessageHandler(char *);
static VOID NewSystemHandler(char *,char *,char *,int,struct FunctionDefinition *);
static HANDLER *InsertHandlerHeader(CLASS_TYPE *,SYMBOL_HN *,int);
static EXPRESSION *ParseHandlerParameters(char *,SYMBOL_HN **,struct token *);
static int CheckHandlerBindList(void);
static int ReplaceHandlerParameters(EXPRESSION *,EXPRESSION *,SYMBOL_HN *);
static int FindHandlerParameter(SYMBOL_HN *,EXPRESSION *);
#endif

#else

#if (! BLOAD_ONLY) && (! RUN_TIME)
static int ParseDefmessageHandler();
static VOID NewSystemHandler();
static HANDLER *InsertHandlerHeader();
static EXPRESSION *ParseHandlerParameters();
static int CheckHandlerBindList();
static int ReplaceHandlerParameters();
static int FindHandlerParameter();
#endif

#endif
      
/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle SYMBOL_HN *INIT_SYMBOL = NULL,
globle           *DELETE_SYMBOL = NULL;

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
#if (! BLOAD_ONLY) && (! RUN_TIME)
static SYMBOL_HN *SELF_SYMBOL = NULL;
#endif

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
globle VOID SetupMessageHandlers()
  {
#if ! RUN_TIME  
   INIT_SYMBOL = (SYMBOL_HN *) AddSymbol(INIT_STRING);
   IncrementSymbolCount(INIT_SYMBOL);
   
   DELETE_SYMBOL = (SYMBOL_HN *) AddSymbol(DELETE_STRING);
   IncrementSymbolCount(DELETE_SYMBOL);

#if ! BLOAD_ONLY
   SELF_SYMBOL = (SYMBOL_HN *) AddSymbol(SELF_STRING);
   IncrementSymbolCount(SELF_SYMBOL);
  
   AddConstruct("defmessage-handler",ParseDefmessageHandler);
   DefineFunction("undefmessage-handler",'v',PTIF CmdUndefmessageHandler,
                  "CmdUndefmessageHandler");
   AddSystemHandlers();
#endif

   DefineFunction("send",'u',PTIF DispatchMessage,"DispatchMessage");
   
#if DEBUGGING_FUNCTIONS
   DefineFunction("preview-send",'v',PTIF PreviewMessageCmd,"PreviewMessageCmd");
   
   DefineFunction("ppdefmessage-handler",'v',PTIF PPDefmessageHandler,"PPDefmessageHandler");
   DefineFunction("list-defmessage-handlers",'v',PTIF CmdListDefmessageHandlers,
                  "CmdListDefmessageHandlers");
#endif

   DefineFunction("next-handlerp",'b',PTIF NextHandlerAvailable,"NextHandlerAvailable");
   DefineFunction("call-next-handler",'u',
                  PTIF CallNextHandler,"CallNextHandler");
   DefineFunction("override-next-handler",'u',
                  PTIF CallNextHandler,"CallNextHandler");
   
   DefineFunction("(hndunknown)",'u',PTIF HandlerRtnUnknown,"HandlerRtnUnknown");
   DefineFunction("(hndwildargs)",'u',PTIF HandlerWildargs,"HandlerWildargs");
   DefineFunction("(hndgetbind)",'u',PTIF HandlerGetBind,"HandlerGetBind");
   DefineFunction("get",'u',PTIF HandlerGetSlot,"HandlerGetSlot");
   DefineFunction("put",'b',PTIF HandlerPutSlot,"HandlerPutSlot");
   DefineFunction("direct-mv-replace",'b',PTIF HandlerReplaceSlot,"HandlerReplaceSlot");
   DefineFunction("direct-mv-insert",'b',PTIF HandlerInsertSlot,"HandlerInsertSlot");
   DefineFunction("direct-mv-delete",'b',PTIF HandlerDeleteSlot,"HandlerDeleteSlot");
#endif
   
#if DEBUGGING_FUNCTIONS
   AddWatchItem("messages",&WatchMessages,36);
   AddWatchItem("message-handlers",&WatchHandlers,35);
#endif
  }
   
/*****************************************************
  NAME         : GetDefmessageHandlerName
  DESCRIPTION  : Gets the name of a message-handler
  INPUTS       : 1) Pointer to a class
                 2) Array index of handler in class's
                    message-handler array (+1)
  RETURNS      : Name-string of message-handler
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
char *GetDefmessageHandlerName(ptr,index)
  VOID *ptr;
  unsigned index;
  {
   return(ValueToString(((CLASS_TYPE *) ptr)->handlers[index-1].name));
  }
  
/*****************************************************
  NAME         : GetDefmessageHandlerType
  DESCRIPTION  : Gets the type of a message-handler
  INPUTS       : 1) Pointer to a class
                 2) Array index of handler in class's
                    message-handler array (+1)
  RETURNS      : Type-string of message-handler
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
globle char *GetDefmessageHandlerType(ptr,index)
  VOID *ptr;
  unsigned index;
  {
   return(hndquals[((CLASS_TYPE *) ptr)->handlers[index-1].type]);
  }
  
/**************************************************************
  NAME         : GetNextDefmessageHandler
  DESCRIPTION  : Finds first or next handler for a class
  INPUTS       : 1) The address of the handler's class
                 2) The array index of the current handler (+1)
  RETURNS      : The array index (+1) of the next handler, or 0
                   if there is none
  SIDE EFFECTS : None
  NOTES        : If index == 0, the first handler array index
                 (i.e. 1) returned
 **************************************************************/
globle unsigned GetNextDefmessageHandler(ptr,index)
  VOID *ptr;
  unsigned index;
  {
   CLASS_TYPE *cls;
   
   cls = (CLASS_TYPE *) ptr;
   if (index == 0)
     return((cls->handlers != NULL) ? 1 : 0);
   if (index == cls->handler_cnt)
     return(0); 
   return(index+1);
  }
  
/***************************************************
  NAME         : FindDefmessageHandler
  DESCRIPTION  : Determines the index of a specfied
                  message-handler
  INPUTS       : 1) A pointer to the class
                 2) Name-string of the handler
                 3) Handler-type: "around","before",
                    "primary", or "after"
  RETURNS      : The index of the handler
                   (0 if not found)
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle unsigned FindDefmessageHandler(ptr,hname,htypestr)
  VOID *ptr;
  char *hname,*htypestr;
  {
   int htype;
   SYMBOL_HN *hsym;
   CLASS_TYPE *cls;
   int index;
   
   htype = HandlerType("handler-lookup",htypestr);
   if (htype == MERROR)
     return(0);
   hsym = FindSymbol(hname);
   if (hsym == NULL)
     return(0);
   cls = (CLASS_TYPE *) ptr;
   index = (int) FindHandler(cls,hsym,(unsigned) htype,LOOKUP_HANDLER_INDEX);
   return((unsigned) (index+1));
  }

/***************************************************
  NAME         : IsDefmessageHandlerDeletable
  DESCRIPTION  : Determines if a message-handler
                   can be deleted
  INPUTS       : 1) Address of the handler's class
                 2) Index of the handler
  RETURNS      : TRUE if deletable, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
globle int IsDefmessageHandlerDeletable(ptr,index)
  VOID *ptr;
  unsigned index;
  {
#if BLOAD_ONLY || RUN_TIME
   return(FALSE);
#else
   CLASS_TYPE *cls;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     return(FALSE);
#endif
   cls = (CLASS_TYPE *) ptr;
   if (cls->handlers[index-1].system == 1)
     return(FALSE);
   return((HandlersExecuting(cls) == FALSE) ? TRUE : FALSE);
#endif
  }

/******************************************************************************
  NAME         : CmdUndefmessageHandler
  DESCRIPTION  : Deletes a handler from a class
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Handler deleted if possible
  NOTES        : CLIPS Syntax: (undefmessage-handler <class> <handler> [<type>])
 ******************************************************************************/
globle VOID CmdUndefmessageHandler()
  {
   int argno;
   char *mname,*tname;
   DATA_OBJECT tmp;
   CLASS_TYPE *cls;

   argno = ArgRangeCheck("undefmessage-handler",2,3);
   if (argno == -1)
     return;
   if (ArgTypeCheck("undefmessage-handler",1,SYMBOL,&tmp) == FALSE)
     return;
   cls = FindDefclassBySymbol((SYMBOL_HN *) tmp.value);
   if ((cls == NULL) ? (strcmp(DOToString(tmp),"*") != 0) : FALSE)
     {
      ClassExistError("undefmessage-handler",DOToString(tmp));
      return;
     }
   if (ArgTypeCheck("undefmessage-handler",2,SYMBOL,&tmp) == FALSE)
     return;
   mname = DOToString(tmp);
   if (strcmp(mname,"*") == 0)
     mname = NULL;
   if (argno == 3)
     {
      if (ArgTypeCheck("undefmessage-handler",3,SYMBOL,&tmp) == FALSE)
        return;
      tname = DOToString(tmp);
      if (strcmp(tname,"*") == 0)
        tname = NULL;
     }
   else
     tname = hndquals[MPRIMARY];
   WildDeleteHandler((VOID *) cls,mname,tname);
  }
  
/****************************************************************
  NAME         : HandlerRtnUnknown
  DESCRIPTION  : COOL internal function for accessing the
                   CurrentMessageFrame
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Caller's result <==> CurrentMessageFrame element
  NOTES        : CLIPS Syntax: ((hndunknown) <index>)
                 Assumes index is within range
 ****************************************************************/
globle VOID HandlerRtnUnknown(result)
  DATA_OBJECT *result;
  {
   CopyMemory(DATA_OBJECT,1,result,
                            &CurrentMessageFrame[(DOPToInteger(GetFirstArgument())) - 1]);
  }
  
/*********************************************************
  NAME         : HandlerWildargs
  DESCRIPTION  : Groups handler arguments from specified
                  index to end or "tail" into a
                  multi-field varaible
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Segment allocated and caller's result set
  NOTES        : CLIPS Syntax: ((hndwildargs) <index>)
                 Assumes index is within range
 *********************************************************/
globle VOID HandlerWildargs(result)
  DATA_OBJECT *result;
  {
   GroupHandlerWildargs(result,DOPToInteger(GetFirstArgument()));
  }
  
  
/*******************************************************************
  NAME         : HandlerGetBind
  DESCRIPTION  : COOL internal function for accessing the
                   handler's bound variables and CurrentMessageFrame
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Caller's result <==> bound variable or 
                                      CurrentMessageFrame element
  NOTES        : CLIPS Syntax: ((hndgetbind) <var-name> <index>)
                 Assumes index is within range
 *******************************************************************/
globle VOID HandlerGetBind(result)
  DATA_OBJECT *result;
  {
   DATA_OBJECT *bind_ptr;
   SYMBOL_HN *var_name;
   EXPRESSION *test_ptr;
   int index;
   
   /*============================================*/
   /* Get the name of the variable being sought. */
   /*============================================*/

   test_ptr = GetFirstArgument();
   var_name = (SYMBOL_HN *) test_ptr->value;

   /*===============================================*/
   /* Search for the variable in the list of binds. */
   /*===============================================*/

   bind_ptr = BindList;
   while (bind_ptr != NULL)
     {
      if (bind_ptr->name == var_name)
        {
         CopyMemory(DATA_OBJECT,1,result,bind_ptr);
         return; 
        }

      bind_ptr = bind_ptr->next;
     }

   /*=====================================================*/
   /* If the variable was not found in the list of binds, */
   /* determine if the variable is only bound on the rhs  */
   /* of the rule.                                        */
   /*=====================================================*/
  
   test_ptr = test_ptr->next_arg;
   index = ValueToInteger(test_ptr->value);
   if (index == 0)
     {    
      SetEvaluationError(TRUE);
      PrintCLIPS(WERROR,"Variable ");
      PrintCLIPS(WERROR,ValueToString(var_name));
      PrintCLIPS(WERROR," unbound in message-handler ");
      PrintCurrentMessage(WERROR); 
      result->type = SYMBOL;
      result->value = (VOID *) CLIPSFalseSymbol;
      return;
     }

   /*=============================================================*/
   /* Variable was bound in the argument list of the deffunction. */
   /*=============================================================*/
   if (index > 0)
     CopyMemory(DATA_OBJECT,1,result,&CurrentMessageFrame[ValueToInteger(test_ptr->value) - 1]);
   else
     GroupHandlerWildargs(result,-index);
  }
  
/***************************************************
  NAME         : HandlerGetSlot
  DESCRIPTION  : Directly references a slot's value
  INPUTS       : The caller's result buffer
  RETURNS      : Nothing useful
  SIDE EFFECTS : Caller's result buffer set
  NOTES        : CLIPS Syntax: (get <slot>)
 ***************************************************/
globle VOID HandlerGetSlot(result)
  DATA_OBJECT *result;
  {
   INSTANCE_SLOT *sp;
   INSTANCE_TYPE *ins;
   DATA_OBJECT temp;
   
   result->type = SYMBOL;
   result->value = (VOID *) CLIPSFalseSymbol;
   if ((GetFirstArgument() != NULL) ? (GetFirstArgument()->next_arg != NULL) : TRUE)
     {
      ExpectedCountError("get",EXACTLY,1);
      SetEvaluationError(TRUE);
      return;
     }
   if (CheckCurrentMessage("get",TRUE) == FALSE)
     return;
   EvaluateExpression(GetFirstArgument(),&temp);
   if (temp.type != SYMBOL)
     {
      ExpectedTypeError("get",1,"symbol");
      SetEvaluationError(TRUE);
      return;
     }
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   sp = FindInstanceSlot(ins,(SYMBOL_HN *) temp.value);
   if (sp == NULL)
     {
      SlotExistError(ValueToString(temp.value),"get");
      return;
     }
   if (*(sp->valaddr) != NULL)
     CopyDataObject(result,*(sp->valaddr));
   else
     {
      PrintCLIPS(WERROR,ValueToString(sp->desc->name));
      PrintCLIPS(WERROR," slot unbound in instance ");
      PrintCLIPS(WERROR,ValueToString(ins->name));
      PrintCLIPS(WERROR,".\n");
      SetEvaluationError(TRUE);
     }
  }
  
/****************************************************
  NAME         : HandlerPutSlot
  DESCRIPTION  : Directly puts a slot's value
  INPUTS       : None
  RETURNS      : TRUE if put OK, FALSE otherwise
  SIDE EFFECTS : Slot modified
  NOTES        : CLIPS Syntax: (put <slot> <value>*)
 ***************************************************/
globle int HandlerPutSlot()
  {
   INSTANCE_SLOT *sp;
   INSTANCE_TYPE *ins;
   DATA_OBJECT temp;
   register int ertn;
   
   if (GetFirstArgument() == NULL)
     {
      ExpectedCountError("put",AT_LEAST,1);
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   if (CheckCurrentMessage("put",TRUE) == FALSE)
     return(FALSE);
   EvaluateExpression(GetFirstArgument(),&temp);
   if (temp.type != SYMBOL)
     {
      ExpectedTypeError("put",1,"symbol");
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   sp = FindInstanceSlot(ins,(SYMBOL_HN *) temp.value);
   if (sp == NULL)
     {
      SlotExistError(ValueToString(temp.value),"put");
      return(FALSE);
     }
   if ((sp->desc->nowrite == 0) ? FALSE :
       ((sp->desc->initonly == 0) || (WithinInit == FALSE)))
     {
      PrintCLIPS(WERROR,ValueToString(sp->desc->name));
      PrintCLIPS(WERROR," slot in instance ");
      PrintCLIPS(WERROR,ValueToString(ins->name));
      PrintCLIPS(WERROR,": write access denied.\n");
      SetEvaluationError(TRUE);
      return(FALSE);
     }
   if ((ertn = EvaluateAndStoreInDataObject(sp->desc->multiple,
                          GetFirstArgument()->next_arg,&temp)) == FALSE)
     return(FALSE);
   if (ValidSlotValue(&temp,ValueToString(CurrentMessageName)) == FALSE)
     return(FALSE);
   if (PutSlotValue(ins,sp,(ertn == MULTI_CLEAR) ? NULL : &temp,TRUE) == SLOT_ERROR)
     return(FALSE);
   return(TRUE);
  }

/*****************************************************************
  NAME         : HandlerReplaceSlot
  DESCRIPTION  : Directly replaces a slot's value
  INPUTS       : None
  RETURNS      : TRUE if put OK, FALSE otherwise
  SIDE EFFECTS : Slot modified
  NOTES        : CLIPS Syntax: (direct-mv-replace <slot> 
                                <range-begin> <range-end> <value>)
 *****************************************************************/
globle int HandlerReplaceSlot()
  {
   INSTANCE_SLOT *sp;
   INSTANCE_TYPE *ins;
   int rb,re;
   DATA_OBJECT newval,newseg;
   
   if (CheckCurrentMessage("direct-mv-replace",TRUE) == FALSE)
     return(FALSE);
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   if (ArgCountCheck("direct-mv-replace",AT_LEAST,4) == -1)
     return(FALSE);
   sp = CheckMultifieldSlotModify(REPLACE,"direct-mv-replace",ins,
                            GetFirstArgument(),&rb,&re,&newval);
   if (sp == NULL)
     return(FALSE);
   if (ReplaceMultiValueField(&newseg,*(sp->valaddr),
                              rb,re,&newval,"direct-mv-replace") == FALSE)
     return(FALSE);
   if (PutSlotValue(ins,sp,&newseg,TRUE) == SLOT_ERROR)
     return(FALSE);
   return(TRUE);
  }

/************************************************************************
  NAME         : HandlerInsertSlot
  DESCRIPTION  : Directly inserts a slot's value
  INPUTS       : None
  RETURNS      : TRUE if put OK, FALSE otherwise
  SIDE EFFECTS : Slot modified
  NOTES        : CLIPS Syntax: (direct-mv-insert <slot> <index> <value>)
 ************************************************************************/
globle int HandlerInsertSlot()
  {
   INSTANCE_SLOT *sp;
   INSTANCE_TYPE *ins;
   int index;
   DATA_OBJECT newval,newseg;
   
   if (CheckCurrentMessage("direct-mv-insert",TRUE) == FALSE)
     return(FALSE);
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   if (ArgCountCheck("direct-mv-insert",AT_LEAST,3) == -1)
     return(FALSE);
   sp = CheckMultifieldSlotModify(INSERT,"direct-mv-insert",ins,
                            GetFirstArgument(),&index,NULL,&newval);
   if (sp == NULL)
     return(FALSE);
   if (InsertMultiValueField(&newseg,*(sp->valaddr),index,
                             &newval,"direct-mv-insert") == FALSE)
     return(FALSE);
   if (PutSlotValue(ins,sp,&newseg,TRUE) == SLOT_ERROR)
     return(FALSE);
   return(TRUE);
  }

/*****************************************************************
  NAME         : HandlerDeleteSlot
  DESCRIPTION  : Directly deletes a slot's value
  INPUTS       : None
  RETURNS      : TRUE if put OK, FALSE otherwise
  SIDE EFFECTS : Slot modified
  NOTES        : CLIPS Syntax: (direct-mv-delete <slot> 
                                <range-begin> <range-end>)
 *****************************************************************/
globle int HandlerDeleteSlot()
  {
   INSTANCE_SLOT *sp;
   INSTANCE_TYPE *ins;
   int rb,re;
   DATA_OBJECT newseg;
   
   if (CheckCurrentMessage("direct-mv-delete",TRUE) == FALSE)
     return(FALSE);
   ins = (INSTANCE_TYPE *) CurrentMessageFrame->value;
   if (ArgCountCheck("direct-mv-delete",EXACTLY,3) == -1)
     return(FALSE);
   sp = CheckMultifieldSlotModify(DELETE,"direct-mv-delete",ins,GetFirstArgument(),&rb,&re,NULL);
   if (sp == NULL)
     return(FALSE);
   if (DeleteMultiValueField(&newseg,*(sp->valaddr),rb,re,"direct-mv-delete") == FALSE)
     return(FALSE);
   if (PutSlotValue(ins,sp,&newseg,TRUE) == SLOT_ERROR)
     return(FALSE);
   return(TRUE);
  }

#if (! BLOAD_ONLY) && (! RUN_TIME)

/*****************************************************
  NAME         : AddSystemHandlers
  DESCRIPTION  : Creates the built-in system class
                   system handlers
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Handler arrays allocated and sorted
  NOTES        : CLIPS System Error is generated if
                   CLIPS Kernel function (hndunknown)
                   could  not be found
 *****************************************************/
globle VOID AddSystemHandlers()
  {
   struct FunctionDefinition *clru;
   
   clru = FindFunction("(hndunknown)");     
   NewSystemHandler(USER_CLASS_STRING,INIT_STRING,"init-slots",FALSE,clru);
   NewSystemHandler(USER_CLASS_STRING,DELETE_STRING,"delete-instance",FALSE,clru);
#if DEBUGGING_FUNCTIONS
   NewSystemHandler(USER_CLASS_STRING,PRINT_STRING,"ppinstance",FALSE,clru);
#endif
  }
  
/**************************************************************
  NAME         : CheckHandlerAgainstSlots
  DESCRIPTION  : Insures that a handler does not conflict
                   with the implicit handlers (slot-acessors)
                   defined for a class
  INPUTS       : 1) The class
                 2) The handler name
                 3) The handler type
  RETURNS      : TRUE if handler does not conflict
                   with slot-accessors, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : The handler is checked only against slots
                   defined in the class (not inherited slots) -
                   conflicts on inheritance will be handled as
                   shadowed handlers.
 **************************************************************/
globle int CheckHandlerAgainstSlots(cls,mname,mtype)
  CLASS_TYPE *cls;
  SYMBOL_HN *mname;
  int mtype;
  {
   register int i;
   
   /* ==================================================
      All slot-accessors are treated as primary handlers
      ================================================== */
   if (mtype != MPRIMARY)
     return(TRUE);
   
   for (i = 0 ; i < cls->slot_cnt ; i++)
     if ((mname == cls->slots[i].gethandler) || (mname == cls->slots[i].puthandler))
       {
        PrintCLIPS(WERROR,ValueToString(mname));
        PrintCLIPS(WERROR," message-handler conflicts with an accessor for slot ");
        PrintCLIPS(WERROR,ValueToString(cls->slots[i].name));
        PrintCLIPS(WERROR,".\n");
        return(FALSE);
       }
   return(TRUE);
  }
  
#endif
  
/************************************************************
  NAME         : WildDeleteHandler
  DESCRIPTION  : Deletes a handler from a class
  INPUTS       : 1) Class address    (Can be NULL)
                 2) Message Handler Name (Can be NULL)
                 3) Type name ("primary", etc.) (Can be NULL)
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Handler deleted if possible
  NOTES        : None
 ************************************************************/
globle int WildDeleteHandler(vptr,mname,tname)
  VOID *vptr;
  char *mname,*tname;
  {
#if RUN_TIME || BLOAD_ONLY
   PrintCLIPS(WERROR,"Unable to delete message-handlers.\n");
   return(0);
#else
   CLASS_TYPE *cls = NULL;
   int mtype;
   SYMBOL_HN *msym;
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Unable to delete message-handlers.\n");
      return(0);
     }
#endif
   cls = (CLASS_TYPE *) vptr;
   msym = (SYMBOL_HN *) AddSymbol((mname != NULL) ? mname : "*");
   if (tname != NULL)
     {
      mtype = HandlerType("undefmessage-handler",tname);
      if (mtype == MERROR)
        return(0);
     }
   else
     mtype = -1;
   if (cls == NULL)
     {
      int success = 1;
     
      for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
        if (DeleteHandler(cls,msym,mtype,FALSE) == 0)
          success = 0;
      return(success);
     }
   return(DeleteHandler(cls,msym,mtype,TRUE));
#endif
  }

/*********************************************************
  NAME         : GroupHandlerWildargs
  DESCRIPTION  : Groups handler arguments from specified
                  index to end or "tail" into a
                  multi-field variable
  INPUTS       : 1) The caller's result buffer
                 2) The starting index
  RETURNS      : Nothing useful
  SIDE EFFECTS : Segment allocated and caller's result set
  NOTES        : None
 *********************************************************/
globle VOID GroupHandlerWildargs(result,index)
  DATA_OBJECT *result;
  int index;
  {
   register int i,j,k;
   int size;
   struct fact *fptr;

   result->type = MULTIFIELD;
   result->begin = 0;
   result->end = -1;
   if (CurrentMessageSize-index < 0)
     {
      result->value = (VOID *) CreateMultifield(0);
      return;
     }
   for (i = index-1 , size = 0 ; i < CurrentMessageSize ; i++)
     {
      if (CurrentMessageFrame[i].type != MULTIFIELD)
        size++;
      else
        size += CurrentMessageFrame[i].end - CurrentMessageFrame[i].begin + 1;
     }
   result->end = size-1;
   result->value = (VOID *) CreateMultifield(size);
   fptr = (struct fact *) result->value;
   for (i = index-1 , j = 0 ; i < CurrentMessageSize ; i++)
     {
      if (CurrentMessageFrame[i].type != MULTIFIELD)
        {
         fptr->atoms[j].type = CurrentMessageFrame[i].type; 
         fptr->atoms[j].value = CurrentMessageFrame[i].value; 
         j++;
        }
      else
        {
         for (k = CurrentMessageFrame[i].begin ;
              k <= CurrentMessageFrame[i].end ; k++ , j++)
           {
            fptr->atoms[j].type = 
                     ((struct fact *) CurrentMessageFrame[i].value)->atoms[k].type;
            fptr->atoms[j].value = 
                     ((struct fact *) CurrentMessageFrame[i].value)->atoms[k].value;
           }         
        }
     }
  }

/***********************************************************
  NAME         : DeleteDefmessageHandler
  DESCRIPTION  : Deletes a handler from a class
  INPUTS       : 1) Class address    (Can be NULL)
                 2) Handler index (can be 0)
  RETURNS      : 1 if successful, 0 otherwise
  SIDE EFFECTS : Handler deleted if possible
  NOTES        : None
 ***********************************************************/
globle int DeleteDefmessageHandler(vptr,mhi)
  VOID *vptr;
  unsigned mhi;
  {
#if RUN_TIME || BLOAD_ONLY
   PrintCLIPS(WERROR,"Unable to delete message-handlers.\n");
   return(0);
#else
   CLASS_TYPE *cls;

#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Unable to delete message-handlers.\n");
      return(0);
     }
#endif
   if (vptr == NULL)
     {
      if (mhi != 0)
        {
         PrintCLIPS(WERROR,"Incomplete message-handler specification for deletion.\n");
         return(0);
        }
      return(WildDeleteHandler(NULL,NULL,NULL));
     }
   if (mhi == 0)
     return(WildDeleteHandler(vptr,NULL,NULL));
   cls = (CLASS_TYPE *) vptr;
   if (HandlersExecuting(cls))
     {
      HandlerDeleteError(ValueToString(cls->name));
      return(0);
     }
   cls->handlers[mhi-1].mark = 1;
   DeallocateMarkedHandlers(cls);
   return(1);
#endif
  }
  
#if DEBUGGING_FUNCTIONS

/*******************************************************************************
  NAME         : PPDefmessageHandler
  DESCRIPTION  : Displays the pretty-print form (if any) for a handler
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax: (ppdefmessage-handler <class> <message> [<type>])
 *******************************************************************************/
globle VOID PPDefmessageHandler()
  {
   int argno;
   DATA_OBJECT temp;
   SYMBOL_HN *csym,*msym;
   char *tname;
   CLASS_TYPE *cls = NULL;
   int mtype;
   HANDLER *hnd;
   
   argno = ArgRangeCheck("ppdefmessage-handler",2,3);
   if (argno == -1)
     return;
   if (ArgTypeCheck("ppdefmessage-handler",1,SYMBOL,&temp) == FALSE)
     return;
   csym = FindSymbol(DOToString(temp));
   if (ArgTypeCheck("ppdefmessage-handler",2,SYMBOL,&temp) == FALSE)
     return;
   msym = FindSymbol(DOToString(temp));
   if (argno == 3)
     {
      if (ArgTypeCheck("ppdefmessage-handler",3,SYMBOL,&temp) == FALSE)
        return;
      tname = DOToString(temp);
     }
   else
     tname = hndquals[MPRIMARY];
   mtype = HandlerType("ppdefmessage-handler",tname);
   if (mtype == MERROR)
     {
      SetEvaluationError(TRUE);
      return;
     }
   if (csym != NULL)
     cls = FindDefclassBySymbol(csym);
   if (((cls == NULL) || (msym == NULL)) ? TRUE :
       ((hnd = (HANDLER *) FindHandler(cls,msym,(unsigned) mtype,
                                       LOOKUP_HANDLER_ADDRESS)) == NULL))
     {
      PrintCLIPS(WERROR,"Unable to find message-handler ");
      PrintCLIPS(WERROR,ValueToString(msym));
      PrintCLIPS(WERROR," ");
      PrintCLIPS(WERROR,tname);
      PrintCLIPS(WERROR," for class ");
      PrintCLIPS(WERROR,ValueToString(csym));
      PrintCLIPS(WERROR," in function ppdefmessage-handler.\n");
      SetEvaluationError(TRUE);
      return;
     }
   if (hnd->pp_form != NULL)
     PrintCLIPS(WDISPLAY,hnd->pp_form);
  }

/*******************************************************************
  NAME         : CmdListDefmessageHandlers
  DESCRIPTION  : Depending on arguments, does lists handlers which
                   match restrictions
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : CLIPS Syntax: (list-defmessage-handlers [<args>])
                   wher <args> ::= <class> | <name> | <type> |
                                   <class> <name> | <class> <type> |
                                   <name> <type>
 *******************************************************************/
globle VOID CmdListDefmessageHandlers()
  {
   int argno;
   DATA_OBJECT temp;
   char *mname,*tname;
   CLASS_TYPE *cls = NULL;
   
   argno = ArgCountCheck("list-defmessage-handlers",NO_MORE_THAN,3);
   if (argno == -1)
     return;
   if (argno > 0)
     {
      if (ArgTypeCheck("list-defmessage-handlers",1,SYMBOL,&temp) == FALSE)
        return;
      cls = (CLASS_TYPE *) FindDefclass(DOToString(temp));
      if ((cls == NULL) ? (strcmp(DOToString(temp),"*") != 0) : FALSE)
        {
         ClassExistError("list-defmessage-handlers",DOToString(temp));
         return;
        }
     }
   if (argno > 1)
     {
      if (ArgTypeCheck("list-defmessage-handlers",2,SYMBOL,&temp) == FALSE)
        return;
      mname = DOToString(temp);
     }
   else
     mname = NULL;
   if (argno > 2)
     {
      if (ArgTypeCheck("list-defmessage-handlers",3,SYMBOL,&temp) == FALSE)
        return;
      tname = DOToString(temp);
      if (strcmp(tname,"*") == 0)
        tname = NULL;
     }
   else
     tname = NULL;
   ListDefmessageHandlers((VOID *) cls,mname,tname);
  }
  
/********************************************************************
  NAME         : PreviewMessageCmd
  DESCRIPTION  : Displays a list of the core for a message describing
                   shadows,etc.
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Temporary core created and destroyed
  NOTES        : CLIPS Syntax: (preview-send <class> <msg>)
 ********************************************************************/
globle VOID PreviewMessageCmd()
  {
   CLASS_TYPE *cls;
   DATA_OBJECT temp;
   
   if (ArgCountCheck("preview-send",EXACTLY,2) == -1)
     return;
     
   /* ============================= 
      Get the class for the message
      ============================= */
   if (ArgTypeCheck("preview-send",1,SYMBOL,&temp) == FALSE)
     return;
   cls = FindDefclassBySymbol((SYMBOL_HN *) temp.value);
   if (cls == NULL)
     {
      ClassExistError("preview-send",ValueToString(temp.value));
      return;
     }
     
   if (ArgTypeCheck("preview-send",2,SYMBOL,&temp) == FALSE)
     return;
   PreviewMessage((VOID *) cls,DOToString(temp));
  }
  
/********************************************************
  NAME         : GetDefmessageHandlerPPForm
  DESCRIPTION  : Gets a message-handler pretty print form
  INPUTS       : 1) Address of the handler's class
                 2) Index of the handler
  RETURNS      : TRUE if printable, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ********************************************************/
globle char *GetDefmessageHandlerPPForm(ptr,index)
  VOID *ptr;
  unsigned index;
  {
   return(((CLASS_TYPE *) ptr)->handlers[index-1].pp_form);
  }

/*******************************************************************
  NAME         : ListDefmessageHandlers
  DESCRIPTION  : Depending on arguments, does lists handlers which
                   match restrictions
  INPUTS       : 1) Class name (Can be NULL)
                 2) Handler name (Can be NULL)
                 3) Type Name (Can be NULL)
  RETURNS      : Nothing useful
  SIDE EFFECTS : None
  NOTES        : None
 *******************************************************************/
globle VOID ListDefmessageHandlers(vptr,mname,tname)
  VOID *vptr;
  char *mname,*tname;
  {
   int mtype;
   CLASS_TYPE *cls,*cptr;
   SYMBOL_HN *msym;
   HANDLER *mptr;
   register int i;
   long cnt;
   
   cls = (CLASS_TYPE *) vptr;
   if (mname != NULL)
     {
      msym = FindSymbol(mname);
      if (msym == NULL)
        return;
     }
   else
     msym = NULL;
   if (tname != NULL)
     {
      mtype = HandlerType("list-defmessage-handlers",tname);
      if (mtype == MERROR)
        return;
     }
   else
     mtype = -1;
   cnt = 0L;
   for (cptr = ClassList ; cptr != NULL ; cptr = cptr->nxt_lst)
     {
      if ((cptr == cls) || (cls == NULL))
        {
         for (i = 0 ; i < cptr->handler_cnt ; i++)
           {
            mptr = &cptr->handlers[i];
            if ((mtype == -1) || (mtype == mptr->type))
              {
               if (((msym == NULL) || (msym == mptr->name)) ? 1 :
                    ((strcmp(msym->contents,"*") == 0) && 
                     (FindHandlerNameGroup(cptr,msym) == -1)))
                 {
                  PrintHandler(WDISPLAY,ValueToString(cptr->name),
                               ValueToString(mptr->name),hndquals[mptr->type]);
                  cnt++;
                 }
              }
           }
        }
     }
   PrintTally(WDISPLAY,cnt,"message-handler","message-handlers");
  }
  
/********************************************************************
  NAME         : PreviewMessage
  DESCRIPTION  : Displays a list of the core for a message describing
                   shadows,etc.
  INPUTS       : 1) Class pointer
                 2) Message name-string
  RETURNS      : Nothing useful
  SIDE EFFECTS : Temporary core created and destroyed
  NOTES        : None
 ********************************************************************/
globle VOID PreviewMessage(clsptr,msgname)
  VOID *clsptr;
  char *msgname;
  {
   HANDLER_LINK *core;
   SYMBOL_HN *msym;
   
   msym = FindSymbol(msgname);
   if (msym == NULL)
     return;
   core = FindPreviewApplicableHandlers((CLASS_TYPE *) clsptr,msym);
   if (core != NULL)
     {
      DisplayCore(core,0);
      DestroyHandlerLinks(core);
     }
  }

#endif

/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
  
#if (! BLOAD_ONLY) && (! RUN_TIME)

/***********************************************************************
  NAME         : ParseDefmessageHandler
  DESCRIPTION  : Parses a message-handler for a class of objects
  INPUTS       : The logical name of the input source
  RETURNS      : FALSE if successful parse, TRUE otherwise
  SIDE EFFECTS : Handler allocated and inserted into class
  NOTES        : CLIPS Syntax:
  
                 (defmessage-handler <class> <name> [<type>] [<comment>]
                    (<params>)
                    <action>*)
                    
                 <params> ::= <var>* | <var>* $?<name>
 ***********************************************************************/
static int ParseDefmessageHandler(read_source)
  char *read_source;
  {
   CLASS_TYPE *cls;
   SYMBOL_HN *mname,*wildcard;
   unsigned mtype = MPRIMARY,pcnt;
   EXPRESSION *hndParams,*actions;
   HANDLER *hnd;
   
   SetPPBufferStatus(ON);
   FlushPPBuffer();          
   SetIndentDepth(3);    
   SavePPBuffer("(defmessage-handler ");
   
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded())
     {
      PrintCLIPS(WERROR,"Cannot load defmessage-handlers while binary load is in effect.\n");
      return(TRUE);
     }
#endif
   GetToken(read_source,&ObjectParseToken);
   if (GetType(ObjectParseToken) != SYMBOL)
     {
      SyntaxErrorMessage("defmessage-handler");
      return(TRUE);
     }
   SavePPBuffer(" ");
   cls = FindDefclassBySymbol((SYMBOL_HN *) GetValue(ObjectParseToken));
   if (cls == NULL)
     {
      PrintCLIPS(WERROR,"A class must be defined before its message-handlers.\n");
      return(TRUE);
     }
   if (HandlersExecuting(cls))
     {
      PrintCLIPS(WERROR,"Cannot (re)define message-handlers during execution of \n");
      PrintCLIPS(WERROR,"  other message-handlers for the same class.\n");
      return(TRUE);
     }
   GetToken(read_source,&ObjectParseToken);
   if (GetType(ObjectParseToken) != SYMBOL)
     {
      SyntaxErrorMessage("defmessage-handler");
      return(TRUE);
     }
   SavePPBuffer(" ");
   mname = (SYMBOL_HN *) GetValue(ObjectParseToken);
   GetToken(read_source,&ObjectParseToken);
   if (GetType(ObjectParseToken) != LPAREN)
     {
      SavePPBuffer(" ");
      if (GetType(ObjectParseToken) != STRING)
        {
         if (GetType(ObjectParseToken) != SYMBOL)
           {
            SyntaxErrorMessage("defmessage-handler");
            return(TRUE);
           }
         mtype = HandlerType("defmessage-handler",DOToString(ObjectParseToken));
         if (mtype == MERROR)
           return(TRUE);
         GetToken(read_source,&ObjectParseToken);
         if (GetType(ObjectParseToken) == STRING)
           {
            SavePPBuffer(" ");
            GetToken(read_source,&ObjectParseToken);
           }
        }
      else
        {
         SavePPBuffer(" ");
         GetToken(read_source,&ObjectParseToken);
        }
     }
   PPBackup();
   PPBackup();
   PPCRAndIndent();
   SavePPBuffer(ObjectParseToken.print_rep);

   hnd = (HANDLER *) FindHandler(cls,mname,mtype,LOOKUP_HANDLER_ADDRESS);
   if (GetPrintWhileLoading())
     {
      if (GetCompilationsWatch())
        {
         PrintCLIPS(WDIALOG,(hnd != NULL) ? "Redefining" : "Defining");
         PrintCLIPS(WDIALOG," defmessage-handler ");
         PrintCLIPS(WDIALOG,ValueToString(mname));
         PrintCLIPS(WDIALOG," ");
         PrintCLIPS(WDIALOG,hndquals[mtype]);
         PrintCLIPS(WDIALOG," in class ");
         PrintCLIPS(WDIALOG,ValueToString(cls->name));
         PrintCLIPS(WDIALOG,".\n");
        }
      else
        PrintCLIPS(WDIALOG,"~");
     }
   
   if (hnd != NULL)
     {
      if (hnd->system == 1)
        {
         PrintCLIPS(WERROR,"System message-handlers may not be modified.\n");
         return(TRUE);
        }
     }
   else if (CheckHandlerAgainstSlots(cls,mname,mtype) == FALSE)
     return(TRUE);
      
   hndParams = ParseHandlerParameters(read_source,&wildcard,&ObjectParseToken);
   if (hndParams == NULL)
     return(TRUE);
   PPCRAndIndent();
   actions = GroupActions(read_source,&ObjectParseToken,TRUE,NULL);
   if (actions == NULL)
     {
      ReturnExpression(hndParams);
      return(TRUE);
     }
   actions = CompactActions(actions);
   if (GetType(ObjectParseToken) != RPAREN)
     {
      SyntaxErrorMessage("defmessage-handler");
      ReturnExpression(hndParams);
      ReturnExpression(actions);
      return(TRUE);
     }
   PPBackup();
   PPBackup();
   SavePPBuffer(")\n");

   pcnt = CountArguments(hndParams);
   if (ReplaceHandlerParameters(actions,hndParams,wildcard) == TRUE)
     {
      ReturnExpression(hndParams);
      ReturnExpression(actions);
      return(TRUE);
     }
   if (CheckHandlerBindList() == FALSE)
     {
      ReturnExpression(hndParams);
      ReturnExpression(actions);
      return(TRUE);
     }
   if (hnd != NULL)
     {
      ExpressionDeinstall(hnd->actions);
      ReturnPackedExpression(hnd->actions);
      if (hnd->pp_form != NULL)
        rm((VOID *) hnd->pp_form,(int) (sizeof(char) * (strlen(hnd->pp_form)+1)));
     }
   else
     {
      hnd = InsertHandlerHeader(cls,mname,mtype);
      IncrementSymbolCount(hnd->name);
     }
   ReturnExpression(hndParams);
   hndParams = actions;
   actions = PackExpression(actions);
   ReturnExpression(hndParams);
   
   if (wildcard == NULL)
     hnd->maxParams = hnd->minParams = pcnt;
   else
     {
      hnd->minParams = pcnt-1;
      hnd->maxParams = -1;
     }
   hnd->actions = actions;
   ExpressionInstall(hnd->actions);
#if DEBUGGING_FUNCTIONS
   if (GetConserveMemory() == FALSE)
     hnd->pp_form = CopyPPBuffer();
   else
#endif
     hnd->pp_form = NULL;
   return(FALSE);
  }

/******************************************************************************
  NAME         : NewSystemHandler
  DESCRIPTION  : Adds a new system handler for a system class
  
                 The handler is assumed to be primary and of
                 the form:
                 
                 (defmessage-handler <class> <handler> () (<func>))
                 
                 OR
                 
                 (defmessage-handler <class <handler> (<param>) (<func> <param>))
                 
  INPUTS       : 1) Name-string of the system class
                 2) Name-string of the system handler
                 3) Name-string of the internal CLIPS function to implement
                      this handler
                 4) Flag indicating whether there is an extra parameter or not
                 5) Address of internal (cl_runknown) CLIPS Kernel function
  RETURNS      : Nothing useful
  SIDE EFFECTS : Creates the new handler and inserts it in the system class's
                   handler array
                 On errors, generate a system error and exits.
  NOTES        : Does not check to see if handler already exists
 *******************************************************************************/
static VOID NewSystemHandler(cname,mname,fname,param,clru)
  char *cname,*mname,*fname;
  int param;
  struct FunctionDefinition *clru;
  {
   CLASS_TYPE *cls;
   HANDLER *hnd;
   struct FunctionDefinition *mfunc;
   
   cls = FindDefclassBySymbol((SYMBOL_HN *) AddSymbol(cname));
   mfunc = FindFunction(fname);
   hnd = InsertHandlerHeader(cls,(SYMBOL_HN *) AddSymbol(mname),MPRIMARY);
   IncrementSymbolCount(hnd->name);
   hnd->system = 1;
   if (param == TRUE)
     {
      hnd->minParams = hnd->maxParams = 2;
      hnd->actions = (EXPRESSION *) gm2((int) (sizeof(EXPRESSION) * 3));
      hnd->actions[0].arg_list = &hnd->actions[1];
      hnd->actions[1].type = FCALL;
      hnd->actions[1].value = (VOID *) clru;
      hnd->actions[1].arg_list = &hnd->actions[2];
      hnd->actions[1].next_arg = NULL;
      hnd->actions[2].type = INTEGER;
      hnd->actions[2].value = (VOID *) AddLong(2L);
      IncrementIntegerCount((INTEGER_HN *) hnd->actions[2].value);
      hnd->actions[2].arg_list = NULL;
      hnd->actions[2].next_arg = NULL;
     }
   else
     {
      hnd->minParams = hnd->maxParams = 1;
      hnd->actions = get_struct(expr);
      hnd->actions->arg_list = NULL;
     }
   hnd->actions->type = FCALL;
   hnd->actions->value = (VOID *) mfunc;
   hnd->actions->next_arg = NULL;
  }
  
/***************************************************
  NAME         : InsertHandlerHeader
  DESCRIPTION  : Allocates a new handler header and
                   inserts it in the proper (sorted)
                   position in the class hnd array
  INPUTS       : 1) The class
                 2) The handler name
                 3) The handler type
  RETURNS      : The address of the new handler
                   header, NULL on errors
  SIDE EFFECTS : Class handler array reallocated
                   and resorted
  NOTES        : Assumes handler does not exist
 ***************************************************/
static HANDLER *InsertHandlerHeader(cls,mname,mtype)
  CLASS_TYPE *cls;
  SYMBOL_HN *mname;
  int mtype;
  {
   HANDLER *nhnd,*hnd;
   unsigned *narr,*arr;
   register int i,j,ni = -1;
   
   hnd = cls->handlers;
   arr = cls->hordered;
   nhnd = (HANDLER *) gm2((int) (sizeof(HANDLER) * (cls->handler_cnt+1)));
   narr = (unsigned *) gm2((int) (sizeof(unsigned) * (cls->handler_cnt+1)));
   CopyMemory(HANDLER,cls->handler_cnt,nhnd,hnd);
   for (i = 0 , j = 0 ; i < cls->handler_cnt ; i++ , j++)
     {
      if (ni == -1)
        {
         if ((hnd[arr[i]].name->bucket > mname->bucket) ? TRUE :
             (hnd[arr[i]].name == mname))
           {
            ni = i;
            j++;
           }
        }
      narr[j] = arr[i];
     }
   if (ni == -1)
     ni = cls->handler_cnt;
   narr[ni] = cls->handler_cnt;
   nhnd[cls->handler_cnt].system = 0;
   nhnd[cls->handler_cnt].type = mtype;
   nhnd[cls->handler_cnt].busy = 0;
   nhnd[cls->handler_cnt].mark = 0;
   nhnd[cls->handler_cnt].name = mname;
   nhnd[cls->handler_cnt].cls = cls;
   nhnd[cls->handler_cnt].minParams = 0;
   nhnd[cls->handler_cnt].maxParams = 0;
   nhnd[cls->handler_cnt].actions = NULL;
   nhnd[cls->handler_cnt].pp_form = NULL;
   if (cls->handler_cnt != 0)
     {
      rm((VOID *) hnd,(int) (sizeof(HANDLER) * cls->handler_cnt));
      rm((VOID *) arr,(int) (sizeof(unsigned) * cls->handler_cnt));
     }
   cls->handlers = nhnd;
   cls->hordered = narr;
   cls->handler_cnt++;
   return(&nhnd[cls->handler_cnt-1]);
  }
  
/*****************************************************************
  NAME         : ParseHandlerParameters
  DESCRIPTION  : Forms parameter expression list for handler
                 ?self is added first
                 Duplicate parameters are not allowed
                 Parameters of the form ?self:<name> are not
                   allowed since this is slot reference syntax
  INPUTS       : 1) The logical name of the input source
                 2) Caller's buffer for a wildcard symbol (if any)
                 3) Caller's buffer for scanner token
  RETURNS      : The address of the parameter list, NULL on errors
  SIDE EFFECTS : Expressions allocated for parameter list
                 Caller's error flag set on errors
  NOTES        : ?self parameter prepended to parameter list
 *****************************************************************/
static EXPRESSION *ParseHandlerParameters(read_source,wildcard,itkn)
  char *read_source;
  SYMBOL_HN **wildcard;
  struct token *itkn;
  {
   EXPRESSION *parameterList = NULL, *nextOne, *lastOne = NULL;
   
   *wildcard = NULL;
   if (itkn->type != LPAREN)
     {
      SyntaxErrorMessage("defmessage-handler");
      return(NULL);
     }    
   
   parameterList = lastOne = GenConstant(SYMBOL,(VOID *) SELF_SYMBOL);
   GetToken(read_source,itkn);
   
   while ((itkn->type == BWORD) || (itkn->type == BWORDS))
     {
      if (*wildcard != NULL)
        {
         PrintCLIPS(WERROR,"No parameters allowed after wildcard parameter.\n");
         ReturnExpression(parameterList);
         return(NULL);
        }
      for (nextOne = parameterList ; nextOne != NULL ; nextOne = nextOne->next_arg)
        {
         if (itkn->value == nextOne->value)
           {
            PrintCLIPS(WERROR,"Duplicate parameter names not allowed.\n");
            ReturnExpression(parameterList);
            return(NULL);
           }
         else if ((strncmp(ValueToString(itkn->value),SELF_STRING,SELF_LEN) == 0) ?
                  (ValueToString(itkn->value)[SELF_LEN] == SELF_SLOT_REF) : FALSE)
           {
            PrintCLIPS(WERROR,"Illegal slot reference in parameter list.\n");
            ReturnExpression(parameterList);
            return(NULL);
           }
        }
      nextOne = GenConstant(itkn->type,itkn->value);
      if (itkn->type == BWORDS)
        *wildcard = (SYMBOL_HN *) itkn->value;
      lastOne->next_arg = nextOne;
      lastOne = nextOne;
      SavePPBuffer(" ");
      GetToken(read_source,itkn);
     }  
   
   if (itkn->type != RPAREN)
     {
      SyntaxErrorMessage("defmessage-handler");
      ReturnExpression(parameterList);
      return(NULL);
     }
     
   if (parameterList->next_arg != NULL)
     {
      PPBackup();
      PPBackup();
      SavePPBuffer(")");
     }
   
   return(parameterList);
  }
  
/****************************************************
  NAME         : CheckHandlerBindList
  DESCRIPTION  : Insures that there are no variables
                   of the form ?self or ?self:<name>
                   in a handler's bind list
  INPUTS       : None
  RETURNS      : TRUE if all OK, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : Uses bind_names_list from VARIABLE.C
 ****************************************************/
static int CheckHandlerBindList()
  {
   struct BindInfo *tmp;
   
   for (tmp = GetParsedBindNames() ; tmp != NULL ; tmp = tmp->next)
     {
      if (tmp->name == SELF_SYMBOL)
        {
         PrintCLIPS(WERROR,"Active instance parameter cannot be changed.\n");
         return(FALSE);
        }
      else if ((strncmp(ValueToString(tmp->name),SELF_STRING,SELF_LEN) == 0) ?
               (ValueToString(tmp->name)[SELF_LEN] == SELF_SLOT_REF) : FALSE)
        {
         PrintCLIPS(WERROR,"Direct slot references cannot be changed.\n");
         return(FALSE);
        }
     }
   return(TRUE);
  }
  
/***************************************************************************
  NAME         : ReplaceHandlerParameters
  DESCRIPTION  : Replaces variable references in handler body
                   with internal function calls to reference
                   elements of the CurrentMessageFrame at run-time
                 Parameter references directly reference CurrentMessageFrame
                   via the function (hndunknown)
                 The special wildcard parameter reference pulls the "tail"
                   of the CurrentMessageFrame into a multi-field variable
                 References to bound variables can be either parameters
                   or bound (since the replacement does not occur until after
                   all the binds were parsed - but in execution, the variable
                   may not actually be bound yet) - thus another function
                   (hndgetbind) gets the reference from the bind list
                   (higher priority) then goes to the CurrentMessageFrame
                 Slot references of the form ?self:<name> directly reference
                   CurrentMessageFrame[0] (the message object - ?self) to
                   find the specified slot at run-time via get
  INPUTS       : 1) The handler body expressions
                 2) The parameter list
                 3) The wildcard parameter symbol (NULL if none)
  RETURNS      : TRUE on ERRORS, FALSE if everything is OK
  SIDE EFFECTS : Handler body BWORD and BWORDS replaced with internal
                   function call expressions
  NOTES        : Objects are allowed to directly access their own slots
                 without sending a message to themselves.  Since the object
                 is "within the boundary of its internals", this does not
                 violate the encapsulation principle of OOP.
 ***************************************************************************/
static int ReplaceHandlerParameters(actions,hndParams,wildcard)
  EXPRESSION *actions,*hndParams;
  SYMBOL_HN *wildcard;
  {
   int position,oldpp;
   BOOLEAN boundAlso;
   EXPRESSION *arg_lvl;
   SYMBOL_HN *bindName;
   VOID *fparam,*fbind,*fwild,*fslot;
   
   fparam = (VOID *) FindFunction("(hndunknown)");
   fbind = (VOID *) FindFunction("(hndgetbind)");
   fwild = (VOID *) FindFunction("(hndwildargs)");
   fslot = (VOID *) FindFunction("get");
   while (actions != NULL)
     {
      if ((actions->type == BWORD) || (actions->type == BWORDS))
        {
         /*===============================================*/
         /* See if the variable is in the parameter list. */
         /*===============================================*/
          
         bindName = (SYMBOL_HN *) actions->value;
         position = FindHandlerParameter(bindName,hndParams);
           
         /*=================================================================*/
         /* Check to see if the variable is bound within the deffunction.   */
         /*=================================================================*/
   
         boundAlso = SearchParsedBindNames(bindName);
         
         /*=============================================*/
         /* If variable is not defined in the parameter */
         /* list or as part of a bind action then...    */
         /*=============================================*/
         
         if ((position == 0) && (boundAlso == FALSE))
           {
            /* ======================================
               Variable is not a parameter or bound -
               Check to see if it is a slot reference
               ====================================== */
            if ((strncmp(ValueToString(bindName),SELF_STRING,SELF_LEN) == 0) ?
                (ValueToString(bindName)[SELF_LEN] == SELF_SLOT_REF) : FALSE)
              {
               struct token itkn;
                             
               /* ===============================================
                  Allow slot-specifier to be a variable reference
                  =============================================== */
               OpenStringSource("hnd-var",ValueToString(bindName) + (SELF_LEN+1),0);
               oldpp = GetPPBufferStatus();
               SetPPBufferStatus(OFF);
               GetToken("hnd-var",&itkn);
               SetPPBufferStatus(oldpp);
               CloseStringSource("hnd-var");
               if (itkn.type == STOP)
                 {
                  PrintCLIPS(WERROR,"\nUndefined variable ");
                  PrintCLIPS(WERROR,ValueToString(bindName));        
                  PrintCLIPS(WERROR," referenced within defmessage-handler.\n");
                  return(TRUE);
                 }
               actions->type = FCALL;
               actions->value = fslot;
               actions->arg_list = GenConstant(itkn.type,itkn.value);
              }
            else
              {
               PrintCLIPS(WERROR,"\nUndefined variable ");
               PrintCLIPS(WERROR,ValueToString(bindName));        
               PrintCLIPS(WERROR," referenced within defmessage-handler.\n");
               return(TRUE);
              }
           }
           
         /*===================================================*/
         /* Else if variable is defined in the parameter list */
         /* and not rebound within the handler then...        */
         /*===================================================*/
         
         else if ((position > 0) && (boundAlso == FALSE))
           {
            actions->type = FCALL;
            if (bindName != wildcard)
              actions->value = fparam;
            else
              actions->value = fwild;
            actions->arg_list = GenConstant(INTEGER,(VOID *) AddLong((long) position));
           }
           
         /*=======================================================*/
         /* Else the variable is rebound within the handler so... */
         /*=======================================================*/
         
         else 
           {
            actions->type = FCALL;
            actions->value = fbind;
            arg_lvl = GenConstant(SYMBOL,(VOID *) bindName);
            actions->arg_list = arg_lvl;
            if (bindName == wildcard)
              position = -position;
            arg_lvl->next_arg = GenConstant(INTEGER,(VOID *) AddLong((long) position));
           }
        }
        
#if DEFGLOBAL_CONSTRUCT
      else if (actions->type == GBWORD)
        { if (ReplaceGlobalVariable(actions) == FALSE) return(TRUE); }
#endif

      if (actions->arg_list != NULL)
        {
         if (ReplaceHandlerParameters(actions->arg_list,hndParams,wildcard) == TRUE)
           return(TRUE);
        }
           
      actions = actions->next_arg;
     }
   return(FALSE);
  }
  
/*******************************************************
  NAME         : FindHandlerParameter
  DESCRIPTION  : Determines posn of variable name in
                   parameter list - used for indices
                   into static CurrentMessageFrame at
                   run-time
  INPUTS       : 1) The name sought
                 2) The parameter expression list
  RETURNS      : The index from 1..n of the name in
                   the parameter list, or 0 if not found
  SIDE EFFECTS : None
  NOTES        : None
 *******************************************************/
static int FindHandlerParameter(bname,plist)
  SYMBOL_HN *bname;
  EXPRESSION *plist;
  {
   register int i;
   
   for (i = 1 ; plist != NULL ; i++ , plist = plist->next_arg)
     if (bname == (SYMBOL_HN *) plist->value)
       return(i);
   return(0);
  }
  
#endif

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
