   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 NETWORK HEADER FILE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_network

#define _H_network

struct patternNode;
struct joinNode;

#include "expressn.h"
#include "match.h"

/*******************************************************************/
/* patternNode:                                                    */
/* Used to join together the tree of elements which constitute     */
/* the pattern network.  For a given node, last_level points to    */
/* the previous element in the pattern, same_level points to other */
/* elements which may fill this slot in the pattern, and next      */
/* level points to the elements which may fill the next position   */
/* assuming that this element is matched.  Check_list holds the    */
/* actual pattern element (e.g. (?, blue | red , ~3, $?, end of    */
/* pattern)).  If the element to be matched is the end of the      */
/* pattern, then var_list will contain the list of variable        */ 
/* bindings, functions call, and predicate tests which must be     */
/* made for the pattern.                                           */
/*******************************************************************/
struct patternNode
  {
   unsigned int singlefieldNode : 1;
   unsigned int multifieldNode : 1;
   unsigned int stopNode : 1;
   unsigned int blocked : 1; 
   unsigned int initialize : 1;
   unsigned int marked : 1;
   unsigned long bsaveID;
   struct expr *eval;
   struct partialMatch *alpha;
   struct joinNode *entry_join;
   struct patternNode *next_level;
   struct patternNode *last_level;
   struct patternNode *left_node;
   struct patternNode *right_node;
  };

/************************************************************/
/* joinNode structure:                                      */
/************************************************************/
struct joinNode
  { 
   unsigned int firstJoin : 1;
   unsigned int lastJoin : 1;
   unsigned int LHSPatternIsNegated : 1;
   unsigned int initialize : 1;
   unsigned int marked : 1;
   unsigned long bsaveID;
   struct partialMatch *beta;        
   struct expr *eval;
   struct expr *not_eval;
   long int id;               
   short int depth;
   struct patternNode *entry_pat;
   struct joinNode *next_level;
   struct joinNode *last_level;
   struct joinNode *right_drive_node;
   struct joinNode *right_match_node;
  };

#endif




