   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                OBJECT SYSTEM DEFINITIONS            */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_object
#define _H_object

typedef struct cls CLASS_TYPE;
typedef struct class_link CLASS_LINK;
typedef struct slot_desc SLOT_DESC;
typedef struct msg_handler HANDLER;
typedef struct instance INSTANCE_TYPE;
typedef struct instance_slot INSTANCE_SLOT;

#ifndef _CLIPS_STDIO_
#define _CLIPS_STDIO_
#include <stdio.h>
#endif
   
#if ANSI_COMPILER
#include <string.h>
#endif

#define BYTE_MASK        0xFF
#define BITS_PER_BYTE    8

/* Maximum # of simultaneous class hierarchy traversals
   should be a multiple of 8 and less than MAX_INT      */

#define MAX_TRAVERSALS  128
#define TRAVERSAL_BYTES (MAX_TRAVERSALS / BITS_PER_BYTE)

#define testbit(n,b)   ((n) & (1 << (b)))
#define setbit(n,b)    (n |= (1 << (b)))
#define clearbit(n,b)  (n &= ~(1 << (b)))

#include "evaluatn.h"
#include "expressn.h"

struct cls
  {
   unsigned installed      : 1;
   unsigned system         : 1;
   unsigned abstract       : 1;
   unsigned primitive      : 1;
   unsigned primitive_type : 5;
   unsigned busy,hashval,predecessors;
   SYMBOL_HN *name;
   CLASS_LINK *superlink,*sublink,*precedence;
   SLOT_DESC *slots,**itemplate;
   unsigned *sordered,slot_cnt,islot_cnt;
   INSTANCE_TYPE *instances;
   HANDLER *handlers;
   unsigned *hordered,handler_cnt;
   CLASS_TYPE *prv_hsh,*nxt_hsh,
              *prv_lst,*nxt_lst;
   long bsaveIndex;
   char *pp_form;
   char *tvids;
  };
  
struct class_link
  {
   CLASS_TYPE *cls;
   struct class_link *nxt;
  };

struct slot_desc
  {
   unsigned default_specified     : 1;
   unsigned cardinality_specified : 1;
   unsigned storage_specified     : 1;
   unsigned access_specified      : 1;
   unsigned shared                : 1;
   unsigned multiple              : 1;
   unsigned composite             : 1;
   unsigned noinherit             : 1;
   unsigned nowrite               : 1;
   unsigned initonly              : 1;
   unsigned override              : 1;
   unsigned dynamic               : 1;
   CLASS_TYPE *cls;
   SYMBOL_HN *name,
             *gethandler,
             *puthandler;
   EXPRESSION *defexp;
   DATA_OBJECT *share_val;
   unsigned share_cnt;
  };

struct msg_handler
  {
   unsigned system : 1;
   unsigned type   : 2;
   unsigned mark   : 1;
   unsigned busy;
   SYMBOL_HN *name;
   CLASS_TYPE *cls;
   int minParams,maxParams;
   EXPRESSION *actions;
   char *pp_form;
  };

struct instance
  {
   unsigned installed : 1;
   unsigned garbage   : 1;
   unsigned init_eval : 1;
   SYMBOL_HN *name;
   unsigned busy;
   int depth;
   unsigned hashval;
   CLASS_TYPE *cls;
   INSTANCE_SLOT *slots;
   INSTANCE_TYPE *prv_cls,*nxt_cls,
                 *prv_hsh,*nxt_hsh,
                 *prv_lst,*nxt_lst;
  };
 
struct instance_slot
  {
   SLOT_DESC *desc;
   DATA_OBJECT **valaddr;
   DATA_OBJECT *val;
   EXPRESSION *valexp;
  };

#endif






