   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*               PARSER UTILITIES MODULE               */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Chris Culbert                                        */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _PARSUTIL_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "expressn.h"
#include "router.h"
#include "constrct.h"
#include "generate.h"
#include "parsutil.h"
#include "utility.h"

#if DEFTEMPLATE_CONSTRUCT
#include "deftempl.h"
#include "deftmfun.h"
#endif
  
#if (! RUN_TIME) && (! BLOAD_ONLY)

/**************************************************************************/
/* ReadUntilClosingParen: Skips over tokens until a ')' is encountered.   */
/* Expects first token after opening parenthesis has already been scanned */
/**************************************************************************/
globle BOOLEAN ReadUntilClosingParen(readSource,inputToken)
  char *readSource;
  struct token *inputToken;
  {
   int cnt = 1,lparen_read = CLIPS_FALSE;
   
   do
     {
      if (lparen_read == CLIPS_FALSE)
        SavePPBuffer(" ");
      GetToken(readSource,inputToken);
      if (inputToken->type == STOP)
        {
         PrintCLIPS(WERROR,"\nExpected ')' to end definition\n");
         return(0);
        }
      else if (inputToken->type == LPAREN)
        {
         lparen_read = CLIPS_TRUE;
         cnt++;
        }
      else if (inputToken->type == RPAREN)
        {
         cnt--;
         if (lparen_read == CLIPS_FALSE)
           {
            PPBackup();
            PPBackup();
            SavePPBuffer(")");
           }
         lparen_read = CLIPS_FALSE;
        }
      else
        lparen_read = CLIPS_FALSE;
     }
   while (cnt > 0);
     
   return(1);
  }
   
/******************************************************************/
/* GetConstructNameAndComment:  Get the name and comment field of */
/*   a construct. Returns name of the construct if no errors are  */
/*   detected, otherwise returns NULL.                            */
/******************************************************************/
globle SYMBOL_HN *GetConstructNameAndComment(readSource,inputToken,constructName,
                                             findFunction,deleteFunction,
                                             constructSymbol,fullMessageCR)
  char *readSource;
  struct token *inputToken;
  char *constructName;
#if ANSI_COMPILER
  VOID *(*findFunction)(char *);
  int (*deleteFunction)(VOID *);
#else
  VOID *(*findFunction)();
  int (*deleteFunction)();
#endif
  char *constructSymbol;
  int fullMessageCR;
  {
   SYMBOL_HN *name;
   int redefining = CLIPS_FALSE;
   VOID *theConstruct;
   
   /*============================================*/
   /* Next token should be the name of the rule. */
   /*============================================*/   

   GetToken(readSource,inputToken);                   
   if (inputToken->type != SYMBOL)
     { 
      PrintCLIPS(WERROR,"\nMissing name for ");
      PrintCLIPS(WERROR,constructName);
      PrintCLIPS(WERROR," construct\n");
      return(NULL);
     }
     
   name = (SYMBOL_HN *) inputToken->value;
 
   /*==============================================================*/
   /* Remove the construct if it is already in the knowledge base. */
   /*==============================================================*/

   if (findFunction != NULL)
     {
      theConstruct = (*findFunction)(ValueToString(name));
      if (theConstruct != NULL) 
        {
         if (deleteFunction != NULL)
           { redefining = (*deleteFunction)(theConstruct); }
         else
           { redefining = CLIPS_TRUE; }
        }
     }

   /*===============================================================*/
   /* If rules being watched, indicate that rule is being compiled. */
   /*===============================================================*/

   if ((GetWatchItem("compilations") == CLIPS_TRUE) && GetPrintWhileLoading())
     {
      if (redefining) PrintCLIPS(WDIALOG,"Redefining ");
      else PrintCLIPS(WDIALOG,"Defining ");
      
      PrintCLIPS(WDIALOG,constructName);
      PrintCLIPS(WDIALOG,": ");
      PrintCLIPS(WDIALOG,ValueToString(name));
      
      if (fullMessageCR) PrintCLIPS(WDIALOG,"\n");
      else PrintCLIPS(WDIALOG," ");
     }
   else if (GetPrintWhileLoading())
     { PrintCLIPS(WDIALOG,constructSymbol); }
     
   /*===========================*/
   /* Get comment if it exists. */
   /*===========================*/

   GetToken(readSource,inputToken);
   if (inputToken->type == STRING)
     {
      PPBackup();
      SavePPBuffer(" ");
      SavePPBuffer(inputToken->print_rep);
      GetToken(readSource,inputToken);
      if (inputToken->type != RPAREN)
        {
         PPBackup();
         SavePPBuffer("\n   "); 
         SavePPBuffer(inputToken->print_rep);
        }          
     }
   else if (inputToken->type != RPAREN)                            
     {
      PPBackup();
      SavePPBuffer("\n   ");
      SavePPBuffer(inputToken->print_rep);
     }

   return(name);
  }

#endif

#if (! RUN_TIME)

/****************************************************************/
/* GroupActions: Returns progn test with actions if successful. */
/*   Returns Null if error detected.                            */
/****************************************************************/
globle struct expr *GroupActions(in_file,rtn_tkn,readFirstToken,endWord)
  char *in_file;
  struct token *rtn_tkn;
  int readFirstToken;
  char *endWord;
  {
   struct expr *top, *next_one, *last_one;
   
   /*========================================================*/
   /* Continue until all appropriate commands are processed. */
   /*========================================================*/

   top = GenConstant(FCALL,FindFunction("progn"));

   last_one = NULL;  

   while (CLIPS_TRUE)
     {
      if (readFirstToken)
        { GetToken(in_file,rtn_tkn); }
      else
        { readFirstToken = CLIPS_TRUE; }
        
      if ((rtn_tkn->type == SYMBOL) && (endWord != NULL))
        {
         if (strcmp(ValueToString(rtn_tkn->value),endWord) == 0) return(top);
        }
        
      if ((rtn_tkn->type == SYMBOL) || (rtn_tkn->type == STRING) ||
          (rtn_tkn->type == INTEGER) || (rtn_tkn->type == FLOAT) ||
#if DEFGLOBAL_CONSTRUCT
          (rtn_tkn->type == GBWORD) ||
#endif
#if OBJECT_SYSTEM
          (rtn_tkn->type == INSTANCE_NAME) ||
#endif
          (rtn_tkn->type == BWORD) || (rtn_tkn->type == BWORDS))
        { next_one = GenConstant(rtn_tkn->type,rtn_tkn->value); }
      else if (rtn_tkn->type == LPAREN)
        { next_one = Function1Parse(in_file); }
      else
        { return(top); }
       
      if (next_one == NULL)
        {
         rtn_tkn->type = KUNKNOWN;
         ReturnExpression(top); 
         return(NULL);
        }        
         
      if (last_one == NULL)
        { top->arg_list = next_one; }
      else
        { last_one->next_arg = next_one; }

      last_one = next_one;

      PPCRAndIndent();
     }
  }

/*************************************************************************/
/* CompactActions: Examines a progn expression chain, and if there       */
/*   is only one action, the progn header is deallocated and the         */
/*   action is returned.  If there are 0 actions, the progn expression   */
/*   is modified to be the CLIPS FALSE symbol and returned.  Otherwise   */
/*   the progn is simply returned.                                       */
/*   Assumes actions is a progn expression and actions->next_arg == NULL */
/*************************************************************************/
globle struct expr *CompactActions(actions)
  struct expr *actions;
  {
   register struct expr *tmp;
   
   if (actions->arg_list == NULL)
     {
      actions->type = SYMBOL;
      actions->value = (VOID *) CLIPSFalseSymbol;
     }
   else if (actions->arg_list->next_arg == NULL)
     {
      tmp = actions;
      actions = actions->arg_list;
      rtn_struct(expr,tmp);
     }
   return(actions);
  }
 
  
/***************************************************************/
/* BuildRHSAssert:                                     */
/***************************************************************/
globle struct expr *BuildRHSAssert(top,in_file,readFirstParen,error)
  struct expr *top;
  char *in_file;
  int readFirstParen, *error;
  {
   struct token act_tkn;
   struct expr *last_one, *next_one, *assert_list, *stub;
   int multi = CLIPS_FALSE;

   *error = CLIPS_FALSE;

   last_one = assert_list = NULL;
   multi = CLIPS_FALSE;
   while((next_one = GetRHSPattern(in_file,&act_tkn,&multi,
                                   error,CLIPS_FALSE,readFirstParen,RPAREN)) != NULL)
     {
      PPCRAndIndent();
        
      stub = GenConstant(FCALL,FindFunction("assert"));
      stub->arg_list = next_one;
      next_one = stub;
   
      if (last_one == NULL)
        { assert_list = next_one; }
      else
        { last_one->next_arg = next_one; }
      last_one = next_one;
      
      multi = CLIPS_FALSE;
      readFirstParen = CLIPS_TRUE;
     }
   
   if (*error)
     {
      ReturnExpression(top);
      ReturnExpression(assert_list);
      return(NULL);
     } 
     
   if (act_tkn.type == RPAREN) 
     {
      PPBackup();
      PPBackup();
      SavePPBuffer(")");
     }
     
   if (assert_list == NULL)
     {
      PrintCLIPS(WERROR,"\nAt least one fact must be asserted\n");
      ReturnExpression(top);
      return(NULL);
     }
   
   if (assert_list->next_arg != NULL)
     {
      top->value = (VOID *) FindFunction("progn");
      top->arg_list = assert_list;
     }
   else
     { 
      top->value = assert_list->value;
      top->arg_list = assert_list->arg_list;
      assert_list->arg_list = NULL;
      ReturnExpression(assert_list);
     }
   
   return(top);
  }
#endif

/***************************************************************/
/* GetRHSPattern:                                              */
/* Returns NULL if ) is first token.                           */
/* Returns assert arguments in test format otherwise.          */
/* Error flag is set to true if an error occurs.               */
/***************************************************************/
globle struct expr *GetRHSPattern(readSource,tempToken,multi,error,
                           constantsOnly,readFirstParen,endType)
  char *readSource;
  struct token *tempToken;
  int *multi, *error, constantsOnly, readFirstParen, endType;
  {
   struct expr *last_one, *next_one, *first_one;
   int printError;

   *error = CLIPS_FALSE;
   last_one = NULL;
   
   /*=================================================*/
   /* Get the opening parenthesis of the RHS pattern. */
   /*=================================================*/
   
   if (readFirstParen)
     {
      GetToken(readSource,tempToken);
      if (tempToken->type == RPAREN) return(NULL);
      if (tempToken->type != LPAREN)
        {
         SyntaxErrorMessage("RHS patterns");
         *error = CLIPS_TRUE;
         return(NULL);
        }
     }
   
   /*================================================================*/
   /* Get the first field and determine if a template exists for it. */
   /*================================================================*/
   
   first_one = GetAssertArgument(readSource,tempToken,multi,error,endType,
                                 constantsOnly,&printError);
   if (*error)
     {
      if (printError) SyntaxErrorMessage("RHS patterns");
      return(NULL);
     }
     
   if (first_one == NULL)
     {
      PrintCLIPS(WERROR,"\nRHS patterns must contain at least one field\n");
      *error = CLIPS_TRUE;
      return(NULL);
     }
     
#if DEFTEMPLATE_CONSTRUCT && (! BLOAD_ONLY)
   if (first_one->type == SYMBOL)
     {
      if (QFindDeftemplate((SYMBOL_HN *) tempToken->value) != NULL)
        {  
         first_one->next_arg = ParseAssertTemplate(readSource,tempToken,
                                                   multi,error,
                                                   tempToken->value,endType,
                                                   constantsOnly);
         if (*error)
           {
            ReturnExpression(first_one);
            first_one = NULL;
           }
           
         return(first_one);
        }
     }
#endif
    
   last_one = first_one;
   
   /*===========================*/
   /* Get the remaining fields. */
   /*===========================*/
   
   SavePPBuffer(" ");
   while ((next_one = GetAssertArgument(readSource,tempToken,multi,
                                        error,endType,constantsOnly,&printError)) != NULL)
     { 
      last_one->next_arg = next_one;
      last_one = next_one;
      SavePPBuffer(" ");
     }
 
   if (*error)
     {
      if (printError) PrintCLIPS(WERROR,"\nBad argument for RHS pattern\n");        
      ReturnExpression(first_one); 
      return(NULL);
     }
       
   PPBackup();
   PPBackup();
   SavePPBuffer(tempToken->print_rep);
   
   return(first_one);
  }
  
/***********************************/
/* GetAssertArgument:              */
/***********************************/
globle struct expr *GetAssertArgument(in_file,act_tkn,multi,error,endType,constantsOnly,printError)
  char *in_file;
  struct token *act_tkn;
  int *multi, *error;
  int endType, constantsOnly;
  int *printError;
  {
   struct expr *next_one = NULL;
   
   *printError = CLIPS_TRUE;
   GetToken(in_file,act_tkn);
   if (act_tkn->type == endType)
     { return(NULL); }
   
   if ((act_tkn->type == SYMBOL) ? (strcmp(ValueToString(act_tkn->value),"=") == 0) : CLIPS_FALSE)
     {
      if (constantsOnly)
        {
         *error = CLIPS_TRUE;
         return(NULL);
        }
        
#if ! RUN_TIME
      next_one = Function0Parse(in_file);
      if (next_one != NULL)
        {
         if (ExpressionFunctionType(next_one) == 'm')
           { *multi = CLIPS_TRUE; }
        } 
      else
#endif
        { 
         *printError = CLIPS_FALSE;
         *error = CLIPS_TRUE; 
        }
     }
   else if ((act_tkn->type == SYMBOL) || (act_tkn->type == STRING) ||
#if OBJECT_SYSTEM
           (act_tkn->type == INSTANCE_NAME) ||
#endif
           (act_tkn->type == FLOAT) || (act_tkn->type == INTEGER))
     { next_one = GenConstant(act_tkn->type,act_tkn->value); } 
   else if ((act_tkn->type == BWORD) || 
#if DEFGLOBAL_CONSTRUCT
            (act_tkn->type == GBWORD) ||
#endif
            (act_tkn->type == BWORDS))
     {
      if (constantsOnly)
        {
         *error = CLIPS_TRUE;
         return(NULL);
        }
        
      next_one = GenConstant(act_tkn->type,act_tkn->value);
         
      if (act_tkn->type == BWORDS)
        { *multi = CLIPS_TRUE; }
     } 
   else
     { 
      *error = CLIPS_TRUE;
      next_one = NULL;
     }
     
   return(next_one);
  }
  


 
