   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 RETE UTILITY MODULE                 */
   /*******************************************************/
   
/*************************************************************/
/* Purpose:                                                  */
/*      Provides a set of utility functions useful to other  */
/*      modules.                                             */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _RETEUTIL_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "router.h"
#include "network.h"
#include "expressn.h"
#include "evaluatn.h"
#include "facts.h"
#include "match.h"
#include "generate.h"
#include "utility.h"
#include "retract.h"
#include "reteutil.h"
#include "drive.h"
#include "engine.h"
#include "build.h"
#include "sysprime.h"

#if DEFRULE_CONSTRUCT

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/ 

   static long int               PseudoFactIndex = -2;
   
/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/ 

   globle struct partialMatch   *GlobalLHSBinds = NULL;
   globle struct partialMatch   *GlobalRHSBinds = NULL;
   
/************************************************************/
/* PrintPartialMatch: Prints out a list of fact id numbers  */
/*   associated with a partial match or rule instantiation. */
/************************************************************/
globle VOID PrintPartialMatch(logicalName,list)
  char *logicalName;
  struct partialMatch *list;
  {
   char print_space[20];
   short int i;
   
   for (i = 0; i < list->bcount;)
     {
      if (list->binds[i]->whoset >= 0)
        {
         sprintf(print_space,"f-%ld",list->binds[i]->whoset);
         PrintCLIPS(logicalName,print_space);
        }
      i++;
      if (i < list->bcount) PrintCLIPS(logicalName,",");
     }     
  }
  
/******************************************************/
/* CopyPartialMatch:  Copies a list of fact bindings. */
/******************************************************/
globle struct partialMatch *CopyPartialMatch(list)
  struct partialMatch *list;
  {
   struct partialMatch *linker;
   short int i;

   
   linker = get_var_struct(partialMatch,sizeof(struct alphaMatch *) * (list->bcount - 1));
      
   linker->next = NULL;
   linker->betaMemory = CLIPS_TRUE;
   linker->busy = CLIPS_FALSE;
   linker->activationf = CLIPS_FALSE;
   linker->dependentsf = CLIPS_FALSE;
   linker->notOriginf = CLIPS_FALSE;
   linker->counterf = CLIPS_FALSE;
   linker->dependents = NULL;
   linker->bcount = list->bcount;
   
   for (i = 0; i < linker->bcount; i++) linker->binds[i] = list->binds[i];

   return(linker);
  }
  
/*******************************************************/
/* MergePartialMatches:  Merges two lists of bindings. */
/*******************************************************/
globle struct partialMatch *MergePartialMatches(list1,list2)
  struct partialMatch *list1, *list2;
  {
   struct partialMatch *linker;
   short int i, j;
   
   linker = get_var_struct(partialMatch,sizeof(struct alphaMatch *) * (list1->bcount + list2->bcount - 1));
      
   linker->next = NULL;
   linker->betaMemory = CLIPS_TRUE;
   linker->busy = CLIPS_FALSE;
   linker->activationf = CLIPS_FALSE;
   linker->dependentsf = CLIPS_FALSE;
   linker->notOriginf = CLIPS_FALSE;
   linker->counterf = CLIPS_FALSE;
   linker->dependents = NULL;
   linker->bcount = list1->bcount + list2->bcount;
   
   for (i = 0; i < list1->bcount; i++) 
     { linker->binds[i] = list1->binds[i]; }
     
   for (i = list1->bcount, j = 0; i < linker->bcount; i++, j++)
     { linker->binds[i] = list2->binds[j]; }

   return(linker);
  }
  
/*********************************************************/
/* AddSingleMatch: Adds a single bind to a list of binds */
/*   creating a completely new list.                     */
/*********************************************************/
globle struct partialMatch *AddSingleMatch(list,afb)
  struct partialMatch *list;
  struct alphaMatch *afb;
  {
   struct partialMatch *linker;
   short int i;
   
   linker = get_var_struct(partialMatch,sizeof(struct alphaMatch *) * (list->bcount));
   
   linker->next = NULL;
   linker->betaMemory = CLIPS_TRUE;
   linker->busy = CLIPS_FALSE;
   linker->activationf = CLIPS_FALSE;
   linker->dependentsf = CLIPS_FALSE;
   linker->notOriginf = CLIPS_FALSE;
   linker->counterf = CLIPS_FALSE;
   linker->dependents = NULL;
   linker->bcount = list->bcount + 1;
   
   for (i = 0; i < list->bcount; i++)
     { linker->binds[i] = list->binds[i]; }

   linker->binds[i] = afb;
   
   return(linker);
  }
  
/***************************************************************/
/* FindFactInPartialMatch: Searches for a fact id number in a list of fact id */
/*   numbers.  Returns CLIPS_TRUE (1) if the number is found in the  */
/*   list, otherwise CLIPS_FALSE (0) is returned.                    */
/***************************************************************/
globle int FindFactInPartialMatch(factIndex,id_list)
  FACT_ID factIndex;
  struct partialMatch *id_list;
  {
   short int i;
   
   for (i = 0 ; i < id_list->bcount; i++)
     {
      if (id_list->binds[i]->whoset == factIndex)
        { return(CLIPS_TRUE); }
     }
     
   return(CLIPS_FALSE);
  }
  
/************************************************************************/
/* AdjustFieldPosition:  Given a list of segment markers and the index  */
/*   to a variable in a pattern, this function computes the index       */
/*   to the element in the fact where the variable begins.  In the      */
/*   case of segment variables, it also computes the extent (or length) */
/*   of the segment.  Note that the extent should be given a default    */
/*   value of either -1 or 1 for variables other than segment variables */
/*   before calling this routine.  An extent of -1 for these variables  */
/*   will distinguish their extent as being different when it is        */
/*   necessary to note their difference from a segment variable with an */
/*   extent of 1. For example, given the pattern (data $?x c $?y ?z)    */
/*   and the fact (data a b c d e f x), the actual index in the fact    */
/*   for the 5th item in the pattern (the variable ?z) would be 8 since */
/*   $?x binds to 2 fields and $?y binds to 3 fields.                   */
/************************************************************************/
globle int AdjustFieldPosition(markList,num,extent)
  struct multifieldMarker *markList;
  int num, *extent;
  {
   int actualIndex;

   actualIndex = num;
   while (markList != NULL)
     {
      if (markList->element == num)
        {
         *extent = (markList->end - markList->start) + 1;
         return(actualIndex);
        }
      else if (markList->element > num)
        { return(actualIndex); }

      actualIndex += (markList->end - markList->start);
      markList = markList->next;
     }

   return(actualIndex);
  }

/**********************************/
/* SetPseudoFactIndex:                */
/**********************************/
globle VOID SetPseudoFactIndex(value)
  long int value;
  {
   PseudoFactIndex = value;
  }
  
/****************************************************************/
/* NewPseudoFactPartialMatch:  Creates a partial structure that */
/*   indicates the "pseudo" fact to which a not pattern CE has  */
/*   been bound. Since a non-existant fact has no fact index,   */
/*   the partial match structure is given a pseudo fact index   */
/*   (a unique negative integer).                               */ 
/****************************************************************/
globle struct partialMatch *NewPseudoFactPartialMatch()
  {
   struct partialMatch *linker;
   struct alphaMatch *tempAlpha;
   
   linker = get_struct(partialMatch);
   linker->next = NULL;
   linker->betaMemory = CLIPS_TRUE;
   linker->busy = CLIPS_FALSE;
   linker->activationf = CLIPS_FALSE;
   linker->dependentsf = CLIPS_FALSE;
   linker->notOriginf = CLIPS_TRUE;
   linker->counterf = CLIPS_FALSE;
   linker->dependents = NULL;
   linker->bcount = 0;               
   tempAlpha = get_struct(alphaMatch);
   tempAlpha->aorigin = tempAlpha;
   tempAlpha->whoset = PseudoFactIndex;
   tempAlpha->origin = NULL;
   tempAlpha->marker = NULL;
   
   linker->binds[0] = tempAlpha;
   PseudoFactIndex--;
   return(linker);
  }
  
/******************************************************************/
/* IncrementPseudoFactIndex: Returns the current not ID and then increments it. */
/******************************************************************/
globle long int IncrementPseudoFactIndex()
  {
   long int rv;
   
   rv = PseudoFactIndex;
   PseudoFactIndex--;
   return(rv);
  }

/******************************************************************/
/* FlushAlphaBetaMemory: Returns all partial matches in a list of */
/*   partial matches either directly to the pool of free memory   */
/*   or to the list of GarbagePartialMatches. Partial matches     */
/*   stored in alpha memories and partial matches which store the */
/*   information for pseudo facts (for not CEs) may be referred   */
/*   to by other data structures and thus must be placed on the   */
/*   list of GarbagePartialMatches.                               */
/******************************************************************/
globle VOID FlushAlphaBetaMemory(pfl)
  struct partialMatch *pfl;
  {
   struct partialMatch *pfltemp;

   while (pfl != NULL)
     {
      pfltemp = pfl->next;
      if ((pfl->notOriginf) || (pfl->betaMemory == CLIPS_FALSE))
        {
         pfl->next = GarbagePartialMatches;
         GarbagePartialMatches = pfl;
        }
      else
        { ReturnPartialMatch(pfl); }
        
      pfl = pfltemp;
     }
  }

#if (! RUN_TIME) && (! BLOAD_ONLY)

/***************************************************************/
/* ClearPatternMatches: Clears the fact list of all pointers   */
/*   which point to a specific pattern.  The pointers are used */
/*   to remember which patterns were matched by a fact to      */
/*   make retraction easier.  When a rule is excised, the      */
/*   pointers need to be removed.                              */
/***************************************************************/
globle VOID ClearPatternMatches(pat_ptr)
  struct patternNode *pat_ptr;
  {
   struct fact *fact_ptr;
   struct patternMatch *match_before, *match_ptr;
 
   /*===========================================*/
   /* Loop through every fact in the fact list. */
   /*===========================================*/
  
   fact_ptr = (struct fact *) GetNextFact(NULL);
   while (fact_ptr != NULL)
     {
      match_before = NULL;
      match_ptr = (struct patternMatch *) fact_ptr->list;

      /*========================================*/
      /* Loop through every match for the fact. */
      /*========================================*/

      while (match_ptr != NULL)
        {
         if (match_ptr->matchingPattern == pat_ptr)
           {
            /* Remove the match. */
            if (match_before == NULL)
              {
               fact_ptr->list = (VOID *) match_ptr->next;
               rtn_struct(patternMatch,match_ptr);
               match_ptr = (struct patternMatch *) fact_ptr->list;
              }
            else
             {
              match_before->next = match_ptr->next;
              rtn_struct(patternMatch,match_ptr);
              match_ptr = match_before->next;
             }
           }
         else
          {  
           /* Move on to the next match. */
           match_before = match_ptr;
           match_ptr = match_ptr->next;
          }
       }
     fact_ptr = (struct fact *) GetNextFact(fact_ptr);
    }
  }
 
/**********************************************************************/
/* ResetNotedJoin: The purpose of this code is to set the patterns    */
/*   that have a not condition before them to true until a pattern    */
/*   matches them.                                                    */
/**********************************************************************/
globle VOID ResetNotedJoin(joinPtr)
  struct joinNode *joinPtr;
  {
   struct joinNode *join_list_2;
   struct patternNode *pat_ptr;     
   struct partialMatch *linker;
   int join_test;
   int count;
       
   if ((joinPtr->LHSPatternIsNegated == CLIPS_TRUE) &&
       (joinPtr->firstJoin == CLIPS_TRUE) &&
       (joinPtr->initialize == CLIPS_TRUE))
     {  
      pat_ptr = joinPtr->entry_pat;
      
      /*============================================================*/
      /* If the pattern node has already been reset, then count the */ 
      /* number of alpha matches which prevent this join from being */
      /* satisified.  If there are no alpha matches then see if the */
      /* test CE associated with this join is satisified indicating */
      /* whether the join as a whole is satisified.                 */
      /*============================================================*/
      
      if (pat_ptr->initialize == CLIPS_FALSE)
        {
         linker = pat_ptr->alpha;

         count = 0;
         while (linker != NULL)
           {
            count++;
            linker = linker->next;
           }
         
         joinPtr->id = count;
         
         if (count == 0)
           {
            join_test = EvaluateJoinExpression(joinPtr->not_eval,NULL,NULL,joinPtr);
            EvaluationError = CLIPS_FALSE;
           }
         else
           { join_test = CLIPS_FALSE; }
        }
        
      /*=================================================================*/ 
      /* If the pattern node hasn't been reset, then nothing has matched */
      /* yet and technically the pattern is satisfied if any associated  */
      /* test CE is satisfied. Activations caused by this may be removed */
      /* when facts are filtered through the pattern network later in    */
      /* the incremental reset.                                          */
      /*=================================================================*/ 
      
      else if (joinPtr->not_eval != NULL)
        { 
         join_test = EvaluateJoinExpression(joinPtr->not_eval,NULL,NULL,joinPtr);
         EvaluationError = CLIPS_FALSE;
        }
      else
        { join_test = CLIPS_TRUE; }
        
      /*==================================================*/
      /* If the join is satisfied, create a partial match */
      /*  and send it on down to subsequent joins.        */
      /*==================================================*/
      
      if (join_test == CLIPS_TRUE)
        {
         linker = NewPseudoFactPartialMatch();
         joinPtr->beta = linker;
         joinPtr->id = linker->binds[0]->whoset;
 
         join_list_2 = joinPtr->next_level;
         linker->bcount =  1;
         while (join_list_2 != NULL)
           {
            Drive(linker,join_list_2,LHS,CLIPS_FALSE);
            join_list_2 = join_list_2->right_drive_node;
           }
         linker->bcount =  0;
        }
     }
  }
  
/****************************************************************************/
/* PrimeJoin: Updates a join in a rule for an incremental reset. Joins are  */
/*   updated by "priming" them only if the join (or its associated pattern) */
/*   is shared with other rules that have already been incrementally reset. */
/*   A join for a new rule will be updated if it is marked for              */
/*   initialization and either its parent join or its associated entry      */
/*   pattern node has not been marked for initialization.                   */
/****************************************************************************/
globle VOID PrimeJoin(joinPtr)
  struct joinNode *joinPtr;
  {
   struct joinNode *tempPtr;
   struct partialMatch *copyList = NULL, *orgList, *nextAdd;
   struct partialMatch *alphaList;
   
   /*===========================================================*/
   /* If the join is the first join of a rule, then send all of */
   /* the partial matches from the alpha memory of the pattern  */
   /* associated with this join to the join for processing and  */
   /* the priming process is then complete.                     */
   /*===========================================================*/
               
   if (joinPtr->firstJoin == CLIPS_TRUE)
     {
      alphaList = joinPtr->entry_pat->alpha;
      while (alphaList != NULL)
        {
         Drive(alphaList,joinPtr,RHS,CLIPS_FALSE);
         alphaList = alphaList->next;
        }
      return;
     }
   
   /*======================================================*/
   /* If the join already has partial matches in its beta  */
   /* memory, then don't bother priming it. I don't recall */
   /* if this situation is possible.                       */
   /*======================================================*/
   
   if (joinPtr->beta != NULL) return;
   
   /*=============================================================*/
   /* Find another join that shares the same parent join with the */
   /* join being primed that has already been initialized. Such a */
   /* join must exist since this is one of the criteria for using */
   /* the PrimeJoin function.                                     */
   /*=============================================================*/
   
   tempPtr = joinPtr->right_drive_node;
   while ((tempPtr != NULL) ? (tempPtr->initialize == CLIPS_TRUE) : CLIPS_FALSE)
     { tempPtr = tempPtr->right_drive_node; }
   
   if (tempPtr == NULL)
     {
      CLIPSSystemError("RETEUTIL",2);
      ExitCLIPS(6);
     }
   
   /*=======================================================*/
   /* Copy the contents of the beta memory in reverse order */
   /* from the sibling join of the join being primed.       */
   /*=======================================================*/
   
   orgList = tempPtr->beta;
   while (orgList != NULL)
     {      
      if (joinPtr->LHSPatternIsNegated)
        {
         nextAdd = AddSingleMatch(orgList,NULL);
         nextAdd->notOriginf = CLIPS_TRUE;
         nextAdd->counterf = CLIPS_TRUE;
         nextAdd->bcount--; 
        }
      else if (joinPtr->lastJoin)
        {
         nextAdd = AddSingleMatch(orgList,NULL);
         nextAdd->activationf = CLIPS_TRUE;
         nextAdd->bcount--;
        }
      else 
       { nextAdd = CopyPartialMatch(orgList); }
      
      nextAdd->next = copyList;
      copyList = nextAdd;
      
      orgList = orgList->next;
     }

   /*===================================================================*/
   /* Copy the reversed beta memory list to the beta memory of the join */
   /* to be primed. The list is reversed  again during the copy process */
   /* to its original order. Process any activations or partial matches */
   /* created in conjunction with not CEs as they occur.                */
   /*===================================================================*/
   
   while (copyList != NULL)
     {
      nextAdd = copyList->next;
     
      copyList->next = joinPtr->beta;
      joinPtr->beta = copyList;
   
      if (joinPtr->lastJoin == CLIPS_FALSE)
        { 
         alphaList = joinPtr->entry_pat->alpha; 
     
         if ((joinPtr->LHSPatternIsNegated) && (alphaList == NULL))
           {  PNLDrive(joinPtr,copyList); }
        }
      else 
        { AddActivation(joinPtr,copyList); }
     
      copyList = nextAdd;
     }
   
   /*===============================================================*/
   /* If the join being primed is not the last (activation) join of */
   /* a rule and is not associated with a not CE, then process the  */
   /* partial matches in the alpha memory associated with this join */
   /* against the newly created beta memory.                        */
   /*===============================================================*/
   
   if ((joinPtr->lastJoin == CLIPS_FALSE) && 
       (joinPtr->LHSPatternIsNegated == CLIPS_FALSE))
     { 
      alphaList = joinPtr->entry_pat->alpha; 
      while (alphaList != NULL)
        {
         Drive(alphaList,joinPtr,RHS,CLIPS_FALSE);
         alphaList = alphaList->next;
        }
     }
  }
  
  
#endif

#if RUN_TIME || BLOAD_AND_BSAVE || BLOAD_ONLY || BLOAD
  
/****************************************************************/
/* ResetDeployedRuleImage:                                      */
/****************************************************************/
globle VOID ResetDeployedRuleImage()
  {
   struct fact *factPtr;
   
   ResetNotedPatterns(NetworkPointer());
   
   factPtr = (struct fact *) GetNextFact(NULL);
   
   while (factPtr != NULL)
     {
      PatternMatch(factPtr,factPtr->atoms,NetworkPointer(),1,0,NULL,NULL);
      factPtr = (struct fact *) GetNextFact(factPtr);
     }
   }

#endif

/***********************************************************************/
/* ResetNotedPatterns: The purpose of this code is to set the patterns */
/*   that have a not condition before them to true until a pattern     */
/*   matches them. Should be called only during an incremental reset.  */
/***********************************************************************/
globle VOID ResetNotedPatterns(pat_net)
  struct patternNode *pat_net;
  {
   struct joinNode *join_list_1, *join_list_2, *tempList;
   struct patternNode *pat_ptr;     
   struct partialMatch *linker;
   int join_test;

   pat_ptr = pat_net;

   while (pat_ptr != NULL)
     { 
      if (pat_ptr->stopNode)
        {
         join_list_1 = pat_ptr->entry_join;
         while (join_list_1 != NULL)
           {
            if ((join_list_1->LHSPatternIsNegated == CLIPS_TRUE) &&
                (join_list_1->firstJoin == CLIPS_TRUE))
              {  
               linker = pat_ptr->alpha;
                  
               if (join_list_1->id < 0)
                 {
                 
                  if (join_list_1->beta == NULL)
                    {
                      CLIPSSystemError("RETEUTIL",3);
                      ExitCLIPS(5);
                     }
                   else
                     {
                      join_list_1->beta->next = GarbagePartialMatches;
                      GarbagePartialMatches = join_list_1->beta;
                      join_list_1->beta = NULL;
                     } 
        
                  tempList = join_list_1->next_level;
                  while (tempList != NULL)
                    {
                     ClearLowerBetaMemory(tempList);
                     tempList = tempList->right_drive_node;
                    }
                 }
                 
               join_list_1->id = 0;
               
               while (linker != NULL)
                 {
                  join_list_1->id++;
                  linker = linker->next;
                 }
                    
               if (join_list_1->id == 0)
                 {
                  join_test = EvaluateJoinExpression(join_list_1->not_eval,NULL,NULL,join_list_1);
                  EvaluationError = CLIPS_FALSE;
                 }
               else
                 { join_test = CLIPS_FALSE; }
        
               if (join_test == CLIPS_TRUE)
                 {
                  linker = NewPseudoFactPartialMatch();
                  join_list_1->beta = linker;
                  join_list_1->id = linker->binds[0]->whoset;
                 
                  linker->bcount =  1;
                  join_list_2 = join_list_1->next_level;
                  while (join_list_2 != NULL)
                    {
                     Drive(linker,join_list_2,LHS,CLIPS_FALSE);
                     join_list_2 = join_list_2->right_drive_node;
                    }
                  linker->bcount =  0;
                 }
              }          
            join_list_1 = join_list_1->right_match_node;
           }
        }
      else
        { ResetNotedPatterns(pat_ptr->next_level); }
        
      pat_ptr = pat_ptr->right_node;
     }
     
  }
  
#if (CONSTRUCT_COMPILER || BLOAD_AND_BSAVE) && (! RUN_TIME)

/********************************************************************/
/* MarkRuleNetwork: Marks the pattern and join network marked slot. */
/********************************************************************/
globle VOID MarkRuleNetwork(value)
  int value;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct patternNode *pnPtr;
   
   /*====================================*/
   /* Mark the nodes as being unmarked. */
   /*====================================*/
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         pnPtr = ppPtr->pptr;
         while (pnPtr != NULL)
           {
            pnPtr->marked = value;
            pnPtr = pnPtr->last_level;
           }
        
         if (ppPtr->jptr != NULL) ppPtr->jptr->marked = value;
         
         ppPtr = ppPtr->next;
        }
        
      if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
      else rulePtr = rulePtr->next;
     }
  }
  
/******************************************************************/
/* TagRuleNetwork: Marks the pattern network with unique IDs. */
/******************************************************************/
globle VOID TagRuleNetwork(patternCount,joinCount)
  unsigned long int *patternCount, *joinCount;
  {
   struct defrule *rulePtr;
   struct patptr *ppPtr;
   struct patternNode *pnPtr;
   
   MarkRuleNetwork(0);
   
   /*============================*/
   /* Mark the nodes with an ID. */
   /*============================*/
   
   rulePtr = (struct defrule *) GetNextDefrule(NULL);
   
   while (rulePtr != NULL)
     {
      ppPtr = rulePtr->pats;
      while (ppPtr != NULL)
        {
         pnPtr = ppPtr->pptr;
         while (pnPtr != NULL)
           {
            if (pnPtr->marked == 0)
              { 
               pnPtr->marked = 1;
               pnPtr->bsaveID = *patternCount;
               (*patternCount)++;
               pnPtr = pnPtr->last_level;
              }
            else
              { pnPtr = NULL; }
           }
        
         if (ppPtr->jptr != NULL) 
           {
            if (ppPtr->jptr->marked == 0)
              { 
               ppPtr->jptr->marked = 1;
               ppPtr->jptr->bsaveID = *joinCount;
               (*joinCount)++;
              }
           }
         ppPtr = ppPtr->next;
        }
      
      if (rulePtr->disjunct != NULL) rulePtr = rulePtr->disjunct;
      else rulePtr = rulePtr->next;
     }
  }

#endif

#endif

/****************************************/
/* GetNumericArgument                               */
/****************************************/
globle BOOLEAN GetNumericArgument(test_ptr,fun_name,result,convertToFloat,position)
  struct expr *test_ptr;
  char *fun_name;
  DATA_OBJECT *result;
  BOOLEAN convertToFloat;
  int position;
  {
   int ntype;
   VOID *value;

   switch(test_ptr->type)
     {
      case FLOAT:
      case INTEGER:
        ntype = test_ptr->type;
        value = test_ptr->value;
        break;
        
#if DEFRULE_CONSTRUCT
      case SCALL_GET_VAR:
        GetVarSysFunction(result,test_ptr->value);
        ntype = result->type;
        value = result->value;
        break;
#endif
      default:      
        EvaluateExpression(test_ptr,result);
        ntype = result->type;
        value = result->value;
        break;
     }
   
   if ((ntype != FLOAT) && (ntype != INTEGER))
     {
      ExpectedTypeError(fun_name,position,"integer or float");
      SetHaltExecution(CLIPS_TRUE);
      SetEvaluationError(CLIPS_TRUE);
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return(CLIPS_FALSE);
     }

   if ((convertToFloat) && (ntype == INTEGER))
     {
      ntype = FLOAT;
      value = (VOID *) AddDouble((double) ValueToLong(value));
     }
   
   result->type = ntype;
   result->value = value;
   
   return(CLIPS_TRUE);
  }
  
  
  
