   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                RULE COMMANDS MODULE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
 
#define _RULECOM_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include "engine.h" 
#include "utility.h"
#include "constrct.h"
#include "drive.h"
#include "expressn.h"
#include "evaluatn.h"
#include "clipsmem.h"
#include "router.h"

#include "defrule.h"

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   VOID                           DefruleCommands(void);
   VOID                           RunCommand(void);
   VOID                           SetBreakCommand(void);
   VOID                           RemoveBreakCommand(void);
   VOID                           ShowBreaksCommand(void);
   VOID                          *SetStrategyCommand(void);
   VOID                          *GetStrategyCommand(void);
   VOID                           RefreshAgendaCommand(void);
   static char                   *GetStrategyName(int);
   VOID                           RulesCommand(void);
   VOID                           ListDefrules(void);
   VOID                           PPRuleCommand(void);
   BOOLEAN                        PPDefrule(char *,char *);
   VOID                           UndefruleCommand(void);
   VOID                           RefreshCommand(void);
   VOID                           MatchesCommand(void);
   VOID                           AgendaCommand(void);
   VOID                          *SetSalienceEvaluationCommand(void);
   VOID                          *GetSalienceEvaluationCommand(void);
   static char                   *SalienceEvaluationName(int);
   int                            GetIncrementalResetCommand(void);
   int                            SetIncrementalResetCommand(void);
   VOID                           HaltCommand(void);
#else
   VOID                           DefruleCommands();
   VOID                           RunCommand();
   VOID                           SetBreakCommand();
   VOID                           RemoveBreakCommand();
   VOID                           ShowBreaksCommand();
   VOID                          *SetStrategyCommand();
   VOID                          *GetStrategyCommand();
   VOID                           RefreshAgendaCommand();
   static char                   *GetStrategyName();
   VOID                           RulesCommand();
   VOID                           ListDefrules();
   VOID                           PPRuleCommand();
   BOOLEAN                        PPDefrule();
   VOID                           UndefruleCommand();
   VOID                           RefreshCommand();
   VOID                           MatchesCommand();
   VOID                           AgendaCommand();
   VOID                          *SetSalienceEvaluationCommand();
   VOID                          *GetSalienceEvaluationCommand();
   static char                   *SalienceEvaluationName();
   int                            GetIncrementalResetCommand();
   int                            SetIncrementalResetCommand();
   VOID                           HaltCommand();
#endif 
   
/****************************************/
/* GLOBAL EXTERNAL FUNCTION DEFINITIONS */
/****************************************/

#if ANSI_COMPILER  
   VOID                           DefruleBinarySetup(void);
#else
   VOID                           DefruleBinarySetup();
#endif

/************************************************************/
/* DefruleCommands:                                         */
/************************************************************/
#if ! RUN_TIME
globle VOID DefruleCommands()
  {
   DefineFunction("run",          'v', PTIF RunCommand,          "RunCommand");
   DefineFunction("excise",       'v', PTIF UndefruleCommand,    "UndefruleCommand");
   DefineFunction("undefrule",    'v', PTIF UndefruleCommand,    "UndefruleCommand");
   DefineFunction("refresh",      'v', PTIF RefreshCommand,      "RefreshCommand");
   DefineFunction("halt",         'v', PTIF HaltCommand,   "HaltCommand");
#if DEBUGGING_FUNCTIONS
   DefineFunction("rules",        'v', PTIF RulesCommand,        "RulesCommand");
   DefineFunction("list-defrules",'v', PTIF RulesCommand,        "RulesCommand");
   DefineFunction("pprule",       'v', PTIF PPRuleCommand,       "PPRuleCommand");
   DefineFunction("ppdefrule",    'v', PTIF PPRuleCommand,       "PPRuleCommand");
   DefineFunction("set-break",    'v', PTIF SetBreakCommand,     "SetBreakCommand");
   DefineFunction("remove-break", 'v', PTIF RemoveBreakCommand,  "RemoveBreakCommand");
   DefineFunction("show-breaks",  'v', PTIF ShowBreaksCommand,   "ShowBreaksCommand");
   DefineFunction("matches",      'v', PTIF MatchesCommand,      "MatchesCommand");
   DefineFunction("agenda",       'v', PTIF AgendaCommand,       "AgendaCommand");
#endif
   DefineFunction("get-strategy", 'w', PTIF GetStrategyCommand,  "GetStrategyCommand");
   DefineFunction("set-strategy", 'w', PTIF SetStrategyCommand,  "SetStrategyCommand");
   DefineFunction("get-salience-evaluation",'w',
                  PTIF GetSalienceEvaluationCommand,"GetSalienceEvaluationCommand");
   DefineFunction("set-salience-evaluation",'w',
                  PTIF SetSalienceEvaluationCommand,"SetSalienceEvaluationCommand");
   DefineFunction("refresh-agenda",'v',
                  PTIF RefreshAgendaCommand,"RefreshAgendaCommand");
   DefineFunction("get-incremental-reset",'b',
                  GetIncrementalResetCommand,"GetIncrementalResetCommand");
   DefineFunction("set-incremental-reset",'b',
                  SetIncrementalResetCommand,"SetIncrementalResetCommand");
          
   DefruleBinarySetup();
  }
#endif
 
/***********************************************************/
/* RunCommand: begins execution of the production rules.  */
/*   Takes an optional argument which indicates the number */
/*   of rules to fire before stopping.                     */
/*   Syntax: (run [ <number> ])                            */
/***********************************************************/
globle VOID RunCommand()
  {
   int numArgs;
   long int runLimit = -1;
   DATA_OBJECT argPtr;

   if ((numArgs = ArgCountCheck("run",NO_MORE_THAN,1)) == -1) return;

   if (numArgs == 0)
     { runLimit = -1; }
   else if (numArgs == 1)
     { 
      if (ArgTypeCheck("run",1,INTEGER,&argPtr) == CLIPS_FALSE) return;
      runLimit = DOToLong(argPtr);
     }

   RunCLIPS(runLimit); 
   
   return;
  }

#if DEBUGGING_FUNCTIONS
/************************************************************/
/* RulesCommand: displays the list of rules.                */
/*   Syntax: (rules)                                        */
/************************************************************/
globle VOID RulesCommand()
  {
   if (ArgCountCheck("rules",EXACTLY,0) == -1) return;
   
   ListDefrules();
  }
  
/************************************************************/
/* ListDefrules:                                           */
/************************************************************/
globle VOID ListDefrules()
  {
   RULEPTR rule_list;
   long count = 0;
   
   rule_list = NULL;
   while ((rule_list = GetNextDefrule(rule_list)) != NULL)
     {         
      if (GetHaltExecution() == CLIPS_TRUE) return;

      PrintCLIPS(WDISPLAY,GetDefruleName(rule_list));
      PrintCLIPS(WDISPLAY,"\n");
      count++;
     }
   
   PrintTally(WDISPLAY,count,"rule","rules");
     
   return;
  }
  
/************************************************************/
/* PPRuleCommand: pretty prints a rule.                     */
/*   Syntax: (ppdefrule <rule name>)                           */
/************************************************************/
globle VOID PPRuleCommand()
  {
   char *ruleName;

   ruleName = GetConstructName(1,1,"ppdefrule","rule name");
   if (ruleName == NULL) return;

   PPDefrule(ruleName,WDISPLAY);

   return;
  }

/*************************************************************/
/* PPDefrule: the driver which actually does the pretty      */
/*   printing of the rule.                                   */
/*************************************************************/
globle BOOLEAN PPDefrule(ruleName,fileid)
  char *ruleName, *fileid;
  {
   VOID *rule_ptr;

   rule_ptr = FindDefrule(ruleName);
   if (rule_ptr == NULL) 
     { 
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,ruleName);
      PrintCLIPS(WERROR,"\n");
      return(CLIPS_FALSE);
     }

   if (GetDefrulePPForm(rule_ptr) == NULL) return(CLIPS_TRUE);
   PrintInChunks(fileid,GetDefrulePPForm(rule_ptr));
   return(CLIPS_TRUE);
  }
  
/************************************************************/
/* SetBreakCommand:                                               */
/************************************************************/
globle VOID SetBreakCommand()
  {
   DATA_OBJECT argPtr;
   char *argument;
   VOID *defrulePtr;

   if (ArgCountCheck("set-break",EXACTLY,1) == -1) return;

   if (ArgTypeCheck("set-break",1,SYMBOL,&argPtr) == CLIPS_FALSE) return;
   
   argument = DOToString(argPtr);
   
   if ((defrulePtr = FindDefrule(argument)) == NULL)
     {
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,argument);
      PrintCLIPS(WERROR,"\n");
      return;
     }
   
   AddBreakpoint(defrulePtr);
  }
  
/************************************************************/
/* RemoveBreakCommand:                                            */
/************************************************************/
globle VOID RemoveBreakCommand()
  {
   DATA_OBJECT argPtr;
   char *argument;
   int nargs;
   VOID *defrulePtr;

   if ((nargs = ArgCountCheck("remove-break",NO_MORE_THAN,1)) == -1) 
     { return; }
     
   if (nargs == 0)
     {
      RemoveAllBreakpoints();
      return;
     }

   if (ArgTypeCheck("remove-break",1,SYMBOL,&argPtr) == CLIPS_FALSE) return;
   
   argument = DOToString(argPtr);
     
   if ((defrulePtr = FindDefrule(argument)) == NULL)
     {
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,argument);
      PrintCLIPS(WERROR,"\n");
      return;
     }
     
   if (RemoveBreakpoint(defrulePtr) == CLIPS_FALSE)
     {
      PrintCLIPS(WERROR,"Rule ");
      PrintCLIPS(WERROR,argument);
      PrintCLIPS(WERROR," does not have a breakpoint set.\n");
     }
  }

/************************************************************/
/* ShowBreaksCommand:                                             */
/************************************************************/
globle VOID ShowBreaksCommand()
  {
   if (ArgCountCheck("show-breaks",EXACTLY,0) == -1) return;
   
   ListBreakpoints();
  }
  
/********************************************************************/
/* MatchesCommand:  Displays the set of partial matches for a rule. */
/********************************************************************/
globle VOID MatchesCommand()
  {
   char *ruleName;
   VOID *rulePtr;
   
   ruleName = GetConstructName(1,1,"matches","rule name");
   if (ruleName == NULL) return;
   
   rulePtr = FindDefrule(ruleName);
   if (rulePtr == NULL) 
     {
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,ruleName);
      PrintCLIPS(WERROR,"\n");
      return;
     }

   ListMatches(rulePtr);
  }
  
/************************************************/
/* AgendaCommand: Prints out the agenda of the */
/*   rules that are ready to fire.              */
/*   Syntax: (agenda)                           */
/************************************************/
globle VOID AgendaCommand()
  {   
   if (ArgCountCheck("agenda",EXACTLY,0) == -1) return;

   ListAgenda();
  }
  
#endif

/*******************************************************/
/* UndefruleCommand: excises a rule from the database. */
/*   Syntax: (undefrule <rule name>)                   */
/*******************************************************/
globle VOID UndefruleCommand()
  {
   char *ruleName;

   ruleName = GetConstructName(1,1,"undefrule","rule name");
   if (ruleName == NULL) return;
   
#if (! RUN_TIME) && (! BLOAD_ONLY)
   if ((FindDefrule(ruleName) == NULL) && (strcmp("*",ruleName) != 0))       
     {
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,ruleName);
      PrintCLIPS(WERROR,"\n");
      return;
     }
   else if (DeleteNamedDefrule(ruleName) == CLIPS_FALSE)        
     {
      PrintCLIPS(WERROR,"Unable to delete defrule ");
      PrintCLIPS(WERROR,ruleName);
      PrintCLIPS(WERROR,"\n");
      return;
     }
#else
   PrintCLIPS(WERROR,"Unable to delete defrule ");
   PrintCLIPS(WERROR,ruleName);
   PrintCLIPS(WERROR,"\n");
#endif

   return;
  }
  
/*******************************************************************/
/* RefreshCommand:  Refreshes a rule reactivating all activations  */
/*   that have fired but still satify the LHS of the rule.         */
/*******************************************************************/
globle VOID RefreshCommand()
  {
   char *ruleName;
   VOID *rulePtr;
   
   ruleName = GetConstructName(1,1,"refresh","rule name");
   if (ruleName == NULL) return;

   rulePtr = FindDefrule(ruleName);
   if (rulePtr == NULL) 
     {
      PrintCLIPS(WERROR,"Unable to find defrule ");
      PrintCLIPS(WERROR,ruleName);
      PrintCLIPS(WERROR,"\n");
      return;
     }
     
   RefreshDefrule(rulePtr);
  }
  
/************************************************************/
/* GetStrategyCommand:                        */
/************************************************************/
globle VOID *GetStrategyCommand()
  {
   ArgCountCheck("get-strategy",EXACTLY,0);
       
   return(AddSymbol(GetStrategyName(GetStrategy())));
  }
  
/************************************************************/
/* SetStrategyCommand:                        */
/************************************************************/
globle VOID *SetStrategyCommand()
  {
   DATA_OBJECT argPtr;
   char *argument;

   if (ArgCountCheck("set-strategy",EXACTLY,1) == -1)
     { return(AddSymbol(GetStrategyName(GetStrategy()))); }
   
   if (ArgTypeCheck("set-strategy",1,SYMBOL,&argPtr) == CLIPS_FALSE)
     { return(AddSymbol(GetStrategyName(GetStrategy()))); }
     
   argument = DOToString(argPtr);
   
   if (strcmp(argument,"depth") == 0)            
     { SetStrategy(DEPTH_STRATEGY); } 
   else if (strcmp(argument,"breadth") == 0)  
     { SetStrategy(BREADTH_STRATEGY); }
   else if (strcmp(argument,"lex") == 0)  
     { SetStrategy(LEX_STRATEGY); }
   else if (strcmp(argument,"mea") == 0)  
     { SetStrategy(MEA_STRATEGY); }
   else if (strcmp(argument,"complexity") == 0)  
     { SetStrategy(COMPLEXITY_STRATEGY); }
   else if (strcmp(argument,"simplicity") == 0)  
     { SetStrategy(SIMPLICITY_STRATEGY); }
   else if (strcmp(argument,"random") == 0)  
     { SetStrategy(RANDOM_STRATEGY); }
   else
     {
      ExpectedTypeError("set-strategy",1,
      "symbol with value depth, breadth, lex, mea, complexity, simplicity, or random");
      return(AddSymbol(GetStrategyName(GetStrategy())));
     }

   return(AddSymbol(argument));
  }

/************************************************************/
/* RefreshAgendaCommand:                                    */
/************************************************************/
globle VOID RefreshAgendaCommand()
  {
   ArgCountCheck("refresh-agenda",EXACTLY,0);
       
   RefreshAgenda();
  }
  
/*****************************/
/* GetStrategyName:          */
/*****************************/
static char *GetStrategyName(strategy)
  int strategy;
  {
   char *sname;
   
   switch (strategy)
     {
      case DEPTH_STRATEGY:
        sname = "depth";
        break;
      case BREADTH_STRATEGY:
        sname = "breadth";
        break;
      case LEX_STRATEGY:
        sname = "lex";
        break;
      case MEA_STRATEGY:
        sname = "mea";
        break;
      case COMPLEXITY_STRATEGY:
        sname = "complexity";
        break;
      case SIMPLICITY_STRATEGY:
        sname = "simplicity";
        break;
      case RANDOM_STRATEGY:
        sname = "random";
        break;
      default:
        sname = "unknown";
        break;
     }
     
   return(sname);
  }

/************************************************************/
/* SetSalienceEvaluationCommand:                            */
/************************************************************/
globle VOID *SetSalienceEvaluationCommand()
  {
   DATA_OBJECT argPtr;
   char *argument;

   if (ArgCountCheck("set-salience-evaluation",EXACTLY,1) == -1)
     { return(AddSymbol(SalienceEvaluationName(GetSalienceEvaluation()))); }
   
   if (ArgTypeCheck("set-salience-evaluation",1,SYMBOL,&argPtr) == CLIPS_FALSE)
     { return(AddSymbol(SalienceEvaluationName(GetSalienceEvaluation()))); }
     
   argument = DOToString(argPtr);
   
   if (strcmp(argument,"when-defined") == 0)            
     { SetSalienceEvaluation(WHEN_DEFINED); } 
   else if (strcmp(argument,"when-activated") == 0)  
     { SetSalienceEvaluation(WHEN_ACTIVATED); }
   else if (strcmp(argument,"every-cycle") == 0)  
     { SetSalienceEvaluation(EVERY_CYCLE); }
   else
     {
      ExpectedTypeError("set-salience-evaluation",1,
      "symbol with value when-defined, when-activated, or every-cycle");
      return(AddSymbol(SalienceEvaluationName(GetSalienceEvaluation())));
     }

   return(AddSymbol(argument));
  }
  
/************************************************************/
/* GetSalienceEvaluationCommand:                        */
/************************************************************/
globle VOID *GetSalienceEvaluationCommand()
  {
   ArgCountCheck("get-salience-evaluation",EXACTLY,0);
       
   return(AddSymbol(SalienceEvaluationName(GetSalienceEvaluation())));
  }
  
/*****************************/
/* SalienceEvaluationName:   */
/*****************************/
static char *SalienceEvaluationName(strategy)
  int strategy;
  {
   char *sname;
   
   switch (strategy)
     {
      case WHEN_DEFINED:
        sname = "when-defined";
        break;
      case WHEN_ACTIVATED:
        sname = "when-activated";
        break;
      case EVERY_CYCLE:
        sname = "every-cycle";
        break;
      default:
        sname = "unknown";
        break;
     }
     
   return(sname);
  } 
  
/************************************************************/
/* SetIncrementalResetCommand:                        */
/************************************************************/
globle int SetIncrementalResetCommand()
  {
   int oldValue;
   DATA_OBJECT argPtr;
   
   oldValue = GetIncrementalReset();
   
   if (ArgCountCheck("set-incremental-reset",EXACTLY,1) == -1) 
     { return(oldValue); }
     
   if (GetNextDefrule(NULL) != NULL)
     {
      PrintCLIPS(WERROR,"The incremental reset behavior cannot be changed with rules loaded.\n");
      SetEvaluationError(CLIPS_TRUE);
      return(oldValue);
     }
     
   RtnUnknown(1,&argPtr);
     
   if ((argPtr.value == (VOID *) CLIPSFalseSymbol) && (argPtr.type == SYMBOL))
     { SetIncrementalReset(CLIPS_FALSE); }
   else
     { SetIncrementalReset(CLIPS_TRUE); }
     
   return(oldValue);
  }
  
/************************************************************/
/* GetIncrementalResetCommand:                        */
/************************************************************/
globle int GetIncrementalResetCommand()
  {
   int oldValue;
   
   oldValue = GetIncrementalReset();
   
   if (ArgCountCheck("get-incremental-reset",EXACTLY,0) == -1) 
     { return(oldValue); }
     
   return(oldValue);
  }

/***********************************************/
/* HaltCommand: Causes rule execution to halt. */
/***********************************************/
globle VOID HaltCommand() 
  { 
   ArgCountCheck("halt",EXACTLY,0);
   HaltRules = CLIPS_TRUE;
  }
  
#endif /* DEFRULE_CONSTRUCT */

