   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    SCANNER MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Chris Culbert                                        */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _SCANNER_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>
#include <ctype.h>

#include "setup.h"
#include "constant.h"
#include "scanner.h"
#include "clipsmem.h"
#include "symbol.h"
#include "utility.h"
#include "router.h"

#if ANSI_COMPILER
#include <stdlib.h>
#else
extern double atof(); 
extern long atol();
#endif

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   static VOID                   *ScanSymbol(char *,int,int *);
   static VOID                   *ScanString(char *);
   static VOID                    ScanNumber(char *,struct token *); 
   static char                   *AppendStrings(char *,char *);
#else
   static VOID                   *ScanSymbol();
   static VOID                   *ScanString();
   static VOID                    ScanNumber();
   static char                   *AppendStrings();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static char            *GlobalString = NULL; 
   static int              GlobalMax = 0;
   static int              GlobalPos = 0;   
   static int              PPBufferStatus = OFF; 
   static int              IndentationDepth = 0;
   
   static int              PPBufferPos = 0;
   static int              PPBufferMax = 0;
   static int              PPBackupOnce = 0;
   static int              PPBackupTwice = 0;
   static char            *PrettyPrintBuffer = NULL;
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   int                     IgnoreCompletionErrors = CLIPS_FALSE;
   
/*####################################################################*/
/*####################################################################*/
/*###                                                              ###*/
/*###                          SCANNER                             ###*/
/*###                                                              ###*/
/*####################################################################*/
/*####################################################################*/

/**********************************************************************/
/* GetToken: Scans infile for the next token.  The type of the token  */
/*   (symbol, string, number, variable, etc.) is return in the global */
/*   variable token.  If the token has an associated string, then the */
/*   value of this string and its hash table entry are stored in the  */
/*   global variables TKNWORD and HASHWORD respectively.  If the      */
/*   token is a number, then its value is stored in the global        */
/*   variable TKNNUMBER.                                              */
/**********************************************************************/
globle VOID GetToken(infile,tknstruct)
 char *infile;
 struct token *tknstruct;
 {
   int inchar;
   int type;

   /*=======================================*/
   /* Set Unknown default values for token. */
   /*=======================================*/

   tknstruct->type = KUNKNOWN;
   tknstruct->value = NULL;
   tknstruct->print_rep = "unknown";
   GlobalPos = 0;
   GlobalMax = 0;

   /*==============================================*/
   /* Remove all white space before processing the */ 
   /* GetToken() request.                          */
   /*==============================================*/

   inchar = GetcCLIPS(infile);
   while ((inchar == ' ') || (inchar == '\n') || (inchar == '\f') ||
          (inchar == '\r') || (inchar == ';') || (inchar == '\t'))
     {    
      /*=======================*/
      /* Remove comment lines. */
      /*=======================*/
 
      if (inchar == ';')               
        {
         inchar = GetcCLIPS(infile);
         while ((inchar != '\n') && (inchar != '\r') && (inchar != EOF) )
           { inchar = GetcCLIPS(infile); }
        }
      inchar = GetcCLIPS(infile);
     }

   /*==========================*/
   /* Process Symbolic Tokens. */
   /*==========================*/

   if (isalpha(inchar))
     {
      tknstruct->type = SYMBOL;
      UngetcCLIPS(inchar,infile);
      tknstruct->value = (VOID *) ScanSymbol(infile,0,&type);
      tknstruct->print_rep = ValueToString(tknstruct->value);                    
     }
     
   /*===============================================*/
   /* Process Number Tokens beginning with a digit. */
   /*===============================================*/
   
   else if (isdigit(inchar))     
     {
      UngetcCLIPS(inchar,infile);
      ScanNumber(infile,tknstruct);
     }

   else switch (inchar)
     {
      /*========================*/
      /* Process String Tokens. */
      /*========================*/

      case '"': 
         tknstruct->value = (VOID *) ScanString(infile);
         tknstruct->type = STRING; 
         tknstruct->print_rep = StringPrintForm(ValueToString(tknstruct->value));
         break;

      /*=======================================*/
      /* Process Tokens that might be numbers. */
      /*=======================================*/

      case '-':
      case '.':
      case '+':
         UngetcCLIPS(inchar,infile);
         ScanNumber(infile,tknstruct);
         break;

      /*============================*/
      /* Process ? and ?var Tokens. */
      /*============================*/
   
       case '?':
          inchar = GetcCLIPS(infile);
          if (isalpha(inchar) 
#if DEFGLOBAL_CONSTRUCT
              || (inchar == '*'))
#else
              )
#endif
            {
             UngetcCLIPS(inchar,infile);
             tknstruct->value = (VOID *) ScanSymbol(infile,0,&type);
             tknstruct->type = BWORD;
#if DEFGLOBAL_CONSTRUCT
             if ((ValueToString(tknstruct->value)[0] == '*') && 
                 (strlen(ValueToString(tknstruct->value)) > 1) &&
                 (ValueToString(tknstruct->value)[strlen(ValueToString(tknstruct->value)) - 1] == '*'))
               { tknstruct->type = GBWORD; }
#endif
             tknstruct->print_rep = AppendStrings("?",ValueToString(tknstruct->value));
            }
          else
            {
             tknstruct->type = SINGLE;
             tknstruct->value = (VOID *) AddSymbol("?");
             UngetcCLIPS(inchar,infile);
             tknstruct->print_rep = "?";
            } 
          break;

      /*==============================*/
      /* Process $? and $?var Tokens. */
      /*==============================*/

      case '$':
         if ((inchar = GetcCLIPS(infile)) == '?')
           {
            inchar = GetcCLIPS(infile);
            if (isalpha(inchar))
              {
               UngetcCLIPS(inchar,infile);
               tknstruct->value = (VOID *) ScanSymbol(infile,0,&type);
               tknstruct->type = BWORDS;
               tknstruct->print_rep = AppendStrings("$?",ValueToString(tknstruct->value));     
              }
            else
              {
               tknstruct->type = MULTIFIELD;
               tknstruct->value = (VOID *) AddSymbol("$?");
               tknstruct->print_rep = "$?";              
               UngetcCLIPS(inchar,infile);
              }
           }
         else 
           { 
            tknstruct->type = SYMBOL; 
            GlobalString = ExpandStringWithChar('$',GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            UngetcCLIPS(inchar,infile); 
            tknstruct->value = (VOID *) ScanSymbol(infile,1,&type); 
            tknstruct->print_rep = ValueToString(tknstruct->value); 
           }
         break;
   
      case '<':
         tknstruct->type = SYMBOL;
         GlobalString = ExpandStringWithChar('<',GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80); 
         tknstruct->value = (VOID *) ScanSymbol(infile,1,&type);
         tknstruct->print_rep = ValueToString(tknstruct->value); 
         break;
         
      /*=============================================*/
      /* Process "(", ")", "~", "|", and "&" Tokens. */
      /*=============================================*/

      case '(':
         tknstruct->type = LPAREN;
         tknstruct->value = (VOID *) AddSymbol("(");
         tknstruct->print_rep = "("; 
         break;
      
      case ')': 
         tknstruct->type= RPAREN;
         tknstruct->value = (VOID *) AddSymbol(")");
         tknstruct->print_rep = ")"; 
         break;
      
      case '~': 
         tknstruct->type = LNOT;
         tknstruct->value = (VOID *) AddSymbol("~");
         tknstruct->print_rep = "~"; 
         break;
         
      case '|':
         tknstruct->type = LOR;
         tknstruct->value = (VOID *) AddSymbol("|");
         tknstruct->print_rep = "|"; 
         break;

      case '&':   
         tknstruct->type =  LAND;
         tknstruct->value = (VOID *) AddSymbol("&");
         tknstruct->print_rep = "&"; 
         break;
 
      /*============================*/
      /* Process End-of-File Token. */
      /*============================*/
      
      case EOF:
      case 0:
      case 3:
         tknstruct->type = STOP;
         tknstruct->value = (VOID *) AddSymbol("stop");
         tknstruct->print_rep = "";   
         break;
   
      /*=======================*/
      /* Process Other Tokens. */
      /*=======================*/

      default:
         if (isprint(inchar))
           {
            UngetcCLIPS(inchar,infile);
            tknstruct->value = (VOID *) ScanSymbol(infile,0,&type);
            tknstruct->type = type; 
            tknstruct->print_rep = ValueToString(tknstruct->value); 
           }
         else
           { tknstruct->print_rep = "<<<unprintable character>>>"; }
         break;
     } 
     
   if (tknstruct->type == INSTANCE_NAME)
     {
      SavePPBuffer("["); 
      SavePPBuffer(tknstruct->print_rep); 
      SavePPBuffer("]"); 
     }
   else 
     { SavePPBuffer(tknstruct->print_rep); }
      
   if (GlobalString != NULL) 
     {
      rm(GlobalString,GlobalMax);
      GlobalString = NULL;
     }
     
   return;     
  }

/******************************************************************/
/* ScanSymbol:  The purpose of ScanSymbol is to input one legal   */
/*   variable name. A legal variable name consists of a letter    */
/*   followed by any printable ascii character with the exception */
/*   of '<', '(', ')', '&', '|', '~'  ' ', '"', and ';'.          */
/******************************************************************/
static VOID *ScanSymbol(infile,count,type)
  char *infile;
  int count;
  int *type;
  {
   int inchar;
#if OBJECT_SYSTEM
   VOID *symbol;
#endif
  
   inchar = GetcCLIPS(infile);
   while ( (inchar != '<') && (inchar != '"') &&
           (inchar != '(') && (inchar != ')') &&
           (inchar != '&') && (inchar != '|') && (inchar != '~') &&
           (inchar != ' ') && (inchar != ';') && 
           isprint(inchar) )
     {             
      GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);


      count++;
      inchar = GetcCLIPS(infile);
     }
   
   /*=======================================*/
   /* Stuff last character back into buffer */
   /* and return the symbol through hasher. */
   /*=======================================*/

   UngetcCLIPS(inchar,infile);
   
   /*====================================================*/
   /* Symbols of the form [<symbol>] are instance names. */ 
   /*====================================================*/
   
#if OBJECT_SYSTEM
   if (count > 2)
     {
      if ((GlobalString[0] == '[') ? (GlobalString[count-1] == ']') : CLIPS_FALSE)
        {
         *type = INSTANCE_NAME;
         inchar = ']';
        }
      else
        {
         *type = SYMBOL;
         return(AddSymbol(GlobalString));
        }
      GlobalString[count-1] = EOS;
      symbol = AddSymbol(GlobalString+1);
      GlobalString[count-1] = (char) inchar;
      return(symbol);
     }
   else
     {
      *type = SYMBOL;
      return(AddSymbol(GlobalString));
     }
#else
   *type = SYMBOL;
   return(AddSymbol(GlobalString));
#endif
  }
 
/********************************************************************/
/* ScanString:  The purpose of ScanString is to input one string with any */
/*   embedded characters. It is delimited by a leading and trailing */
/*   double quote.                                                  */
/********************************************************************/
static VOID *ScanString(infile)
  char *infile;
  {
   int inchar;
   int pos = 0, max = 0;
   char *theString = NULL;
   VOID *thePtr;

   inchar = GetcCLIPS(infile);                          
   while ((inchar != '"') && (inchar != EOF))      
     {
      if (inchar == '\\')
        { inchar = GetcCLIPS(infile); }
 
      theString = ExpandStringWithChar(inchar,theString,&pos,&max,max+80);
      inchar = GetcCLIPS(infile);                 
     }
       
   if ((inchar == EOF) && (IgnoreCompletionErrors == CLIPS_FALSE))
     { PrintCLIPS(WERROR,"\nEncountered End-Of-File while scanning a string\n"); }
     
   if (theString == NULL)
     { thePtr = AddSymbol(""); }
   else
     { 
      thePtr = AddSymbol(theString);
      rm(theString,max);
     }
     
   return(thePtr);
  }
  
/********************************************************************/
/* ScanNumber:  The purpose of ScanNumber is to input a numeric value   */
/*   and return it as a token. The number may be proceeded by a '+' */
/*   or a '-' sign, and may be in exponential form.                 */
/********************************************************************/
static VOID ScanNumber(infile,tknstruct)
  char *infile;
  struct token *tknstruct;
  {
   int count = 0;
   int inchar, phase;
   int digitFound = CLIPS_FALSE;
   int processFloat = CLIPS_FALSE;
   double fvalue;
   long lvalue;
   int type;
   
   /* Phases:              */
   /*  -1 = sign           */
   /*   0 = integral       */
   /*   1 = decimal        */
   /*   2 = exponent-begin */
   /*   3 = exponent-value */
   /*   5 = done           */
   /*   9 = error          */
      
   inchar = GetcCLIPS(infile);
   phase = -1; 
   
   while ((phase != 5) && (phase != 9))
     {
      if (phase == -1)
        { 
         if (isdigit(inchar))
           { 
            phase = 0;
            digitFound = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
         else if ((inchar == '+') || (inchar == '-'))
           {
            phase = 0;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
         else if (inchar == '.')
           {
            processFloat = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
            phase = 1; 
           }
         else if ((inchar == 'E') || (inchar == 'e'))
           {
            processFloat = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;  
            phase = 2; 
           }
         else if ( (inchar == '<') || (inchar == '"') ||
                   (inchar == '(') || (inchar == ')') ||
                   (inchar == '&') || (inchar == '|') || (inchar == '~') ||
                   (inchar == ' ') || (inchar == ';') || 
                   (isprint(inchar) == 0) )
           { phase = 5; }
         else 
           {
            phase = 9;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;  
           }
        }
      else if (phase == 0)
        { 
         if (isdigit(inchar))
           { 
            digitFound = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;  
           }
         else if (inchar == '.')
           {
            processFloat = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
            phase = 1; 
           }
         else if ((inchar == 'E') || (inchar == 'e'))
           {
            processFloat = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;   
            phase = 2; 
           }
         else if ( (inchar == '<') || (inchar == '"') ||
                   (inchar == '(') || (inchar == ')') ||
                   (inchar == '&') || (inchar == '|') || (inchar == '~') ||
                   (inchar == ' ') || (inchar == ';') ||
                   (isprint(inchar) == 0) )
           { phase = 5; }
         else 
           {
            phase = 9;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;  
           }
        }
      else if (phase == 1)
        {
         if (isdigit(inchar))
           { 
            digitFound = CLIPS_TRUE;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++;  
           }
         else if ((inchar == 'E') || (inchar == 'e'))
           {
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
            phase = 2; 
           }
         else if ( (inchar == '<') || (inchar == '"') ||
                   (inchar == '(') || (inchar == ')') ||
                   (inchar == '&') || (inchar == '|') || (inchar == '~') ||
                   (inchar == ' ') || (inchar == ';') ||
                   (isprint(inchar) == 0) )
           { phase = 5; }
         else
           {
            phase = 9;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
        }
      else if (phase == 2)
        {
         if (isdigit(inchar))
           { 
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
            phase = 3; 
           }
         else if ((inchar == '+') || (inchar == '-'))
           {
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
            phase = 3; 
           }
         else if ( (inchar == '<') || (inchar == '"') ||
                   (inchar == '(') || (inchar == ')') ||
                   (inchar == '&') || (inchar == '|') || (inchar == '~') ||
                   (inchar == ' ') || (inchar == ';') ||
                   (isprint(inchar) == 0) )
           { 
            digitFound = CLIPS_FALSE;
            phase = 5; 
           }
         else
           {
            phase = 9;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
        }
      else if (phase == 3)
        { 
         if (isdigit(inchar))
           { 
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
         else if ( (inchar == '<') || (inchar == '"') ||
                   (inchar == '(') || (inchar == ')') ||
                   (inchar == '&') || (inchar == '|') || (inchar == '~') ||
                   (inchar == ' ') || (inchar == ';') || 
                   (isprint(inchar) == 0) )
           { 
            if ((GlobalString[count-1] == '+') || (GlobalString[count-1] == '-'))
              { digitFound = CLIPS_FALSE; }
            phase = 5; 
           }
         else
           {
            phase = 9;
            GlobalString = ExpandStringWithChar(inchar,GlobalString,&GlobalPos,&GlobalMax,GlobalMax+80);
            count++; 
           }
        }  
        
      if ((phase != 5) && (phase != 9)) 
        { inchar = GetcCLIPS(infile); }
     }
    
   if (phase == 9)
     { 
      tknstruct->value = (VOID *) ScanSymbol(infile,count,&type);
      tknstruct->type = type;
      tknstruct->print_rep = ValueToString(tknstruct->value);
      return;
     }

   /*=======================================*/
   /* Stuff last character back into buffer */
   /* and return the number.                */
   /*=======================================*/

   UngetcCLIPS(inchar,infile);  
   
   if (! digitFound)
     {
      tknstruct->type = SYMBOL;
      tknstruct->value = (VOID *) AddSymbol(GlobalString);
      tknstruct->print_rep = ValueToString(tknstruct->value);
      return;
     }
     
   if (processFloat)
     {
      fvalue = atof(GlobalString);
      tknstruct->type = FLOAT;
      tknstruct->value = (VOID *) AddDouble(fvalue);
      tknstruct->print_rep = FloatToString(ValueToDouble(tknstruct->value));
     }
   else
     {
      lvalue = atol(GlobalString);
      tknstruct->type = INTEGER;
      tknstruct->value = (VOID *) AddLong(lvalue);
      tknstruct->print_rep = LongIntegerToString(ValueToLong(tknstruct->value));
     }

   return;
  }
  
/********************************************************************/
/* CopyToken:                                                        */ 
/********************************************************************/
globle VOID CopyToken(token1,token2)
  struct token *token1, *token2;
  {
   token1->type = token2->type;
   token1->value = token2->value;
   token1->print_rep = token2->print_rep;
  }

/*####################################################################*/
/*####################################################################*/
/*###                                                              ###*/
/*###                    PRETTY PRINT FUNCTIONS                    ###*/
/*###                                                              ###*/
/*####################################################################*/
/*####################################################################*/

/********************************************************/
/* FlushPPBuffer:  Resets the pretty print save buffer. */
/********************************************************/
globle int FlushPPBuffer()
  {
   if (PrettyPrintBuffer == NULL) return(0);
   PPBackupOnce = 0;
   PPBackupTwice = 0;
   PPBufferPos = 0;
   PrettyPrintBuffer[0] = EOS;
   return(1);
  }
  
/**********************************************************************/
/* DestroyPPBuffer:  Resets and removes the pretty print save buffer. */
/**********************************************************************/
globle VOID DestroyPPBuffer()
  {
   PPBackupOnce = 0;
   PPBackupTwice = 0;
   PPBufferPos = 0;
   if (PrettyPrintBuffer != NULL) rm(PrettyPrintBuffer,PPBufferMax);
   PrettyPrintBuffer = NULL;
   PPBufferMax = 0;
  }

/*******************************************************/
/* SavePPBuffer:  Appends a string to the end of the */
/*   pretty print save buffer.                         */
/*******************************************************/
globle VOID SavePPBuffer(str)
  char *str;
  {
   if (PPBufferStatus == OFF) return;

   if (strlen(str) + PPBufferPos + 1 >= PPBufferMax)
     { 
      PrettyPrintBuffer = genrealloc(PrettyPrintBuffer,PPBufferMax,PPBufferMax + 512);
      PPBufferMax += 512; 
     }
  
   PPBackupTwice = PPBackupOnce;
   PPBackupOnce = PPBufferPos;

   PrettyPrintBuffer = AppendToString(str,PrettyPrintBuffer,&PPBufferPos,&PPBufferMax);
  }

/****************************************************/
/* PPBackup:  Removes the last string added to the  */
/*   pretty print save buffer.  Only capable of     */
/*   backing up for the two most recent additions.  */
/****************************************************/
globle VOID PPBackup()
  {
   if ((PPBufferStatus == OFF) || (PrettyPrintBuffer == NULL)) return;
   
   PPBufferPos = PPBackupOnce;
   PPBackupOnce = PPBackupTwice;
   PrettyPrintBuffer[PPBufferPos] = EOS;
  }

/*****************************************************/
/* CopyPPBuffer:  Makes a copy of the pretty print */
/*   save buffer.                                    */
/*****************************************************/
globle char *CopyPPBuffer()
  {
   int length;
   char *newString;

   length = (1 + strlen(PrettyPrintBuffer)) * (int) sizeof (char);
   newString = (char *) gm2(length);
   
   strcpy(newString,PrettyPrintBuffer);
   return(newString);
  }
  
/*****************************************************/
/* GetPPBuffer                                    */
/*****************************************************/
globle char *GetPPBuffer()
  {
   return(PrettyPrintBuffer);
  }
  
/********************************************************/
/* PPCRAndIndent:  Prints white spaces into the pretty */
/*   print buffer.                                      */
/********************************************************/
globle VOID PPCRAndIndent()
  {
   int i;
   char buffer[120];

   buffer[0] = '\n';

   for (i = 1 ; i <= IndentationDepth ; i++)
     { buffer[i] = ' '; }
   buffer[i] = EOS;

   SavePPBuffer(buffer);
  }
  
/********************************************************/
/* IncrementIndentDepth:                                    */
/********************************************************/
globle VOID IncrementIndentDepth(value)
  int value;
  {
   IndentationDepth += value;
  }
  
/********************************************************/
/* DecrementIndentDepth:                                    */
/********************************************************/
globle int DecrementIndentDepth(value)
  int value;
  {
   IndentationDepth -= value;
   return(IndentationDepth);
  }
  
/********************************************************/
/* SetIndentDepth:                                    */
/********************************************************/
globle int SetIndentDepth(value)
  int value;
  {
   IndentationDepth = value;
   return(IndentationDepth);
  }
  
/***************************************************************/
/* SetPPBufferStatus: Sets PPBufferStatus flag to boolean */
/*   value of ON or OFF.                                       */
/***************************************************************/
globle VOID SetPPBufferStatus(value)
  int value;
  {
   PPBufferStatus = value;
  }
  
/*****************************************************************/
/* GetPPBufferStatus: Returns value of PPBufferStatus flag. */ 
/*****************************************************************/
globle int GetPPBufferStatus()
  {
   return(PPBufferStatus);
  }

/********************************************************/
/* StringPrintForm:                                     */
/********************************************************/
globle char *StringPrintForm(str)
  char *str;
  {
   int i = 0, pos = 0, max = 0;
   char *theString = NULL;
   VOID *thePtr;
   
   theString = ExpandStringWithChar('"',theString,&pos,&max,max+80);
   while (str[i] != EOS)
     {
      if ((str[i] == '"') || (str[i] == '\\'))
        { 
         theString = ExpandStringWithChar('\\',theString,&pos,&max,max+80);
         theString = ExpandStringWithChar(str[i],theString,&pos,&max,max+80);
        }
      else
        { theString = ExpandStringWithChar(str[i],theString,&pos,&max,max+80); }
      i++;
     }
    
   theString = ExpandStringWithChar('"',theString,&pos,&max,max+80);
   
   thePtr = AddSymbol(theString);
   rm(theString,max);
   return(ValueToString(thePtr));
  }
  
/********************************************************/
/* AppendStrings:                                     */
/********************************************************/
static char *AppendStrings(str1,str2)
  char *str1, *str2;
  {
   int pos = 0, max = 0;
   char *theString = NULL;
   VOID *thePtr;
   
   theString = AppendToString(str1,theString,&pos,&max);
   theString = AppendToString(str2,theString,&pos,&max);
      
   thePtr = AddSymbol(theString);
   rm(theString,max);
   return(ValueToString(thePtr));
  }

    
