   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 SYMBOL HEADER FILE                  */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_symbol
#define _H_symbol

struct symbolHashNode;
struct floatHashNode;
struct integerHashNode;
struct symbolMatch;

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _SYMBOL_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#define SYMBOL_HASH_SIZE        1013
#define FLOAT_HASH_SIZE          503 
#define INTEGER_HASH_SIZE        167

/************************************************************/
/* symbolHashNode STRUCTURE:                                */
/************************************************************/
struct symbolHashNode
  {
   struct symbolHashNode *next;
   int count;
   int depth;
   unsigned int markedEphemeral : 1;
   unsigned int neededSymbol : 1;
   unsigned int bucket : 14;
   char *contents;
   struct symbolHashNode *relatedSymbol;
  };
  
/************************************************************/
/* floatHashNode STRUCTURE:                                  */
/************************************************************/
struct floatHashNode
  {
   struct floatHashNode *next;
   int count;
   int depth;
   unsigned int markedEphemeral : 1;
   unsigned int neededFloat : 1;
   unsigned int bucket : 14;
   double contents;
  };

/************************************************************/
/* integerHashNode STRUCTURE:                               */
/************************************************************/
struct integerHashNode
  {
   struct integerHashNode *next;
   int count;
   int depth;
   unsigned int markedEphemeral : 1;
   unsigned int neededInteger : 1;
   unsigned int bucket : 14;
   long int contents;
  };

/************************************************************/
/* symbolMatch STRUCTURE:                               */
/************************************************************/
struct symbolMatch
  {
   struct symbolHashNode *match;
   struct symbolMatch *next;
  };
  
typedef struct symbolHashNode SYMBOL_HN;
typedef struct floatHashNode FLOAT_HN;
typedef struct integerHashNode INTEGER_HN;

#define ValueToString(target) (((struct symbolHashNode *) (target))->contents)
#define ValueToDouble(target) (((struct floatHashNode *) (target))->contents)
#define ValueToLong(target) (((struct integerHashNode *) (target))->contents)
#define ValueToInteger(target) ((int) (((struct integerHashNode *) (target))->contents))

#if ANSI_COMPILER
LOCALE VOID                          *AddSymbol(char *);
LOCALE SYMBOL_HN                     *FindSymbol(char *);
LOCALE VOID                          *AddDouble(double);
LOCALE VOID                          *AddLong(long int);
LOCALE VOID                           InitializeAtomTables(void);
LOCALE int                            HashSymbol(char *,int);
LOCALE int                            HashFloat(double,int);
LOCALE int                            HashInteger(long int,int);
LOCALE VOID                           IncrementSymbolCount(struct symbolHashNode *);
LOCALE VOID                           IncrementFloatCount(struct floatHashNode *);
LOCALE VOID                           IncrementIntegerCount(struct integerHashNode *);
LOCALE VOID                           DecrementSymbolCount(struct symbolHashNode *);
LOCALE VOID                           DecrementFloatCount(struct floatHashNode *);
LOCALE VOID                           DecrementIntegerCount(struct integerHashNode *);
LOCALE VOID                           RemoveEphemeralAtoms(void); 
LOCALE struct symbolHashNode        **GetSymbolTable(void);
LOCALE VOID                           SetSymbolTable(struct symbolHashNode **);
LOCALE struct floatHashNode          **GetFloatTable(void);
LOCALE VOID                           SetFloatTable(struct floatHashNode **);
LOCALE struct integerHashNode       **GetIntegerTable(void);
LOCALE VOID                           SetIntegerTable(struct integerHashNode **);
LOCALE VOID                           RefreshBooleanSymbols(void);
LOCALE struct symbolMatch            *FindSymbolMatches(char *,int *);
LOCALE VOID                           ReturnSymbolMatches(struct symbolMatch *);
LOCALE SYMBOL_HN                     *GetNextSymbolMatch(char *,int,SYMBOL_HN *);
#else
LOCALE VOID                          *AddSymbol();
LOCALE SYMBOL_HN                     *FindSymbol();
LOCALE VOID                          *AddDouble();
LOCALE VOID                          *AddLong();
LOCALE VOID                           InitializeAtomTables();
LOCALE int                            HashSymbol();
LOCALE int                            HashFloat();
LOCALE int                            HashInteger();
LOCALE VOID                           IncrementSymbolCount();
LOCALE VOID                           IncrementFloatCount();
LOCALE VOID                           IncrementIntegerCount();
LOCALE VOID                           DecrementSymbolCount();
LOCALE VOID                           DecrementFloatCount();
LOCALE VOID                           DecrementIntegerCount();
LOCALE VOID                           RemoveEphemeralAtoms();
LOCALE struct symbolHashNode        **GetSymbolTable();
LOCALE VOID                           SetSymbolTable();
LOCALE struct floatHashNode          **GetFloatTable();
LOCALE VOID                           SetFloatTable();
LOCALE struct integerHashNode       **GetIntegerTable();
LOCALE VOID                           SetIntegerTable();
LOCALE VOID                           RefreshBooleanSymbols();
LOCALE struct symbolMatch            *FindSymbolMatches();
LOCALE VOID                           ReturnSymbolMatches();
LOCALE SYMBOL_HN                     *GetNextSymbolMatch();
#endif


#ifndef _SYMBOL_SOURCE
   extern VOID                   *CLIPSTrueSymbol;
   extern VOID                   *CLIPSFalseSymbol;
#endif

#endif



