#include <stdio.h>
#include "crsv.h"

/********************************************************************
 *  This file contains all the functions used to process defexternal
 *  structures and store or retreive defexternal information.
 ********************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */
extern DE_PTR     AllocDefexternal();
extern DA_PTR     AllocDeffield();
extern WL_PTR     AllocWordList();
extern void       FreeWordList();
/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern int      GetToken();

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern void     ErrorMessage();
extern void     SendMessage();

/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern int      FindRhtParen();
extern WL_PTR   AddToWordList();
extern IL_PTR	AddToIntegerList();
extern FL_PTR	AddToFloatList();
extern int      IsWordInList();
extern int      IsIntegerInList();
extern int      IsFloatInList();

/* -----------------------------
 *  From the file: CRSVPROC.C
 * -----------------------------
 */

extern EX_FUNC  *FindExternalFunction();


/* -----------------------------
 * From the file : CRSVHASH.c
 * -----------------------------
 */
 extern HASH  *AddSymbol();

/* ---------------------------------------------------------------
 *   From the file: CRSVDEFREL.C
 *
 *   These functions are used for both defrelation and defexternal
 *   in order to process the arguments of an external function or
 *   the fields of a relation.
 * ---------------------------------------------------------------
 */
 extern void   ProcessFieldStringList(); /* Process the string list of
                                             defexternal                     */
 extern void   ProcessFieldNumberList(); /* Process the number list of
                                             defexternal                     */
 extern void   ProcessFieldRange();       /* Process the number range of
                                             defexternal                     */
 extern void   ProcessFieldType();        /* Process the argument type of
                                             defexternal                     */
 extern void   ProcessFieldWordList();   /* Process the word list of
                                             defexternal                     */
 extern int    CompareFieldTypes();
 extern int    CompareFieldWordLists();
 extern int    CompareFieldStringLists();
 extern int    CompareFieldNumbers();

/* -----------------------------
 *  From the system
 * -----------------------------
 */

extern double atof();


/* ===========  Functions defined here for Global use  ================ */

int      ProcessDefexternal();
void     CheckArgumentInfo();
void     CheckVariableArgumentInfo();

/* ===========  Functions defined here for internal use  ============== */

DEF_EXT   *FindCreateDefexternal();
DEF_EXT   *FindDefexternal();
DEF_EXT   *CreateDefexternal();
DEF_EXT   *InsertDefexternal();
int        ProcessDefargument();
DEF_ARG   *FindDefargument();
void       DuplicateDefexternal();


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern char        *CurrentObjectName;         /* Name of current object         */
extern FILE        *CurrentFile;             /* current file pointer           */
extern char        *CurrentFileName;        /* Name of the current file       */

extern DEF_EXT     *DefexternalHead;         /* First Element in Defexternal
                                             list                           */
extern EX_FUNC     *externalFunctionHead;        /* First element in external list */

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern long    INTNUMBER;                  /* gettoken number return  */
extern double  FLOATNUMBER;
extern char    *TKNWORD;                    /* gettoken string storage */


/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined here for internal use  ============== */

DEF_EXT *current_de;


/* ======================================================================= */
/*                  FUNCTIONS TO PROCESS DEFEXTERNALS                      */
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessDefexternal()
 *
 *  PURPOSE: This function reads and stores in the
 *           appropriate data structures all of the
 *           information defined in a defexternal
 *           construct.
 *
 *  INPUTS:  None.
 *
 *  RETURNS: An integer, -99 (RESTART) if it got
 *           hopelessly confused while trying to
 *           read the defexternal, a -1 if it found
 *           some kind of syntax error, otherwise, a
 *           1 (OK).
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

int ProcessDefexternal()
{
   int                    token;
   DEF_EXT                *CurDefexternal;
   HASH *temp_hash;

   /*============================*/
   /* Check for defexternal name */
   /*============================*/

   token = GetToken(CurrentFile);

   IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expected a name following defexternal keyword!");
      return(RESTART);
   END_IF

   CurrentObjectName  = TKNWORD;

   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf,"\n\nProcessing defexternal: %.40s",CurrentObjectName);
      SendMessage(MsgBuf,NO);
   END_IF

   /*============================================*/
   /* Check if function has been previously used */
   /*============================================*/

   IF(FindExternalFunction(externalFunctionHead, CurrentObjectName) NEQ NULL) THEN
      sprintf(MsgBuf, "External function %.40s defined after being used.\n",
                    CurrentObjectName);
      ErrorMessage(WARNING, MsgBuf);
   END_IF

   /*======================================*/
   /* Get pointer to defexternal structure */
   /*======================================*/
 
   if(FindDefexternal(DefexternalHead,CurrentObjectName) != NULL)
     {
       sprintf(MsgBuf, "External function %.40s is being redefined, ignore new defintion\n",
                    CurrentObjectName);
       ErrorMessage(ERROR, MsgBuf);
       return(RESTART);
     }
   CurDefexternal = CreateDefexternal(CurrentObjectName);

   /*===================================*/ 
   /* Begin processing defexternal info */
   /*===================================*/

   token = GetToken(CurrentFile);
   /*===================================*/
   /* Read and process defexternal info */
   /*===================================*/

   WHILE((token NEQ RPAREN) && (token != STOP))DO

      /*======================================*/
      /* Find opening paren, and get key word */
      /*======================================*/

     IF(token NEQ LPAREN) THEN
         ErrorMessage(ERROR, "Expecting a left paren inside a defexternal");
         return(RESTART);
     END_IF

     token = GetToken(CurrentFile);
     IF((token EQ WORD) AND
         (strcmp("true-function-name",TKNWORD) EQ 0)) THEN

           /*============================================*
            *  Process true function name                *
            *============================================*/
       if(ProcessTrueFuncInDefexternal(CurDefexternal) != OK)
        {
          FreeDefexternal(CurDefexternal);
          return(RESTART);
        }
           
     ELSE_IF((token EQ WORD) AND (strcmp("asserts",TKNWORD) EQ 0))

          /*===========================================================*
           *  Process the relations asserted by the external function  *
           *===========================================================*/
       if(ProcessAssertInDefexternal(CurDefexternal) != OK)
        {
          FreeDefexternal(CurDefexternal);
          return(RESTART);
        }
          
     ELSE_IF((token EQ WORD) AND (strcmp("retracts",TKNWORD) EQ 0))

          /*============================================================*
           *  Process the relations retracted by the external function  *
           *============================================================*/
       if(ProcessRetractInDefexternal(CurDefexternal) != OK)
        {
          FreeDefexternal(CurDefexternal);
          return(RESTART);
        }
        
      ELSE_IF((token EQ WORD) AND (strcmp("min-number-of-args", TKNWORD) EQ 0))

         /*=================================*/
         /* Process min-number-of-arguments */
         /*=================================*/
      if(ProcessMinNumberOfArgs(CurDefexternal) != OK)
       {
         FreeDefexternal(CurDefexternal);
         return(RESTART);
       }
         
      ELSE_IF((token EQ WORD) AND (strcmp("max-number-of-args", TKNWORD) EQ 0))

         /*=================================*/
         /* Process max-number-of-arguments */
         /*=================================*/
       if(ProcessMaxNumberOfArgs(CurDefexternal) != OK)
         {
           FreeDefexternal(CurDefexternal);
           return(RESTART);
         }
         
      ELSE_IF((token EQ WORD) AND  (strcmp("return-type",TKNWORD) == 0))

         /*================================*
          * Process return-type            *
          *================================*/
        if(ProcessReturnType(CurDefexternal) != OK)
         {
           FreeDefexternal(CurDefexternal);
           return(RESTART);
         }
          
      ELSE_IF((token EQ WORD) AND (strcmp("argument", TKNWORD) EQ 0))

         /*=============================== */
         /* Process an argument definition */
         /*=============================== */
         if(ProcessDefargument(CurDefexternal) != OK)
          {
            FreeDefexternal(CurDefexternal);
            return(RESTART);
          }

      ELSE

         /*========================*/
         /* Unrecognized key word! */
         /*========================*/

         sprintf(MsgBuf,
            "Unrecognized attribute: %.40s inside a defexternal", TKNWORD);
         ErrorMessage(ERROR, MsgBuf);
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);
   END_WHILE
   if(CurDefexternal->min_arguments > CurDefexternal->max_arguments)
     {
       ErrorMessage(ERROR,"Min. number of arguments is greater than Max. of arguments\n");
       FreeDefexternal(CurDefexternal);
       return(RESTART);
     }
   DefexternalHead = InsertDefexternal(DefexternalHead,CurDefexternal);
  return(OK);
}
 /* ------------------------------------------------------------------
 *  NAME   : FindDefexternal()
 *
 *  PURPOSE: This function searches through a list of
 *           defexternals to find a specific name.
 *
 *  INPUTS : A pointer to the list of defexternals (DE_PTR)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns NULL.
 * ------------------------------------------------------------------
 */

DE_PTR FindDefexternal(node,name)
char *name;
DE_PTR node;
{
   int rtn;

   IF(node EQ NULL) THEN                       /* A new defexternal! */
      return(NULL);
   ELSE
      rtn = strcmp(name, node->name);
      IF(rtn < 0) THEN
         return(FindDefexternal(node->lft_def_ext, name));
      ELSE_IF(rtn > 0)
         return(FindDefexternal(node->rht_def_ext, name));
      ELSE
         return(node);                         /* An old defrelation! */
      END_IF
   END_IF
}

/* ------------------------------------------------------------------
 *  NAME   : CreateDefexternal()
 *
 *  PURPOSE: This function creates a defexternal node.
 *
 *  INPUTS : a pointer to the name 
 *
 *  RETURNS: a new node
 * ------------------------------------------------------------------
 */
DE_PTR CreateDefexternal(name)
char   *name;
{
   DE_PTR node;                      /* A new defrelation! */
      
      node        = AllocDefexternal();
      node->name  = name;
      node->file  = CurrentFileName;
   return(node);
}
  
/* ------------------------------------------------------------------
 *  NAME   :  InsertDefexternal()
 *
 *  PURPOSE: This function inserts a defexternal in to the binary tree list.
 *
 *  INPUTS : a pointer pointing to the defexternal  being inserted
 *         : a poiter pointing to the root of the list
 *
 *  RETURNS: new root of the list
 *
 *  Note:    
 * ------------------------------------------------------------------
 */ 
 DE_PTR InsertDefexternal(list,node)
 DE_PTR list,node;
 {
    DE_PTR DefExtPtr = NULL,DefExtParent = NULL, NewDefExt;
    int rtn;
   
   if (list == NULL)
     return(node);
   DefExtPtr = list;
   while(DefExtPtr != NULL)
    {
       rtn = strcmp(node->name,DefExtPtr->name);
       if(rtn < 0)
        {
          DefExtParent = DefExtPtr;
          DefExtPtr = DefExtPtr->lft_def_ext;   
        }
       else if(rtn > 0)
        {
          DefExtParent = DefExtPtr;
          DefExtPtr = DefExtPtr->rht_def_ext;   
        }
       else{
         ErrorMessage(ERROR,"CRSV error in InsertDefexternal, Please report this error\n");
         return(list);}
    }
   if(rtn < 0)
      DefExtParent->lft_def_ext = node;
   else if(rtn > 0)
      DefExtParent->rht_def_ext = node;
   else
      ErrorMessage(ERROR,"CRSV error in InsertDefexternal, Please report this error\n");
   return(list);
 }

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindCreateDefexternal()
 *
 *  PURPOSE: This function searches through a list of
 *           defexternals to find a specific name. If it doesn't
 *           find an entry with that name, it creates one.
 *
 *  INPUTS : A pointer to the list of defexternals (DEF_EX_FUNC *)
 *           and a pointer to the name to search for (char *)
 *
 *  RETURNS: If the name is found, it returns a pointer to
 *           that structure. If it isn't found it returns a
 *           pointer to the newly created structure.
 * ------------------------------------------------------------------
 */

static DEF_EXT *FindCreateDefexternal(node, name,type)
DE_PTR  node;
char   *name;
int type;
{
   int rtn;

   IF(node EQ NULL) THEN                       /* A new defexternal! */
      node        = AllocDefexternal();
      node->name  = name;
      node->file  = CurrentFileName;
      current_de  = node;
   ELSE
      rtn = strcmp(name, node->name);
      IF(rtn < 0) THEN
         node->lft_def_ext = FindCreateDefexternal(node->lft_def_ext, name,type);
      ELSE_IF(rtn > 0)
         node->rht_def_ext = FindCreateDefexternal(node->rht_def_ext, name,type);
      ELSE_IF((rtn == 0) && (type != DEFFUNCTION) && 
              (type != DEFMETHOD) && (type != DEFGENERIC))
         DuplicateDefexternal(name);
         node->max_arguments = 10000;
         node->min_arguments = 0;
         node->file          = CurrentFileName;
         current_de          = node;
      END_IF
   END_IF

   return(node);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : DuplicateDefexternal()
 *
 *  PURPOSE: This function merely prints an error message if
 *           a duplicate external function definition is found.
 *
 *  INPUTS : A pointer to the external function name.
 *
 *  RETURNS: Nothing
 *
 *  NOTES  : This function is called to avoid putting a large
 *           string buffer on the stack many times.
 * ------------------------------------------------------------------
 */

static void DuplicateDefexternal(name)
char *name;
{
   sprintf(MsgBuf,
      "Duplicate function definition for <%.40s>, erasing old info", name);
   ErrorMessage(WARNING, MsgBuf);

   return;
}

/* ======================================================================= */
/* ------------------------------------------------------------------
 *  ProcessTrueFuncInDefexternal
 *  This function processes the true function name
 * Input:
 *   CurDefexternal : the current defexternal
 * Return:
 *   OK : if everything is OK
 *   RESTART : if found error
 * ------------------------------------------------------------------
 */
static int ProcessTrueFuncInDefexternal(CurDefexternal)
DE_PTR CurDefexternal;
{
  int token;
  HASH *TempHash;
     
    token = GetToken(CurrentFile);
    IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expecting a SYMBOL for ");
      SendMessage("true-function-name in defexternal",NO);
    ELSE
      TempHash = AddSymbol(TKNWORD);
      CurDefexternal->true_name = TempHash->contents;
    END_IF
      if(GetToken(CurrentFile) != RPAREN)
        {
          ErrorMessage(ERROR,"Check appropriate syntax for defexternal construct \n");
          return(RESTART);
        }
    return (OK);

}
/* ------------------------------------------------------------------
 *  ProcessRetractInDefexternal
 *  This function processes the retraction list of a defexternal
 * Input:
 *   CurDefexternal : the current defexternal
 * Return:
 *   OK : if everything is OK
 *   RESTART : if found error
 * ------------------------------------------------------------------
 */
static int ProcessRetractInDefexternal(CurDefexternal)
DE_PTR CurDefexternal;
{
  int token;
  struct word_list* relation;
  HASH *TempHash;
  
  token = GetToken(CurrentFile);
  IF(token == RPAREN) THEN
    ErrorMessage(ERROR,"Expecting a SYMBOL, a STRING, a INTEGER, a FLOAT, a INSTANCE, ?NONE or \n");
    SendMessage("?VARIABLE after the retracts attribute in a defexternal construct",NO);
    return(RESTART);
  END_IF
  while ((token NEQ RPAREN) && (token NEQ STOP)) DO
    IF(token EQ BWORD) THEN
       IF((strcmp(TKNWORD,"VARIABLE") NEQ 0) AND
          (strcmp(TKNWORD,"NONE") NEQ 0)) THEN
          ErrorMessage(ERROR,
           	"Unidentified word, expected VARIABLE or NONE after ?");
          return(RESTART);
       END_IF
    ELSE_IF((token NEQ WORD) AND (token NEQ STRING))
       ErrorMessage(ERROR, "Expecting a word, string, ?NONE, or \n");
       SendMessage("?VARIABLE after the retracts attribute in a defexternal construct",NO);
       return(RESTART);
    ELSE
       relation = AllocWordList();
       TempHash = AddSymbol(TKNWORD);
       relation->word = TempHash->contents;
       relation->next_word = CurDefexternal->retract_list;
       CurDefexternal->retract_list = relation;
    END_IF
    token = GetToken(CurrentFile);
  END_WHILE
  return(OK);

}
/* ------------------------------------------------------------------
 *  ProcessAssertInDefexternal
 *  This function processes the assertion list of a defexternal
 * Input:
 *   CurDefexternal : the current defexternal
 * Return:
 *   OK : if everything is OK
 *   RESTART : if found error
 * ------------------------------------------------------------------
 */
static int ProcessAssertInDefexternal(CurDefexternal)
DE_PTR CurDefexternal;
{
  int token;
  struct word_list *relation;
  HASH *TempHash;
  
  token = GetToken(CurrentFile);
  IF(token == RPAREN) THEN
    ErrorMessage(ERROR, "Expecting a SYMBOL STRING ?NONE  or \n");
               SendMessage("?VARIABLE after the asserts attribute in a defexternal construct",NO);
  END_IF

  while ((token NEQ RPAREN) && (token NEQ STOP)) DO
    IF(token EQ BWORD) THEN
       IF((strcmp(TKNWORD,"VARIABLE") NEQ 0) AND
         (strcmp(TKNWORD,"NONE") NEQ 0)) THEN
         ErrorMessage(ERROR,
           "Unidentified word, expected VARIABLE or NONE after ?");
         return(RESTART);
       END_IF
    ELSE_IF((token NEQ WORD) AND (token NEQ STRING))
      ErrorMessage(ERROR, "Expecting a SYMBOL, a STRING, a INTEGER, a FLOAT, a INSTANCE, ?NONE, or \n");
      SendMessage("?VARIABLE after the asserts attribute in a defexternal construct",NO);
      return(RESTART);
    ELSE
      relation = AllocWordList();
      TempHash = AddSymbol(TKNWORD);
      relation->word = TempHash->contents;
      relation->next_word = CurDefexternal->assert_list;
      CurDefexternal->assert_list = relation;
    END_IF
      token = GetToken(CurrentFile);
  END_WHILE
  return(OK);
}

/* ---------------------------------------------------------------------------
 * ProcessReturnType
 *   This function processes the "return-type" attribute in a defexternal
 * Input:
 *   CurDefexternal : the current defexternal
 * Return:
 *   OK : if everything is OK
 *   RESTART : if found error
 * ---------------------------------------------------------------------------
 */

static int ProcessReturnType(CurDefexternal)
DE_PTR CurDefexternal;
{

  int token;
  
  token = GetToken(CurrentFile);
   IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expecting INTEGER,FLOAT,BOOLEAN, STRING, WORD, ");
      SendMessage(" EXTERNAL-ADDRESS, MULTI, or NONE for return-type attribute",NO);
      return(RESTART);
   ELSE
     IF(strcmp(TKNWORD,"MULTI") EQ 0) THEN
      CurDefexternal->return_word    = YES;
      CurDefexternal->return_string  = YES;
      CurDefexternal->return_integer = YES;
      CurDefexternal->return_float   = YES;
      CurDefexternal->return_boolean = YES;
      CurDefexternal->return_address = YES;
     ELSE_IF(strcmp(TKNWORD,"NONE") EQ 0)
      CurDefexternal->return_word    = NO;
      CurDefexternal->return_string  = NO;
      CurDefexternal->return_integer = NO;
      CurDefexternal->return_float   = NO;
      CurDefexternal->return_boolean = NO;
      CurDefexternal->return_address = NO;
     ELSE_IF(strcmp(TKNWORD,"BOOLEAN") EQ 0)
      CurDefexternal->return_boolean = YES;
     ELSE_IF(strcmp(TKNWORD,"STRING") EQ 0)
      CurDefexternal->return_string = YES;
     ELSE_IF(strcmp(TKNWORD,"WORD") EQ 0)
      CurDefexternal->return_word = YES;
     ELSE_IF(strcmp(TKNWORD,"FLOAT") EQ 0)
      CurDefexternal->return_float = YES;
     ELSE_IF ( strcmp(TKNWORD,"INTEGER")EQ 0)
      CurDefexternal->return_integer = YES;
     ELSE_IF(strcmp(TKNWORD,"NUMBER") EQ 0)
      CurDefexternal->return_float = YES;
      CurDefexternal->return_integer = YES;
     ELSE_IF(strcmp(TKNWORD,"EXTERNAL-ADDRESS") EQ 0)
      CurDefexternal->return_address = YES;
     ELSE
      ErrorMessage(ERROR, "Expecting NUMBER,INTEGER,FLOAT,BOOLEAN, STRING, WORD, ");
      SendMessage(" EXTERNAL-ADDRESS, MULTI, or NONE for return-type attribute",NO);
      return(RESTART);
     END_IF
  END_IF
  if(GetToken(CurrentFile) != RPAREN)
    {
     ErrorMessage(ERROR,"Check appropriate syntax for defexternal construct \n");
     return(RESTART);
    }
  return(OK);
}
/* ----------------------------------------------------------------------
 * ProcessMinNumberOfArgs
 * This function processes the minimum number of arguments in a 
 * defexternal
 * Input:
 *  CurDefexternal : the current defexternal
 * REturn :
 *   Ok if everything is OK
 *   RESTART if found an error
 * ----------------------------------------------------------------------
 */
 static int ProcessMinNumberOfArgs(CurDefexternal)
 DE_PTR CurDefexternal;
{
   int token;
   token = GetToken(CurrentFile);
   IF((token EQ BWORD) AND (strcmp("VARIABLE", TKNWORD) EQ 0)) THEN
     CurDefexternal->min_arguments = 0;
   ELSE_IF(token NEQ INTEGER)
     ErrorMessage(ERROR, 
     "Expecting an integer for min-number-of-arguments attribute in defexternal");
     return(RESTART);
   ELSE
     CurDefexternal->min_arguments = (int)INTNUMBER;
   END_IF
   if(GetToken(CurrentFile) != RPAREN)
     {
        ErrorMessage(ERROR,"Check appropriate syntax for defexternal construct \n");
        return(RESTART);
     }
   return(OK);
}

/* ----------------------------------------------------------------------
 * ProcessMaxNumberOfArgs
 * This function processes the maximum number of arguments in a 
 * defexternal
 * Input:
 *  CurDefexternal : the current defexternal
 * REturn :
 *   Ok if everything is OK
 *   RESTART if found an error
 * ----------------------------------------------------------------------
 */
 static int ProcessMaxNumberOfArgs(CurDefexternal)
 DE_PTR CurDefexternal;
 {
   int token;

   token = GetToken(CurrentFile);
   IF((token EQ BWORD) AND (strcmp("VARIABLE", TKNWORD) EQ 0)) THEN
     CurDefexternal->max_arguments = 10000;
   ELSE_IF(token NEQ INTEGER)
     ErrorMessage(ERROR, 
        "Expecting an integer for max-number-of-arguments attribute in defexternal");
     return(RESTART);
   ELSE
     CurDefexternal->max_arguments = (int)INTNUMBER;
   END_IF
   if(GetToken(CurrentFile) != RPAREN)
     {
       ErrorMessage(ERROR,"Check appropriate syntax for defexternal construct \n");
       return(RESTART);
     }
   return(OK);
  }
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : ProcessDefargument()
 *
 *  PURPOSE: This function reads and stores in the
 *           appropriate data structures all of the
 *           information about arguments from a defexternal.
 *
 *  INPUTS:  A pointer to the current defexternal.
 *
 *  RETURNS: An integer, either OK or ERROR.
 *
 *  NOTES:   None.
 * ------------------------------------------------------------------
 */

static int ProcessDefargument(CurDefexternal)
DEF_EXT *CurDefexternal;
{
   int token;
   DA_PTR CurDefargument;

   token = GetToken(CurrentFile);

   /*=======================*/
   /* Check position number */
   /*=======================*/

   IF(token NEQ INTEGER) THEN
      ErrorMessage(ERROR, "Argument keyword must be followed by an integer");
      return(RESTART);
   ELSE_IF((INTNUMBER < 1) OR (INTNUMBER > CurDefexternal->max_arguments))
      sprintf(MsgBuf,"Argument position, %d, is too big or too small",
                  INTNUMBER);
      ErrorMessage(ERROR, MsgBuf);
      return(RESTART);
   END_IF

   /*=================================================*/
   /* Check if argument postion has already been used */
   /*=================================================*/

   CurDefargument = FindDefargument((int)INTNUMBER, CurDefexternal->arguments);

   IF(CurDefargument NEQ NULL) THEN
      ErrorMessage(WARNING, "Duplicate argument position, redefining values");

      FreeWordList(CurDefargument->possible_words);
      FreeWordList(CurDefargument->possible_strings);
      FreeIntegerList(CurDefargument->possible_integers);
      FreeFloatList(CurDefargument->possible_floats);
      FreeWordList(CurDefargument->possible_instances);

      CurDefargument->possible_words    = NULL;
      CurDefargument->possible_strings  = NULL;
      CurDefargument->possible_integers = NULL;
      CurDefargument->possible_floats   = NULL;
      CurDefargument->possible_instances = NULL;
      CurDefargument->allow_instance    = YES;
      CurDefargument->allow_word        = YES;
      CurDefargument->allow_string      = YES;
      CurDefargument->allow_integer     = YES;
      CurDefargument->allow_float       = YES;
      CurDefargument->set_int_max       = NO;
      CurDefargument->set_int_min       = NO;
      CurDefargument->set_float_max       = NO;
      CurDefargument->set_float_min       = NO;
   ELSE
      CurDefargument             = AllocDeffield();
      CurDefargument->position   = (int)INTNUMBER;
      CurDefargument->next_field = CurDefexternal->arguments;
      CurDefexternal->arguments     = CurDefargument;
   END_IF


   /*==========================================*/
   /* Argument found, start examining keywords */
   /*==========================================*/

   token = GetToken(CurrentFile);

   while((token NEQ RPAREN) && (token NEQ STOP))DO

      /*======================================*/
      /* Find opening paren, and get key word */
      /*======================================*/

      IF(token NEQ LPAREN) THEN
         ErrorMessage(ERROR,
            "Expecting a left paren inside argument definition");
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);

      /*===========================*/
      /* Process argument keywords */
      /*===========================*/

      IF((token EQ WORD) AND (strcmp("type", TKNWORD) EQ 0)) THEN
         ProcessFieldType(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-strings", TKNWORD) EQ 0))
		   ProcessFieldStringList(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-words", TKNWORD) EQ 0))
         ProcessFieldWordList(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-instances",TKNWORD) EQ 0))
         ProcessFieldInstanceList(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-numbers", TKNWORD) EQ 0))
          ProcessFieldNumberList(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-integers", TKNWORD) EQ 0))
         IF(CurDefargument->allow_integer IS_YES) THEN
           ErrorMessage(WARNING,
               "ALLOWED_INTEGER has already been defined");
           if(FindRhtParen() != OK)
             return(RESTART);
         ELSE
           ProcessFieldIntegerList(CurDefargument);
         END_IF
      ELSE_IF((token EQ WORD) AND (strcmp("allowed-floats", TKNWORD) EQ 0))
           ProcessFieldFloatList(CurDefargument);
      ELSE_IF((token EQ WORD) AND (strcmp("range", TKNWORD) EQ 0))
            ProcessFieldRange(CurDefargument);

      ELSE

         /*========================*/
         /* Unrecognized key word! */
         /*========================*/

         sprintf(MsgBuf,
            "Unrecognized key word: %.40s inside a argument definition",
            TKNWORD);
         ErrorMessage(ERROR, MsgBuf);
         return(RESTART);
      END_IF

      token = GetToken(CurrentFile);
   END_WHILE
   CheckDefField(CurDefargument);
  return(OK);
}

/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : FindDefargument()
 *
 *  PURPOSE: This function searches through a list of
 *           argument definitions to find a specific argument number.
 *
 *  INPUTS : A pointer to the list of argument definitions (DA_PTR *)
 *           and a argument number to search for (int).
 *
 *  RETURNS: If the number is found, it returns a pointer to
 *           that structure. If it isn't found, it returns NULL.
 * ------------------------------------------------------------------
 */

static DA_PTR FindDefargument(number, list)
int number;
DA_PTR   list;
{
   DA_PTR temp;

   temp = list;
   while(temp NEQ NULL) DO
      IF(number EQ temp->position) THEN
         return(temp);
      END_IF

      temp = temp->next_field;
   END_WHILE

   return(NULL);
}




/* ------------------------------------------------------------------
 *  NAME   : CopyArgumentInfo()
 *
 *  PURPOSE: This function copies the argument info from the current
 *           argument into a new def_argument structure and returns a
 *           pointer to that structure.
 *
 *  INPUTS:  Two arguments,
 *           A pointer to the function name, (char *)
 *           the position in the pattern, (int)
 *
 *  RETURNS: A pointer to the new structure (DEF_ARG *)
 *
 *  NOTES:   Does not check argument number here!
 * ------------------------------------------------------------------
 */

DEF_ARG *CopyArgumentInfo(ex_func_name,argument_pos)
   char *ex_func_name;
   int   argument_pos;
{
   DEF_ARG *new_def_arg,*cur_arg;
   DEF_EXT *ex_func;
   WL_PTR   wl_ptr;
   IL_PTR   il_ptr;
   FL_PTR   fl_ptr;

   /*===========================*/
   /* Find defexternal (if any) */
   /*===========================*/

   ex_func = FindDefexternal(DefexternalHead, ex_func_name);

   /*=======================================*/
   /* No defexternal, flag error and return */
   /*=======================================*/

   IF(ex_func EQ NULL) THEN
      DefexternalHead = FindCreateDefexternal(DefexternalHead, ex_func_name,DEFEXTERNAL);
      sprintf(MsgBuf,
         "External function <%.40s> used prior to its definition.",
         ex_func_name);
      ErrorMessage(WARNING, MsgBuf);
      return(NULL);
   END_IF

   cur_arg = FindDefargument(argument_pos, ex_func->arguments);
   IF(cur_arg EQ NULL) THEN

      /*================================*/
      /* This argument not defined, return */
      /*================================*/

      return(NULL);
   END_IF

   new_def_arg = AllocDeffield();

   /*================*/
   /* Copy type info */
   /*================*/

   new_def_arg->allow_word   = cur_arg->allow_word;
   new_def_arg->allow_string = cur_arg->allow_string;
   new_def_arg->allow_instance = cur_arg->allow_instance;
   new_def_arg->allow_integer = cur_arg->allow_integer;
   new_def_arg->allow_float = cur_arg->allow_float;
   new_def_arg->allow_address = cur_arg->allow_address;

   /*================*/
   /* Copy word list */
   /*================*/

   for (wl_ptr = cur_arg->possible_words; wl_ptr NEQ NULL;
         wl_ptr = wl_ptr->next_word) DO
      new_def_arg->possible_words =
         AddToWordList(new_def_arg->possible_words, wl_ptr->word, WORD,NO);
   END_FOR

   /*==================*/
   /* Copy string list */
   /*==================*/

   for (wl_ptr = cur_arg->possible_strings; wl_ptr NEQ NULL;
         wl_ptr = wl_ptr->next_word) DO
      new_def_arg->possible_strings =
         AddToWordList(new_def_arg->possible_strings, wl_ptr->word, STRING,
            NO);
   END_FOR

   /*====================*/
   /* Copy instance list */
   /*====================*/

   for (wl_ptr = cur_arg->possible_instances; wl_ptr NEQ NULL;
         wl_ptr = wl_ptr->next_word) DO
      new_def_arg->possible_instances =
         AddToWordList(new_def_arg->possible_instances, wl_ptr->word, STRING,
            NO);
   END_FOR
   /*==================*/
   /* Copy number list */
   /*==================*/

   for (il_ptr = cur_arg->possible_integers; il_ptr NEQ NULL;
         il_ptr = il_ptr->next_number) DO
      new_def_arg->possible_integers =
         AddToIntegerList(new_def_arg->possible_integers, il_ptr->number, NO);
   END_FOR
   
   for (fl_ptr = cur_arg->possible_floats; fl_ptr NEQ NULL;
         fl_ptr = fl_ptr->next_number) DO
      new_def_arg->possible_floats =
         AddToFloatList(new_def_arg->possible_floats, fl_ptr->number, NO);
   END_FOR

   /*=================*/
   /* Copy range info */
   /*=================*/

   new_def_arg->set_int_max     = cur_arg->set_int_max;
   new_def_arg->set_int_min     = cur_arg->set_int_min;
   new_def_arg->set_float_max   = cur_arg->set_float_max;
   new_def_arg->set_float_min   = cur_arg->set_float_min;
   new_def_arg->int_max_range   = cur_arg->int_max_range;
   new_def_arg->int_min_range   = cur_arg->int_min_range;
   new_def_arg->float_max_range = cur_arg->float_max_range;
   new_def_arg->float_min_range = cur_arg->float_min_range;

   return(new_def_arg);
}


/* ======================================================================= */
/*                  FUNCTIONS TO CHECK DEFEXTERNAL INFO                    */
/* ======================================================================= */

/* ------------------------------------------------------------------
 *  NAME   : CheckArgumentInfo()
 *
 *  PURPOSE: This function compares info from functions
 *           used in a rule or deffact to the defexternal
 *           paramaters and prints appropriate error
 *           messages.
 *
 *  INPUTS:  Four arguments,
 *           A pointer to the function name, (char *)
 *           the type of the value, WORD, STRING, INSTANCE, NUMBER,(int)
 *           the position in the pattern, (int)
 *           and the literal value of the argument (char *).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   Does not check argument number here!
 * ------------------------------------------------------------------
 */

void CheckArgumentInfo(name, type, position, value)
char *name, *value;
int   type,  position;
{
   DEF_EXT  *def_ex_func;
   DEF_ARG  *arg;
   int       intnumber;
   float     floatnumber;

   /*===========================*/
   /* Find defexternal (if any) */
   /*===========================*/

   def_ex_func = FindDefexternal(DefexternalHead, name);

   /*=======================================*/
   /* No defexternal, flag error and return */
   /*=======================================*/

   IF(def_ex_func EQ NULL) THEN
      DefexternalHead = FindCreateDefexternal(DefexternalHead, name,DEFEXTERNAL);
      return;
   END_IF

   /*=====================*/
   /* Check argument info */
   /*=====================*/
   arg = FindDefargument(position, def_ex_func->arguments);
   IF(arg EQ NULL) THEN

      /*===================================*/
      /* This argument not defined, return */
      /*===================================*/

      return;
   ELSE

      /*========================================*/
      /* Check argument type and literal values */
      /*========================================*/

      switch (type) {

         case WORD:
              /*=====================*/
              /* Check WORDS allowed */
              /*=====================*/

              IF(arg->allow_word NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used a SYMBOL in argument %d, which isn't allowed",
                    position);
                 ErrorMessage(ERROR, MsgBuf);
                 sprintf(MsgBuf, "\n        by defexternal %.40s", name);
                 SendMessage(MsgBuf,NO);

              /*=================*/
              /* Check WORD list */
              /*=================*/

              ELSE_IF(arg->possible_words NEQ NULL)
                 IF(IsWordInList(arg->possible_words, value) IS_NO) THEN
                    sprintf(MsgBuf,
                       "Used a SYMBOL (%.40s) in argument %d that is not in the",
                            value, position);
                    ErrorMessage(ERROR, MsgBuf);
                    SendMessage("\n        ALLOWED_WORDS list",NO);
                 END_IF
              END_IF
              break;


         case STRING:
              /*=======================*/
              /* Check STRINGS allowed */
              /*=======================*/

              IF(arg->allow_string NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used a STRING in argument %d, which isn't allowed",
                    position);
                 ErrorMessage(ERROR,MsgBuf);
                 sprintf(MsgBuf, "\n        by defexternal %.40s", name);
                 SendMessage(MsgBuf,NO);

              /*====================*/
              /* Check STRINGS list */
              /*====================*/

              ELSE_IF(arg->possible_strings NEQ NULL)
                 IF(IsWordInList(arg->possible_strings, value) IS_NO) THEN
                    sprintf(MsgBuf,
                    "Used a STRING (%.40s) in argument %d that is not in the",
                            value, position);
                    ErrorMessage(ERROR, MsgBuf);
                    SendMessage("\n        ALLOWED_STRINGS list",NO);
                 END_IF
              END_IF
              break;

         case INSTANCE:
              /*=========================*/
              /* Check INSTANCES allowed */
              /*=========================*/

              IF(arg->allow_instance NEQ YES) THEN
                 sprintf(MsgBuf,
                    "Used an INSTANCE in argument %d, which isn't allowed",
                    position);
                 ErrorMessage(ERROR,MsgBuf);
                 sprintf(MsgBuf, "\n        by defexternal %.40s", name);
                 SendMessage(MsgBuf,NO);

              /*======================*/
              /* Check INSTANCES list */
              /*======================*/

              ELSE_IF(arg->possible_instances NEQ NULL)
                 IF(IsWordInList(arg->possible_instances, value) IS_NO) THEN
                    sprintf(MsgBuf,
                    "Used an INSTANCE (%.40s) in argument %d that is not in the",
                            value, position);
                    ErrorMessage(ERROR, MsgBuf);
                    SendMessage("\n        ALLOWED_INSTANCES list",NO);
                 END_IF
              END_IF
              break;

         case INTEGER:
              /*=======================*/
              /* Check INTEGERS allowed */
              /*=======================*/

              IF(arg->allow_integer NEQ YES) THEN
                 sprintf(MsgBuf,
                         "Used an INTEGER in argument %d, which isn't allowed",
                         position);
                 ErrorMessage(ERROR,MsgBuf);
                 sprintf(MsgBuf, "\n        by defexternal %.40s", name);
                 SendMessage(MsgBuf,NO);
              ELSE
                 /*====================*/
                 /* Check INTEGERS list */
                 /*====================*/

                 intnumber = atoi(value);
                 IF(arg->possible_integers NEQ NULL) THEN
                    IF(IsIntegerInList(arg->possible_integers,
                                         intnumber) IS_NO) THEN
                       sprintf(MsgBuf,
                          "Used an INTEGER (%.40s) in argument %d that is not ",
                          value, position);
                       ErrorMessage(ERROR, MsgBuf);
                       SendMessage("in the\n        ALLOWED_INTEGERS list",NO);
                    END_IF
                 END_IF

                 /*=====================*/
                 /* Check INTEGERS range */
                 /*=====================*/

                 IF(arg->set_int_max EQ YES) THEN
                    IF(intnumber > arg->int_max_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%d) in argument %d greater than range max (%d)", 
                               intnumber, position, arg->int_max_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF

                 IF(arg->set_int_min EQ YES) THEN
                    IF(intnumber < arg->int_min_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%d) in argument %d smaller than range min (%d)", 
                               intnumber, position, arg->int_min_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF
              END_IF

              break;
              
         case FLOAT:
              /*=======================*/
              /* Check FLOATS allowed */
              /*=======================*/

              IF(arg->allow_float NEQ YES) THEN
                 sprintf(MsgBuf,
                         "Used a FLOAT in argument %d, which isn't allowed",
                         position);
                 ErrorMessage(ERROR,MsgBuf);
                 sprintf(MsgBuf, "\n        by defexternal %.40s", name);
                 SendMessage(MsgBuf,NO);
              ELSE
                 /*====================*/
                 /* Check FLOATS list */
                 /*====================*/

                 floatnumber = atof(value);
                 IF(arg->possible_floats NEQ NULL) THEN
                    IF(IsFloatInList(arg->possible_floats,
                                         floatnumber) IS_NO) THEN
                       sprintf(MsgBuf,
                          "Used a FLOAT (%.40s) in argument %d that is not ",
                          value, position);
                       ErrorMessage(ERROR, MsgBuf);
                       SendMessage("in the\n        ALLOWED_FLOATS list",NO);
                    END_IF
                 END_IF

                 /*=====================*/
                 /* Check FLOATS range */
                 /*=====================*/

                 IF(arg->set_float_max EQ YES) THEN
                    IF(floatnumber > arg->float_max_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%f) in argument %d greater than range max (%f)", 
                               floatnumber, position, arg->float_max_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF

                 IF(arg->set_float_min EQ YES) THEN
                    IF(floatnumber < arg->float_min_range) THEN
                       sprintf(MsgBuf,
                          "Used a value (%f) in argument %d smaller than range min (%f)", 
                               floatnumber, position, arg->float_min_range);
                       ErrorMessage(ERROR, MsgBuf);
                    END_IF
                 END_IF
              END_IF
            break;
      case UNKNOWN:
        ErrorMessage(WARNING,"Global variable is used, and its type is unknown\n");
        SendMessage("        Unable to verify with the definition\n");
        break;
         }
   END_IF
}

/* ------------------------------------------------------------------
 *  NAME   : CheckVariableArgumentInfo()
 *
 *  PURPOSE: This function checks to see that a variable is used
 *           in a argument whose defexternal info is consistent with
 *           the info from arguments in which the variable was used
 *           previously.
 *
 *  INPUTS:  Four arguments,
 *           A pointer to the function name, (char *)
 *           a pointer to the variable name, (char *),
 *           a pointer to the structure containing the cumulative
 *              argument info for that variable (DEF_ARG *),
 *           and the position in the pattern, (int).
 *
 *  RETURNS: Nothing.
 *
 *  NOTES:   Does not check argument number here!
 * ------------------------------------------------------------------
 */

void CheckVariableArgumentInfo(ex_func_name, var_name, var_arg, position,side)
   char    *ex_func_name, *var_name;
   DEF_ARG *var_arg;
   int      position;
   short    side;
{
   DEF_EXT  *def_ex_func;
   DEF_ARG  *arg;
   int       found_common;
   int       num_in_common,tot_num_in_common = 0;

   /*===========================*/
   /* Find defexternal (if any) */
   /*===========================*/

   def_ex_func = FindDefexternal(DefexternalHead, ex_func_name);

   /*=======================================*/
   /* No defexternal, flag error and return */
   /*=======================================*/

   IF(def_ex_func EQ NULL) THEN
      DefexternalHead = FindCreateDefexternal(DefexternalHead, ex_func_name,DEFEXTERNAL);
      sprintf(MsgBuf,
         "check_var_argument External function <%.40s> used prior to its definition.",
         ex_func_name);
      ErrorMessage(WARNING, MsgBuf);
      return;
   END_IF

   /*=====================*/
   /* Check argument info */
   /*=====================*/

   arg = FindDefargument(position, def_ex_func->arguments);
   IF(arg EQ NULL) THEN

      /*===================================*/
      /* This argument not defined, return */
      /*===================================*/

      return;
   END_IF

   /*============================*/
   /* Compare data types allowed */
   /*============================*/

   found_common = CompareFieldTypes(var_arg,arg);
   IF(found_common IS_NO) THEN
      IF(side == LHS) THEN
         ErrorMessage(WARNING,
           "Due to the inconsistency of the field definitions, this rule MAY NOT fire ");
         sprintf(MsgBuf, "\n        variable <%.40s> has incompatible type definitions",
            var_name);
         SendMessage(MsgBuf,YES);
      ELSE
         sprintf(MsgBuf,
            "Variable <%.40s> is of the wrong type for argument %d ",
            var_name, arg->position);
         ErrorMessage(ERROR, MsgBuf);
         sprintf(MsgBuf,"\n        of the defexternal <%.40s>", ex_func_name);
         SendMessage(MsgBuf,NO);	
      END_IF

   /*===========================================================*/
   /* If data types are not exclusive, look at lists and ranges */
   /*===========================================================*/

   ELSE

      /*====================*/
      /* Compare word lists */
      /*====================*/

      num_in_common = CompareFieldWordLists(var_name,var_arg,arg);
      if (num_in_common == 0)
       {
          sprintf(MsgBuf,
          "Variable <%.40s> used in arguments whose ALLOWED_WORDS definitions", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
          SendMessage("\n        have no common values",NO);
      }
     else if (num_in_common > 0)
      tot_num_in_common += num_in_common;
      /*======================*/
      /* Compare string lists */
      /*======================*/

      num_in_common = CompareFieldStringLists(var_name,var_arg,arg);
      if (num_in_common == 0)
       {
          sprintf(MsgBuf,
          "Variable <%.40s> used in arguments whose ALLOWED_STRINGS definitions", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
          SendMessage("\n        have no common values",NO);
      }
     else if (num_in_common > 0)
      tot_num_in_common += num_in_common;
      
      /*======================*/
      /* Compare instance lists */
      /*======================*/

      num_in_common = CompareFieldInstanceLists(var_name,var_arg,arg);
      if (num_in_common == 0)
       {
          sprintf(MsgBuf,
          "Variable <%.40s> used in arguments whose ALLOWED_INSTANCES definitions", 
              var_name);
          ErrorMessage(ERROR, MsgBuf);
          SendMessage("\n        have no common values",NO);
      }
     else if (num_in_common > 0)
      tot_num_in_common += num_in_common;
      /*=================*/
      /* Compare numbers */
      /*=================*/

     num_in_common = CompareFieldNumbers(var_name,var_arg,arg);
     if (num_in_common > 0)
      tot_num_in_common += num_in_common;
      /*=====================================================*/
      /* If no common values found in ANY of the categories, */
      /*    print error message                              */
      /*=====================================================*/

      IF((tot_num_in_common EQ 0) AND (side EQ LHS)) THEN
	     ErrorMessage(ERROR,
                "Argument definitions won't allow this rule to fire since ");
	     sprintf(MsgBuf,
                "variable\n        <%.40s> has no common allowed values",
                var_name);
	     SendMessage(MsgBuf,NO);

      /*============================================*/
      /* Otherwise, if only one common value found, */
      /*    print warning message                   */
      /*============================================*/

      ELSE_IF((tot_num_in_common EQ 1) AND (side EQ LHS))
         sprintf(MsgBuf,
         "Argument definitions allow variable <%.40s> only one possible value",
                 var_name);
         ErrorMessage(WARNING,MsgBuf);
      END_IF
   END_IF
}


/*
 *============================================================
 * ProcessDefgeneric
 *   This function only adds the name of the generic function
 *   in the defexternal list so CRSV would not mistaken it for 
 *   an undefined function.
 *=============================================================
 */
 int ProcessDefgeneric()
 {
     
   int                    token;

   /*============================*/
   /* Check for defexternal name */
   /*============================*/

   token = GetToken(CurrentFile);

   IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expected a name following defgeneric keyword!");
      return(RESTART);
   END_IF

   CurrentObjectName  = TKNWORD;

   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf,"\n\nProcessing defgeneric: %.40s",CurrentObjectName);
      SendMessage(MsgBuf,NO);
   END_IF

   /*======================================*/
   /* Get pointer to defexternal structure */
   /*======================================*/
   DefexternalHead = FindCreateDefexternal(DefexternalHead, CurrentObjectName,DEFGENERIC);
   return(FindRhtParen());
 }
 
 
 /*
 *============================================================
 * ProcessDeffunction
 *   This function only adds the name of the function
 *   in the defexternal list so CRSV would not mistaken it for 
 *   an undefined function.
 *=============================================================
 */
 
 int ProcessDeffunction()
 {
    
   int                    token;

   /*============================*/
   /* Check for defexternal name */
   /*============================*/

   token = GetToken(CurrentFile);

   IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expected a name following deffunction keyword!");
      return(RESTART);
   END_IF

   CurrentObjectName  = TKNWORD;

   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf,"\n\nProcessing deffunction: %.40s",CurrentObjectName);
      SendMessage(MsgBuf,NO);
   END_IF

   /*======================================*/
   /* Get pointer to defexternal structure */
   /*======================================*/
   DefexternalHead = FindCreateDefexternal(DefexternalHead, CurrentObjectName,DEFFUNCTION);
   return(FindRhtParen());
   
 }
 
 
 
 /*
 *============================================================
 * ProcessDefmethod
 *   This function only adds the name of the method
 *   in the defexternal list so CRSV would not mistaken it for 
 *   an undefined function.
 *=============================================================
 */
 
 int ProcessDefmethod()
 {
    
   int                    token;

   /*============================*/
   /* Check for defexternal name */
   /*============================*/

   token = GetToken(CurrentFile);

   IF(token NEQ WORD) THEN
      ErrorMessage(ERROR, "Expected a name following defmethod keyword!");
      return(RESTART);
   END_IF

   CurrentObjectName  = TKNWORD;

   IF(VERBOSE IS_ON) THEN
      sprintf(MsgBuf,"\n\nProcessing defmethod: %.40s",CurrentObjectName);
      SendMessage(MsgBuf,NO);
   END_IF

   /*======================================*/
   /* Get pointer to defexternal structure */
   /*======================================*/
   DefexternalHead = FindCreateDefexternal(DefexternalHead, CurrentObjectName,DEFMETHOD);
   return(FindRhtParen());
   
 }