#include <stdio.h>
#include "crsv.h"
#include <string.h>

/******************************************************
 *  This file contains the functions used to store
 *  strings in CRSV, a basic hash table.
 ******************************************************
 */


/* ===========  Functions defined Externally but used here  =========== */

/*=============================
 * in CRSV.C
 *=============================
 */
 
extern void EndCRSV();

/*=============================
 * in CRSVPRNT.C
 *=============================
 */
 
extern void SendMessage();

/*=============================
 * in CRSVMEM.C
 *=============================
 */
 
extern char      *GenAlloc();
extern void       GenFree();
extern char      *gm2();
extern int        rm();


/* ===========  Functions defined here for Global use  ================ */

HASH     *AddSymbol();
void      InitSymbolTable();


/* ===========  Functions defined here for internal use  ============== */

struct MemPtr
  {
   struct MemPtr *next;
  };

struct MemPtr *TempMemPtr;

#define GetStruct(type) \
  ((MemoryTable[sizeof(struct type)] == NULL) ? \
   ((struct type *) GenAlloc(sizeof(struct type),"AddSymbol")) :\
   ((TempMemPtr = MemoryTable[sizeof(struct type)]),\
    MemoryTable[sizeof(struct type)] = TempMemPtr->next,\
    ((struct type *) TempMemPtr)))
    
#define rtn_struct(type,struct_ptr) \
  (TempMemPtr = (struct MemPtr *) struct_ptr,\
   TempMemPtr->next = MemoryTable[sizeof(struct type)], \
   MemoryTable[sizeof(struct type)] = TempMemPtr)
   
static int       HashSymbol();
static void      ClearSymbolTable();

/* ===========  Variables defined here for Global use  ================ */

/* ===========  Variables defined Externally but used here  =========== */


/*=============================
 * in CRSVMEM.C
 *=============================
 */
 
extern struct MemPtr *MemoryTable[];


/* ===========  Variables defined here for internal use  ============== */

static HASH       **SymbolTable; 


/* ======================================================================== */

   
/*####################################################################*/
/*####################################################################*/
/*###                                                              ###*/
/*###                      HASHING FUNCTIONS                       ###*/
/*###                                                              ###*/
/*####################################################################*/
/*####################################################################*/
	  
/************************************************************************/
/* AddSymbol:  Searches for the string in the hash table. If the       */ 
/*   string is already in the hash table, then the address of the       */
/*   string hash is returned.  Otherwise, the string is hashed into the */
/*   table and the address of the string hash is also returned.         */
/************************************************************************/
HASH *AddSymbol(str)
   char *str;
   {
    int tally;
    unsigned int length;
    HASH *past, *peek;
    
    /*====================================*/
    /* Get the hash value for the string. */
    /*====================================*/

    if (str == NULL)
      {
       SendMessage("\n**ERROR**  Null string passed to AddSymbol",NO);
       EndCRSV();
      }

    tally = HashSymbol(str,HASHSIZE);
    peek  = SymbolTable[tally];

    /*==================================================*/
    /* If the hash table entry for the string is empty, */
    /* then place the string at that location, and      */
    /* return the address of the string.                */
    /*==================================================*/

    if (peek == NULL)
      {
       peek = GetStruct(hash_entry); 
       SymbolTable[tally] = peek;
       length = strlen(str) + 1;
       peek->contents = (char *) gm2(length);
       if (peek->contents == NULL)
         {
          SendMessage
             ("\n**ERROR** Unable to allocate memory in AddSymbol\n",NO);
          EndCRSV();
         }
       peek->next = NULL;
       peek->bucket = tally;
       (void)strcpy(peek->contents,str);
       return(peek);
      }

    /*==================================================*/
    /* Search for the string in the list of entries for */ 
    /* this hash location.  If the string is found,     */
    /* then return the address of the string.           */
    /*==================================================*/

    while (peek != NULL)
      {
       if (strcmp(str,peek->contents) == 0)
	     { return(peek); }
       past = peek;
       peek = peek->next;
      }

    /*==================================================*/
    /* Add the string at the end of the list of entries */
    /* for this hash location.  Return the address of   */
    /* the string.                                      */
    /*==================================================*/

    past->next = GetStruct(hash_entry);
    peek = past->next;
    length = strlen(str) + 1;
    peek->contents = (char *) gm2(length);
    if (peek->contents == NULL)
      {
       SendMessage("\n**ERROR** Unable to allocate memory in AddSymbol\n",NO);
       EndCRSV();
      }
    peek->next = NULL;
    peek->bucket = tally;
    (void)strcpy(peek->contents,str);
    
    return(peek);
   }

/***********************************************************************/
/* InitSymbolTable: Purpose is to initialize the hash table to NULL. */
/***********************************************************************/

void InitSymbolTable()
{
    int i;

    IF(SymbolTable NEQ NULL) THEN
       ClearSymbolTable();

    ELSE
       SymbolTable = (HASH **) gm2(sizeof (HASH *) * HASHSIZE);
       IF(SymbolTable EQ NULL) THEN
          SendMessage
             ("\n**ERROR** Unable to allocate memory in hash table\n",NO);
          EndCRSV();
       END_IF
    END_IF

    for (i=0; i< HASHSIZE; i++)
        SymbolTable[i] = NULL;
}

/*********************************************************************/
/* CLEAR_SYMBOL_TABLE: Purpose is free memory tied up in hash table. */
/*********************************************************************/
static void ClearSymbolTable()
{
   int   i;
   HASH *temp1, *temp2;

   for (i = 0; i < HASHSIZE; i++) DO
      temp1 = SymbolTable[i];
      
      WHILE(temp1 NEQ NULL) DO
         temp2 = temp1->next;
         (void)rm(temp1->contents,(unsigned)(strlen(temp1->contents) + 1));
         rtn_struct(hash_entry,temp1);
         temp1 = temp2;
      END_WHILE
   END_FOR

   return;
}

/************************************************************************/
/* HASH_SYMBOL:  Computes the hash table location for the given string. */
/************************************************************************/
static int HashSymbol(word,range)
  char *word;
  int range;
  {
   int tally = 0, posn = 0;

   while (*word != EOS)
     {
      tally = (tally + (*word) * (posn % 16)) % 8192;
      word++; 
      posn++;           
     }

   return(tally % range);
  }





