#include <stdio.h>
#include "crsv.h"
#if MAC_LSC
#include <stdlib.h>
#endif

/******************************************************
 *  This file contains the functions used to allocate
 *  new memory blocks for each of the data structures
 *  used in CRSV
 ******************************************************
 */


/* ===========  Functions defined Externally but used here  =========== */
#if !MAC_LSC
 extern char *malloc();
#endif

/*=============================
 * in CRSVPRNT.C
 *=============================
 */

extern void SendMessage();
extern void ErrorMessage();

/*=============================
 * in CRSV.C
 *=============================
 */

extern void EndCRSV();


/* ===========  Functions defined here for Global use  ================ */

FB_PTR     AllocFactBlock();
FBR_PTR    AllocFactBlockRelation();
REL_PTR    AllocRelation();
FLD_PTR    AllocField();
EXF_PTR    AllocExternalFunction();
WL_PTR     AllocWordList();
RULE_PTR   AllocRule();
RR_PTR     AllocRuleRelation();
VAR_PTR    AllocVariable();
DR_PTR     AllocDefrelation();
DF_PTR     AllocDeffield();
DT_PTR     AllocDeftemplate();
TN_PTR     AllocTokenNode();
DS_PTR     AllocDefslot();
TK_PTR     AllocToken();
COUNTER   *AllocCounter();
ACT_R_PTR      AllocActivatedRule();
ACT_REL_PTR    AllocActivatedRelation();
FIELD_PTR      AllocFieldValues();
TMP_TKN_PTR    AllocTempToken();
SPE_FACT_PTR   AllocFactList();
RULE_NAME_PTR  AllocRuleList();
MULT_REL_PTR   AllocMultipleRelation();
IN_OR_VAR_PTR  AllocVariableInOrPattern();
PAT_LIST_PTR   AllocPatternList();
PAT_TKN_PTR    AllocTokenInPattern();
DGPTR          AllocDefglobalPtr();
MY_TMPLATE_PTR AllocMyTemplate();
AGEN_RULE_PTR  AllocAgendaRule();
FL_PTR         AllocFloatList();

void   FreeFactBlock();
void   FreeFactBlockRelation();
void   FreeRelation();
void   FreeField();
void   FreeExternalFunction();
void   FreeWordList();
void   FreeRule();
void   FreeRuleRelation();
void   FreeVariable();
void   FreeDefrelation();
void   FreeDefexternal();
void   FreeDeffield();
void   FreeDefslots();
void   FreeDeftemplate();
void   FreeTokenNode();
void   FreeFieldValue();
void   FreeToken();
WL_PTR FreeWordAndGetNext();
IL_PTR FreeIntegerAndGetNext();
FL_PTR FreeFloatAndGetNext();
void   FreeActivatedRule();
void   FreeRuleList();
void   FreeActivatedRelation();
void   FreeFactList();
void   FreeTempToken();
void   FreeDefglobalPtr();
void   FreeMyTemplate();
void   FreeAgendaRule();
void   FreeVariableInOrPattern();


void       MemoryError();
char      *GenAlloc();
void       GenFree();
char      *gm2();
int        rm();
long int   MemoryUsed();
long int   MaximumMemoryUsed();
long int   MemoryRequests();

int        ReleaseMemory();


/* ===========  Functions defined here for internal use  ============== */

#if MAC_LSC
static int   ReturnBlock();
static int   AllocateBlock();
static char *RequestBlock();
static int   InitNewBuffer();
static int   InitMem();
#endif


/* ===========  Variables defined here for Global use  ================ */

/* ===========  Variables defined Externally but used here  =========== */

/* ===========  Variables defined here for internal use  ============== */

#if MAC_LSC
#define INITBUFFERSIZE   8192
#define BUFFERSIZE       8192
#endif

#define MEM_TABLE_SIZE    300

struct MemPtr
  {
   struct MemPtr *next;
  };

static long int         MemoryAmount = 0;
static long int         MemoryCalls  = 0;
static long int         MaxMemory   = 0;

struct MemPtr         *MemoryTable[MEM_TABLE_SIZE];

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   GenAlloc()
 *
 *  PURPOSE:   This function is a general
 *             purpose memory allocation
 *             function. All CRSV memory
 *             allocations are through
 *             this function.
 *
 *  INPUTS:    The number of bytes needed.
 *
 *  OUTPUT:    A pointer to the new memory
 *             chunk.
 * ----------------------------------------
 */

char *GenAlloc(size, called_from)
unsigned int size;
char *called_from;
{
   char *MemPtr;

#if   MAC_LSC
   extern char *RequestBlock();

   MemPtr = RequestBlock(size);
   if (MemPtr == NULL)
     {
      ReleaseMemory(((size * 5 > 1024) ? size * 5 : 1024));
      MemPtr = RequestBlock(size);
      if (MemPtr == NULL)
        {
         ReleaseMemory(-1);
         MemPtr = RequestBlock(size);
         if (MemPtr == NULL)
           {
            SendMessage("\nERROR: out of memory",NO);
            EndCRSV();
           }
        }
     }
#else
   MemPtr = malloc(size);
   if (MemPtr == NULL)
     {
      ReleaseMemory((int)(size * 5 > 1024 ? size * 5 : 1024));
      MemPtr = malloc(size);
      if (MemPtr == NULL)
        {
         ReleaseMemory(-1);
         MemPtr = malloc(size);
         if (MemPtr == NULL)
           {
            return(NULL);
           }
        }
      }
#endif

   IF(CHECK_DEBUG IS_ON) THEN
      sprintf(MsgBuf,"\nMEM %ld ALLOC by: %s (%d)", MemPtr, called_from,
         size);
      SendMessage(MsgBuf,NO);
   END_IF

   MemoryAmount += size;
   MemoryCalls++;
   IF(MemoryAmount > MaxMemory) THEN
      MaxMemory= MemoryAmount;
   END_IF

   return(MemPtr);
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   disp_alloc()
 *
 *  PURPOSE:   This function is a slight mod
 *             to the general
 *             purpose memory allocation
 *             function. It is only for
 *             use by Mac display functions
 *             which need memory.
 *
 *  INPUTS:    The number of bytes needed.
 *
 *  OUTPUT:    A pointer to the new memory
 *             chunk.
 * ----------------------------------------
 */

/*^char *disp_alloc(size)
unsigned int size;
{
   char *MemPtr;

#if   MAC_LSC
   extern char *RequestBlock();

   MemPtr = RequestBlock(size);
   if (MemPtr == NULL)
     {
      ReleaseMemory(((size * 5 > 1024) ? size * 5 : 1024));
      MemPtr = RequestBlock(size);
      if (MemPtr == NULL)
        {
         ReleaseMemory(-1);
         MemPtr = RequestBlock(size);
         if (MemPtr == NULL)
           {
            EndCRSV();
           }
        }
     }
#else
   MemPtr = malloc(size);
   if (MemPtr == NULL)
     {
      ReleaseMemory((int)(size * 5 > 1024 ? size * 5 : 1024));
      MemPtr = malloc(size);
      if (MemPtr == NULL)
        {
         ReleaseMemory(-1);
         MemPtr = malloc(size);
         if (MemPtr == NULL)
           {
            return(NULL);
           }
        }
      }
#endif

   MemoryAmount += size;
   MemoryCalls++;
   IF(MemoryAmount > max_mem) THEN
      MaxMemory= MemoryAmount;
   END_IF

   return(MemPtr);
}*/

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   GenFree()
 *
 *  PURPOSE:   This function is a general
 *             purpose memory release
 *             function. All CRSV memory
 *             releases are through
 *             this function.
 *
 *  INPUTS:    A pointer to the memory to
 *             be released.
 *
 *  OUTPUT:    Nothing.
 * ----------------------------------------
 */

void GenFree(waste, size)
char *waste;
unsigned int size;
{
#if    MAC_LSC
   int  rtn = ReturnBlock(waste,size);

   if (rtn < 0)
     {
      sprintf(MsgBuf, "\nRelease error in genfree (%d)", rtn);
      SendMessage(MsgBuf,NO);
      return;
     }
#else
   free(waste);
#endif

   IF(CHECK_DEBUG IS_ON) THEN
      sprintf(MsgBuf,"\nMEM %ld FREE (%d)",waste, size);
      SendMessage(MsgBuf,NO);
   END_IF
   MemoryAmount -= size;
   MemoryCalls--;

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   disp_free()
 *
 *  PURPOSE:   This function is a slight mod
 *             to the general
 *             purpose memory release
 *             function. It is only for
 *             use by Mac display functions
 *             which need memory.
 *
 *  INPUTS:    A pointer to the memory to
 *             be released.
 *
 *  OUTPUT:    Nothing.
 * ----------------------------------------
 */

/*^void disp_free(waste,size)
char *waste;
int size;
{

#if    MAC_LSC
   int rtn;

   rtn = ReturnBlock(waste,size);
   if (rtn < 0)
     {
      return;
     }
#else
   (void) free (waste);
#endif

   MemoryAmount -= size;
   MemoryCalls--;

   return;
}*/

/*******************************************/
/* MemoryUsed:  Returns the amount of memory */
/*   currently allocated by CLIPS.         */
/*******************************************/

long int MemoryUsed()
  {
   return(MemoryAmount);
  }

/***************************************************/
/* MaximumMemoryUsed:  Returns the maximum amount  */
/*   of memory ever allocated by CRSV.             */
/***************************************************/

long int MaximumMemoryUsed()
  {
   return(MaxMemory);
  }

/***********************************************************/
/* MemoryRequests:  Returns the number of outstanding memory */
/*   memory calls made through memory functions.           */
/***********************************************************/

long int MemoryRequests()
  {
   return( MemoryCalls);
  }


/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  ALLOC_FB
 *
 *  PURPOSE:  This function allocates
 *            memory for a new fact
 *            block structure and fills
 *            in the appropriate fields.
 *
 *  INPUTS:   None.
 *
 *  OUTPUT:   A pointer to a fact block
 *            structure.
 * ----------------------------------------
 */

FB_PTR AllocFactBlock()
{
   FB_PTR temp_fb;

   IF((temp_fb = (FB_PTR) GenAlloc(sizeof(FACT_BLOCK),"AllocFactBlock")) == NULL)THEN
      MemoryError();                             /* Allocate memory */
   END_IF

   temp_fb->name       = NULL;                    /* Clear name pointer */
   temp_fb->file       = NULL;                    /* Clear file pointer */
   temp_fb->next_block = NULL;                    /* Clear FB pointer   */
   temp_fb->num_facts  = 0;                       /* Clear num of facts */

   return(temp_fb);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeFactBlock
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            fact blocks.
 *
 *  INPUTS:   A pointer to the head of
 *            the fact block list (FACT_BLOCK *).
 *
 *  OUTPUT:   Nothing
 *
 *  NOTE  :   HACK warning! I do not free the
 *            file name stored with each
 *            fact block since they are
 *            currently muliply referenced.
 *            This means I am creating
 *            garbage if the program runs
 *            multiple times!
 * ----------------------------------------
 */

void FreeFactBlock(list)
FB_PTR list;
{
   FB_PTR temp;

   while(list NEQ NULL)
    {
      temp = list;
      list = list->next_block;
      GenFree((char *)temp, (unsigned)sizeof(FACT_BLOCK));
    }
   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_FBR
 *
 *  PURPOSE:   This function allocates
 *             memory for a new fact
 *             block/relation structure
 *             and fills in the appropriate
 *             fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a fact block/
 *             relation structure.
 * ----------------------------------------
 */

FBR_PTR AllocFactBlockRelation()
{
   FBR_PTR temp;

   if ((temp = (FBR_PTR) GenAlloc(sizeof(struct fbr_list),"AllocFactBlockRelation")) ==
      NULL)
      MemoryError();

   temp->fact_block = NULL;
   temp->next_fbr   = NULL;
   temp->num_fields = 0;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeFactBlockRelation
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            fact blocks/relations.
 *
 *  INPUTS:   A pointer to the head of the
 *            fact block/rel list (FBR_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeFactBlockRelation(list)
FBR_PTR list;
{
  FBR_PTR temp;

  while(list NEQ NULL)
    {
       temp = list;
       list = list->next_fbr;
       GenFree((char *)temp, (unsigned)sizeof(struct fbr_list));
    }
   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :  ALLOC_RR
 *
 *  PURPOSE:   This function allocates
 *             memory for a new rule/
 *             relation structure and
 *             fills in the appropriate
 *             fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a rule/
 *             relation structure.
 * ----------------------------------------
 */

RR_PTR AllocRuleRelation()
{
   RR_PTR temp;

   if((temp = (RR_PTR) GenAlloc(sizeof(struct rr_list),"AllocRuleRelation")) == NULL)
      MemoryError();

   temp->rule       = NULL;
   temp->next_rr    = NULL;
   temp->num_fields = 0;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeRuleRelation
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            rule/relation structures.
 *
 *  INPUTS:   A pointer to the head of the
 *            rule/rel list (RR_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeRuleRelation(list)
RR_PTR list;
{
   RR_PTR temp;

   while(list != NULL)
    {
      temp = list;
      list = list->next_rr;
      GenFree((char *)temp, (unsigned)sizeof(struct rr_list));
    }

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_REL
 *
 *  PURPOSE:   This function allocates
 *             memory for a new relation
 *             structure and fills in
 *             the appropriate fields.
 *
 *   INPUTS:   None.
 *
 *   OUTPUT:   A pointer to a fact relation
 *             structure.
 * ----------------------------------------
 */

REL_PTR AllocRelation()
{
   REL_PTR temp;

   if((temp = (REL_PTR) GenAlloc(sizeof(RELATION),"AllocRelation")) == NULL)
      MemoryError();                            /* Allocate memory */

   temp->name         = NULL;                    /* Clear all pointers */
   temp->lft_rel      = NULL;
   temp->rht_rel      = NULL;
   temp->rule_list    = NULL;
   temp->fb_list      = NULL;
   temp->assert_list  = NULL;
   temp->retract_list = NULL;
   temp->field_list   = NULL;
   temp->is_template  = NO;
   
   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeRelation
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            relations.
 *
 *  INPUTS:   A pointer to the head of the
 *            relation list (RELATION *).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeRelation(list)
REL_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_rel NEQ NULL) THEN
         FreeRelation(list->lft_rel);
      END_IF

      IF(list->rht_rel NEQ NULL) THEN
         FreeRelation(list->rht_rel);
      END_IF

      FreeRuleRelation(list->rule_list);
      FreeFactBlockRelation(list->fb_list);
      FreeRuleRelation(list->retract_list);
      FreeRuleRelation(list->assert_list);
      FreeField(list->field_list);
      GenFree((char *)list, (unsigned)sizeof(RELATION));
   END_IF

   return;
}

/* =========================================================
 *  AllocTokenInPattern()
 *   Allocate memory for a field in each pattern
 * ========================================================
 */

PAT_TKN_PTR AllocTokenInPattern()
{
   PAT_TKN_PTR temp;

 if((temp = (PAT_TKN_PTR) GenAlloc(sizeof(PAT_TKN), "AllocTokenInPattern")) == NULL)
       MemoryError();

   temp->next = NULL;
   temp->type = 0;
   temp->num = 0;
   temp->word = NULL;
   return(temp);
}

/*
 *  ======================================================
 *  FreeTokenInPattern()
 *    free memory for the field in a pattern
 *  ======================================================
 */
void FreeTokenInPattern(list)
PAT_TKN_PTR list;
{
   PAT_TKN_PTR temp;

   while(list != NULL)
     {
        temp = list;
        list = list->next;
        temp->word = NULL;
        GenFree((char *)temp, (unsigned)sizeof(PAT_TKN));
     }

   
}

/* =========================================================
 * AllocPatternList()
 *    allocate memeory for each pattern in the rule.
 *
 * =========================================================
 */

PAT_LIST_PTR AllocPatternList()
{
    PAT_LIST_PTR temp;
    
    if((temp = (PAT_LIST_PTR) GenAlloc(sizeof(PAT_LIST), "AllocPatternList")) == NULL)
       MemoryError();
    
    temp->num     = 0;
    temp->in_or   = NO;
    temp->in_not  = NO;
    temp->in_and  = NO;
    temp->has_wildcard = NO;
    temp->pattern = NULL;
    temp->next_pattern = NULL;
    return(temp);
}

/* =============================================================
 *  Free the memory for the rule's pattern
 * =============================================================
 */
void FreePatternList(list)
PAT_LIST_PTR list;
 {
   PAT_LIST_PTR temp;
   
   while(list != NULL )
    {
       temp = list;
       FreeTokenInPattern(temp->pattern);
       list = list->next_pattern;
       GenFree((char *)temp, (unsigned)sizeof(PAT_LIST));
    }
 }

/* =============================================================
 *  Allocate memory for a rule in the agenda.
 * =============================================================
 */
 
 AGEN_RULE_PTR AllocAgendaRule()
 {
    AGEN_RULE_PTR temp;
    
    if((temp = (AGEN_RULE_PTR) GenAlloc(sizeof(AGEN_RULE), "AllocAgendaRule")) == NULL)
       MemoryError();
    temp->rule = NULL;
    temp->next_rule = NULL;
    return(temp);
 }
 
/* =============================================================
 *  FreeAgendaRule
 * =============================================================
 */
 void FreeAgendaRule(list)
 AGEN_RULE_PTR list;
 {
    AGEN_RULE_PTR temp;
    
    while(list != NULL)
      {
        temp = list;
        temp->rule = NULL;
        list = list->next_rule;
        GenFree((char *)temp,(unsigned)sizeof(AGEN_RULE));
      }
 }
 
/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   AllocRule
 *
 *  PURPOSE:    This function allocates
 *              memory for a new rule
 *              structure and fills in
 *              the appropriate fields.
 *
 *   INPUTS:   None.
 *
 *   OUTPUT:   A pointer to a rule structure.
 * ----------------------------------------
 */

RULE_PTR AllocRule()
{
   RULE_PTR temp;

   if ((temp = (RULE_PTR) GenAlloc(sizeof(RULE),"AllocRule")) == NULL)
      MemoryError();                            /* Allocate memory */

   temp->name          = NULL;                   /* Clear all pointers */
   temp->file          = NULL;
   temp->comment       = NULL;
   temp->lft_rule      = NULL;
   temp->rht_rule      = NULL;
   temp->lhs_patterns  = NULL;
   temp->salience_set  = NO;
   temp->salience_val  = 0;
   temp->num_patterns  = 0;
   temp->num_actions   = 0;
   temp->num_retracts  = 0;
   temp->num_asserts   = 0;
   temp->num_printouts = 0;
   temp->num_ex_funcs  = 0;
   temp->num_nots      = 0;
   temp->num_ifs       = 0;
   temp->num_whiles    = 0;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeRule
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            rules.
 *
 *  INPUTS:   A pointer to the head of
 *            the rule list (RULE *).
 *
 *  OUTPUT:   Nothing
 *
 * ----------------------------------------
 */

void FreeRule(list)
RULE_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_rule NEQ NULL) THEN
         FreeRule(list->lft_rule);
      END_IF

      IF(list->rht_rule NEQ NULL) THEN
         FreeRule(list->rht_rule);
      END_IF
      FreePatternList(list->lhs_patterns);
      GenFree((char *)list, (unsigned)sizeof(RULE));
   END_IF

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_FLD
 *
 *  PURPOSE:   This function allocates
 *             memory for a new field
 *             structure and fills in
 *             the appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a field
 *             structure.
 * ----------------------------------------
 */

FLD_PTR AllocField()
{
   FLD_PTR temp;

   if ((temp = (FLD_PTR) GenAlloc(sizeof(struct field),"AllocField")) == NULL)
      MemoryError();

   temp->field_num  = 0;
   temp->next_field = NULL;
   temp->lit_values = NULL;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeField
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            fields.
 *
 *  INPUTS:   A pointer to the head of
 *            the field list (FIELD *).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeField(list)
FLD_PTR list;
{
   FLD_PTR temp;

   while(list NEQ NULL)
    {
      temp = list;
      list  = list->next_field;
      if(temp->lit_values != NULL)
        FreeWordList(temp->lit_values);
      GenFree((char *)temp, (unsigned)sizeof(FIELD));
    }

   return;
}

/* -------------------------------------------------------
 *  Name : ALLOC_EX_FUNC()
 *         This function allocate the memory for an external
 *         function during processing external function
 *  Return:
 *         Pointer poiniting to the allocated structure (EXF_PTR)
 * --------------------------------------------------------
 */

EXF_PTR AllocExternalFunction()
{
   EXF_PTR temp;

   if ((temp = (EXF_PTR) GenAlloc(sizeof(struct ex_func),"AllocExternalFunction")) == NULL)
      MemoryError();

   temp->LHS_rule_list = NULL;
   temp->RHS_rule_list = NULL;
   temp->global_list   = NULL;
   temp->lft_ex_func   = NULL;
   temp->rht_ex_func   = NULL;
   temp->argument_list = NULL;

   return(temp);
}

/* -------------------------------------------------------------------------
 *  Name: FreeExternalFunction();
 *          This function frees the memory that ocuppied by the
 *          external function (EXF_PTR)
 *  Return: Nothing
 * --------------------------------------------------------------------------
 */

void FreeExternalFunction(list)
EXF_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_ex_func NEQ NULL) THEN
         FreeExternalFunction(list->lft_ex_func);
      END_IF

      IF(list->rht_ex_func NEQ NULL) THEN
         FreeExternalFunction(list->rht_ex_func);
      END_IF

      FreeRuleRelation(list->LHS_rule_list);
      FreeRuleRelation(list->RHS_rule_list);
      FreeDefglobalPtr(list->global_list);
      FreeField(list->argument_list);

      GenFree((char *)list, (unsigned)sizeof(EX_FUNC));
   END_IF

   return;
}


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   AllocWordList
 *
 *  PURPOSE:   This function allocates
 *             memory for a new word list
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a word list
 *             structure.
 * ----------------------------------------
 */

WL_PTR AllocWordList()
{
   WL_PTR temp;

   if ((temp = (WL_PTR) GenAlloc(sizeof(struct word_list),"AllocWordList")) ==
      NULL)
      MemoryError();

   temp->next_word  = NULL;
   temp->word       = NULL;
   temp->times_used = 0;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeWordList
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            words.
 *
 *  INPUTS:   A pointer to the head of
 *            the word list (WL_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeWordList(list)
WL_PTR list;
{
   WL_PTR temp;

   while(list != NULL)
    {
      temp = list;
      list = list->next_word;
      GenFree((char *)temp, (unsigned)sizeof(WORD_LIST));
    }
   return;
}

/* ======================================================================== */

/* ------------------------------------------------------------------
 *  NAME   : FreeWordAndGetNext()
 *
 *  PURPOSE: This function frees the current word and returns a
 *           pointer to the next word in the list.
 *
 *  INPUTS:  Three arguments,
 *           The address of the head-of-the-list pointer (WL_PTR *),
 *           a pointer to the word to be freed (WL_PTR),
 *           and a pointer to the previous word (WL_PTR).
 *
 *  RETURNS: A pointer to the next word in the list (WL_PTR).
 * ------------------------------------------------------------------
 */

WL_PTR FreeWordAndGetNext(head_word,cur_word,last_word)
   WL_PTR *head_word,cur_word,last_word;
{
   WL_PTR old_word;

   old_word = cur_word;

   IF (last_word EQ NULL) THEN
      *head_word = old_word->next_word;
   ELSE
      last_word->next_word = old_word->next_word;
   END_IF

   cur_word = old_word->next_word;

   GenFree((char *)old_word, (unsigned)sizeof(WORD_LIST));

   return(cur_word);
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_FL
 *
 *  PURPOSE:   This function allocates
 *             memory for a new number list
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a number list
 *             structure.
 * ----------------------------------------
 */

FL_PTR AllocFloatList()
{
   FL_PTR temp;

   temp = (FL_PTR) GenAlloc(sizeof(struct float_list),"AllocFloatList");

   IF(temp EQ NULL) THEN
      MemoryError();
   END_IF

   temp->next_number = NULL;
   temp->number      = 0.0;
   temp->times_used  = 0;

   return(temp);
}


/* ----------------------------------------
 *  NAME   :  FreeFloatList
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            floats.
 *
 *  INPUTS:   A pointer to the head of
 *            the number list (FL_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeFloatList(list)
FL_PTR list;
{
   FL_PTR temp;

   while ( list != NULL)
    {
      temp = list;
      list = list->next_number;
      GenFree((char *)temp, (unsigned)sizeof(FLOAT_LIST));
    }

   return;
}

/* ----------------------------------------
 *  NAME   :   AllocIntegerList
 *
 *  PURPOSE:   This function allocates
 *             memory for a new integer list
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a number list
 *             structure.
 * ----------------------------------------
 */

IL_PTR AllocIntegerList()
{
   IL_PTR temp;

   temp = (IL_PTR) GenAlloc(sizeof(struct integer_list),"AllocIntegerList");

   IF(temp EQ NULL) THEN
      MemoryError();
   END_IF

   temp->next_number = NULL;
   temp->number      = 0;
   temp->times_used  = 0;

   return(temp);
}
/* ==================================================================== */


/* ----------------------------------------
 *  NAME   :  FreeIntegerList
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            integers.
 *
 *  INPUTS:   A pointer to the head of
 *            the number list (IL_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeIntegerList(list)
IL_PTR list;
{
   IL_PTR temp;

   while ( list != NULL)
    {
      temp = list;
      list = list->next_number;
      GenFree((char *)temp, (unsigned)sizeof(INT_LIST));
    }

   return;
}

/* ======================================================================== */


/* ------------------------------------------------------------------
 *  NAME   : FreeIntegerAndGetNext()
 *
 *  PURPOSE: This function frees the current number and returns a
 *           pointer to the next number in the list.
 *
 *  INPUTS:  Three arguments,
 *           The address of the head-of-the-list pointer (IL_PTR *),
 *           a pointer to the number to be freed (IL_PTR),
 *           and a pointer to the previous number (IL_PTR).
 *
 *  RETURNS: A pointer to the next number in the list (IL_PTR).
 * ------------------------------------------------------------------
 */

IL_PTR FreeIntegerAndGetNext(head_num,cur_num,last_num)
   IL_PTR *head_num,cur_num,last_num;
{
   IL_PTR old_num;

   old_num = cur_num;

   IF (last_num EQ NULL) THEN
      *head_num = old_num->next_number;
   ELSE
      last_num->next_number = old_num->next_number;
   END_IF

   cur_num = old_num->next_number;
   GenFree((char *)old_num, (unsigned)sizeof(INT_LIST));

   return(cur_num);
}
/* ------------------------------------------------------------------
 *  NAME   : FreeFloatAndGetNext()
 *
 *  PURPOSE: This function frees the current number and returns a
 *           pointer to the next number in the list.
 *
 *  INPUTS:  Three arguments,
 *           The address of the head-of-the-list pointer (FL_PTR *),
 *           a pointer to the number to be freed (FL_PTR),
 *           and a pointer to the previous number (FL_PTR).
 *
 *  RETURNS: A pointer to the next number in the list (FL_PTR).
 * ------------------------------------------------------------------
 */

FL_PTR FreeFloatAndGetNext(head_num,cur_num,last_num)
   FL_PTR *head_num,cur_num,last_num;
{
   FL_PTR old_num;

   old_num = cur_num;

   IF (last_num EQ NULL) THEN
      *head_num = old_num->next_number;
   ELSE
      last_num->next_number = old_num->next_number;
   END_IF

   cur_num = old_num->next_number;
   GenFree((char *)old_num, (unsigned)sizeof(FLOAT_LIST));

   return(cur_num);
}
/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_VAR
 *
 *  PURPOSE:   This function allocates
 *             memory for a variable list
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a variable
 *             list structure.
 * ----------------------------------------
 */

VAR_PTR AllocVariable()
{
   VAR_PTR temp;

   if ((temp = (VAR_PTR) GenAlloc(sizeof(struct variable_list),
      "AllocVariable")) == NULL)
      MemoryError();

   temp->next_variable   = NULL;
   temp->prev_variable   = NULL;
   temp->relation        = NULL;
   temp->ex_func         = NULL;
   temp->var             = NULL;
   temp->field_info      = NULL;
   temp->level_retracted = 0;
   temp->block_type      = 0;
   temp->count           = 0;
   temp->type            = 0;
   temp->bound_in_not    = NO;
   temp->bound_in_or     = FALSE;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeVariable
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            variables.
 *
 *  INPUTS:   A pointer to the head of the
 *            variable list (VAR_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeVariable(list)
VAR_PTR list;
{
   VAR_PTR temp;
     while(list != NULL){
       temp = list;
       list = list->next_variable;
       temp->prev_variable = NULL;
       temp->relation   = NULL;
       temp->field_info = NULL;
       temp->ex_func    = NULL;
       GenFree((char *)temp, (unsigned)sizeof(VARIABLE_LIST));}
     return;
}

/* -------------------------------------------------
 *
 * 
 * -------------------------------------------------
 */

IN_OR_VAR_PTR AllocVariableInOrPattern()
{

   IN_OR_VAR_PTR  temp;
	
   if ((temp = (IN_OR_VAR_PTR) GenAlloc(sizeof(struct in_or_var),
      "AllocVariableInOrPattern")) == NULL)
      MemoryError();
    temp->num = 0;
    temp->check = OFF;
    temp->var_info = NULL;
    temp->word     = NULL;
    temp->next     = NULL;
    return(temp);
}


/* --------------------------------------------------------------
 *  FreeVariableInOrPattern
 *
 * --------------------------------------------------------------
 */
 
 void FreeVariableInOrPattern(list)
 IN_OR_VAR_PTR list;
 {
    IN_OR_VAR_PTR temp;
    
 	while(list != NULL)
 	{
 	  temp = list;
 	  list = list->next;
 	  GenFree((char *)temp, (unsigned)sizeof(struct in_or_var));
 	}
   return;	
 }
/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_DEF_REL
 *
 *  PURPOSE:   This function allocates
 *             memory for a defrelation
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a defrelation
 *             structure.
 * ----------------------------------------
 */

DR_PTR AllocDefrelation()
{
   DR_PTR temp;

   if ((temp = (DR_PTR) GenAlloc(sizeof(struct def_relation),
      "AllocDefrelation")) == NULL)
      MemoryError();

   temp->name        = NULL;
   temp->file        = NULL;
   temp->lft_def_rel = NULL;
   temp->rht_def_rel = NULL;
   temp->max_fields  = 10000;          /* Defaults to essentially any */
   temp->min_fields  = 0;              /* number of fields */
   temp->fields      = NULL;
   temp->created_by  = CRSV_MADE;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeDefrelation
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            def relations
 *
 *  INPUTS:   A pointer to the head of the
 *            def relations list (DR_PTR).
 *
 *  OUTPUT:   Nothing
 *
 *  NOTE  :   HACK warning! I do not free the
 *            file name stored with each
 *            fact block since they are
 *            currently muliply referenced.
 *            This means I am creating
 *            garbage if the program runs
 *            multiple times!
 * ----------------------------------------
 */

void FreeDefrelation(list)
DR_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_def_rel NEQ NULL) THEN
         FreeDefrelation(list->lft_def_rel);
      END_IF

      IF(list->rht_def_rel NEQ NULL) THEN
         FreeDefrelation(list->rht_def_rel);
      END_IF

      FreeDeffield(list->fields);
      GenFree((char *)list, (unsigned)sizeof(DEF_REL));
   END_IF

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_DEF_TMP
 *
 *  PURPOSE:   This function allocates
 *             memory for a deftemplate
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a deftemplate
 *             structure.
 * ----------------------------------------
 */

DT_PTR AllocDeftemplate()
{
   DT_PTR temp;

   if ((temp = (DT_PTR) GenAlloc(sizeof(struct def_template),
      "AllocDeftemplate")) == NULL)
      MemoryError();

   temp->name        = NULL;
   temp->file        = NULL;
   temp->lft_def_tmp = NULL;
   temp->rht_def_tmp = NULL;
   temp->slots       = NULL;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeDeftemplate
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            def templates
 *
 *  INPUTS:   A pointer to the head of the
 *            def templates list (DT_PTR).
 *
 *  OUTPUT:   Nothing
 *
 *  NOTE  :   HACK warning! I do not free the
 *            file name stored with each
 *            fact block since they are
 *            currently muliply referenced.
 *            This means I am creating
 *            garbage if the program runs
 *            multiple times!
 * ----------------------------------------
 */

void FreeDeftemplate(list)
DT_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_def_tmp NEQ NULL) THEN
         FreeDeftemplate(list->lft_def_tmp);
      END_IF

      IF(list->rht_def_tmp NEQ NULL) THEN
         FreeDeftemplate(list->rht_def_tmp);
      END_IF

      FreeDefslots(list->slots);
      GenFree((char *)list, (unsigned)sizeof(DEF_TMP));
   END_IF

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeDefslots
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            def templates slots
 *
 *  INPUTS:   A pointer to the head of the
 *            def templates slots list (DS_PTR).
 *
 *  OUTPUT:   Nothing
 *
 * ----------------------------------------
 */

void FreeDefslots(slot)
DS_PTR slot;
{
   WHILE (slot) DO
      DS_PTR previous_slot = slot;
      TN_PTR defaults = previous_slot->defaults;

      slot = slot->next_slot;
      GenFree ((char *)previous_slot, (unsigned)sizeof (DEF_SLOT));

      WHILE (defaults) DO
         TN_PTR previous_default = defaults;

         defaults = defaults->next;
         FreeToken (previous_default->token);
         FreeTokenNode (previous_default);
      DONE;
   DONE;

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   ALLOC_DEF_SLOT
 *
 *  PURPOSE:   This function allocates
 *             memory for a new def slot
 *             structure and fills in
 *             the appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a def slot
 *             structure.
 * ----------------------------------------
 */

DS_PTR AllocDefslot()
{
   DS_PTR temp;

   if ((temp = (DS_PTR) GenAlloc(sizeof(struct def_slot),"AllocDefslot")) ==
      NULL)
      MemoryError();

   temp->name  = NULL;
   temp->position = 0;
   temp->next_slot = NULL;
   temp->max_elements = 9900;
   temp->min_elements = 0;
   temp->defaults = NULL;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :   AllocDeffield
 *
 *  PURPOSE:   This function allocates
 *             memory for a def_field
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a def_field
 *             structure.
 * ----------------------------------------
 */

DF_PTR AllocDeffield()
{
   DF_PTR temp;

   if ((temp = (DF_PTR) GenAlloc(sizeof(struct def_field),"AllocDeffield")) ==
      NULL)
      MemoryError();

   temp->position         = 0;
   temp->next_field       = NULL;
   temp->allow_word       = YES;
   temp->allow_string     = YES;
   temp->allow_instance   = YES;
   temp->allow_integer    = YES;
   temp->allow_float      = YES;
   temp->allow_address    = YES;
   temp->possible_words   = NULL;
   temp->possible_strings = NULL;
   temp->possible_instances = NULL;
   temp->possible_integers = NULL;
   temp->possible_floats  = NULL;
   temp->set_int_max      = NO;
   temp->set_int_min      = NO;
   temp->set_float_max    = NO;
   temp->set_float_min    = NO;

   return(temp);
}


/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeDeffield
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            field definitions.
 *
 *  INPUTS:   A pointer to the head of the
 *            field list (DF_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeDeffield(list)
DF_PTR list;
{
   DF_PTR temp;

   while(list != NULL)
    {
      temp = list;
      list = list->next_field;
      FreeWordList(temp->possible_words);
      FreeWordList(temp->possible_strings);
      FreeWordList(temp->possible_instances);
      FreeIntegerList(temp->possible_integers);
      FreeFloatList(temp->possible_floats);
      GenFree((char *)temp, (unsigned)sizeof(DEF_FLD));
    }

   return;
}

/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   AllocDefexternal
 *
 *  PURPOSE:   This function allocates
 *             memory for a defrelation
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a defrelation
 *             structure.
 * ----------------------------------------
 */

DE_PTR AllocDefexternal()
{
   DE_PTR temp;

   if ((temp = (DE_PTR)GenAlloc(sizeof(struct def_external),
      "AllocDefexternal")) == NULL)
      MemoryError();

   temp->name             = NULL;
   temp->true_name        = NULL;
   temp->assert_list      = NULL;
   temp->retract_list     = NULL;
   temp->file             = NULL;
   temp->lft_def_ext      = NULL;
   temp->rht_def_ext      = NULL;
   temp->return_word      = NO;
   temp->return_string    = NO;
   temp->return_integer   = NO;
   temp->return_float     = NO;
   temp->return_address   = NO;
   temp->return_boolean   = NO;
   temp->max_arguments    = 10000;          /* Defaults to essentially any */
   temp->min_arguments    = 0;              /* number of arguments */
   temp->arguments        = NULL;

   return(temp);
}

/* ==================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeDefexternal
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            def externals
 *
 *  INPUTS:   A pointer to the head of the
 *            def relations list (DE_PTR).
 *
 *  OUTPUT:   Nothing
 *
 *  NOTE  :   HACK warning! I do not free the
 *            file name stored with each
 *            function block since they are
 *            currently muliply referenced.
 *            This means I am creating
 *            garbage if the program runs
 *            multiple times!
 * ----------------------------------------
 */

void FreeDefexternal(list)
DE_PTR list;
{
   IF(list NEQ NULL) THEN
      IF(list->lft_def_ext NEQ NULL) THEN
         FreeDefexternal(list->lft_def_ext);
      END_IF

      IF(list->rht_def_ext NEQ NULL) THEN
         FreeDefexternal(list->rht_def_ext);
      END_IF

      IF(list->assert_list NEQ NULL) THEN
         FreeWordList(list->assert_list);
      END_IF

      IF(list->retract_list NEQ NULL) THEN
         FreeWordList(list->retract_list);
      END_IF

      FreeDeffield(list->arguments);
      GenFree((char *)list, (unsigned)sizeof(DEF_EXT));
   END_IF

   return;
}


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   AllocCounter
 *
 *  PURPOSE:   This function allocates
 *             memory for a counter
 *             structure and fills in the
 *             appropriate fields.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    A pointer to a counter
 *             structure.
 * ----------------------------------------
 */

COUNTER *AllocCounter()
{
   COUNTER *temp;

   if ((temp = (struct counter *) GenAlloc(sizeof(struct counter),
      "AllocCounter")) == NULL)
      MemoryError();

   temp->prime_count   = 0;
   temp->count_A       = 0;
   temp->count_B       = 0;
   temp->count_C       = 0;
   temp->count_D       = 0;

   return(temp);
}

/**********************************************************************/

/* ----------------------------------------
 *  NAME   :  AllocTokenNode
 *
 *  PURPOSE:  This function allocates the
 *            memory required by a token
 *            node
 *
 *  INPUTS:   None
 *
 *  OUTPUT:   A pointer to an empty
 *            structure (TN_PTR).
 *
 *  NOTE  :   None
 * ----------------------------------------
 */

TN_PTR AllocTokenNode ()
{
   TN_PTR temp;

   if ((temp = (struct token_node *) GenAlloc(sizeof(struct token_node),
      "AllocTokenNode")) == NULL)
      MemoryError();

   temp->token   = NULL;
   temp->next    = NULL;

   return(temp);
}


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeTokenNode
 *
 *  PURPOSE:  This function frees the
 *            memory used by a token
 *            node
 *
 *  INPUTS:   A pointer to the node which is
 *            to be freed.
 *
 *  OUTPUT:   Nothing
 *
 *  NOTE  :   A token pointed at by the
 *            node is not freed.
 * ----------------------------------------
 */

void FreeTokenNode (node)
 TN_PTR node;
 {
   GenFree ((char *)node, (unsigned)sizeof(TOKEN_NODE));
   return;
 }


/* ======================================================================== */

MULT_REL_PTR AllocMultipleRelation()
{
   MULT_REL_PTR temp;

   if ((temp = (MULT_REL_PTR) GenAlloc(sizeof(MULT_REL),"AllocMultipleRelation")) ==
      NULL)
      MemoryError();

   temp->rel_name = NULL;
   temp->next_rel = NULL;

   return(temp);
}

 /********************************************************************/

 void FreeMultipleRelation(list)
 MULT_REL_PTR list;
 {
     MULT_REL_PTR temp,temp1;

     temp = list;
     while( temp != NULL)
      {
        temp1 = temp;
        temp = temp->next_rel;
        GenFree((char *)temp1, (unsigned)sizeof(MULT_REL));
      }

     return;
 }


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :  AllocToken
 *
 *  PURPOSE:  This function allocates the
 *            memory required by a token
 *
 *  INPUTS:   None
 *
 *  OUTPUT:   A pointer to an empty
 *            structure (TK_PTR).
 *
 *  NOTE  :   None
 * ----------------------------------------
 */

TK_PTR AllocToken ()
{
   TK_PTR temp;

   if ((temp = (struct token *) GenAlloc(sizeof(struct token),
      "AllocToken")) == NULL)
      MemoryError();

   temp->token   = 0;
   temp->intnumber = 0;
   temp->floatnumber = 0.0;
   temp->tknword = NULL;
   temp->hashword = NULL;
   temp->print_rep = NULL;
   temp->field = NULL;

   return(temp);
}


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :  FreeToken
 *
 *  PURPOSE:  This function frees the
 *            memory used by a token
 *
 *  INPUTS:   A pointer to the node which is
 *            to be freed.
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */

void FreeToken (node)
 TK_PTR node;
 {
   GenFree ((char *)node, (unsigned)sizeof(TOKEN));
   return;
 }



DG_PTR AllocDefglobal()
{
    DG_PTR temp;
    
   if ((temp = (struct def_global *) GenAlloc(sizeof(struct def_global),
      "AllocDefglobal")) == NULL)
      MemoryError();
   temp->variable = NULL;
   temp->value = NULL;
   temp->next = NULL;
   
   return(temp);
}

void FreeDefglobal(node)
DG_PTR node;
{
   DG_PTR temp;
   
   while(node != NULL)
    {
      temp = node;
      node = node->next;
      GenFree ((char *)temp, (unsigned)sizeof(struct def_global));
    }
   return;
}


struct DefglobalPtr *AllocDefglobalPtr()
{
    struct DefglobalPtr *temp;
    
   if ((temp = (struct DefglobalPtr *) GenAlloc(sizeof(struct DefglobalPtr),
      "AllocDefglobalPtr")) == NULL)
      MemoryError();
   temp->defglobal_list = NULL;
   temp->next = NULL;
   
   return(temp);
}

void FreeDefglobalPtr(node)
DGPTR node;
{
   DGPTR temp;
   
   while(node != NULL)
    {
      temp = node;
      node->defglobal_list = NULL;
      node = node->next;
      GenFree ((char *)temp, (unsigned)sizeof(DG));
    }
   return;
}
/* ======================================================================== */

/* ============================================== *
 *   THE BELOW FUCTIONS ARE SPECIALIZED FOR       *
 *   PROCESSING DRIBBLE FILE.                     *
 * ============================================== */

/* ----------------------------------------------- *
 * NAME    : ALLOC_A_RULE                          *
 *                                                 *
 * PURPOSE :                                       *
 *    This function allocates memory for ACT_RULE  *
 *    structure and initialize it.                 *
 *                                                 *
 * INPUT   : Nothing.                              *
 *                                                 *
 * RETURN  : pointer to the ACT_RULE structure     *
 * ----------------------------------------------- */

 ACT_R_PTR AllocActivatedRule()

 {
   ACT_R_PTR temp;

   if((temp  = (ACT_R_PTR) GenAlloc(sizeof(ACT_RULE),"AllocActivatedRule")) == NULL)
      MemoryError();

   temp->next = NULL;
   temp->fire_num = 0;
   temp->activate_num  = 0;
   temp->deactivate_num = 0;
   return(temp);
 }


/* ----------------------------------------
 *  NAME   :  FreeActivatedRule()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            the activated rules.
 *
 *  INPUTS:   A pointer to the head of the
 *            rule_list (ACT_R_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeActivatedRule(list)
 ACT_R_PTR list;
 {
    ACT_R_PTR temp;

    while(list != NULL)
     {
       temp = list;
       list = list->next;
       GenFree((char *)temp, (unsigned)sizeof(ACT_RULE));
     }

   return;
 }



/* ----------------------------------------------- *
 * NAME    : ALLOC_FIELD                           *
 *                                                 *
 * PURPOSE :                                       *
 *    This function allocates memory for FIELD_LIST*
 *    structure and initialize it.                 *
 *                                                 *
 * INPUT   : Nothing.                              *
 *                                                 *
 * RETURN  : pointer to the FIELD_LIST structure   *
 * ----------------------------------------------- */

 FIELD_PTR AllocFieldValues()

{
   FIELD_PTR temp;

   if((temp = (FIELD_PTR) GenAlloc(sizeof(FIELD_LIST),"AllocFieldValues")) ==
      NULL)
      MemoryError();
   temp->num = 0;
   temp->next = NULL;
   temp->type = UNKNOWN;
   return(temp);
}

/* ----------------------------------------
 *  NAME   :  FreeFieldValue()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            the fields.
 *
 *  INPUTS:   A pointer to the head of the
 *            rule_list (FIELD_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeFieldValue(list)
 FIELD_PTR list;
 {
   FIELD_PTR temp;

   while(list != NULL)
    {
      temp =list;
      list = list->next;
      GenFree((char *)temp, (unsigned)sizeof(FIELD));
    }

   return;
 }

/* ----------------------------------------------- *
 * NAME    : AllocTempToken()                       *
 *                                                 *
 * PURPOSE :                                       *
 *    This function allocates memory for TMP_TKN   *
 *    structure and initialize it.                 *
 *                                                 *
 * INPUT   : Nothing.                              *
 *                                                 *
 * RETURN  : pointer to the TMP_TKN structure   *
 * ----------------------------------------------- */

TMP_TKN_PTR AllocTempToken()

{
   TMP_TKN_PTR temp;

   if((temp = (TMP_TKN_PTR) GenAlloc(sizeof(TMP_TKN),"AllocTempToken")) ==
      NULL)
      MemoryError();
   temp->type  = 0;
   temp->field = NULL;
   temp->next = NULL;
   return(temp);
}

/* ----------------------------------------
 *  NAME   :  FreeTempToken()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            the temp_tokens.
 *
 *  INPUTS:   A pointer to the head of the
 *            token_list (TMP_TKN_PTR)
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeTempToken(list)
 TMP_TKN_PTR list;
 {
   TMP_TKN_PTR temp;

   while(list != NULL)
    {
      temp =list;
      list = list->next;
      temp->field = NULL;
      GenFree((char *)temp, (unsigned)sizeof(TMP_TKN));
    }

   return;
 }
 
 /* ------------------------------------------------*
  *  AllocMyTemplate()								*
  * ----------------------------------------------- */
  
  MY_TMPLATE_PTR AllocMyTemplate()
  {
    MY_TMPLATE_PTR temp;
    
    if((temp = (MY_TMPLATE_PTR)GenAlloc(sizeof(MY_TMPLATE),
        "AllocMyTemplate")) == NULL)
        MemoryError();
    temp->name = NULL;
    temp->value_list = NULL;
    temp->next = NULL;
    return(temp);
  }
/* ------------------------------------------------*
  *  FreeMyTemplate()								*
  * ----------------------------------------------- */
    
  void FreeMyTemplate(list)
  MY_TMPLATE_PTR list;
  {
     MY_TMPLATE_PTR temp;
     
     
      while(list != NULL)
       {
          temp = list;
          list = list->next;
          FreeFieldValue(temp->value_list);
          GenFree((char *)temp,(unsigned)sizeof(MY_TMPLATE));
       }
  }

/* ----------------------------------------------- *
 * AllocFactList()                               *
 *   allocate the memory to store the detail       *
 *   information a relation.                       *
 * ----------------------------------------------- */


 SPE_FACT_PTR AllocFactList()
 {
     SPE_FACT_PTR temp;

     if((temp = (SPE_FACT_PTR) GenAlloc(sizeof (SPE_FACT_LIST),
        "AllocFactList")) == NULL)
        MemoryError();
     temp->next = NULL;
     temp->field = NULL;
     temp->templ_slot = NULL;
     temp->num_activated_a = 0;
     temp->num_deactivated_a = 0;
     temp->activate_list_a = NULL;
     temp->deactivate_list_a = NULL;
     temp->num_activated_d = 0;
     temp->num_deactivated_d = 0;
     temp->activate_list_d = NULL;
     temp->deactivate_list_d = NULL;
     temp->retracted = NO;
     temp->deffact_name = NULL;
     temp->asserted_by_rule = NULL;
     return(temp);
 }

 /* ----------------------------------------
 *  NAME   :  FreeFactList()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            instances of a relation.
 *
 *  INPUTS:   A pointer to the head of the
 *            fact_list (SPE_FACT_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeFactList(list)
 SPE_FACT_PTR list;
 {
   SPE_FACT_PTR temp;

      while(list != NULL)
    {
      temp = list;
      list  = list->next;
      FreeMyTemplate(temp->templ_slot);
      FreeFieldValue(temp->field);
      FreeRuleList(temp->activate_list_a);
      FreeRuleList(temp->activate_list_d);
      FreeRuleList(temp->deactivate_list_a);
      FreeRuleList(temp->deactivate_list_d);
      temp->deffact_name = NULL;
      temp->asserted_by_rule = NULL;
      GenFree((char *)temp, (unsigned)sizeof(SPE_FACT_LIST));
    }

   return;
 }

/* ----------------------------------------------- *
 * NAME    : ALLOC_A_REL                           *
 *                                                 *
 * PURPOSE :                                       *
 *    This function allocates memory for           *
 *     ACT_RELATION structure and initialize it.   *
 *                                                 *
 * INPUT   : Nothing.                              *
 *                                                 *
 * RETURN  : pointer to the ACT_RELATION structure *
 * ----------------------------------------------- */

 ACT_REL_PTR AllocActivatedRelation()

 {
   ACT_REL_PTR temp;

   if((temp = (ACT_REL_PTR) GenAlloc (sizeof(ACT_RELATION),"AllocActivatedRelation")) ==
      NULL)
      MemoryError();

   temp->left = NULL;
   temp->right = NULL;
   temp->known = NO;
   temp->current_active = 0;
   temp->max_occurences = 0;
   temp->spec_fact_list = NULL;
   temp->num_assert = 0;
   temp->num_retract = 0;
   temp->activate_num  = 0;
   temp->deactivate_num = 0;
   return(temp);

 }

  /* ----------------------------------------
 *  NAME   :  FreeActivatedRelation()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            the activated relations.
 *
 *  INPUTS:   A pointer to the head of the
 *            relation_list (ACT_REL_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeActivatedRelation(list)
 ACT_REL_PTR list;
 {
  IF(list NEQ NULL) THEN
      IF(list->left NEQ NULL) THEN
         FreeActivatedRelation(list->left);
      END_IF
      IF(list->right NEQ NULL) THEN
         FreeActivatedRelation(list->right);
      END_IF

      FreeFactList(list->spec_fact_list);
      GenFree((char *)list, (unsigned)sizeof(ACT_RELATION));
   END_IF

   return;
 }
/* ------------------------------------------------------ *
 *  allocate memory to store information about a rule     *
 *  that activated aor deactivated because of the         *
 *  retraction or assertion of a relation.                *
 * ------------------------------------------------------ */


 RULE_NAME_PTR AllocRuleList()
 {
   RULE_NAME_PTR temp;

   if((temp = (RULE_NAME_PTR) GenAlloc(sizeof(struct rule_name_list),
      "AllocRuleList")) == NULL)
      MemoryError();
   temp->name = NULL;
   temp->counter = 0;
   temp->last_fact = NO;
   temp->next = NULL;
   return(temp);
 }

/* ----------------------------------------
 *  NAME   :  FreeRuleList()
 *
 *  PURPOSE:  This function frees all the
 *            memory tied up in a list of
 *            rules that were involed in the
 *            retraction or assertion of rules
 *
 *  INPUTS:   A pointer to the head of the
 *            fact_list (RULE_NAME_PTR).
 *
 *  OUTPUT:   Nothing
 * ----------------------------------------
 */
 void FreeRuleList(list)
 RULE_NAME_PTR list;
 {
   RULE_NAME_PTR temp;

      while(list != NULL)
    {
      temp = list;
      list = list->next;
      GenFree((char *)temp, (unsigned)sizeof(RULE_NAME));
    }

   return;
 }
/* ====================== END OF DRIBBLE FILE STUFFS ==================== */


/* ======================================================================== */

/* ----------------------------------------
 *  NAME   :   MemoryError
 *
 *  PURPOSE:   This function calls
 *             ErrorMessage() and
 *             EndCRSV() to exit from the
 *             program when malloc()
 *             returns NULL during a
 *             memory-allocation routine.
 *
 *  INPUTS:    None.
 *
 *  OUTPUT:    Nothing useful.
 * ----------------------------------------
 */

void MemoryError()
{
   ErrorMessage(ERROR,"Out of memory.");
   EndCRSV();
}


/* ======================================================================== */

/****************************************************/
/*        CLIPS Memory Module Stuff                 */
/****************************************************/

/*****************************************************************/
/* ReleaseMemory:  Returns all the lists of stored data structures */
/*   to the pool of free memory.                                 */
/*****************************************************************/

int ReleaseMemory(maximum)
int maximum;
{
   struct MemPtr *tmp_mem, *mptr;
   unsigned int i;
   int amount = 0;

   SendMessage("\n*** DEALLOCATING MEMORY ***\n",NO);

   for (i = (MEM_TABLE_SIZE - 1) ; i >= sizeof(char *) ; i--)
     {
      mptr = MemoryTable[i];
      while (mptr != NULL)
        {
         tmp_mem = mptr->next;
         GenFree((char *)mptr, i);
         mptr = tmp_mem;
         amount += i;
        }
      MemoryTable[i] = NULL;
      if ((amount > maximum) && (maximum > 0))
        {
         SendMessage("\n*** MEMORY  DEALLOCATED ***\n",NO);
         return;
        }
     }

   SendMessage("\n*** MEMORY  DEALLOCATED ***\n",NO);
  }


/*****************************************************************/
/* GM2: Allocates memory and does not initialize it              */
/*      (used by hash stuff)                                     */
/*****************************************************************/
char *gm2(size)
   unsigned int size;
  {
   struct MemPtr *strmem;

   if (size < sizeof(char *)) size = sizeof(char *);

   if (size >= MEM_TABLE_SIZE) return(GenAlloc(size,"gm2"));

   strmem = (struct MemPtr *) MemoryTable[size];
   if (strmem == NULL)
     {
      return(GenAlloc(size, "gm2"));
     }

   MemoryTable[size] = strmem->next;

   return ( (char *) strmem);
  }

/*****************************************************************/
/* RM: Returns memory (used by hash stuff)                       */
/*****************************************************************/
int rm(str,size)
  char *str;
  unsigned int size;
  {
   struct MemPtr *strmem;

   IF(size < sizeof(char *)) THEN
      size = sizeof(char *);
   END_IF

   IF(size >= MEM_TABLE_SIZE) THEN
      GenFree((char *)str, size);
      return(0);
   END_IF

   strmem = (struct MemPtr *) str;
   strmem->next = MemoryTable[size];
   MemoryTable[size] = strmem;
   return(1);
  }

/*****************************************************************/
/* BLOCK MEMORY FUNCTIONS (for Mac)                              */
/*****************************************************************/

#if MAC_LSC

struct mem_info
  {
   struct mem_info *prev_mem;
   struct mem_info *next_free;
   struct mem_info *last_free;
   int size;
  };

struct buf_info
  {
   struct buf_info *next_buf;
   struct buf_info *prev_buf;
   struct mem_info *buf_top_mem;
   int size;
  };

static struct buf_info   *top_buf;
static int                buf_info_size;
static int                info_size;
static int                init_alloc = FALSE;

/***********************/
/* InitMem:           */
/***********************/
static int InitMem(req_size)
  int req_size;
  {
  
   struct mem_info *mem_buf;
   unsigned int init_req;
   int usable_buf_size;

   if (sizeof(char) != 1)
     {
      SendMessage("\nSize of character data is not 1\n",NO);
      SendMessage("\nMemory allocation functions may not work\n",NO);
      return(0);
     }

   info_size = sizeof(struct mem_info);
   info_size = (((info_size - 1) / 4) + 1) * 4;

   buf_info_size = sizeof(struct buf_info);
   buf_info_size = (((buf_info_size - 1) / 4) + 1) * 4;

   init_req = (INITBUFFERSIZE > req_size ? INITBUFFERSIZE : req_size);
   init_req += info_size * 2 + buf_info_size;
   init_req = (((init_req - 1) / 4) + 1) * 4;

   usable_buf_size = init_req - (2 * info_size) - buf_info_size;

   /* make sure we get a buffer big enough to be usable */
   if ((req_size < INITBUFFERSIZE) &&
       (usable_buf_size <= req_size + info_size))
     {
      init_req = req_size + info_size * 2 + buf_info_size;
      init_req = (((init_req - 1) / 4) + 1) * 4;
      usable_buf_size = init_req - (2 * info_size) - buf_info_size;
     }

   top_buf = (struct buf_info *) malloc(init_req);

   if (top_buf == NULL)
     {
      SendMessage("\nUnable to allocate initial memory pool\n",NO);
      return(0);
     }

   top_buf->next_buf = NULL;
   top_buf->prev_buf = NULL;
   top_buf->buf_top_mem = (struct mem_info *) ((char *) top_buf +
      buf_info_size);
   top_buf->size = usable_buf_size;

   mem_buf = (struct mem_info *) ((char *) top_buf + buf_info_size +
      info_size + usable_buf_size);
   mem_buf->next_free = NULL;
   mem_buf->last_free = NULL;
   mem_buf->prev_mem = top_buf->buf_top_mem;
   mem_buf->size = 0;

   top_buf->buf_top_mem->next_free = NULL;
   top_buf->buf_top_mem->last_free = NULL;
   top_buf->buf_top_mem->prev_mem = NULL;
   top_buf->buf_top_mem->size = usable_buf_size;

   init_alloc = TRUE;
   return(1);
  }

/********************************/
/* InitNewBuffer:             */
/********************************/
static int InitNewBuffer(buf_ptr,req_size)
  struct buf_info *buf_ptr;
  int req_size;
  {
  
   unsigned int buf_size;
   int usable_buf_size;
   struct buf_info *new_buf;
   struct mem_info *top_new_buf;

   buf_size = (BUFFERSIZE > req_size ? BUFFERSIZE : req_size);
   buf_size += buf_info_size + info_size * 2;
   buf_size = (((buf_size - 1) / 4) + 1) * 4;

   usable_buf_size = buf_size - buf_info_size - (2 * info_size);

   /* make sure we get a buffer big enough to be usable */
   if ((req_size < BUFFERSIZE) &&
       (usable_buf_size <= req_size + info_size))
     {
      buf_size = req_size + info_size * 2 + buf_info_size;
      buf_size = (((buf_size - 1) / 4) + 1) * 4;
      usable_buf_size = buf_size - (2 * info_size) - buf_info_size;
     }

   new_buf = (struct buf_info *) malloc(buf_size);

   if (new_buf == NULL) return(0);

   new_buf->next_buf = NULL;
   new_buf->prev_buf = buf_ptr;
   new_buf->buf_top_mem = (struct mem_info *) ((char *) new_buf +
      buf_info_size);
   new_buf->size = usable_buf_size;
   buf_ptr->next_buf = new_buf;

   top_new_buf = (struct mem_info *) ((char *) new_buf + buf_info_size +
      info_size + usable_buf_size);
   top_new_buf->next_free = NULL;
   top_new_buf->last_free = NULL;
   top_new_buf->size = 0;
   top_new_buf->prev_mem = new_buf->buf_top_mem;

   new_buf->buf_top_mem->next_free = NULL;
   new_buf->buf_top_mem->last_free = NULL;
   new_buf->buf_top_mem->prev_mem = NULL;
   new_buf->buf_top_mem->size = usable_buf_size;

   return(1);
  }

/********************************/
/* RequestBlock:               */
/********************************/
static char *RequestBlock(req_size)
  int req_size;
  {
   struct mem_info *MemPtr, *last_free;
   struct buf_info *buf_ptr;

   /*==================================================*/
   /* Allocate initial memory pool block if it has not */
   /* already been allocated.                          */
   /*==================================================*/

   if (init_alloc == FALSE)
      {
       if (InitMem(req_size) == 0) return(NULL);
      }

   /*====================================================*/
   /* Make sure that the amount of memory requested will */
   /* fall on a word boundary.  Assume a word size of 4. */
   /*====================================================*/

   req_size = (((req_size - 1) / 4) + 1) * 4;

   /*=====================================================*/
   /* Search through the list of free memory for a block  */
   /* of the appropriate size.  If a block is found, then */
   /* allocate and return a pointer to it.                */
   /*=====================================================*/

   buf_ptr = top_buf;

   while (buf_ptr != NULL)
     {
      MemPtr = buf_ptr->buf_top_mem;

      while (MemPtr != NULL)
        {
         if ((MemPtr->size == req_size) ||
             (MemPtr->size > (req_size + info_size)))
           {
            AllocateBlock(buf_ptr,MemPtr,req_size);

            return((char *) MemPtr + info_size);
           }
         MemPtr = MemPtr->next_free;
        }

      if (buf_ptr->next_buf == NULL)
        {
         if (InitNewBuffer(buf_ptr,req_size) == 0)  /* get another buffer */
           { return(NULL); }
        }
      buf_ptr = buf_ptr->next_buf;
     }

   EndCRSV();
  }

/*************************/
/* AllocateBlock:       */
/*************************/
static int AllocateBlock(buffer_ptr,block_ptr,req_size)
  struct buf_info *buffer_ptr;
  struct mem_info *block_ptr;
  int req_size;
  {
   struct mem_info *split_mem, *next_mem;

   /*================================================*/
   /* Size of memory block is an exact match for the */
   /* requested amount of memory.                    */
   /*================================================*/
   if (req_size == block_ptr->size)
     {
      block_ptr->size = - req_size;
      if (block_ptr->last_free == NULL)
        {
         if (block_ptr->next_free != NULL)
           { buffer_ptr->buf_top_mem = block_ptr->next_free; }
         else
           { buffer_ptr->buf_top_mem = (struct mem_info *) ((char *)
                buffer_ptr + buf_info_size + info_size + buffer_ptr->size); }

        }
      else
        { block_ptr->last_free->next_free = block_ptr->next_free; }

      if (block_ptr->next_free != NULL)
        { block_ptr->next_free->last_free = block_ptr->last_free; }

      block_ptr->last_free = NULL;
      block_ptr->next_free = NULL;
      return(1);
     }

   /*=============================================*/
   /* Memory block is larger than memory request. */
   /* Split it.                                   */
   /*=============================================*/

   next_mem = (struct mem_info *)
              ((char *) block_ptr + info_size + block_ptr->size);

   split_mem = (struct mem_info *)
                  ((char *) block_ptr + (info_size + req_size));

   split_mem->size = block_ptr->size - (req_size + info_size);
   split_mem->prev_mem = block_ptr;

   split_mem->next_free = block_ptr->next_free;
   split_mem->last_free = block_ptr->last_free;

   next_mem->prev_mem = split_mem;

   if (split_mem->last_free == NULL)
     { buffer_ptr->buf_top_mem = split_mem; }
   else
     { split_mem->last_free->next_free = split_mem; }

   if (split_mem->next_free != NULL)
     { split_mem->next_free->last_free = split_mem; }

   block_ptr->size = - req_size;
   block_ptr->last_free = NULL;
   block_ptr->next_free = NULL;

   return(1);
  }

/*************************/
/* ReturnBlock:         */
/*************************/
static int ReturnBlock(MemPtr,size)
  char *MemPtr;
  int size;
  {
   struct mem_info *info_ptr, *last_mem, *next_mem, *nn_mem;
   struct buf_info *buf_ptr;
   char             buf[128];

   size = (((size - 1) / 4) + 1) * 4;

   info_ptr = (struct mem_info *) ((char *) MemPtr - info_size);

   IF(info_ptr == NULL) THEN
      return(-1);
   END_IF

   IF(info_ptr->size >= 0) THEN
      return(-2);
   END_IF

   IF(info_ptr->size != -size) THEN
      return(-3);
   END_IF

   info_ptr->size = - info_ptr->size;

   /* find out which buffer you're in */

   nn_mem = info_ptr;
   while (nn_mem->prev_mem != NULL)
     { nn_mem = nn_mem->prev_mem; }
   buf_ptr = (struct buf_info *) ((char *) nn_mem - buf_info_size);

   last_mem = info_ptr->prev_mem;
   next_mem = (struct mem_info *) ((char *) MemPtr + size);

   if (buf_ptr->buf_top_mem == NULL)
     { return(-4); }

   buf_ptr->buf_top_mem->last_free = info_ptr;
   info_ptr->next_free = buf_ptr->buf_top_mem;
   info_ptr->last_free = NULL;

   buf_ptr->buf_top_mem = info_ptr;

   /*=====================================================*/
   /* Combine this block with previous block if possible. */
   /*=====================================================*/

   if (last_mem != NULL)
     {
      if (last_mem->size > 0)
        {
         last_mem->size += (info_size + info_ptr->size);

         if (next_mem != NULL)
           { next_mem->prev_mem = last_mem; }
         else
           { return(-5); }

         /* Detach last free from list. */

         if (last_mem->last_free != NULL)
           { last_mem->last_free->next_free = last_mem->next_free; }

         if (last_mem->next_free != NULL)
           { last_mem->next_free->last_free = last_mem->last_free; }

         last_mem->next_free = info_ptr->next_free;
         if (info_ptr->next_free != NULL)
           { info_ptr->next_free->last_free = last_mem; }
         last_mem->last_free = NULL;
         buf_ptr->buf_top_mem = last_mem;
         info_ptr->last_free = NULL;
         info_ptr->next_free = NULL;
         info_ptr = last_mem;
        }
     }

   /*=================================================*/
   /* Combine this block with next block if possible. */
   /*=================================================*/

   if (next_mem == NULL) return(-6);
   if (info_ptr == NULL) return(-7);

   if (next_mem->size > 0)
     {
      info_ptr->size += (info_size + next_mem->size);

      nn_mem = (struct mem_info *) ((char *) next_mem + next_mem->size +
         info_size);
      if (nn_mem != NULL)
        { nn_mem->prev_mem = info_ptr; }
      else
        { return(-8); }

      if (next_mem->last_free != NULL)
        { next_mem->last_free->next_free = next_mem->next_free; }

      if (next_mem->next_free != NULL)
        { next_mem->next_free->last_free = next_mem->last_free; }

     }

   /* Free the buffer if we can, but */
   /* don't free the first buffer    */
   /* if it's the only one.          */

   if ((info_ptr->prev_mem == NULL) &&
       (info_ptr->size == buf_ptr->size))
     {
      if (buf_ptr->prev_buf != NULL)
        {
         buf_ptr->prev_buf->next_buf = buf_ptr->next_buf;
         if (buf_ptr->next_buf != NULL)
           { buf_ptr->next_buf->prev_buf = buf_ptr->prev_buf; }
         free(buf_ptr);
        }
      else
        {
         if (buf_ptr->next_buf != NULL)
           {
            buf_ptr->next_buf->prev_buf = NULL;
            top_buf = buf_ptr->next_buf;
            free((char *) buf_ptr);
           }
        }
     }

   return(1);
  }
#endif


