#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>

#include "crsv.h"

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVPRNT.C
 * -----------------------------
 */

extern void     ErrorMessage();
extern void     SendMessage();

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern int      GetToken();

/* -----------------------------
 *  From the file: CRSVHASH.C
 * -----------------------------
 */

extern HASH    *AddSymbol();

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern WL_PTR     	AllocWordList();
extern IL_PTR       AllocIntegerList();
extern FL_PTR       AllocFloatList();
extern TN_PTR     	AllocTokenNode();
extern PAT_TKN_PTR  AllocTokenInPattern();
extern PAT_LIST_PTR AllocPatternList();
extern void         FreeTokenInPattern();


/* -----------------------------
 *  From the file: CRSVPC.C
 * -----------------------------
 */

#if IBM_TBC || IBM_MSC
extern int        pc_window();
#endif


/* -----------------------------
 *  From the file: CRSVUNIX.C
 * -----------------------------
 */

#if UNIX_V
extern int        pc_window();
#endif


/* -----------------------------
 *  From the file: CRSVMISC.C
 * -----------------------------
 */

extern void       FreeTokenNode();


/* -----------------------------
 *  From the system:
 * -----------------------------
 */

extern double atof();


/* ===========  Functions defined here for Global use  ================ */

int      CheckInputs();
int      CheckTopConstruct();
int      FindTopConstruct();
int      FindRhtParen();
void     FindNewConstruct();
WL_PTR   AddToWordList();
IL_PTR   AddToIntegerList();
FL_PTR   AddToFloatList();
int      IsWordInList();
int      IsIntegerInList();
int      IsFloatInList();
FLD_PTR  SortFieldList();

void    SetCHECK_RULES();
void    SetCHECK_RELATIONS();
void    SetCHECK_EX_FLAG();
void    SetCHECK_COMMENTS();
void    SetVERBOSE();
void    SetCHECK_STYLE();
void    SetCREATE_DEFRELS();
void    SetCHECK_DEFRELS();
void    SetCHECK_DEBUG();
void    SetANALYZE_TRACE();
void    SetLITERAL();
void    SetINFERENCE_TRACE();

void QueueToken ();
TK_PTR DequeueToken ();
PAT_TKN_PTR    InsertTokenIntoPattern();
void           DequeueTokenFromPattern();
PAT_LIST_PTR   CreateAndLinkPatToList();

/* ===========  Functions defined here for internal use  ============== */

static char *ExtractArg();
static void  DisplayOptions();


/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSV.C
 * -----------------------------
 */

extern FILE     *CurrentFile;                    /* Pointer to current file */
extern int      MAX_SINGLE_LIST;


/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern double  FLOATNUMBER;                    /* GetToken number return  */
extern long    INTNUMBER;
extern char   *TKNWORD;                      /* GetToken string storage */

/* -----------------------------
 * From CRSVPROC.C
 * -----------------------------
 */
extern PAT_LIST_PTR CurrentPattern;               /* Pattern in process */
extern PAT_TKN_PTR  CurrentTokenInPattern;


/* ===========  Variables defined here for internal use  ============== */


/* ==============================================================
 * ================  Miscellaneous Functions  ===================
 * ==============================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :     QueueToken
 *
 *  PURPOSE:     Adds a token to a specified queue.
 *
 *  INPUTS :     A doubly dereferenced pointer to the head of
 *               the queue, the pointer to the token being added.
 *
 *  RETURNS:     None
 * -------------------------------------------------------------
 */

void QueueToken (token_list_ptr, token)
 TN_PTR *token_list_ptr;
 TK_PTR token;
 {
   TN_PTR new_node = AllocTokenNode();

   new_node->token = token;

   IF (*token_list_ptr EQ NULL) THEN
      *token_list_ptr = new_node;
   ELSE
      TN_PTR tail = *token_list_ptr;

      WHILE (tail->next NEQ NULL) DO
         tail = tail->next;
      DONE

      tail->next = new_node;
   END_IF
 }

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :     DequeueToken
 *
 *  PURPOSE:     Removes (and returns) the oldest token in a
 *               queue.
 *
 *  INPUTS :     A doubly dereferenced pointer to the head of
 *               the queue.
 *
 *  RETURNS:     The pointer to the token which was removed from
 *               the queue.
 * -------------------------------------------------------------
 */

TK_PTR DequeueToken (token_list_ptr)
 TN_PTR *token_list_ptr;
 {
   TN_PTR old_head = *token_list_ptr;
   TK_PTR token = (*token_list_ptr)->token;

   *token_list_ptr = (*token_list_ptr)->next;
   FreeTokenNode (old_head);

   return (token);
 }

/* ===============================================================
 *  InsertTokenIntoPattern()
 *  This function links the new node at the end of the list.
 * ===============================================================
 */
PAT_TKN_PTR InsertTokenIntoPattern(tkn,word,type)
PAT_TKN_PTR tkn;
char *word;
int type;
{
   PAT_TKN_PTR temp_word;
   HASH   *temp_hash;
   
   temp_word = tkn;
   IF(temp_word == NULL)THEN
       temp_word 	    = AllocTokenInPattern();
       temp_hash            = AddSymbol(word);
       temp_word->word      = temp_hash->contents;
       temp_word->type      = type;
       CurrentTokenInPattern          = temp_word;
       return(temp_word);
   END_IF
   WHILE(temp_word->next != NULL) DO
       temp_word = temp_word->next;
   END_WHILE
    
   temp_word->next    	      = AllocTokenInPattern();
   temp_hash                  = AddSymbol(word);
   temp_word->next->word      = temp_hash->contents;
   temp_word->next->type      = type;
   CurrentTokenInPattern 	  = temp_word->next;
   IF(type == WORD)THEN
      if(strcmp(word,"or") == 0)
          CurrentPattern->in_or = YES;
      else if(strcmp(word,"and") == 0)
          CurrentPattern->in_and = YES;
      else if(strcmp(word,"not") == 0)
          CurrentPattern->in_not = YES;
   ELSE_IF((type == MULTIPLE) || (type == BWORDS))
      CurrentPattern->has_wildcard = YES;
   END_IF
   return(tkn);
}

/* ===============================================================
 *  DequeueTokenFromPattern()
 *   This funtion takes a node away from the list
 * ===============================================================
 */
void DequeueTokenFromPattern(pat)
PAT_TKN_PTR pat;
{
   PAT_TKN_PTR temp_word1,temp_word2;
   
   temp_word1 = pat;
   IF(temp_word1 == NULL)
     return;
   while(temp_word1->next != NULL) DO
       temp_word2 = temp_word1;
       temp_word1 = temp_word1->next;
   END_WHILE
   IF(temp_word2 != NULL) THEN
     temp_word2->next = NULL;
   ELSE
     pat = NULL;
   END_IF
   FreeTokenInPattern(temp_word1);
}

/* ===================================================================
 *  CreateAndLinkPatToList()
 *    This function will create a new node for new_pattern and link it 
 *    to the list
 * ===================================================================
 */
 PAT_LIST_PTR CreateAndLinkPatToList(list)
 PAT_LIST_PTR list;
 {
   PAT_LIST_PTR temp;
   
   IF(list == NULL)THEN
       list = AllocPatternList();
       CurrentPattern = list;
       return(list);
   END_IF
   temp = list;
   while(temp->next_pattern != NULL)
     temp = temp->next_pattern;
   temp->next_pattern = AllocPatternList();
   CurrentPattern = temp->next_pattern;
   return(list);
 }
 

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :      CheckInputs
 *
 *  PURPOSE:      This function checks to see that all command
 *                line options are valid, that enough arguments
 *                are input, and builds the list of files to be
 *                processed.
 *
 *  INPUTS:
 *   int   *argc         Pointer to the number of arguments on
 *                       the command line.
 *   char *argv[]        An array of pointers to the arguments.
 *   char **c_file       Pointer to character pointer which is
 *                       to point at the file name for storing
 *                       defrelations.
 *   char **t_file       Pointer to character pointer which is
 *                       to point at the dribble file name.
 *
 *  RETURNS:       An integer flag. ERROR (-1) if an error is
 *                 found, or OK (0) if all arguments are OK.
 *                 Also, argv and argc are modified to contain
 *                 only the names of files to be processed.
 * -------------------------------------------------------------
 */

int CheckInputs(argc, argv, c_file, t_file)
int *argc;
char **argv, **c_file, **t_file;
{
   int i, j;

   IF((*argc EQ 2) AND (strcmp("?", argv[1]) EQ 0)) THEN
      DisplayOptions();
      return(ERROR);
   END_IF

   /*=============================*/
   /* Set default flag conditions */
   /*=============================*/

   ANALYZE_TRACE   = OFF;              /*  Option flag T  */
   CREATE_DEFRELS  = OFF;              /*  Option flag C  */
   CHECK_DEFRELS   = ON;               /*  Option flag D  */
   CHECK_EX_FLAG   = OFF;              /*  Option flag E  */
   LITERAL         = OFF;              /*  Option flag L  */
   CHECK_RULES     = OFF;              /*  Option flag R  */
   CHECK_STYLE     = ON;               /*  Option flag S  */
   VERBOSE         = OFF;              /*  Option flag V  */
   CHECK_RELATIONS = OFF;              /*  Option flag X  */
   CHECK_DEBUG     = OFF;              /*  Option flag Z  */
   CHECK_COMMENTS  = OFF;              /*  Not used currently, may use later */
   INFERENCE_TRACE  = OFF;             /*  Option flag I  */

   FOR (i = 1; i < *argc;) DO
      IF(argv[i][0] EQ '-') THEN
         char *option = ExtractArg (argv, argc, i);
         int suffix_arg = OFF; /* Set to ON when 'c', 'm', or 't' code found */

         FOR(j = 1; j < strlen(option); j++) DO
            char c = option[j];

            IF (isalpha(c) AND islower(c)) THEN
               c = toupper (c);
            END_IF

            IF(((c EQ 'T')OR (c EQ 'I') OR (c EQ 'C') OR (c EQ 'M')) AND suffix_arg IS_ON) THEN
               SendMessage
                  ("\nSyntax error: Simultaneous 'c', 'm', and 't' options\n",NO);
               DisplayOptions();
               return(ERROR);
            ELSE_IF(c EQ 'C')
               CREATE_DEFRELS = ON;
               suffix_arg = ON;

               IF ((*c_file = ExtractArg (argv, argc, i)) == NULL) THEN
                  SendMessage
                     ("\nNo defrelation file specified after '-c'.\n",NO);
                  DisplayOptions();
                  return (ERROR);
               END_IF
            ELSE_IF(c EQ 'D')
               CHECK_DEFRELS = OFF;
            ELSE_IF(c EQ 'E')
               CHECK_EX_FLAG = ON;
            ELSE_IF(c EQ 'H')
               DisplayOptions();
               return(ERROR);
            ELSE_IF(c EQ 'L')
               LITERAL = ON;
            ELSE_IF(c EQ 'M')
               int max;
               char *max_str = ExtractArg (argv, argc, i);

               suffix_arg = ON;
               max = atoi(max_str);
               IF (max > 0) THEN
                 MAX_SINGLE_LIST = max;
               ELSE
                 SendMessage("\n'-m' must be followed by a max length.\n",NO);
                 DisplayOptions();
                 return(ERROR);
               END_IF
            ELSE_IF(c EQ 'R')
               CHECK_RULES = ON;
            ELSE_IF(c EQ 'S')
               CHECK_STYLE = OFF;
            ELSE_IF(c EQ 'T')
               ANALYZE_TRACE = ON;
               IF(INFERENCE_TRACE IS_OFF) THEN
                 suffix_arg = ON;

                 IF ((*t_file = ExtractArg (argv, argc, i)) == NULL) THEN
                   SendMessage ("\nNo trace file specified after '-t'.\n",NO);

                   DisplayOptions();
                   return (ERROR);
                 END_IF
               END_IF
            ELSE_IF(c EQ 'V')
               VERBOSE = ON;
            ELSE_IF(c EQ 'X')
               CHECK_RELATIONS = ON;
            ELSE_IF(c EQ 'Z')
               CHECK_DEBUG = ON;
            ELSE
               SendMessage("\nInvalid option code.",NO);
               DisplayOptions();
               return(ERROR);
            END_IF
         END_FOR
      ELSE
         i++;
      END_IF
   END_FOR

   IF(*argc <= 1) THEN
      SendMessage
         ("\nNot enough arguments! No file(s) to cross-reference.\n",NO);
      DisplayOptions();
      return(ERROR);
   END_IF

   return(OK);
}


/* ==================================================================== */



/*
 * -------------------------------------------------------------
 *  NAME   :  ExtractArg
 *
 *  PURPOSE:  Removes an argument from the argument list
 *            specified via argv.
 *
 *  INPUTS :  The argument list and the argument to remove.
 *
 *  RETURNS:  The removed argument.
 * -------------------------------------------------------------
 */

static char *ExtractArg (argv, argc, arg)
   char **argv;
   int *argc;
   int arg;
{
   char *extracted_arg = argv[arg];
   char *str1, *str2;
   int i;

   IF (arg != --(*argc)) THEN
     for(;arg < *argc;)
        {
           argv[arg] = argv[arg+1];
           arg++;
        }
   END_IF

   return (extracted_arg);
}


/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  DisplayOptions
 *
 *  PURPOSE:  This function prints a list of the valid CRSV
 *            options.
 *
 *  INPUTS :  None.
 *
 *  RETURNS:  Nothing
 * -------------------------------------------------------------
 */

static void DisplayOptions()
{
   SendMessage("\nSyntax: crsv [-options [file-name]] rule-files ",NO);
   SendMessage("[-options [file-name]] ...",NO);
   SendMessage("\nValid options are:\n",NO);
   SendMessage("     r - Display rules or deffacts summary info\n",NO);
   SendMessage("     x - Display cross referenced relations and field info\n",NO);
   SendMessage("     e - Do not display external function info\n",NO);
   SendMessage("     c - Create a file with defrelation info\n",NO);
   SendMessage("         'file-name' names the output file\n",NO);
   SendMessage("     s - Turn off style warnings and code analysis\n",NO);
   SendMessage("     l - Add rule field values to defrelation defaults\n",NO);
   SendMessage("     m - Maximum number of single field values printed\n",NO);
   SendMessage("         (follow m by a space and the new max number)\n",NO);
   SendMessage("     d - Do not use defrelations to verify rules\n",NO);
   SendMessage("     t - Analyze trace file \n",NO);
   SendMessage("         'file-name' names the trace file\n",NO);
   SendMessage("? or h - Display valid options\n",NO);
   SendMessage("     v - Be verbose\n",NO);
   SendMessage("NOTE: Options 'c', 'm', and 't' may NOT be simultaneously\n",NO);
   SendMessage("      used in the same options argument.\n",NO);
}


/* ==================================================================== */

/*
 * ----------------------------------------------------------------
 *  NAME   :       CheckTopConstruct
 *
 *  PURPOSE:       This function checks the input file for
 *                 the start of a rule or fact block. If neither
 *                 construct is found, an error message is
 *                 printed out.
 *
 *  INPUTS:        Two arguments,
 *                 a pointer to the input file (FILE *), and
 *                 the current token (int).
 *
 *  RETURNS:       An integer flag which means:
 *     ERROR       (-1)  if an error is found,
 *     DEFRULE     (1)   if at the beginning of a rule,
 *     DEFFACTS    (2)   if at the beginning of a fact block,
 *     DEFRELATION (3)   if at the beginning of a defrelation,
 *     DEFEXTERNAL (4)   if at the beginning of a defexternal.
 *     DEFTEMPLATE (5)   if at the beginning of a deftemplate.
 *     DEFCLASS    (6)   if at the beginning of a defclass.
 *     DEFINSTANCES(7)   if at the beginning of a definstances.
 *     DEFMESSAGE-HANDLER (8) if at the beginning of a defmessage-handler
 *     DEFGLOBAL   (9)   if at the beginning of a defglobal.
 *     DEFGENERIC  (10)  if at the beginning of a defgeneric.
 *     DEFMETHOD   (11)  if at the beginning of a defmethod.
 *
 *  NOTES  :       This function is called when I would normally
 *                 expect the next item to be a def... something.
 *                 It returns error if the very next thing ISN'T
 *                 a def... something.
 *
 *                 Note that the opening parenthesis is supposed
 *                 to have already been stripped off.
 * ----------------------------------------------------------------
 */

int CheckTopConstruct(source, token)
FILE *source;
int   token;
{
   int    rtn;

   IF(token NEQ LPAREN) THEN
      ErrorMessage(ERROR,
       "Expected left parenthesis to begin defrule or deffacts statement");
      return(ERROR);
   ELSE
      token = GetToken(source);

      IF(token NEQ WORD) THEN
         rtn = ERROR;
      ELSE_IF(strcmp(TKNWORD,"deffacts") EQ 0)
         rtn = DEFFACTS;
      ELSE_IF(strcmp(TKNWORD,"defrule") EQ 0)
         rtn = DEFRULE;
      ELSE_IF(strcmp(TKNWORD,"defrelation") EQ 0)
         rtn = DEFRELATION;
      ELSE_IF(strcmp(TKNWORD,"defexternal") EQ 0)
         rtn = DEFEXTERNAL;
      ELSE_IF(strcmp(TKNWORD,"deftemplate") EQ 0)
         rtn = DEFTEMPLATE;
      ELSE_IF(strcmp(TKNWORD,"defclass") EQ 0)
         rtn = DEFCLASS;
      ELSE_IF(strcmp(TKNWORD,"definstances") EQ 0)
         rtn = DEFINSTANCES;
      ELSE_IF(strcmp(TKNWORD,"defmethod") EQ 0)
         rtn = DEFMETHOD;
      ELSE_IF(strcmp(TKNWORD,"defmessage-handler") EQ 0)
         rtn = DEFMESSAGE_HANDLER;
      ELSE_IF(strcmp(TKNWORD,"defglobal") EQ 0)
         rtn = DEFGLOBAL;
      ELSE_IF(strcmp(TKNWORD,"defgeneric") EQ 0)
         rtn = DEFGENERIC;
      ELSE_IF(strcmp(TKNWORD,"deffunction") EQ 0)
         rtn = DEFFUNCTION;
      ELSE
         rtn = ERROR;
      END_IF

      IF(rtn EQ ERROR) THEN
         ErrorMessage (ERROR, "Expected a defrule, deffacts, defrelation, ");
         SendMessage("\n      defexternal,deftemplate,defclass,definstances,",NO);
         SendMessage("\n      defmessage-handler,defglobal,defgeneric,defmethod",NO);
         SendMessage("\n      deffunction construct but found a ",NO);
         SendMessage (TKNWORD,NO);
         SendMessage (" construct.\n        Check that all constructs ",NO);
         SendMessage("have the proper number of matching parentheses.\n",NO);
         return(ERROR);
      ELSE
         return(rtn);
      END_IF
   END_IF
}

/* ======================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :       FindTopConstruct
 *
 *  PURPOSE:       This function searches the input file for
 *                 the start of a rule or fact block. If neither
 *                 construct is found, an error message is
 *                 printed out.
 *
 *  INPUTS:        Two arguments,
 *                 a pointer to the input file (FILE *), and
 *                 the current token (int).
 *
 *  RETURNS:       An integer flag which means:
 *     ERROR       (-1)  if an error is found,
 *     DEFRULE     (1)   if at the beginning of a rule,
 *     DEFFACTS    (2)   if at the beginning of a fact block.
 *     DEFRELATION (3)   if at the beginning of a defrelation block
 *     DEFEXTERNAL (4)   if at the beginning of a defexternal block
 *     DEFTEMPLATE (5)   if at the beginning of a deftemplate block
 *     DEFOBJECT   (6)	 if at the beginning of a defobjection block
 *
 *  NOTES  :       This function is called when I have gotten
 *                 really confused somehow. It searches all
 *                 tokens until a def... something is found. If
 *                 one is never found, it returns ERROR when
 *                 it reaches the end of the file.
 *
 *                 Note that the opening parenthesis is supposed
 *                 to have already been stripped off.
 * -------------------------------------------------------------
 */

int FindTopConstruct(source, token)
FILE *source;
int token;
{
   while(TRUE) DO
      IF(token EQ STOP) THEN
         return(ERROR);
      ELSE_IF(token EQ LPAREN)
         token = GetToken(source);
         IF((token EQ WORD) AND (strcmp(TKNWORD,"deffacts") EQ 0)) THEN
            return(DEFFACTS);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defrule") EQ 0))
            return(DEFRULE);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defrelation") EQ 0))
            return(DEFRELATION);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defexternal") EQ 0))
            return(DEFEXTERNAL);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"deftemplate") EQ 0))
            return(DEFTEMPLATE);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defclass") EQ 0))
            return(DEFCLASS);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"definstances") EQ 0))
            return(DEFINSTANCES);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defmethod") EQ 0))
            return(DEFMETHOD);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defmessage-handler") EQ 0))
            return(DEFMESSAGE_HANDLER);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defglobal") EQ 0))
            return(DEFGLOBAL);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defgeneric") EQ 0))
            return(DEFGENERIC);
         ELSE_IF((token EQ WORD) AND (strcmp(TKNWORD,"defmethod") EQ 0))
            return(DEFMETHOD);
         ELSE
            SaveToken(token,TKNWORD,FLOATNUMBER,INTNUMBER);
         END_IF
      END_IF
      token = GetToken(source);
   END_WHILE
}

/* ======================================================================== */

/* -----------------------------------------------
 *  NAME   :         FindRhtParen
 *
 *  PURPOSE:         This function searches the
 *                   input file for an unmatched
 *                   right paren. It uses GetToken
 *                   and so is not confused by
 *                   strings.
 *
 *  INPUTS:          None.
 *
 *  RETURNS:         Nothing.
 * -------------------------------------------------
 */

int FindRhtParen()
{
   int count;
   int token;
   char *TempTKNWORD;

   count = 1;

   while (count NEQ 0) DO
      token = GetToken(CurrentFile);
      IF(token EQ LPAREN) THEN
         token = GetToken(CurrentFile);
         if(token == WORD)
           {
             if((strcmp(TKNWORD,"defrule") == 0) ||
                (strcmp(TKNWORD,"deffacts") == 0) ||
                (strcmp(TKNWORD,"defglobal") == 0) ||
                (strcmp(TKNWORD,"deftemplate") == 0)||
                (strcmp(TKNWORD,"defclass") == 0) ||
                (strcmp(TKNWORD,"definstances") == 0)||
                (strcmp(TKNWORD,"defgeneric") == 0)||
                (strcmp(TKNWORD,"defrelation") == 0)||
                (strcmp(TKNWORD,"defexternal") == 0)||
                (strcmp(TKNWORD,"deffunction") == 0)||
                (strcmp(TKNWORD,"defmethod") == 0)||
                (strcmp(TKNWORD,"defmessage-handler") == 0))
              {
                TempTKNWORD = TKNWORD;
                TKNWORD = "(";
                token = LPAREN;
                INTNUMBER = 0;
                FLOATNUMBER = 0.0;
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
                TKNWORD = TempTKNWORD;
                token = WORD;
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
                ErrorMessage(ERROR,"Missing the closing parenthesis\n");
                return(ERROR);
              }
             else
              {
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
                count++;
              }
           }
           else
            {
              SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
              count++;
            }
      ELSE_IF(token EQ RPAREN)
         count--;
      ELSE_IF (token EQ STOP)
         ErrorMessage(ERROR,"Missing the closing parenthesis\n");
         return(ERROR);
      END_IF
   END_WHILE
   return(OK);
}

/* -----------------------------------------------
 *  NAME   :         FindNewConstruct
 *
 *  PURPOSE:         This function searches the
 *                   input file for a new construct. 
 *                   It uses GetToken
 *                   and so is not confused by
 *                   strings.
 *
 *  INPUTS:          None.
 *
 *  RETURNS:         Nothing.
 * -------------------------------------------------
 */

void FindNewConstruct()
{
   int count;
   int token;
   char *TempTKNWORD;

   count = 1;
   token = GetToken(CurrentFile);
   while (token != STOP) DO
      IF(token EQ LPAREN) THEN
         token = GetToken(CurrentFile);
         if(token == WORD)
           {
             if((strcmp(TKNWORD,"defrule") == 0) ||
                (strcmp(TKNWORD,"deffacts") == 0) ||
                (strcmp(TKNWORD,"defglobal") == 0) ||
                (strcmp(TKNWORD,"deftemplate") == 0)||
                (strcmp(TKNWORD,"defclass") == 0) ||
                (strcmp(TKNWORD,"definstances") == 0)||
                (strcmp(TKNWORD,"defgeneric") == 0)||
                (strcmp(TKNWORD,"defrelation") == 0)||
                (strcmp(TKNWORD,"defexternal") == 0)||
                (strcmp(TKNWORD,"deffunction") == 0)||
                (strcmp(TKNWORD,"defmethod") == 0)||
                (strcmp(TKNWORD,"defmessage-handler") == 0))
              {
                TempTKNWORD = TKNWORD;
                TKNWORD = "(";
                token = LPAREN;
                INTNUMBER = 0;
                FLOATNUMBER = 0.0;
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
                TKNWORD = TempTKNWORD;
                token = WORD;
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
                return;
              }
             else
                SaveToken (token, TKNWORD, FLOATNUMBER,INTNUMBER);
          }
       END_IF
      token = GetToken(CurrentFile);
   END_WHILE
}
/* ======================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   : SortFieldList
 *
 *  PURPOSE: This function sorts a list of FLD_PTR's so that
 *           the fields are in numerical order.
 *
 *  INPUTS : A pointer to the field list (FLD_PTR),
 *
 *  RETURNS: A pointer to the sorted list (FLD_PTR).
 * -------------------------------------------------------------
 */

FLD_PTR SortFieldList(list)
FLD_PTR list;
{
   FLD_PTR temp1, temp2, temp3, prev, new_list;

   IF(list->next_field EQ NULL) THEN
      return(list);                     /* Dont sort a single element list! */
   END_IF

   temp1    = list->next_field;
   new_list = list;
   new_list->next_field = NULL;

   WHILE(temp1 NEQ NULL) DO
      temp3 = temp1->next_field;

      temp2 = new_list;
      prev  = NULL;
      while(temp2 NEQ NULL) DO
         IF(temp1->field_num <= temp2->field_num) THEN
            break;
         END_IF
         prev  = temp2;
         temp2 = temp2->next_field;
      END_WHILE

      IF(prev NEQ NULL) THEN
         temp1->next_field = prev->next_field;
         prev->next_field  = temp1;
      ELSE
         temp1->next_field = new_list;
         new_list          = temp1;
      END_IF

      temp1 = temp3;
   END_WHILE

   return(new_list);
}

/* ======================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   : AddToWordList
 *
 *  PURPOSE: This function checks to see if a word is already
 *           in a word list. If it is, it increments the times
 *           used counter. If it isn't, it adds the word to
 *           the FRONT of the list.
 *
 *  INPUTS : Three arguments, a pointer to the word list (WL_PTR),
 *           the word to search for (char *), and the type of the
 *           word.
 *
 *  RETURNS: The modified word list (WL_PTR).
 * -------------------------------------------------------------
 */

WL_PTR AddToWordList(list, word, type, duplication_check)
WL_PTR  list;
char   *word;
int     type,duplication_check;
{
   int     rtn;
   WL_PTR  temp_wl;
   HASH   *temp_hash;
   char   buf[80];

   temp_wl  = list;

   while(temp_wl NEQ NULL) DO

      /* --------------------------------------------------
       *  If the type is a NUMBER then convert the string
       *  representation to a float value and then compare
       *  values.  This allows for such cases as 200 and
       *  200.0 to be considered the same number.
       * ---------------------------------------------------
       */

      IF(type EQ NUMBER) THEN
        IF (atof(word) EQ atof(temp_wl->word)) THEN
           rtn = 0;
        ELSE
           rtn = 1;
        END_IF
      ELSE
        rtn = strcmp(word, temp_wl->word);
      END_IF
      IF(rtn EQ 0) THEN
         IF(duplication_check)THEN
            sprintf(buf,
               "<%s> has been listed more than once in the allowed list.",
               word);
            ErrorMessage(WARNING,buf);
         END_IF
         temp_wl->times_used++;
         return(list);
      END_IF

      temp_wl = temp_wl->next_word;
   END_WHILE

   temp_wl              = AllocWordList();
   temp_hash            = AddSymbol(word);
   temp_wl->word        = temp_hash->contents;
   temp_wl->next_word   = list;
   temp_wl->times_used  = 1;
   temp_wl->type        = type;

   return(temp_wl);
}

/* ======================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   : IsWordInList
 *
 *  PURPOSE: This function checks to see if a word is in a
 *           word list. If it is, it return YES, otherwise
 *           it returns NO.
 *
 *  INPUTS : Two arguments, a pointer to the word list (WL_PTR),
 *           and the word to search for (char *).
 *
 *  RETURNS: An integer, either YES or NO.
 * -------------------------------------------------------------
 */

int IsWordInList(list, word)
WL_PTR  list;
char   *word;
{
   int    rtn;
   WL_PTR temp_wl;

   temp_wl  = list;

   while(temp_wl NEQ NULL) DO
      rtn = strcmp(word, temp_wl->word);
      IF(rtn EQ 0) THEN
         return(YES);
      END_IF

      temp_wl = temp_wl->next_word;
   END_WHILE

   return(NO);
}

/* ======================================================================== */


/*
 * -------------------------------------------------------------
 *  NAME   : AddToFloatList
 *
 *  PURPOSE: This function checks to see if a number is already
 *           in a number list. If it is, it increments the times
 *           used counter. If it isn't, it adds the number to
 *           the FRONT of the list.
 *
 *  INPUTS : Two arguments, a pointer to the number list (FL_PTR),
 *           and the number to search for (char *).
 *
 *  RETURNS: The modified number list (FL_PTR).
 * -------------------------------------------------------------
 */

FL_PTR AddToFloatList(list, number,duplication_check)
FL_PTR  list;
double   number;
int     duplication_check;
{
   FL_PTR temp_fl;

   temp_fl  = list;

   WHILE(temp_fl NEQ NULL) DO
      IF(number EQ temp_fl->number) THEN
        IF(duplication_check)THEN
           sprintf(MsgBuf,
              "<%f> has been listed more than once in the allowed list.",
              number);
           ErrorMessage(WARNING, MsgBuf);
         END_IF
         temp_fl->times_used++;
         return(list);
      END_IF

      temp_fl = temp_fl->next_number;
   END_WHILE

   temp_fl              = AllocFloatList();
   temp_fl->number      = number;
   temp_fl->next_number = list;
   temp_fl->times_used  = 1;

   return(temp_fl);
}



/*
 * -------------------------------------------------------------
 *  NAME   : AddToIntegerList
 *
 *  PURPOSE: This function checks to see if a number is already
 *           in a number list. If it is, it increments the times
 *           used counter. If it isn't, it adds the number to
 *           the FRONT of the list.
 *
 *  INPUTS : Two arguments, a pointer to the number list (iL_PTR),
 *           and the number to search for (char *).
 *
 *  RETURNS: The modified number list (IL_PTR).
 * -------------------------------------------------------------
 */

IL_PTR AddToIntegerList(list, number,duplication_check)
IL_PTR  list;
long    number;
int     duplication_check;
{
   IL_PTR temp_il;

   temp_il  = list;

   while(temp_il NEQ NULL) DO
      IF(number EQ temp_il->number) THEN
        IF(duplication_check)THEN
           sprintf(MsgBuf,
              "<%f> has been listed more than once in the allowed list.",
              number);
           ErrorMessage(WARNING, MsgBuf);
         END_IF
         temp_il->times_used++;
         return(list);
      END_IF

      temp_il = temp_il->next_number;
   END_WHILE

   temp_il              = AllocIntegerList();
   temp_il->number      = number;
   temp_il->next_number = list;
   temp_il->times_used  = 1;

   return(temp_il);
}

/* ======================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   : IsFloatInList
 *
 *  PURPOSE: This function checks to see if a number is in a
 *           number list. If it is, it return YES, otherwise
 *           it returns NO.
 *
 *  INPUTS : Two arguments, a pointer to the number list (FL_PTR),
 *           and the number to search for (char *).
 *
 *  RETURNS: An integer, either YES or NO.
 * -------------------------------------------------------------
 */

int IsFloatInList(list, number)
FL_PTR  list;
double   number;
{
   FL_PTR temp_fl;

   temp_fl = list;

   while(temp_fl NEQ NULL) DO
      IF(number EQ temp_fl->number) THEN
         return(YES);
      END_IF

      temp_fl = temp_fl->next_number;
   END_WHILE

   return(NO);
}

/*
 * -------------------------------------------------------------
 *  NAME   : IsIntegerInList
 *
 *  PURPOSE: This function checks to see if an integer is in an
 *           integer list. If it is, it return YES, otherwise
 *           it returns NO.
 *
 *  INPUTS : Two arguments, a pointer to the number list (IL_PTR),
 *           and the number to search for (char *).
 *
 *  RETURNS: An integer, either YES or NO.
 * -------------------------------------------------------------
 */

int IsIntegerInList(list, number)
IL_PTR  list;
long   number;
{
   IL_PTR temp_il;

   temp_il = list;

   while(temp_il NEQ NULL) DO
      IF(number EQ temp_il->number) THEN
         return(YES);
      END_IF

      temp_il = temp_il->next_number;
   END_WHILE

   return(NO);
}


/* ------------------------------------------------------------*/
/*   These funtions are for the user interface                 */
/* ------------------------------------------------------------*/

void SetCHECK_RULES(flag)
int flag;
{
   CHECK_RULES = flag;
}

void SetLITERAL(flag)
int flag;
{
   LITERAL = flag;
}

void SetCHECK_RELATIONS(flag)
int flag;
{
   CHECK_RELATIONS = flag;
}

void SetCHECK_COMMENTS(flag)
int flag;
{
   CHECK_COMMENTS = flag;
}

void SetCHECK_STYLE(flag)
int flag;
{
   CHECK_STYLE = flag;
}

void SetCREATE_DEFRELS(flag)
int flag;
{
   CREATE_DEFRELS = flag;
}

void SetCHECK_DEFRELS(flag)
int flag;
{
   CHECK_DEFRELS = flag;
}

void SetCHECK_DEBUG(flag)
int flag;
{
   CHECK_DEBUG = flag;
}

void SetVERBOSE(flag)
int flag;
{
   VERBOSE = flag;
}

void SetCHECK_EX_FLAG(flag)
int flag;
{
   CHECK_EX_FLAG = flag;
}

void SetANALYZE_TRACE(flag)
int flag;
{
   ANALYZE_TRACE = flag;
}


void SetINFERENCE_TRACE(flag)
int flag;
{
   INFERENCE_TRACE = flag;
}
