#include <stdio.h>
#include <string.h>
#include "crsv.h"

/********************************************************************
 *  This file contains all the functions used to display messages
 *  during execution as well as the functions which display summary
 *  information at the end of a run. There are entry points for
 *  each type of summary printout; rules, deffacts, relations
 *  and external functions, as well as message functions.
 ********************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

#if(IBM_TBC || IBM_MSC) 

/* -------------------
 *  from CRSVPC.C
 * -------------------
 */

extern void CrsvErrorMessage();
extern void CrsvMessage();

#endif

/* -------------------------
 *  From the file: CRSV.C
 * -------------------------
 */

extern DT_PTR DeftemplateHead;

/* ---------------------------
 *  From the file: CRSVTMP.C
 * ---------------------------
 */

extern DT_PTR FindDeftemplate();
extern DS_PTR FindDefslotByPosition();

/* ---------------------------
 *  From the file: CRSVTR.C
 * ---------------------------
 */

extern ACT_REL_PTR FindActivatedRelation();


/* ===========  Functions defined here for Global use  ================ */

void RelationsPrintout();
void RulePrintout();
void FactBlockPrintout();
void ExternalFuncsPrintout();
void ErrorMessage();
void SendMessage();
void SendText ();

/* ===========  Functions defined here for internal use  ============== */

static void PrintRuleInformation();
static void PrintRules();
static void PrintRelations();
static void PrintRelationInformation();
static void PrintExtFuncInformation();
static void PrintExternalFunctions();
static void PrintDefrelationInformation();
static void PrintDefrelations();
static void PrintDeffields();


/* ===========  Variables defined here for Global use  ================ */

/* -----------------------
 *  From the file: CRSV.C
 * -----------------------
 */

extern int NUM_ERRORS;
extern int NUM_WARNINGS;


/* ===========  Variables defined Externally but used here  =========== */

extern char         *CurrentObjectName;          /* Name of current object   */
extern int           CurrentLineNumber;          /* Current line number      */
extern short         IssuedErrorMessage;       /* Set when ErrorMessage called */
extern ACT_REL_PTR   ActvatedRelationHead;       /* List of dribble file relations */


/* ===========  Variables defined here for internal use  ============== */


/* ===================================================================
 *                       FACT BLOCK Printout Functions
 * ===================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :  FactBlockPrintout
 *
 *  PURPOSE:  This function prints summary information about
 *            all the deffacts used in a KB.
 *
 *  INPUTS :  A pointer to a list of deffacts (FB_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
void FactBlockPrintout(list_head)
FB_PTR list_head;
{
   FB_PTR   temp;

   IF(list_head NEQ NULL) THEN   
      SendMessage
         ("\n\n===================== DEFFACTS SUMMARY ====================",NO);
      SendMessage
         ("\n  Name                                      # of facts  File",NO);

      temp = list_head;
      while(temp NEQ NULL) DO
         sprintf(MsgBuf, "\n  %-43s  %6d   %-25s", temp->name, 
            temp->num_facts, temp->file);
         SendMessage(MsgBuf,NO);
         temp = temp->next_block;
      END_WHILE

      SendMessage("\n\n",NO);
   END_IF
}

/* ===================================================================
 *                          RULE Printout Functions
 * ===================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :  RulesPrintout
 *
 *  PURPOSE:  This function prints summary information about
 *            all the rules used in a KB.
 *
 *  INPUTS :  A pointer to a list of rules (RULE_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Merely prints a header then calls a recursive
 *            function which actually works through the
 *            binary tree.
 * -------------------------------------------------------------
 */

/*ARGSUSED*/ /* Make lint happy */
void RulesPrintout(node, drb_node)
RULE_PTR   node;
ACT_R_PTR  drb_node;
{
   ACT_R_PTR  rule_ptr;

   IF(node NEQ NULL) THEN
      SendMessage
         ("\n\n======================= RULE SUMMARY ======================",NO);
      PrintRules(node);
   END_IF
   
   IF((ANALYZE_TRACE IS_ON AND VERBOSE IS_ON) AND (drb_node NEQ NULL)) THEN
      SendMessage
("\n\n===================== Dribble file rule summary ====================",NO);

      SendMessage
         ("Rule                                                    Times  Times  Times\n",NO);
      SendMessage
         ("Names                                                   Fired  Acti   Deact\n",NO);
      SendMessage
         ("                                                               vated  ivated\n",NO);
      SendMessage
         ("-----                                                   -----  -----  ------\n",NO);

      rule_ptr = drb_node;
      while(rule_ptr NEQ NULL) DO
         sprintf(MsgBuf,"%-54.54s  %5d  %5d  %5d\n",
                 rule_ptr->name,
                 rule_ptr->fire_num, 
                 rule_ptr->activate_num, 
                 rule_ptr->deactivate_num);
          SendMessage(MsgBuf,NO);
          rule_ptr = rule_ptr->next;
      END_WHILE
   END_IF
}

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintRules
 *
 *  PURPOSE:  This is a recursive function which processes
 *            through all nodes in a binary tree of rule
 *            structures.
 *
 *  INPUTS :  A pointer to a list of rules (RULE_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Just processes binary tree nodes, actual
 *            printing done in PrintRuleInformation.
 * -------------------------------------------------------------
 */
 
static void PrintRules(node)
RULE_PTR node;
{
   IF(node NEQ NULL) THEN
     PrintRules(node->lft_rule);
     PrintRuleInformation(node);
     PrintRules(node->rht_rule);
   END_IF
}

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintRuleInformation
 *
 *  PURPOSE:  This is the function which actually prints
 *            rule summary information.
 *
 *  INPUTS :  A pointer to a rule (RULE_PTR *)
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

static void PrintRuleInformation(node)
RULE_PTR node;
{
   char *dum;

   IF(node->salience_set IS_YES) THEN
      dum = "Yes";
   ELSE
      dum = "No ";
   END_IF

   sprintf(MsgBuf, "\n\nINFO FOR RULE: %.40s\nFrom file    : %.20s",
      node->name, node->file);
   SendMessage(MsgBuf,NO);
   SendMessage(
   "\nSalience  Salience   # of      # of     # of      # of     # of",NO);
   SendMessage(
   "\n  Set?     Value    patterns  Actions  Retracts  Asserts  Ex Funcs",NO); 

   sprintf(MsgBuf, "\n  %s  %8d  %7d  %8d  %7d  %8d  %7d", dum,
      node->salience_val, node->num_patterns, node->num_actions,
      node->num_retracts, node->num_asserts,  node->num_ex_funcs);
   SendMessage(MsgBuf,NO);
}
      
/* ===================================================================
 *                       Relation Printout Functions
 * ===================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :  RelationsPrintout
 *
 *  PURPOSE:  This function prints summary information about
 *            all the relations used in a KB.
 *
 *  INPUTS :  A pointer to a list of relations (REL_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Merely prints a header then calls a recursive
 *            function which actually works through the
 *            binary tree.
 * -------------------------------------------------------------
 */

void RelationsPrintout(node /* , drb_node*/)
REL_PTR      node;
/*ACT_REL_PTR  drb_node; Not currently used. */
{
   SendMessage
      ("\n\n===============  RELATION/TEMPLATE SUMMARY  ===============",NO);

   PrintRelations(node);
}


/*
 * -------------------------------------------------------------
 *  NAME   :  PrintRelations
 *
 *  PURPOSE:  This is a recursive function which processes
 *            through all nodes in a binary tree of relation
 *            structures.
 *
 *  INPUTS :  A pointer to a list of relations (REL_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Just processes binary tree nodes, actual
 *            printing done in PrintRelationInformation.
 * -------------------------------------------------------------
 */
 
static void PrintRelations(node)
REL_PTR node;
{
   IF(node NEQ NULL) THEN
     PrintRelations(node->lft_rel);
     PrintRelationInformation(node);
     PrintRelations(node->rht_rel);
   END_IF
}

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintRelationInformation
 *
 *  PURPOSE:  This is the function which actually prints
 *            relation summary information.
 *
 *  INPUTS :  A pointer to a relation (REL_PTR *)
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

static void PrintRelationInformation(node)
REL_PTR node;
{
   FBR_PTR     temp_fbr;
   RR_PTR      temp_rr;
   FLD_PTR     temp_fld;
   WL_PTR      temp_wl;
   DT_PTR      template;
   DS_PTR      slot;
   int         MULTI;
   ACT_REL_PTR temp_act;
 
   SendMessage
      ("\n\n\n-----------------------------------------------------------\n",NO);

   template = FindDeftemplate(DeftemplateHead,node->name);
   IF(template) THEN
     sprintf(MsgBuf,"   TEMPLATE NAME: %.40s",template->name);
   ELSE
     sprintf(MsgBuf,"   RELATION NAME: %.40s",node->name);
   END_IF
   SendMessage(MsgBuf,NO);
   SendMessage
      ("\n-----------------------------------------------------------",NO);
   
   IF(node->fb_list NEQ NULL) THEN
      SendMessage("\n\nAppears in the following deffacts:",NO);
      temp_fbr = node->fb_list;
      while(temp_fbr NEQ NULL) DO
         IF(temp_fbr->num_fields > 0) THEN
            sprintf(MsgBuf, "\n%.40s  (%d fields)",
                    temp_fbr->fact_block->name, temp_fbr->num_fields);
            SendMessage(MsgBuf,NO);
         ELSE
            sprintf(MsgBuf, "\n%.40s  (%d fields, with $?)",
                    temp_fbr->fact_block->name, abs(temp_fbr->num_fields));
            SendMessage(MsgBuf,NO);
         END_IF
         temp_fbr = temp_fbr->next_fbr;
      END_WHILE
   END_IF

   IF(node->rule_list NEQ NULL) THEN
      SendMessage("\n\nAppears on the LHS of the following rules:",NO);
      temp_rr = node->rule_list;
      while(temp_rr NEQ NULL) DO
         IF(temp_rr->num_fields > 0) THEN
            sprintf(MsgBuf, "\n   %.40s  (%d fields)",
                    temp_rr->rule->name, temp_rr->num_fields);
            SendMessage(MsgBuf,NO);
         ELSE
            sprintf(MsgBuf, "\n   %.40s  (%d fields, with $?)",
                    temp_rr->rule->name, abs(temp_rr->num_fields));
            SendMessage(MsgBuf,NO);
         END_IF
         temp_rr = temp_rr->next_rr;
      END_WHILE
   END_IF

   IF(node->assert_list NEQ NULL) THEN
      SendMessage("\n\nAsserted by the following rules:",NO);
      temp_rr = node->assert_list;
      while(temp_rr NEQ NULL) DO
         IF(temp_rr->num_fields > 0) THEN
            sprintf(MsgBuf, "\n   %.40s  (%d fields)",
                    temp_rr->rule->name, temp_rr->num_fields);
            SendMessage(MsgBuf,NO);
         ELSE
            sprintf(MsgBuf, "\n   %.40s  (%d fields, with $?)",
                    temp_rr->rule->name, abs(temp_rr->num_fields));
            SendMessage(MsgBuf,NO);
         END_IF
         temp_rr = temp_rr->next_rr;
      END_WHILE
   END_IF

   IF(node->retract_list NEQ NULL) THEN
      SendMessage("\n\nRetracted by the following rules:",NO);
      temp_rr = node->retract_list;
      while(temp_rr NEQ NULL) DO
         sprintf(MsgBuf, "\n   %.40s", temp_rr->rule->name);
         SendMessage(MsgBuf,NO);
         temp_rr = temp_rr->next_rr;
      END_WHILE
   END_IF

   temp_fld = node->field_list;
   IF(temp_fld) THEN
      SendMessage("\n\nThe following literal values were used:",NO);

      MULTI = 0;

      WHILE (temp_fld) DO 

         IF(template) THEN
           slot = FindDefslotByPosition(template,temp_fld->field_num);
           IF((slot->min_elements NEQ 1) OR (slot->max_elements NEQ 1)) THEN
             IF(!MULTI) THEN
               sprintf(MsgBuf,"\n   In multi-field <%.40s>",slot->name);
               SendMessage(MsgBuf,NO);
               MULTI = 1;
             END_IF
           ELSE
             sprintf(MsgBuf,"\n   In field <%.40s>",slot->name);
             SendMessage(MsgBuf,NO);
           END_IF
         ELSE
           sprintf(MsgBuf, "\n   In field %d:", temp_fld->field_num);
           SendMessage(MsgBuf,NO);
         END_IF

         temp_wl = temp_fld->lit_values;
         WHILE (temp_wl) DO
            sprintf(MsgBuf, "\n\t%.40s (%d times)", 
                    temp_wl->word, temp_wl->times_used);
            SendMessage(MsgBuf,NO);
            temp_wl = temp_wl->next_word;
         END_WHILE
         temp_fld = temp_fld->next_field;
      END_WHILE
   END_IF

   IF(ANALYZE_TRACE IS_ON AND VERBOSE IS_ON) THEN
      SendMessage("\n\nIn the dribble file, this relation was ",NO);
      temp_act = FindActivatedRelation(ActvatedRelationHead, node->name);
      IF(temp_act EQ NULL) THEN
         SendMessage("never been asserted or retracted",NO);
      ELSE
         sprintf(MsgBuf,"\n   Asserted  %5d times", temp_act->num_assert);
         SendMessage(MsgBuf,NO);
         sprintf(MsgBuf,"\n   Retracted %5d times", temp_act->num_retract);
         SendMessage(MsgBuf,NO);
         sprintf(MsgBuf,
              "\n   A max of  %5d facts with this relation existed at one time",
              temp_act->max_occurences);
         SendMessage(MsgBuf,NO);
      END_IF
   END_IF
}


/* ===================================================================
 *                 External Function Printout Functions
 * ===================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME   :  ExternalFuncsPrintout
 *
 *  PURPOSE:  This function prints summary information about
 *            all the external functions used in a KB.
 *
 *  INPUTS :  A pointer to a list of external function
 *            (EXF_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  
 * -------------------------------------------------------------
 */
 
void ExternalFuncsPrintout(list_head)
EXF_PTR list_head;
{
   SendMessage
      ("\n\n================ EXTERNAL FUNCTIONS SUMMARY ===============",NO);
   PrintExternalFunctions(list_head);
}

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintExternalFunctions
 *
 *  PURPOSE:  This is a recursive function which processes
 *            through all nodes in a binary tree of external
 *            function structures.
 *
 *  INPUTS :  A pointer to a list of functions (EXF_PTR *)
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Just processes binary tree nodes, actual
 *            printing done in PrintExtFuncInformation.
 * -------------------------------------------------------------
 */
 
static void PrintExternalFunctions(node)
EXF_PTR node;
{
   IF(node NEQ NULL) THEN
     PrintExternalFunctions(node->lft_ex_func);
     PrintExtFuncInformation(node);
     PrintExternalFunctions(node->rht_ex_func);
   END_IF
}

/*
 * -------------------------------------------------------------
 *  NAME   :  PrintExtFuncInformation
 *
 *  PURPOSE:  This is the function which actually prints
 *            external function summary information.
 *
 *  INPUTS :  A pointer to a function (EXF_PTR *)
 *
 *  RETURNS:  Nothing.
 * -------------------------------------------------------------
 */

static void PrintExtFuncInformation(node)
EXF_PTR node;
{
   RR_PTR   temp_rr;
 
   sprintf(MsgBuf, "\n\n----- FUNCTION: %.40s -----", node->name);
   SendMessage(MsgBuf,NO);

   IF(node->LHS_rule_list NEQ NULL) THEN
      SendMessage("\n\nAppears on the LHS of the following rules:",NO);
      temp_rr = node->LHS_rule_list;
      while(temp_rr NEQ NULL) DO
         IF(temp_rr->num_fields EQ 1) THEN
            sprintf(MsgBuf, "\n%.40s  (1 argument)",
                    temp_rr->rule->name);
         ELSE
            sprintf(MsgBuf, "\n%.40s  (%d arguments)",
                    temp_rr->rule->name, temp_rr->num_fields);
         END_IF
         SendMessage(MsgBuf,NO);
         temp_rr = temp_rr->next_rr;
      END_WHILE
   END_IF

   IF(node->RHS_rule_list NEQ NULL) THEN
      SendMessage("\n\nAppears on the RHS of the following rules:",NO);
      temp_rr = node->RHS_rule_list;
      while(temp_rr NEQ NULL) DO
         IF(temp_rr->num_fields EQ 1) THEN
            sprintf(MsgBuf, "\n%.40s  (1 argument)",
                    temp_rr->rule->name);
         ELSE
            sprintf(MsgBuf, "\n%.40s  (%d arguments)",
                    temp_rr->rule->name, temp_rr->num_fields);
         END_IF
         SendMessage(MsgBuf,NO);
        temp_rr = temp_rr->next_rr;
      END_WHILE
   END_IF
}
	   
/* ===================================================================
 *                 Defrelation Printout Functions
 *               (primarily for debugging purposes)
 * ===================================================================
 */

void DefrelationsPrintout(node)
DR_PTR node;
{
   IF(node NEQ NULL) THEN
      SendMessage
         ("\n\n=================== Defrelation SUMMARY ==================",NO);
      PrintDefrelations(node);
   END_IF
}

static void PrintDefrelations(node)
DR_PTR node;
{
   IF(node NEQ NULL) THEN
     PrintDefrelations(node->lft_def_rel);
     PrintDefrelationInformation(node);
     PrintDefrelations(node->rht_def_rel);
   END_IF
}

static void PrintDefrelationInformation(node)
DR_PTR node;
{
   sprintf(MsgBuf, "\n\nINFO FOR Defrelation: %.40s\nFrom file    : %.20s",
           node->name, node->file);
   SendMessage(MsgBuf,NO);
   sprintf(MsgBuf, "\nMax number fields: %d   Min number fields: %d", 
           node->max_fields, node->min_fields);
   SendMessage(MsgBuf,NO);
   PrintDeffields(node->fields);
}

static void PrintDeffields(list)
DF_PTR list;
{
   DF_PTR temp;
   WL_PTR temp_wl;
   IL_PTR temp_nl;
   FL_PTR temp_fl;
   
   temp = list;
   
   while(temp NEQ NULL) DO
      sprintf(MsgBuf, "\n\nFor Field %d words: %d strings: %d integers: %d  floats: %d",
              temp->position, temp->allow_word, 
              temp->allow_string, temp->allow_integer,temp->allow_float);
      SendMessage(MsgBuf,NO);

      SendMessage("\nAllowed strings: ",NO);
      temp_wl = temp->possible_strings;
      IF(temp_wl EQ NULL) THEN
         SendMessage("pointer is null",NO);
      ELSE
         while(temp_wl NEQ NULL) DO
            sprintf(MsgBuf, "%.40s ",temp_wl->word);
            SendMessage(MsgBuf,NO);
            temp_wl = temp_wl->next_word;
         END_WHILE
      END_IF

      SendMessage("\nAllowed words: ",NO);
      temp_wl = temp->possible_words;
      IF(temp_wl EQ NULL) THEN
         SendMessage("pointer is null",NO);
      ELSE
         while(temp_wl NEQ NULL) DO
            sprintf(MsgBuf, "%.40s ",temp_wl->word);
            SendMessage(MsgBuf,NO);
            temp_wl = temp_wl->next_word;
         END_WHILE
      END_IF

      SendMessage("\nAllowed integers: ",NO);
      temp_nl = temp->possible_integers;
      IF(temp_nl EQ NULL) THEN
         SendMessage("pointer is null",NO);
      ELSE
         while(temp_nl NEQ NULL) DO
            sprintf(MsgBuf, "%f ",temp_nl->number);
            SendMessage(MsgBuf,NO);
            temp_nl = temp_nl->next_number;
         END_WHILE
      END_IF

      SendMessage("\nAllowed floats: ",NO);
      temp_fl = temp->possible_floats;
      IF(temp_nl EQ NULL) THEN
         SendMessage("pointer is null",NO);
      ELSE
         while(temp_nl NEQ NULL) DO
            sprintf(MsgBuf, "%f ",temp_fl->number);
            SendMessage(MsgBuf,NO);
            temp_fl = temp_fl->next_number;
         END_WHILE
      END_IF
      if(temp->set_int_min OR temp->set_float_min)
        SendMessage("\nRANGE min set: TRUE",NO);
      else
        SendMessage("\nRANGE min set: FALSE",NO);
      if((float)temp->int_min_range > temp->float_min_range)
        sprintf(MsgBuf," min value %d ",temp->int_min_range);
      else
        sprintf(MsgBuf," min value %f ",temp->float_min_range);
      SendMessage(MsgBuf,NO);
      if(temp->set_int_max OR temp->set_float_max)
        SendMessage("range max set: TRUE",NO);
      else
        SendMessage("range max set: FALSE",NO);
      if((float)temp->int_max_range > temp->float_max_range)
        sprintf(MsgBuf," max value %d \n",temp->int_max_range);
      else
        sprintf(MsgBuf," max value %f \n",temp->float_max_range);
      SendMessage(MsgBuf,NO);  
      

      temp = temp->next_field;
   END_WHILE
}
      
/* ===================================================================
 *                 Processing Message Functions
 *
 *  All outputs from anywhere in CRSV are sent through these two
 *  functions. Any future outputs should also be sent through here!
 * ===================================================================
 */

/*
 * -------------------------------------------------------------
 *  NAME:          ErrorMessage
 *
 *  PURPOSE:       This function prints error messages that are
 *                 generated at various points during processing.
 *                 It adds information about where the error
 *                 occurred (file, object, approx. line #).
 *                 Currently, it merely prints to stdout using a
 *                 call to printf.
 *
 *  INPUTS:        Two arguments. The first is an integer which
 *                 represents the type, WARNING or ERROR. The
 *                 second is a pointer to the message.
 *
 *  RETURNS:       Nothing.
 * -------------------------------------------------------------
 */

void ErrorMessage(type, str)
int   type;
char *str;
{
   char *buf = type EQ ERROR ? "ERROR  " : "Warning";
   char  big_buf[WORDLENGTH];

   IF (type EQ ERROR) THEN
     NUM_ERRORS++;
   ELSE
     NUM_WARNINGS++;
   END_IF
   
   /* --------------------------------------------------
    *  Only output style warnings if CHECK_STYLE is 
    *  turned on (the s option was not specified)
    * ---------------------------------------------------
    */

   IF ((type EQ WARNING AND CHECK_STYLE IS_ON) OR
       (type EQ ERROR)) THEN

     IssuedErrorMessage = YES;


#if MAC_LSC
       sprintf(big_buf, "\n%s (%5d) in: %-.58s\n        %s", 
         buf, CurrentLineNumber, CurrentObjectName, str);
       display_print(big_buf);
#else /* !MAC_LSC */
#if IBM_TBC || IBM_MSC
   IF (!IBM_COM_LINE) THEN
         char error_str[25];

         sprintf(error_str,"\n%s (%5d) in: ", buf, CurrentLineNumber);
         sprintf(big_buf, "%-.55s\n       %s", CurrentObjectName, str);

         CrsvErrorMessage(error_str,big_buf);
   ELSE
     	 sprintf(big_buf, "\n%s (%5d) in: %-.58s\n        %s", 
        	 buf, CurrentLineNumber, CurrentObjectName, str);

      	(void)printf(big_buf);
   END_IF
#else /* !(IBM_TBC || IBM_MSC) */
   sprintf(big_buf, "\n%s (%5d) in: %-.58s\n        ",
      buf, CurrentLineNumber, CurrentObjectName);

   SendText (big_buf, 8);
   SendText (str, 8);
#endif /* IBM_TBC || IBM_MSC */
#endif /* MAC_LSC*/

   END_IF

   return;
}

/*
 * -------------------------------------------------------------
 *  NAME:         SendMessage
 *
 *  PURPOSE:      This function prints messages to the same
 *                place ErrorMessage does, but without the
 *                added info. Currently, it merely
 *                prints to stdout using a call to printf.
 *
 *  INPUTS:       A single argument, a string pointer to the 
 *                message to be printed.
 *
 *  RETURNS:      Nothing.
 *
 *  NOTES:        Does not add leading CR/LF.
 * -------------------------------------------------------------
 */

void SendMessage(str,is_warning)
char *str;
int is_warning;
{

if((is_warning) && (CHECK_STYLE IS_OFF))
  return;
#if MAC_LSC
   display_print(str);
#else /*  not MAC_LSC */
#if IBM_TBC || IBM_MSC
   IF (!IBM_COM_LINE) THEN
      CrsvMessage(str);
   ELSE
      (void)printf("%s", str);
   END_IF
#else /*!( IBM_TBC || IBM_MSC) */
   SendText (str, 0);
#endif/*  IBM_TBC || IBM_MSC */
#endif /* MAC_LSC */

   return;
}

/*
 * -------------------------------------------------------------
 *  NAME:         SendText
 *
 *  PURPOSE:      This function prints word-wrapped text to the
 *                output device.
 *
 *  INPUTS:       The new text to print and the line pad when
 *                a wrap occurs.
 *
 *  RETURNS:      Nothing.
 *
 *  NOTES:        Does not add leading CR/LF.
 *                The line pad is ignored when explicit new
 *                line characters are processed.
 * -------------------------------------------------------------
 */

void SendText(str, pad)
char *str;
int pad;
{
   static int space_left = LINELENGTH;
   char *next_line = strchr (str, '\n');

   IF (next_line NEQ NULL) THEN
      IF (next_line NEQ str) THEN
         *next_line = '\0';
         SendText (str, pad);
         *next_line = '\n';
      END_IF

      (void)printf ("\n");
      space_left = LINELENGTH;

      IF (*(next_line + 1) NEQ '\0') THEN
         SendText (next_line + 1, pad);
      END_IF
   ELSE
      int len = strlen (str), p;

      IF (len > space_left) THEN
         char *continuation_string = &str[space_left];
   
         WHILE (*continuation_string != ' ' AND continuation_string > str) DO
            continuation_string --;
         DONE
   
         IF (continuation_string == str AND (continuation_string = strchr (&str[space_left], ' ')) EQ NULL) THEN
            (void)printf ("%s\n", str);
            space_left = LINELENGTH;
         ELSE
            *continuation_string = '\0';
            (void)printf ("%s\n", str);
            *continuation_string = ' ';
   
            FOR (p = 0; p < pad; p++) DO
              (void)printf (" ");
              SendMessage(" ",NO);
            DONE

            space_left = LINELENGTH - pad;

            SendText (continuation_string + 1, pad);
         END_IF
      ELSE
        (void)printf ("%s", str);
         space_left -= len;
      END_IF
   END_IF
   
   return;
}

