 /* ========================================================= *
 *  This file contains all the fuctions that are used to     *
 *  process the trace file.                                  *
 * ========================================================= */

#include <stdio.h>
#include <ctype.h>
#include "crsv.h"
#include <string.h>


#define ASSERTION           1
#define RETRACTION          2

/* Extern functions  */

 extern double atof();
 extern int	   atoi();
 extern long   atol();

 extern ACT_R_PTR      	AllocActivatedRule();       /* From csrvmem.c */
 extern ACT_REL_PTR    	AllocActivatedRelation();
 extern SPE_FACT_PTR   	AllocFactList();
 extern FIELD_PTR      	AllocFieldValues();
 extern RULE_NAME_PTR  	AllocRuleList();
 extern AGEN_RULE_PTR  	AllocAgendaRule();
 extern PAT_LIST_PTR   	AllocPatternList();
 extern MY_TMPLATE_PTR 	AllocMyTemplate();
 extern void           	FreeAgendaRule();
 extern void           	EvaluationRelation();       /* from crsvtr.c */
 extern void		  	EvaluateTemplate(); 
 extern HASH          	*AddSymbol();         /* from crsvhash.c */
 extern void          	SendMessage();        /* From crsvprnt.c*/
 extern void          	ErrorMessage();
 extern RULE_PTR      	SearchRule();
 extern PAT_TKN_PTR   	InsertTokenIntoPattern();
 
 extern int 			GetTokenWithEndOfLine();          /* From crsvtkn.c */


/* global variables which have been defined from another file */

 extern ACT_R_PTR    ActivatedRuleHead;
 extern RULE_PTR     RuleHead;
 extern RULE_PTR     CurrentRule;
 extern ACT_REL_PTR  ActvatedRelationHead;
 extern DR_PTR       DefrelationHead;
 extern DT_PTR		 DeftemplateHead;
 extern FB_PTR       FactBlockHead;
 extern ACT_REL_PTR  LastActivatedRelation; /* last active relation            */
 
 extern short        LastOperation;         /* last operation code , RETRACTION*/
                                            /* or ASSERTION                    */
 extern short        FirstRelation;         /* Is it first relation in the     */
                                            /* trace file ?                    */
 extern short        FirstRule;             /* Is it first rule in the trace   */
                                            /* file ?               */
 extern PAT_TKN_PTR  CurrentTokenInPattern;
 extern PAT_LIST_PTR CurrentPattern;        /* Pattern in process */
 extern ACT_REL_PTR  CurrentActivatedRelation;  /* current relation returned by             */
                                             /* FindCreateActRelation                      */
 extern SPE_FACT_PTR CurrentFact;            /* specific information about the current   */
                                             /* relation                                 */
 extern AGEN_RULE   *AgendaRules;            /* Rules in the agenda                      */
 extern int          FactNumber;             /* the number corresponds to the            */
                                             /* current fact in the trace  file          */
 extern int          cycle;
 extern int          FactNumberInDeffact;
 extern FB_PTR       LastFactBlock;
 extern short        start;
 extern ACT_R_PTR    LastFiredRule;
 extern char         *TKNWORD;
 extern long         INTNUMBER;
 extern FILE         *CurrentFile;

 /* ---------   Local variables  ---------- */

 

 HASH    *save_word;             /* A pointer to a hash word for saving      */
                                 /* words in the symbol table                */
 FILE    *logfile = NULL;
 
 /* --------- functions for global uses---------------*/

 
 int                ProcessRelationActivation();
 int                ProcessRuleActivation();
 int                ProcessRelationDactivation();
 int                ProcessRuleDactivation();
 int                ProcessRuleFire();
 void 		        PrintRuleInAgenda();
 void				ProcessRelationInTrace();
 void		        ProcessTemplateInTrace();
 void		        QueueSlotInTemplate();
 void				ProcessRelForInfTrace();
 void   		    SendMessageToFile();
 int                OpenLogFile();
 void 		        CloseLogFile();
 void               PrintRelToFile();
 void				FindEndOfLine();
 
 /* --------- functions for local uses---------------*/
 
 static ACT_REL_PTR        FindCreateActRelation();
 static void               SearchActivatedFact();
 static ACT_R_PTR          InserRule();
 static int                SearchFact();
 static int                StoreRuleName();
 static PAT_LIST_PTR       SearchAndPattern();
 static PAT_LIST_PTR	   SearchOrPattern();
/*
  * -------------------------------------------------------------
  *  FindEndOfLine()
  *		This function will read a file until end of the line.
  * -------------------------------------------------------------
 */

void FindEndOfLine(CurrentFile)
FILE *CurrentFile;
{
  int token;
  if(TKNWORD[0] == '\n')
    return;
  while((token = GetTokenWithEndOfLine(CurrentFile)) != EOL) DO
    if(token == STOP)
      return;
  DONE
  return;
}

/*
 * -------------------------------------------------------------
 *  NAME   :  FindCreateActRelation()
 *
 *  PURPOSE:  This function searches through a list of
 *            relations to find a specific name. If it doesn't
 *            find the name, it creates a new relation structure
 *            and adds it to the list.
 *
 *  INPUTS :  A pointer to the list of relations (ACT_REL_PTR)
 *            and the name of the node being searched (char *)
 *
 *  RETURNS:  A pointer to the relation structure which has a same
 *            name as the passed argument.
 *
 *  NOTES  :
 * ------------------------------------------------------------------
 */

static ACT_REL_PTR FindCreateActRelation(node, name)
ACT_REL_PTR  node;
char    *name;
{
   int rtn;

   IF(node EQ NULL) THEN        /* A new relation! */
      node        = AllocActivatedRelation();
      save_word   = AddSymbol(name);
      node->name  = save_word->contents;
      CurrentActivatedRelation = node;
   ELSE
      rtn = strcmp(name, node->name);
      IF(rtn < 0) THEN
         node->left = FindCreateActRelation(node->left, name);
      ELSE_IF(rtn > 0)
         node->right = FindCreateActRelation(node->right, name);
      ELSE
         CurrentActivatedRelation = node;
      END_IF
   END_IF
   return(node);
}
/*
 * -----------------------------------------------------------------------
 *   Fuction name : ProcessRelationActivation
 *         This function will put a relation's info. into an act_relation
 *         structure and link it to the global link list (ActvatedRelationHead).
 *   Argument(s) :
 *      str : char *str;
 *   Return : Nothing
 * -----------------------------------------------------------------------
*/
 ProcessRelationActivation()
  {
    SPE_FACT_PTR fact_ptr;
    FIELD_PTR  field_ptr,temp_fld_ptr;
    int num,rtn,token;
    char *relation_name;

    /* ================================== *
     *  Form to be scanned:               *
     *   0        (initial-fact)          *
     * ================================== */

    token = GetTokenWithEndOfLine(CurrentFile);
    IF (token != INTEGER)THEN
        FindEndOfLine(CurrentFile);
        return(ERROR);
    END_IF
    num = atoi(TKNWORD);       /* get the fact number */
    IF(FirstRelation)THEN
          IF(num > 1)THEN
               ErrorMessage(WARNING, "Incomplete trace file!\n        ");
               SendMessage(
                  "The first asserted relation in the trace file is\n    ",YES);
               SendMessage(
                  "    not fact 0. Trace file analysis may be incomplete",YES);
          END_IF
          FirstRelation = FALSE;
    END_IF

    /* -------------------------------------------------- *
     *  Allocate space for new relation and initialize it *
     * -------------------------------------------------- */

    /* scanning the relation name  */

    token = GetTokenWithEndOfLine(CurrentFile);
    IF(token != LPAREN)THEN
        FindEndOfLine(CurrentFile);
        return(ERROR);
    END_IF
    token = GetTokenWithEndOfLine(CurrentFile);
    IF((token != WORD) AND (token != STRING) 
       AND (token != INTEGER) AND (token != FLOAT) AND
       (token != INSTANCE))THEN
          FindEndOfLine(CurrentFile);
          return(ERROR);
    END_IF
    relation_name = TKNWORD;
    IF(VERBOSE IS_ON)THEN
         sprintf(MsgBuf,"\n\nProcessing activated fact f-%d <%.40s>",num,
                 relation_name);
         SendMessage(MsgBuf,NO);
    END_IF
    /* ---------------------------------------------------- *
     *  Search through the relation list to see if relation *
     *  has existed.If it has,increase the assertion counter*
     *  by 1,else allocate new node and link it to the list.*
     *  All of these will be done by "FindCreateActRelation", *
     *  and use the relation name as the key.               *
     * -----------------------------------------------------*/

    ActvatedRelationHead = FindCreateActRelation(ActvatedRelationHead,relation_name);
    (CurrentActivatedRelation->current_active)++;
    if(CurrentActivatedRelation->max_occurences < CurrentActivatedRelation->current_active)
            CurrentActivatedRelation->max_occurences = CurrentActivatedRelation->current_active;
    (CurrentActivatedRelation->num_assert)++;

    /* ------------------------------------------------------------- *
     * Get new maximum occurences of a relation at one time          *
     * if neccesary.                                                 *
     * From here down <fact> will be mentioned instead of <relation> *
     * since I assume fact is an instance of a relation.             *
     * ------------------------------------------------------------- */

    /* -------------------------------- *
     *  Store the fields of a fact.     *
     * -------------------------------- */

    fact_ptr = AllocFactList();
    fact_ptr->number_field = 1;
    fact_ptr->fact_num = num;     		  /* num = 0 for <initial-fact> */
    /* scanning for fact's fields */
    
    token = GetTokenWithEndOfLine(CurrentFile);
    IF (token == LPAREN) THEN
      ProcessTemplateInTrace(fact_ptr);
    ELSE
	  ProcessRelationInTrace(fact_ptr,token);
	END_IF
    /* --------------------------------------------
     *  Insert the fact into the global list.
     * --------------------------------------------
     */
     fact_ptr->next = CurrentActivatedRelation->spec_fact_list;
     CurrentActivatedRelation->spec_fact_list = fact_ptr;
     LastOperation = ASSERTION;
     FactNumber = num;
     IF(INFERENCE_TRACE) 
       ProcessRelForInfTrace(fact_ptr);
     IF(CHECK_DEFRELS IS_ON)
       ProcessRelWithDefrel(fact_ptr,relation_name);
    FindEndOfLine(CurrentFile);
    return(OK);

   }

/* ---------------------------------------------------------
 *  ProcessRelWithDefrel()
 *   This function will call the appropriate functions to 
 *   cross checking the relations found in the trace file with
 *   their definitions.
 *
 *  Arguments:
 *    - fact_ptr : 
 *		poiter point to the structure which contains
 *      all the information of the fields in the fact.
 *    - relation_name :
 *	    name of the rellation.
 *
 *  return: nothing.
 * ---------------------------------------------------------
 */

ProcessRelWithDefrel(fact_ptr,relation_name)
SPE_FACT_PTR fact_ptr;
char *relation_name;
{
    
  IF(fact_ptr->field != NULL) THEN              /* Relation is in regular form */
     IF(DefrelationHead == NULL)THEN
       ErrorMessage(WARNING, "Relations have not been defined,\n   ");
       SendMessage("      relations in the trace file can't be verified.",YES);
     ELSE
       EvaluationRelation(fact_ptr,relation_name);
     END_IF
  ELSE_IF (fact_ptr->templ_slot != NULL)    /*Relation is in deftemplate form */
     IF(DeftemplateHead == NULL) THEN
       ErrorMessage(WARNING, "Deftemplates were not provided,\n   ");
       SendMessage("      templates in the trace file can't be verified.",YES);
     ELSE
       EvaluateTemplate(fact_ptr,relation_name);
     END_IF   
  END_IF
 }
/* ---------------------------------------------------------
 *  ProcessRelationInTrace()
 *  Process the regular relation it found in the trace file.
 *
 *  Arguments:
 *    - fact_ptr:
 *		poiter point to the structure which contains
 *      all the information of the fields in the fact.
 *	  - token :
 *      the value of the current token in the file.
 *
 *  Return: Nothing.
 * ---------------------------------------------------------
 */
 
void ProcessRelationInTrace(fact_ptr,token)
SPE_FACT_PTR fact_ptr;
int token;
{

   FIELD_PTR field_ptr = NULL,temp_fld_ptr = NULL;
   
     while((token != RPAREN) && (token != EOL) && (token != STOP))DO
         
         (fact_ptr->number_field)++;      /* Increase number of fields by one */

         /* ------------------------------------------------ *
          *  Allocate the space to store a field of the      *
          *  fact and initialize it.                         *
          * ------------------------------------------------ */

         field_ptr       = AllocFieldValues();
         field_ptr->num  = fact_ptr->number_field;
         save_word       = AddSymbol(TKNWORD);
         field_ptr->word = save_word->contents;
         field_ptr->type = token;

         /* ------------------------------------------------ *
          *  Insert the field to the fact                    *
          * ------------------------------------------------ */

         IF(fact_ptr->field == NULL)THEN
           fact_ptr->field = field_ptr;
         ELSE_IF (fact_ptr->field->next == NULL)
           fact_ptr->field->next = field_ptr;
         ELSE
              temp_fld_ptr = fact_ptr->field;
              while (temp_fld_ptr->next != NULL)
                 temp_fld_ptr  = temp_fld_ptr->next;
              temp_fld_ptr->next = field_ptr;

         END_IF
         token = GetTokenWithEndOfLine(CurrentFile);
      END_WHILE
}

/*
 * ============================================================
 *   ProcessTemplateInTrace()
 *   Process the template it found in the trace file
 *
 *   Argument:
 *   fact_ptr:
 *		poiter point to the structure which contains
 *      all the information of the fields in the fact.
 * ============================================================
 */
 
 void ProcessTemplateInTrace(fact_ptr)
 SPE_FACT_PTR fact_ptr;
 {
    int token, count = 1;
    MY_TMPLATE_PTR temp;
    HASH *temp_hash;
    FIELD_PTR afield,last_field;
    
    while(count)
     {
        token = GetTokenWithEndOfLine(CurrentFile);
        IF(token == LPAREN) THEN /* Process a slot in the template */
            temp = AllocMyTemplate();
            count++;
            token = GetTokenWithEndOfLine(CurrentFile);
            temp_hash = AddSymbol(TKNWORD);
            temp->name = temp_hash->contents;
            last_field = NULL;
            WHILE(((token = GetTokenWithEndOfLine(CurrentFile)) != RPAREN) 
                  && (token != STOP))DO
            	temp_hash = AddSymbol(TKNWORD);
            	afield = AllocFieldValues();
            	afield->word = temp_hash->contents;
            	afield->type = token;
            	IF(last_field == NULL)THEN
            	    temp->value_list = afield; 
            	    last_field = afield;
            	ELSE
            	     last_field->next = afield;
            	     last_field = afield;
            	END_IF
            END_WHILE
            QueueSlotInTemplate(fact_ptr,temp);
        ELSE_IF(token == RPAREN)
            count--;
        ELSE
         break;
        END_IF   
     }
 
 }
/*
 * ----------------------------------------------------------------------
 *  QueueSlotInTemplate()
 *   queue a slot in the template by it's position
 * ----------------------------------------------------------------------
 */ 
 void QueueSlotInTemplate(fact_ptr,temp_slot)
 SPE_FACT_PTR fact_ptr;
 MY_TMPLATE_PTR temp_slot;
 {
    MY_TMPLATE_PTR  temp = NULL;
    
    IF(fact_ptr->templ_slot == NULL) THEN
      fact_ptr->templ_slot = temp_slot;
    ELSE
         temp = fact_ptr->templ_slot;
         while(temp->next != NULL)
           temp = temp->next;
         temp->next = temp_slot;
    END_IF
 }
 
 /*
 * ----------------------------------------------------------------------
 *  ProcessRelForInfTrace()
 *   This is for the inference trace option (just for Joe)
 * ----------------------------------------------------------------------
 */ 
 void ProcessRelForInfTrace(fact_ptr)
 SPE_FACT_PTR fact_ptr;
 {
   FB_PTR cur_fb  = NULL;
   int rtn;
   
    IF(start)THEN
         start = FALSE;
         LastFactBlock = FactBlockHead;
         if(LastFactBlock != NULL)
         while(LastFactBlock->next_block != NULL)
            LastFactBlock = LastFactBlock->next_block;
    END_IF

	/* 
	 * =======================================================
	 * to see if the fact is asserted by deffact and record 
	 * the deffact's name.
	 * =======================================================
	 */
    IF((cycle == 1) && 
       	  ((rtn = strcmp(CurrentActivatedRelation->name,"initial-fact")) != 0)&&
       	  (LastFactBlock != NULL))THEN
       IF(FactNumberInDeffact > LastFactBlock->num_facts)THEN
               cur_fb = FactBlockHead;
               while(cur_fb->next_block != LastFactBlock)
                 cur_fb = cur_fb->next_block;
               LastFactBlock = cur_fb;
               FactNumberInDeffact = 0;
       END_IF
	   fact_ptr->deffact_name = LastFactBlock->name;
       FactNumberInDeffact++;
     ELSE_IF (cycle != 1)
           fact_ptr->asserted_by_rule = LastFiredRule->name;
     END_IF
 }
 
  /*
  * ----------------------------------------------------------------------
  *   Name: process_rule_activate();
  *         This function records all informations about the activated
  *         rule.
  * ----------------------------------------------------------------------
  */

 int ProcessRuleActivation()
 {
    int token ,i = 0,num;
    char *name;
    AGEN_RULE_PTR temp_rule;

    /* =============================================================== *
     *  Form to be scanned :                                           *
     *      <salience> <rule-name>: <num>, ,<num>                      *
     * =============================================================== */


    /* ------------------------------------------------ *
     *   Get salience value of the rule                 *
     *   This value is stored but have not been used yet*
     * ------------------------------------------------ */

    token = GetTokenWithEndOfLine(CurrentFile);
    IF(token != INTEGER)THEN
        FindEndOfLine(CurrentFile);
        return(ERROR);
     END_IF
     num = (int)INTNUMBER;
    /* ------------------------------------------------- *
     *    Get the name of the rule.                      *
     * ------------------------------------------------- */
    token = GetTokenWithEndOfLine(CurrentFile);
    IF((token != WORD) AND (token != STRING ) AND 
       (token != INTEGER)AND (token != FLOAT)AND
       (token != INSTANCE))THEN
        FindEndOfLine(CurrentFile);
        return(ERROR);
    END_IF
    i = strlen(TKNWORD);
    IF(TKNWORD[i-1] == ':')
      TKNWORD[i-1] = EOS;
    IF(VERBOSE IS_ON)THEN
        sprintf(MsgBuf,"\n\nProcessing activated rule <%.40s>",TKNWORD);
        SendMessage(MsgBuf,NO);
    END_IF
   
    name = TKNWORD;


    /* ------------------------------------------------------ *
     *    Allocate space for new rule,initialize it,and       *
     *    link it to the rule list.                           *
     * ------------------------------------------------------ */
    IF(INFERENCE_TRACE IS_ON)THEN
       temp_rule = AllocAgendaRule();
       temp_rule->rule = InserRule(name,num);
       temp_rule->next_rule = AgendaRules;
       AgendaRules = temp_rule;
    ELSE
       (void)InserRule(name,num);
    END_IF
      /* -------------------------------------------------- *
       * Looking for the facts that activate this rule.     *
       * -------------------------------------------------- */
    SearchActivatedFact(name);
    FindEndOfLine(CurrentFile);
    return(OK);
 }



 /* --------------------------------------------------- *
  *  InserRule()                                      *
  *    Allocate space for new rule,initialize it,and    *
  *    link it to the rule list.                        *
  * --------------------------------------------------- */

 static ACT_R_PTR InserRule(name,num)
 char *name;
 int num;
  {
    ACT_R_PTR  rule_ptr,temp_rule_ptr;

    
    IF(ActivatedRuleHead == NULL) THEN  /* The list is empty  */
         ActivatedRuleHead           = AllocActivatedRule();
         save_word                   = AddSymbol(name);
         ActivatedRuleHead->name     = save_word->contents;
         ActivatedRuleHead->salience = num;
         ActivatedRuleHead->activate_num = 1;
         return(ActivatedRuleHead);
    ELSE
         temp_rule_ptr = ActivatedRuleHead;
         rule_ptr = ActivatedRuleHead;
         WHILE(temp_rule_ptr != NULL)DO
             if(strcmp(temp_rule_ptr->name,name) == 0)
               break;
             rule_ptr = temp_rule_ptr;
             temp_rule_ptr = temp_rule_ptr->next;
         END_WHILE

         IF(temp_rule_ptr != NULL) THEN  /* rule has been inserted into the list */
              (temp_rule_ptr->activate_num)++;
         ELSE       /* new rule */
	         temp_rule_ptr           = AllocActivatedRule();
	         save_word               = AddSymbol(name);
	         temp_rule_ptr->name     = save_word->contents;
	         /* salience is currently not used */
	         temp_rule_ptr->salience = num;
	         temp_rule_ptr->activate_num = 1;
	         rule_ptr->next = temp_rule_ptr;
	     END_IF
    END_IF
    return(temp_rule_ptr);
  }
  
 /* -------------------------------------------------- *
  * SearchActivatedFact()                                 *
  *  Looking for the facts that activate this rule.    *
  * -------------------------------------------------- */
static void SearchActivatedFact(name)
char *name;
{
     short flag;
     int token, i, num;

     /* the rule was activated because of the retraction of a certain fact,*/
     /* and there is only a (not) pattern on the left-hand side of the rule */

   token = GetTokenWithEndOfLine(CurrentFile);
   IF ((token == EOL)&&(LastOperation == RETRACTION)) THEN
          flag = FALSE;
          (void)SearchFact(ActvatedRelationHead,FactNumber,&flag);
      IF (flag == TRUE) THEN
         (CurrentFact->num_activated_d)++;
         IF ((ANALYZE_TRACE IS_ON) AND (VERBOSE IS_ON)) THEN
	          StoreRuleName(name,FactNumber);
         END_IF
      END_IF
   END_IF

     i = 0;
   WHILE((token  != EOL) && (token != STOP))DO
   									                                
      IF(token == INTEGER) THEN

        num = atoi(TKNWORD);
	 /* ------------------------------------------------- *
	  *  Increases the relations' activate counter by 1.  *
	  * ------------------------------------------------- */
	
	    flag = FALSE;
	    CurrentFact = NULL;
        (void)SearchFact(ActvatedRelationHead,num,&flag);
        IF (flag == TRUE) THEN
            (CurrentActivatedRelation->activate_num)++;
            IF (LastOperation == ASSERTION) THEN
               (CurrentFact->num_activated_a)++;
            ELSE_IF (LastOperation == RETRACTION)
               (CurrentFact->num_activated_d)++;
            END_IF
		
		 /* ---------------------------------------------------- *
		  *  Be more specific about the facts that activate the  *
		  *  rule if the user requested.                         *
		  * ---------------------------------------------------- */
		
             IF ((ANALYZE_TRACE IS_ON) AND (VERBOSE IS_ON)) THEN
		              StoreRuleName(name,num);
             END_IF
         END_IF /* flag == TRUE */
      END_IF /*token == INTEGER */
      token = GetTokenWithEndOfLine(CurrentFile);
   DONE
}


/* ---------------------------------------------- *
 * SearchFact()                                   *
 *   This function will look for the fact that    *
 *   has same number  as <num> from the global    *
 *   list.                                        *
 * Argumments                                     *
 *   relation_ptr : head of the list              *
 *   num          : fact's number                 *
 *   flag         : TRUE if found                 *
 * Returns:  ERROR or OK                          *
 * ---------------------------------------------- */

static int SearchFact(relation_ptr,num,flag)
ACT_REL_PTR relation_ptr;
int num;
short *flag;

{
   SPE_FACT_PTR temp_ptr;

   if(relation_ptr == NULL)
      return(ERROR);
   temp_ptr = relation_ptr->spec_fact_list;
   WHILE((temp_ptr != NULL)&&(!(*flag)))DO
     IF(temp_ptr->fact_num == num)THEN
        *flag = TRUE;
         CurrentActivatedRelation = relation_ptr;   /* global current active relation */
         CurrentFact = temp_ptr;          /* global current active fact */
     END_IF
     temp_ptr = temp_ptr->next;
   END_WHILE
   if((*flag) == FALSE)
      (void)SearchFact(relation_ptr->left,num,flag);
   if((*flag) == FALSE)
      (void)SearchFact(relation_ptr->right,num,flag);
   return (OK);
}


 /* ----------------------------------------------------------- *
  * This function adds rules to the activated rule list in the  *
  * spec_fact_list field of a relation.                         *
  * ----------------------------------------------------------- */

 static int StoreRuleName(name,num)
  char *name;
  int num;
  {
     RULE_NAME_PTR rule_name_ptr,a_rule_name;

     IF(LastOperation == ASSERTION)THEN
         rule_name_ptr = CurrentFact->activate_list_a;
         WHILE(rule_name_ptr != NULL) DO
             if(strcmp(name,rule_name_ptr->name) == 0)
                break;
             rule_name_ptr = rule_name_ptr->next;
         END_WHILE
         IF(rule_name_ptr == NULL)THEN
             a_rule_name       = AllocRuleList();
             save_word         = AddSymbol(name);
             a_rule_name->name = save_word->contents;
             if(FactNumber == num)
               a_rule_name->last_fact = YES;
             a_rule_name->counter = 1;
             a_rule_name->next = CurrentFact->activate_list_a;
             CurrentFact->activate_list_a = a_rule_name;
          ELSE
             (rule_name_ptr->counter)++;
             if(FactNumber == num)
               rule_name_ptr->last_fact = YES;
          END_IF
     ELSE_IF (LastOperation == RETRACTION)
         rule_name_ptr = CurrentFact->activate_list_d;
         WHILE(rule_name_ptr != NULL)DO
             if(strcmp(name,rule_name_ptr->name) == 0)
               break;
             rule_name_ptr = rule_name_ptr->next;
         END_WHILE
         IF(rule_name_ptr == NULL)THEN
             a_rule_name          = AllocRuleList();
             save_word            = AddSymbol(name);
             a_rule_name->name    = save_word->contents;
             a_rule_name->counter = 1;
             if(FactNumber == num)
               a_rule_name->last_fact = YES;
             a_rule_name->next = CurrentFact->activate_list_d;
             CurrentFact->activate_list_d = a_rule_name;
         ELSE
             (rule_name_ptr->counter)++;
             if(FactNumber == num)
               rule_name_ptr->last_fact = YES;
         END_IF
      END_IF
  }

 /*
  * ----------------------------------------------------------------------
  *   Name : ProcessRelationDactivation();
  *      This function is used for setting the flag of the relation which
  *      has been retract from the agenda.
  *
  *   Arguments : char *str;
  *               a linefrom the trace file that indicates the
  *               retraction is taken place.
  *   Return: nothing .
  * ----------------------------------------------------------------------
  */
int ProcessRelationDactivation() 
{

      int token,num;
      short flag;

      /* ==================================================== *
       *  Form to be scanned :                                *
       *     0     (initial-fact)                             *
       * ==================================================== */
       
       token   = GetTokenWithEndOfLine(CurrentFile);
      IF(token != INTEGER)THEN
           FindEndOfLine(CurrentFile);
           return(ERROR);
      END_IF
      num =  atoi(TKNWORD);   /* Get the fact number */

      IF (VERBOSE IS_ON) THEN
          sprintf(MsgBuf,"\n\nProcessing retracted fact f-%d",num);
          SendMessage(MsgBuf,NO);
      END_IF

      flag = FALSE;

      /* Look for the fact that being retracted,*/
      
      (void)SearchFact(ActvatedRelationHead,num,&flag);
      IF (flag == TRUE) THEN
          /* and store its address in cur_ac_rel    */
        (CurrentActivatedRelation->num_retract)++;
         CurrentActivatedRelation->current_active -= 1;
    	 LastOperation = RETRACTION;
	     FactNumber = num;
      END_IF
      FindEndOfLine(CurrentFile);
      return(OK);
}


 /*
  * ----------------------------------------------------------------------
  *   name : ProcessRuleDactivation
  *       This function will get the name of the rule has been deactivated
  *       and use it to search for the rule in the list.Once it is found
  *       the deactivate_num of the rule will be increased by one.
  *   arguments:
  *      char *str
  *        a line in the trace file that indicates a rule has been
  *        deactivated.
  *   return: nothing.
  * ----------------------------------------------------------------------
  */

 int ProcessRuleDactivation()

 {
   ACT_R_PTR        rule_ptr;
   RULE_NAME_PTR    name_ptr;
   int              token,i = 0,j;
   short            flag;

	token = GetTokenWithEndOfLine(CurrentFile);
	IF(token != INTEGER)THEN
	     FindEndOfLine(CurrentFile);
	     return(ERROR);
	END_IF

   /* get the name of the rule that been deactivated*/
   
    token = GetTokenWithEndOfLine(CurrentFile);
    i = strlen(TKNWORD);
    if(TKNWORD[i-1] == ':')
     TKNWORD[i-1] = EOS;
    IF(VERBOSE IS_ON)THEN
       sprintf(MsgBuf,"\n\nProcessing deactivated rule <%.40s>",TKNWORD);
       SendMessage(MsgBuf,NO);
    END_IF
    IF(INFERENCE_TRACE IS_ON) THEN
	  ProcessRuleDeactForInfTrace();
    END_IF
   /*----- Search for the rule -------*/

   rule_ptr = ActivatedRuleHead;
   WHILE(rule_ptr != NULL)DO
     if(strcmp(TKNWORD,rule_ptr->name) ==  0)
       break;
     rule_ptr = rule_ptr->next;
   END_WHILE
   if(rule_ptr != NULL)
      (rule_ptr->deactivate_num)++;

   /* ================================================= *
    *  If the last operation is RETRACTION, increase    *
    *  the number of rules being deactivated by the     *
    *  retraction of the relation by 1                  *
    * ================================================= */

   if(LastOperation == RETRACTION)
       ProcessRuleDeactDueToRetract();


   /* ================================================= *
    *  If the last operation is ASSERTION, increase     *
    *  the number of rules being deactivated by the     *
    *  assertion of the relation by 1                   *
    * ================================================= */

   else if (LastOperation == ASSERTION)
       ProcessRuleDeactDueToAssert(); 
   FindEndOfLine(CurrentFile);
   return(OK);
 }
/*
 *================================================= *
 * ProcessRuleDeactForInfTrace()
 *  This for the inference trace option (Joe's idea)
 *  Currently, it is always Off.
 *================================================= *
 */  
 
 ProcessRuleDeactForInfTrace()
 {
    AGEN_RULE_PTR    ag_rule = NULL,temp_ag_rule = NULL;
    
       ag_rule = AgendaRules;
       WHILE(ag_rule != NULL)DO
	      if(strcmp(ag_rule->rule->name,TKNWORD) == 0)
	        break;
	      temp_ag_rule = ag_rule;
	      ag_rule = ag_rule->next_rule;
	   END_WHILE
       IF((ag_rule != NULL)&&(temp_ag_rule != NULL))THEN
	      temp_ag_rule->next_rule = ag_rule->next_rule;
	      ag_rule->next_rule = NULL;
	      FreeAgendaRule(ag_rule);
	   ELSE_IF(ag_rule == NULL)
          SendMessageToFile(logfile,"Incomplete Trace File \n");
       ELSE_IF(temp_ag_rule == NULL)    /* Rule match at the begining of the list */
          AgendaRules = ag_rule->next_rule;
          ag_rule->next_rule = NULL;
          FreeAgendaRule(ag_rule);
       END_IF
 }

/*
 *================================================= *
 * ProcessRuleDeactDueToRetract()
 *================================================= *
 */   
 ProcessRuleDeactDueToRetract()
 {
   short flag;
   RULE_NAME_PTR    name_ptr;
   
       flag = FALSE;
       (void)SearchFact(ActvatedRelationHead,FactNumber,&flag);
       IF(flag == TRUE)THEN
           (CurrentActivatedRelation->deactivate_num)++;
           name_ptr = CurrentFact->deactivate_list_d;
           (CurrentFact->num_deactivated_d)++;

       /* ====================================== *
        * Store the names of the rules that are  *
        * deactivated by the retraction of this  *
        * relation.                              *
        * ====================================== */

         IF(ANALYZE_TRACE IS_ON AND VERBOSE IS_ON)THEN
            WHILE(name_ptr != NULL)DO
                 if(strcmp(name_ptr->name,TKNWORD) == 0)
                    break;
                 name_ptr = name_ptr->next;
            END_WHILE
            IF(name_ptr != NULL)THEN
                 (name_ptr->counter)++;
            ELSE
                 name_ptr          = AllocRuleList();
                 save_word         = AddSymbol(TKNWORD);
                 name_ptr->name    = save_word->contents;
                 name_ptr->counter = 1;
                 name_ptr->next    = CurrentFact->deactivate_list_d;
                 CurrentFact->deactivate_list_d = name_ptr;
            END_IF
         END_IF
      END_IF
 }
/*
 *================================================= *
 * ProcessRuleDeactDueToAssert()
 *================================================= *
 */   
 
 ProcessRuleDeactDueToAssert()
 {
     short 				flag = FALSE;
     RULE_NAME_PTR    	name_ptr;
     
     (void)SearchFact(ActvatedRelationHead,FactNumber,&flag);
     IF(flag == TRUE)THEN
          (CurrentActivatedRelation->deactivate_num)++;
          name_ptr = CurrentFact->deactivate_list_a;
          (CurrentFact->num_deactivated_a)++;

       /* ====================================== *
        * Store the names of the rules that are  *
        * deactivated by the assertion of this   *
        * relation.                              *
        * ====================================== */

         IF(ANALYZE_TRACE IS_ON AND VERBOSE IS_ON)THEN
            WHILE(name_ptr != NULL)DO
                 if(strcmp(name_ptr->name,TKNWORD) == 0)
                    break;
                 name_ptr = name_ptr->next;
            END_WHILE
            IF(name_ptr != NULL)THEN
                      (name_ptr->counter)++;
            ELSE
                 name_ptr          = AllocRuleList();
                 save_word         = AddSymbol(TKNWORD);
                 name_ptr->name    = save_word->contents;
                 name_ptr->counter = 1;
                 name_ptr->next    = CurrentFact->deactivate_list_a;
                 CurrentFact->deactivate_list_a = name_ptr;
            END_IF
         END_IF
    END_IF
  }
  	
 /*
  * ----------------------------------------------------------------------
  *  void ProcessRuleFire()
  *        recording all the informations about the rule being fired
  * ----------------------------------------------------------------------
  */
  int ProcessRuleFire()

  {
     int token ,i = 0,j,fire_num;
     char  *name;
     ACT_R_PTR rule_ptr;
     

	 token = GetTokenWithEndOfLine(CurrentFile);
	 IF(token != INTEGER)THEN
	     FindEndOfLine(CurrentFile);
	     return(ERROR);
	 END_IF
     
     fire_num = atoi(TKNWORD);
     IF(FirstRule)THEN
       IF(fire_num > 1)THEN
           ErrorMessage(WARNING, "Incomplete trace file!");
           SendMessage
                  ("\n         the first rule firing in the trace file is not",YES);
           SendMessage
                  ("\n         FIRE 1. Trace file analysis may be incomplete",YES);
        END_IF
        FirstRule = FALSE;
     END_IF
     token = GetTokenWithEndOfLine(CurrentFile);
     i = strlen(TKNWORD);
     if(TKNWORD[i-1] == ':')
      TKNWORD[i-1] = EOS;
     name = TKNWORD;
     IF(VERBOSE IS_ON)THEN
           sprintf(MsgBuf,"\n\nProcessing rule firing <%.40s>",TKNWORD);
           SendMessage(MsgBuf,NO);
     END_IF
     IF(INFERENCE_TRACE IS_ON) THEN
        ProcessRuleFireForInfTrace(name);
     END_IF
     rule_ptr = ActivatedRuleHead;
     WHILE(rule_ptr != NULL)DO
        if(strcmp(rule_ptr->name,name) == 0)
             break;
        rule_ptr = rule_ptr->next;
     END_WHILE
     IF(rule_ptr != NULL)THEN
         (rule_ptr->fire_num)++;
         LastFiredRule = rule_ptr;
     ELSE
       SendMessage("Incomplete trace file \n");
     END_IF
     cycle++;
     FindEndOfLine(CurrentFile);
     return(OK);
   }

/*
 *------------------------------------------------------------
 * ProcessRuleFireForInfTrace()
 *------------------------------------------------------------
 */

ProcessRuleFireForInfTrace(name)
char *name;
{
  AGEN_RULE_PTR ag_rule = NULL,temp_ag_rule = NULL;
   
      sprintf(MsgBuf,"\nCycle %d\n",cycle);
      SendMessageToFile(logfile,MsgBuf);
      ag_rule = AgendaRules;
      WHILE(ag_rule != NULL)DO
         if(strcmp(ag_rule->rule->name,name) == 0)
           break;
         temp_ag_rule = ag_rule;
         ag_rule = ag_rule->next_rule;
      END_WHILE
      IF((ag_rule != NULL)&&(temp_ag_rule != NULL))THEN
         temp_ag_rule->next_rule = ag_rule->next_rule;
         ag_rule->next_rule = NULL;
         FreeAgendaRule(ag_rule);
      ELSE_IF(ag_rule == NULL)
         SendMessageToFile(logfile,"Incomplete Trace File ");
      ELSE_IF(temp_ag_rule == NULL)  /* Rule match at the begining of the list */
         AgendaRules = ag_rule->next_rule;
         ag_rule->next_rule = NULL;
         FreeAgendaRule(ag_rule);
      ELSE
         PrintRuleInAgenda(AgendaRules,name);
      END_IF
}
/*
 * ===========================================================
 *  Print the information about the fired rule and
 *  print all the rules in agenda
 * ===========================================================
 */
void PrintRuleInAgenda(list,name)
AGEN_RULE_PTR list;
char *name;
{
   AGEN_RULE_PTR temp = NULL;
   PAT_LIST_PTR pat = NULL,temp_pat = NULL;
   PAT_TKN_PTR cur_field = NULL,start_field;
   char num_str[20];
   int token,field_num,count, num = 0;
   short found;
   
   CurrentRule = SearchRule(RuleHead,name);
   IF(CurrentRule != NULL)THEN
       sprintf(MsgBuf,"Rule: %s\n",name);
       SendMessageToFile(logfile,MsgBuf,NO);
       pat = CurrentRule->lhs_patterns;
   END_IF
   start_field = pat->pattern;
   token = GetTokenWithEndOfLine(CurrentFile);
   
   /******* Print first "not" patterns if any *********/
   
   IF(((token == EOL) OR (token == STOP))&&(pat != NULL))THEN
     WHILE(pat!= NULL)DO
         PrintPattern(pat,NULL,NULL);
         pat = pat->next_pattern;
     END_WHILE
    END_IF
   /********** Print the regular patterns ***********/
   
   WHILE ((token!= EOL) && (token != STOP)) DO
     IF(token == INTEGER)THEN
        field_num = atoi(TKNWORD);
        found = FALSE;
        (void)SearchFact(ActvatedRelationHead,field_num,&found);
        IF(found)THEN
	         WHILE(pat!=NULL)DO
	            if(pat->in_not != YES)
			      break;
	            PrintPattern(pat,NULL,NULL);
	            pat = pat->next_pattern;
	         END_WHILE
	         IF(pat != NULL)THEN
	             IF(pat->in_or)THEN
	               temp_pat = SearchOrPattern(pat,CurrentFact,CurrentActivatedRelation->name);
	               IF(temp_pat != NULL)THEN
	                 temp_pat->num = pat->num;
	                 pat = pat->next_pattern;
	               END_IF
	             ELSE_IF (pat->in_and)
	               temp_pat = SearchAndPattern(&start_field);
	               IF(temp_pat != NULL)THEN
	                 temp_pat->num = pat->num + num;
	                 num++;
	               END_IF
	             ELSE
	               temp_pat = AllocPatternList();
	               CurrentPattern = temp_pat;
	               temp_pat->num = pat->num;
	               cur_field = pat->pattern;
	               while(cur_field->type != LPAREN)
	                 cur_field = cur_field->next;
	               temp_pat->pattern = InsertTokenIntoPattern(temp_pat->pattern,
					cur_field->word,cur_field->type);
	               CurrentTokenInPattern->num = cur_field->num;
	               count = 1;
	               WHILE(count)DO
	               	   cur_field = cur_field->next;
	                   if(cur_field->type == LPAREN)
	                     count++;
	                   else if (cur_field->type == RPAREN)
	                     count--;
	                   temp_pat->pattern = 
			           InsertTokenIntoPattern(temp_pat->pattern,
			   		   cur_field->word,cur_field->type);
	                   CurrentTokenInPattern->num = cur_field->num;
	               END_WHILE
	               pat = pat->next_pattern;
	             END_IF
	         END_IF
	         if(temp_pat != NULL)
	           PrintPattern(temp_pat,CurrentFact,CurrentActivatedRelation->name);
	         FreePatternList(temp_pat);
	         temp_pat = NULL;
	    ELSE
         SendMessageToFile(logfile,"Incomplete trace file \n");
        END_IF
     END_IF
     token = GetTokenWithEndOfLine(CurrentFile);
    DONE
    
   /****** Print the left over patterns (possibly  "not" patterns) ******/
   WHILE(pat != NULL)DO
       PrintPattern(pat,NULL,NULL);
       pat = pat->next_pattern;
   END_WHILE
    
    /****** Print Rules in agenda ******/
   IF(list != NULL)THEN
      temp = list->next_rule;
      if(temp!= NULL)
         SendMessageToFile(logfile,"Rules in the agenda \n",NO);
   END_IF
   WHILE(temp != NULL)DO
       sprintf(MsgBuf,"   Rule: %s\n",temp->rule->name);
       SendMessageToFile(logfile,MsgBuf,NO);
       temp = temp->next_rule;
   END_WHILE

}

/*****************************************************
 *  SearchOrPattern:
 *    Find the pattern in "or" pattern that matches with
 *    the fact.
 *****************************************************
 */

static PAT_LIST_PTR SearchOrPattern(pat,fact,name)
PAT_LIST_PTR pat;
SPE_FACT_PTR fact;
char *name;
{
   PAT_LIST_PTR temp_pat = NULL;
   PAT_TKN_PTR  field = NULL,start_ptr = NULL;
   int count,start_process = YES,rtn;
   FIELD_PTR temp_field = NULL;
   
   field  = pat->pattern;
   temp_field = fact->field;
   while(strcmp(field->word,"or") != 0)
     field = field->next;
   while(field->type != LPAREN)
     field = field->next;
   start_ptr = field;
   while(field->num == 0)
     field = field->next;
   while(temp_field != NULL)  DO /*search for the pattern that matches with the fact */
	   IF((field->type == WORD) || (field->type == STRING) ||
		(field->type == INTEGER) || (field->type == FLOAT) ||
		(field->type == INSTANCE)) THEN
	      IF(start_process) THEN    /* check the relation's name */
			   IF((field->type != INTEGER) AND (field->type != FLOAT))THEN
	 	          rtn = strcmp(field->word,name);
	 	       ELSE_IF (field->type == FLOAT)
		          if(atof(field->word) == atof(name))
			        rtn = 0;
		          else
			        rtn = 1;
			   ELSE_IF (field->type == INTEGER)
		          if(atol(field->word) == atol(name))
			        rtn = 0;
		          else
			        rtn = 1;
			   END_IF                 
	         ELSE    				/* check the field */
               IF((field->type != INTEGER) && (field->type != FLOAT)) THEN
	              rtn = strcmp(field->word,temp_field->word);
               ELSE_IF(field->type == FLOAT)
                 if(atof(field->word) == atof(temp_field->word))
		             rtn = 0;
	             else
		             rtn = 1;
		       ELSE_IF(field->type == INTEGER)
		          if(atol(field->word) == atol(temp_field->word))
			        rtn = 0;
		          else
			        rtn = 1; 
			   END_IF
	         END_IF
	         IF( rtn != 0) THEN /* Not the correct pattern in the OR statement */
	           count = 1;
	           while((count) && (field != NULL))
	              {
	               field = field->next;
	               if(field->type == RPAREN)
	                 count--;
	               else if (field->type == LPAREN)
	                 count++;       
	              }   
	            while((field != NULL)&&(field->type != LPAREN))
                    field = field->next;
                    start_ptr = field;
                    field = field->next;
                    temp_field = fact->field;
                    start_process = TRUE;      
	          ELSE  /* rtn == 0 */
                  if(start_process)
                    start_process = FALSE;
                  else
                    temp_field = temp_field->next;
                  IF(temp_field)THEN
  		    	    field = field->next;
                   while(field->num == 0)     /* searching for real field */
                      field = field->next;
                  ELSE
                    while(field->type != RPAREN)
                      field = field->next;
                  END_IF
             END_IF
         ELSE  /* Variable */
		   if(start_process)
		      start_process = FALSE;
		   else
		      temp_field = temp_field->next;
	       IF(temp_field)THEN
	  	     field = field->next;
	         while(field->num == 0) /* searching for real field */
	           field = field->next;
	       ELSE
	         while(strcmp(field->word,")") != 0)
	            field = field->next;
	       END_IF
         END_IF
   END_WHILE
   temp_pat = AllocPatternList();
   CurrentPattern = temp_pat;
   WHILE(start_ptr != field->next) DO
        temp_pat->pattern = InsertTokenIntoPattern(temp_pat->pattern,
        					start_ptr->word,start_ptr->type);
        CurrentTokenInPattern->num = start_ptr->num;
        start_ptr = start_ptr->next;
   END_WHILE
   return(temp_pat);
}

/************************************************************************/
/* SearchAndPattern                                                       */
/*    return the pattern(in the and pattern) that matches with the fact */
/************************************************************************/

static PAT_LIST_PTR SearchAndPattern(start_field)
PAT_TKN_PTR *start_field;
{
    PAT_TKN_PTR temp_field = NULL;
    PAT_LIST_PTR temp_pat = NULL;
    int count = 1;
    
    temp_field = *start_field;
    while((temp_field != NULL) && (strcmp(temp_field->word,"and") != 0))
      temp_field = temp_field->next;
    while((temp_field != NULL)&&(temp_field->type != LPAREN))
      temp_field = temp_field->next;
    if(temp_field == NULL)
      return(NULL);
    temp_pat = AllocPatternList();
    CurrentPattern = temp_pat;
    WHILE(count)DO
        temp_pat->pattern = InsertTokenIntoPattern(temp_pat->pattern,temp_field->word,
        				temp_field->type);
        CurrentTokenInPattern->num = temp_field->num;
        temp_field = temp_field->next;
        if(temp_field->type == LPAREN)
           count++;
        else if (temp_field->type == RPAREN)
           count--;
    END_WHILE
    temp_pat->pattern = InsertTokenIntoPattern(temp_pat->pattern,temp_field->word,
        				temp_field->type);
    CurrentTokenInPattern->num = temp_field->num;
    *start_field = temp_field->next;
    return(temp_pat);
    
}
/* ========================================================
 *  Print the pattern and the value of the variables that 
 *    appear in the pattern.
 * ========================================================
 */

PrintPattern(pat,fact,name)
PAT_LIST_PTR pat;
SPE_FACT_PTR fact;
char *name;
{
   PAT_TKN_PTR cur_word = NULL,last_word = NULL;
   
       sprintf(MsgBuf,"Pattern%d  :",pat->num);
       SendMessageToFile(logfile,MsgBuf,NO);
       cur_word = pat->pattern;
       SendMessageToFile(logfile," ",NO);
       WHILE(cur_word != NULL)DO
            IF(last_word != NULL)THEN
		        if((strcmp(cur_word->word,"(") == 0) && 
		   			(strcmp(last_word->word,":") != 0))
                   SendMessageToFile(logfile," ",NO);
				else if(((strcmp(last_word->word,"~") != 0) &&
						(strcmp(last_word->word,"&") != 0)) &&
						(strcmp(cur_word->word,"&") != 0))
		  			SendMessageToFile(logfile," ",NO);
            END_IF
            if(cur_word->type == BWORD)
              SendMessageToFile(logfile,"?",NO);
            else if(cur_word->type == BWORDS)
              SendMessageToFile(logfile,"$?",NO);
            sprintf(MsgBuf,"%s",cur_word->word);
	        SendMessageToFile(logfile,MsgBuf,NO);
            last_word = cur_word;
            cur_word = cur_word->next;
       END_WHILE
      SendMessageToFile(logfile,"\n",NO);
      IF(pat->has_wildcard)  THEN/* If wildcard found, will print the whole fact */
           SendMessageToFile(logfile,"Fact :",NO);
           PrintRelToFile(fact,name);
           SendMessageToFile(logfile,"\n",NO);
           return;
      END_IF
      IF (fact != NULL) THEN
      
        /* pattern is in relation form */
      
        IF(fact->templ_slot == NULL) THEN 
           PrintVariableInField(fact->field,pat->pattern,name);
           
        /* pattern is in template form */
        
        ELSE 							  
           PrintVariableInSlots(fact->templ_slot,pat->pattern);
        END_IF 
    
       /* Print the originated point of the fact */ 
       
	    IF(fact->deffact_name != NULL)THEN
	      sprintf(MsgBuf,"  From deffact: %s \n",fact->deffact_name);
	      SendMessageToFile(logfile,MsgBuf,NO);
	    ELSE_IF(fact->asserted_by_rule)
	      sprintf(MsgBuf,"  Asserted by rule: %s \n",fact->asserted_by_rule);
	      SendMessageToFile(logfile,MsgBuf,NO);
	    END_IF
      END_IF
 }
/*
 * ==========================================================
 *  Print the value of the variables that found in template.
 * ==========================================================
 */
PrintVariableInSlots(slots,pat_tkn)
MY_TMPLATE_PTR slots;
PAT_TKN_PTR pat_tkn;
{
   MY_TMPLATE_PTR temp_slot;
   PAT_TKN_PTR cur_tkn;
   FIELD_PTR afield;
   char *aslot;
   int count1;
   
   
   cur_tkn = pat_tkn;
   while(cur_tkn->num != 1)
     cur_tkn = cur_tkn->next;
   cur_tkn = cur_tkn->next;
   WHILE(cur_tkn != NULL) DO
      IF (cur_tkn->type == LPAREN) THEN
      	 cur_tkn =cur_tkn->next;
         aslot = cur_tkn->word;
         count1 = 1;
         WHILE(cur_tkn->num == 0)DO
            if (cur_tkn->type == LPAREN)
              count1++;
            else if(cur_tkn->type == RPAREN)
              count1--;
            cur_tkn = cur_tkn->next;
         END_WHILE
         IF((cur_tkn->type == BWORD) || (cur_tkn->type == SINGLE) ||
         	 (cur_tkn->type == BWORDS)|| (cur_tkn->type == MULTIPLE)) THEN
         	 
             IF((cur_tkn->type == BWORD) || (cur_tkn->type == SINGLE))THEN
               sprintf(MsgBuf,"  Slot %s Fact ?%s :",aslot, cur_tkn->word);
               SendMessageToFile(logfile,MsgBuf,NO);
             ELSE
               sprintf(MsgBuf,"  Slot %s Fact $?%s :",aslot, cur_tkn->word);
               SendMessageToFile(logfile,MsgBuf,NO);
             END_IF 
             temp_slot = slots;
             while((temp_slot != NULL) &&(strcmp(temp_slot->name,aslot) != 0)) 
              temp_slot = temp_slot->next;
             IF(temp_slot != NULL) THEN
                 afield = temp_slot->value_list;
             ELSE
                  afield = NULL;
                  sprintf(MsgBuf,"Slot %s was not found in trace file \n",aslot);
                  ErrorMessage(ERROR,MsgBuf);
             END_IF
             WHILE(afield != NULL)DO
                 sprintf(MsgBuf," %s ",afield->word);
                 SendMessageToFile(logfile,MsgBuf,NO);
                 afield = afield->next;
             END_WHILE
             WHILE(count1)DO
           		cur_tkn = cur_tkn->next;
           		if (cur_tkn->type == LPAREN)
            		count1++;
	            else if (cur_tkn->type == RPAREN)
	            	count1--;
           	 END_WHILE
           	 SendMessageToFile(logfile,"\n",NO);
         END_IF
      END_IF
      cur_tkn = cur_tkn->next;
   END_WHILE
   SendMessageToFile(logfile,"\n",NO);
}

/*
 * =======================================================
 * Print the value of the variables that found in relation.
 * =======================================================
 */
PrintVariableInField(field,patterns,name)
FIELD_PTR field;
PAT_TKN_PTR patterns;
char *name;
{
   FIELD_PTR temp_field;
   PAT_TKN_PTR cur_word;
   
		temp_field = field;
	    cur_word = patterns;
	    cur_word = cur_word->next;
	    IF((cur_word->type != WORD) && (cur_word->type != STRING)
		&& (cur_word->type != INTEGER) && (cur_word->type != FLOAT) 
		&& (cur_word->type != INSTANCE))THEN
	         sprintf(MsgBuf,"Relation ?%s : %s \n",cur_word->word,name);
	         SendMessageToFile(logfile,MsgBuf,NO);
	    END_IF
	    cur_word = cur_word->next;
        IF(temp_field == NULL)THEN
               sprintf(MsgBuf,"Fact : ( %s )\n",CurrentActivatedRelation->name);
               SendMessageToFile(logfile,MsgBuf,NO);
        END_IF
	    WHILE((cur_word != NULL)&&(temp_field != NULL))DO
            IF((cur_word->type == BWORD)&&(cur_word->num > 0))THEN
                      sprintf(MsgBuf,"  Field %d Fact ?%s : %s\n",cur_word->num,
	           			cur_word->word,temp_field->word);
 	              	  SendMessageToFile(logfile,MsgBuf,NO);
                      temp_field = temp_field->next;
            ELSE_IF((cur_word->type == BWORDS)&& (cur_word->num > 0))
	              sprintf(MsgBuf,"  Field %d Fact $?%s : %s\n",cur_word->num,
	           			cur_word->word,temp_field->word);
	              SendMessageToFile(logfile,MsgBuf,NO);
                      temp_field = temp_field->next;
	        ELSE_IF(((cur_word->type == SINGLE) OR 
			         (cur_word->type == MULTIPLE))AND
                         (cur_word->num> 0))
	              sprintf(MsgBuf,"  Field %d Fact %s : %s\n",cur_word->num,
	           			cur_word->word,temp_field->word);
	              SendMessageToFile(logfile,MsgBuf,NO);
                      temp_field = temp_field->next;
	        ELSE_IF ((cur_word->type == STRING) OR 
			         (cur_word->type == INTEGER) OR
			         (cur_word->type == FLOAT) OR
			         (cur_word->type == INSTANCE))
                  temp_field = temp_field->next;
            ELSE_IF (cur_word->type == WORD)
                  if((strcmp(cur_word->word,"&") != 0) AND
				     (strcmp(cur_word->word,":") != 0))
                       temp_field = temp_field->next;
            END_IF
              cur_word = cur_word->next;
       END_WHILE
 }

/*
 * ----------------------------------------------------------
 *   OpenLogFile()
 * ----------------------------------------------------------
 */

int OpenLogFile(filename)
char *filename;
{
   if((logfile = fopen(filename,"w")) == NULL)
     return(ERROR);
   return(1);
}

/*
 * ----------------------------------------------------------
 *   CloseLogFile()
 * ----------------------------------------------------------
 */

void CloseLogFile()
{
  fclose(logfile);
}


/*
 * ----------------------------------------------------------
 *   SendMessageToFile()
 * ----------------------------------------------------------
 */

void SendMessageToFile(logfile,buf)
FILE *logfile;
char *buf;
{
  if((logfile != NULL) && ((IBM_COM_LINE)||( !MAC_LSC && !IBM_TBC && !IBM_MSC)))
    fprintf(logfile,"%s",buf);
  else
    SendMessage(buf,NO);
}
/* ------------------------------------------------ *
 *  PrintRelToFile()
 *    Print a relation, includes the relation's name
 *         and its fields.
 *
 *  Argument:
 *     fact_ptr : pointer pointing to the relation
 *     name     : name of the relation.
 *
 *  Returns: nothing.
 * ------------------------------------------------ */

void PrintRelToFile(fact_ptr,name)
SPE_FACT_PTR fact_ptr;
char *name;
{
     FIELD_PTR fld_ptr;
     MY_TMPLATE_PTR slot_ptr;

     sprintf(MsgBuf,"(%s",name);
     SendMessageToFile(logfile,MsgBuf);
      IF(fact_ptr->field != NULL) THEN
	     fld_ptr = fact_ptr->field;
	     while(fld_ptr!=NULL)
	       {
	         sprintf(MsgBuf,"  %s",fld_ptr->word);
	         SendMessageToFile(logfile,MsgBuf);
	         fld_ptr = fld_ptr->next;
	       }
	  ELSE_IF (fact_ptr->templ_slot != NULL)
	    slot_ptr = fact_ptr->templ_slot; 
	    while(slot_ptr != NULL)
	     {
	        sprintf(MsgBuf,"(%s ",slot_ptr->name);
	        SendMessageToFile(logfile,MsgBuf);
	        fld_ptr = slot_ptr->value_list;
	        while(fld_ptr != NULL)
	          {
	            sprintf(MsgBuf," %s",fld_ptr->word);
	            SendMessageToFile(logfile,MsgBuf);
	            fld_ptr = fld_ptr->next;
	          }
	         SendMessageToFile(logfile,")");
	        slot_ptr = slot_ptr->next;
	     }
	  END_IF
     SendMessageToFile(logfile,")");

}

