#include <stdio.h>
#include "crsv.h"

/*********************************************************************
 * This file contains the functions used to compare RHS actions to a
 * list of CLIPS defined actions (the reserved words). If a token is
 * not found in that list, it can be assumed to be an external function
 * call. The list of reserved words will have to be updated when new
 * actions are added to CLIPS. Only entry point is through the function
 * CheckExternalFuncs().
 *********************************************************************
 */

/* ===========  Functions defined Externally but used here  =========== */

/* -----------------------------
 *  From the file: CRSVMEM.C
 * -----------------------------
 */

extern void  MemoryError();
extern char *GenAlloc();


/* ===========  Functions defined here for Global use  ================ */

int  CheckExternalFuncs();


/* ===========  Functions defined here for internal use  ============== */

void            Lower();
void            StoreToken();
struct letter  *NewLetter();
int             LoadReservedWords();
void            ResetWordList();
short int       NextWordLetter();


/* ===========  Variables defined here for Global use  ================ */


/* ===========  Variables defined Externally but used here  =========== */


/* ===========  Variables defined here for internal use  ============== */

/* ------------------------- Macro Definitions --------------------------- */

#define TOKEN_SIZE         40
#define DUPLICATE_TOKEN    -2
#define NOT_TOKEN          -1
#define NULL_TOKEN          0

#define CASE_OFFSET ('a' - 'A')


/* ----------------------- Structure Definitions ------------------------- */

struct letter
  {
   char value;
   struct letter *prev;
   struct letter *step;
   struct letter *next;
   short int key;
   int return_type;
  };

struct ReservedWord {
   char *word;
   int return_type;
   short int key;
   
   };


/* -------------------- Internal Global Variables ------------------------- */

/* -----------------------
 *  CLIPS Actions. This
 *  list is valid for
 *  CLIPS Release 5.0
 * -----------------------
 */

struct ReservedWord ReservedList[] = {
"!=",					BOOLEAN,	1,
"*",        			UNKNOWN,	2,
"**",       			FLOAT,		3,
"+",        			UNKNOWN,	4,
"-",        			UNKNOWN,	5,
"/",        			UNKNOWN,	6,
"<",        			BOOLEAN,	7,
"<=",       			BOOLEAN,	8,
"=",        			BOOLEAN,	9,
">",        			BOOLEAN,	10,
">=",       			BOOLEAN,	11,
"ABS",           		NUMBER,		12,
"ACOS",     			FLOAT,		13,
"ACOSH",    			FLOAT,		14,
"ACOT",     			FLOAT,		15,
"ACOTH",    			FLOAT,		16,
"ACSC",     			FLOAT,		17,
"ACSCH",    			FLOAT,		18,
"AGENDA",       		VOID,		19,
"AND",      			BOOLEAN,	20,
"ANY-INSTANCEP",		BOOLEAN,	21,
"ASEC",     			FLOAT,		22,
"ASECH",    			FLOAT,		23,
"ASIN",     			FLOAT,		24,
"ASINH",    			FLOAT,		25,
"ASSERT",       		VOID,		26,
"ATAN",     			FLOAT,		27,
"ATANH",    			FLOAT,		28,
"BATCH",         		VOID,		29,
"BIND",         		UNKNOWN,	30,
"BLOAD",				VOID,		31,
"BSAVE",				VOID,		32,
"BUILD",       			BOOLEAN,	33,
"CALL-NEXT-HANDLER",	UNKNOWN,	34,
"CALL-NEXT-METHOD",		UNKNOWN,	35,
"CLASS",				UNKNOWN,	36,
"CLASS-INSTANCES",		VOID,		37,
"CLASS-MESSAGE-HANDLER-EXISTP",	BOOLEAN,38,
"CLASS-SLOT-EXISTP",		BOOLEAN,39,
"CLEAR",   				VOID,		40,
"CLOSE",      			BOOLEAN,	41,
"CONSERVE-MEM",  		VOID,		42,
"COS",      			FLOAT,		43,
"COSH",     			FLOAT,		44,
"COT",      			FLOAT,		45,
"COTH",     			FLOAT,		46,
"CREATE$",     			MULTI_FIELD,47,
"CRSV-TRACE-OFF",		VOID,		48,
"CRSV-TRACE-ON", 		VOID,		49,
"CSC",      			FLOAT,		50,
"CSCH",     			FLOAT,		51,
"DEG-GRAD", 			FLOAT,		52,
"DEG-RAD",  			FLOAT,		53,
"DELAYED-DO-FOR-ALL-INSTANCES",	UNKNOWN,54,
"DELETE$",     			MULTI_FIELD,55,
"DEPENDENCIES", 		VOID,		56,
"DEPENDENTS",   		VOID,		57,
"DESCRIBE-CLASS",		VOID,		58,
"DIRECT-MV-DELETE",		BOOLEAN,	59,
"DIRECT-MV-INSERT",		BOOLEAN,	60,
"DIRECT-MV-REPLACE",	BOOLEAN,	61,
"DIV",      			LONG,		62,
"DO-FOR-ALL-INSTANCES",	UNKNOWN,	63,
"DO-FOR-INSTANCE",		UNKNOWN,	64,
"DRIBBLE-OFF",   		VOID,		65,
"DRIBBLE-ON",    		VOID,		66,
"DUPLICATE",			UNKNOWN,	67,
"EDIT",					VOID,		68,
"EQ",       			BOOLEAN,	69,
"EVAL",        			UNKNOWN,	70,
"EVENP",    			BOOLEAN,	71,
"EXCISE",       		VOID,		72,
"EXIT",    				VOID,		73,
"EXP",      			FLOAT,		74,
"FACT-NUMBER",   		LONG,		75,
"FACTS",        		VOID,		76,
"FETCH",        		INTEGER,	77,
"FIND-ALL-INSTANCES",	UNKNOWN,	78,
"FIND-INSTANCE",		UNKNOWN,	79,
"FLOAT",         		FLOAT,		80,
"FLOATP",   			BOOLEAN,	81,
"FORMAT",     			STRING,		82,
"FPRINTOUT",  			VOID,		83,
"GENSYM",        		WORD,		84,
"GENSYM*",       		WORD,		85,
"GET",					UNKNOWN,	86,
"GET-AUTO-FLOAT-DIVIDEND", 	BOOLEAN,87,
"GET-DYNAMIC-DEFTEMPLATE-CHECKING",BOOLEAN,88,
"GET-FACT-DUPLICATION", 	BOOLEAN,89,
"GET-INCREMENTAL-RESET", 	BOOLEAN,90,
"GET-RESET-GLOBALS", 		BOOLEAN,91,
"GET-SALIENCE-EVALUATION", 	WORD,	92,
"GET-STRATEGY", 		WORD,		93,
"GRAD-DEG", 			FLOAT,		94,
"HALT",         		VOID,		95,
"HELP",         		VOID,		96,
"HELP-PATH",    		VOID,		97,
"IF",           		UNKNOWN,	98,
"INIT-SLOTS",			UNKNOWN,	99,
"INITIALIZE-INSTANCE",	UNKNOWN,	100,
"INSTANCE-ADDRESS",		UNKNOWN,	101,
"INSTANCE-NAME",		UNKNOWN,	102,
"INSTANCE-NAME-TO-SYMBOL",	UNKNOWN,103,
"INSTANCE-NAMEP",		BOOLEAN,	104,
"INSTANCEP",			BOOLEAN,	105,
"INSTANCES",			VOID,		106,
"INTEGER",       		LONG,		107,
"INTEGERP", 			BOOLEAN,	108,
"LENGTH",        		LONG,		109,
"LENGTH$",       		LONG,		110,
"LEXEMEP",  			BOOLEAN,	111,
"LIST-DEFCLASSES",		VOID,		112,
"LIST-DEFFACTS", 		VOID,		113,
"LIST-DEFFUNCTIONS",	VOID,		114,
"LIST-DEFGENERICS",		VOID,		115,
"LIST-DEFGLOBALS", 		VOID,		116,
"LIST-DEFINSTANCES",	VOID,		117,
"LIST-DEFMESSAGE-HANDLERS",	VOID,	118,
"LIST-DEFMETHODS",		VOID,		119,
"LIST-DEFRULES",		VOID,		120,
"LIST-DEFTEMPLATES",		VOID,	121,
"LOAD",          		VOID,		122,
"LOAD-FACTS",    		VOID,		123,
"LOAD-INSTANCES",		BOOLEAN,	124,
"LOG",      			FLOAT,		125,
"LOG10",    			FLOAT,		126,
"LOWCASE",     			UNKNOWN,	127,
"MAKE-INSTANCE",		UNKNOWN,	128,
"MATCHES",      		VOID,		129,
"MAX",           		UNKNOWN,	130,
"MEM-REQUESTS",  		LONG,		131,
"MEM-USED",      		LONG,		132,
"MEMBER",      			UNKNOWN,	133,
"MEMBER$",     			UNKNOWN,	134,
"MIN",           		NUMBER,		135,
"MOD",      			NUMBER,		136,
"MODIFY",			 	UNKNOWN,	137,
"MV-APPEND",   			MULTI_FIELD,138,
"MV-DELETE",   			MULTI_FIELD,139,
"MV-REPLACE",  			MULTI_FIELD,140,
"MV-SLOT-DELETE",		UNKNOWN,	141,
"MV-SLOT-INSERT",		UNKNOWN,	142,
"MV-SLOT-REPLACE",		UNKNOWN,	143,
"MV-SUBSEQ",   			MULTI_FIELD,144,
"NEQ",      			BOOLEAN,	145,
"NEXT-HANDLERP",		BOOLEAN,	146,
"NEXT-METHODP",			BOOLEAN,	147,
"NOT",      			BOOLEAN,	148,
"NTH",         			UNKNOWN,	149,
"NTH$",        			UNKNOWN,	150,
"NUMBERP",  			BOOLEAN,	151,
"ODDP",     			BOOLEAN,	152,
"OPEN",       			BOOLEAN,	153,
"OPTIONS",       		VOID,		154,
"OR",       			BOOLEAN,	155,
"OVERRIDE-NEXT-HANDLER",	UNKNOWN,156,
"PI",       			FLOAT,		157,
"POINTERP", 			BOOLEAN,	158,
"PPDEFCLASS",			VOID,		159,
"PPDEFFACTS",    		VOID,		160,
"PPDEFFUNCTION",		VOID,		161,
"PPDEFGENERIC",			VOID,		162,
"PPDEFINSTANCES",		VOID,		163,
"PPDEFMESSAGE-HANDLER",		VOID,	164,
"PPDEFMETHOD",			VOID,		165,
"PPDEFRULE",    		VOID,		166,
"PPDEFTEMPLATE",		VOID,		167,
"PPINSTANCE",			VOID,		168,
"PPRULE",       		VOID,		169,
"PREVIEW-GENERIC",		VOID,		170,
"PREVIEW-SEND",			VOID,		171,
"PRINT-REGION", 		INTEGER,	172,
"PRINTOUT",   			VOID,		173,
"PROGN",        		UNKNOWN,	174,
"PUT",					BOOLEAN,	175,
"RAD-DEG",  			FLOAT,		176,
"RANDOM",        		LONG,		177,
"READ",       			UNKNOWN,	178,
"READLINE",   			UNKNOWN,	179,
"REFRESH",      		VOID,		180,
"REFRESH-AGENDA",		VOID,		181,
"RELEASE-MEM",   		VOID,		182,
"REMOVE-BREAK", 		VOID,		183,
"RESET",   				VOID,		184,
"RETRACT",      		VOID,		185,
"ROUND",    			LONG,		186,
"RULES",        		VOID,		187,
"RULES-TO-C", 			VOID,		188,
"RUN",          		VOID,		189,
"SAVE",          		VOID,		190,
"SAVE-FACTS",    		VOID,		191,
"SAVE-INSTANCES",		BOOLEAN,	192,
"SEC",      			FLOAT,		193,
"SECH",     			FLOAT,		194,
"SEED",          		VOID,		195,
"SEND",					UNKNOWN,	196,
"SEQUENCEP",			BOOLEAN,	197,
"SET-AUTO-FLOAT-DIVIDEND", 	BOOLEAN,198,
"SET-BREAK",    		VOID,		199,
"SET-DYNAMIC-DEFTEMPLATE-CHECKING",BOOLEAN,200,
"SET-FACT-DUPLICATION",		BOOLEAN,201,
"SET-INCREMENTAL-RESET",	BOOLEAN,202,
"SET-RESET-GLOBALS",		BOOLEAN,203,
"SET-SALIENCE-EVALUATION",	WORD,	204,
"SET-STRATEGY", 		WORD,		205,
"SETGEN",        		LONG,		206,
"SHOW-BREAKS",  		VOID,		207,
"SIN",      			FLOAT,		208,
"SINH",     			FLOAT,		209,
"SLOT-BOUNDP",			BOOLEAN,	210,
"SLOT-EXISTP",			BOOLEAN,	211,
"SLOT-INITABLEP",		BOOLEAN,	212,
"SLOT-WRITABLEP",		BOOLEAN,	213,
"SQRT",     			FLOAT,		214,
"STR-ASSERT",    		VOID,		215,
"STR-CAT",     			UNKNOWN,	216,
"STR-COMPARE", 			LONG,		217,
"STR-EXPLODE", 			MULTI_FIELD,218,
"STR-IMPLODE", 			STRING,		219,
"STR-INDEX",   			LONG,		220,
"STR-LENGTH",  			LONG,		221,
"STRINGP",  			BOOLEAN,	222,
"SUB-STRING",  			STRING,		223,
"SUBCLASSP",			BOOLEAN,	224,
"SUBSET",      			BOOLEAN,	225,
"SUPERCLASSP",			BOOLEAN,	226,
"SYM-CAT",     			UNKNOWN,	227,
"SYMBOL-TO-INSTANCE-NAME",	UNKNOWN,228,
"SYMBOLP",  			BOOLEAN,	229,
"SYSTEM",        		VOID,		230,
"TAN",      			FLOAT,		231,
"TANH",     			FLOAT,		232,
"TIME",          		FLOAT,		233,
"TOSS",         		INTEGER,	234,
"TRUNC",         		LONG,		235,
"TYPE",				    UNKNOWN,	236,
"UNDEFCLASS",			VOID,		237,
"UNDEFFACTS",			VOID,		238,
"UNDEFFUNCTION",		VOID,		239,
"UNDEFGENERIC",			VOID,		240,
"UNDEFINSTANCES",		VOID,		241,
"UNDEFMESSAGE-HANDLER",	VOID,		242,
"UNDEFMETHOD",			VOID,		243,
"UNDEFRULE",    		VOID,		244,
"UNDEFTEMPLATE",		VOID,		245,
"UNMAKE-INSTANCE",		VOID,		246,
"UNWATCH", 				VOID,		247,
"UPCASE",      			UNKNOWN,	248,
"WATCH",   				VOID,		249,
"WHILE",        		UNKNOWN,	250,
"WORDP",    			BOOLEAN,	251,
   };

#define NUM_RSV_WORDS     251

static struct letter *TokenTree     = NULL;
static struct letter *current_letter = NULL;


/* ======================================================================== */

/* --------------------------------------------------
 *  NAME   :  CheckExternalFuncs()
 *
 *  PURPOSE:  This function takes a word
 *            and compares it to the list
 *            of reserved words. If the
 *            word is found, it returns
 *            TRUE, otherwise FALSE.
 *
 *  INPUTS:   A pointer to the word to be checked.
 *
 *  RETURNS:  An integer, either TRUE or FALSE.
 *
 *  NOTES:    The token comparison functions were
 *            provided by Brian Donnell. He used
 *            them for the CLIPS tutorial parser.
 * --------------------------------------------------
 */

int CheckExternalFuncs(word,rtn)
char *word;
int *rtn;
{
   int index;
   short int key;

   IF(TokenTree EQ NULL) THEN
      (void)LoadReservedWords();
   ELSE
      ResetWordList();
   END_IF

   for(index = 0; word[index] NEQ EOS; index++) DO
      key = NextWordLetter(word[index],rtn);
   END_FOR

   IF(key > 0) THEN
      return(YES);
   ELSE
      return(NO);
   END_IF
}

/* ======================================================================== */

static short int NextWordLetter(ch,rtn)
char ch;
int *rtn;
  {
   struct letter *lptr;

   lptr = current_letter;
   if (current_letter == NULL)
     return(NOT_TOKEN);
   else if (ch == EOS)
     return(NULL_TOKEN);
   if (lptr->step != NULL)
     lptr = lptr->step;
   else
     return(NOT_TOKEN);
   for ( ; lptr != NULL ; lptr = lptr->next)
     if (lptr->value == ch)
       break;
   if (lptr == NULL)
     return(NOT_TOKEN);
   current_letter = lptr;
   *rtn = current_letter->return_type;
   return(current_letter->key);
  }

/* ------------------------------------------------------------------------ */

static void ResetWordList()
  {
   current_letter = TokenTree;
  }

/* ------------------------------------------------------------------------ */

static int LoadReservedWords()
   {
   int i;

   TokenTree = NewLetter();
   TokenTree->key = NULL_TOKEN;
   for(i = 0; i < NUM_RSV_WORDS; i++)
     {
      Lower(ReservedList[i].word);
      StoreToken(ReservedList[i].word,ReservedList[i].key,ReservedList[i].return_type);
     }
   current_letter = TokenTree;
   return(TRUE);
   }

/* ------------------------------------------------------------------------ */

static void Lower(str)
  char *str;
  {
   int index;

   for (index = 0 ; str[index] != EOS ; index++)
    if ((str[index] >= 'A') && (str[index] <= 'Z'))
      str[index] += CASE_OFFSET;
  }

/* ------------------------------------------------------------------------ */

static void StoreToken(token,key,rtn)
  char *token;
  short int key;
  int rtn;
  {
   int index;
   struct letter *lptr,*prev,*lnode;

   lptr = TokenTree;
   for (index = 0 ; token[index] != EOS ; index++)
     {
      if (lptr->step == NULL)
        {
         lptr->step = NewLetter();
         lptr->step->value = token[index];
         lptr->step->prev = lptr;
         lptr = lptr->step;
        }
      else
        {
         prev = lptr;
         for (lptr = lptr->step ; lptr != NULL ; lptr = lptr->next)
           {
            lnode = lptr;
            if (lptr->value == token[index])
              break;
           }
         if (lptr == NULL)
           {
            lnode->next = NewLetter();
            lnode->next->value = token[index];
            lnode->next->prev = prev;
            lptr = lnode->next;
           }
       }
     }
   if (lptr->key == NOT_TOKEN)
    {
       lptr->key = key;
       lptr->return_type = rtn;
    }
   else
     lptr->key = DUPLICATE_TOKEN;
  }

/* ------------------------------------------------------------------------ */

static struct letter *NewLetter()
  {
   struct letter *lptr;

   if ((lptr = (struct letter *) GenAlloc(sizeof(struct letter),
      "NewLetter")) == NULL)
      MemoryError();
   lptr->value = EOS;
   lptr->prev = NULL;
   lptr->step = NULL;
   lptr->next = NULL;
   lptr->key = NOT_TOKEN;
   return(lptr);
  }

/* =============================================================
 * AInstanceQueryFunction 
 * This function returns YES if the name matches with one of the 
 * instance query function.
 * =============================================================
 */

int AInstanceQueryFunction(name)
char *name;
{
  if(strcmp(name,"any-instancep") == 0)
     return(YES);
  if(strcmp(name,"delayed-do-all-instances") == 0)
     return(YES);
  if(strcmp(name,"do-for-all-instances") == 0)
     return(YES);
  if(strcmp(name,"do-for-instance") == 0)
     return(YES);
  if(strcmp(name,"find-all-instances") == 0)
     return(YES);
  if(strcmp(name,"find-instance") == 0)
     return(YES);
  return(NO);
}