#include <stdio.h>
#include "crsv.h"
#include <string.h>

void       		CheckVariables ();
VAR_PTR    		FindVariable ();
IN_OR_VAR_PTR 	FindAndCreateVariableInTheOrPat();
void       		StoreNotVariable ();
void       		LinkToNotVariableList ();
int 			ProcessLHSVariable();
int 			ProcessRHSVariable();
int             FindVariableInList();
void 			InitVarListInOrPattern();

/* ===========  Variables defined Externally but used here  =========== */

extern VAR_PTR      RetractHead;        /* First Element in Retract list    */
extern VAR_PTR      VariableHead;       /* First Element in Variable list   */
extern VAR_PTR      NegativeVariableHead;   /* First Element in the negation of
                                            variable list                    */
extern RULE_PTR     CurrentRule;            /* Current rule pointer */
extern DR_PTR       DefrelationHead;        /* First Element in Defrelation list*/
extern IN_OR_VAR_PTR InOrVariableList;           /* Variables in the OR pattern */
                                            
                                            

/* -----------------------------
 *  From the file: CRSVTKN.C
 * -----------------------------
 */

extern double  FLOATNUMBER;                    /* gettoken number return  */
extern long    INTNUMBER;
extern char  *TKNWORD;                      /* gettoken string storage */
                                            

/* Functions defined externally, but called here */

extern VAR_PTR    		AllocVariable ();
extern IN_OR_VAR_PTR 	AllocVariableInOrPattern();
extern DEF_FLD 			*CopyFieldInfo ();
extern DA_PTR   		CopyArgumentInfo ();
extern DEF_REL  		*FindDefrelation ();
extern DEF_FLD  		*FindDeffield ();
extern DR_PTR			FindCreateDefrelation();
extern HASH             *AddSymbol();

/* ===========  Variables defined here for internal use  ============== */



/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  FindVariable () 
 *
 *  PURPOSE:  This function will search a variable list for
 *            a specific variable name.
 *
 *  INPUTS :  A pointer to the variable name to search for
 *            (char *), and a pointer to the variable list.
 *
 *  RETURNS:  If the variable name is found in the list, it
 *            returns a pointer to that list node (VAR_PTR), 
 *            otherwise, it returns NULL.
 *
 *  NOTES  :  Used for both retract variables and bound
 *            variable lists.
 * -------------------------------------------------------------
 */

VAR_PTR FindVariable (word, list) 
char    *word;
VAR_PTR   list;
{
   VAR_PTR temp;

   temp = list;
   while (temp NEQ NULL) DO
      IF (strcmp (word, temp->var) EQ 0) THEN
         return (temp);
      END_IF

      temp = temp->next_variable;
   END_WHILE

   return (NULL);
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  CheckVariables () 
 *
 *  PURPOSE:  This function is called after all other rule
 *            processing has been completed and scans through
 *            both the retract variable list and the bound
 *            variable list for problems.
 *
 *  INPUTS :  None.
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Uses global variables for both variable lists.
 * -------------------------------------------------------------
 */

void CheckVariables () 
{
   VAR_PTR  t1, t2;

   /*=====================================*/
   /* Check for bound facts not retracted */
   /*=====================================*/

   t1 = RetractHead;
   while (t1 NEQ NULL) DO
      IF (t1->level_retracted EQ NOT_RETRACTED) THEN
         sprintf (MsgBuf, "Fact bound to <%.40s>, but not retracted.", t1->var);
         ErrorMessage (WARNING, MsgBuf);
      END_IF
      t2 = t1->next_variable;
      t1->next_variable = NULL;
      t1->prev_variable = NULL;
      FreeVariable(t1);
      t1 = t2;
   END_WHILE

   RetractHead  = NULL;

   /*============================*/
   /* Clean up regular variables */
   /*============================*/

   t1 = VariableHead;
   while (t1 NEQ NULL) DO
      IF (t1->count <= 1) THEN
         sprintf (MsgBuf, "Variable <%.40s> bound, but not used.", t1->var);
         ErrorMessage (WARNING, MsgBuf);
      END_IF
      t2 = t1->next_variable;
      t1->next_variable = NULL;
      t1->prev_variable = NULL;
      FreeVariable(t1);
      t1 = t2;
   END_WHILE

   VariableHead = NULL;
}

/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  ProcessLHSVariable () 
 *
 *  PURPOSE:  This function is called when a bound word is
 *            found anywhere on the LHS of a rule. It stores
 *            all the appropriate information about that
 *            variable.
 *
 *  INPUTS :  Four arguments, 
 *             A pointer to the current relation (REL_PTR), 
 *             A flag describing the variable type, either a
 *               BWORD or BWORDS (int), 
 *             A number telling the field position (if any) 
 *               in which the variable was found (int), 
 *             A flag telling if variable was found inside a
 *               bind operation (YES or NO) (int), and
 *             A flag telling if variable was found inside a
 *               not pattern (YES or NO) (int).
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :  Certain operations on the RHS of a rule also call
 *            this, such as bind, since they can create new
 *            variables.
 * -------------------------------------------------------------
 */

int ProcessLHSVariable (CurrentRule,current_rel, current_ex_func, type, field_pos, 
                                 in_bind, in_not, in_or,code, not_flag) 
int     type, field_pos, in_bind, in_not,in_or, code;
short   not_flag;
REL_PTR current_rel;
RULE_PTR CurrentRule;
EX_FUNC *current_ex_func;
{
   VAR_PTR   		temp_rl;
   IN_OR_VAR_PTR 	temp_var;
   HASH            *temp_hash;
   
   /*============================================*/
   /*  Make sure variable isn't bound to a fact  */
   /*============================================*/

   IF ((temp_rl = FindVariable (TKNWORD, RetractHead)) NEQ NULL) THEN
      sprintf (MsgBuf, "Fact address ?%s also used as variable name\n", 
         TKNWORD);
      ErrorMessage (ERROR, MsgBuf);
      return(RESTART);
   ELSE

      /*==========================================*/
      /*  Check if variable has already been used */
      /*==========================================*/

      temp_rl = FindVariable (TKNWORD, VariableHead);

      IF (temp_rl EQ NULL) THEN
        if (not_flag) 
          {
           sprintf (MsgBuf, 
           "Undefined variable  %s referenced", TKNWORD);
           ErrorMessage (ERROR, MsgBuf);
           return(RESTART);
          }
         /*===================================*/
         /*  New Variable, put name on stack  */
         /*===================================*/

         temp_rl       = AllocVariable ();
         temp_hash     = AddSymbol(TKNWORD);
         temp_rl->var  = temp_hash->contents;

         /*====================================*/
         /*  Set values in new variable object */
         /*====================================*/

         temp_rl->next_variable   = VariableHead;
         temp_rl->prev_variable   = NULL;
         if (code == REL) 
           temp_rl->relation      = current_rel;
         else
           temp_rl->ex_func       = current_ex_func;
         temp_rl->type            = type;
         temp_rl->count++;

         IF ((CHECK_DEFRELS IS_ON) AND (type NEQ MULTIPLE) AND
             (field_pos > 0)) THEN
            IF (code == REL) THEN
               temp_rl->field_info = CopyFieldInfo (current_rel->name, field_pos);
            ELSE_IF (code == FUNCTION) 
               temp_rl->field_info = CopyArgumentInfo (current_ex_func->name, 
                                                        field_pos);
            END_IF
         END_IF

         IF (in_not > 0) THEN
            temp_rl->bound_in_not = in_not;
         ELSE
            temp_rl->bound_in_not = NO;
         END_IF

         IF (VariableHead NEQ NULL) THEN
            VariableHead->prev_variable = temp_rl;
         END_IF

         VariableHead = temp_rl;
      ELSE /* temp_rl != NULL */
         if ((not_flag) && (CHECK_DEFRELS IS_ON) && (field_pos > 0)) 
           {
             temp_rl->count += 1;
             StoreNotVariable (current_rel, temp_rl->var, field_pos);
             return(OK);
           }

         /*==========================================================*/
         /*  Variable previously used, check whether single or multi */
         /*==========================================================*/

         IF (type NEQ temp_rl->type) THEN
            sprintf (MsgBuf, 
               "Variable <%.40s> bound to both single and multiple fields.", 
               temp_rl->var);
            ErrorMessage (WARNING, MsgBuf);
         END_IF

         /*==============================*/
         /* Check defrelation field info */
         /*==============================*/

         IF ((CHECK_DEFRELS IS_ON) AND (type NEQ MULTIPLE) AND
                 (field_pos > 0)) THEN
            IF (code EQ REL) THEN
               IF (temp_rl->field_info EQ NULL) THEN
                  temp_rl->field_info = CopyFieldInfo (current_rel->name, 
                     field_pos);
               ELSE
                  CheckVarFieldInfo (LHS, current_rel->name, temp_rl->var, 
                                       temp_rl->field_info, field_pos);
               END_IF
            ELSE_IF (code EQ FUNCTION) 
               IF (temp_rl->field_info EQ NULL) THEN
                  temp_rl->field_info = CopyArgumentInfo (current_ex_func->name, 
                     field_pos);
               ELSE
                  CheckVariableArgumentInfo (current_ex_func->name, temp_rl->var, 
                                          temp_rl->field_info, field_pos, LHS);
               END_IF
            END_IF
         END_IF

         /*=============================================*/
         /*  Check if it was originally bound in a not  */
         /*=============================================*/

         IF (temp_rl->bound_in_not > 0) THEN
            IF ((in_not EQ NO)&&(!in_or)) THEN
	       sprintf(MsgBuf,
                   "Variable <%s> : Can only use variables bound in a NOT inside that NOT.",temp_rl->var);
               ErrorMessage (WARNING, MsgBuf);
            ELSE_IF (CurrentRule->num_nots NEQ temp_rl->bound_in_not) 
               sprintf(MsgBuf,
               " Variable <%s> bound inside a NOT is not constrained by\n        the previous NOT pattern(s) ",
			temp_rl->var);
               ErrorMessage(WARNING,MsgBuf);
           END_IF
         END_IF

         /*=============================================*/
         /*  Don't increment counter if used in a bind  */
         /*=============================================*/

         IF (in_bind NEQ YES) THEN
            temp_rl->count++;
         END_IF
      END_IF /* end if temp_rl == NULL */
       /* ============================================*/
         /* Check if it was in the or pattern           */
         /* ============================================*/
         
      IF(in_or) THEN
           InOrVariableList = FindAndCreateVariableInTheOrPat(InOrVariableList,&temp_var);
           IF(temp_var->check == OFF)THEN
           	     temp_var->check = ON;
           	     temp_var->num += 1;
           END_IF  
           IF(temp_var->var_info == NULL)
                temp_var->var_info = temp_rl;
           END_IF
      END_IF
   return(OK);
}


/* ==================================================================== *
 * 	FindAndCreateVariableInTheOrPat: create a node for a variable in an OR pattern												*
 * ==================================================================== *
 */
 
 IN_OR_VAR_PTR FindAndCreateVariableInTheOrPat(list,temp_var)
 IN_OR_VAR_PTR list,*temp_var;
 {
 	IN_OR_VAR_PTR temp;
 	HASH          *temp_hash;
 	
 	IF(list == NULL)THEN
 	    list = AllocVariableInOrPattern();
 	    temp_hash = AddSymbol(TKNWORD);
 	    list->word = temp_hash->contents;
 	    *temp_var = list;
 	ELSE
 	     temp = list;
 	     WHILE(temp!= NULL)DO
 	         if((strcmp (TKNWORD,temp->word)) == 0)
 	           break;
 	         temp = temp->next;  
 	     END_WHILE
 	     IF (temp == NULL)THEN
 	        temp = AllocVariableInOrPattern();
 	        temp_hash = AddSymbol(TKNWORD);
 	        temp->word = temp_hash->contents;
 	        temp->next = list;
 	        *temp_var = temp;
 	         return(temp);
 	     ELSE
 	        *temp_var = temp;
 	     END_IF
 	 END_IF
 	 return(list);
 }
/* ==================================================================== */
/*   Store the information of the negative variable for latter checking */
/* ==================================================================== */
void StoreNotVariable (CurrentRelation, var_name, position) 
REL_PTR CurrentRelation;
char  *var_name;
int position;

{
  DEF_REL *def_rel;
  DEF_FLD *fld;

  def_rel = FindDefrelation (DefrelationHead, CurrentRelation->name);
  IF(def_rel == NULL) THEN
      DefrelationHead = FindCreateDefrelation (DefrelationHead, CurrentRelation->name);
      sprintf (MsgBuf, 
         "Relation <%.40s> was referenced before it is defined.", CurrentRelation->name);
      ErrorMessage (ERROR, MsgBuf);
      return;
  END_IF
  fld = FindDeffield (position, def_rel->fields);
  IF ((fld == NULL) && (def_rel->created_by == USER_MADE))THEN
       sprintf (MsgBuf, "field %d is not defined.", position);
       ErrorMessage (WARNING, MsgBuf);
       return;
  ELSE_IF(fld == NULL)
      return;
  END_IF
/*===============================================================================*/
/* Store the variable in a list if there is only one value allowed for the field */
/*===============================================================================*/

  IF((fld->allow_word == YES) && (fld->allow_string == NO) &&
  	  (fld->allow_integer == NO) && (fld->allow_float == NO)) THEN
     IF((fld->possible_words != NULL) &&
         (fld->possible_words->next_word == NULL)) THEN
        LinkToNotVariableList (CurrentRelation, var_name, position);
     END_IF
  ELSE_IF((fld->allow_word == NO) && (fld->allow_string == YES) &&
  	  (fld->allow_integer == NO) && (fld->allow_float == NO)) 
     IF ((fld->possible_strings != NULL) &&
         (fld->possible_strings->next_word == NULL)) THEN
        LinkToNotVariableList (CurrentRelation, var_name, position);
     END_IF
  ELSE_IF ((fld->allow_word == NO) && (fld->allow_string == NO) &&
  	  (fld->allow_integer == YES) && (fld->allow_float == NO)) 
     IF((fld->possible_integers != NULL) &&
         (fld->possible_integers->next_number == NULL)) THEN
        LinkToNotVariableList (CurrentRelation, var_name, position);
     END_IF
  ELSE_IF ((fld->allow_word == NO) && (fld->allow_string == NO) &&
  	  (fld->allow_integer == NO) && (fld->allow_float == YES)) 
     IF ((fld->possible_floats != NULL) &&
         (fld->possible_floats->next_number == NULL)) THEN
        LinkToNotVariableList (CurrentRelation, var_name, position);
     END_IF
  END_IF
}
/* ======================================================== *
 * LINK_TO_NOT_VARIABLE_LIST:                               *
 * ======================================================== *
 */

void LinkToNotVariableList (current_rel, var_name, position) 
REL_PTR current_rel;
char  *var_name;
int position;

{
   VAR_PTR temp_var;
   HASH *temp_hash;

        temp_var = AllocVariable ();
        temp_hash = AddSymbol(var_name);
        temp_var->var = temp_hash->contents;
        temp_var->next_variable = NegativeVariableHead;
        temp_var->prev_variable = NULL;
        temp_var->relation = current_rel;
        temp_var->field_info = CopyFieldInfo (current_rel->name, position);
        if (NegativeVariableHead != NULL) 
          NegativeVariableHead->prev_variable = temp_var;
        NegativeVariableHead = temp_var;

}
/* ==================================================================== */

/*
 * -------------------------------------------------------------
 *  NAME   :  ProcessRHSVariable () 
 *
 *  PURPOSE:  This function is called when a bound word is
 *            found anywhere on the RHS of a rule. It stores
 *            all the appropriate information about that
 *            variable.
 *
 *  INPUTS :  Three arguments, 
 *             A flag describing the variable type, either a
 *               BWORD or BWORDS (int), 
 *             A number indicating the field position (if any) 
 *               in which the variable was found (int), 
 *             A flag telling if variable was found inside a
 *               not pattern (YES or NO) (int).
 *
 *  RETURNS:  Nothing.
 *
 *  NOTES  :
 * -------------------------------------------------------------
 */

int ProcessRHSVariable (CurrentRelation,CurrentExtFunc,type, 
			field_pos, in_not,in_or, code,var_in_tmp_list) 
REL_PTR  CurrentRelation;
EXF_PTR  CurrentExtFunc; 
int     type, field_pos, in_not,in_or, code;
FIELD_PTR var_in_tmp_list;
{
   VAR_PTR   temp_rl;
   
   /*============================================*/
   /*  Make sure variable isn't bound to a fact  */
   /*============================================*/

   IF ((temp_rl = FindVariable (TKNWORD, RetractHead)) NEQ NULL) THEN
      sprintf (MsgBuf, "Fact address ?%s also used as variable name\n", 
         TKNWORD);
      ErrorMessage (ERROR, MsgBuf);
      return(RESTART);
   ELSE

      /*==========================================*/
      /*  Check if variable has already been used */
      /*  or it is used in the template pattern   */
      /*==========================================*/
      
      temp_rl = FindVariable (TKNWORD, VariableHead);

      IF (temp_rl EQ NULL) THEN
         IF (!(FindVariableInList(var_in_tmp_list,TKNWORD))) THEN
         /*==================================*/
         /*  New Variable, probably unbound  */
         /*==================================*/

           sprintf (MsgBuf, "Variable <%.40s> used before being bound!", TKNWORD);
           ErrorMessage (ERROR, MsgBuf);
           return(RESTART);
         END_IF
      ELSE

         /*==============================*/
         /* Check defrelation field info */
         /*==============================*/

         IF ((CHECK_DEFRELS IS_ON) AND (type NEQ MULTIPLE) AND
                 (field_pos > 0)) THEN
            IF (code EQ REL) THEN
               IF (temp_rl->field_info EQ NULL) THEN
                  temp_rl->field_info = CopyFieldInfo (CurrentRelation->name, 
                     field_pos);
               ELSE
                  CheckVarFieldInfo (RHS, CurrentRelation->name, temp_rl->var, 
                                       temp_rl->field_info, field_pos);
               END_IF
            ELSE_IF (code EQ FUNCTION) 
               IF (temp_rl->field_info EQ NULL) THEN
                  temp_rl->field_info = CopyArgumentInfo (CurrentExtFunc->name, 
                     field_pos);
               ELSE
                  CheckVariableArgumentInfo (CurrentExtFunc->name, temp_rl->var, 
                                          temp_rl->field_info, field_pos, RHS);
               END_IF
            END_IF
         END_IF


         IF (temp_rl->bound_in_or) THEN
            sprintf(MsgBuf,"Variable <%s> : Due to the OR pattern this variable was not contrained",
                temp_rl->var);
            ErrorMessage(ERROR,MsgBuf);
          END_IF
         /*=============================================*/
         /*  Check if it was originally bound in a not  */
         /*=============================================*/

         IF (temp_rl->bound_in_not > 0) THEN
            IF ((in_not EQ NO) && (!in_or))THEN
               sprintf(MsgBuf,"Variable <%s>: Can only use variables bound in a NOT inside that NOT pattern",
			temp_rl->var);
               ErrorMessage (ERROR, MsgBuf);
            ELSE_IF (CurrentRule->num_nots NEQ temp_rl->bound_in_not) 
	       sprintf(MsgBuf,
				"Variable <%s> bound inside a NOT is not constrained by \n         the previous NOT pattern(s)",
			temp_rl->var);
               ErrorMessage (WARNING,MsgBuf);
            END_IF
         END_IF

         temp_rl->count++;
      END_IF
   END_IF
   return(OK);
}

/* ==================================================================== */


		 /* ================================================ *
          * This function is used  to detect the variable is *
          *  bound in all patterns of the OR pattern on the  *
          *  left hand side of a rule.                       *
          *  variable appears in the OR pattern can be       *
          *  referenced on the RHS if only it appears in all *
          *  the children patterns of the OR pattern.        *
          *  For example :                                   *
          *    (defrule foo									 *
          *       (or (fifi  ?x  ?y)						 *
          *		      (bar   ?x))							 *
          *     =>											 *
          *     (assert (more-bar ?x ?y)))                   *
          *                            						 *
          *    <?x> is valid but not <?y>                    *
          ****************************************************/
 void InitVarListInOrPattern(list,val)
 IN_OR_VAR_PTR list;
 short val;
 { 
    IN_OR_VAR_PTR temp_var;
            
     temp_var = list;
     while(temp_var != NULL)
        {
          temp_var->check = val;
          temp_var = temp_var->next;
         }
  }