   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                  ANALYSIS MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
#define _ANALYSIS_SOURCE_

#include "setup.h"

#if (! RUN_TIME) && (! BLOAD_ONLY) && DEFRULE_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_ 

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "reorder.h"
#include "variable.h"
#include "analysis.h"
#include "generate.h"
#include "utility.h"
#include "router.h"

#if DEFTEMPLATE_CONSTRUCT
#include "deftempl.h"
#endif

#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif

#define ANALYSIS_SWITCH 0
   
/****************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS  */
/****************************************/

#if ANSI_COMPILER  
   static VOID                    BuildNetworkExpressions(struct node *);
   static VOID                    FieldConversion(struct node *,int,int,struct field_tests *);
   static VOID                    ExtractAnds(struct node *,int,int,int,struct field_tests *);
   static int                     AllVariablesInPattern(struct node *,int,int); 
   static struct var_info        *GetVariables(struct node *,int);
   static BOOLEAN                 CheckFactAddress(struct symbolHashNode *,int);
   static BOOLEAN                 CheckPattern(struct node *,int,int);
#else
   static VOID                    BuildNetworkExpressions();
   static VOID                    FieldConversion();
   static VOID                    ExtractAnds();
   static int                     AllVariablesInPattern(); 
   static struct var_info        *GetVariables();
   static BOOLEAN                 CheckFactAddress();
   static BOOLEAN                 CheckPattern();
#endif

/****************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS  */
/****************************************/

   static int                DeftemplatePattern;
   
/***********************************/
/* RuleAnalysis:  Analyzes a rule. */
/***********************************/
globle BOOLEAN RuleAnalysis(rule)
  struct node *rule;
  {
   VariableAnalysis(rule);
   if (CheckVariables(rule) == CLIPS_TRUE) return(CLIPS_TRUE);
   BuildNetworkExpressions(rule);
   return(CLIPS_FALSE);
  }

/************************************/
/* BuildNetworkExpressions:         */
/************************************/
static VOID BuildNetworkExpressions(rule)
  struct node *rule;
  {
   int pattern, element;
   struct expr *top_jn;
   struct node *field_ptr;
   struct field_tests rtn_expr;
   struct pn_test *new_fs, *last_fs, *top_fs;
   struct expr_info *next_pat, *last_pat = NULL, *top_pat = NULL;
   struct expr *top_test, *last_test, *new_test;
   
   FlushAnalysisExpressions();
   
   if (rule->type != PATTERN)
     {
      CLIPSSystemError("ANALYSIS",1);
      ExitCLIPS(4);
     }
     
   pattern = 1;
   while (rule != NULL)
     {
      DeftemplatePattern = PatternHasTemplate(pattern);
      
      next_pat = get_struct(expr_info);
      next_pat->pattern = pattern;
      if (rule->negated) next_pat->logic = NEGATIVE;
      else next_pat->logic  = POSITIVE;
      next_pat->next = NULL;
      next_pat->otest = NULL;
      next_pat->pntl = NULL;
      next_pat->jnt = NULL;
      
      /*==============================*/
      /* Get the pattern information. */
      /*==============================*/
      
      element = 1;
      top_jn = NULL;
      top_fs = last_fs = NULL;
      field_ptr = rule->right;
      while (field_ptr != NULL)
        {
         FieldConversion(field_ptr,pattern,element,&rtn_expr);
         
         new_fs = get_struct(pn_test);
         new_fs->element = element;
         if ((field_ptr->type == SINGLE) || (field_ptr->type == BWORD))
           { new_fs->type = SINGLE; }
         else if ((field_ptr->type == MULTIFIELD) || 
                  (field_ptr->type == BWORDS))
           { new_fs->type = MULTIFIELD; }
         else
           { 
            CLIPSSystemError("ANALYSIS",2);
            ExitCLIPS(4);
           }
         new_fs->pnt = rtn_expr.pnt;
         new_fs->next = NULL;
         
         if (last_fs == NULL)
           { top_fs = new_fs; }
         else
           { last_fs->next = new_fs; }
         last_fs = new_fs;
         
         
         top_jn = CombineExpressions(top_jn,rtn_expr.jnt);
         
         field_ptr = field_ptr->right;
         element++;
        }
      
      /*==============================*/
      /* Get any test information. */
      /*==============================*/  
      
      top_test = last_test = NULL;
      rule = rule->bottom;
      while ((rule != NULL) ? (rule->type == PAT_TEST) : CLIPS_FALSE)
        {
         new_test = rule->expression;
         rule->expression = NULL;
         GetvarReplace(new_test,pattern,element,0);
         if (last_test == NULL)
           { top_test = new_test; }
         else
           { last_test->next_arg = new_test; }
           
         last_test = new_test;
         rule = rule->bottom;
        } 
        
      if (top_test != NULL)
        {
         if (top_test->next_arg != NULL)
           {
            new_test = GenAnd();
            new_test->arg_list = top_test;
            top_test = new_test;
           }
         
         if (next_pat->logic == NEGATIVE)
           { next_pat->otest = top_test; }
         else
           { top_jn = CombineExpressions(top_jn,top_test); }
        }   
        
      next_pat->jnt = top_jn;
      next_pat->pntl = top_fs;
      
      if (last_pat == NULL)
        { top_pat = next_pat; }
      else 
        { last_pat->next = next_pat; }
      last_pat = next_pat;
      
#if ANALYSIS_SWITCH 
     
      PrintCLIPS(WDIALOG,"Pattern ");
      PrintLongInteger(WDIALOG,(long int) pattern);
      PrintCLIPS(WDIALOG,"\n");
      PrintCLIPS(WDIALOG," Join test: ");
      PrintExpression(WDIALOG,next_pat->jnt);
      PrintCLIPS(WDIALOG,"\n");
      PrintCLIPS(WDIALOG," Other tests: ");
      PrintExpression(WDIALOG,next_pat->otest);
      PrintCLIPS(WDIALOG,"\n");
      
      while (top_fs != NULL)
        {
         PrintCLIPS(WDIALOG," Field Pattern Test ");
         PrintLongInteger(WDIALOG,(long int) top_fs->element);
         PrintCLIPS(WDIALOG,": ");
         PrintExpression(WDIALOG,top_fs->pnt);
         PrintCLIPS(WDIALOG,"\n");
         top_fs = top_fs->next;
        }
      

#endif
      
      pattern++;
     }
   
   SetRuleInformation(top_pat);
  }
   
/*******************************************************************/
/* FIELD_CONVERSION: Generates join and pattern tests for a field. */
/*******************************************************************/
static VOID FieldConversion(pat_field,pattern,element,rtn_vals)
  struct node *pat_field;
  int pattern, element;
   struct field_tests *rtn_vals;
  {
   struct field_tests test_vals;
   int tests_in_pn = CLIPS_TRUE;
   struct node *pat_ptr;
   struct expr *top_pn, *top_jn, *last_pn, *last_jn, *new_node;
   struct expr *bind_test;
   int logic;
   
   if (pat_field == NULL)
     { 
      CLIPSSystemError("ANALYSIS",3);
      ExitCLIPS(4);
     }
   
   /*=================================================*/
   /* Determine if constant testing must be performed */
   /* in the join network. Only possible when a field */
   /* contains an or ('|') and references are made to */
   /* variables outside the pattern.                  */
   /*=================================================*/
   
   if (pat_field->bottom != NULL)
     { 
      if (pat_field->bottom->bottom != NULL)
        { tests_in_pn = AllVariablesInPattern(pat_field->bottom,pattern,element); }
     }
   
   /*======================================*/
   /* Extract pattern network expressions. */
   /*======================================*/
   
   top_pn = last_pn = NULL;
   top_jn = last_jn = NULL;
    
   pat_ptr = pat_field->bottom;
   while (pat_ptr != NULL)
     {
      ExtractAnds(pat_ptr,pattern,element,tests_in_pn,&test_vals);
      
      if (test_vals.pnt != NULL)
        {
         if (last_pn == NULL)
           { top_pn = test_vals.pnt; }
         else
           { last_pn->next_arg = test_vals.pnt; }
         last_pn = test_vals.pnt;
        }
        
      if (test_vals.jnt != NULL)
        {
         if (last_jn == NULL)
           { top_jn = test_vals.jnt; }
         else
           { last_jn->next_arg = test_vals.jnt; }
         last_jn = test_vals.jnt;
        }
        
      pat_ptr = pat_ptr->bottom;
     }
  
   if ((top_pn != NULL) ? (top_pn->next_arg != NULL) : CLIPS_FALSE)
     {
      new_node = GenOr();
      new_node->arg_list = top_pn;
      top_pn = new_node;
     }
     
   if ((top_jn != NULL) ? (top_jn->next_arg != NULL) : CLIPS_FALSE)
     {
      new_node = GenOr();
      new_node->arg_list = top_jn;
      top_jn = new_node;
     }
 
   /*=============================================*/
   /* Attach expression for the binding instance. */
   /*=============================================*/
   
   if ((pat_field->type == BWORDS) || (pat_field->type == BWORD))
     {
      if (pat_field->negated) logic = NEGATIVE;
      else logic = POSITIVE;
      if (FindVariable(pat_field->value,
                            pattern,pattern,element,INSIDE) != NULL)
        { 
         bind_test = GenPNVariableComparison(logic,pat_field->value,pattern,element); 
         top_pn = CombineExpressions(bind_test,top_pn);
        }
      else if (FindVariable(pat_field->value,
                             1,pattern,element,INSIDE) != NULL)
        { 
         bind_test = GenJNVariableComparison(logic,pat_field->value,pattern,element,0); 
         top_jn = CombineExpressions(bind_test,top_jn);
        }
     }
   
      
   rtn_vals->pnt = top_pn;
   rtn_vals->jnt = top_jn;
  }
  
/****************************************************************/
/* EXTRACT_ANDS:   */
/****************************************************************/
static VOID ExtractAnds(and_field,pattern,element,test_in_pn,test_ptr)
  struct node *and_field;
  int pattern, element, test_in_pn;
  struct field_tests *test_ptr;
  {
   struct expr *new_node;
   struct expr *new_pn, *new_jn;
   struct expr *top_pn = NULL, *top_jn = NULL;
   struct expr *last_pn = NULL, *last_jn = NULL;
   int logic;
   
   /*=========================================================*/
   /* Loop through the list of field constraints and generate */
   /* test for those that can have pattern network tests.     */
   /*=========================================================*/
   
   while (and_field != NULL)
     {
      if (and_field->negated) logic = NEGATIVE;
      else logic = POSITIVE;
      
      new_pn = NULL;
      new_jn = NULL;
      if ((and_field->type == STRING) || 
          (and_field->type == SYMBOL) ||
          (and_field->type == FLOAT) ||
#if OBJECT_SYSTEM
          (and_field->type == INSTANCE_NAME) ||
#endif
          (and_field->type == INTEGER))
        { 
         if (test_in_pn == CLIPS_TRUE)
           { new_pn = GenPNConstant(logic,and_field->type,and_field->value); }
         else
           { new_jn = GenJNConstant(logic,and_field->type,and_field->value,pattern,element); }
        }
      else if (and_field->type == COAMP)
        {
         if ((test_in_pn == CLIPS_TRUE) &&
             (CheckExpression(and_field->expression,pattern,pattern,
                              element + 1,INSIDE) == NULL))
           { 
            new_pn = GenPNColon(logic,and_field->expression,pattern,element); 
            and_field->expression = NULL;
           }
         else
           { 
            new_jn = GenJNColon(logic,and_field->expression,pattern,element,DeftemplatePattern); 
            and_field->expression = NULL;
           }
        }
      else if ((and_field->type == FCALL)
#if DEFGENERIC_CONSTRUCT
               || (and_field->type == GCALL)
#endif
#if DEFFUNCTION_CONSTRUCT
               || (and_field->type == PCALL)
#endif
              )
        {
         if ((test_in_pn == CLIPS_TRUE) &&
             (CheckExpression(and_field->expression,pattern,pattern,
                              element + 1,INSIDE) == NULL))
           { 
            new_pn = GenPNEq(logic,and_field->expression,pattern,element); 
            and_field->expression = NULL;
           }
         else
           { 
            new_jn = GenJNEq(logic,and_field->expression,pattern,element,DeftemplatePattern); 
            and_field->expression = NULL;
           }
        }
      else if ((and_field->type == BWORD) || (and_field->type == BWORDS))
        {
         if ((test_in_pn == CLIPS_TRUE) &&
             (FindVariable(and_field->value,
                            pattern,pattern,element+1,INSIDE) != NULL))
           { new_pn = GenPNVariableComparison(logic,and_field->value,pattern,element); }
         else
           { new_jn = GenJNVariableComparison(logic,and_field->value,pattern,element,DeftemplatePattern); }
        }
        
      if (new_pn != NULL)
        {
         if (last_pn == NULL)
           { top_pn = new_pn; }
         else
           { last_pn->next_arg = new_pn; }
         last_pn = new_pn;
        }
        
      if (new_jn != NULL)
        {
         if (last_jn == NULL)
           { top_jn = new_jn; }
         else
           { last_jn->next_arg = new_jn; }
         last_jn = new_jn;
        }
        
      and_field = and_field->right;
     }
     
   /*===================================================*/
   /* Determine if an and function call must be wrapped */
   /* around the arguments.                             */
   /*===================================================*/
   
   if ((top_pn != NULL) ? (top_pn->next_arg != NULL) : CLIPS_FALSE)
     {
      new_node = GenAnd();
      new_node->arg_list = top_pn;
      top_pn = new_node;
     }
     
   if ((top_jn != NULL) ? (top_jn->next_arg != NULL) : CLIPS_FALSE)
     {
      new_node = GenAnd();
      new_node->arg_list = top_jn;
      top_jn = new_node;
     }
     
   test_ptr->pnt = top_pn;
   test_ptr->jnt = top_jn;
  }
  
/****************************************************************/
/* AllVariablesInPattern: Determines if all of the variable references */
/*   in a field can be referenced within the pattern.           */
/****************************************************************/
static int AllVariablesInPattern(or_field,pattern,element)
  struct node *or_field;
  int pattern, element;
  {
   struct node *and_field;
   
   while (or_field != NULL)
     {
      and_field = or_field;
      while (and_field != NULL)
        {
         if ((and_field->type == BWORD) || (and_field->type == BWORDS))
           {
            if (FindVariable(and_field->value,
                           pattern,pattern,element+1,INSIDE) == NULL)
              { return(CLIPS_FALSE); }
           }
         else if ((and_field->type == COAMP) ||
#if DEFGENERIC_CONSTRUCT
                  (and_field->type == GCALL) ||
#endif
#if DEFFUNCTION_CONSTRUCT
                  (and_field->type == PCALL) ||
#endif
                  (and_field->type == FCALL)) 
           {
            if (CheckExpression(and_field->expression,pattern,pattern,
                              element + 1,INSIDE) != NULL)
              { return(CLIPS_FALSE); }
           }
           
         and_field = and_field->right;
        }
      or_field = or_field->bottom;
     }
     
   return(CLIPS_TRUE);
  }
  
/*#####################################*/
/* FUNCTIONS FOR GETTING THE VARIABLES */
/*#####################################*/
  
/**************************************************************/
/* VariableAnalysis:  For each pattern, determines and stores */
/*   the following information: pattern position, type of     */
/*   pattern (regular or not), fact-address (if any).         */
/**************************************************************/
globle VOID VariableAnalysis(pat_list)
  struct node *pat_list;
  {
   struct pat_info *new_pat, *head = NULL, *last_pat = NULL;
   int count = 1;
   
   FlushVariableAnalysis();
   while (pat_list != NULL)
     {
      if (pat_list->type == PATTERN)
        {    
         new_pat = get_struct(pat_info);
         new_pat->position = count;
         if (pat_list->negated) new_pat->logic = NEGATIVE;
         else new_pat->logic = POSITIVE;
         new_pat->fact_address = (SYMBOL_HN *) pat_list->value;         
         new_pat->variables = GetVariables(pat_list->right,count);
         new_pat->next = NULL;
         
         /*================================*/
         /* Save the relation information. */
         /*================================*/
         
         if ((pat_list->right->type != SINGLE) || (pat_list->right->bottom == NULL))
           { new_pat->relation = NULL; }
         else if ((pat_list->right->bottom->type != SYMBOL) ||
                  (pat_list->right->bottom->right != NULL) ||
                  (pat_list->right->bottom->bottom != NULL))
           { new_pat->relation = NULL; }
         else
           { new_pat->relation = (SYMBOL_HN *) pat_list->right->bottom->value; }
         
         /*==================*/
         /* Add to the list. */
         /*==================*/
         
         if (last_pat == NULL)
           { head = new_pat; }
         else
           { last_pat->next = new_pat; }
        
         last_pat = new_pat;
         count++;
        }

      pat_list = pat_list->bottom;
     }

   SetVariableInformation(head);
  }
  
/*************************************************************/
/* GetVariables:       */
/*************************************************************/
static struct var_info *GetVariables(pat_fields,position)
  struct node *pat_fields;
  int position;
  { 
   int elem_count = 1;
   struct var_info *head = NULL, *last_var = NULL, *new_var;
  
   while (pat_fields != NULL)
     {
      if ((pat_fields->type == BWORD) || 
          (pat_fields->type == BWORDS))
        {    
         new_var = get_struct(var_info);
         new_var->name = (SYMBOL_HN *) pat_fields->value;
         new_var->pattern = position;
         new_var->element = elem_count;
         new_var->next = NULL;
         
         if (last_var == NULL)
           { head = new_var; }
         else
           { last_var->next = new_var; }
        
         last_var = new_var;
        }
      
      elem_count++;
      pat_fields = pat_fields->right;
     }
     
   return(head);
  }
  
/*###################################*/
/* FUNCTIONS FOR CHECKING VARIABLES. */
/*###################################*/

/*****************************************************/
/* CheckVariables: Verifies proper use of variables. */
/*****************************************************/
globle int CheckVariables(pat_list)
  struct node *pat_list;
  {
   int error_flag = CLIPS_FALSE;
   struct pat_info *temp_pat;
   int pattern, element;
   struct expr *rv;
   
   /*=====================================================*/
   /* Check that fact-addresses are not reused or used as */
   /* variables within patterns.                          */
   /*=====================================================*/
   
   temp_pat = GetVariableInformation();
   while (temp_pat != NULL)
     {
      if (temp_pat->fact_address != NULL)
        { 
         if (CheckFactAddress(temp_pat->fact_address,
                                         temp_pat->position) == CLIPS_TRUE)
           { error_flag = CLIPS_TRUE; }
        }
      temp_pat = temp_pat->next;
     }
     
   /*===========================================*/
   /* Check that variables used in patterns are */
   /* referenced properly.                      */         
   /*===========================================*/
   
   pattern = 0;
   temp_pat = GetVariableInformation();
   while (pat_list != NULL)
     {
      element = 1;
      if (pat_list->type == PATTERN)
        {
#if DEFTEMPLATE_CONSTRUCT 
         if (temp_pat->relation == NULL)
           { DeftemplatePattern = CLIPS_FALSE; }
         else if (FindDeftemplate(temp_pat->relation->contents) != NULL)
           { DeftemplatePattern = CLIPS_TRUE; }
         else
           { DeftemplatePattern = CLIPS_FALSE; }
         
#else
         DeftemplatePattern = CLIPS_FALSE;
#endif 

         pattern++;
         if (CheckPattern(pat_list->right,pattern,element) == CLIPS_TRUE)
           { error_flag = CLIPS_TRUE; }
         temp_pat = temp_pat->next;
        }
      else if (pat_list->type == PAT_TEST)
        {
         if ((rv = CheckExpression(pat_list->expression,1,pattern,element,OUTSIDE))
             != NULL)
           { 
            PrintCLIPS(WERROR,"\nUndefined variable ");
            PrintCLIPS(WERROR,ValueToString(rv->value));
            PrintCLIPS(WERROR," referenced in test conditional element:\n");
            PrintExpression(WERROR,pat_list->expression);
            PrintCLIPS(WERROR,"\n");
            error_flag = CLIPS_TRUE; 
           }
        }
      else
        {
         CLIPSSystemError("VARIABLE",1);
         ExitCLIPS(4);
        }
   
      pat_list = pat_list->bottom;
     }
   
   return(error_flag);
  }
  
/************************************************************/
/* CheckFactAddress: */
/************************************************************/
static BOOLEAN CheckFactAddress(name,position)
  SYMBOL_HN *name;
  int position;
  {
   struct pat_info *temp_pat;
   struct var_info *temp_var;
   int error_flag = CLIPS_FALSE;
   
   temp_pat = GetVariableInformation();
   while (temp_pat != NULL)
     {
      if ((temp_pat->fact_address == name) && 
          (temp_pat->position > position))
        { 
         PrintCLIPS(WERROR,"\nDuplicate fact-address ?");
         PrintCLIPS(WERROR,ValueToString(name));
         PrintCLIPS(WERROR," used\n");
         error_flag = CLIPS_TRUE;
        }
        
      temp_var = temp_pat->variables;
      while (temp_var != NULL)
        {
         if (temp_var->name == name)
           { 
            PrintCLIPS(WERROR,"\nFact-address ?");
            PrintCLIPS(WERROR,ValueToString(name));
            PrintCLIPS(WERROR," also used as variable name\n");
            error_flag = CLIPS_TRUE;
           }
         temp_var = temp_var->next;
        }
      
      temp_pat = temp_pat->next;
     }
     
   return(error_flag);
  }
  
/************************************************************/
/* CheckPattern: */
/************************************************************/
static BOOLEAN CheckPattern(pat_fields,pattern,element)
  struct node *pat_fields;
  int pattern, element;
  {
   int error_flag = CLIPS_FALSE;
   struct node *and_field, *or_field;
   int end;
   struct expr *rv;
   
   while (pat_fields != NULL)
     {    
      /*======================================================*/
      /* Check that the field variables are previously bound. */
      /*======================================================*/
      
      or_field = pat_fields->bottom;
      while (or_field != NULL)
        {
         and_field = or_field;
         while (and_field != NULL)
           { 
            if ((and_field->type == BWORD) || (and_field->type == BWORDS))
              {
               if (DeftemplatePattern) end = -1;
               else end = element + 1;
                 
               if (FindVariable(and_field->value,1,pattern,end,INSIDE) == NULL)
                 {
                  PrintCLIPS(WERROR,"\nVariable name ");
                  PrintCLIPS(WERROR,ValueToString(and_field->value));
                  PrintCLIPS(WERROR," was referenced in pattern #");
                  PrintLongInteger(WERROR,(long int) pattern);
                  PrintCLIPS(WERROR," field #");
                  PrintLongInteger(WERROR,(long int) element);
                  PrintCLIPS(WERROR," before it is defined\n");
                  error_flag = CLIPS_TRUE;
                 }
              }
            else if ((and_field->type == COAMP) ||
#if DEFGENERIC_CONSTRUCT
                     (and_field->type == GCALL) ||
#endif
#if DEFFUNCTION_CONSTRUCT
                     (and_field->type == PCALL) ||
#endif
                     (and_field->type == FCALL)) 
              {
               if (DeftemplatePattern) end = -1;
               else end = element + 1;
               
               if ((rv = CheckExpression(and_field->expression,1,pattern,end,
                                   INSIDE)) != NULL)
                 {
                  PrintCLIPS(WERROR,"\nVariable name ");
                  PrintCLIPS(WERROR,ValueToString(rv->value));
                  PrintCLIPS(WERROR," was referenced in pattern #");
                  PrintLongInteger(WERROR,(long int) pattern);
                  PrintCLIPS(WERROR," field #");
                  PrintLongInteger(WERROR,(long int) element);
                  PrintCLIPS(WERROR," before it is defined\n");
                  error_flag = CLIPS_TRUE;
                 }
              }
            and_field = and_field->right;
           }
         or_field = or_field->bottom;
        }
        
      /*========================*/
      /* Move to the next field */
      /*========================*/
      
      element++;
      pat_fields = pat_fields->right;
     }
     
   return(error_flag);
  }
     
/************************************************************/
/* CheckExpression */
/************************************************************/
globle struct expr *CheckExpression(test_ptr,start,pattern,element,location)
  struct expr *test_ptr;
  int start, pattern, element, location;
  {
   struct expr *rv;
   int fap;
   
   while (test_ptr != NULL)
     {
      if ((test_ptr->type == BWORD) || 
          (test_ptr->type == BWORDS))
        {
         if (FindVariable(test_ptr->value,start,pattern,element,location) == NULL)
           { 
            fap = GetFactAddressPosition(test_ptr->value);
            if ((fap == 0) || 
                (fap < start) ||
                ((location == INSIDE) && (fap >= pattern)) ||
                ((location == OUTSIDE) && (fap > pattern)))
              { return(test_ptr); }
           }
        }
#if DEFGLOBAL_CONSTRUCT
      else if (test_ptr->type == GBWORD)
        {
         if (QFindDefglobal((SYMBOL_HN *) test_ptr->value) < 0) return(test_ptr);
        }
#endif
      else if (((test_ptr->type == FCALL)
#if DEFGENERIC_CONSTRUCT
             || (test_ptr->type == GCALL)
#endif
#if DEFFUNCTION_CONSTRUCT
             || (test_ptr->type == PCALL)
#endif
         ) && (test_ptr->arg_list != NULL))
        { 
         if ((rv = CheckExpression(test_ptr->arg_list,start,pattern,
                              element,location)) != NULL)
           { return(rv); }
        }
      test_ptr = test_ptr->next_arg;
     }
   
   return(NULL);
  }

/***************************************************************/
/* LogicalAnalysis:  Analyzes the use of the logical construct */
/*   within the LHS of a rule.                                 */
/***************************************************************/
globle int LogicalAnalysis(patternList)
  struct node *patternList;
  {
   int firstLogical, logicalsFound = CLIPS_FALSE, logicalJoin = 1;
   int gap = CLIPS_FALSE;
   
   firstLogical = patternList->logical;
   
   while (patternList != NULL)
     {
      if (patternList->type == PATTERN)
        {    
         if (patternList->logical) 
           {
            if (! firstLogical)
              {
               PrintCLIPS(WERROR,"\nLogical CEs must be placed first in a rule\n");
               return(-1);
              }
              
            if (gap)
              {
               PrintCLIPS(WERROR,"\nGaps may not exist between logical CEs\n");
               return(-1);
              }
              
            logicalJoin++;
            logicalsFound = CLIPS_TRUE;
           }
         else gap = CLIPS_TRUE;
        }
        
      patternList = patternList->bottom;
     }
   
   if (logicalsFound) return(logicalJoin);
   
   return(0);
  }
#endif
