   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                     BUILD MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _BUILD_SOURCE_

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_

#include "constant.h"
#include "clipsmem.h"
#include "router.h"
#include "expressn.h"
#include "constrct.h" 
#include "network.h"
#include "variable.h"
#include "generate.h"
#include "utility.h" 
#include "reteutil.h" 
#include "drive.h"
#include "build.h"

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER 
#if (! RUN_TIME) && (! BLOAD_ONLY) 
   static VOID                    RemoveIntranetworkLink(struct joinNode *);
   static VOID                    DetachPattern(struct patternNode *);
   static struct joinNode        *ReuseJoin(struct joinNode *,int,int,struct expr *,
                                             struct expr *,struct joinNode *);
   static struct patternNode     *PlacePattern(struct patternNode *,struct patternNode *,
                                                int,int,int,struct patternNode **);
#endif
#else
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static VOID                    RemoveIntranetworkLink();
   static VOID                    DetachPattern();
   static struct joinNode        *ReuseJoin();
   static struct patternNode     *PlacePattern();
#endif
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct patternNode   *PatternNetworkPointer = NULL;
      
#if (! RUN_TIME) && (! BLOAD_ONLY)

/********************************************************************/
/* ConstructJoins:  Takes the logical representation of a rule and */
/*   builds the set of linked joins which represents the left hand  */
/*   side of the rule.  Returns the last join constructed which     */
/*   functions as a link between the join net and the right hand    */
/*   side actions of the rule.                                      */
/********************************************************************/
globle struct joinNode *ConstructJoins(top_pat_node)
  struct patternNode **top_pat_node;
  {
   struct joinNode *last_join, *join;
   struct patternNode *last_pat;
   int first_join = CLIPS_TRUE, negatedRHSPattern = CLIPS_FALSE;
   char logic = '+';
   int try_to_reuse = CLIPS_TRUE;
   struct joinNode *list_of_joins;
   struct joinNode *old_join;
   int total_joins, join_number;
   struct expr *join_test, *not_test;

   last_join = NULL;
   
   /*=======================================================*/
   /* Process each of the patterns and/or test constructs   */
   /* of the rule.  At this point, there should be no lower */
   /* level 'and' or 'or' logic structures.                 */
   /*=======================================================*/ 
   
   total_joins = CountJoins();
   join_number = 1;
   
   while (join_number <= total_joins)
     {
      logic = GetJoinLogic(join_number);
      if (logic == '?') 
        {
         CLIPSSystemError("BUILD",1);
         ExitCLIPS(4);
        }
      else if (logic == '-')
        { negatedRHSPattern = CLIPS_TRUE; }
      else
        { negatedRHSPattern = CLIPS_FALSE; }

      last_pat = PlacePattern(*top_pat_node,NULL,
                               join_number,1,CountPatternFields(join_number),top_pat_node);

      if (first_join == CLIPS_TRUE)
        { list_of_joins = last_pat->entry_join; }
      else
        { list_of_joins = last_join->next_level; }
        
      join_test = GetPrimaryJoinExpression(join_number);
      not_test = GetNotJoinExpression(join_number);

      if ((try_to_reuse == CLIPS_TRUE) && 
          ((old_join = ReuseJoin(list_of_joins,first_join,negatedRHSPattern,join_test,
                                  not_test,last_pat->entry_join)) != NULL) ) 
        {
#if DEBUGGING_FUNCTIONS
         if ((GetWatchItem("compilations") == CLIPS_TRUE) && GetPrintWhileLoading())
           { PrintCLIPS(WDIALOG,"=j"); }
#endif
         last_join = old_join;
         ReturnExpression(join_test);
         ReturnExpression(not_test);
        }
      else
        {
         try_to_reuse = CLIPS_FALSE;

#if DEBUGGING_FUNCTIONS
         if ((GetWatchItem("compilations") == CLIPS_TRUE) && GetPrintWhileLoading())
           { PrintCLIPS(WDIALOG,"+j"); }
#endif
         join = get_struct(joinNode);
         join->beta = NULL;
         join->eval = NULL;
         join->not_eval = NULL;
         join->entry_pat = last_pat;
         join->next_level = NULL;
         join->last_level = last_join;
         join->right_drive_node = NULL;
         join->right_match_node = NULL;
         join->firstJoin = first_join;
         join->LHSPatternIsNegated = negatedRHSPattern;
         join->initialize = GetIncrementalReset();
         join->lastJoin = CLIPS_FALSE;
         join->id = 0;
         join->depth = join_number;

         if (! first_join)
           {
            join->right_drive_node = last_join->next_level;
            last_join->next_level = join;     
           }

         /*======================================================*/
         /* Connect the pattern net to the join net for a single */
         /* pattern of a rule.                                   */
         /*======================================================*/
   
         join->right_match_node = last_pat->entry_join;
         last_pat->entry_join = join;
         last_join = join;

         ExpressionInstall(join_test);
         join->eval = PackExpression(join_test);
         ReturnExpression(join_test);
         ExpressionInstall(not_test); 
         join->not_eval = PackExpression(not_test);
         ReturnExpression(not_test);
        }

      /*===============================================*/
      /* Point to the next pattern and the expressions */
      /* associated with the join for that pattern.    */
      /*===============================================*/
          
      join_number++;
      first_join = CLIPS_FALSE;
     }

#if DEBUGGING_FUNCTIONS   
   if ((GetWatchItem("compilations") == CLIPS_TRUE) && GetPrintWhileLoading())
     { PrintCLIPS(WDIALOG,"\n"); }
#endif

   return(last_join);
  }

/**********************************************************************/
/* ReuseJoin:                                                        */
/**********************************************************************/
static struct joinNode *ReuseJoin(test_join,first_join,negatedRHS,
                                    join_test,not_test,pat_join)
  struct joinNode *test_join, *pat_join;
  int first_join, negatedRHS;
  struct expr *join_test, *not_test;
  {
   struct joinNode *join_ptr;

   while (test_join != NULL)
     {
      if ((test_join->firstJoin == first_join) &&
          (test_join->LHSPatternIsNegated == negatedRHS) &&
          (IdenticalExpression(test_join->eval,join_test) == CLIPS_TRUE) &&
          (IdenticalExpression(test_join->not_eval,not_test) == CLIPS_TRUE) )
        {
         join_ptr = pat_join;
         while (join_ptr != NULL)
           {
            if (join_ptr == test_join) 
              { return (test_join); }
            join_ptr = join_ptr->right_match_node;
           }
        }
        
      if (first_join)
        { test_join = test_join->right_match_node; }
      else
        { test_join = test_join->right_drive_node; }
     }

   return (NULL);
  }

/*****************************************/
/* PlacePattern                          */
/*****************************************/
static struct patternNode *PlacePattern(pattern_list,upper_level,
                                      pat,elem,last_elem,top_pat_node)
  struct patternNode *pattern_list;
  struct patternNode *upper_level;
  int pat, elem, last_elem;
  struct patternNode **top_pat_node;
  {
   struct expr *basic_list;
   struct patternNode *cur_elem, *new_elem;
   struct patternNode *last_look = NULL;
   int same_type, cn_test;
   int pn_type;

   /*=================================================*/
   /* Check for a match on this level of the pattern  */
   /* network.                                        */
   /*=================================================*/

   cur_elem = pattern_list;
   basic_list = GetPatternExpression(pat,elem);
      
   while (cur_elem != NULL)
     {
      if ((cur_elem->stopNode) && (elem > last_elem))
        { return(cur_elem); }
        
      pn_type = GetNodeType(pat,elem);
      if (pn_type == KUNKNOWN)
        {
         CLIPSSystemError("BUILD",2);
         ExitCLIPS(4);
        }

      if (((cur_elem->singlefieldNode == CLIPS_TRUE) && (pn_type == SINGLE)) ||
          ((cur_elem->multifieldNode == CLIPS_TRUE) && (pn_type == MULTIFIELD)) ||
          ((cur_elem->stopNode == CLIPS_TRUE) && (pn_type == STOP)))
        { same_type = CLIPS_TRUE; }
      else
        { same_type = CLIPS_FALSE; }
        
        
      if ((same_type == CLIPS_TRUE) &&
          (IdenticalExpression(cur_elem->eval,basic_list) == CLIPS_TRUE))
        {
         ReturnExpression(basic_list);   
         cur_elem = PlacePattern(cur_elem->next_level,cur_elem,
                                  pat, elem+1,last_elem,top_pat_node);
         return(cur_elem);         
        }
      else
        {
         last_look = cur_elem; 
         cur_elem = cur_elem->right_node;
        }
     }
   
   cn_test = CLIPS_FALSE;
   if (basic_list != NULL)
     {
      if (basic_list->value == PTR_CONSTANT)
        { cn_test = CLIPS_TRUE; }
     }

   /*==================================================*/
   /* Pattern node not found.  Need to add new pattern */
   /* node into the pattern network.                   */
   /*==================================================*/

   new_elem = get_struct(patternNode);
   new_elem->singlefieldNode = CLIPS_FALSE;
   new_elem->multifieldNode = CLIPS_FALSE;
   new_elem->stopNode = CLIPS_FALSE;
   new_elem->blocked = CLIPS_FALSE;
   new_elem->initialize = GetIncrementalReset();
   new_elem->last_level = upper_level;
   new_elem->next_level = NULL;
   new_elem->right_node = NULL;
   new_elem->left_node = NULL;
   new_elem->eval = NULL;
   new_elem->entry_join = NULL;
   new_elem->alpha = NULL;

   if (last_look == NULL)
     {
      if (upper_level == NULL)
        { *top_pat_node = new_elem; }
      else
        { upper_level->next_level = new_elem; }
     }
   else
     {
      if (cn_test == CLIPS_TRUE)
        {
         last_look->right_node = new_elem;
         new_elem->left_node = last_look;
        }
      else if (upper_level != NULL)
        {
         new_elem->right_node = upper_level->next_level;
         if (upper_level->next_level != NULL)
           { upper_level->next_level->left_node = new_elem; }
         upper_level->next_level = new_elem;
        }
      else
        {
         new_elem->right_node = *top_pat_node;
         if (*top_pat_node != NULL)
           { (*top_pat_node)->left_node = new_elem; }
         *top_pat_node = new_elem;
        }
     }

   /*=============================================================*/
   /* No test indicates that the end of pattern has been reached. */
   /* The pattern has now been installed in the pattern net.      */
   /*=============================================================*/

   if (elem > last_elem)
     {
      new_elem->stopNode = CLIPS_TRUE;
      new_elem->eval = NULL;
      return(new_elem);
     }

   /*=============================================================*/
   /* If this element of the pattern was a segment variable, then */
   /* the pattern node will perform a multiple bind when pattern  */
   /* matching.  Otherwise a single bind will occur.              */
   /*=============================================================*/  
   
    pn_type = GetNodeType(pat,elem);
    if (pn_type == KUNKNOWN)
      {
       CLIPSSystemError("BUILD",3);
       ExitCLIPS(4);
      }
        
    if (pn_type == SINGLE)
      { new_elem->singlefieldNode = CLIPS_TRUE; }
    else if (pn_type == MULTIFIELD)
      { new_elem->multifieldNode = CLIPS_TRUE; }
    else if (pn_type == STOP)
      { new_elem->stopNode = CLIPS_TRUE; }

   /*=============================================================*/
   /* If the test to be performed at the pattern node is a "nop"  */
   /* (i.e. no test), then set the evaluation slot to NULL rather */
   /* than storing the test.  Otherwise store the test.           */
   /*=============================================================*/
    
   ExpressionInstall(basic_list);
   new_elem->eval = PackExpression(basic_list);
   ReturnExpression(basic_list);

   /*===========================================================*/
   /* Place the next level of the pattern into the pattern net. */
   /*===========================================================*/
         
   new_elem = PlacePattern(NULL,new_elem,pat,elem+1,last_elem,top_pat_node);
   return(new_elem);
  }
  
/***********************************************************/
/* DetachJoins:                                            */
/***********************************************************/
globle VOID DetachJoins(join)
  struct joinNode *join;
  {
   struct joinNode *prev_join;
   struct joinNode *join_ptr, *last_join;   

   
   while (join != NULL)
     {
      /*======================================*/
      /* Remember the join "above" this join. */
      /*======================================*/

      prev_join = join->last_level;

      /*=================================================*/
      /* If the join was attached to a pattern, remove   */
      /* any structures associated with the pattern that */
      /* are no longer needed.                           */
      /*=================================================*/
      
      if (join->entry_pat != NULL) 
        { RemoveIntranetworkLink(join); }

      /*===========================================================*/
      /* Remove any bindings or activation links left in the join. */
      /*===========================================================*/

      FlushAlphaBetaMemory(join->beta);
        
      join->beta = NULL;
      
      /*=================================================*/
      /* Remove the expression associated with the join. */
      /* If it is the last join then the slot points to  */
      /* the defrule structure and should be left alone. */
      /*=================================================*/
 
      if (join->lastJoin == FALSE)
        {
         ExpressionDeinstall(join->eval);
         ReturnPackedExpression(join->eval);
         ExpressionDeinstall(join->not_eval);
         ReturnPackedExpression(join->not_eval);
        }
        
      rtn_struct(joinNode,join);

      if (prev_join == NULL) return;
      
      last_join = NULL;
      join_ptr = prev_join->next_level;
      while (join_ptr != NULL)
        {
         if (join_ptr == join)
           {
            if (last_join == NULL)
              { prev_join->next_level = join_ptr->right_drive_node; }
            else
              { last_join->right_drive_node = join_ptr->right_drive_node; }

            join_ptr = NULL;
           }
         else
           {
            last_join = join_ptr;
            join_ptr = join_ptr->right_drive_node; 
           }
         }  

      if (prev_join->next_level == NULL) 
        { join = prev_join; }
      else
        { join = NULL; }
     }
  
  }

/**********************************************************************/
/* RemoveIntranetworkLink:  Removes the pattern structures associated */
/*   with a join.  This includes the list structure which connects    */
/*   the pattern to the join along with any parts of the pattern that */
/*   are no longer needed if this is the last join to which the       */
/*   pattern points.                                                  */
/**********************************************************************/
static VOID RemoveIntranetworkLink(join)
  struct joinNode *join;
  {
   struct patternNode *pat_ptr;
   struct joinNode *join_ptr, *last_join;

   /**************************************************/
   /* Determine the pattern that enters this join.   */
   /* Determine the list of joins which this pattern */
   /* points to.                                     */
   /**************************************************/

   pat_ptr = join->entry_pat;
   join_ptr = pat_ptr->entry_join;
   last_join = NULL;
   
   /***************************************************/
   /* Loop through the list of join pointers that the */
   /* the pattern feeds to until this join is found.  */
   /* Remove the list pointer from that pattern.      */
   /***************************************************/

   while (join_ptr != NULL)
     {
      if (join_ptr == join)
        {
         if (last_join == NULL)
           { pat_ptr->entry_join = join_ptr->right_match_node; }
         else
           { last_join->right_match_node = join_ptr->right_match_node; }

         join_ptr = NULL;
        }
      else
        {
         last_join = join_ptr;
         join_ptr = join_ptr->right_match_node; 
        }
     }

   /*****************************************************/
   /* If the terminal node of the pattern doesn't point */
   /* to any joins, then start removing the pattern.    */
   /*****************************************************/

   if (pat_ptr->entry_join == NULL)
     {   
      ClearPatternMatches(pat_ptr);
      DetachPattern(pat_ptr); 
     }  
  }   
  
/*************************************************************/
/* DetachPattern:  Detaches a pattern from the pattern net. */
/*   The pattern net is a tree-like structure.               */
/*   Detachment of a pattern involves starting at the end    */ 
/*   of the pattern in the net (the terminating leaf) and    */ 
/*   removing the longest "limb" of the pattern.             */
/*   Example:                                                */
/*     Patterns (a b c d) and (a b e f) would be represented */
/*     by the pattern net shown on the left.  If (a b c d)   */
/*     was detached, the resultant pattern net would be the  */
/*     one shown on the right. The '=' represents an end of  */
/*     of pattern or STOP marker.                            */  
/*                                                           */
/*           a                  a                            */
/*           |                  |                            */
/*           b                  b                            */
/*           |                  |                            */
/*           c--e               e                            */
/*           |  |               |                            */
/*           d  f               f                            */
/*           |  |               |                            */
/*           =  =               =                            */
/*                                                           */
/*************************************************************/
static VOID DetachPattern(pat_ptr)
  struct patternNode *pat_ptr;
  {
   struct patternNode *upper_level;

   /*========================*/
   /* System Error Checking. */
   /*========================*/

   if (pat_ptr->next_level != NULL) 
     {
      CLIPSSystemError("RULEMNGR",2);
      ExitCLIPS(5);
     }

   /*==============================================================*/
   /* Loop until all appropriate pattern nodes have been detached. */
   /*==============================================================*/

   upper_level = pat_ptr;
   while (upper_level != NULL)
     {
      if ((upper_level->left_node == NULL) &&
          (upper_level->right_node == NULL))
        {
         /*===============================================*/
         /* Pattern node is the only node on this level.  */
         /* Remove it and continue detaching other nodes  */
         /* above this one, because no other patterns are */
         /* dependent upon this node.                     */
         /*===============================================*/
         
         pat_ptr = upper_level;
         upper_level = pat_ptr->last_level;
         if (upper_level == NULL) 
           { SetNetworkPointer(NULL); }

         FlushAlphaBetaMemory(pat_ptr->alpha);
         ExpressionDeinstall(pat_ptr->eval);
         ReturnPackedExpression(pat_ptr->eval);
         rtn_struct(patternNode,pat_ptr);
        }
      else if (upper_level->left_node != NULL)
        {
         /*====================================================*/
         /* Pattern node has another pattern node which must   */
         /* be checked preceding it.  Remove the pattern node, */
         /* but do not detach any nodes above this one.        */      
         /*====================================================*/

         pat_ptr = upper_level;
         upper_level->left_node->right_node = upper_level->right_node;
         if (upper_level->right_node != NULL)
           { upper_level->right_node->left_node = upper_level->left_node; }

         FlushAlphaBetaMemory(pat_ptr->alpha);
         ExpressionDeinstall(pat_ptr->eval);
         ReturnPackedExpression(pat_ptr->eval);
         rtn_struct(patternNode,pat_ptr);
         upper_level = NULL;
        }
      else
        {
         /*====================================================*/
         /* Pattern node has no pattern node preceding it, but */
         /* does have one succeeding it.  Remove the pattern   */
         /* node, but do not detach any nodes above this one.  */      
         /*====================================================*/

         pat_ptr = upper_level;
         upper_level = upper_level->last_level;
         if (upper_level == NULL)
           { SetNetworkPointer(pat_ptr->right_node); }
         else
           { upper_level->next_level = pat_ptr->right_node; }
         pat_ptr->right_node->left_node = NULL;

         FlushAlphaBetaMemory(pat_ptr->alpha);
         ExpressionDeinstall(pat_ptr->eval);
         ReturnPackedExpression(pat_ptr->eval);
         rtn_struct(patternNode,pat_ptr);
         upper_level = NULL;
        }
     }
  }

#endif

/*********************************************************************/
/* NetworkPointer: */
/*********************************************************************/
globle struct patternNode *NetworkPointer()
  {
   return(PatternNetworkPointer);
  }
  
/*********************************************************************/
/* SetNetworkPointer: */
/*********************************************************************/
globle VOID SetNetworkPointer(value)
  struct patternNode *value;
  {
   PatternNetworkPointer = value;
  }

#endif



