   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                   DEFFACTS MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
#define _DEFFACTS_SOURCE_

#include "setup.h"

#if DEFFACTS_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "router.h"
#include "expressn.h"
#include "evaluatn.h"
#include "constrct.h"
#include "utility.h"
#include "parsutil.h"
#include "generate.h"
#include "deffacts.h" 
  
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#endif

#if CONSTRUCT_COMPILER
#include "conscomp.h"
#endif
   
/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
    
   static struct deffacts             *ListOfDeffacts = NULL;
   static BOOLEAN                      DeletionsLegal = CLIPS_TRUE;
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static struct deffacts             *LastDeffacts = NULL;
#endif
   
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static struct deffacts HUGE_ADDR   *DeffactsArray = NULL;
   static unsigned long                NumberOfDeffacts = 0;
#endif

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static VOID                    ClearDeffacts(void);
   static VOID                    SaveDeffacts(char *);
   static int                     ParseDeffacts(char *);
#endif
   static VOID                    ResetDeffacts(void);
   VOID                           UndeffactsCommand(void);
   VOID                           PpdeffactsCommand(void);
   VOID                           ListDeffactsCommand(void);
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind(void);
   static VOID                    BsaveExpressions(FILE *);
   static VOID                    BsaveDeffacts(FILE *);
#endif
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static VOID                    BloadDeffacts(void);
   static VOID                    ClearBload(void);
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeffactsToCode(char *,int,FILE *,int,int);
   static VOID                    InitDeffactsCode(FILE *,int,int);
#endif
#else
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static VOID                    ClearDeffacts();
   static VOID                    SaveDeffacts();
   static int                     ParseDeffacts();
#endif
   static VOID                    ResetDeffacts();
   VOID                           UndeffactsCommand();
   VOID                           PpdeffactsCommand();
   VOID                           ListDeffactsCommand();
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind();
   static VOID                    BsaveExpressions();
   static VOID                    BsaveDeffacts();
#endif
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static VOID                    BloadDeffacts();
   static VOID                    ClearBload();
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeffactsToCode();
   static VOID                    InitDeffactsCode();
#endif
#endif

/*****************************/
/* InitializeDeffacts:       */
/*****************************/
globle VOID InitializeDeffacts()
  {
   AddResetFunction("deffacts",ResetDeffacts,0);
   
#if (! RUN_TIME) && (! BLOAD_ONLY)
   CreateInitialFactDeffacts();
   AddClearFunction("deffacts",ClearDeffacts,0);
#endif

#if ! RUN_TIME
   DefineFunction("undeffacts",'v', PTIF UndeffactsCommand,"UndeffactsCommand");
#if DEBUGGING_FUNCTIONS
   DefineFunction("list-deffacts", 'v', PTIF ListDeffactsCommand, "ListDeffactsCommand");
   DefineFunction("ppdeffacts",    'v', PTIF PpdeffactsCommand,   "PpdeffactsCommand"); 
#endif 
#endif

#if (! BLOAD_ONLY) && (! RUN_TIME)
   AddConstruct("deffacts",ParseDeffacts);
   AddSaveFunction("deffacts", PTIF SaveDeffacts,10);
#endif

#if (BLOAD_AND_BSAVE || BLOAD)
   AddBeforeBloadFunction("deffacts",RemoveAllDeffacts,0);
#endif

#if BLOAD_AND_BSAVE
   AddBinaryItem("deffacts",0,BsaveFind,BsaveExpressions,BsaveDeffacts,BloadDeffacts,
                              NULL,ClearBload);
#endif

#if (BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY)
   AddAbortBloadFunction("deffacts",CreateInitialFactDeffacts,0);
#endif 
   
#if (BLOAD || BLOAD_ONLY)
   AddBinaryItem("deffacts",0,NULL,NULL,NULL,BloadDeffacts,NULL,ClearBload);
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
   AddCodeGeneratorItem("deffacts",20,NULL,InitDeffactsCode,DeffactsToCode);
#endif
  }

#if (! RUN_TIME) && (! BLOAD_ONLY)
/*****************************/
/* ClearDeffacts:            */
/*****************************/
static VOID ClearDeffacts()
  {
#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return;
#endif

   RemoveAllDeffacts();
   CreateInitialFactDeffacts();
  }
  
/*****************************/
/* SaveDeffacts:          */
/*****************************/
static VOID SaveDeffacts(logicalName)
  char *logicalName;
  {
   VOID *deffactsPtr;
   char *ppform;
   
   deffactsPtr = GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      ppform = GetDeffactsPPForm(deffactsPtr);
      if (ppform != NULL)
        {
         PrintInChunks(logicalName,ppform);
         PrintCLIPS(logicalName,"\n");
        }
      deffactsPtr = GetNextDeffacts(deffactsPtr);
     }
  }

/***************************************************************/
/* ParseDeffacts:  The purpose of this function is to parse    */
/*   the deffacts statement into a list of facts which can be  */
/*   asserted when a reset is performed.  The name of the      */ 
/*   deffacts block is saved along with each fact for use with */
/*   the undeffacts statement.                                 */
/***************************************************************/
static int ParseDeffacts(readSource)
  char *readSource;
  {
   SYMBOL_HN *deffactsName;
   struct expr *temp;
   struct deffacts *newDeffacts;
   int deffactsError;
   struct token inputToken;

   deffactsError = CLIPS_FALSE;
   SetPPBufferStatus(ON);

   FlushPPBuffer();          
   SetIndentDepth(3);    
   SavePPBuffer("(deffacts ");

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) 
     {
      PrintCLIPS(WERROR,"\nCannot load deffacts with binary load in effect.\n");
      return(CLIPS_TRUE);
     }
#endif

   /*====================================================*/
   /* Parse the name and comment fields of the deffact.  */
   /* Excise the deffact if it already exists.           */
   /*====================================================*/  

   deffactsName = GetConstructNameAndComment(readSource,&inputToken,"deffacts",
                                             FindDeffacts,DeleteDeffacts,"$",CLIPS_TRUE);
   if (deffactsName == NULL) { return(CLIPS_TRUE); } 
   
   /*========================================*/
   /* Check that next token is a '(' or ')'. */
   /*========================================*/
   
   if ((inputToken.type != LPAREN) && (inputToken.type != RPAREN))
     {
      SyntaxErrorMessage("deffacts");
      return(CLIPS_TRUE);
     }

   if (inputToken.type == LPAREN)
     {
      temp = GenConstant(FCALL,FindFunction("assert"));
      temp = BuildRHSAssert(temp,readSource,CLIPS_FALSE,&deffactsError);

      if (deffactsError == CLIPS_TRUE) { return(deffactsError); }
   
      if (ExpressionContainsVariables(temp,CLIPS_FALSE))
        {
         PrintCLIPS(WERROR,"Variables cannot be accessed by a deffacts construct\n");
         ReturnExpression(temp);
         return(CLIPS_TRUE);
        }
     }
   else
     { temp = NULL; }
   
   SavePPBuffer("\n");

   ExpressionInstall(temp);
   newDeffacts = get_struct(deffacts);
   newDeffacts->name = deffactsName;
   IncrementSymbolCount(deffactsName);
   newDeffacts->alist = PackExpression(temp);
   newDeffacts->next = NULL;
   ReturnExpression(temp);

   if (GetConserveMemory() == CLIPS_TRUE)
     { newDeffacts->pp_form = NULL; }
   else
     { newDeffacts->pp_form = CopyPPBuffer(); } 
   
   if (LastDeffacts == NULL)
     { ListOfDeffacts = newDeffacts; }
   else
     { LastDeffacts->next = newDeffacts; }
   LastDeffacts = newDeffacts;

   return(deffactsError);       
  }

#endif

/**************************************************************/
/* ResetDeffacts:  Copies the deffacts list to the fact list. */
/**************************************************************/
static VOID ResetDeffacts()
  {
   struct deffacts *deffactsPtr;
   DATA_OBJECT result;
     
   DeletionsLegal = CLIPS_FALSE;
   deffactsPtr = ListOfDeffacts;
   while (deffactsPtr != NULL)
     {
      if (GetHaltExecution() == CLIPS_TRUE)
        {
         DeletionsLegal = CLIPS_TRUE;
         return;
        }
        
      if (deffactsPtr->alist != NULL)
        { 
         SetEvaluationError(CLIPS_FALSE);
         EvaluateExpression(deffactsPtr->alist,&result); 
        }
      deffactsPtr = deffactsPtr->next;
     }
   DeletionsLegal = CLIPS_TRUE;
  }
  
#if (! RUN_TIME) && (! BLOAD_ONLY)
/*********************************************************/
/* CreateInitialFactDeffacts:  Creates the initial fact, */
/*  (initial-fact), and places it on the deffacts list.  */
/*********************************************************/
globle VOID CreateInitialFactDeffacts()
  {
   struct expr *stub;
   struct deffacts *newDeffacts;

   /*==========================*/
   /* Create the initial fact. */
   /*==========================*/

   stub = GenConstant(FCALL,FindFunction("assert"));
   stub->arg_list = GenConstant(SYMBOL,AddSymbol("initial-fact"));
   ExpressionInstall(stub);

   newDeffacts = get_struct(deffacts);
   newDeffacts->name = (SYMBOL_HN *) AddSymbol("initial-fact");
   IncrementSymbolCount(newDeffacts->name);
   newDeffacts->alist = PackExpression(stub);
   newDeffacts->next = NULL;
   newDeffacts->pp_form = NULL;
   ReturnExpression(stub);
   
   LastDeffacts = ListOfDeffacts = newDeffacts;  
  }

/***************************************************************/
/* RemoveAllDeffacts:                  */
/***************************************************************/
globle VOID RemoveAllDeffacts()
  {
   struct expr *stub;
   struct deffacts *nextDeffacts;
   
   while (ListOfDeffacts != NULL)
     {
      stub = ListOfDeffacts->alist;
      ExpressionDeinstall(stub);
      ReturnPackedExpression(stub);
      
      nextDeffacts = ListOfDeffacts->next;

      DecrementSymbolCount(ListOfDeffacts->name);
      if (ListOfDeffacts->pp_form != NULL)
        {
         rm(ListOfDeffacts->pp_form,
                    (int) sizeof(char) * ((int) strlen(ListOfDeffacts->pp_form) + 1));
        }
      rtn_struct(deffacts,ListOfDeffacts);

      ListOfDeffacts = nextDeffacts;
     }
   LastDeffacts = NULL;
  }

/*****************************************************/
/* DeleteNamedDeffacts: Delete all facts in the deffacts  */
/*   list which have the identifying name of the     */
/*   deffacts block to be removed.                   */
/*****************************************************/
globle BOOLEAN DeleteNamedDeffacts(deffactsName)
  char *deffactsName;
  {
#if (! BLOAD_ONLY)
   VOID *deffactsPtr;

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(0);
#endif

   if (DeletionsLegal == CLIPS_FALSE) return(0);
   
   deffactsPtr = FindDeffacts(deffactsName);
   
   if (deffactsPtr != NULL)
     { return(DeleteDeffacts(deffactsPtr)); }
   else
     { 
      if (strcmp("*",deffactsName) == 0)
        {
         RemoveAllDeffacts();
         return(CLIPS_TRUE);
        }
        
      return(CLIPS_FALSE); 
     }
#else
   return(CLIPS_FALSE);
#endif
  }

#endif

/******************************************************************/
/* DeleteDeffacts: Deletes a deffacts from the list of deffacts.  */
/*   Returns 1 if the deffacts was successfully deleted,          */
/*   otherwise 0 is returned. If the NULL pointer is passed as    */
/*   the pointer to the deffacts to be deleted, then all deffacts */ 
/*   are deleted.                                                 */
/******************************************************************/
globle BOOLEAN DeleteDeffacts(theDeffacts)
  VOID *theDeffacts;
  {
#if BLOAD_ONLY || RUN_TIME
   return(0); 
#else
   struct deffacts *nextDeffacts, *lastDeffacts;
   struct deffacts *currentDeffacts;
   struct expr *stub;
   
   /*=============================================*/
   /* Return FALSE if deffacts cannot be deleted. */
   /*=============================================*/
   
#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return(CLIPS_FALSE);
#endif

   if (DeletionsLegal == CLIPS_FALSE) return(CLIPS_FALSE);
   
   /*==================================================================*/
   /* Delete all deffacts if the deffacts pointer is the NULL pointer. */
   /*==================================================================*/
   
   if (theDeffacts == NULL)
     {
      RemoveAllDeffacts();
      return(CLIPS_TRUE);
     }
         
   /*=================================================*/
   /* Search for the deffacts in the list of deffacts */
   /* and delete it when found.                       */
   /*=================================================*/
   
   lastDeffacts = NULL;
   currentDeffacts = ListOfDeffacts;
   while (currentDeffacts != NULL)
     {
      nextDeffacts = currentDeffacts->next;
      if (currentDeffacts == (struct deffacts *) theDeffacts)
        {
         stub = currentDeffacts->alist;
         ExpressionDeinstall(stub);
         ReturnPackedExpression(stub);

         if (lastDeffacts == NULL)
           { ListOfDeffacts = nextDeffacts; }
         else
           { lastDeffacts->next = nextDeffacts; }

         if (LastDeffacts == currentDeffacts)
           { LastDeffacts = lastDeffacts; }

         DecrementSymbolCount(currentDeffacts->name);
         if (currentDeffacts->pp_form != NULL)
           {
            rm(currentDeffacts->pp_form,
                    (int) sizeof(char) * ((int) strlen(currentDeffacts->pp_form) + 1));
           }
         rtn_struct(deffacts,currentDeffacts);
         return(1);
        }
  
      lastDeffacts = currentDeffacts;
      currentDeffacts = nextDeffacts;
     }
        
   return(0); 
#endif
  }
  
/******************************************************************/
/* FindDeffacts:  Searches for a deffact in the list of deffacts. */
/*   Returns a pointer to the deffact if found, otherwise NULL.   */
/******************************************************************/
globle VOID *FindDeffacts(deffactsName)
  char *deffactsName;
  {
   struct deffacts *deffactsPtr;
   SYMBOL_HN *findValue;
   
   if ((findValue = (SYMBOL_HN *) FindSymbol(deffactsName)) == NULL) return(NULL);

   deffactsPtr = ListOfDeffacts;
   while (deffactsPtr != NULL)
     {
      if (deffactsPtr->name == findValue)
        { return((VOID *) deffactsPtr); }
      
      deffactsPtr = deffactsPtr->next; 
     }

   return(NULL);
  }

/************************************************************/
/* SetListOfDeffacts:                            */
/************************************************************/
globle VOID SetListOfDeffacts(deffactsPtr)
  VOID *deffactsPtr;
  {
   ListOfDeffacts = (struct deffacts *) deffactsPtr; 
   /* Need to set the value of LastDeffacts. */
  }
  
/************************************************************/
/* GetNextDeffacts:                            */
/************************************************************/
globle VOID *GetNextDeffacts(deffactsPtr)
  VOID *deffactsPtr;
  {
   if (deffactsPtr == NULL)
     { return((VOID *) ListOfDeffacts); }
   else
     { return((VOID *) (((struct deffacts *) deffactsPtr)->next)); }
  }

/************************************************************/
/* GetDeffactsName:                            */
/************************************************************/
globle char *GetDeffactsName(deffactsPtr)
  VOID *deffactsPtr;
  { return(ValueToString(((struct deffacts *) deffactsPtr)->name)); }
  
/************************************************************/
/* GetDeffactsPPForm:                            */
/************************************************************/
globle char *GetDeffactsPPForm(deffactsPtr)
  VOID *deffactsPtr;
  { return(((struct deffacts *) deffactsPtr)->pp_form); }

/**************************************************************/
/* IsDeffactsDeletable: Returns TRUE if a particular deffacts */
/*   can be deleted otherwise returns FALSE.                  */
/**************************************************************/
globle BOOLEAN IsDeffactsDeletable(ptr)
  VOID *ptr;
  {
#if MAC_MPW
#pragma unused(ptr)
#endif
#if BLOAD_ONLY || RUN_TIME
   return(FALSE);
#else
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded()) return(FALSE);
#endif
   return(DeletionsLegal);
#endif
  }
  
/*****************************************************/
/* UndeffactsCommand: removes a deffacts statement. */
/*   Syntax: (undeffacts <deffacts name>)            */
/*****************************************************/
globle VOID UndeffactsCommand()
  {
   char *deffactsName;

   deffactsName = GetConstructName(1,1,"undeffacts","deffacts name");
   if (deffactsName == NULL) return;

#if (! RUN_TIME) && (! BLOAD_ONLY)
   if ((FindDeffacts(deffactsName) == CLIPS_FALSE) && (strcmp("*",deffactsName) != 0))
     { 
      PrintCLIPS(WERROR,"Unable to find deffacts ");
      PrintCLIPS(WERROR,deffactsName);
      PrintCLIPS(WERROR,"\n");
      return;
     }
   else if (DeleteNamedDeffacts(deffactsName) == CLIPS_FALSE)
     { 
      PrintCLIPS(WERROR,"Unable to delete deffacts ");
      PrintCLIPS(WERROR,deffactsName);
      PrintCLIPS(WERROR,"\n");
      return;
     }

   return;
#else
   PrintCLIPS(WERROR,"Unable to delete deffacts ");
   PrintCLIPS(WERROR,deffactsName);
   PrintCLIPS(WERROR,"\n");
   return;
#endif
  }
 
#if DEBUGGING_FUNCTIONS
/**********************************************************/
/* PPDeffacts: the driver which actually does the pretty  */
/*   printing of the deffact.                             */
/**********************************************************/
globle int PPDeffacts(deffactsName,logicalName)
  char *deffactsName, *logicalName;
  {
   VOID *deffactsPtr;

   deffactsPtr = FindDeffacts(deffactsName);
   if (deffactsPtr == NULL)        
     {
      PrintCLIPS(WERROR,"Unable to find deffacts ");
      PrintCLIPS(WERROR,deffactsName);
      PrintCLIPS(WERROR,"\n");
      return(CLIPS_FALSE);
     }
     
   if (GetDeffactsPPForm(deffactsPtr) == NULL) return(CLIPS_TRUE);
   PrintInChunks(logicalName,GetDeffactsPPForm(deffactsPtr));
   return(CLIPS_TRUE);
  }
  
/************************************************************/
/* PpdeffactsCommand: pretty prints a deffact.                  */
/*   Syntax: (pp_deffact <deffact name>)                    */
/************************************************************/
globle VOID PpdeffactsCommand()
  {
   char *deffactsName;

   deffactsName = GetConstructName(1,1,"ppdeffacts","deffact name");
   if (deffactsName == NULL) return;

   PPDeffacts(deffactsName,WDISPLAY);

   return;
  }

/************************************************************/
/* ListDeffacts: displays the list of deffacts.            */
/*   Syntax: (list-deffacts)                                */
/************************************************************/
globle VOID ListDeffacts()
  {
   VOID *deffactsPtr;
   char *deffactsName;
   long count = 0;

   deffactsPtr = GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {         
      if (GetHaltExecution() == CLIPS_TRUE) return;

      deffactsName = GetDeffactsName(deffactsPtr);
      if (deffactsName != NULL)
        {
         PrintCLIPS(WDISPLAY,deffactsName);
         PrintCLIPS(WDISPLAY,"\n");
        }
      deffactsPtr = GetNextDeffacts(deffactsPtr);
      count++;
     }

   PrintTally(WDISPLAY,count,"deffacts","deffacts");

   return;
  }

/************************************************************/
/* ListDeffactsCommand: displays the list of deffacts.        */
/*   Syntax: (list-deffacts)                                */
/************************************************************/
globle VOID ListDeffactsCommand()
  {
   if (ArgCountCheck("list-deffacts",EXACTLY,0) == -1) return;

   ListDeffacts();
  }
   
#endif

#if BLOAD_AND_BSAVE
/********************************/
/* BsaveFind:                   */
/********************************/
static VOID BsaveFind()
  {
   struct deffacts *deffactsPtr;
   
   deffactsPtr = (struct deffacts *) GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      deffactsPtr->name->neededSymbol = CLIPS_TRUE;
      ExpressionCount += ExpressionSize(deffactsPtr->alist);
      MarkNeededItems(deffactsPtr->alist);
      deffactsPtr = deffactsPtr->next;
     }
  }

/********************************/
/* BsaveExpressions:            */
/********************************/
static VOID BsaveExpressions(fp)
  FILE *fp;
  {
   struct deffacts *deffactsPtr;
   
   deffactsPtr = (struct deffacts *) GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      BsaveExpression(deffactsPtr->alist,fp);
      deffactsPtr = deffactsPtr->next;
     }
  }

/*****************************************************/
/* BsaveDeffacts: Writes out all deffacts structures */ 
/*   to the binary file                              */
/*****************************************************/
static VOID BsaveDeffacts(fp)
  FILE *fp;
  {
   unsigned long int count = 0, space;
   struct deffacts *deffactsPtr;
   struct deffacts newDeffacts;
   
   /*============================================*/
   /* Write out the size and number of deffacts. */
   /*============================================*/
  
   deffactsPtr = (struct deffacts *) GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      count++;
      deffactsPtr = deffactsPtr->next;
     }
   
   space = count * sizeof(struct deffacts) + sizeof(unsigned long);
   GenWrite(&space,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
   if (count == 0) return;
     
   /*=========================*/
   /* Write out each deffact. */
   /*=========================*/
   
   count = 0;
   deffactsPtr = (struct deffacts *) GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      newDeffacts.name = (SYMBOL_HN *) deffactsPtr->name->bucket;
      
      newDeffacts.pp_form = NULL;
      if (deffactsPtr->alist != NULL)
        { 
         newDeffacts.alist = (struct expr *) ExpressionCount;
         ExpressionCount += ExpressionSize(deffactsPtr->alist); 
        }
      else
        { newDeffacts.alist = (struct expr *) -1L; }
        
      count++;
      if (deffactsPtr->next != NULL)
        { newDeffacts.next = (struct deffacts *) count; }
      else
        { newDeffacts.next = (struct deffacts *) -1L; }
        
      GenWrite(&newDeffacts,(unsigned long) sizeof(struct deffacts),fp);
      deffactsPtr = deffactsPtr->next;
     }
  }
#endif 

#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
/****************************************************************/
/* BloadDeffacts: Loads the deffacts used by this binary image. */
/****************************************************************/
static VOID BloadDeffacts()
  {
   unsigned long int space, i;
   long int index;
   
   /*=============================*/
   /* Get the number of deffacts. */
   /*=============================*/
  
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   space -= sizeof(unsigned long);
   GenRead(&NumberOfDeffacts,(unsigned long) sizeof(unsigned long int));
   if (NumberOfDeffacts == 0)
     {
      DeffactsArray = NULL;
      return;
     }
     
   /*=======================*/
   /* Read in the deffacts. */
   /*=======================*/
   
   DeffactsArray = (struct deffacts HUGE_ADDR *) genlongalloc(space);
   GenRead((VOID *) DeffactsArray,space);
           
   /*==========================*/
   /* Reset deffacts pointers. */
   /*==========================*/

   for (i = 0; i < NumberOfDeffacts; i++)
     {
      DeffactsArray[i].name = SymbolPointer(DeffactsArray[i].name);
      IncrementCount(DeffactsArray[i].name);
      DeffactsArray[i].alist = (struct expr *)
                         ExpressionPointer(DeffactsArray[i].alist);
     
      index = (long int) DeffactsArray[i].next;
      if (index != -1)
        { DeffactsArray[i].next = (struct deffacts *) &DeffactsArray[index]; }
      else
        { DeffactsArray[i].next = NULL; }
     }
     
   /*============================================*/
   /* Set the deffacts to the new deffacts list. */
   /*============================================*/
   
   SetListOfDeffacts((VOID *) DeffactsArray); 
  }
  
/****************************************************************/
/* ClearBload: */
/****************************************************************/
static VOID ClearBload()
  {
   unsigned long int i, space;
   
   for (i = 0; i < NumberOfDeffacts; i++)
     { DecrementSymbolCount(DeffactsArray[i].name); }
     
   space = NumberOfDeffacts * sizeof(struct deffacts);
   if (space != 0) genlongfree((VOID *) DeffactsArray,space); 
   
   SetListOfDeffacts(NULL);
#if ! BLOAD_ONLY
   CreateInitialFactDeffacts();
#endif
  }
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)

/***********************************************************/
/* InitDeffactsCode:  */
/***********************************************************/
static VOID InitDeffactsCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   if (GetNextDeffacts(NULL) != NULL)
     { fprintf(initFP,"   SetListOfDeffacts(dft%d_1);\n",imageID); }
   else
     { fprintf(initFP,"   SetListOfDeffacts(NULL);\n"); }
  }

/***********************************************************/
/* DeffactsToCode:  Produces all the deffacts definitions. */
/***********************************************************/
static int DeffactsToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   struct deffacts *deffactsPtr;
   FILE *fp;
   int version = 1;
   int i = 1;
   int newHeader = CLIPS_TRUE;
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   fprintf(headerFP,"#include \"deffacts.h\"\n");
   if (GetNextDeffacts(NULL) == NULL) return(-1);
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
   
   deffactsPtr = (struct deffacts *) GetNextDeffacts(NULL);
   while (deffactsPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct deffacts dft%d_%d[] = {\n",imageID,version);
         fprintf(headerFP,"extern struct deffacts dft%d_%d[];\n",imageID,version);
         newHeader = CLIPS_FALSE;
        }
     
      fprintf(fp,"{"); 
      PrintSymbolReference(fp,deffactsPtr->name); /* name */
      fprintf(fp,",NULL,");                 /* ppform */
      ExpressionToCode(fp,deffactsPtr->alist);
      fprintf(fp,",");

      if (deffactsPtr->next == NULL)
        { fprintf(fp,"NULL"); }
      else
        { 
         if ((i + 1) > maxIndices)
           { fprintf(fp,"&dft%d_%d[%d]",imageID,version + 1,0); }
         else
           { fprintf(fp,"&dft%d_%d[%d]",imageID,version,i); }
        }
        
      i++;
      deffactsPtr = deffactsPtr->next;
      if ((i > maxIndices) || (deffactsPtr == NULL))
        {
         fprintf(fp,"}};\n");
         fclose(fp);
         i = 1;
         version++;
         if (deffactsPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,"},\n"); }
     }
   return(1);
  }
#endif

#endif /* DEFFACTS_CONSTRUCT */


