   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                 RULE MANAGER MODULE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _DEFRULE_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "scanner.h"
#include "expressn.h"
#include "evaluatn.h"
#include "router.h"
#include "constrct.h"
#include "network.h"
#include "facts.h"
#include "reorder.h"
#include "analysis.h"
#include "ruleprsr.h"
#include "match.h"
#include "retract.h"
#include "build.h"
#include "drive.h"
#include "engine.h"
#include "reteutil.h"
#include "utility.h"
#include "defrule.h"
#include "variable.h"
#include "spclform.h"
#include "parsutil.h"
#include "generate.h"
    
#if DEFTEMPLATE_CONSTRUCT
#include "deftmfun.h"
#endif

#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#endif
  
/****************************************/
/* GLOBAL EXTERNAL FUNCTION DEFINITIONS */
/****************************************/

#if ANSI_COMPILER
   int                            DefruleCommands(void);
#else
   int                            DefruleCommands();
#endif

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static int                     ParseDefrule(char *); 
   static VOID                    RememberJoinsForRule(struct joinNode *,struct defrule *,int);
   static VOID                    SaveDefrules(char *);
   static struct joinNode        *AddTerminatorJoin(void);
   static VOID                    ReplaceExpressionVariables(struct expr *,int *);
   static BOOLEAN                 ReplaceRHSVariable(struct expr *);
   static VOID                    MarkNetworkForIncrementalReset(struct defrule *,int);
   static VOID                    IncrementalReset(struct defrule *);
   static VOID                    CheckForPrimableJoins(struct defrule *);
   static VOID                    AddDefrule(struct defrule *);
   static VOID                    RemoveRuleNetwork(struct defrule *);
#endif
   static VOID                    ResetDefrules(void);
#else
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static int                     ParseDefrule();
   static VOID                    RememberJoinsForRule();
   static VOID                    SaveDefrules();
   static struct joinNode        *AddTerminatorJoin();
   static VOID                    ReplaceExpressionVariables();
   static BOOLEAN                 ReplaceRHSVariable();
   static VOID                    MarkNetworkForIncrementalReset();
   static VOID                    IncrementalReset();
   static VOID                    CheckForPrimableJoins();
   static VOID                    AddDefrule();
   static VOID                    RemoveRuleNetwork();
#endif
   static VOID                    ResetDefrules();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct defrule    *ListOfDefrules = NULL;
   static BOOLEAN            DeletionsAllowed = CLIPS_TRUE;
#if DEBUGGING_FUNCTIONS
   static int                WatchRules = OFF;
#endif
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static struct defrule    *LastDefrule = NULL;
#if DEBUGGING_FUNCTIONS
   static int                DeletedRuleHadBreakpoint = CLIPS_FALSE;
#endif
#endif

/******************************************************************/
/* InitializeDefrules: Initializes the defrule construct for use. */
/******************************************************************/
globle VOID InitializeDefrules()
  {
#if ! RUN_TIME
   DefruleCommands();
#endif

#if DEBUGGING_FUNCTIONS
   AddWatchItem("rules",&WatchRules,70);
#endif

   InitializeEngine();
   
   AddResetFunction("defrule",ResetDefrules,-10);
   
#if (! BLOAD_ONLY) && (! RUN_TIME)
   AddClearFunction("defrule",ClearDefrules,10);
   AddConstruct("defrule",ParseDefrule);
   AddSaveFunction("defrule", PTIF SaveDefrules,0);
#endif
  }
  
/*************************************************************/
/* FindDefrule:  Searches for a rule in the list of rules.   */
/*   Returns a pointer to the rule if found, otherwise NULL. */
/*************************************************************/
globle VOID *FindDefrule(ruleName)
  char *ruleName;
  {
   struct defrule *rulePtr;
   SYMBOL_HN *findValue;
   
   if ((findValue = (SYMBOL_HN *) FindSymbol(ruleName)) == NULL) return(NULL);
   
   rulePtr = ListOfDefrules;
   while (rulePtr != NULL)
     {
      if (rulePtr->name == findValue) return((VOID *) rulePtr);
      
      rulePtr = rulePtr->next; 
     }

   return(NULL);
  }
  
/***********************************************************/
/* SetRuleDeletions:                         */
/***********************************************************/
globle BOOLEAN SetRuleDeletions(value)
  int value;
  { 
   int ov;
   
   ov = DeletionsAllowed;
   DeletionsAllowed = value;
   return(ov);
  }
  
/***********************************************************/
/* GetRuleDeletions:                         */
/***********************************************************/
globle BOOLEAN GetRuleDeletions()
  { 
   return(DeletionsAllowed);
  }
  
/*********************************************************************/
/* ResetDefrules: */
/*********************************************************************/   
static VOID ResetDefrules()
  {
   ResetNotedPatterns(NetworkPointer());
  }

#if (! RUN_TIME) && (! BLOAD_ONLY)

/***********************************************************/
/* DeleteNamedDefrule: Remove a named rule from the database. */
/*   Returns 1 if the rule was found and removed.  Returns */
/*   0 if the rule was not found.                          */
/***********************************************************/
globle BOOLEAN DeleteNamedDefrule(ruleName)
  char *ruleName;
  {
#if (! BLOAD_ONLY)
   VOID *defrulePtr;

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(0);
#endif

   if (DeletionsAllowed == CLIPS_FALSE) return(CLIPS_FALSE);
   
   defrulePtr = FindDefrule(ruleName);
   
   if (defrulePtr != NULL)
     { return(DeleteDefrule(defrulePtr)); }
   else
     { 
      if (strcmp("*",ruleName) == 0)
        {
         ClearDefrules();
         return(CLIPS_TRUE);
        }
        
      return(CLIPS_FALSE); 
     }
#else
   return(CLIPS_FALSE);
#endif
  }

/***********************************************************/
/* RemoveRuleNetwork:                                    */
/***********************************************************/
static VOID RemoveRuleNetwork(rulePtr)
   struct defrule *rulePtr;
   {
    struct patptr *pat_del;
    struct patptr *next_del;
    
    while (rulePtr != NULL)
      {
       pat_del = rulePtr->pats;
       while (pat_del != NULL)
         {
          next_del = pat_del->next;

          if (pat_del->pptr == NULL)
            { DetachJoins(pat_del->jptr); } 
  
          rtn_struct(patptr,pat_del);
          pat_del = next_del; 
         }
       rulePtr->pats = NULL;
       rulePtr = rulePtr->disjunct;
      }
   }
   
#endif

/***************************************************************/
/* DeleteDefrule: Deletes a defrule from the list of defrules. */
/*   Returns 1 if the defrule was successfully deleted,        */
/*   otherwise 0 is returned. If the NULL pointer is passed as */
/*   the pointer to the rule to be deleted, then all rules are */
/*   deleted.                                                  */
/***************************************************************/
globle BOOLEAN DeleteDefrule(theRule)
  VOID *theRule;
  {
#if BLOAD_ONLY || RUN_TIME
   return(0);
#else
   struct defrule *rulePtr, *ruleBefore, *tempRule;
   int ruleFound;
   
   /*=============================================*/
   /* Return FALSE if defrules cannot be deleted. */
   /*=============================================*/
   
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(CLIPS_FALSE);
#endif

   if (DeletionsAllowed == CLIPS_FALSE) return(CLIPS_FALSE);
   
   /*==============================================================*/
   /* Delete all defrules if the rule pointer is the NULL pointer. */
   /*==============================================================*/
   
   if (theRule == NULL)
     {
      if (GetRuleFiring() != NULL) DeletedFiringRule = CLIPS_TRUE;
      ClearDefrules();
      return(CLIPS_TRUE);
     }
   
   /*=================================================*/
   /* Search for the defrule in the list of defrules. */
   /*=================================================*/
   
   ruleBefore = NULL;
   ruleFound = CLIPS_FALSE;
   rulePtr = ListOfDefrules;
   while ((rulePtr != NULL) && (ruleFound == CLIPS_FALSE))
     {
      if (rulePtr == (struct defrule *) theRule)
        { ruleFound = CLIPS_TRUE; }
      else
        {  
         ruleBefore = rulePtr;
         rulePtr = rulePtr->next;
        } 
     }

   /*=====================================================================*/
   /* If a rule is redefined, then we want to save its breakpoint status. */
   /*=====================================================================*/
   
#if DEBUGGING_FUNCTIONS
   DeletedRuleHadBreakpoint = rulePtr->afterBreakpoint;
#endif
   
   /*====================================================*/
   /* If the currently executed rule has been deleted,   */
   /* then set a flag indicating that this has happened. */
   /*====================================================*/
   
   if (ValueToString(rulePtr->name) == GetRuleFiring())
     { DeletedFiringRule = CLIPS_TRUE; }
     
   /*============================================================*/
   /* Clear the agenda of all the activations added by the rule. */
   /*============================================================*/

   ClearRuleFromAgenda(rulePtr->name);

   /*=============================================================*/
   /* Remove all structures associated with each pattern on the   */
   /* lhs of a rule from both the pattern and join nets.  This    */
   /* will also remove the structures associated with the actions */
   /* of the rhs of the rule.                                     */
   /*=============================================================*/

   RemoveRuleNetwork(rulePtr);
   
   /*=========================================*/
   /* Remove the rule from the list of rules. */
   /*=========================================*/

   DecrementSymbolCount(rulePtr->name);
   
   if (rulePtr == LastDefrule) LastDefrule = ruleBefore;
   
   if (ruleBefore == NULL)
     {
      ListOfDefrules = rulePtr->next;
      ReturnDefrule(rulePtr);
     }
   else
     {
      tempRule = ruleBefore;
      while (tempRule != NULL)
        {
         tempRule->next = rulePtr->next;
         tempRule = tempRule->disjunct;
        }
      ReturnDefrule(rulePtr);
     }

   if (ExecutingRule == NULL)
     { FlushGarbagePartialMatches(); }
     
   /*=====================*/
   /* System Error Check. */
   /*=====================*/

   if ((ListOfDefrules == NULL) && (NetworkPointer() != NULL))
     {
      CLIPSSystemError("RULEMNGR",1);
      ExitCLIPS(5);
     }

   return(CLIPS_TRUE);
#endif
  }
  
/*************************************************************/
/* GetNextDefrule:                                            */
/*************************************************************/
globle RULEPTR GetNextDefrule(rulePtr)
  RULEPTR rulePtr;
  {
   if (rulePtr == NULL)
     { return((RULEPTR) ListOfDefrules); }
     
   return((RULEPTR) (((struct defrule *) rulePtr)->next));
  }
  
/************************************************************/
/* SetListOfDefrules:                            */
/************************************************************/
globle VOID SetListOfDefrules(listPtr)
  VOID *listPtr;
  {
   ListOfDefrules = (struct defrule *) listPtr; 
  } 
  
/*************************************************************/
/* GetDefrulePPForm:                                            */
/*************************************************************/
globle char *GetDefrulePPForm(rulePtr)
  RULEPTR rulePtr;
  {
   if (rulePtr == NULL)
     { return(NULL); }
     
   return(((struct defrule *) rulePtr)->pp_form);
  }
  
/*************************************************************/
/* GetDefruleName:                                            */
/*************************************************************/
globle char *GetDefruleName(rulePtr)
  RULEPTR rulePtr;
  {
   if (rulePtr == NULL)
     { return(NULL); }
     
   return(ValueToString(((struct defrule *) rulePtr)->name));
  }
  
/*************************************************************/
/* GetIndexedDefrule:                                            */
/*************************************************************/
globle RULEPTR GetIndexedDefrule(count)
  int count;
  {
   struct defrule *rulePtr;
   int i = 1;
   
   rulePtr = ListOfDefrules;
   while ((rulePtr != NULL) && (i < count))
     { 
      rulePtr = rulePtr->next; 
      i++;
     }
   return((RULEPTR) rulePtr);
  }

/************************************************************/
/* IsDefruleDeletable: Returns TRUE if a particular defrule */
/*   can be deleted otherwise returns FALSE.                */
/************************************************************/
globle BOOLEAN IsDefruleDeletable(ptr)
  VOID *ptr;
  {
#if MAC_MPW
#pragma unused(ptr)
#endif
#if BLOAD_ONLY || RUN_TIME
   return(FALSE);
#else
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded()) return(FALSE);
#endif
   return(DeletionsAllowed);
#endif
  }
  
/*************************************************************/
/* GetDisjunctIndex:                                            */
/*************************************************************/
globle int GetDisjunctIndex(rPtr)
  RULEPTR rPtr;
  {
   struct defrule *rulePtr, *tempPtr;
   int i = 0;
   
   rulePtr = ListOfDefrules;
   while (rulePtr != NULL)
     { 
      tempPtr = rulePtr;
      while (tempPtr != NULL)
        {
         if (tempPtr == (struct defrule *) rPtr) return(i);
         i++;
         tempPtr = tempPtr->disjunct;
        }
        
      rulePtr = rulePtr->next; 
     }
     
   return(-1);
  }

/*************************************************************/
/* EvaluateSalience:                                         */
/*************************************************************/
globle int EvaluateSalience(vPtr)
 RULEPTR vPtr;
 {
  struct defrule *rPtr;
  DATA_OBJECT salienceValue;
  int salience;
  
  rPtr = (struct defrule *) vPtr;
  
  if (GetSalienceEvaluation() != WHEN_DEFINED)
    {
     if (rPtr->dynamicSalience == NULL) return(rPtr->salience);
     
     SetEvaluationError(CLIPS_FALSE);
     if (EvaluateExpression(rPtr->dynamicSalience,&salienceValue))
       {
        SalienceInformationError(ValueToString(rPtr->name));
        return(rPtr->salience);
       }
     
     if (salienceValue.type != INTEGER)
       {
        PrintCLIPS(WERROR,"\nSalience value must be an integer value.\n");
        SalienceInformationError(ValueToString(rPtr->name));
        SetEvaluationError(CLIPS_TRUE);
        return(rPtr->salience);
       }

     salience = (int) ValueToLong(salienceValue.value);
     
     if ((salience > MAX_SALIENCE) || (salience < MIN_SALIENCE))
        {  
         PrintCLIPS(WERROR,"\nSalience value out of range ");
         PrintLongInteger(WERROR,(long int) MIN_SALIENCE);
         PrintCLIPS(WERROR," to ");
         PrintLongInteger(WERROR,(long int) MAX_SALIENCE);
         PrintCLIPS(WERROR,"\n");
         SetEvaluationError(CLIPS_TRUE);
         SalienceInformationError(ValueToString(((struct defrule *) rPtr)->name));
         return(rPtr->salience);
        } 
     rPtr->salience = salience;
    }
    
  return(rPtr->salience);
 }   
   
/*************************************************************/
/* SalienceInformationError:                                  */
/*************************************************************/
globle VOID SalienceInformationError(ruleName)
  char *ruleName;
  {
   PrintCLIPS(WERROR,"This error occurred while evaluating the salience");
   if (ruleName == NULL)
     {
      PrintCLIPS(WERROR," for rule ");
      PrintCLIPS(WERROR,ruleName);
     }
   PrintCLIPS(WERROR,"\n");
  }
 
#if DEBUGGING_FUNCTIONS
/*************************************************************/
/* GetRulesWatch:                                            */
/*************************************************************/
globle BOOLEAN GetRulesWatch()
  {
   return(WatchRules);
  }
#endif
  
#if (! RUN_TIME) && (! BLOAD_ONLY)

/*********************************************************************/
/* AddDefrule: */
/*********************************************************************/
static VOID AddDefrule(rulePtr)
  struct defrule *rulePtr;
  {   
   struct defrule *tempRule;
   
   IncrementSymbolCount(rulePtr->name);
   ExpressionInstall(rulePtr->dynamicSalience);
   
   if (LastDefrule == NULL)
     { ListOfDefrules = rulePtr; }
   else
     { 
      tempRule = LastDefrule;
      while (tempRule != NULL)
        {
         tempRule->next = rulePtr;
         tempRule = tempRule->disjunct;
        } 
     }
     
   LastDefrule = rulePtr;
  }
  
/*********************************************************************/
/* MarkNetworkForIncrementalReset: */
/*********************************************************************/
static VOID MarkNetworkForIncrementalReset(tempRule,value)
  struct defrule *tempRule;
  int value;
  {
   struct patptr *iptr;
   struct patternNode *pptr;
   
   while (tempRule != NULL)
     {
      iptr = tempRule->pats;
      while (iptr != NULL)
        {
         if (iptr->jptr->initialize)
           {
            iptr->jptr->initialize = value;
            pptr = iptr->jptr->entry_pat;
            if (pptr != NULL)
              { if (pptr->initialize == CLIPS_FALSE) pptr = NULL; }
            
            while (pptr != NULL)
              {
               pptr->initialize = value;
               pptr = pptr->last_level;
              }
           }
        
         iptr = iptr->next;
        }
      tempRule = tempRule->disjunct;
     }
  }

/*********************************************************************/
/* IncrementalReset: */
/*********************************************************************/
static VOID IncrementalReset(tempRule)
  struct defrule *tempRule;
  {
   struct defrule *tempPtr;
   struct fact *factPtr;
   struct patptr *iptr;
      
   IncrementalResetFlag = CLIPS_TRUE;
   
   tempPtr = tempRule;
   while (tempPtr != NULL)
     {
      iptr = tempPtr->pats;
      while (iptr != NULL)
        { 
         ResetNotedJoin(iptr->jptr);
         iptr = iptr->next;
        }
      tempPtr = tempPtr->disjunct;
     }
     
   tempPtr = tempRule;
   while (tempPtr != NULL)
     {
      CheckForPrimableJoins(tempRule);
      tempPtr = tempPtr->disjunct;
     }
   
   factPtr = (struct fact *) GetNextFact(NULL);
   
   while (factPtr != NULL)
     {
      PatternMatch(factPtr,factPtr->atoms,NetworkPointer(),1,0,NULL,NULL);
      factPtr = (struct fact *) GetNextFact(factPtr);
     }
   
   IncrementalResetFlag = CLIPS_FALSE;
  }
  
/*******************************************************************************/
/* CheckForPrimableJoins: Updates the joins of a rule for an incremental reset */
/*   if portions of that rule are shared with other rules that have already    */
/*   been incrementally reset. A join for a new rule will be updated if it is  */
/*   marked for initialization and either its parent join or its associated    */
/*   entry pattern node has not been marked for initialization. The function   */
/*   PrimeJoin is used to update joins which meet these criteria.              */
/*******************************************************************************/
static VOID CheckForPrimableJoins(tempRule)
  struct defrule *tempRule;
  {
   struct patptr *iptr;
   
   iptr = tempRule->pats;
   while (iptr != NULL)
     {
      if (iptr->jptr->initialize)
        {
         if (iptr->jptr->firstJoin == CLIPS_TRUE)
           {
            if (iptr->jptr->entry_pat->initialize == CLIPS_FALSE)
              { PrimeJoin(iptr->jptr); }
           }
         else if (iptr->jptr->last_level->initialize == CLIPS_FALSE)
           { PrimeJoin(iptr->jptr); }
        }
        
      iptr = iptr->next;
     }
  }

/*********************************************************************/
/* SaveDefrules: */
/*********************************************************************/   
static VOID SaveDefrules(logicalName)
  char *logicalName;
  {
   RULEPTR rulePtr;
   char *ppform;

   rulePtr = GetNextDefrule(NULL);
   while (rulePtr != NULL)
     {
      ppform = GetDefrulePPForm(rulePtr);
      if (ppform != NULL)
        {
         PrintInChunks(logicalName,ppform);
         PrintCLIPS(logicalName,"\n");
        }
      rulePtr = GetNextDefrule(rulePtr);
     }
  }
  
/*********************************************************************/
/* ClearDefrules: */
/*********************************************************************/   
globle VOID ClearDefrules()
  {
   RULEPTR rulePtr;

#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return;
#endif
   
   /*===================================*/
   /* Remove all rules from the system. */
   /*===================================*/

   while ((rulePtr = GetNextDefrule(NULL)) != NULL)
     { DeleteDefrule(rulePtr); }
   
   /*========================*/
   /* System Error Checking. */
   /*========================*/
   
#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_FALSE)
#endif

   if (NetworkPointer() != NULL)
     { 
      CLIPSSystemError("RULEMNGR",3);
      ExitCLIPS(5);
     }
  }

/*********************************************************************/
/* ParseDefrule():                                                   */
/*   This function compiles the left and right hand sides of a rule. */
/*********************************************************************/
static int ParseDefrule(readSource)
  char *readSource;
  {
   SYMBOL_HN *ruleName;
   struct joinNode *rhs_link;
   struct node *convert_rep, *temp_node = NULL;
   struct joinNode *last_join;
   struct expr *actions, *packPtr;
   struct patternNode *old_network_pointer;
   struct expr *new_actions;
   struct token inp_tkn;
   struct defrule *topDisjunct = NULL, *currentDisjunct,
                  *lastDisjunct = NULL;
   int logicalJoin = 0;
   int error = CLIPS_FALSE;

   SetPPBufferStatus(ON);

   /*================================================*/
   /* Flush the buffer which stores the pretty print */ 
   /* representation for a rule.  Add the already    */ 
   /* parsed keyword defrule to this buffer.         */
   /*================================================*/
   
   FlushPPBuffer();            
   SavePPBuffer("(defrule ");

   /*=========================================================*/
   /* Rules cannot be loaded when a binary load is in effect. */
   /*=========================================================*/
   
#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) 
     {
      PrintCLIPS(WERROR,"\nCannot load defrules with binary load in effect.\n");
      return(CLIPS_TRUE);
     }
#endif

   /*================================================*/
   /* Parse the name and comment fields of the rule, */
   /* deleting the rule if it already exists.        */
   /*================================================*/  

#if DEBUGGING_FUNCTIONS
   DeletedRuleHadBreakpoint = CLIPS_FALSE;
#endif
   ruleName = GetConstructNameAndComment(readSource,&inp_tkn,"defrule",
                                         FindDefrule,DeleteDefrule,"*",CLIPS_FALSE);
   if (ruleName == NULL) return(CLIPS_TRUE);
   
   /*============================*/
   /* Parse the LHS of the rule. */
   /*============================*/

   convert_rep = ParseRuleLHS(readSource,&inp_tkn);
   if (convert_rep == NULL)
     {
      ReturnPackedExpression(SalienceExpression);
      SalienceExpression = NULL;
      return(CLIPS_TRUE);
     }

   /*============================*/
   /* Parse the RHS of the rule. */
   /*============================*/

   ClearParsedBindNames();
   actions = ParseRuleRHS(readSource);
   
   if (actions == NULL)
     {
      ReturnNodes(convert_rep); 
      return(CLIPS_TRUE);
     }   
   
   /*===========================================================*/
   /* The top level of the construct representing the LHS of a  */
   /* rule is assumed to be an OR.  If the implied OR is at the */
   /* top level of the pattern construct, then remove it.       */
   /*===========================================================*/

   if (convert_rep->type == PAT_OR)
     { 
      temp_node = convert_rep;
      convert_rep = convert_rep->right;
      temp_node->right = NULL;
      temp_node->bottom = NULL;
      ReturnNodes(temp_node);
     }

   while (convert_rep != NULL)
     {
      currentDisjunct = get_struct(defrule);
      currentDisjunct->pp_form = NULL;
      currentDisjunct->pats = NULL;
      currentDisjunct->next = NULL;
      currentDisjunct->logicalJoin = NULL;
      currentDisjunct->disjunct = NULL;
      currentDisjunct->name = ruleName;
      currentDisjunct->actions = NULL;
      currentDisjunct->salience = GlobalSalience;
      currentDisjunct->afterBreakpoint = 0;
      currentDisjunct->dynamicSalience = SalienceExpression;
      
      if (topDisjunct == NULL) topDisjunct = currentDisjunct;
      else lastDisjunct->disjunct = currentDisjunct;
   
      if (convert_rep->type == PAT_AND)
        { temp_node = convert_rep->right; }
      else if (convert_rep->type == PATTERN)
        { temp_node = convert_rep; }
      else
        {
         CLIPSSystemError("RULEPARS",1);
         ExitCLIPS(4);
        }
 
      if (temp_node->type == PAT_TEST)
        {
         error = CLIPS_TRUE;
         PrintCLIPS(WERROR,"\nFirst CE in a rule may not be a test CE\n");
        }
      else if ((logicalJoin = LogicalAnalysis(temp_node)) < 0)
        { error = CLIPS_TRUE; }
      else
        { error = RuleAnalysis(temp_node); }
        
      if (error == CLIPS_TRUE) 
        {
         RemoveRuleNetwork(topDisjunct);
         ReturnDefrule(topDisjunct);
         SalienceExpression = NULL;
         ReturnNodes(convert_rep);
         ReturnExpression(actions);
         FlushAnalysisExpressions();  /* new */
         FlushVariableAnalysis();     /* new */
         ClearParsedBindNames();      /* new */
         return(CLIPS_TRUE);
        }    
        
      currentDisjunct->complexity = RuleComplexity();
      
      /*==================================*/
      /* Determine if the actions are ok. */
      /*==================================*/
      
      new_actions = CopyExpression(actions);
      ReplaceExpressionVariables(new_actions,&error);
      if (error == CLIPS_TRUE)
        {
         RemoveRuleNetwork(topDisjunct);
         ReturnDefrule(topDisjunct);
         SalienceExpression = NULL;
         ReturnExpression(actions);
         ReturnExpression(new_actions);
         ReturnNodes(convert_rep);
         FlushAnalysisExpressions();  /* new */
         FlushVariableAnalysis();     /* new */
         ClearParsedBindNames();      /* new */
         return(CLIPS_TRUE);
        }
      
      old_network_pointer = NetworkPointer();
      last_join = ConstructJoins(&old_network_pointer);
      SetNetworkPointer(old_network_pointer);
       
      if (last_join == NULL)
        {
         CLIPSSystemError("RULEPARS",2);
         ExitCLIPS(4);
        }

      /*=============================================*/
      /* Connect last join of the rule to link node. */
      /*=============================================*/

      rhs_link = AddTerminatorJoin();
      rhs_link->eval = (struct expr *) currentDisjunct;
      
      ExpressionInstall(new_actions);
      packPtr = PackExpression(new_actions);
      ReturnExpression(new_actions);
      currentDisjunct->actions = packPtr;

      rhs_link->last_level = last_join;
      rhs_link->entry_pat = NULL; 
         
      rhs_link->right_drive_node = last_join->next_level;
      last_join->next_level = rhs_link;
     
      RememberJoinsForRule(rhs_link,currentDisjunct,logicalJoin);

      temp_node = convert_rep->bottom;
      convert_rep->bottom = NULL;
      ReturnNodes(convert_rep);
      convert_rep = temp_node;
      lastDisjunct = currentDisjunct;
     }

   ReturnExpression(actions);
   SalienceExpression = NULL;
   FlushAnalysisExpressions();
   FlushVariableAnalysis();
   ClearParsedBindNames();
    
   /*======================================*/
   /* Save the nice printout of the rules. */
   /*======================================*/

   SavePPBuffer("\n");
   if (GetConserveMemory() == CLIPS_TRUE)
     { topDisjunct->pp_form = NULL; }
   else
     { topDisjunct->pp_form = CopyPPBuffer(); } 
   
   /*===============================================*/
   /* Rule completely parsed. Add to list of rules. */
   /*===============================================*/

   AddDefrule(topDisjunct);

   /*========================================================================*/
   /* If a rule is redefined, then we want to restore its breakpoint status. */
   /*========================================================================*/
   
#if DEBUGGING_FUNCTIONS   
   if (DeletedRuleHadBreakpoint) AddBreakpoint(topDisjunct);
#endif
   
   /*================================*/
   /* Perform the incremental reset. */
   /*================================*/
   
   if (GetIncrementalReset())
     {
      MarkNetworkForIncrementalReset(topDisjunct,CLIPS_TRUE);
      IncrementalReset(topDisjunct);
      MarkNetworkForIncrementalReset(topDisjunct,CLIPS_FALSE);
     }
     
   return(0);
  }
  
/*********************************************************************/
/* RememberJoinsForRule:  Stores information for each rule of how it */
/*   is attached to the pattern and join networks. This information  */
/*   is stored in a linked list attached to the defrule data         */
/*   structure for the rule.                                         */
/*********************************************************************/
static VOID RememberJoinsForRule(join_ptr,rulePtr,ljoinNumber)
  struct joinNode *join_ptr;
  struct defrule *rulePtr;
  int ljoinNumber;
  {
   struct patptr *tmp_ptr;

   while (join_ptr != NULL)
     {
      if (join_ptr->depth == ljoinNumber) rulePtr->logicalJoin = join_ptr;
      tmp_ptr = get_struct(patptr);
      tmp_ptr->next = rulePtr->pats;
      tmp_ptr->pptr = join_ptr->entry_pat;
      tmp_ptr->jptr = join_ptr;

      rulePtr->pats = tmp_ptr; 
      join_ptr = join_ptr->last_level;    
     }
  }   
  
/**************************************************************/
/* AddTerminatorJoin:  Creates a "link" which will be used to */
/*   connect the join network with the actions of the rule.   */
/**************************************************************/
static struct joinNode *AddTerminatorJoin()
  {
   struct joinNode *rhs_link;

   /*=======================================================*/
   /* Create link which connects to the actions of the rule */
   /*=======================================================*/

   rhs_link = get_struct(joinNode);          
   rhs_link->id = 0;
   rhs_link->beta = NULL;
   rhs_link->depth = CountJoins() + 1;
   rhs_link->firstJoin = CLIPS_FALSE;
   rhs_link->lastJoin = CLIPS_TRUE;
   rhs_link->LHSPatternIsNegated = CLIPS_FALSE; 
   rhs_link->initialize = GetIncrementalReset(); 
   rhs_link->eval = NULL;
   rhs_link->not_eval = NULL;
   rhs_link->entry_pat = NULL;
   rhs_link->last_level = NULL;
   rhs_link->next_level = NULL;
   rhs_link->right_drive_node = NULL;
   rhs_link->right_match_node = NULL;
   
   return(rhs_link);
  }

/*****************************************************************/
/* ReplaceExpressionVariables:                                 */
/*****************************************************************/
static VOID ReplaceExpressionVariables(list,error)
  struct expr *list;
  int *error;
  {
   while (list != NULL)
     {
      if (list->type == BWORDS)
        { 
         if (ReplaceRHSVariable(list) == CLIPS_FALSE) 
           {
            *error = CLIPS_TRUE;
            return;
           } 
        }
      else if (list->type == BWORD)
        { 
         if (ReplaceRHSVariable(list) == CLIPS_FALSE)
           {
            *error = CLIPS_TRUE;
            return;
           } 
        }
     
#if DEFGLOBAL_CONSTRUCT
      else if (list->type == GBWORD)
        { if (ReplaceGlobalVariable(list) == CLIPS_FALSE) return; }
#endif

#if DEFTEMPLATE_CONSTRUCT
      else if ((list->type == FCALL) && 
               (list->value == (VOID *) FindFunction("modify")))
        { 
         if (UpdateModifyDuplicate(list,"modify")) 
           { ReplaceExpressionVariables(list->arg_list,error); }
         else
           {
            *error = CLIPS_TRUE;
            return;
           }
        }
      else if ((list->type == FCALL) && 
               (list->value == (VOID *) FindFunction("duplicate")))
        { 
         if (UpdateModifyDuplicate(list,"duplicate")) 
           { ReplaceExpressionVariables(list->arg_list,error); }
         else
           {
            *error = CLIPS_TRUE;
            return;
           }
        }
#endif
      else if (list->arg_list != NULL)
       { ReplaceExpressionVariables(list->arg_list,error); }

      list = list->next_arg;
     }
  }

/******************************************/
/* ReplaceRHSVariable:                    */
/******************************************/
static BOOLEAN ReplaceRHSVariable(list)
  struct expr *list;
  {
   SYMBOL_HN *var_name;
   struct expr *arg_lvl;
   struct var_info *var_ptr1;
   int fap, bound_also;
   
   /*=================================*/
   /* Check to see if the variable is */
   /* bound on the LHS of the rule.   */
   /*=================================*/
   
   var_name = (SYMBOL_HN *) list->value;
   var_ptr1 = FindVariable(var_name,1,CountJoins(),-1,OUTSIDE);
      
   /*=================================*/
   /* Check to see if the variable is */
   /* bound on the RHS of the rule.   */
   /*=================================*/
   
   bound_also = SearchParsedBindNames(var_name);
     
   /*========================================*/
   /* Check also if the variable is bound as */
   /* a fact-address on the LHS of the rule. */
   /*========================================*/
     
   if (list->type == BWORD)
     { fap = GetFactAddressPosition(var_name); }
   else 
     { fap = 0; }
        
   /*==========================================================*/
   /* If variable is not defined on LHS or RHS of rule then... */
   /*==========================================================*/
   
   if ((var_ptr1 == NULL) && (bound_also == CLIPS_FALSE) && (fap == 0))
     {
      PrintCLIPS(WERROR,"\nUndefined variable ");
      PrintCLIPS(WERROR,ValueToString(list->value));
      PrintCLIPS(WERROR," referenced on the RHS of the rule\n");  
      return(CLIPS_FALSE);
     }
      
   /*==================================================*/
   /* Else if variable is bound on the LHS of the rule */
   /* and not rebound on the RHS of the rule then...   */
   /*==================================================*/
   
   else if ((var_ptr1 != NULL) && (bound_also == CLIPS_FALSE))
     {
      list->type = SCALL_GET_VAR;
      list->value = GenGetvarValue(var_ptr1->pattern,var_ptr1->element);
     }
     
   else if ((fap > 0) && (bound_also == CLIPS_FALSE))
     {
      list->type = SCALL_GET_VAR;
      list->value = GenGetvarValue(fap,-1);
     }
     
   /*=====================================================*/
   /* Else if variable is rebound on RHS of rule then ... */
   /*=====================================================*/
   
   else if (bound_also == CLIPS_TRUE)
     {
      list->type = FCALL;
      list->value = (VOID *) FindFunction("(get_bind)");
      list->arg_list = GenConstant(SYMBOL,var_name);
      arg_lvl = list->arg_list;

      if (var_ptr1 != NULL)
        {
         arg_lvl->next_arg = GenConstant(SCALL_GET_VAR,
                                         GenGetvarValue(var_ptr1->pattern,
                                                        var_ptr1->element));
        } 
      else if (fap > 0)
        {
         arg_lvl->next_arg = GenConstant(SCALL_GET_VAR,
                                         GenGetvarValue(fap,-1));
        }
     }
     
   return(CLIPS_TRUE);
  }
  
/***************************************/
/* ReturnDefrule:                      */
/***************************************/
globle int ReturnDefrule(waste)
  struct defrule *waste;
  {
   int first = CLIPS_TRUE;
   struct defrule *nextPtr;
   
   if (waste == NULL) { return(0); }
   
   while (waste != NULL)
     {
      if (first)
        {
         if (waste->pp_form != NULL)
           { 
            rm(waste->pp_form,(int) strlen(waste->pp_form) + 1);
            waste->pp_form = NULL; 
           }
     
         if (waste->dynamicSalience != NULL) 
           { 
            ExpressionDeinstall(waste->dynamicSalience);
            ReturnPackedExpression(waste->dynamicSalience);
            waste->dynamicSalience = NULL; 
           }
   
         first = CLIPS_FALSE;
        }
        
      if ((waste->actions != NULL) && (waste != ExecutingRule))
        {
         ExpressionDeinstall(waste->actions);
         ReturnPackedExpression(waste->actions);
        }
        
      nextPtr = waste->disjunct;
      if (waste != ExecutingRule) rtn_struct(defrule,waste);
      else waste->disjunct = NULL;
      waste = nextPtr;
     }

   return(1);
  }

#endif

#if DEBUGGING_FUNCTIONS

/*****************************************************************/
/* ListMatches:  Displays the set of partial matches for a rule. */
/*****************************************************************/
globle BOOLEAN ListMatches(theRule)
  VOID *theRule;
  {
   struct defrule *rulePtr, *tmpPtr;
   struct partialMatch *match_list;
   struct patptr *pat_del;
   VOID *agenda_ptr;
   int pat_count, flag;

   rulePtr = (struct defrule *) theRule;
  
   tmpPtr = rulePtr;
   while (rulePtr != NULL) 
     {
     
   /*=======================================*/
   /* List individual matches for patterns. */
   /*=======================================*/

   pat_count = 1;
   pat_del = rulePtr->pats;
   while (pat_del != NULL)
     {         
      if (GetHaltExecution() == CLIPS_TRUE) return(CLIPS_TRUE);
      
      if (pat_del->jptr->firstJoin == CLIPS_TRUE)
        { pat_count = 1; }
      if (pat_del->pptr != NULL)
        { match_list = pat_del->pptr->alpha; }
      else
        { match_list = NULL; }
        
      if (pat_del->jptr->next_level != NULL)
        {
         PrintCLIPS(WDISPLAY,"Matches for Pattern ");
         PrintLongInteger(WDISPLAY,(long int) pat_count);
         PrintCLIPS(WDISPLAY,"\n");
         if (match_list == NULL) PrintCLIPS(WDISPLAY," None\n");
        }
      
      while (match_list != NULL)
        {      
         if (GetHaltExecution() == CLIPS_TRUE) return(CLIPS_TRUE);
         PrintPartialMatch(WDISPLAY,match_list);
         PrintCLIPS(WDISPLAY,"\n");
         match_list = match_list->next;
        }
      pat_count++;
      pat_del = pat_del->next;
     }

   /* old code for partial matches */
   pat_count = 0;
   pat_del = rulePtr->pats;
   while (pat_del != NULL)
     {         
      if (GetHaltExecution() == CLIPS_TRUE) return(CLIPS_TRUE);

      if (pat_del->jptr->firstJoin == CLIPS_TRUE)
        { 
         pat_count = 0;
         match_list = NULL;
        }
      else
        { match_list = pat_del->jptr->beta; }
      
      if (pat_count > 1)
        {
         PrintCLIPS(WDISPLAY,"Partial matches for patterns 1 - ");
         PrintLongInteger(WDISPLAY,(long int) pat_count);
         PrintCLIPS(WDISPLAY,"\n");
         if (match_list == NULL) { PrintCLIPS(WDISPLAY," None\n"); }
        }
      else
        { match_list = NULL; }

      while (match_list != NULL)
        {         
         if (GetHaltExecution() == CLIPS_TRUE) return(CLIPS_TRUE);

         PrintPartialMatch(WDISPLAY,match_list);
         PrintCLIPS(WDISPLAY,"\n");
         match_list = match_list->next;
        }
      pat_count++;
      pat_del = pat_del->next;
     }
   
   rulePtr = rulePtr->disjunct;
   }
   
   rulePtr = tmpPtr;
   PrintCLIPS(WDISPLAY,"Activations\n");
   agenda_ptr = GetNextActivation(NULL);
   flag = 1;
   while (agenda_ptr != NULL)
     {        
      if (GetHaltExecution() == CLIPS_TRUE) return(CLIPS_TRUE);

      if (((struct activation *) agenda_ptr)->theRule->name == rulePtr->name) 
        {
         flag = 0; 
         PrintPartialMatch(WDISPLAY,ActivationBasis(agenda_ptr));
         PrintCLIPS(WDISPLAY,"\n");
        }
      agenda_ptr = GetNextActivation(agenda_ptr);
     }
   if (flag) PrintCLIPS(WDISPLAY," None\n");
   
   
   return(CLIPS_TRUE);
  }
#endif

/************************************************************************/
/* RefreshDefrule:                                                      */
/************************************************************************/
globle BOOLEAN RefreshDefrule(theRule)
  VOID *theRule;
  {
   struct defrule *rulePtr;
   struct partialMatch *match_list;
   struct patptr *pat_del;

   rulePtr = (struct defrule *) theRule;

   while (rulePtr != NULL)
     {
      pat_del = rulePtr->pats;
      while (pat_del != NULL)
        {               
         if (pat_del->jptr->lastJoin == CLIPS_TRUE)
           {  
            match_list = pat_del->jptr->beta;
         
            while (match_list != NULL)
              {
               if ((match_list->activationf) ? 
                   (match_list->binds[match_list->bcount] == NULL) : FALSE)
                 { AddActivation(pat_del->jptr,match_list); }
              
               match_list = match_list->next;
              }
           }
      
         pat_del = pat_del->next;
        }
        
      rulePtr = rulePtr->disjunct;  
     }
     
   return(CLIPS_TRUE);
  }
  
#endif




