   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*             DEFTEMPLATE COMMANDS MODULE             */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define  _DEFTMCOM_SOURCE_

#include "setup.h"

#if DEFTEMPLATE_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "utility.h"
#include "constrct.h"
#include "router.h"
#include "facts.h"

#include "deftempl.h"
#include "deftmpsr.h"
#include "deftmfun.h"

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
#include "bload.h"
#include "bsave.h"
#endif

#if CONSTRUCT_COMPILER
#include "conscomp.h"
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
  
   static struct deftemplate    *ListOfDeftemplates = NULL;
   static struct dhash         **DeftemplateHashtable;
   static BOOLEAN                DynamicDeftemplateChecking = CLIPS_FALSE;
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static struct deftemplate    *LastDeftemplate = NULL;
#endif
#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)
   static unsigned long          NumberOfDeftemplates;
   static unsigned long          NumberOfTemplateSlots;
   static struct deftemplate HUGE_ADDR    *DeftemplateArray;
   static struct templateSlot HUGE_ADDR   *SlotArray;
#endif
   
/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
#if ! RUN_TIME
   static VOID                    DeftemplateCommands(void);
#endif
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static VOID                    InstallDeftemplate(struct deftemplate *);
   static VOID                    DeinstallDeftemplate(struct deftemplate *);
   static VOID                    SaveDeftemplates(char *);
#endif
   VOID                           UndeftemplateCommand(void);
   VOID                           PPDeftemplateCommand(void);
   VOID                           ListDeftemplatesCommand(void);
   VOID                           ModifyCommand(void);
   VOID                           DuplicateCommand(void);
   static VOID                    DuplicateModifyCommand(int);
   static struct fact            *TemplateMultifieldSlotReplace(struct expr *,struct fact *,struct deftemplate *);
   static VOID                    InitializeDeftemplateHashTable(void);
   static VOID                    AddHashDeftemplate(struct deftemplate *);
   static BOOLEAN                 RemoveHashDeftemplate(struct deftemplate *);
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind(void);
   static VOID                    BsaveExpressions(FILE *);
   static VOID                    BsaveDeftemplates(FILE *);
#endif
#if (BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY)
   static VOID                    BloadDeftemplates(void);
   static VOID                    ClearBload(void);
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeftemplatesToCode(char *,int,FILE *,int,int);
   static VOID                    InitDeftemplatesCode(FILE *,int,int);
#endif
#else 
#if ! RUN_TIME
   static VOID                    DeftemplateCommands();
#endif
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static VOID                    InstallDeftemplate();
   static VOID                    DeinstallDeftemplate();
   static VOID                    SaveDeftemplates();
#endif
   VOID                           UndeftemplateCommand();
   VOID                           PPDeftemplateCommand();
   VOID                           ListDeftemplatesCommand();
   VOID                           ModifyCommand();
   VOID                           DuplicateCommand();
   static VOID                    DuplicateModifyCommand();
   static struct fact            *TemplateMultifieldSlotReplace();
   static VOID                    InitializeDeftemplateHashTable();
   static VOID                    AddHashDeftemplate();
   static BOOLEAN                 RemoveHashDeftemplate();
#if BLOAD_AND_BSAVE
   static VOID                    BsaveFind();
   static VOID                    BsaveExpressions();
   static VOID                    BsaveDeftemplates();
#endif
#if (BLOAD_AND_BSAVE || BLOAD || BLOAD_ONLY) 
   static VOID                    BloadDeftemplates();
   static VOID                    ClearBload();
#endif
#if CONSTRUCT_COMPILER && (! RUN_TIME)
   static int                     DeftemplatesToCode();
   static VOID                    InitDeftemplatesCode();
#endif
#endif

/**************************************************************************/
/* InitializeDeftemplates: Initializes the deftemplate construct for use. */
/**************************************************************************/
globle VOID InitializeDeftemplates()
  {
#if (! RUN_TIME) && (! BLOAD_ONLY)
   AddConstruct("deftemplate",ParseDeftemplate);
   AddClearFunction("deftemplate",ClearDeftemplates,0);
   AddSaveFunction("deftemplate",(int (*)()) SaveDeftemplates,20);
#endif

   InitializeDeftemplateHashTable();
   
#if (BLOAD_AND_BSAVE || BLOAD)
   AddBeforeBloadFunction("deftemplate",ClearDeftemplates,0);
#endif
#if BLOAD_AND_BSAVE 
   AddBinaryItem("deftemplate",0,BsaveFind,BsaveExpressions,BsaveDeftemplates,
                               BloadDeftemplates,NULL,ClearBload);
#endif
#if (BLOAD || BLOAD_ONLY) 
   AddBinaryItem("deftemplate",0,NULL,NULL,NULL,BloadDeftemplates,NULL,ClearBload);
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)
   AddCodeGeneratorItem("deftemplate",30,NULL,InitDeftemplatesCode,DeftemplatesToCode);
#endif

#if ! RUN_TIME
   DeftemplateCommands();
#endif
  }

/************************************************************/
/* InitializeDeftemplateHashTable: Purpose is to initialize */
/*   the deftemplate hash table to NULL.                    */
/************************************************************/
static VOID InitializeDeftemplateHashTable()
   {
    int i;

    DeftemplateHashtable = (struct dhash **) gm2((int) sizeof (struct dhash *) * 
                                                       SIZE_DEFTEMPLATE_HASH);
      
    for (i = 0; i < SIZE_DEFTEMPLATE_HASH; i++) DeftemplateHashtable[i] = NULL;
   }
   
#if ! RUN_TIME
/*********************************************************/
/* DeftemplateCommands: Set up the deftemplate commands. */
/*********************************************************/
static VOID DeftemplateCommands()
  {
   DefineFunction("undeftemplate",'v',
                   PTIF UndeftemplateCommand,"UndeftemplateCommand");
#if DEBUGGING_FUNCTIONS
   DefineFunction("ppdeftemplate",'v',
                   PTIF PPDeftemplateCommand,"PPDeftemplateCommand"); 
   DefineFunction("list-deftemplates",'v',
                   PTIF ListDeftemplatesCommand,"ListDeftemplatesCommand");
#endif

   DefineFunction("modify",'v', PTIF ModifyCommand,"ModifyCommand");
   DefineFunction("duplicate",'v', PTIF DuplicateCommand,"DuplicateCommand");
   DefineFunction("get-dynamic-deftemplate-checking",'b',GDDCommand,"GDDCommand");
   DefineFunction("set-dynamic-deftemplate-checking",'b',SDDCommand,"SDDCommand");
   
#if (! BLOAD_ONLY)
   AddFunctionParser("modify",ModifyParse);
   AddFunctionParser("duplicate",DuplicateParse);
#endif
  }
#endif
  
#if (! RUN_TIME) && (! BLOAD_ONLY)


/************************************************************************/
/* DeleteNamedDeftemplate: Deletes the named deftemplate. Returns CLIPS_TRUE */
/*   if the deftemplate was found and deleted, otherwise CLIPS_FALSE.   */
/************************************************************************/
globle BOOLEAN DeleteNamedDeftemplate(deftemplateName)
  char *deftemplateName;
  {
#if (! BLOAD_ONLY)
   VOID *deftemplatePtr;

#if BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE
   if (Bloaded() == CLIPS_TRUE) return(0);
#endif
   
   deftemplatePtr = FindDeftemplate(deftemplateName);
   
   if (deftemplatePtr != NULL)
     { return(DeleteDeftemplate(deftemplatePtr)); }
   else
     { 
      if (strcmp("*",deftemplateName) == 0)
        {
         ClearDeftemplates();
         return(CLIPS_TRUE);
        }
        
      return(CLIPS_FALSE); 
     }
#else
   return(CLIPS_FALSE);
#endif
  }

/***********************************************/
/* ReturnSlots: Returns the slot structures of */
/*   a deftemplate to free memory.             */
/***********************************************/
globle VOID ReturnSlots(slot_ptr)
  struct templateSlot *slot_ptr;
  {
   struct templateSlot *next_slot;
   
   while (slot_ptr != NULL)
     {
      next_slot = slot_ptr->next;
      ReturnExpression(slot_ptr->defaultList);
      ReturnExpression(slot_ptr->restrictionList);
      ReturnExpression(slot_ptr->maxValue);
      ReturnExpression(slot_ptr->minValue);
      rtn_struct(templateSlot,slot_ptr);
      slot_ptr = next_slot;
     }
  }
  
/*********************************************************/
/* AddDeftemplate:          */  
/*********************************************************/
globle VOID AddDeftemplate(dtmpl_ptr)
  struct deftemplate *dtmpl_ptr;
  {
   if (LastDeftemplate == NULL)
     { ListOfDeftemplates = dtmpl_ptr; }
   else
     { LastDeftemplate->next = dtmpl_ptr; }
   LastDeftemplate = dtmpl_ptr;
   
   InstallDeftemplate(dtmpl_ptr);
  }
   
/**************************************************************/
/* InstallDeftemplate: Increments all occurrences in the hash */
/*   table of symbols found in an deftemplate and adds it to  */
/*   the hash table.                                          */  
/**************************************************************/
static VOID InstallDeftemplate(dtmplPtr)
  struct deftemplate *dtmplPtr;
  {
   struct templateSlot *slot_ptr;
   
   IncrementSymbolCount(dtmplPtr->name);
   
   slot_ptr = dtmplPtr->slotList;
   while (slot_ptr != NULL)
     {
      IncrementSymbolCount(slot_ptr->slotName);
      ExpressionInstall(slot_ptr->defaultList);
      ExpressionInstall(slot_ptr->restrictionList);
      ExpressionInstall(slot_ptr->maxValue);
      ExpressionInstall(slot_ptr->minValue);
      slot_ptr = slot_ptr->next;
     }
     
   AddHashDeftemplate(dtmplPtr);
  }

/***********************************************************/
/* DeinstallDeftemplate: Decrements all occurrences in the */
/*   hash table of symbols found in an deftemplate and     */  
/*   removes it from the hash table.                       */
/***********************************************************/
static VOID DeinstallDeftemplate(dtmplPtr)
  struct deftemplate *dtmplPtr;
  {
   struct templateSlot *slot_ptr;
   
   DecrementSymbolCount(dtmplPtr->name);
         
   slot_ptr = dtmplPtr->slotList;
   while (slot_ptr != NULL)
     {
      DecrementSymbolCount(slot_ptr->slotName);
      ExpressionDeinstall(slot_ptr->defaultList);
      ExpressionDeinstall(slot_ptr->restrictionList);
      ExpressionDeinstall(slot_ptr->maxValue);
      ExpressionDeinstall(slot_ptr->minValue);
      slot_ptr = slot_ptr->next;
     }
     
   RemoveHashDeftemplate(dtmplPtr);
  }
  
/************************************************/
/* ClearDeftemplates: Removes all deftemplates. */
/************************************************/
globle VOID ClearDeftemplates()
  {
#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return;
#endif

#if ! BLOAD_ONLY   
   while (ListOfDeftemplates != NULL)
     { DeleteDeftemplate(ListOfDeftemplates); }
#endif
  }

/***************************************************/
/* SaveDeftemplates: Saves deftemplates to a file. */
/***************************************************/
static VOID SaveDeftemplates(logicalName)
  char *logicalName;
  {
   VOID *dtm_ptr;
   char *ppform;
   
   dtm_ptr = GetNextDeftemplate(NULL);
   while (dtm_ptr != NULL)
     {
      ppform = GetDeftemplatePPForm(dtm_ptr);
      if (ppform != NULL)
        {
         PrintInChunks(logicalName,ppform);
         PrintCLIPS(logicalName,"\n");
        }
      dtm_ptr = GetNextDeftemplate(dtm_ptr);
     }
  }

#endif

/***************************************************************/
/* DeleteDeftemplate: Deletes a deftemplate from the list of   */
/*   deftemplates. Returns 1 if the deffacts was successfully  */
/*   deleted, otherwise 0 is returned. If the NULL pointer is  */
/*   passed as the pointer to the deffacts to be deleted, then */ 
/*   all deffacts are deleted.                                 */
/***************************************************************/
globle int DeleteDeftemplate(theDeftemplate)
  VOID *theDeftemplate;
  {
#if BLOAD_ONLY || RUN_TIME
   return(0);
#else
   struct deftemplate *deftemplatePtr;
   struct deftemplate *lastPtr, *currentPtr, *nextPtr;
   
   /*=================================================*/
   /* Return FALSE if deftemplates cannot be deleted. */
   /*=================================================*/
   
#if BLOAD || BLOAD_AND_BSAVE || BLOAD_ONLY
   if (Bloaded() == CLIPS_TRUE) return(0);
#endif

   deftemplatePtr = (struct deftemplate *) theDeftemplate;
   
   /*=========================================================================*/
   /* Delete all deftemplates if the deftemplate pointer is the NULL pointer. */
   /*=========================================================================*/
   
   if (theDeftemplate == NULL)
     {
      ClearDeftemplates();
      return(CLIPS_TRUE);
     }
   
   /*===========================================*/
   /* Search for the deftemplate in the list of */
   /* deftemplates and delete it when found.    */
   /*===========================================*/
   
   lastPtr = NULL;
   currentPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (currentPtr != NULL)
     {
      nextPtr = currentPtr->next;
      if (currentPtr == deftemplatePtr)
        {
         if (lastPtr == NULL)
           { ListOfDeftemplates = nextPtr; }
         else
           { lastPtr->next = nextPtr; }

         if (currentPtr == LastDeftemplate) LastDeftemplate = lastPtr;
         
         DeinstallDeftemplate(currentPtr);
         ReturnSlots(currentPtr->slotList);
         if (currentPtr->pp_form != NULL)
           {
            rm(currentPtr->pp_form,
                    (int) sizeof(char) * ((int) strlen(currentPtr->pp_form) + 1));
           }
         rtn_struct(deftemplate,currentPtr);
         return(CLIPS_TRUE);
        }
  
      lastPtr = currentPtr;
      currentPtr = nextPtr;
     }
     
   return(CLIPS_FALSE);
#endif 
  }
  
/************************************************/
/* SetListOfDeftemplates:  */
/************************************************/
globle VOID SetListOfDeftemplates(dtPtr)
  VOID *dtPtr;
  {
   struct deftemplate *tempPtr;
#if ! RUN_TIME
   struct templateSlot *tsPtr;
#endif
   
   if (dtPtr == NULL)
     {
      tempPtr = ListOfDeftemplates;
      while (tempPtr != NULL)
        {
#if ! RUN_TIME
         DecrementSymbolCount(tempPtr->name);
         for (tsPtr = tempPtr->slotList ; tsPtr != NULL ; tsPtr = tsPtr->next)
           { DecrementSymbolCount(tsPtr->slotName); }
#endif
         RemoveHashDeftemplate(tempPtr);
         tempPtr = tempPtr->next;
        }
     }
     
   ListOfDeftemplates = (struct deftemplate *) dtPtr;
   
   tempPtr = (struct deftemplate *) dtPtr;
   while (tempPtr != NULL)
     {
      AddHashDeftemplate(tempPtr);
      tempPtr = tempPtr->next;
     }
  }
  
/************************************************/
/* QSetListOfDeftemplates:  */
/************************************************/
globle VOID QSetListOfDeftemplates(dtPtr)
  VOID *dtPtr;
  {     
   ListOfDeftemplates = (struct deftemplate *) dtPtr;
  }
  
/**************************************************************************/
/* AddHashDeftemplate:  Adds a deftemplate to the deftemplate hash table. */
/**************************************************************************/
static VOID AddHashDeftemplate(dptr)
  struct deftemplate *dptr;
  {
   struct dhash *newhash, *temp;
   int hashValue;

   newhash = get_struct(dhash);
   newhash->dptr = dptr;

   hashValue = HashSymbol(dptr->name->contents,SIZE_DEFTEMPLATE_HASH);
   
   temp = DeftemplateHashtable[hashValue];
   DeftemplateHashtable[hashValue] = newhash;
   newhash->next = temp;
  }

/*************************************************/
/* RemoveHashDeftemplate:  Removes a deftemplate */
/*   from the deftemplate hash table.            */
/*************************************************/
static BOOLEAN RemoveHashDeftemplate(dptr)
  struct deftemplate *dptr;
  {
   int hashValue;
   struct dhash *hptr, *prev = NULL;

   hashValue = HashSymbol(dptr->name->contents,SIZE_DEFTEMPLATE_HASH);
   
   hptr = DeftemplateHashtable[hashValue];
   while (hptr != NULL)
     {
      if (hptr->dptr == dptr)
        {
         if (prev == NULL)
           { 
            DeftemplateHashtable[hashValue] = hptr->next;
            rtn_struct(dhash,hptr);
            return(1);
           }
         else
           {
            prev->next = hptr->next;
            rtn_struct(dhash,hptr);
            return(1);
           }
        }
      prev = hptr;
      hptr = hptr->next;
     }
   return(0);
  }

/****************************************************************/
/* FindSlot: Finds a specified slot in a deftemplate structure. */
/****************************************************************/
globle struct templateSlot *FindSlot(dtmpl_ptr,name)
  struct deftemplate *dtmpl_ptr;
  SYMBOL_HN *name;
  {
   struct templateSlot *slotPtr;
   
   slotPtr = dtmpl_ptr->slotList;
   while (slotPtr != NULL)
     {
      if (slotPtr->slotName == name)
        { return(slotPtr); }
      slotPtr = slotPtr->next;
     }
     
   return(NULL);
  }
  
/*******************************************************/
/* FindSlotPosition: Finds the position of a specified */
/*   slot in a deftemplate structure.                  */
/*******************************************************/
globle int FindSlotPosition(dtmpl_ptr,name)
  struct deftemplate *dtmpl_ptr;
  SYMBOL_HN *name;
  {
   struct templateSlot *slotPtr;
   int position = 1;
   
   slotPtr = dtmpl_ptr->slotList;
   while (slotPtr != NULL)
     {
      if (slotPtr->slotName == name)
        { return(position); }
      slotPtr = slotPtr->next;
      position++;
     }
     
   return(0);
  }

/***************************************************************/
/* FindSlotItem: Finds a particular slot in a list of slots. */
/***************************************************************/
globle struct slp *FindSlotItem(slot_ptr,slp_ptr)
  struct templateSlot *slot_ptr;
  struct slp *slp_ptr;
  {
   while (slp_ptr != NULL)
     {
      if (slp_ptr->name == slot_ptr->slotName) return (slp_ptr);
      slp_ptr = slp_ptr->next;
     }
  
   return(NULL);
  }
  
/************************************************************/
/* GetDeftemplateName:                                    */
/************************************************************/
globle char *GetDeftemplateName(deftemplatePtr)
  VOID *deftemplatePtr;
  { return(((struct deftemplate *) deftemplatePtr)->name->contents); }
  
/******************************************************************/
/* QFindDeftemplate:  Searches for a deftemplate in the list of    */
/*   deftemplates. Returns a pointer to the deftemplate if found, */
/*   otherwise NULL.                                              */
/******************************************************************/
globle VOID *QFindDeftemplate(df_name)
  SYMBOL_HN *df_name;
  {
   struct dhash *hPtr;
   int hashValue;
   
   hashValue = HashSymbol(df_name->contents,SIZE_DEFTEMPLATE_HASH);
   hPtr = DeftemplateHashtable[hashValue];
   
   while (hPtr != NULL)
     {
      if (hPtr->dptr->name == df_name)
        { return((VOID *) hPtr->dptr); }
     
      hPtr = hPtr->next;
     }

   return(NULL);
  }
  
/******************************************************************/
/* FindDeftemplate:  Searches for a deftemplate in the list of    */
/*   deftemplates. Returns a pointer to the deftemplate if found, */
/*   otherwise NULL.                                              */
/******************************************************************/
globle VOID *FindDeftemplate(df_name)
  char *df_name;
  {
   struct dhash *hPtr;
   int hashValue;
   SYMBOL_HN *findValue;
   
   hashValue = HashSymbol(df_name,SIZE_DEFTEMPLATE_HASH);
   hPtr = DeftemplateHashtable[hashValue];
   
   findValue = (SYMBOL_HN *) FindSymbol(df_name);
   
   while (hPtr != NULL)
     {
      if (hPtr->dptr->name == findValue)
        { return((VOID *) hPtr->dptr); }
     
      hPtr = hPtr->next;
     }

   return(NULL);
  }
  
/**********************************************************/
/* UndeftemplateCommand: Removes a deftemplate statement. */
/*   Syntax: (undeftemplate <deftemplate name>)           */
/**********************************************************/
globle VOID UndeftemplateCommand()
  {
   char *name;
#if (! RUN_TIME) && (! BLOAD_ONLY)
   VOID *deftemplatePtr;
#endif

   name = GetConstructName(1,1,"undeftemplate","deftemplate name");
   if (name == NULL) return;
   
#if (! RUN_TIME) && (! BLOAD_ONLY)
   if (((deftemplatePtr = FindDeftemplate(name)) == NULL) && (strcmp("*",name) != 0))
     { 
      PrintCLIPS(WERROR,"Unable to find deftemplate ");
      PrintCLIPS(WERROR,name);
      PrintCLIPS(WERROR,"\n");
      return;
     }
   else if (DeleteDeftemplate(deftemplatePtr) == CLIPS_FALSE)
     { 
      PrintCLIPS(WERROR,"Unable to delete deftemplate ");
      PrintCLIPS(WERROR,name);
      PrintCLIPS(WERROR,"\n");
      return;
     }

   return;
#else
   PrintCLIPS(WERROR,"Unable to delete deftemplate ");
   PrintCLIPS(WERROR,name);
   PrintCLIPS(WERROR,"\n");
   return;
#endif
  }
  
/**********************************************/
/* GetDeftemplatePPForm: Returns the pretty */ 
/*   print string for the deftemplate.        */
/**********************************************/
globle char *GetDeftemplatePPForm(deftemplatePtr)
  VOID *deftemplatePtr;
  { return(((struct deftemplate *) deftemplatePtr)->pp_form); }

#if DEBUGGING_FUNCTIONS
/*******************************************************/
/* PPDeftemplateCommand: pretty prints a deftemplate.  */
/*   Syntax: (pp_deftemplate <deftemplate name>)       */
/*******************************************************/
globle VOID PPDeftemplateCommand()
  {
   char *deftemplateName;

   deftemplateName = GetConstructName(1,1,"ppdeftemplate","deftemplate name");
   if (deftemplateName == NULL) return;

   PPDeftemplate(deftemplateName,WDISPLAY);

   return;
  }
  
/******************************************************/
/* PPDeftemplate: the driver which actually does the */
/*   pretty printing of the deftemplate.              */
/******************************************************/
globle int PPDeftemplate(df_name,fileid)
  char *df_name, *fileid;
  {
   VOID *deftemplatePtr;

   deftemplatePtr = FindDeftemplate(df_name);
   if (deftemplatePtr == NULL)        
     {
      PrintCLIPS(WERROR,"Unable to find deftemplate ");
      PrintCLIPS(WERROR,df_name);
      PrintCLIPS(WERROR,"\n");
      return(CLIPS_FALSE);
     }
     
   if (GetDeftemplatePPForm(deftemplatePtr) == NULL) return(CLIPS_TRUE);
   PrintInChunks(fileid,GetDeftemplatePPForm(deftemplatePtr));
   return(CLIPS_TRUE);
  }
  
/****************************************************/
/* ListDeftemplatesCommand: Interface glue function */ 
/*   for displaying the list of deftemplates.       */
/*   Syntax: (list-deftemplates)                    */
/****************************************************/
globle VOID ListDeftemplatesCommand()
  {
   if (ArgCountCheck("list-deftemplates",EXACTLY,0) == -1) return;

   ListDeftemplates();
  }

/*********************************************************/
/* ListDeftemplates: Displays the list of deftemplates. */
/*********************************************************/
globle VOID ListDeftemplates()
  {
   VOID *dr_ptr;
   long count = 0;

   dr_ptr = GetNextDeftemplate(NULL);
   while (dr_ptr != NULL)
     {      
      if (GetHaltExecution() == CLIPS_TRUE) return;

      PrintCLIPS(WDISPLAY,GetDeftemplateName(dr_ptr));
      PrintCLIPS(WDISPLAY,"\n");
      dr_ptr = GetNextDeftemplate(dr_ptr);
      count++;
     }
     
   PrintTally(WDISPLAY,count,"deftemplate","deftemplates");

   return;
  }

#endif
 
/*******************************************************************/
/* PrintTemplateFact: Prints a fact using the deftemplate format.  */
/*   Returns CLIPS_TRUE if the fact was printed using this format, */
/*   otherwise CLIPS_FALSE.                                        */
/*******************************************************************/
globle int PrintTemplateFact(logicalName,fact_ptr)
  char *logicalName;
  struct fact *fact_ptr;
  {
   struct element *sublist;
   int length, i;
   struct deftemplate *dtmpl_ptr;
   struct templateSlot *slot_ptr;
   
   sublist = fact_ptr->atoms;
   length = fact_ptr->factLength;
   
   /*========================================================*/
   /* If the first field of the fact is not a symbol, then the */
   /* fact cannot be printed using the deftemplate format.   */
   /*========================================================*/
   
   if (sublist[0].type != SYMBOL) return(CLIPS_FALSE);
   
   /*========================================================*/
   /* If the deftemplate corresponding to the first field of */
   /* of the fact cannot be found, then the fact cannot be   */
   /* printed using the deftemplate format.                  */
   /*========================================================*/
   
   dtmpl_ptr = (struct deftemplate *) FindDeftemplate(ValueToString(sublist[0].value));
   if (dtmpl_ptr == NULL) return(CLIPS_FALSE);
   
   /*=============================================*/
   /* Print the relation name of the deftemplate. */
   /*=============================================*/
   
   PrintCLIPS(logicalName,dtmpl_ptr->name->contents);
   PrintCLIPS(logicalName," ");
   
   /*===================================================*/
   /* Print each of the field slots of the deftemplate. */
   /*===================================================*/
   
   slot_ptr = dtmpl_ptr->slotList;
   
   i = 1;
   while (slot_ptr != NULL)
     {    
      /*===========================================*/
      /* Print the closing parenthesis of the slot */
      /* and the slot name.                        */
      /*===========================================*/
      
      PrintCLIPS(logicalName,"(");
      PrintCLIPS(logicalName,slot_ptr->slotName->contents);
     
      /*======================================================*/
      /* Print the value of the slot for a single field slot. */
      /*======================================================*/
      
      if (slot_ptr->multiSlot == CLIPS_FALSE)
        {
         PrintCLIPS(logicalName," ");
         if (i < length)
           { PrintAtom(logicalName,sublist[i].type,sublist[i].value); }
         else
           { PrintCLIPS(logicalName,"<<<MISSING FIELD>>>"); }
         i++;
        }
        
      /*==========================================================*/
      /* Else print the value of the slot for a multi field slot. */
      /*==========================================================*/
      
      else
        {
         while (i < length)
           {
            PrintCLIPS(logicalName," ");
            PrintAtom(logicalName,sublist[i].type,sublist[i].value);
            i++;
           }
        }
        
      /*============================================*/
      /* Print the closing parenthesis of the slot. */
      /*============================================*/
      
      PrintCLIPS(logicalName,")");
      slot_ptr = slot_ptr->next;
      if (slot_ptr != NULL) PrintCLIPS(logicalName," ");
     }
     
    
    /*========================================================*/
    /* Print any extraneous values placed in the deftemplate. */
    /*========================================================*/
      
    while (i < length)
      {
       PrintCLIPS(logicalName," ");
       PrintAtom(logicalName,sublist[i].type,sublist[i].value);
       i++;
      }
      
   /*==========================================*/
   /* Return CLIPS_TRUE indicating that the fact was */
   /* printed using the deftemplate format.    */
   /*==========================================*/
   
   return(CLIPS_TRUE);
  }

/****************************************************/ 
/* CheckSlotType:                                   */
/****************************************************/ 
globle BOOLEAN CheckSlotType(type,slotPtr)
  int type;
  struct templateSlot *slotPtr;
  {
   if ((slotPtr->anyAllowed == CLIPS_FALSE) &&
       (((type == INTEGER) && (slotPtr->integersAllowed != CLIPS_TRUE)) ||
        ((type == FLOAT) && (slotPtr->floatsAllowed != CLIPS_TRUE)) ||
        ((type == STRING) && (slotPtr->stringsAllowed != CLIPS_TRUE)) ||
        ((type == EXTERNAL_ADDRESS) && (slotPtr->addressesAllowed != CLIPS_TRUE)) ||
#if OBJECT_SYSTEM
        (((type == INSTANCE_NAME) || (type == INSTANCE)) && (slotPtr->instancesAllowed != CLIPS_TRUE)) ||
#endif
        ((type == SYMBOL) && (slotPtr->symbolsAllowed != CLIPS_TRUE))))
     { return(CLIPS_FALSE); }
       
   return(CLIPS_TRUE);
  }
  
/****************************************************/ 
/* CheckSlotAllowedValues:                          */
/****************************************************/ 
globle BOOLEAN CheckSlotAllowedValues(type,vPtr,slotPtr)
  int type;
  VOID *vPtr;
  struct templateSlot *slotPtr;
  {
   struct expr *tmpPtr;
   switch (type)
     {
      case SYMBOL:
#if OBJECT_SYSTEM
      case INSTANCE_NAME:
#endif
        if ((slotPtr->symbolRestriction == CLIPS_FALSE) && (slotPtr->anyRestriction == CLIPS_FALSE))
          return(CLIPS_TRUE);
        break;
      
      case STRING:
        if ((slotPtr->stringRestriction == CLIPS_FALSE) && (slotPtr->anyRestriction == CLIPS_FALSE))
          return(CLIPS_TRUE);
        break;
      
      case INTEGER:
        if ((slotPtr->integerRestriction == CLIPS_FALSE) && (slotPtr->anyRestriction == CLIPS_FALSE))
          return(CLIPS_TRUE);
        break;
      
      case FLOAT:
        if ((slotPtr->floatRestriction == CLIPS_FALSE) && (slotPtr->anyRestriction == CLIPS_FALSE))
          return(CLIPS_TRUE);
        break;
      
      default:
        return(CLIPS_TRUE);
     }
     
   tmpPtr = slotPtr->restrictionList;
   while (tmpPtr != NULL)
     {
      if ((tmpPtr->type == type) && (tmpPtr->value == vPtr)) return(CLIPS_TRUE);
      
      else if (slotPtr->numberRestriction)
        {
         if ((tmpPtr->type == INTEGER) && (type == FLOAT))
           { if (((double) ValueToLong(tmpPtr->value)) == ValueToDouble(vPtr)) return(CLIPS_TRUE); }
        
         else if ((tmpPtr->type == FLOAT) && (type == INTEGER))
           { if (((double) ValueToLong(vPtr)) == ValueToDouble(tmpPtr->value)) return(CLIPS_TRUE); }
        }
        
      tmpPtr = tmpPtr->next_arg;
     }
     
   return(CLIPS_FALSE);
  }
 
/****************************************************/ 
/* CheckSlotRange:    */
/****************************************************/ 
globle BOOLEAN CheckSlotRange(type,vPtr,slotPtr)
  int type;
  VOID *vPtr;
  struct templateSlot *slotPtr;
  {
   if ((type != INTEGER) && (type != FLOAT)) return(CLIPS_TRUE);
   
   /*==================================*/
   /* Check against the minimum value. */
   /*==================================*/

   if (slotPtr->minValue != NULL)
     {
      if ((type == INTEGER) && (slotPtr->minValue->type == INTEGER))
        {
         if (ValueToLong(vPtr) < ValueToLong(slotPtr->minValue->value))
           { return(CLIPS_FALSE); }
        }
      else if ((type == INTEGER) && (slotPtr->minValue->type == FLOAT))
        {
         if (((double) ValueToLong(vPtr)) < ValueToDouble(slotPtr->minValue->value))
           { return(CLIPS_FALSE); }
        }
      else if ((type == FLOAT) && (slotPtr->minValue->type == INTEGER))
        {
         if (ValueToDouble(vPtr) < ((double) ValueToLong(slotPtr->minValue->value)))
           { return(CLIPS_FALSE); }
        
        }
      else if ((type == FLOAT) && (slotPtr->minValue->type == FLOAT))
        {
         if (ValueToDouble(vPtr) < ValueToDouble(slotPtr->minValue->value))
           { return(CLIPS_FALSE); }
        }
     }
     
   /*==================================*/
   /* Check against the maximum value. */
   /*==================================*/

   if (slotPtr->maxValue != NULL)
     {
      if ((type == INTEGER) && (slotPtr->maxValue->type == INTEGER))
        {
         if (ValueToLong(vPtr) > ValueToLong(slotPtr->maxValue->value))
           { return(CLIPS_FALSE); }
        }
      else if ((type == INTEGER) && (slotPtr->maxValue->type == FLOAT))
        {
         if (((double) ValueToLong(vPtr)) > ValueToDouble(slotPtr->maxValue->value))
           { return(CLIPS_FALSE); }
        }
      else if ((type == FLOAT) && (slotPtr->maxValue->type == INTEGER))
        {
         if (ValueToDouble(vPtr) > ((double) ValueToLong(slotPtr->maxValue->value)))
           { return(CLIPS_FALSE); }
        
        }
      else if ((type == FLOAT) && (slotPtr->maxValue->type == FLOAT))
        {
         if (ValueToDouble(vPtr) > ValueToDouble(slotPtr->maxValue->value))
           { return(CLIPS_FALSE); }
        }
     }
     
   return(CLIPS_TRUE);
  }
   
/**********************/
/* ModifyCommand:     */
/**********************/
globle VOID ModifyCommand()
  {
   DuplicateModifyCommand(CLIPS_TRUE);
  }
  
/**********************/
/* DuplicateCommand:  */
/**********************/
globle VOID DuplicateCommand()
  {
   DuplicateModifyCommand(CLIPS_FALSE);
  }
  
/************************************************************/
/* DuplicateModifyCommand:                                  */
/************************************************************/
static VOID DuplicateModifyCommand(retractIt)
  int retractIt;
  {
   long int factNum;
   struct fact *oldFact, *newFact;
   struct expr *testPtr;
   DATA_OBJECT computeResult;
   struct deftemplate *templatePtr;
   struct element *sublist;
   struct templateSlot *slotPtr;  
   int multiSlotPosition, i, position, found;

   /*=============================================*/
   /* Get a pointer to the fact to be duplicated. */
   /*=============================================*/
   
   testPtr = GetFirstArgument();

   EvaluateExpression(testPtr,&computeResult);
   if (computeResult.type == INTEGER)
     {
      factNum = ValueToLong(computeResult.value);
      if (factNum < 0)
        { 
         ExpectedTypeError("modify/duplicate",1,"fact-address or fact-index");
         SetEvaluationError(CLIPS_TRUE);
         return;
        }
        
      oldFact = (struct fact *) GetNextFact(NULL);
      while (oldFact != NULL)
        {
         if (oldFact->factIndex == factNum)
           { break; }
         else
           { oldFact = oldFact->next; }
        }
         
      if (oldFact == NULL)
        {
         PrintCLIPS(WERROR,"Unable to find fact f-");
         PrintLongInteger(WERROR,factNum);
         PrintCLIPS(WERROR,"\n");
         return;
        }
     }
   else if (computeResult.type == FACT_ADDRESS)
     { oldFact = (struct fact *) computeResult.value; }
   else
     {
      ExpectedTypeError("modify/duplicate",1,"fact-address or fact-index");
      SetEvaluationError(CLIPS_TRUE);
      return;
     }
     
   testPtr = testPtr->next_arg;

   /*===============================*/
   /* See if it is a template fact. */
   /*===============================*/
       
   sublist = oldFact->atoms;
   if (sublist[0].type != SYMBOL) templatePtr = NULL;
   else templatePtr = (struct deftemplate *) QFindDeftemplate((SYMBOL_HN *) sublist[0].value);
     
   if (templatePtr == NULL)
     {
      ExpectedTypeError("modify/duplicate",1,"deftemplate fact");
      SetEvaluationError(CLIPS_TRUE);
      return;
     }
     
   /*===================================*/
   /* Determine the multislot position. */
   /*===================================*/
   
   if (templatePtr->hasMultiSlot) multiSlotPosition = templatePtr->numberOfSlots;
   else multiSlotPosition = 0;
   
   /*=====================*/
   /* Duplicate the fact. */
   /*=====================*/
   
   newFact = (struct fact *) CreateFact(oldFact->factLength);
   for (i = 0; i < oldFact->factLength; i++) 
     {
      newFact->atoms[i].type = oldFact->atoms[i].type;
      newFact->atoms[i].value = oldFact->atoms[i].value;
     }
     
   /*========================*/
   /* Start replacing slots. */
   /*========================*/
   
   while (testPtr != NULL)
     {
      /*============================================================*/
      /* If the slot identifier is an integer, then the slot was    */
      /* previously identified and its position within the template */
      /* was stored. Otherwise, the position of the slot within the */
      /* deftemplate has to be determined by comparing the name of  */
      /* the slot against the list of slots for the deftemplate.    */
      /*============================================================*/
      
      if (testPtr->type == INTEGER)
        { position = (int) ValueToLong(testPtr->value); }
      else
        {
         found = CLIPS_FALSE;
         position = 1;
         slotPtr = templatePtr->slotList;
         while (slotPtr != NULL)
           {
            if (slotPtr->slotName == (SYMBOL_HN *) testPtr->value)
              { 
               found = CLIPS_TRUE;
               slotPtr = NULL; 
              }
            else
              {
               slotPtr = slotPtr->next;
               position++;
              }
           }
           
         if (! found)
           {
            PrintCLIPS(WERROR,"\nInvalid slot ");
            PrintCLIPS(WERROR,(ValueToString(testPtr->value)));
            PrintCLIPS(WERROR," not defined in corresponding deftemplate ");
            PrintCLIPS(WERROR,ValueToString(templatePtr->name));
            PrintCLIPS(WERROR,"\n");
            SetEvaluationError(CLIPS_TRUE);
            ReturnElements(newFact);
            return;
           }
        }
        
      /*===================================================*/
      /* If a single field slot is being replaced, then... */
      /*===================================================*/
        
      if (multiSlotPosition != position)
        {
         if ((testPtr->arg_list == NULL) ? TRUE : (testPtr->arg_list->next_arg != NULL))
           {
            ReturnElements(newFact);
            MultiIntoSingleFieldSlotError(position,templatePtr->name); 
            return;
           }
           
         EvaluateExpression(testPtr->arg_list,&computeResult);
         SetEvaluationError(CLIPS_FALSE);
         
         /* Multifield values and fact-addresses cannot be place in single field slots. */
         if (computeResult.type == MULTIFIELD)
           {
            ReturnElements(newFact);
            MultiIntoSingleFieldSlotError(position,templatePtr->name); 
            return;
           }
         else if (computeResult.type == FACT_ADDRESS)
           {
            PrintCLIPS(WERROR,"ERROR: Fact-addresses may not be asserted within facts\n");
            SetEvaluationError(CLIPS_TRUE);
            ReturnElements(newFact);
            return;
           }
         
         if (position < newFact->factLength)
           {
            newFact->atoms[position].type = computeResult.type;
            newFact->atoms[position].value = computeResult.value;
           }
         else
           {
            PrintCLIPS(WERROR,"\nInvalid slot <unknown>");
            PrintCLIPS(WERROR," not defined in corresponding deftemplate ");
            PrintCLIPS(WERROR,ValueToString(templatePtr->name));
            PrintCLIPS(WERROR,"\n");
            SetEvaluationError(CLIPS_TRUE);
            ReturnElements(newFact);
            return;
           }
        }
        
      /*=================================*/
      /* Else replace a multifield slot. */
      /*=================================*/
      
      else
        {
         newFact = TemplateMultifieldSlotReplace(testPtr,newFact,templatePtr);
         if (newFact == NULL) return;
        }
            
      testPtr = testPtr->next_arg;
     }
        
   if (retractIt) RetractFact(oldFact);
   AddFact(newFact);
   return;
  }
  
/********************************************************************/
/* TemplateMultifieldSlotReplace: Replaces the multifield slot of a */
/*   deftemplate fact as part of the modify/duplicate command.      */
/********************************************************************/
static struct fact *TemplateMultifieldSlotReplace(testPtr,newFact,templatePtr)
  struct expr *testPtr;
  struct fact *newFact;
  struct deftemplate *templatePtr;
  {
   struct expr *tempTest;
   struct lelement *tempElem, *multiHead, *lastMulti;
   int multiSlotSize, extent;
   DATA_OBJECT computeResult;
   struct element *sublist;
   int i;
   struct fact *tempFact;
  
   /*========================================================*/
   /* Construct the list of values to be placed in the slot. */
   /*========================================================*/
   
   multiSlotSize = 0;
   multiHead = lastMulti = NULL;
   tempTest = testPtr->arg_list;
   while (tempTest != NULL)
     {
      EvaluateExpression(tempTest,&computeResult);
      SetEvaluationError(CLIPS_FALSE);
                    
      if (GetType(computeResult) == MULTIFIELD)
        {
         extent = (computeResult.end - computeResult.begin) + 1;
         multiSlotSize += extent;
  
         sublist = ((struct fact *) computeResult.value)->atoms;
         for (i = computeResult.begin; i <= computeResult.end; i++)
           {
            if (sublist[i].type == FACT_ADDRESS)
              {
               PrintCLIPS(WERROR,"ERROR: Fact-addresses may not be asserted within facts\n");
               SetEvaluationError(CLIPS_TRUE);
               ReturnElements(newFact);
               while (multiHead != NULL)
                 {
                  tempElem = multiHead->next;
                  rtn_struct(lelement,multiHead);
                  multiHead = tempElem;
                 }
               return(NULL);
              } 
            
            tempElem = get_struct(lelement);
            tempElem->type = sublist[i].type;
            tempElem->value = sublist[i].value;
            tempElem->next = NULL;
               
            if (multiHead == NULL) multiHead = tempElem;
            else lastMulti->next = tempElem;
               
            lastMulti = tempElem;
           }
        }      
      else
        {
         if (computeResult.type == FACT_ADDRESS)
           {
            PrintCLIPS(WERROR,"ERROR: Fact-addresses may not be asserted within facts\n");
            SetEvaluationError(CLIPS_TRUE);
            ReturnElements(newFact);
            while (multiHead != NULL)
              {
               tempElem = multiHead->next;
               rtn_struct(lelement,multiHead);
               multiHead = tempElem;
              }
            return(NULL);
           } 
                 
         multiSlotSize += 1;
         tempElem = get_struct(lelement);
         tempElem->type = computeResult.type;
         tempElem->value = computeResult.value;
         tempElem->next = NULL;
            
         if (multiHead == NULL) multiHead = tempElem;
         else lastMulti->next = tempElem;
               
         lastMulti = tempElem;
        }
           
      tempTest = tempTest->next_arg;
     }
           
   /*==========================================================*/
   /* If the number of values in the newly created fact cannot */
   /* exactly hold the value for the multifield slot, then     */
   /* create another fact of the exact size and recopy values. */
   /*==========================================================*/
         
   if (newFact->factLength != (multiSlotSize + templatePtr->numberOfSlots))
     {
      tempFact = (struct fact *) CreateFact(multiSlotSize + templatePtr->numberOfSlots);
      for (i = 0; i < templatePtr->numberOfSlots; i++) 
        {
         tempFact->atoms[i].type = newFact->atoms[i].type;
         tempFact->atoms[i].value = newFact->atoms[i].value;
        }
              
      ReturnElements(newFact);
      newFact = tempFact;
     }
            
   /*===========================================*/
   /* Copy the values into the multifield slot. */
   /*===========================================*/
            
   for (i = templatePtr->numberOfSlots; i < (templatePtr->numberOfSlots + multiSlotSize); i++) 
     {
      newFact->atoms[i].type = multiHead->type;
      newFact->atoms[i].value = multiHead->value;
      tempElem = multiHead->next;
      rtn_struct(lelement,multiHead);
      multiHead = tempElem;
     }
   
   return(newFact);
  }
    
/************************************************************/
/* SDDCommand:                        */
/************************************************************/
globle int SDDCommand()
  {
   int oldValue;
   DATA_OBJECT arg_ptr;
   
   oldValue = GetDynamicDeftemplateChecking();
   
   if (ArgCountCheck("set-dynamic-deftemplate-checking",EXACTLY,1) == -1) 
     { return(oldValue); }
     
   RtnUnknown(1,&arg_ptr);
     
   if ((arg_ptr.value == (VOID *) CLIPSFalseSymbol) && (arg_ptr.type == SYMBOL))
     { SetDynamicDeftemplateChecking(CLIPS_FALSE); }
   else
     { SetDynamicDeftemplateChecking(CLIPS_TRUE); }
     
   return(oldValue);
  }
  
/************************************************************/
/* GDDCommand:                        */
/************************************************************/
globle int GDDCommand()
  {
   int oldValue;
   
   oldValue = GetDynamicDeftemplateChecking();
   
   if (ArgCountCheck("get-dynamic-deftemplate-checking",EXACTLY,0) == -1) 
     { return(oldValue); }
     
   return(oldValue);
  }
  
/**************************************************************/
/* SetDynamicDeftemplateChecking:  */    
/**************************************************************/
globle BOOLEAN SetDynamicDeftemplateChecking(value)
  int value;
  { 
   int ov;
   
   ov = DynamicDeftemplateChecking;
   DynamicDeftemplateChecking = value;
   return(ov);
  }

/**************************************************************/
/* GetDynamicDeftemplateChecking:  */    
/**************************************************************/
globle BOOLEAN GetDynamicDeftemplateChecking()
  { return(DynamicDeftemplateChecking); }
  
/*********************************************************/
/* GetNextDeftemplate: If passed NULL, returns the first */
/*   deftemplate, otherwise the deftemplate after the    */
/*   deftemplate passed as an argument.                  */
/*********************************************************/
globle VOID *GetNextDeftemplate(deftemplatePtr)
  VOID *deftemplatePtr;
  {
   if (deftemplatePtr == NULL)
     { return((VOID *) ListOfDeftemplates); }
   else
     { return((VOID *) (((struct deftemplate *) deftemplatePtr)->next)); }
  }
  
/*********************************************************/
/* IsDeftemplateDeletable: Returns TRUE if a particular  */
/*   deftemplate can be deleted otherwise returns FALSE. */
/*********************************************************/
globle BOOLEAN IsDeftemplateDeletable(ptr)
  VOID *ptr;
  {
#if MAC_MPW
#pragma unused(ptr)
#endif
#if BLOAD_ONLY || RUN_TIME
   return(CLIPS_FALSE);
#else
#if BLOAD || BLOAD_AND_BSAVE
   if (Bloaded()) return(CLIPS_FALSE);
#endif
   return(CLIPS_TRUE);
#endif
  }
  
#if BLOAD_AND_BSAVE 

/*******************************/
/* BsaveFind:       */
/*******************************/
static VOID BsaveFind()
  {
   struct deftemplate *dtPtr;
   struct templateSlot *tsPtr;
  
   dtPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dtPtr != NULL)
     {
      dtPtr->name->neededSymbol = CLIPS_TRUE;
      tsPtr = dtPtr->slotList;
      while (tsPtr != NULL)
        {  
         tsPtr->slotName->neededSymbol = CLIPS_TRUE;
         ExpressionCount += (ExpressionSize(tsPtr->defaultList) +
                             ExpressionSize(tsPtr->restrictionList) +
                             ExpressionSize(tsPtr->minValue) +
                             ExpressionSize(tsPtr->maxValue));
         MarkNeededItems(tsPtr->defaultList);
         MarkNeededItems(tsPtr->restrictionList);
         MarkNeededItems(tsPtr->minValue);
         MarkNeededItems(tsPtr->maxValue);
         tsPtr = tsPtr->next;
        }
      dtPtr = (struct deftemplate *) GetNextDeftemplate(dtPtr);
     }
  }

/******************************************/
/* BsaveExpressions:            */
/******************************************/
static VOID BsaveExpressions(fp)
  FILE *fp;
  {
   struct deftemplate *dtPtr;
   struct templateSlot *tsPtr;
  
   dtPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dtPtr != NULL)
     {
      tsPtr = dtPtr->slotList;
      while (tsPtr != NULL)
        {  
         BsaveExpression(tsPtr->defaultList,fp);
         BsaveExpression(tsPtr->restrictionList,fp);
         BsaveExpression(tsPtr->minValue,fp);
         BsaveExpression(tsPtr->maxValue,fp);
         tsPtr = tsPtr->next;
        }
      dtPtr = (struct deftemplate *) GetNextDeftemplate(dtPtr);
     }
  }

/*************************************************/
/* BsaveDeftemplates: Writes out all deftemplate */
/*   structures to the binary file.              */
/*************************************************/
static VOID BsaveDeftemplates(fp)
  FILE *fp;
  {
   unsigned long int count = 0, space;
   struct deftemplate *dtPtr, tempDeftemplate;
   struct templateSlot *tsPtr, tempTemplateSlot;
   unsigned long int numberOfTemplateSlots = 0;
   
   /*=======================================*/
   /* Write out the number of deftemplates. */
   /*=======================================*/
  
   dtPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dtPtr != NULL)
     {
      count++;
      tsPtr = dtPtr->slotList;
      while (tsPtr != NULL)
        {
         numberOfTemplateSlots++;
         tsPtr = tsPtr->next;
        }
      dtPtr = (struct deftemplate *) GetNextDeftemplate(dtPtr);
     }
     
   space = (count * sizeof(struct deftemplate)) + 
           (numberOfTemplateSlots * sizeof(struct templateSlot)) +
           (2 * sizeof(unsigned long));
   GenWrite(&space,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&count,(unsigned long) sizeof(unsigned long int),fp);
   GenWrite(&numberOfTemplateSlots,(unsigned long) sizeof(unsigned long int),fp);
   if (count == 0) return;
     
   /*=============================*/
   /* Write out each deftemplate. */
   /*=============================*/
   
   numberOfTemplateSlots = 0;
   dtPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dtPtr != NULL)
     {
      tempDeftemplate.name = (SYMBOL_HN *) dtPtr->name->bucket;
      tempDeftemplate.pp_form = NULL;
      tempDeftemplate.hasMultiSlot = dtPtr->hasMultiSlot;
      tempDeftemplate.numberOfSlots = dtPtr->numberOfSlots;
      
      if (dtPtr->slotList != NULL) 
        { tempDeftemplate.slotList = (struct templateSlot *) numberOfTemplateSlots; }
      else tempDeftemplate.slotList = (struct templateSlot *) -1L;
      
      if (dtPtr->next != NULL) tempDeftemplate.next = NULL;
      else tempDeftemplate.next = (struct deftemplate *) -1L;
      
      GenWrite(&tempDeftemplate,(unsigned long) sizeof(struct deftemplate),fp);
      
      tsPtr = dtPtr->slotList;
      while (tsPtr != NULL)
        {
         numberOfTemplateSlots++;
         tsPtr = tsPtr->next;
        }
           
      dtPtr = (struct deftemplate *) GetNextDeftemplate(dtPtr);
     }
     
   /*=============================*/
   /* Write out template slots. */
   /*=============================*/
   
   dtPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dtPtr != NULL)
     {
      tsPtr = dtPtr->slotList;
      while (tsPtr != NULL)
        {
         genmemcpy(&tempTemplateSlot,tsPtr,(unsigned long) sizeof(struct templateSlot));

         tempTemplateSlot.slotName = (SYMBOL_HN *) tsPtr->slotName->bucket;
         
         if (tsPtr->defaultList != NULL)
           { 
            tempTemplateSlot.defaultList = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tsPtr->defaultList); 
           }
         else
           { tempTemplateSlot.defaultList = (struct expr *) -1L; }
         
         if (tsPtr->restrictionList != NULL)
           { 
            tempTemplateSlot.restrictionList = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tsPtr->restrictionList); 
           }
         else
           { tempTemplateSlot.restrictionList = (struct expr *) -1L; }
         
         if (tsPtr->minValue != NULL)
           { 
            tempTemplateSlot.minValue = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tsPtr->minValue); 
           }
         else
           { tempTemplateSlot.minValue = (struct expr *) -1L; }
         
         if (tsPtr->minValue != NULL)
           { 
            tempTemplateSlot.maxValue = (struct expr *) ExpressionCount;
            ExpressionCount += ExpressionSize(tsPtr->maxValue); 
           }
         else
           { tempTemplateSlot.maxValue = (struct expr *) -1L; }
           
         if (tsPtr->next != NULL) tempTemplateSlot.next = NULL;
         else tempTemplateSlot.next = (struct templateSlot *) -1L;
         
         GenWrite(&tempTemplateSlot,(unsigned long) sizeof(struct templateSlot),fp);
         tsPtr = tsPtr->next;
        }
           
      dtPtr = (struct deftemplate *) GetNextDeftemplate(dtPtr);
     }
  }
#endif


#if (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)

/************************************************************************/
/* BloadDeftemplates: Loads the deftemplates used by this binary image. */
/************************************************************************/
static VOID BloadDeftemplates()
  {
   struct deftemplate *dtPtr;
   struct templateSlot *tsPtr;
   unsigned long int space, i;
   long int index;
   
   /*=================================*/
   /* Get the number of deftemplates. */
   /*=================================*/
  
   GenRead(&space,(unsigned long) sizeof(unsigned long int));
   GenRead(&NumberOfDeftemplates,(unsigned long) sizeof(unsigned long int));
   GenRead(&NumberOfTemplateSlots,(unsigned long) sizeof(unsigned long int));
   if (NumberOfDeftemplates == 0)
     {
      DeftemplateArray = NULL;
      SlotArray = NULL;
      return;
     }
     
   /*===========================*/
   /* Read in the deftemplates. */
   /*===========================*/
   
   space = NumberOfDeftemplates * sizeof(struct deftemplate);
   DeftemplateArray = (struct deftemplate HUGE_ADDR *) genlongalloc(space);
   GenRead((VOID *) DeftemplateArray,space);
   
   space =  NumberOfTemplateSlots * sizeof(struct templateSlot);
   SlotArray = (struct templateSlot HUGE_ADDR *) genlongalloc(space);
   GenRead((VOID *) SlotArray,space);
           
   /*=============================*/
   /* Reset deftemplate pointers. */
   /*=============================*/

   for (i = 0; i < NumberOfDeftemplates; i++)
     {
      dtPtr = (struct deftemplate *) &DeftemplateArray[i];
      
      dtPtr->name = SymbolPointer(dtPtr->name);
      IncrementCount(dtPtr->name);
      
      /*==================================================*/
      /* Set up the slot list pointer of the deftemplate. */
      /*==================================================*/
      
      index = (long int) dtPtr->slotList;
      if (index != -1L)
        { dtPtr->slotList = (struct templateSlot *) &SlotArray[index]; }
      else
        { dtPtr->slotList = NULL; }
     
      /*====================================================*/
      /* Set the next pointer of the deftemplate structure. */
      /*====================================================*/
      
      index = (long int) dtPtr->next;
      if (index != -1L)
        { dtPtr->next = (struct deftemplate *) &DeftemplateArray[i+1]; }
      else
        { dtPtr->next = NULL; }
     }
      
   /*==========================================*/
   /* Reset the pointers in the list of slots. */
   /*==========================================*/
      
   for (i = 0; i < NumberOfTemplateSlots; i++)
     {
      tsPtr = (struct templateSlot *) &SlotArray[i];
 
      tsPtr->slotName = SymbolPointer(tsPtr->slotName);
      IncrementCount(tsPtr->slotName);
      tsPtr->defaultList = ExpressionPointer(tsPtr->defaultList);
      tsPtr->restrictionList = ExpressionPointer(tsPtr->restrictionList);
      tsPtr->minValue = ExpressionPointer(tsPtr->minValue);
      tsPtr->maxValue = ExpressionPointer(tsPtr->maxValue);
        
      index = (long int) tsPtr->next;
      if (index != -1L)
        { tsPtr->next = (struct templateSlot *) &SlotArray[i+1]; }
      else
        { tsPtr->next = NULL; }
     }
     
   /*============================================*/
   /* Set the deftemplates to the new deftemplates list. */
   /*============================================*/
   
   SetListOfDeftemplates((VOID *) DeftemplateArray);
  }
  
/****************************************************************/
/* ClearBload: */
/****************************************************************/
static VOID ClearBload()
  {
   unsigned long int space;
   
   SetListOfDeftemplates(NULL);
   
   space = NumberOfDeftemplates * sizeof(struct deftemplate);
   if (space != 0) genlongfree((VOID *) DeftemplateArray,space);
   
   space =  NumberOfTemplateSlots * sizeof(struct templateSlot);
   if (space != 0) genlongfree((VOID *) SlotArray,space);
  }
#endif

#if CONSTRUCT_COMPILER && (! RUN_TIME)

/******************************************************************/
/* InitDeftemplatesCode:   */
/******************************************************************/
static VOID InitDeftemplatesCode(initFP,imageID,maxIndices)
  FILE *initFP;
  int imageID;
  int maxIndices;
  {
#if MAC_MPW
#pragma unused(maxIndices)
#endif
   if (GetNextDeftemplate(NULL) != NULL)
     { fprintf(initFP,"   SetListOfDeftemplates(dtm%d_1);\n",imageID); }
   else
     { fprintf(initFP,"   SetListOfDeftemplates(NULL);\n"); }
  }

/******************************************************************/
/* DeftemplatesToCode:  Produces all the deftemplate definitions. */
/******************************************************************/
static int DeftemplatesToCode(fileName,fileID,headerFP,imageID,maxIndices)
  char *fileName;
  int fileID;
  FILE *headerFP;
  int imageID;
  int maxIndices;
  {
   struct deftemplate *dfPtr;
   struct templateSlot *slotPtr;
   FILE *fp;
   int version = 1;
   int arrayVersion = 1;
   int i = 1;
   int newHeader = CLIPS_TRUE, slotCount = 0;
   
   /*==================*/
   /* Create the file. */
   /*==================*/
   
   fprintf(headerFP,"#include \"deftempl.h\"\n");
   if (GetNextDeftemplate(NULL) == NULL) return(-1);
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
   
   /*==================================*/
   /* Dump the deftemplate structures. */
   /*==================================*/
   
   dfPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   while (dfPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct deftemplate dtm%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct deftemplate dtm%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      fprintf(fp,"{"); 
      PrintSymbolReference(fp,dfPtr->name); /* name */
      fprintf(fp,",NULL,");                 /* ppform */
      
      if (dfPtr->slotList == NULL)
        { fprintf(fp,"NULL,"); }
      else
        {
         fprintf(fp,"&tfl%d_%d[%d],",imageID,
                                     (slotCount / maxIndices) + 1,
                                     slotCount % maxIndices); 
         slotPtr = dfPtr->slotList;
         while (slotPtr != NULL)
           {
            slotCount++;
            slotPtr = slotPtr->next;
           } 
        }
      
      fprintf(fp,"%d,%d,",dfPtr->hasMultiSlot,dfPtr->numberOfSlots);

      if (dfPtr->next == NULL)
        { fprintf(fp,"NULL"); }
      else
        { 
         if ((i + 1) > maxIndices)
           { fprintf(fp,"&dtm%d_%d[%d]",imageID,version + 1,0); }
         else
           { fprintf(fp,"&dtm%d_%d[%d]",imageID,version,i); }
        }
        
      i++;
      dfPtr = dfPtr->next;
      if ((i > maxIndices) || (dfPtr == NULL))
        {
         fprintf(fp,"}};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (dfPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,"},\n"); }
     }
     
   /*===========================*/
   /* Dump the slot structures. */
   /*===========================*/
     
   if (slotCount == 0) return(1);
   
   if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
   newHeader = CLIPS_TRUE;
   
   slotCount = 0;
   arrayVersion = 1;
   i = 1;
   
   dfPtr = (struct deftemplate *) GetNextDeftemplate(NULL);
   slotPtr = dfPtr->slotList;
   while ((slotPtr == NULL) && (dfPtr != NULL))
     {
      dfPtr = dfPtr->next;
      if (dfPtr != NULL) slotPtr = dfPtr->slotList;
     }
     
   while (slotPtr != NULL)
     {
      if (newHeader)
        {
         fprintf(fp,"struct templateSlot tfl%d_%d[] = {\n",imageID,arrayVersion);
         fprintf(headerFP,"extern struct templateSlot tfl%d_%d[];\n",imageID,arrayVersion);
         newHeader = CLIPS_FALSE;
        }
     
      fprintf(fp,"{"); 
      PrintSymbolReference(fp,slotPtr->slotName); /* slotName */
      fprintf(fp,",%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,",
                   slotPtr->multiSlot,slotPtr->anyAllowed,slotPtr->symbolsAllowed,
                   slotPtr->stringsAllowed,slotPtr->floatsAllowed,
                   slotPtr->integersAllowed,
                   slotPtr->instancesAllowed,
                   slotPtr->addressesAllowed,
                   slotPtr->anyRestriction,
                   slotPtr->symbolRestriction,
                   slotPtr->stringRestriction,slotPtr->numberRestriction,
                   slotPtr->floatRestriction,
                   slotPtr->integerRestriction);
                      
      ExpressionToCode(fp,slotPtr->defaultList);
      fprintf(fp,",");
      ExpressionToCode(fp,slotPtr->restrictionList);
      fprintf(fp,",");
      ExpressionToCode(fp,slotPtr->minValue);
      fprintf(fp,",");
      ExpressionToCode(fp,slotPtr->maxValue);
      fprintf(fp,",");
   
      if (slotPtr->next == NULL)
        { fprintf(fp,"NULL}"); }
      else
        {
         fprintf(fp,"&tfl%d_%d[%d]}",imageID,
                                  ((slotCount+1) / maxIndices) + 1,
                                   (slotCount+1) % maxIndices);
        }
         
      slotCount++;
      i++;
      slotPtr = slotPtr->next;
      while ((slotPtr == NULL) && (dfPtr != NULL))
        {
         dfPtr = dfPtr->next;
         if (dfPtr != NULL) slotPtr = dfPtr->slotList;
        }
        
      if ((i == maxIndices) || (slotPtr == NULL))
        {
         fprintf(fp,"};\n");
         fclose(fp);
         i = 1;
         version++;
         arrayVersion++;
         if (slotPtr != NULL)
           { 
            if ((fp = NewCFile(fileName,fileID,version)) == NULL) return(0);
            newHeader = CLIPS_TRUE;
           }
        }
      else
        { fprintf(fp,",\n"); }
     }
     
   return(1);
  }
#endif

#endif
