   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*             DEFTEMPLATE FUNCTIONS MODULE            */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _DEFTMFUN_SOURCE_

#include "setup.h"

#if DEFTEMPLATE_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "expressn.h"
#include "evaluatn.h" 
#include "router.h"
#include "utility.h" 
#include "variable.h" 
#include "facts.h"
#include "generate.h"
#include "commline.h"
#include "parsutil.h"

#include "deftempl.h"
#include "deftmfun.h"

#if ANSI_COMPILER  
   static struct sap             *ParseAssertSlotValues(char *,struct token *,struct templateSlot *,int *,int *,int);
   static struct expr            *ReorderAssertSlotValues(struct templateSlot *,struct sap *);
   static struct expr            *GetSlotAssertValues(struct templateSlot *,struct sap *);
   static VOID                    ReturnSAPs(struct sap *);
   static BOOLEAN                 AssertSlotsMultiplyDefined(struct sap *);
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static struct expr            *ModAndDupParse(struct expr *,char *,char *);
   static BOOLEAN                 ModifySlotsMultiplyDefined(struct expr *);
#endif
   static struct templateSlot    *ParseSlotLabel(char *,struct token *,struct deftemplate *,int *,int);
   static BOOLEAN                 CheckRHSSlotTypes(struct expr *,struct templateSlot *,char *);
   static BOOLEAN                 FieldCheckTemplate(int,VOID *,struct templateSlot *,struct fact *);
#else
   static struct sap             *ParseAssertSlotValues();
   static struct expr            *ReorderAssertSlotValues();
   static struct expr            *GetSlotAssertValues();
   static VOID                    ReturnSAPs();
   static BOOLEAN                 AssertSlotsMultiplyDefined();
#if (! RUN_TIME) && (! BLOAD_ONLY)
   static struct expr            *ModAndDupParse();
   static BOOLEAN                 ModifySlotsMultiplyDefined();
#endif
   static struct templateSlot    *ParseSlotLabel();
   static BOOLEAN                 CheckRHSSlotTypes();
   static BOOLEAN                 FieldCheckTemplate();
#endif
  
/*####################################################*/
/*#                                                  #*/
/*# DEFRELATION MODIFICATIONS FOR THE ASSERT COMMAND #*/
/*#                                                  #*/
/*####################################################*/

/******************************************************************/
/* ParseAssertTemplate: Parses and builds the list of values that */
/*   are used for an assert of a fact with a deftemplate.         */
/******************************************************************/
globle struct expr *ParseAssertTemplate(readSource,temp_tkn,multi,error,dname,endType,constantsOnly)
  char *readSource;
  struct token *temp_tkn;
  int *multi, *error;
  SYMBOL_HN *dname;
  int endType, constantsOnly;
  {
   struct deftemplate *dtmpl_ptr;
   struct sap *first_slot, *last_slot, *next_slot;
   struct expr *first_arg;
   struct templateSlot *slotPtr;
   
   first_slot = NULL;
   last_slot = NULL;
   
   /*===============================================*/
   /* Get a pointer to the deftemplate information. */
   /*===============================================*/
   
   dtmpl_ptr = (struct deftemplate *) FindDeftemplate(dname->contents);
   if (dtmpl_ptr == NULL)
     { CLIPSSystemError("DEFTMFUN",1); }
     
   /*==============================================*/
   /* Parse each of the slot fields in the assert. */
   /*==============================================*/
  
   while ((slotPtr = ParseSlotLabel(readSource,temp_tkn,dtmpl_ptr,error,endType)) != NULL)
     {
      next_slot = ParseAssertSlotValues(readSource,temp_tkn,
                                        slotPtr,multi,error,constantsOnly);
      
      if (*error)
        { 
         ReturnSAPs(first_slot); 
         return(NULL);
        }
        
      if (CheckRHSSlotTypes(next_slot->expr,slotPtr,"assert") == 0)
        {
         *error = CLIPS_TRUE;
         ReturnSAPs(first_slot);
         ReturnSAPs(next_slot);
         return(NULL);
        }
        
      if (last_slot == NULL)
        { first_slot = next_slot; }
      else
        { last_slot->next = next_slot; }
      last_slot = next_slot;
     }
     
   if (*error)
     {
      ReturnSAPs(first_slot); 
      return(NULL);
     }
   
   /*=================================================*/
   /* Check to see if any slot names were duplicated. */
   /*=================================================*/
    
   if (AssertSlotsMultiplyDefined(first_slot))
     {
      *error = CLIPS_TRUE;
      ReturnSAPs(first_slot);
      return(NULL);
     }
     
   /*=============================================================*/
   /* Reorder the arguments to the order used by the deftemplate. */
   /*=============================================================*/
 
   first_arg = ReorderAssertSlotValues(dtmpl_ptr->slotList,first_slot);
   ReturnSAPs(first_slot);
   
   /*==============================*/
   /* Return the assert arguments. */
   /*==============================*/
   
   return(first_arg); 
  }

/****************************************************************/
/* ParseSlotLabel: Parses the beginning of a slot definition.   */
/*   Checks for opening left parenthesis and a valid slot name. */
/****************************************************************/
static struct templateSlot *ParseSlotLabel(inputSource,tempToken,dtmplPtr,error,endType)
  char *inputSource;
  struct token *tempToken;
  struct deftemplate *dtmplPtr;
  int *error;
  int endType;
  {
   struct templateSlot *slotPtr;
   
   /*========================*/
   /* Initialize error flag. */
   /*========================*/
   
   *error = CLIPS_FALSE;
   
   /*============================================*/
   /* If token is a right parenthesis, then fact */
   /* template definition is complete.           */
   /*============================================*/
   
   GetToken(inputSource,tempToken);
   if (tempToken->type == endType)
     { return(NULL); }
   
   /*=======================================*/
   /* Put a space between the template name */ 
   /* and the first slot definition.        */
   /*=======================================*/
   
   PPBackup();
   SavePPBuffer(" ");
   SavePPBuffer(tempToken->print_rep);
      
   /*=======================================================*/
   /* Slot definition begins with opening left parenthesis. */
   /*=======================================================*/
   
   if (tempToken->type != LPAREN)
     {
      SyntaxErrorMessage("deftemplate pattern");  
      *error = CLIPS_TRUE;
      return(NULL);
     }
     
   /*===========================*/
   /* Slot name must be a symbol. */
   /*===========================*/
      
   GetToken(inputSource,tempToken);
   if (tempToken->type != SYMBOL)
     {
      SyntaxErrorMessage("deftemplate pattern");
      *error = CLIPS_TRUE;
      return(NULL);
     }
    
   /*======================================================*/
   /* Check that the slot name is valid for this template. */
   /*======================================================*/
       
   if ((slotPtr = FindSlot(dtmplPtr,tempToken->value)) == NULL)
     {
      PrintCLIPS(WERROR,"\nInvalid slot ");
      PrintCLIPS(WERROR,ValueToString(tempToken->value));
      PrintCLIPS(WERROR," not defined in corresponding deftemplate ");
      PrintCLIPS(WERROR,ValueToString(dtmplPtr->name));
      PrintCLIPS(WERROR,"\n");
      *error = CLIPS_TRUE;
      return(NULL);
     }
     
   /*====================================*/
   /* Return a pointer to the slot name. */
   /*====================================*/
   
   return(slotPtr);
  }

/**************************************************************************/
/* ParseAssertSlotValues: Gets a single assert slot value for a template. */
/**************************************************************************/
static struct sap *ParseAssertSlotValues(inputSource,tempToken,
                                                slotPtr,multi,error,constantsOnly)
  char *inputSource;
  struct token *tempToken;
  struct templateSlot *slotPtr;
  int *multi, *error;
  int constantsOnly;
  {
   struct sap *next_slot;
   struct expr *new_test, *multi_test, *last_test;
   int temp_multi = CLIPS_FALSE;
   int printError;
   
   /*=============================*/
   /* Handle a single field slot. */
   /*=============================*/
   
   if (slotPtr->multiSlot == CLIPS_FALSE)
     {  
      SavePPBuffer(" ");
      new_test = GetAssertArgument(inputSource,tempToken,&temp_multi,
                                   error,RPAREN,constantsOnly,&printError); 
      if (*error)
        { 
         if (printError) SyntaxErrorMessage("deftemplate pattern");
         return(NULL); 
        }
        
      if (new_test == NULL)
       {
        *error = CLIPS_TRUE;
        PrintCLIPS(WERROR,"\nSingle field slot ");
        PrintCLIPS(WERROR,slotPtr->slotName->contents);
        PrintCLIPS(WERROR," must contain a single field value\n");
        return(NULL);
       }
       
      if (temp_multi)
       {
        *error = CLIPS_TRUE;
        PrintCLIPS(WERROR,"\nA multifield value cannot be used ");
        PrintCLIPS(WERROR,"in the single field slot ");
        PrintCLIPS(WERROR,slotPtr->slotName->contents);
        PrintCLIPS(WERROR,"\n");
        ReturnExpression(new_test);
        return(NULL);
       }
       
      GetToken(inputSource,tempToken);
     }
      
   /*============================*/
   /* Handle a multi field slot. */
   /*============================*/
   
   else
     {
      multi_test = GetAssertArgument(inputSource,tempToken,multi,
                                     error,RPAREN,constantsOnly,&printError);
      if (*error)
        { 
         if (printError) SyntaxErrorMessage("deftemplate pattern");
         return(NULL);
        }
        
      last_test = multi_test;
      
      while (tempToken->type != RPAREN)
        {
         PPBackup();  
         SavePPBuffer(" ");
         SavePPBuffer(tempToken->print_rep);
            
         new_test = GetAssertArgument(inputSource,tempToken,multi,error,RPAREN,constantsOnly,&printError);
         if (*error) 
           { 
            if (printError) SyntaxErrorMessage("deftemplate pattern");
            ReturnExpression(multi_test);
            return(NULL); 
           }
             
         last_test->next_arg = new_test;
         last_test = new_test;
        }
        
      new_test = multi_test;
     }
   
   /*==========================================================*/
   /* Slot definition must be closed with a right parenthesis. */
   /*==========================================================*/
   
   if (tempToken->type != RPAREN)   
     {
      PrintCLIPS(WERROR,"\nSingle field slot ");
      PrintCLIPS(WERROR,slotPtr->slotName->contents);
      PrintCLIPS(WERROR," must contain a single field value\n");
      *error = CLIPS_TRUE;
      ReturnExpression(new_test);
      return(NULL);
     }        
     
   /*=========================================================*/
   /* Build and return a structure describing the slot value. */
   /*=========================================================*/
   
   next_slot = get_struct(sap);
   next_slot->name = slotPtr->slotName;
   next_slot->expr = new_test;
   next_slot->next = NULL;
   
   return(next_slot);
  }

/****************************************************/ 
/* CheckRHSSlotTypes:                               */
/****************************************************/
static BOOLEAN CheckRHSSlotTypes(rhsSlots,slotPtr,commandName)
  struct expr *rhsSlots;
  struct templateSlot *slotPtr;
  char *commandName;
  {
   char returnType;
      
   while (rhsSlots != NULL)
     {
      if ((slotPtr->anyAllowed == CLIPS_FALSE) &&
          (((rhsSlots->type == INTEGER) && (slotPtr->integersAllowed != CLIPS_TRUE)) ||
           ((rhsSlots->type == FLOAT) && (slotPtr->floatsAllowed != CLIPS_TRUE)) ||
           ((rhsSlots->type == STRING) && (slotPtr->stringsAllowed != CLIPS_TRUE)) ||
#if OBJECT_SYSTEM
           ((rhsSlots->type == INSTANCE_NAME) && (slotPtr->instancesAllowed != CLIPS_TRUE)) ||
#endif
           ((rhsSlots->type == SYMBOL) && (slotPtr->symbolsAllowed != CLIPS_TRUE))))
        {
         PrintCLIPS(WERROR,"\nLiteral slot value(s) found in the ");
         PrintCLIPS(WERROR,commandName);
         PrintCLIPS(WERROR," command\ndo not match the allowed types for slot ");
         PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
         PrintCLIPS(WERROR,"\n");
         return(0);
        }
      
      if ((rhsSlots->type == FCALL) && (slotPtr->anyAllowed == CLIPS_FALSE))
        {
         returnType = ExpressionFunctionType(rhsSlots);
         if ((((returnType == 'i') || (returnType == 'l')) && (slotPtr->integersAllowed != CLIPS_TRUE)) ||
             (((returnType == 'f') || (returnType == 'd')) && (slotPtr->floatsAllowed != CLIPS_TRUE)) ||
             ((returnType == 's') && (slotPtr->stringsAllowed != CLIPS_TRUE)) ||
             ((returnType == 'a') && (slotPtr->addressesAllowed != CLIPS_TRUE)) ||
             (((returnType == 'w') || (returnType == 'c')) && (slotPtr->symbolsAllowed != CLIPS_TRUE)) ||
             ((returnType == 'b') || (returnType == 'v')))
           {
            PrintCLIPS(WERROR,"\nThe function return value found in the ");
            PrintCLIPS(WERROR,commandName);
            PrintCLIPS(WERROR," command\ndoes not match the allowed types for slot ");
            PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
            PrintCLIPS(WERROR,"\n");
            return(0);
           }
        }
      
      if (CheckSlotAllowedValues(rhsSlots->type,rhsSlots->value,slotPtr) == CLIPS_FALSE)
        {
         PrintCLIPS(WERROR,"\nLiteral slot value(s) found in the ");
         PrintCLIPS(WERROR,commandName);
         PrintCLIPS(WERROR," command\ndo not match the allowed values for slot ");
         PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
         PrintCLIPS(WERROR,"\n");
         return(0);
        }
      else if (CheckSlotRange(rhsSlots->type,rhsSlots->value,slotPtr) == CLIPS_FALSE)
        {
         PrintCLIPS(WERROR,"\nLiteral slot value(s) found in the ");
         PrintCLIPS(WERROR,commandName);
         PrintCLIPS(WERROR," command\ndo not fall in the allowed range for slot ");
         PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
         PrintCLIPS(WERROR,"\n");
         return(0);
        }
         
      rhsSlots = rhsSlots->next_arg;
     }
     
   return(1);
  }
  
/*************************************************************************/
/* ReorderAssertSlotValues: Rearranges the asserted values to correspond */
/*   to the order of the values described by the deftemplate.            */
/*************************************************************************/
static struct expr *ReorderAssertSlotValues(slotPtr,first_slot)
  struct templateSlot *slotPtr;
  struct sap *first_slot;
  {
   struct expr *first_arg = NULL;
   struct expr *last_arg = NULL, *new_arg;
   
   while (slotPtr != NULL)
     {
      new_arg = GetSlotAssertValues(slotPtr,first_slot);
        
      if (new_arg != NULL)
        {
         if (last_arg == NULL)
           { first_arg = new_arg; }
         else
           { last_arg->next_arg = new_arg; }
           
         while (new_arg->next_arg != NULL)
           { new_arg = new_arg->next_arg; }
         last_arg = new_arg;
        }
        
      slotPtr = slotPtr->next;
     }
     
   return(first_arg);
  }

/***************************************************************/  
/* GetSlotAssertValues: Gets the assert value for a given slot */
/*   of a deftemplate. If the value was supplied by the user,  */
/*   it will be used. If not the default value or default      */
/*   default value will be used.                               */
/***************************************************************/  
static struct expr *GetSlotAssertValues(slotPtr,first_slot)
  struct templateSlot *slotPtr;
  struct sap *first_slot;
  {
   struct sap *slotItem;
   struct expr *new_arg;
   
   /*==================================================*/
   /* Determine if the slot is assigned in the assert. */
   /*==================================================*/
   
   slotItem = (struct sap *) FindSlotItem(slotPtr,(struct slp *) first_slot);
   
   /*==========================================*/
   /* If the slot is assigned, use that value. */
   /*==========================================*/
   
   if (slotItem != NULL)
     {
      new_arg = slotItem->expr;
      slotItem->expr = NULL; 
     }
     
   /*=================================*/
   /* Otherwise, use a default value. */
   /*=================================*/
   
   else
     {
      if ((slotPtr->multiSlot == CLIPS_FALSE) && (slotPtr->defaultList == NULL))
        { 
         if (slotPtr->anyAllowed || slotPtr->symbolsAllowed)
           { new_arg = GenConstant(SYMBOL,AddSymbol("nil")); }
         else if (slotPtr->stringsAllowed)
           { new_arg = GenConstant(STRING,AddSymbol("")); }
         else if (slotPtr->integersAllowed)
           { new_arg = GenConstant(INTEGER,AddLong(0L)); }
         else if (slotPtr->floatsAllowed)
           { new_arg = GenConstant(FLOAT,AddDouble(0.0)); }
         else if (slotPtr->addressesAllowed)
           { new_arg = GenConstant(EXTERNAL_ADDRESS,NULL); }
         else 
           { new_arg = GenConstant(SYMBOL,AddSymbol("nil")); }
        }
      else
        { new_arg = CopyExpression(slotPtr->defaultList); }
     }
     
   /*==============================================*/
   /* Return the value to be asserted in the slot. */
   /*==============================================*/
   
   return(new_arg);
  }
          
/********************************************/
/* ReturnSAPs: Returns data structures used */ 
/*   for parsing deftemplate assert.        */
/********************************************/
static VOID ReturnSAPs(saps)
  struct sap *saps;
  {
   struct sap *tempSaps;

   while (saps != NULL)
     {
      tempSaps = saps->next;
      if (saps->expr != NULL) ReturnExpression(saps->expr);
      rtn_struct(sap,saps);
      saps = tempSaps;
     }
  }
  
/************************************************************/
/* AssertSlotsMultiplyDefined: Determines if any slots were */
/*   defined more that once for an assert statement.        */
/************************************************************/
static BOOLEAN AssertSlotsMultiplyDefined(first_slot)
  struct sap *first_slot;
  {
   struct sap *next_slot;
   int rv = CLIPS_FALSE;
   
   while (first_slot != NULL)
     {
      next_slot = first_slot->next;
      while (next_slot != NULL)
        {
         if (next_slot->name == first_slot->name)
           {
            PrintCLIPS(WERROR,"\nMultiple occurences of slot ");
            PrintCLIPS(WERROR,first_slot->name->contents);
            PrintCLIPS(WERROR,"\n");
            rv = CLIPS_TRUE;
           }
         next_slot = next_slot->next;
        }
      first_slot = first_slot->next;
     }
  
   return(rv);
  }

#if (! RUN_TIME) && (! BLOAD_ONLY)

/*################################################*/
/*#                                              #*/
/*# DEFRELATION ADDITIONS FOR THE MODIFY COMMAND #*/
/*#                                              #*/
/*################################################*/

/***************************************************************/
/* UpdateModifyDuplicate: Changes the modify/duplicate command */
/*   found on the RHS of a rule such that the positions of the */
/*   slots for replacement are stored rather than the slot     */
/*   name which allows quicker replacement of slots. This      */
/*   substitution can only take place when the deftemplate     */
/*   type is known (i.e. if a fact-index is used you don't     */
/*   know which type of deftemplate is going to be replaced    */
/*   until you actually do the replacement of slots).          */
/***************************************************************/
globle BOOLEAN UpdateModifyDuplicate(top,name)
  struct expr *top;
  char *name;
  {
   struct expr *modify_args, *temp_ptr;
   int fap = 0;
   SYMBOL_HN *dtmpl_name = NULL;
   struct deftemplate *dtmpl_ptr;
   struct templateSlot *slotPtr;
     
   /*========================================*/
   /* Determine the fact-address or index to */ 
   /* be retracted by the modify command.    */
   /*========================================*/
   
   modify_args = top->arg_list;
   if (modify_args->type == BWORD)
     {
      fap = GetFactAddressPosition(modify_args->value);
      dtmpl_name = GetRelationForPattern(fap);
      if (dtmpl_name == NULL) return(CLIPS_TRUE);
     }
   else 
     { return(CLIPS_TRUE); }
     
   /*========================================*/
   /* Make sure that the fact being modified */
   /* has a corresponding deftemplate.       */
   /*========================================*/
   
   if ((dtmpl_ptr = (struct deftemplate *) FindDeftemplate(dtmpl_name->contents)) == NULL)
     { return(CLIPS_TRUE); }
           
   /*=============================================================*/
   /* Make sure all the slot names are valid for the deftemplate. */
   /*=============================================================*/
   
   temp_ptr = modify_args->next_arg;
   while (temp_ptr != NULL)
     { 
      /*======================*/
      /* Does the slot exist? */
      /*======================*/
      
      if ((slotPtr = FindSlot(dtmpl_ptr,temp_ptr->value)) == NULL)
        {
         PrintCLIPS(WERROR,"\nInvalid slot ");
         PrintCLIPS(WERROR,((SYMBOL_HN *) temp_ptr->value)->contents);
         PrintCLIPS(WERROR," not defined in corresponding deftemplate ");
         PrintCLIPS(WERROR,ValueToString(dtmpl_ptr->name));
         PrintCLIPS(WERROR,"\n");
         return(CLIPS_FALSE);
        }
      
      /*=========================================================*/
      /* Is a multifield value being put in a single field slot? */
      /*=========================================================*/
      
      if (slotPtr->multiSlot == CLIPS_FALSE)
        {
         if (temp_ptr->arg_list == NULL)
           { 
            PrintCLIPS(WERROR,"\nSingle field slot ");
            PrintCLIPS(WERROR,slotPtr->slotName->contents);
            PrintCLIPS(WERROR," must contain a single field value\n");
            return(CLIPS_FALSE);
           }
         else if (temp_ptr->arg_list->next_arg != NULL)
           {
            PrintCLIPS(WERROR,"\nSingle field slot ");
            PrintCLIPS(WERROR,slotPtr->slotName->contents);
            PrintCLIPS(WERROR," must contain a single field value\n");
            return(CLIPS_FALSE);
           }
         else if ((temp_ptr->arg_list->type == BWORDS) ||
                  ((temp_ptr->arg_list->type == FCALL) ? 
                   (((struct FunctionDefinition *) temp_ptr->arg_list->value)->returnValueType == 'm') :
                      CLIPS_FALSE))
           {
            PrintCLIPS(WERROR,"\nA multifield value cannot be used ");
            PrintCLIPS(WERROR,"in the single field slot ");
            PrintCLIPS(WERROR,slotPtr->slotName->contents);
            PrintCLIPS(WERROR,"\n");
            return(CLIPS_FALSE);
           }
        }
      
      /*======================================*/
      /* Are the slot restrictions satisfied? */
      /*======================================*/
      
      if (CheckRHSSlotTypes(temp_ptr->arg_list,slotPtr,name) == 0)
        return(CLIPS_FALSE);
      
      /*=============================================*/
      /* Replace the slot with the integer position. */
      /*=============================================*/
      
      temp_ptr->type = INTEGER;
      temp_ptr->value = (VOID *) AddLong((long) FindSlotPosition(dtmpl_ptr,temp_ptr->value));
      
      temp_ptr = temp_ptr->next_arg;
     }
         
   return(CLIPS_TRUE);
  }
  
/*******************************************/
/* ModifyParse: Parses the modify command. */
/*******************************************/
globle struct expr *ModifyParse(top,in_file)
  struct expr *top;
  char *in_file;
  {
   return(ModAndDupParse(top,in_file,"modify"));
  }
  
/*************************************************/
/* DuplicateParse: Parses the duplicate command. */
/*************************************************/
globle struct expr *DuplicateParse(top,in_file)
  struct expr *top;
  char *in_file;
  {
   return(ModAndDupParse(top,in_file,"duplicate"));   
  }
  
/*************************************************************/
/* ModAndDupParse: Parses the modify and duplicate commands. */
/*************************************************************/
static struct expr *ModAndDupParse(top,in_file,name)
  struct expr *top;
  char *in_file;
  char *name;
  {
   int multi, error = CLIPS_FALSE;
   struct token act_tkn;
   struct expr *next_one;
   struct expr *new_test, *multi_test, *last_test;
   int printError;
   
   /*==================================================================*/
   /* Parse the fact-address or index to the modify/duplicate command. */
   /*==================================================================*/
   
   SavePPBuffer(" ");
   GetToken(in_file,&act_tkn);
   
   if ((act_tkn.type == BWORD) || (act_tkn.type == GBWORD))
     { next_one = GenConstant(act_tkn.type,act_tkn.value); }
   else if (act_tkn.type == INTEGER)
     {
      if (! TopLevelCommand())
        {
         PrintCLIPS(WERROR,"\nFact-indexes can only be used by ");
         PrintCLIPS(WERROR,name);
         PrintCLIPS(WERROR," as a top level command\n");
         ReturnExpression(top);
         return(NULL);
        }
        
      next_one = GenConstant(INTEGER,act_tkn.value);
     }
   else
     {
      ExpectedTypeError(name,1,"fact-address or fact-index");
      ReturnExpression(top); 
      return(NULL);
     }

   next_one->next_arg = NULL;
   next_one->arg_list = NULL;
   top->arg_list = next_one;
   next_one = top->arg_list;

   /*==============================================*/
   /* Parse the remaining modify/duplicate fields. */
   /*==============================================*/
   
   GetToken(in_file,&act_tkn);
   while (act_tkn.type != RPAREN)
     {
      PPBackup();
      SavePPBuffer(" ");
      SavePPBuffer(act_tkn.print_rep);
      
      /*=================================================*/
      /* Slot definition begins with a left parenthesis. */
      /*=================================================*/
      
      if (act_tkn.type != LPAREN)
        {
         SyntaxErrorMessage("duplicate/modify function");
         ReturnExpression(top);
         return(NULL);
        }
        
      /*=================================*/
      /* The slot name must be a symbol. */
      /*=================================*/
      
      GetToken(in_file,&act_tkn);
      if (act_tkn.type != SYMBOL)
        {
         SyntaxErrorMessage("duplicate/modify function");
         ReturnExpression(top);
         return(NULL);
        }
      
      next_one->next_arg = GenConstant(SYMBOL,act_tkn.value);
      next_one = next_one->next_arg;

      SavePPBuffer(" ");

      /*====================================================*/
      /* Get the values to be stored in the specified slot. */
      /*====================================================*/
      
      multi_test = NULL;
      last_test = NULL;
      while (act_tkn.type != RPAREN)
        {
         multi = CLIPS_FALSE;
         new_test = GetAssertArgument(in_file,&act_tkn,&multi,&error,
                                      RPAREN,CLIPS_FALSE,&printError); 
          
         if (error)
           {
            if (printError) SyntaxErrorMessage("deftemplate pattern");
            ReturnExpression(top);
            return(NULL);
           }
          
         if (last_test == NULL)
           { multi_test = new_test; }
         else
           { last_test->next_arg = new_test; }
         last_test = new_test;
        }
   
      /*==================================================*/
      /* Slot definition begins with a right parenthesis. */
      /*==================================================*/
      
      if (act_tkn.type != RPAREN)   
        {
         SyntaxErrorMessage("duplicate/modify function");
         ReturnExpression(top);
         ReturnExpression(multi_test);
         return(NULL);
        }        

      next_one->arg_list = multi_test;
         
      GetToken(in_file,&act_tkn);
     }
   
   /*========================================================*/
   /* Check to see that no slot name is used more than once. */
   /*========================================================*/
    
   if (ModifySlotsMultiplyDefined(top->arg_list->next_arg->next_arg))
     {
      ReturnExpression(top);
      return(NULL);
     }
          
   return(top); 
  }

  
/*********************************************************/
/* ModifySlotsMultiplyDefined: Determines if a slot name */  
/*   has been used more than once in a modify command.   */
/*********************************************************/
static int ModifySlotsMultiplyDefined(first_slot)
  struct expr *first_slot;
  {
   struct expr *next_slot;
   int rv = CLIPS_FALSE;
   
   while (first_slot != NULL)
     {
      next_slot = first_slot->next_arg;
      while (next_slot != NULL)
        {
         if (next_slot->value == first_slot->value)
           {
            PrintCLIPS(WERROR,"\nMultiple occurences of slot ");
            PrintCLIPS(WERROR,((SYMBOL_HN *) first_slot->value)->contents);
            PrintCLIPS(WERROR,"\n");
            rv = CLIPS_TRUE;
           }
         next_slot = next_slot->next_arg;
        }
      first_slot = first_slot->next_arg;
     }
  
   return(rv);
  }
  
#endif

/*##############################################*/
/*#                                            #*/
/*# DEFTEMPLATE FUNCTIONS FOR DYNAMIC CHECKING #*/
/*#                                            #*/
/*##############################################*/

/******************************************************************/
/* GetMultiSlotPosition: Determines if a given name is associated */
/*   with a deftemplate and whether that deftemplate has a multi  */
/*   field slot. Returns -1 if the name is not a deftemplate, 0   */
/*   the name is a deftemplate but doesn't have a multifield,     */
/*   and the position of the multifield slot if one exists for    */
/*   the deftemplate.                                             */    
/******************************************************************/
globle int GetMultiSlotPosition(name)
  SYMBOL_HN *name;
  {
   struct deftemplate *dtmpl_ptr;
   
   /*==============================================================*/
   /* If the corresponding deftemplate cannot be found, return -1. */
   /*==============================================================*/
   
   dtmpl_ptr = (struct deftemplate *) QFindDeftemplate(name);
   if (dtmpl_ptr == NULL) return(-1);
      
   /*=========================================*/
   /* Return 0 if it has no multifield slot. */
   /*=========================================*/
   
   if (dtmpl_ptr->hasMultiSlot == CLIPS_FALSE) return(0);
   
   /*=========================================================*/
   /* Return the last position which is the multifield slot. */
   /*=========================================================*/
   
   return(dtmpl_ptr->numberOfSlots);
  }
  
/******************************************************************/
/* MultiIntoSingleFieldSlotError:                */    
/******************************************************************/
globle VOID MultiIntoSingleFieldSlotError(currentPosition,dName)
  int currentPosition;
  SYMBOL_HN *dName;
  {
   struct deftemplate *dtmpl_ptr;
   struct templateSlot *slotPtr;
   
   dtmpl_ptr = (struct deftemplate *) QFindDeftemplate(dName);
   if (dtmpl_ptr == NULL) slotPtr = NULL;
   else slotPtr = dtmpl_ptr->slotList;
   
   while ((slotPtr != NULL) && (currentPosition > 1))
     { 
      slotPtr = slotPtr->next;
      currentPosition--;
     }
     
   PrintCLIPS(WERROR,"\nError: Attempted to assert a multifield value \n"); 
   PrintCLIPS(WERROR,"into the single field slot ");
   if (slotPtr != NULL) PrintCLIPS(WERROR,slotPtr->slotName->contents);
   else PrintCLIPS(WERROR,"<<unknown>>");
   PrintCLIPS(WERROR," of deftemplate ");
   if (dtmpl_ptr != NULL) PrintCLIPS(WERROR,dtmpl_ptr->name->contents);
   else PrintCLIPS(WERROR,"<<unknown>>");
   PrintCLIPS(WERROR,"\n");
      
   SetEvaluationError(CLIPS_TRUE);
  }

/**************************************************************/
/* CheckTemplateFact: Checks a fact to see if it violates any */
/*   deftemplate type, allowed-..., or range specifications.  */    
/**************************************************************/
globle VOID CheckTemplateFact(fact_ptr)
  struct fact *fact_ptr;
  {
   struct element *sublist;
   int length, i;
   struct deftemplate *dtmpl_ptr;
   struct templateSlot *slotPtr;
   
   if (! GetDynamicDeftemplateChecking()) return;
   
   sublist = fact_ptr->atoms;
   length = fact_ptr->factLength;
   
   /*==========================================================*/
   /* If the first field of the fact is not a symbol, then the   */
   /* fact need not be checked against the deftemplate format. */
   /*==========================================================*/
   
   if (sublist[0].type != SYMBOL) return;
   
   /*========================================================*/
   /* If the deftemplate corresponding to the first field of */
   /* of the fact cannot be found, then the fact cannot be   */
   /* checked against the deftemplate format.                */
   /*========================================================*/
   
   dtmpl_ptr = (struct deftemplate *) QFindDeftemplate((SYMBOL_HN *) sublist[0].value);
   if (dtmpl_ptr == NULL) return;
      
   /*===================================================*/
   /* Check each of the field slots of the deftemplate. */
   /*===================================================*/
   
   slotPtr = dtmpl_ptr->slotList;
   
   i = 1;
   while (slotPtr != NULL)
     {     
      /*======================================================*/
      /* Print the value of the slot for a single field slot. */
      /*======================================================*/
      
      if (slotPtr->multiSlot == CLIPS_FALSE)
        {
         FieldCheckTemplate(sublist[i].type,sublist[i].value,slotPtr,fact_ptr);
         i++;
        }
        
      /*==========================================================*/
      /* Else print the value of the slot for a multi field slot. */
      /*==========================================================*/
      
      else
        {
         while (i < length)
           {
            if (FieldCheckTemplate(sublist[i].type,sublist[i].value,slotPtr,fact_ptr))
              { i = length; }
            else
              { i++; }
           }
        }
        
      slotPtr = slotPtr->next;
     }
    
   return;
  }

/************************************/
/* FieldCheckTemplate:              */
/************************************/
static BOOLEAN FieldCheckTemplate(type,vPtr,slotPtr,factPtr)
  int type;
  VOID *vPtr;
  struct templateSlot *slotPtr;
  struct fact *factPtr;
  {
   char print_space[20];
   
   if (CheckSlotType(type,slotPtr) == CLIPS_FALSE)
     {
      PrintCLIPS(WERROR,"\nSlot value ");
      PrintAtom(WERROR,type,vPtr);
      PrintCLIPS(WERROR," found in fact ");
      sprintf(print_space,"f-%-5ld ",factPtr->factIndex);
      PrintCLIPS(WERROR,print_space);
      PrintCLIPS(WERROR,"\ndoes not match the allowed types for slot ");
      PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
      PrintCLIPS(WERROR,"\n");
      SetHaltExecution(CLIPS_TRUE);
      return(CLIPS_TRUE);
     }
   else if (CheckSlotAllowedValues(type,vPtr,slotPtr) == CLIPS_FALSE)
     {
      PrintCLIPS(WERROR,"\nSlot value ");
      PrintAtom(WERROR,type,vPtr);
      PrintCLIPS(WERROR," found in fact ");
      sprintf(print_space,"f-%-5ld ",factPtr->factIndex);
      PrintCLIPS(WERROR,print_space);
      PrintCLIPS(WERROR,"\ndoes not match the allowed values for slot ");
      PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
      PrintCLIPS(WERROR,"\n");
      SetHaltExecution(CLIPS_TRUE);
      return(CLIPS_TRUE);
     }
   else if (CheckSlotRange(type,vPtr,slotPtr) == CLIPS_FALSE)
     {
      PrintCLIPS(WERROR,"\nSlot value ");
      PrintAtom(WERROR,type,vPtr);
      PrintCLIPS(WERROR," found in fact ");
      sprintf(print_space,"f-%-5ld ",factPtr->factIndex);
      PrintCLIPS(WERROR,print_space);
      PrintCLIPS(WERROR,"\ndoes not fall in the allowed range for slot ");
      PrintCLIPS(WERROR,ValueToString(slotPtr->slotName));
      PrintCLIPS(WERROR,"\n");
      SetHaltExecution(CLIPS_TRUE);
      return(CLIPS_TRUE);
     }
     
   return(CLIPS_FALSE);
  }

#endif

