   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10 07/17/91             */
   /*                                                     */
   /*              DEFTEMPLATE PARSER MODULE              */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
  
#define _DEFTMPSR_SOURCE_

#include "setup.h"

#if DEFTEMPLATE_CONSTRUCT && (! RUN_TIME) && (! BLOAD_ONLY)

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "clipsmem.h"
#include "symbol.h"
#include "scanner.h"
#include "expressn.h"
#include "router.h"
#include "utility.h"
#include "constrct.h"
#include "reorder.h"
#include "ruleprsr.h"
#include "parsutil.h"
#include "generate.h"

#include "deftempl.h"
#include "deftmpsr.h"
  
#if ANSI_COMPILER  
   static struct templateSlot    *SlotDeclarations(char *,struct token *);
   static struct templateSlot    *ParseSlot(char *,struct token *);
   static struct templateSlot    *DefinedSlots(char *,SYMBOL_HN *,int,struct token *);
   static struct expr            *ParseDefault(char *,int,struct token *);
   static BOOLEAN                 MultiplyDefinedSlots(struct templateSlot *);
   static BOOLEAN                 ParseTypeAttribute(char *,struct templateSlot *);
   static BOOLEAN                 ParseAllowedValuesAttribute(char *,struct templateSlot *,char *);
   static BOOLEAN                 CheckSlotConflicts(struct templateSlot *);
   static BOOLEAN                 ParseRangeAttribute(char *,struct templateSlot *);
#else
   static struct templateSlot    *SlotDeclarations();
   static struct templateSlot    *ParseSlot();
   static struct templateSlot    *DefinedSlots();
   static struct expr            *ParseDefault();
   static BOOLEAN                 MultiplyDefinedSlots();
   static BOOLEAN                 ParseTypeAttribute();
   static BOOLEAN                 ParseAllowedValuesAttribute();
   static BOOLEAN                 CheckSlotConflicts();
   static BOOLEAN                 ParseRangeAttribute();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/
  
   static int              DeftemplateError;
   
/*******************************************************/
/* ParseDeftemplate: Parses the deftemplate construct. */
/*******************************************************/
globle int ParseDeftemplate(readSource)
  char *readSource;
  {
   SYMBOL_HN *dr_name;
   struct deftemplate *new_dtmpl;
   struct templateSlot *slots;
   struct token inputToken;
   
   /*================================================*/
   /* Initialize pretty print and error information. */
   /*================================================*/
   
   DeftemplateError = CLIPS_FALSE;
   SetPPBufferStatus(ON);
   FlushPPBuffer();            
   SavePPBuffer("(deftemplate ");

   /*=======================================================*/
   /* Parse the name and comment fields of the deftemplate. */
   /*=======================================================*/  

   dr_name = GetConstructNameAndComment(readSource,&inputToken,"deftemplate",
                                        FindDeftemplate,DeleteDeftemplate,"%",CLIPS_TRUE);
   if (dr_name == NULL) return(CLIPS_TRUE);
   
   /*===========================================*/
   /* Parse the slot fields of the deftemplate. */
   /*===========================================*/
   
   slots = SlotDeclarations(readSource,&inputToken);
   if (DeftemplateError == CLIPS_TRUE) return(CLIPS_TRUE);

   /*=====================================*/
   /* Create a new deftemplate structure. */
   /*=====================================*/
   
   new_dtmpl = get_struct(deftemplate);
   new_dtmpl->name =  dr_name;
   new_dtmpl->slotList = slots;
   new_dtmpl->hasMultiSlot = CLIPS_FALSE;
   new_dtmpl->numberOfSlots = 0;
   while (slots != NULL)
     {
      if (slots->multiSlot == CLIPS_TRUE) new_dtmpl->hasMultiSlot = CLIPS_TRUE;
      new_dtmpl->numberOfSlots++;
      slots = slots->next;
     }
     
   new_dtmpl->next = NULL;

   /*====================================*/
   /* Store pretty print representation. */
   /*====================================*/
   
   if (GetConserveMemory() == CLIPS_TRUE)
     { new_dtmpl->pp_form = NULL; }
   else
     { new_dtmpl->pp_form = CopyPPBuffer(); } 
   
   /*==============================================*/
   /* Add deftemplate to the list of deftemplates. */
   /*==============================================*/
   
   AddDeftemplate(new_dtmpl);
   
   return(CLIPS_FALSE);       
  }

/********************************************************************/
/* SlotDeclarations: Parses the slot declarations of a deftemplate. */
/********************************************************************/
static struct templateSlot *SlotDeclarations(readSource,inputToken)
  char *readSource;
  struct token *inputToken;
  {
   struct templateSlot *newSlot, *slotList = NULL, *last_slot = NULL;
   int parse_multi = CLIPS_FALSE, multi_found = CLIPS_FALSE;
   struct templateSlot *multiSlot = NULL;

   while (inputToken->type != RPAREN)
     {
      /*====================================================*/
      /* Slots begin with a '(' followed by a slot keyword. */
      /*====================================================*/
      
      if (inputToken->type != LPAREN)
        {
         SyntaxErrorMessage("deftemplate");
         ReturnSlots(slotList);
         ReturnSlots(multiSlot);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }

      GetToken(readSource,inputToken);
      if (inputToken->type != SYMBOL)
        {
         SyntaxErrorMessage("deftemplate");
         ReturnSlots(slotList);
         ReturnSlots(multiSlot);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }
        
      /*=================*/
      /* Parse the slot. */
      /*=================*/
      
      newSlot = ParseSlot(readSource,inputToken);
      if (DeftemplateError == CLIPS_TRUE)
        {
         ReturnSlots(newSlot);
         ReturnSlots(slotList);
         ReturnSlots(multiSlot);
         return(NULL);
        }
       
      /*=============================*/
      /* Keep track of the new slot. */
      /*=============================*/
      
      if (newSlot != NULL)
        {
         parse_multi = newSlot->multiSlot;
          
         if ((parse_multi == CLIPS_TRUE) && (multi_found == CLIPS_TRUE))
           {
            PrintCLIPS(WERROR,"\nTwo multifield slot definitions\n");
            ReturnSlots(newSlot); 
            ReturnSlots(slotList);
            ReturnSlots(multiSlot);
            DeftemplateError = CLIPS_TRUE;
            return(NULL);
           } 
           
         if (parse_multi == CLIPS_TRUE)
           { 
            multiSlot = newSlot;
            multi_found = CLIPS_TRUE;
           }
         else
           {
            if (last_slot == NULL)
              { slotList = newSlot; }
            else
              { last_slot->next = newSlot; }
            last_slot = newSlot;
           }
        }
       
      /*================================*/
      /* Check for closing parenthesis. */
      /*================================*/
      
      GetToken(readSource,inputToken);
      if (inputToken->type != RPAREN)
        {
         PPBackup();
         SavePPBuffer("\n   ");
         SavePPBuffer(inputToken->print_rep);
        }
     }
     
  SavePPBuffer("\n");
      
  /*====================================================*/
  /* Attach the multi slot to the end of the slot list. */
  /*====================================================*/
  
  if (last_slot != NULL)
    { last_slot->next = multiSlot; }
  else
    { slotList = multiSlot; }
     
     
  if (MultiplyDefinedSlots(slotList))
    {
     DeftemplateError = CLIPS_TRUE;
     ReturnSlots(slotList);
     return(NULL);
    }
    
  /*=======================*/
  /* Return the slot list. */
  /*=======================*/
  
  return(slotList);
 }

/*****************************************************/
/* ParseSlot: Parses a single slot of a deftemplate. */
/*****************************************************/
static struct templateSlot *ParseSlot(readSource,inputToken)
  char *readSource;
  struct token *inputToken;
  { 
   int parse_multi;
   SYMBOL_HN *slotName;
   struct templateSlot *newSlot;
   
   /*=====================================================*/
   /* Slots must  begin with keyword field or multifield. */
   /*=====================================================*/
   
   if ((strcmp(ValueToString(inputToken->value),"field") != 0) &&
       (strcmp(ValueToString(inputToken->value),"multi-field") != 0) &&
       (strcmp(ValueToString(inputToken->value),"multifield") != 0))
     {
      SyntaxErrorMessage("deftemplate");  
      DeftemplateError = CLIPS_TRUE;
      return(NULL);
     }
   
   /*===============================================*/
   /* Determine if multifield slot is being parsed. */
   /*===============================================*/
     
   if ((strcmp(ValueToString(inputToken->value),"multi-field") == 0) ||
       (strcmp(ValueToString(inputToken->value),"multifield") == 0))
     { parse_multi = CLIPS_TRUE; }
   else
     { parse_multi = CLIPS_FALSE; }
   
   /*========================================*/
   /* The name of the slot must be a symbol. */
   /*========================================*/
          
   SavePPBuffer(" ");
   GetToken(readSource,inputToken);
   if (inputToken->type != SYMBOL)
     {
      SyntaxErrorMessage("deftemplate");  
      DeftemplateError = CLIPS_TRUE;
      return(NULL);
     }
        
   slotName = (SYMBOL_HN *) inputToken->value;
     
   /*===================================*/
   /* Parse the attributes of the slot. */
   /*===================================*/
    
   newSlot = DefinedSlots(readSource,slotName,parse_multi,inputToken);
   if (newSlot == NULL) 
     {
      DeftemplateError = CLIPS_TRUE;
      return(NULL);
     }
   
   /*=================================*/
   /* Check for slot conflict errors. */
   /*=================================*/
   
   if (CheckSlotConflicts(newSlot) == CLIPS_FALSE)
     {
      ReturnSlots(newSlot);
      DeftemplateError = CLIPS_TRUE;
      return(NULL);
     }
   
   /*==================*/
   /* Return the slot. */
   /*==================*/
   
   return(newSlot);
  }
   
/**************************************************************/
/* DefinedSlots: Parses a field or multifield slot attribute. */
/**************************************************************/
static struct templateSlot *DefinedSlots(readSource,defined_slot,parse_multi,inputToken)
  char *readSource;
  SYMBOL_HN *defined_slot;
  int parse_multi;
  struct token *inputToken;
  {
   struct templateSlot *newSlot;
   struct expr *default_list = NULL;
   int default_found = CLIPS_FALSE;
   int type_found = CLIPS_FALSE;
   
   /*===========================*/
   /* Build the slot container. */
   /*===========================*/
   
   newSlot = get_struct(templateSlot);
   newSlot->slotName = defined_slot;
   newSlot->defaultList = NULL;
   newSlot->restrictionList = NULL;
   newSlot->minValue = NULL;
   newSlot->maxValue = NULL;
   newSlot->multiSlot = parse_multi;
   newSlot->next = NULL;
   newSlot->anyAllowed = CLIPS_TRUE;
   newSlot->symbolsAllowed = CLIPS_FALSE;
   newSlot->stringsAllowed = CLIPS_FALSE;
   newSlot->floatsAllowed = CLIPS_FALSE;
   newSlot->integersAllowed = CLIPS_FALSE;
   newSlot->instancesAllowed = CLIPS_FALSE;
   newSlot->addressesAllowed = CLIPS_FALSE;
   newSlot->anyRestriction = CLIPS_FALSE;
   newSlot->symbolRestriction = CLIPS_FALSE;
   newSlot->stringRestriction = CLIPS_FALSE;
   newSlot->numberRestriction = CLIPS_FALSE;
   newSlot->floatRestriction = CLIPS_FALSE;
   newSlot->integerRestriction = CLIPS_FALSE;

   /*========================================*/
   /* Parse the primitive slot if it exists. */
   /*========================================*/

   GetToken(readSource,inputToken);
   
   while (inputToken->type != RPAREN)
     {
      PPBackup();
      SavePPBuffer(" ");
      SavePPBuffer(inputToken->print_rep);
      
      /*================================================*/
      /* Slot attributes begin with a left parenthesis. */
      /*================================================*/
      
      if (inputToken->type != LPAREN)
        { 
         SyntaxErrorMessage("deftemplate");   
         ReturnSlots(newSlot);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }
        
      /*=============================================*/
      /* The name of the attribute must be a symbol. */
      /*=============================================*/
      
      GetToken(readSource,inputToken);
      if (inputToken->type != SYMBOL)
        {
         SyntaxErrorMessage("deftemplate");  
         ReturnSlots(newSlot);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }

      /*========================================================*/
      /* If the attribute is only used by CRSV, then ignore it. */
      /*========================================================*/
      
      if (strcmp(ValueToString(inputToken->value),"range") == 0)
        { 
         if (ParseRangeAttribute(readSource,newSlot) == 0)
           { 
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           }; 
        }  
      else if ((strcmp(ValueToString(inputToken->value),"min-number-of-elements") == 0) ||
               (strcmp(ValueToString(inputToken->value),"max-number-of-elements") == 0))
        { 
         if (parse_multi)
           { 
            if (ReadUntilClosingParen(readSource,inputToken) == 0)
              { 
               DeftemplateError = CLIPS_TRUE;
               ReturnSlots(newSlot);
               return(NULL);
              }; 
           }
         else
           {
            PrintCLIPS(WERROR,"\nmin and max number-of-elements attributes ");
            PrintCLIPS(WERROR,"are only allowed for multifield slots\n"); 
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           }
        }
        
      /*======================================================*/
      /* else if the attribute is an allowed value attribute, */
      /* then add it to the allowed values list.              */
      /*======================================================*/
      
      else if ((strcmp(ValueToString(inputToken->value),"allowed-words") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-symbols") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-strings") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-numbers") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-integers") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-floats") == 0) ||
               (strcmp(ValueToString(inputToken->value),"allowed-values") == 0))
        {
         if (ParseAllowedValuesAttribute(readSource,newSlot,
                                         ValueToString(inputToken->value)) == 0)
           { 
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           }; 
        }
        
      /*=================================================*/
      /* else if the attribute is the default attribute, */
      /* then get the default list for this slot.        */
      /*=================================================*/
      
      else if (strcmp(ValueToString(inputToken->value),"default") == 0)
        {
         if (default_found)
           {
            PrintCLIPS(WERROR,"\nThe default attribute has already been defined\n");
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           }
            
         default_list = ParseDefault(readSource,parse_multi,inputToken);
         if (DeftemplateError == CLIPS_TRUE)
           { 
            ReturnSlots(newSlot);
            return(NULL); 
           }
         default_found = CLIPS_TRUE;
         newSlot->defaultList = default_list;
        }
        
      /*==============================================*/
      /* else if the attribute is the type attribute, */
      /* then get the allowed types for this slot.    */
      /*==============================================*/
      
      else if (strcmp(ValueToString(inputToken->value),"type") == 0)
        {
         if (type_found)
           {
            PrintCLIPS(WERROR,"\nThe type attribute has already been defined\n");
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           }
           
         if (ParseTypeAttribute(readSource,newSlot) == 0)
           { 
            DeftemplateError = CLIPS_TRUE;
            ReturnSlots(newSlot);
            return(NULL);
           };  
           
         type_found = CLIPS_TRUE;
        }
        
      /*============================================*/
      /* Otherwise the attribute is an invalid one. */
      /*============================================*/
      
      else
        {
         PrintCLIPS(WERROR,"\nIllegal slot attribute\n");
         ReturnSlots(newSlot);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }
      
      /*===================================*/
      /* Begin parsing the next attribute. */
      /*===================================*/
      
      GetToken(readSource,inputToken);
     }

   /*============================*/
   /* Return the attribute list. */
   /*============================*/
   
   return(newSlot);
  }
  
/**********************************************/
/* ParseDefault: Parses a default value list. */
/**********************************************/
static struct expr *ParseDefault(readSource,parse_multi,inputToken)
  char *readSource;
  int parse_multi;
  struct token *inputToken;
  {
   struct expr *default_list = NULL, *last_default = NULL;
   struct expr *new_item;
   
   SavePPBuffer(" ");
   GetToken(readSource,inputToken);

   while (inputToken->type != RPAREN)
     {
      SavePPBuffer(" ");
       
      /*============================================*/
      /* If the value is a symbol, string, or number, */ 
      /* store it as the next default value.        */
      /*============================================*/
      
      if ((inputToken->type == SYMBOL) || (inputToken->type == STRING) || 
#if OBJECT_SYSTEM
          (inputToken->type == INSTANCE_NAME) ||
#endif
          (inputToken->type == FLOAT) || (inputToken->type == INTEGER))
        { new_item = GenConstant(inputToken->type,inputToken->value); }
        
      /*===================================*/
      /* Else the variable ?NONE indicates */
      /* that there is no default value.   */
      /*===================================*/
      
      else if (strcmp(inputToken->print_rep,"?NONE") == 0)
        {
         if (default_list != NULL)
           { 
            PrintCLIPS(WERROR,"\n?NONE cannot be used with other default values\n");
            ReturnExpression(default_list);
            DeftemplateError = CLIPS_TRUE;
            return(NULL);
           }
         GetToken(readSource,inputToken);
         
         if (inputToken->type != RPAREN)
           {
            PrintCLIPS(WERROR,"\nExpected ')' after ?NONE default value\n");
            ReturnExpression(default_list);
            DeftemplateError = CLIPS_TRUE;
            return(NULL);
           }
         
         PPBackup();
         PPBackup();
         SavePPBuffer(")");
         
         return(NULL);
        }
        
      /*===============================================*/
      /* else an invalid default value has been given. */
      /*===============================================*/
      
      else
        {
         PrintCLIPS(WERROR,"\nDefault values must be primitive data types\n");
         ReturnExpression(default_list);
         DeftemplateError = CLIPS_TRUE;
         return(NULL);
        }
      
      /*============================================*/
      /* Add the default value to the default list. */
      /*============================================*/
      
      if (last_default == NULL)
        { default_list = new_item; }
      else
        { last_default->next_arg = new_item; }
      last_default = new_item;
      
      /*=======================================*/
      /* Begin parsing the next default value. */
      /*=======================================*/
      
      GetToken(readSource,inputToken);
     }
     
   /*=====================================*/
   /* Fix up pretty print representation. */
   /*=====================================*/
   
   PPBackup();
   PPBackup();
   SavePPBuffer(")");
   
   /*=========================================*/
   /* A single field slot's default attribute */
   /* must contain a single value.            */
   /*=========================================*/
   
   if ((parse_multi == CLIPS_FALSE) && 
       ((default_list == NULL) ? CLIPS_TRUE :
                                 (default_list->next_arg == NULL) ? CLIPS_FALSE : CLIPS_TRUE))
     {
      PrintCLIPS(WERROR,"\nThe default attribute for a single field slot ");
      PrintCLIPS(WERROR,"must contain a single value\n");
      ReturnExpression(default_list);
      DeftemplateError = CLIPS_TRUE;
      return(NULL);
     }

   /*==========================*/
   /* Return the default list. */
   /*==========================*/
   
   return(default_list);
  }
  
/**************************************************/
/* ParseTypeAttribute: Parses the type attribute. */
/**************************************************/
static BOOLEAN ParseTypeAttribute(readSource,slotPtr)
  char *readSource;
  struct templateSlot *slotPtr;
  {
   int typeParsed = CLIPS_FALSE;
   int variableParsed = CLIPS_FALSE;
   struct token inputToken;
   
   SavePPBuffer(" ");
   GetToken(readSource,&inputToken);

   while (inputToken.type != RPAREN)
     {
      SavePPBuffer(" ");
         
      if (inputToken.type == SYMBOL)
        {
         if (variableParsed == CLIPS_TRUE)
           {
            PrintCLIPS(WERROR,"\n?VARIABLE must appear by itself in the type attribute\n");
            DeftemplateError = CLIPS_TRUE;
            return(CLIPS_FALSE);
           }
         else if (strcmp(inputToken.print_rep,"LEXEME") == 0)
           { 
            if (slotPtr->symbolsAllowed || slotPtr->stringsAllowed) DeftemplateError = CLIPS_TRUE;
            else 
              {
               slotPtr->stringsAllowed = CLIPS_TRUE; 
               slotPtr->symbolsAllowed = CLIPS_TRUE;
              }
           }
         else if ((strcmp(inputToken.print_rep,"SYMBOL") == 0) ||
                  (strcmp(inputToken.print_rep,"WORD") == 0))
           { 
            if (slotPtr->symbolsAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->symbolsAllowed = CLIPS_TRUE; 
           }
         else if (strcmp(inputToken.print_rep,"STRING") == 0)
           { 
            if (slotPtr->stringsAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->stringsAllowed = CLIPS_TRUE; 
           }
         else if (strcmp(inputToken.print_rep,"NUMBER") == 0)
           { 
            if (slotPtr->floatsAllowed || slotPtr->integersAllowed) DeftemplateError = CLIPS_TRUE;
            else 
              {
               slotPtr->floatsAllowed = CLIPS_TRUE; 
               slotPtr->integersAllowed = CLIPS_TRUE;
              }
           }
         else if (strcmp(inputToken.print_rep,"FLOAT") == 0)
           { 
            if (slotPtr->floatsAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->floatsAllowed = CLIPS_TRUE; 
           }
         else if (strcmp(inputToken.print_rep,"INTEGER") == 0)
           { 
            if (slotPtr->integersAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->integersAllowed = CLIPS_TRUE; 
           }
         else if (strcmp(inputToken.print_rep,"INSTANCE") == 0)
           { 
            if (slotPtr->instancesAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->instancesAllowed = CLIPS_TRUE; 
           }        
         else if (strcmp(inputToken.print_rep,"EXTERNAL-ADDRESS") == 0)
           { 
            if (slotPtr->addressesAllowed) DeftemplateError = CLIPS_TRUE;
            else slotPtr->addressesAllowed = CLIPS_TRUE; 
           }
         else
           {
            PrintCLIPS(WERROR,"\nAllowed type attribute values are\n");
            PrintCLIPS(WERROR,"LEXEME, SYMBOL, STRING, NUMBER, INTEGER, FLOAT\n");
            PrintCLIPS(WERROR,"INSTANCE, EXTERNAL-ADDRESS, and ?VARIABLE\n");
            DeftemplateError = CLIPS_TRUE;
            return(CLIPS_FALSE);
           }
           
         slotPtr->anyAllowed = CLIPS_FALSE;
         typeParsed = CLIPS_TRUE;
        }
      else if (inputToken.type == BWORD)
        {
         if (strcmp(inputToken.print_rep,"?VARIABLE") == 0)
           {
            if (typeParsed == CLIPS_TRUE)
              {
               PrintCLIPS(WERROR,"\n?VARIABLE must appear by itself in the type attribute\n");
               DeftemplateError = CLIPS_TRUE;
               return(CLIPS_FALSE);
              }
              
            typeParsed = CLIPS_TRUE;
            variableParsed = CLIPS_TRUE;
           }
         else
           {
            PrintCLIPS(WERROR,"\nAllowed type attribute values are\n");
            PrintCLIPS(WERROR,"LEXEME, SYMBOL, STRING, NUMBER, INTEGER, FLOAT\n");
            PrintCLIPS(WERROR,"INSTANCE, EXTERNAL-ADDRESS, and ?VARIABLE\n");
            DeftemplateError = CLIPS_TRUE;
            return(CLIPS_FALSE);
           }
         }
      else
        { DeftemplateError = CLIPS_TRUE; }
        
      if (DeftemplateError)
        {
         SyntaxErrorMessage("deftemplate type attribute");
         return(CLIPS_FALSE);
        }
      
      /*====================================*/
      /* Begin parsing the next type value. */
      /*====================================*/
      
      GetToken(readSource,&inputToken);
     }
     
   /*=====================================*/
   /* Fix up pretty print representation. */
   /*=====================================*/
   
   PPBackup();
   PPBackup();
   SavePPBuffer(")");
   
   /*=======================================*/
   /* The type attribute must have a value. */
   /*=======================================*/
 
   if ((! typeParsed) && (! variableParsed))
     {
      SyntaxErrorMessage("deftemplate type attribute");  
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
     
   return(CLIPS_TRUE);
  }
  
/**************************************************/
/* ParseRangeAttribute: Parses the range attribute. */
/**************************************************/
static BOOLEAN ParseRangeAttribute(readSource,slotPtr)
  char *readSource;
  struct templateSlot *slotPtr;
  {
   struct token inputToken;
   
   if ((slotPtr->minValue != NULL) || (slotPtr->maxValue != NULL))
     {
      PrintCLIPS(WERROR,"\nThe range attribute has already been defined\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
   
   if ((slotPtr->anyRestriction) || (slotPtr->floatRestriction) ||
       (slotPtr->integerRestriction))
     {
      PrintCLIPS(WERROR,"\nThe allowed-values/numbers/integers/floats attribute\n");
      PrintCLIPS(WERROR,"cannot be used in conjunction with the range attribute\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
     
   /*==========================*/
   /* Parse the minimum value. */
   /*==========================*/
   
   SavePPBuffer(" ");
   GetToken(readSource,&inputToken);
   if ((inputToken.type == INTEGER) || (inputToken.type == FLOAT))
     { slotPtr->minValue = GenConstant(inputToken.type,inputToken.value); }
   else if ((inputToken.type == BWORD) && (strcmp(inputToken.print_rep,"?VARIABLE") == 0))
     {
      /* Do nothing. */
     }
   else
     {
      PrintCLIPS(WERROR,"\nAllowed range attribute values are\n");
      PrintCLIPS(WERROR,"<integers>, <floats>, and ?VARIABLE\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
     
   /*==========================*/
   /* Parse the maximum value. */
   /*==========================*/
   
   SavePPBuffer(" ");
   GetToken(readSource,&inputToken);
   if ((inputToken.type == INTEGER) || (inputToken.type == FLOAT))
     { slotPtr->maxValue = GenConstant(inputToken.type,inputToken.value); }
   else if ((inputToken.type == BWORD) && (strcmp(inputToken.print_rep,"?VARIABLE") == 0))
     {
      /* Do nothing. */
     }
   else
     {
      PrintCLIPS(WERROR,"\nAllowed range attribute values are\n");
      PrintCLIPS(WERROR,"<integers>, <floats>, and ?VARIABLE\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
   
   /*================================*/
   /* Parse the closing parenthesis. */
   /*================================*/
   
   GetToken(readSource,&inputToken);
   if (inputToken.type != RPAREN)
     {
      SyntaxErrorMessage("deftemplate range attribute");  
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
     
   /*====================================================*/
   /* Minimum value must be less than the maximum value. */
   /*====================================================*/
   
   if ((slotPtr->maxValue != NULL) && (slotPtr->minValue != NULL))
     {
      if ((slotPtr->minValue->type == INTEGER) && (slotPtr->maxValue->type == INTEGER))
        {
         if (ValueToLong(slotPtr->minValue->value) > ValueToLong(slotPtr->maxValue->value))
           { DeftemplateError = CLIPS_TRUE; }
        }
      else if ((slotPtr->minValue->type == FLOAT) && (slotPtr->maxValue->type == FLOAT))
        {
         if (ValueToDouble(slotPtr->minValue->value) > ValueToDouble(slotPtr->maxValue->value))
           { DeftemplateError = CLIPS_TRUE; }
        }
      else if ((slotPtr->minValue->type == INTEGER) && (slotPtr->maxValue->type == FLOAT))
        {
         if (((double) ValueToLong(slotPtr->minValue->value)) > ValueToDouble(slotPtr->maxValue->value))
           { DeftemplateError = CLIPS_TRUE; }
        }
      else if ((slotPtr->minValue->type == FLOAT) && (slotPtr->maxValue->type == INTEGER))
        {
         if (ValueToDouble(slotPtr->minValue->value) > ((double) ValueToLong(slotPtr->maxValue->value)))
           { DeftemplateError = CLIPS_TRUE; }
        }
      if (DeftemplateError)
        {
         PrintCLIPS(WERROR,"\nMinimum range value must be less than or equal to\n");
         PrintCLIPS(WERROR,"the maximum range value\n");
         return(0);
        }
     }
     
   return(1);
  }

/***********************************************************/
/* ParseAllowedValuesAttribute: Parses the type attribute. */
/***********************************************************/
static BOOLEAN ParseAllowedValuesAttribute(readSource,slotPtr,attribute)
  char *readSource;
  struct templateSlot *slotPtr;
  char *attribute;
  {
   struct token inputToken;
   int expectedType = KUNKNOWN, error = CLIPS_FALSE;
   struct expr *new_item;
   int typeParsed = CLIPS_FALSE, variableParsed = CLIPS_FALSE;
   
   /*============================================*/
   /* Check that the attribute does not conflict */ 
   /* with previous attribute definitions.       */
   /*============================================*/
   
   if ((strcmp(attribute,"allowed-values") == 0) &&
       ((slotPtr->symbolRestriction) ||
        (slotPtr->stringRestriction) ||
        (slotPtr->integerRestriction) ||
        (slotPtr->floatRestriction)))
     {
      PrintCLIPS(WERROR,"\nThe allowed-values attribute cannot be used in ");
      PrintCLIPS(WERROR,"conjunction with other value restriction attributes\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }   
   else if ((strcmp(attribute,"allowed-values") != 0) && (slotPtr->anyRestriction))
     {
      PrintCLIPS(WERROR,"\nThe allowed-values attribute cannot be used in ");
      PrintCLIPS(WERROR,"conjunction with other value restriction attributes\n");
      DeftemplateError = CLIPS_TRUE;
      return(0);
     }
   else if (strcmp(attribute,"allowed-values") == 0)
     { 
      if (slotPtr->anyRestriction)
        {
         PrintCLIPS(WERROR,"\nThe allowed-values attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      if ((slotPtr->minValue != NULL) || (slotPtr->maxValue != NULL))
        {
         PrintCLIPS(WERROR,"\nThe allowed-values attribute\n");
         PrintCLIPS(WERROR,"cannot be used in conjunction with the range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
      
      slotPtr->anyRestriction = CLIPS_TRUE;
      expectedType = KUNKNOWN;
     }
   else if ((strcmp(attribute,"allowed-words") == 0) ||
            (strcmp(attribute,"allowed-symbols") == 0))
     {
      if (slotPtr->symbolRestriction)
        {
         PrintCLIPS(WERROR,"\nThe allowed-symbols attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
      slotPtr->symbolRestriction = CLIPS_TRUE;
      expectedType = SYMBOL;
     }
   else if (strcmp(attribute,"allowed-strings") == 0) 
     {
      if (slotPtr->stringRestriction)
        {
         PrintCLIPS(WERROR,"\nThe allowed-strings attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
      slotPtr->stringRestriction = CLIPS_TRUE;
      expectedType = STRING;
     }
   else if (strcmp(attribute,"allowed-integers") == 0)
     {
      if (slotPtr->integerRestriction)
        {
         PrintCLIPS(WERROR,"\nThe allowed-integers attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
            
      if ((slotPtr->minValue != NULL) || (slotPtr->maxValue != NULL))
        {
         PrintCLIPS(WERROR,"\nThe allowed-integers attribute\n");
         PrintCLIPS(WERROR,"cannot be used in conjunction with the range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      slotPtr->integerRestriction = CLIPS_TRUE;
      expectedType = INTEGER;
     }
   else if (strcmp(attribute,"allowed-floats") == 0)
     {
      if (slotPtr->floatRestriction)
        {
         PrintCLIPS(WERROR,"\nThe allowed-floats attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      if ((slotPtr->minValue != NULL) || (slotPtr->maxValue != NULL))
        {
         PrintCLIPS(WERROR,"\nThe allowed-floats attribute\n");
         PrintCLIPS(WERROR,"cannot be used in conjunction with the range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      slotPtr->floatRestriction = CLIPS_TRUE;
      expectedType = FLOAT;
     }
   else if (strcmp(attribute,"allowed-numbers") == 0)
     {
      if ((slotPtr->floatRestriction) || (slotPtr->integerRestriction))
        {
         PrintCLIPS(WERROR,"\nThe allowed-numbers attribute has already been defined\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      if ((slotPtr->minValue != NULL) || (slotPtr->maxValue != NULL))
        {
         PrintCLIPS(WERROR,"\nThe allowed-numbers attribute\n");
         PrintCLIPS(WERROR,"cannot be used in conjunction with the range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      slotPtr->numberRestriction = CLIPS_TRUE;
      slotPtr->floatRestriction = CLIPS_TRUE;
      slotPtr->integerRestriction = CLIPS_TRUE;
      expectedType = INTEGER_OR_FLOAT;
     }
     
   /*==================================================*/
   /* Get the allowed values and add them to the list. */
   /*==================================================*/
   
   SavePPBuffer(" ");
   GetToken(readSource,&inputToken);

   while (inputToken.type != RPAREN)
     {
      SavePPBuffer(" ");

      /*=================*/
      /* Check the type. */
      /*=================*/
      
      switch(inputToken.type)
        {
         case INTEGER:
           if ((expectedType != KUNKNOWN) && 
               (expectedType != INTEGER) &&
               (expectedType != INTEGER_OR_FLOAT)) error = CLIPS_TRUE;
           typeParsed = CLIPS_TRUE;
           break;
         
         case FLOAT:
           if ((expectedType != KUNKNOWN) && 
               (expectedType != FLOAT) &&
               (expectedType != INTEGER_OR_FLOAT)) error = CLIPS_TRUE;
           typeParsed = CLIPS_TRUE;
           break;
         
         case STRING:
           if ((expectedType != KUNKNOWN) && 
               (expectedType != STRING)) error = CLIPS_TRUE;
           typeParsed = CLIPS_TRUE;
           break;
         
         case SYMBOL:
#if OBJECT_SYSTEM
         case INSTANCE_NAME:
#endif
           if ((expectedType != KUNKNOWN) && 
               (expectedType != SYMBOL)) error = CLIPS_TRUE;
           typeParsed = CLIPS_TRUE;
           break;
           
         case BWORD:
           if (strcmp(inputToken.print_rep,"?VARIABLE") == 0) variableParsed = CLIPS_TRUE;
           else error = CLIPS_TRUE;
           break;
         
         default: 
           SyntaxErrorMessage("deftemplate allowed-... attribute");
           DeftemplateError = CLIPS_TRUE;
           return(0);
        }
        
      if (typeParsed && variableParsed)
        {
         PrintCLIPS(WERROR,"\n?VARIABLE must appear by itself in the ");
         PrintCLIPS(WERROR,attribute);
         PrintCLIPS(WERROR," attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      if (error)
        {
         PrintCLIPS(WERROR,"\nValue does not match expected type for the ");
         PrintCLIPS(WERROR,attribute);
         PrintCLIPS(WERROR," attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(0);
        }
        
      /*===========================*/
      /* Add the type restriction. */
      /*===========================*/
      
      new_item = GenConstant(inputToken.type,inputToken.value);
      new_item->next_arg = slotPtr->restrictionList;
      slotPtr->restrictionList = new_item;
      
      /*====================================*/
      /* Begin parsing the next type value. */
      /*====================================*/
      
      GetToken(readSource,&inputToken);
     }
        
   /*======================================================*/
   /* There must be at least one value for this attribute. */
   /*======================================================*/
        
   if ((! typeParsed) && (! variableParsed))
     {
      SyntaxErrorMessage("deftemplate allowed-... attribute");  
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
     
   /*======================================*/
   /* If ?VARIABLE was parsed, then remove */ 
   /* the restrictions for that type.      */
   /*======================================*/
     
   if (variableParsed)
     {
      switch(expectedType)
        {
         case KUNKNOWN:
           slotPtr->anyRestriction = CLIPS_FALSE;
           break;
         
         case SYMBOL:
           slotPtr->symbolRestriction = CLIPS_FALSE;
           break;
         
         case STRING:
           slotPtr->stringRestriction = CLIPS_FALSE;
           break;
           
         case INTEGER:
           slotPtr->integerRestriction = CLIPS_FALSE;
           break;
         
         case FLOAT:
           slotPtr->floatRestriction = CLIPS_FALSE;
           break;
 
         case INTEGER_OR_FLOAT:
           slotPtr->floatRestriction = CLIPS_FALSE;
           slotPtr->integerRestriction = CLIPS_FALSE;
           break;
        }
     }
     
   /*=====================================*/
   /* Fix up pretty print representation. */
   /*=====================================*/
   
   PPBackup();
   PPBackup();
   SavePPBuffer(")");
   
   return(1);
  }
  
/***********************************************************/
/* CheckSlotConflicts: Determines a slot has any conflicts */
/*   in the attribute specifications.                      */
/***********************************************************/
static BOOLEAN CheckSlotConflicts(slotPtr)
  struct templateSlot *slotPtr;
  {
   struct expr *defaultPtr;
   
   if (slotPtr->anyAllowed == CLIPS_TRUE)
     { /* Do Nothing */ }
   else if (slotPtr->symbolRestriction && (slotPtr->symbolsAllowed == CLIPS_FALSE))
     {
      PrintCLIPS(WERROR,"\nType attribute conflicts with allowed-symbols attribute\n");
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
   else if (slotPtr->stringRestriction && (slotPtr->stringsAllowed == CLIPS_FALSE))
     {
      PrintCLIPS(WERROR,"\nType attribute conflicts with allowed-strings attribute\n");
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
   else if (slotPtr->integerRestriction && (slotPtr->integersAllowed == CLIPS_FALSE))
     {
      PrintCLIPS(WERROR,"\nType attribute conflicts with allowed-integers/numbers attribute\n");
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
   else if (slotPtr->floatRestriction && (slotPtr->floatsAllowed == CLIPS_FALSE))
     {
      PrintCLIPS(WERROR,"\nType attribute conflicts with allowed-floats/numbers attribute\n");
      DeftemplateError = CLIPS_TRUE;
      return(CLIPS_FALSE);
     }
     
   /*================================================================*/
   /* Check to see if range attribute conflicts with type attribute. */
   /*================================================================*/
   
   if ((slotPtr->maxValue != NULL) && (slotPtr->anyAllowed == CLIPS_FALSE))
     {
      if (((slotPtr->maxValue->type == INTEGER) && 
          (slotPtr->integersAllowed == CLIPS_FALSE)) ||
          ((slotPtr->maxValue->type == FLOAT) && 
           (slotPtr->floatsAllowed == CLIPS_FALSE)))
        {
         PrintCLIPS(WERROR,"\nType attribute conflicts with range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(CLIPS_FALSE);
        }
     }
  
   if ((slotPtr->minValue != NULL) && (slotPtr->anyAllowed == CLIPS_FALSE))
     {
      if (((slotPtr->minValue->type == INTEGER) && 
          (slotPtr->integersAllowed == CLIPS_FALSE)) ||
          ((slotPtr->minValue->type == FLOAT) && 
           (slotPtr->floatsAllowed == CLIPS_FALSE)))
        {
         PrintCLIPS(WERROR,"\nType attribute conflicts with range attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(CLIPS_FALSE);
        }
     }
 
   /*==========================================================*/
   /* Check to see if default values violate other attributes. */
   /*==========================================================*/
   
   defaultPtr = slotPtr->defaultList;
   while (defaultPtr != NULL)
     {
      if (CheckSlotType(defaultPtr->type,slotPtr) == CLIPS_FALSE)
        {
         PrintCLIPS(WERROR,"\nType attribute conflicts with default attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(CLIPS_FALSE);
        }
      else if (CheckSlotAllowedValues(defaultPtr->type,defaultPtr->value,slotPtr) == CLIPS_FALSE)
        {
         PrintCLIPS(WERROR,"\nAllowed-... attribute conflicts with default attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(CLIPS_FALSE);
        }
      else if (CheckSlotRange(defaultPtr->type,defaultPtr->value,slotPtr) == CLIPS_FALSE)
        {
         PrintCLIPS(WERROR,"\nRange attribute conflicts with default attribute\n");
         DeftemplateError = CLIPS_TRUE;
         return(CLIPS_FALSE);
        }
      
      defaultPtr = defaultPtr->next_arg;
     }
     
   return(CLIPS_TRUE);
  }
  
/*******************************************/
/* MultiplyDefinedSlots: Determines if two */
/*   slots have been given the same name.  */
/*******************************************/
static BOOLEAN MultiplyDefinedSlots(first_slot)
  struct templateSlot *first_slot;
  {
   struct templateSlot *next_slot;
   int rv = CLIPS_FALSE;
   
   while (first_slot != NULL)
     {
      next_slot = first_slot->next;
      while (next_slot != NULL)
        {
         if (next_slot->slotName == first_slot->slotName)
           {
            PrintCLIPS(WERROR,"\nMultiple occurences of slot ");
            PrintCLIPS(WERROR,first_slot->slotName->contents);
            PrintCLIPS(WERROR,"\n");
            rv = CLIPS_TRUE;
           }
         next_slot = next_slot->next;
        }
      first_slot = first_slot->next;
     }
  
   return(rv);
  }
  
#endif


