   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*              FACTS MANAGER HEADER FILE              */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_facts

#define _H_facts

struct element;
struct fact;
struct lelement;
struct fhash;

#include "match.h"
#include "evaluatn.h"

struct element
  {
   int type;
   VOID *value;
  };
  
struct fact
  {
   struct element *atoms;   
   VOID *list;
   VOID *dependents;          
   long int factIndex;
   int depth;
   int count;
   unsigned int garbage : 1;
   unsigned int factLength : 15;               
   struct fact *next;     
   struct fact *previous;  
  };

typedef struct fact * SEGMENT;
typedef struct element * ELEMENT_PTR;

struct lelement
  {
   int type;
   VOID *value;
   struct lelement *next;
  };

struct fhash 
  {
   struct fact *fptr;
   struct fhash *next;
  };

#define SIZE_FACT_HASH  1013
#define CreateElements(size) CreateFact(size)

#ifdef LOCALE
#undef LOCALE
#endif
#ifdef _FACTMNGR_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
   LOCALE BOOLEAN                        FactCompare(struct fact *,struct fact *);
   LOCALE VOID                           ListFacts(void);
   LOCALE VOID                           PrintFactWithIdentifier(char *,struct fact *);
   LOCALE VOID                           PrintFact(char *,struct fact *);
   LOCALE int                            HashFact(struct fact *);
   LOCALE struct fact                   *FactExists(struct fact *,int);
   LOCALE VOID                           AddHashedFact(struct fact *,int);
   LOCALE BOOLEAN                        RemoveHashedFact(struct fact *);
   LOCALE BOOLEAN                        RetractFact(VOID *);
   LOCALE VOID                           RemoveOldFacts(void);
   LOCALE VOID                          *AddFact(VOID *);
   LOCALE VOID                           RemoveAllFacts(void);
   LOCALE VOID                          *CreateFact(int);
   LOCALE VOID                           ReturnElements(struct fact *);
   LOCALE VOID                           FactInstall(struct fact *);
   LOCALE VOID                           FactDeinstall(struct fact *);
   LOCALE VOID                           SegmentInstall(struct fact *);
   LOCALE VOID                           SegmentDeinstall(struct fact *);
   LOCALE VOID                           SetFactID(long);
   LOCALE VOID                          *GetNextFact(VOID *);
   LOCALE VOID                           GetFactPPForm(char *,int,VOID *);
   LOCALE long int                       GetFactIndex(VOID *);
   LOCALE VOID                          *AssertString(char *);
   LOCALE struct fact                   *StringToFact(char *);
   LOCALE struct fact                   *StringToMultifield(char *);
   LOCALE int                            GetFactListChanged(void);
   LOCALE VOID                           SetFactListChanged(int);
   LOCALE long int                       GetNumberOfFacts(void);
   LOCALE VOID                          *CreateMultifield(int);
   LOCALE VOID                           AddToSegmentList(struct fact *);
   LOCALE VOID                           FlushSegments(void);
   LOCALE VOID                           DuplicateSegment(struct dataObject *,struct dataObject *);
   LOCALE VOID                           InitializeFacts(void);
   LOCALE BOOLEAN                        GetFactDuplication(void);
   LOCALE BOOLEAN                        SetFactDuplication(int);  
   LOCALE struct fact                   *FindIndexedFact(long);
   LOCALE VOID                           IncrementFactCount(VOID *);
   LOCALE VOID                           DecrementFactCount(VOID *);
#else
   LOCALE BOOLEAN                        FactCompare();
   LOCALE VOID                           ListFacts();
   LOCALE VOID                           PrintFactWithIdentifier();
   LOCALE VOID                           PrintFact();
   LOCALE int                            HashFact();
   LOCALE struct fact                   *FactExists();
   LOCALE VOID                           AddHashedFact();
   LOCALE BOOLEAN                        RemoveHashedFact();
   LOCALE BOOLEAN                        RetractFact();
   LOCALE VOID                           RemoveOldFacts();
   LOCALE VOID                          *AddFact();
   LOCALE VOID                           RemoveAllFacts();
   LOCALE VOID                          *CreateFact();
   LOCALE VOID                           ReturnElements();
   LOCALE VOID                           FactInstall();
   LOCALE VOID                           FactDeinstall();
   LOCALE VOID                           SegmentInstall();
   LOCALE VOID                           SegmentDeinstall();
   LOCALE VOID                           SetFactID();
   LOCALE VOID                          *GetNextFact();
   LOCALE VOID                           GetFactPPForm();
   LOCALE long int                       GetFactIndex();
   LOCALE VOID                          *AssertString();
   LOCALE struct fact                   *StringToFact();
   LOCALE struct fact                   *StringToMultifield();
   LOCALE int                            GetFactlistChanged();
   LOCALE VOID                           SetFactlistChanged();
   LOCALE long int                       GetNumberOfFacts();
   LOCALE VOID                          *CreateMultifield();
   LOCALE VOID                           AddToSegmentList();
   LOCALE VOID                           FlushSegments();
   LOCALE VOID                           DuplicateSegment();
   LOCALE VOID                           InitializeFacts();
   LOCALE BOOLEAN                        GetFactDuplication();
   LOCALE BOOLEAN                        SetFactDuplication();
   LOCALE struct fact                   *FindIndexedFact();
   LOCALE VOID                           IncrementFactCount();
   LOCALE VOID                           DecrementFactCount();
#endif

#ifndef _FACTMNGR_SOURCE_
   extern int                            ChangeToFactList;
#endif

#endif





