   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_genrcfun
#define _H_genrcfun

typedef struct restriction RESTRICTION;
typedef struct method METHOD;
typedef struct generic_func GENERIC_FUNC;

#ifndef _CLIPS_STDIO_
#define _CLIPS_STDIO_
#include <stdio.h>
#endif

#include "symbol.h"
#include "expressn.h"
#include "evaluatn.h"

#if OBJECT_SYSTEM
#include "object.h"
#endif

struct restriction
  {
   unsigned tcnt;
   VOID **types;
   EXPRESSION *query;
  };

struct method
  {
   unsigned index,busy;
   int minRestrictions,maxRestrictions;
   RESTRICTION *restrictions;
   EXPRESSION *actions;
   char *pp_form;
  };

struct generic_func
  {
   SYMBOL_HN *name;
   unsigned busy;
   METHOD *methods;
   struct FunctionDefinition *sysfunc;
   unsigned mcnt,new_index;
   char *pp_form;
   long bsaveIndex;
   struct generic_func *nxt,*prv;
  };

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _GENRCFUN_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER

#if (! BLOAD_ONLY) && (! RUN_TIME)

LOCALE METHOD *AddMethod(GENERIC_FUNC *,METHOD *,int,unsigned,EXPRESSION *,
                      int,SYMBOL_HN *,EXPRESSION *,char *);
LOCALE GENERIC_FUNC *AddGeneric(SYMBOL_HN *,int *);
LOCALE METHOD *FindMethodByRestrictions(GENERIC_FUNC *,EXPRESSION *,int,
                                        SYMBOL_HN *,int *);
LOCALE int ClearDefmethods(void);
LOCALE int ClearDefgenerics(void);
LOCALE VOID MethodAlterError(char *);
LOCALE VOID DeleteMethodInfo(GENERIC_FUNC *,METHOD *);
LOCALE int MethodsExecuting(GENERIC_FUNC *);
#endif

LOCALE GENERIC_FUNC *FindDefgenericBySymbol(SYMBOL_HN *);
LOCALE int FindMethodByIndex(GENERIC_FUNC *,unsigned);
LOCALE VOID GenericDispatch(GENERIC_FUNC *,EXPRESSION *,DATA_OBJECT *);
#if DEBUGGING_FUNCTIONS
LOCALE VOID PreviewGeneric(void);
LOCALE VOID PrintMethod(char *,int,METHOD *);
#endif
LOCALE int NextMethodP(void);
LOCALE VOID CallNextMethod(DATA_OBJECT *);
LOCALE VOID SetGenericList(GENERIC_FUNC *);

#else

#if (! BLOAD_ONLY) && (! RUN_TIME)

LOCALE METHOD *AddMethod();
LOCALE GENERIC_FUNC *AddGeneric();
LOCALE METHOD *FindMethodByRestrictions();
LOCALE int ClearDefmethods();
LOCALE int ClearDefgenerics();
LOCALE VOID MethodAlterError();
LOCALE VOID DeleteMethodInfo();
LOCALE int MethodsExecuting();

#endif

LOCALE GENERIC_FUNC *FindDefgenericBySymbol();
LOCALE int FindMethodByIndex();
LOCALE VOID GenericDispatch();
#if DEBUGGING_FUNCTIONS
LOCALE VOID PreviewGeneric();
LOCALE VOID PrintMethod();
#endif
LOCALE int NextMethodP();
LOCALE VOID CallNextMethod();
LOCALE VOID SetGenericList();

#endif

#ifndef _GENRCFUN_SOURCE_
extern GENERIC_FUNC *GenericList,*CurrentGeneric,*GenericListBottom;
extern METHOD *CurrentMethod;
extern EXPRESSION *GenericStackFrame;
extern int GenericStackSize;
#if DEBUGGING_FUNCTIONS
extern int WatchGenerics,WatchMethods;
#endif
#endif

#endif





