   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*             LOGICAL DEPENDENCIES MODULE             */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _LGCLDPND_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include "constant.h"
#include "clipsmem.h"
#include "engine.h"
#include "reteutil.h"
#include "router.h"
#include "utility.h"
#include "lgcldpnd.h"

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   static struct partialMatch    *FindLogicalBind(struct joinNode *,struct partialMatch *);
   static VOID                    DetachAssociatedPMDependencies(struct partialMatch *,
                                                                    struct fact *);
   static VOID                    DetachAssociatedFactDependencies(struct fact *,
                                                                    struct partialMatch *);
#else
   static struct partialMatch    *FindLogicalBind();
   static VOID                    DetachAssociatedPMDependencies();
   static VOID                    DetachAssociatedFactDependencies();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct dependency     *DependencyList = NULL;
   
/*********************************************************************/
/* AddLogicalDependencies:         */
/*********************************************************************/
globle BOOLEAN AddLogicalDependencies(factPtr,existingFact)
  struct fact *factPtr;
  int existingFact;
  {
   struct partialMatch *theBinds;
   struct dependency *newDependency;
   
   /*==============================================*/
   /* If the rule has no logical patterns, then no */
   /* dependencies have to be established.         */
   /*==============================================*/
   
   if (TheLogicalJoin == NULL) 
     {
      if (existingFact) RemoveFactDependencies(factPtr);
      return(CLIPS_TRUE);
     }
   else if (existingFact && (factPtr->dependents == NULL)) 
     { return(CLIPS_TRUE); }
   
   /*=====================================================*/
   /* Find the binds in the logical join associated with  */
   /* activation binds. If the binds cannot be found,     */
   /* then the binds must have been deleted by a previous */
   /* RHS action and the fact should not be asserted.     */
   /*=====================================================*/
   
   theBinds = FindLogicalBind(TheLogicalJoin,GlobalLHSBinds);
   if (theBinds == NULL) return(CLIPS_FALSE);
      
   /*======================================================*/
   /* Add a dependency link between the fact and the partialMatch */
   /* and between the partialMatch and the fact.                  */
   /*======================================================*/
   
   newDependency = get_struct(dependency);
   newDependency->dPtr = (VOID *) factPtr;
   newDependency->next = (struct dependency *) theBinds->dependents;
   theBinds->dependents = (VOID *) newDependency;
   
   newDependency = get_struct(dependency);
   newDependency->dPtr = (VOID *) theBinds;
   newDependency->next = (struct dependency *) factPtr->dependents;
   factPtr->dependents = (VOID *) newDependency;
   
   /*========================================================*/
   /* Return true to indicate that the fact can be asserted. */
   /*========================================================*/
   
   return(TRUE);
  }

/*********************************************************************/
/* AddLogicalDependencies:         */
/*********************************************************************/
static struct partialMatch *FindLogicalBind(theJoin,theBinds)
  struct joinNode *theJoin;
  struct partialMatch *theBinds;
  {
   struct partialMatch *compPtr;
   unsigned int max, i;
   int found;
   
   compPtr = theJoin->beta;
   while (compPtr != NULL)
     {
      max = compPtr->bcount;
      found = CLIPS_TRUE;
      
      for (i = 0; i < max; i++)
        {
         if (compPtr->binds[i]->whoset != theBinds->binds[i]->whoset)
           {
            i = max;
            found = CLIPS_FALSE;
           }
        }
        
      if (found) return(compPtr);
      compPtr = compPtr->next;
     }
     
   return(NULL);
  }

/*********************************************************************/
/* RemoveFactDependencies:         */
/*********************************************************************/
globle VOID RemoveFactDependencies(theFact)
  struct fact *theFact;
  {
   struct dependency *fdPtr, *nextPtr;
   struct partialMatch *theBinds;
   
   fdPtr = (struct dependency *) theFact->dependents;
   
   while (fdPtr != NULL)
     {
      nextPtr = fdPtr->next;
      
      theBinds = (struct partialMatch *) fdPtr->dPtr;
      DetachAssociatedPMDependencies(theBinds,theFact);
      rtn_struct(dependency,fdPtr);
      fdPtr = nextPtr;
     }
     
   theFact->dependents = NULL;
  }

/*********************************************************************/
/* DetachAssociatedPMDependencies:         */
/*********************************************************************/
static VOID DetachAssociatedPMDependencies(theBinds,theFact)
  struct partialMatch *theBinds;
  struct fact *theFact;
  {
   struct dependency *fdPtr, *nextPtr, *lastPtr = NULL;
   
   fdPtr = (struct dependency *) theBinds->dependents;
   
   while (fdPtr != NULL)
     {
      if (fdPtr->dPtr == (VOID *) theFact)
        {
         nextPtr = fdPtr->next;
         if (lastPtr == NULL) theBinds->dependents = (VOID *) nextPtr;
         else lastPtr->next = nextPtr;
         rtn_struct(dependency,fdPtr);
         fdPtr = nextPtr;
        }
      else
        {
         lastPtr = fdPtr;
         fdPtr = fdPtr->next;
        }
     }
  }
  
/*********************************************************************/
/* RemovePMDependencies:         */
/*********************************************************************/
globle VOID RemovePMDependencies(theBinds)
  struct partialMatch *theBinds;
  {
   struct dependency *fdPtr, *nextPtr;
   struct fact *theFact;
   
   fdPtr = (struct dependency *) theBinds->dependents;
   
   while (fdPtr != NULL)
     {
      nextPtr = fdPtr->next;
      
      theFact = (struct fact *) fdPtr->dPtr;
      DetachAssociatedFactDependencies(theFact,theBinds);
      rtn_struct(dependency,fdPtr);
      fdPtr = nextPtr;
     }
     
   theBinds->dependents = NULL;
  }

/*********************************************************************/
/* DetachAssociatedFactDependencies:         */
/*********************************************************************/
static VOID DetachAssociatedFactDependencies(theFact,theBinds)
  struct fact *theFact;
  struct partialMatch *theBinds;
  {
   struct dependency *fdPtr, *nextPtr, *lastPtr = NULL;
   
   fdPtr = (struct dependency *) theFact->dependents;
   
   while (fdPtr != NULL)
     {
      if (fdPtr->dPtr == (VOID *) theBinds)
        {
         nextPtr = fdPtr->next;
         if (lastPtr == NULL) theFact->dependents = (VOID *) nextPtr;
         else lastPtr->next = nextPtr;
         rtn_struct(dependency,fdPtr);
         fdPtr = nextPtr;
        }
      else
        {
         lastPtr = fdPtr;
         fdPtr = fdPtr->next;
        }
     }
  }
  
/*********************************************************************/
/* AddToDependencyList:                  */
/*********************************************************************/
globle VOID AddToDependencyList(theBinds)
  struct partialMatch *theBinds;
  {
   struct dependency *dlPtr, *tempPtr;
   struct fact *theFact;
   
   dlPtr = (struct dependency *) theBinds->dependents;
   
   while (dlPtr != NULL)
     {
      tempPtr = dlPtr->next;
      
      theFact = (struct fact *) dlPtr->dPtr;
      DetachAssociatedFactDependencies(theFact,theBinds);
      if (theFact->dependents == NULL)
        {
         dlPtr->next = DependencyList;
         DependencyList = dlPtr;
        }
      else
        { rtn_struct(dependency,dlPtr); }
      
      dlPtr = tempPtr;
     }
     
   theBinds->dependents = NULL;
  }
  
/*********************************************************************/
/* GetNextLogicalRetraction:                  */
/*********************************************************************/
globle struct fact *GetNextLogicalRetraction()
  {
   struct dependency *tempPtr;
   struct fact *rv;
   
   if (DependencyList == NULL) return(NULL);
     
   rv = (struct fact *) DependencyList->dPtr;
   tempPtr = DependencyList->next;
   rtn_struct(dependency,DependencyList);
   DependencyList = tempPtr;
   
   return(rv);
  }
  
/*********************************************************************/
/* ForceLogicalRetractions:                  */
/*********************************************************************/
globle VOID ForceLogicalRetractions()
  {
   struct dependency *tempPtr;
   VOID *theFact;
   
   if (DependencyList == NULL) return;
   
   theFact = DependencyList->dPtr;
   
   tempPtr = DependencyList->next;
   rtn_struct(dependency,DependencyList);
   DependencyList = tempPtr;
   
   RetractFact(theFact);
  }

/*********************************************************************/
/* ListDependencies:         */
/*********************************************************************/
globle VOID ListDependencies(theFact)
  struct fact *theFact;
  {
   struct dependency *fdPtr;
   struct partialMatch *theBinds;
   
   fdPtr = (struct dependency *) theFact->dependents;
   
   if (fdPtr == NULL) PrintCLIPS(WDISPLAY,"None\n");
     
   while (fdPtr != NULL)
     {
      theBinds = (struct partialMatch *) fdPtr->dPtr;
      
      PrintPartialMatch(WDISPLAY,theBinds);
      PrintCLIPS(WDISPLAY,"\n");
      
      fdPtr = fdPtr->next;
     }
  }
  
/*********************************************************************/
/* ListDependents:         */
/*********************************************************************/
globle VOID ListDependents(theFact)
  struct fact *theFact;
  {
   struct fact *factPtr;
   struct dependency *fdPtr;
   struct partialMatch *theBinds;
   int found = CLIPS_FALSE;
   
   factPtr = (struct fact *) GetNextFact(NULL);
   
   while (factPtr != NULL)
     {
      fdPtr = (struct dependency *) factPtr->dependents;
      while (fdPtr != NULL)
        {
         theBinds = (struct partialMatch *) fdPtr->dPtr;
         if (FindFactInPartialMatch(theFact->factIndex,theBinds) == CLIPS_TRUE)
           {
            if (found) PrintCLIPS(WDISPLAY,",");
            PrintCLIPS(WDISPLAY,"f-");
            PrintLongInteger(WDISPLAY,factPtr->factIndex);
            found = CLIPS_TRUE;
            fdPtr = NULL;
           }
         else
           { fdPtr = fdPtr->next; }
        }
      factPtr = (struct fact *) GetNextFact(factPtr);
     }
     
   if (! found) PrintCLIPS(WDISPLAY,"None\n");
   else PrintCLIPS(WDISPLAY,"\n");
  }

#endif

