   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    MATCH MODULE                     */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _MATCH_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"

#if DEFRULE_CONSTRUCT

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "facts.h"
#include "match.h"
#include "network.h"
#include "generate.h"
#include "drive.h"
#include "reteutil.h"
#include "router.h"
#include "defrule.h"

/***************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/***************************************/

#if ANSI_COMPILER  
   static BOOLEAN                  EvaluatePatternExpression(struct element *,struct expr *,struct patternNode *);
   static VOID                     TraceErrorToPattern(struct patternNode *);
   static struct multifieldMarker *CopySegmentMarkers(struct multifieldMarker *);
#else
   static BOOLEAN                  EvaluatePatternExpression();
   static VOID                     TraceErrorToPattern();
   static struct multifieldMarker *CopySegmentMarkers();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static struct fact             *CurrentPatternFact;
   static struct multifieldMarker *CurrentPatternMarks;
   
/*****************************************************************/
/* PatternMatch:  Compares the elements of a fact to the nodes in the */
/*   pattern network.  Arguments are:                            */
/*   fact_ptr:  Pointer to the fact being compared.              */
/*   elem_ptr:  Pointer to the list of atoms in the fact.        */
/*              Redundant argument because it can be derived     */
/*              from fact_ptr.                                   */
/*   pat_pos:   Current index into the pattern being matched.    */
/*   cur_elem:  The index of the current atom being pointed to   */
/*              in the fact.                                     */
/*   markers:   The list of segment markers generated during the */
/*              pattern match.                                   */  
/*****************************************************************/
globle VOID PatternMatch(fact_ptr,elem_ptr,patn_ptr,pat_pos,cur_elem,markers,end_mark)
  struct fact *fact_ptr;
  struct element *elem_ptr;
  struct patternNode *patn_ptr;
  int pat_pos;
  int cur_elem;
  struct multifieldMarker *markers, *end_mark;
  {
   struct joinNode *drive_list;
   struct partialMatch *var_list;
   struct patternMatch *list_of_matches;
   int finish_match, length, end_epos;
   struct multifieldMarker *new_mark, *old_mark;
   struct alphaMatch *afbtemp;

   if (patn_ptr == NULL) return;
   
   CurrentPatternFact = fact_ptr;
   CurrentPatternMarks = markers;
   length = fact_ptr->factLength;   
   while (CLIPS_TRUE)
     {
      /*================================================*/
      /* 5.1 Modification: Added this to if conditions. */
      /*================================================*/
      
      /* finish_match = CLIPS_FALSE; */

      /*==========================================================*/
      /* If an incremental reset is being performed and the node  */
      /* is not part of the network to be reset, then the pattern */
      /* matching attempt for this pattern has failed.            */
      /*==========================================================*/
      
      if (IncrementalResetFlag && (patn_ptr->initialize == CLIPS_FALSE))
        { finish_match = CLIPS_TRUE; }
        
      /*========================================================*/
      /* 5.1 Modification: Added this to stop unecessary tests. */
      /*========================================================*/
      
      else if ((cur_elem < length) && (patn_ptr->stopNode != CLIPS_TRUE))
        { finish_match = CLIPS_FALSE; }
        
      /*==========================================================*/
      /* If there are no elements in the fact left to match, and  */
      /* the pattern element being matched against is not an end  */
      /* of pattern marker or a multiple element binder, then the */
      /* pattern matching attempt for this pattern has failed.    */
      /*==========================================================*/

      else if ((cur_elem == length) && 
               (patn_ptr->stopNode == CLIPS_FALSE) &&
               (patn_ptr->multifieldNode == CLIPS_FALSE))
        { finish_match = CLIPS_TRUE; }

      /*===========================================================*/
      /* If there are elements in the fact left to match, and the  */
      /* pattern element being matched against is an end of        */
      /* pattern marker, then the fact is too long for the pattern */
      /* and pattern matching attempt has failed.                  */
      /*===========================================================*/

      else if ((cur_elem < length) && (patn_ptr->stopNode))
        { finish_match = CLIPS_TRUE; }

      /*===========================================================*/
      /* If there are no elements in the fact left to match, and   */
      /* the pattern element being matched against is an end of    */
      /* pattern marker, then the pattern has matched.             */
      /*===========================================================*/
    
      else if ((cur_elem == length) && (patn_ptr->stopNode))
        {
         /*=======================================================*/
         /* Add the pattern to the list of matches for this fact. */
         /*=======================================================*/
        
         list_of_matches = (struct patternMatch *) fact_ptr->list;
         fact_ptr->list = (VOID *) get_struct(patternMatch);
         ((struct patternMatch *) fact_ptr->list)->next = list_of_matches;
         ((struct patternMatch *) fact_ptr->list)->matchingPattern = patn_ptr;

         /*============================================*/
         /* Get the fact bindings associated with this */
         /* pattern, and "drive" them to the joins     */
         /* connected to this pattern.                 */
         /*============================================*/
 
         var_list = get_struct(partialMatch);
         var_list->next = NULL;
         var_list->betaMemory = CLIPS_FALSE;
         var_list->busy = CLIPS_FALSE;
         var_list->activationf = CLIPS_FALSE;
         var_list->dependentsf = CLIPS_FALSE;
         var_list->notOriginf = CLIPS_FALSE;
         var_list->counterf = CLIPS_FALSE;
         var_list->dependents = NULL;
         var_list->bcount = 1;

         afbtemp = get_struct(alphaMatch);
         afbtemp->aorigin = afbtemp;
         afbtemp->whoset = fact_ptr->factIndex;
         afbtemp->origin = fact_ptr;
         if (markers != NULL)
           { afbtemp->marker = CopySegmentMarkers(markers); }
         else
           { afbtemp->marker = NULL; }
         var_list->binds[0] = afbtemp;
         var_list->next = patn_ptr->alpha;
         patn_ptr->alpha = var_list;
            
         drive_list = patn_ptr->entry_join;
         while (drive_list != NULL)
           {   
            Drive(var_list,drive_list,RHS,CLIPS_FALSE);
            drive_list = drive_list->right_match_node;
           }
         finish_match = CLIPS_TRUE;
        }
        
      /*==========================================*/
      /* 5.1 Modification: Added this as default. */
      /*==========================================*/

      else
        { finish_match = CLIPS_FALSE; }
        
      /*====================================================*/
      /* Compare the pattern element with the fact element. */
      /*====================================================*/
     
      if (finish_match == CLIPS_TRUE)
        {
         while ((patn_ptr->right_node == NULL) ||
                (patn_ptr->blocked))
           {
            if (patn_ptr->blocked) patn_ptr->blocked = CLIPS_FALSE;
            patn_ptr = patn_ptr->last_level;
            if (patn_ptr == NULL) 
              { return; }
            else if (patn_ptr->multifieldNode) 
              { return; }

            pat_pos--;
            cur_elem--;
           }
         patn_ptr = patn_ptr->right_node;
        }
      else if (patn_ptr->singlefieldNode)
        {
         /*==================================================*/
         /* Pattern element is a constant.  Check that the   */
         /* fact element is the same type and value, and if  */
         /* it is then proceed to the next level of compare. */
         /*==================================================*/
   
         if (patn_ptr->eval == NULL)
           {
            cur_elem++;
            pat_pos++;
            patn_ptr = patn_ptr->next_level;
           }         
         else if ((patn_ptr->eval->value == PTR_CONSTANT) ?
                  ((patn_ptr->eval->arg_list->type == elem_ptr[cur_elem].type) &&
                   (patn_ptr->eval->arg_list->value == elem_ptr[cur_elem].value)) : CLIPS_FALSE)
           {
            patn_ptr->blocked = CLIPS_TRUE;
            cur_elem++;
            pat_pos++;
            patn_ptr = patn_ptr->next_level;
           }
         else if (EvaluatePatternExpression(&elem_ptr[cur_elem],patn_ptr->eval,patn_ptr) != CLIPS_FALSE)
           {
            if (patn_ptr->eval->value == PTR_CONSTANT) 
              { patn_ptr->blocked = CLIPS_TRUE; }
            cur_elem++;
            pat_pos++;
            patn_ptr = patn_ptr->next_level; 
           }
         else
           {
            EvaluationError = CLIPS_FALSE;
            while ((patn_ptr->right_node == NULL) ||
                   (patn_ptr->blocked))
              {
               if (patn_ptr->blocked) patn_ptr->blocked = CLIPS_FALSE;
               patn_ptr = patn_ptr->last_level;
               if (patn_ptr == NULL) 
                 { return; }
               else if (patn_ptr->multifieldNode) 
                { return; }
               pat_pos--;
               cur_elem--;
              }
            patn_ptr = patn_ptr->right_node;
           }         
        }
      else if (patn_ptr->multifieldNode)
        {
         /*======================================================*/
         /* Pattern element is a '$?'.  The type or value of the */ 
         /* fact element does not matter, however multiple paths */
         /* of comparison have to be followed since the '$?' can */
         /* bind to zero or more fact elements.                  */
         /*======================================================*/

         /* special case: next level is stop and no others at that level */
         /* $? binds to all elements.                                    */

         old_mark = markers;
         if ((patn_ptr->next_level->stopNode) &&
             (patn_ptr->next_level->right_node == NULL))
           {
            new_mark = get_struct(multifieldMarker);
            new_mark->element = pat_pos;
            new_mark->start = cur_elem + 1;
            new_mark->end = length;
            new_mark->next = NULL;
            if (end_mark == NULL)
              { 
               markers = new_mark; 
               CurrentPatternMarks = markers;
              }
            else
              { end_mark->next = new_mark; }
            
            if (patn_ptr->eval == NULL)
              {
               PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                    pat_pos+1,length,markers,new_mark);
              }
            else if (EvaluatePatternExpression(&elem_ptr[cur_elem],patn_ptr->eval,patn_ptr) != CLIPS_FALSE)
              {
               PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                       pat_pos+1,length,markers,new_mark);
              }
           }
         else
           {
            /* Bind to no elements */

            new_mark = get_struct(multifieldMarker);
            new_mark->element = pat_pos;
            new_mark->start = cur_elem + 1;
            new_mark->end = cur_elem;
            new_mark->next = NULL;
            if (end_mark == NULL)
              {
               markers = new_mark; 
               CurrentPatternMarks = markers;
              }
            else
              { end_mark->next = new_mark; }
              
            if (patn_ptr->eval == NULL)
              {
               PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                       pat_pos+1,cur_elem,markers,new_mark);
              }
            else if (EvaluatePatternExpression(&elem_ptr[cur_elem],patn_ptr->eval,patn_ptr) != CLIPS_FALSE)
              {
               PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                       pat_pos+1,cur_elem,markers,new_mark);
              }

            /* Consider rest of _bindings */
        
            end_epos = cur_elem;
            while (end_epos < length)
              {
               new_mark->element = pat_pos;
               new_mark->start = cur_elem + 1;
               new_mark->end = end_epos + 1;
               new_mark->next = NULL;
               end_epos++;
               
               if (patn_ptr->eval == NULL)
                 {
                  PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                          pat_pos+1,end_epos,markers,new_mark);
                 }
               else if (EvaluatePatternExpression(&elem_ptr[cur_elem],patn_ptr->eval,patn_ptr) != CLIPS_FALSE)
                 {
                  PatternMatch(fact_ptr,elem_ptr,patn_ptr->next_level,
                          pat_pos+1,end_epos,markers,new_mark);
                 }
              }
           }

         rtn_struct(multifieldMarker,new_mark);
         if (end_mark != NULL) end_mark->next = NULL;
         markers = old_mark;
         CurrentPatternMarks = old_mark;

         while ((patn_ptr->right_node == NULL) ||
                (patn_ptr->blocked))
           {
            if (patn_ptr->blocked) patn_ptr->blocked = CLIPS_FALSE;
            patn_ptr = patn_ptr->last_level;
            if (patn_ptr == NULL) 
              { return; }
            else if (patn_ptr->multifieldNode) 
              { return; }

            pat_pos--;
            cur_elem--;
           }
         patn_ptr = patn_ptr->right_node;
        }
     }
  }

/***************************************************************/
/* EvaluatePatternExpression: Performs a faster evaluation for */
/*   pattern expressions than if EvaluateExpression were used  */
/*   directly. This function evaluates calls to function       */
/*   constant, function notconstant, function and, and         */
/*   function or. No other function calls are made in pattern  */
/*   expressions.                                              */
/***************************************************************/
static int EvaluatePatternExpression(elem_ptr,pat_test,patn_ptr)
  struct element *elem_ptr;
  struct expr *pat_test;
  struct patternNode *patn_ptr;
  {
   int pass, fail;
   DATA_OBJECT vresult;
   
   int e1, e2;
   struct expr *tptr;
   struct element *elem_a, *elem_b;
   struct fact *fact_ptr;
   int a_extent, b_extent;
   struct multifieldMarker *marks; 
   
   /*=============================================================*/
   /* Evaluate expressions constant and notconstant.  Calls to    */
   /* these functions are expressed in the format:                */
   /*   (constant <value>)                                        */
   /*   (notconstant <value>)                                     */
   /* Function constant returns true (1.0) if the element pointer */
   /* has the same type and value as the expression value.        */  
   /* Function notconstant acts similarly to constant except the  */
   /* return value is negated.                                    */
   /*=============================================================*/
   
   if ((pat_test->value == PTR_CONSTANT) ||
       (pat_test->value == PTR_NOTCONSTANT))
     {
      /*==========================================*/
      /* Determine the appropriate return values. */
      /*==========================================*/

      if (pat_test->value != PTR_NOTCONSTANT)
        {
         pass = CLIPS_TRUE;
         fail = CLIPS_FALSE;
        }
      else
        {
         pass = CLIPS_FALSE;
         fail = CLIPS_TRUE;
        }

      /*========================================*/
      /* Compare the element to the test value. */
      /*========================================*/

      pat_test = pat_test->arg_list;
      
      if (pat_test->type != elem_ptr->type)
        { return(fail); }
      else if (pat_test->value != elem_ptr->value)
        { return(fail); }
      else
        { return(pass); }
     }
     
   /*==============================================================*/
   /* Evaluate expressions eq_vars and neq_vars.  Calls to these   */
   /* functions are expressed in the format:                       */
   /*   (eq_vars  <pattern m> <element n> <pattern p> <element r>) */
   /*   (neq_vars <pattern m> <element n> <pattern p> <element r>) */
   /* Function eq_vars returns true (1.0) if the nth element of    */
   /* the mth pattern is equal to the rth element of the pth       */
   /* pattern.  Function neq_vars acts similarly to eq_vars except */
   /* the return value is negated.                                 */
   /*==============================================================*/

   else if ((pat_test->value == PTR_EQ_FIELD) ||
            (pat_test->value == PTR_NEQ_FIELD))
     {
      /*==========================================*/
      /* Determine the appropriate return values. */
      /*==========================================*/

      if (pat_test->value == PTR_EQ_FIELD)
        {
         pass = CLIPS_TRUE;
         fail = CLIPS_FALSE;
        }
      else
        {
         pass = CLIPS_FALSE;
         fail = CLIPS_TRUE;
        }

      /*==============================================*/
      /* Determine which elements are to be extracted */
      /* out of which patterns.                       */
      /*==============================================*/                  

      tptr = pat_test->arg_list;
      e1 = (int) ValueToLong(tptr->value);
      tptr = tptr->next_arg;
      e2 = (int) ValueToLong(tptr->value);

      /*==============================================*/
      /* Extract the fact pointer and list of segment */
      /* markers for the first pattern.               */
      /*==============================================*/

      fact_ptr = CurrentPatternFact;
      marks = CurrentPatternMarks;
      
      /*=============================================*/
      /* Determine the beginning point and extent of */
      /* the comparision for the first pattern.      */
      /*=============================================*/

      a_extent = b_extent = 1;
     
      if (marks != NULL)
        { 
         if (marks->element < e1)
           { 
            e1 = AdjustFieldPosition(marks,e1,&a_extent);
           }
         else if (marks->element == e1)
           { a_extent = (marks->end - marks->start) + 1; }
           
         if (marks->element < e2)
           { 
            e2 = AdjustFieldPosition(marks,e2,&b_extent);
           }
         else if (marks->element == e2)
           { b_extent = (marks->end - marks->start) + 1; }
        }
  
      /*===============================================*/
      /* If the extents are not the same, then the     */
      /* comparision fails (e.g. two segment variables */
      /* of different length.                          */
      /*===============================================*/

      if (a_extent != b_extent)
        { return(fail); }

      /*=========================================*/
      /* If the extents are both zero, then the  */
      /* comparision succeeds (e.g. two segment  */
      /* variables of length zero).              */
      /*=========================================*/

      if (a_extent == 0)
        { return(pass); }

      /*==============================================*/
      /* Compare the two variables element by element */
      /* for the length of the extent.                */
      /*==============================================*/

      e1--;
      elem_a = &fact_ptr->atoms[e1];
      e2--;
      elem_b = &fact_ptr->atoms[e2];

      while (CLIPS_TRUE)
        {
         if (elem_a->type != elem_b->type)
           { return(fail); }
         else if (elem_a->value != elem_b->value)
           { return(fail); }

         a_extent--;

         if (a_extent == 0) return(pass);

         elem_a++;
         elem_b++;
        }   
     }

   /*=========================================================*/
   /* Evaluate or expressions expressed in the format:        */
   /*   (or <expression 1> <expression 2> ... <expression n>) */
   /* Returns true (1.0) if any of the expression are true,   */
   /* otherwise returns false (0.0).                          */
   /*=========================================================*/

   else if (pat_test->value == PTR_OR)
     {
      pat_test = pat_test->arg_list;
      while (pat_test != NULL)
        {
         if (EvaluatePatternExpression(elem_ptr,pat_test,patn_ptr) == CLIPS_TRUE)
           { 
            if (EvaluationError) return(CLIPS_FALSE);
            return(CLIPS_TRUE); 
           }          
         if (EvaluationError) return(CLIPS_FALSE);
         pat_test = pat_test->next_arg;
        }
      return(CLIPS_FALSE);
     }

   /*==========================================================*/
   /* Evaluate and expressions expressed in the format:        */
   /*   (and <expression 1> <expression 2> ... <expression n>) */
   /* Returns false (0.0) if any of the expression are false,  */
   /* otherwise returns true (1.0).                            */
   /*==========================================================*/

   else if (pat_test->value == PTR_AND)
     {
      pat_test = pat_test->arg_list;
      while (pat_test != NULL)
        {
         if (EvaluatePatternExpression(elem_ptr,pat_test,patn_ptr) == CLIPS_FALSE)
           { return(CLIPS_FALSE); } 
         if (EvaluationError) return(CLIPS_FALSE);
         pat_test = pat_test->next_arg;
        }
      return(CLIPS_TRUE);
     } 
     
   /*=======================================================*/
   /* Evaluate all other expressions using EvaluateExpression. */
   /*=======================================================*/

   else
     {
      if (EvaluateExpression(pat_test,&vresult)) 
        {
         PatternNetErrorMessage(patn_ptr);
         return(CLIPS_FALSE);
        }
      
      if ((vresult.value == (VOID *) CLIPSFalseSymbol) && (vresult.type == SYMBOL))
        { return(CLIPS_FALSE); }
      else
        { return(CLIPS_TRUE); }
     }
  }
  
/*************************************************************/
/* GetFieldSysFunction:  Extracts the mth pattern of a rule. */
/*************************************************************/
globle VOID GetFieldSysFunction(bound_var)
  DATA_OBJECT_PTR bound_var;
  {
   int element;
   struct fact *fact_ptr;
   struct element *elem_ptr;
   struct multifieldMarker *marks;
   struct expr *test_ptr;
   int extent;

   test_ptr = GetFirstArgument();
   element = DOPToInteger(test_ptr);
     
   fact_ptr = CurrentPatternFact;
   marks = CurrentPatternMarks;

   extent = -1;

   if (marks != NULL)
     { 
      if (marks->element < element)
        { element = AdjustFieldPosition(marks,element,&extent); }
      else if (marks->element == element)
        { extent = (marks->end - marks->start) + 1; }
     }

   if (extent != -1)
     {
      bound_var->type = MULTIFIELD;
      bound_var->value = (VOID *) fact_ptr;
      bound_var->begin = --element;
      bound_var->end = element + extent - 1;
      return;
     }
   else
     { extent = 1; }  

   element--;
   elem_ptr = &fact_ptr->atoms[element];

   bound_var->type = elem_ptr->type;
   bound_var->value = elem_ptr->value;
  }
  
/*****************************************/
/* PatternNetErrorMessage:                     */
/*****************************************/
globle VOID PatternNetErrorMessage(patn_ptr)
  struct patternNode *patn_ptr;
  {
   int field = 1;
   struct patternNode *temp_ptr;
   char buffer[60];
   
   PrintCLIPS(WERROR,"\nThis error occurred in the pattern network\n");
   PrintCLIPS(WERROR,"   Currently active fact: ");
   PrintFact(WERROR,CurrentPatternFact);
   PrintCLIPS(WERROR,"\n");
   
   /*========================================================*/
   /* Determine the exact field in which the error occurred. */
   /*========================================================*/
   
   temp_ptr = patn_ptr;
   while (temp_ptr->last_level != NULL)
     {
      field++;
      temp_ptr = temp_ptr->last_level;
     }
     
   sprintf(buffer,"   Problem resides in field #%d\n",field);
   PrintCLIPS(WERROR,buffer);
   TraceErrorToPattern(patn_ptr->next_level); 
   PrintCLIPS(WERROR,"\n");  
  }
  
/***********************************/
/* TraceErrorToPattern:            */
/***********************************/
static VOID TraceErrorToPattern(patn_ptr)
  struct patternNode *patn_ptr;
  {
   struct joinNode *join_ptr;
   char buffer[60];
   
   while (patn_ptr != NULL)
     {
      if (patn_ptr->stopNode)
        {
         join_ptr = patn_ptr->entry_join;
         while (join_ptr != NULL)
           {
            sprintf(buffer,"      Of pattern #%d in rule(s):\n",join_ptr->depth);
            PrintCLIPS(WERROR,buffer);
            TraceErrorToRule(join_ptr->next_level,"         ");
            join_ptr = join_ptr->right_match_node;
           }
        }
      else
        { TraceErrorToPattern(patn_ptr->next_level); }
      patn_ptr = patn_ptr->right_node;
     }
  
  }
 
/***********************************/
/* TraceErrorToRule:            */
/***********************************/ 
globle VOID TraceErrorToRule(join_ptr,indent_spaces)
  struct joinNode *join_ptr;
  char *indent_spaces;
  {   
   char *name;
   
   while (join_ptr != NULL)
     {
      if (join_ptr->next_level == NULL)
        {
         name = GetDefruleName(join_ptr->eval);
         PrintCLIPS(WERROR,indent_spaces);
         PrintCLIPS(WERROR,name);
         PrintCLIPS(WERROR,"\n");
        }
      else
        { TraceErrorToRule(join_ptr->next_level,indent_spaces); }

      join_ptr = join_ptr->right_drive_node;
     }
  
  }
  
/*****************************************************************/
/* CopySegmentMarkers:  Copies a list of structures indicating how many  */
/*   elements a segment variable or wildcard has bound.          */
/*****************************************************************/
static struct multifieldMarker *CopySegmentMarkers(mark_list)
  struct multifieldMarker *mark_list;
  {
   struct multifieldMarker *head = NULL, *last_mark = NULL, *new_mark;

   while (mark_list != NULL)
     {
      new_mark = get_struct(multifieldMarker);
      new_mark->next = NULL;
      new_mark->element = mark_list->element;
      new_mark->start = mark_list->start;
      new_mark->end = mark_list->end;

      if (last_mark == NULL)
        { head = new_mark; }
      else
        { last_mark->next = new_mark; }
      last_mark = new_mark;

      mark_list = mark_list->next;
     }
   
   return(head);
  }
  
#endif

