   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    MEMORY MODULE                    */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _MEMORY_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"    
#include "constant.h"
#include "clipsmem.h"
#include "router.h"
#include "utility.h"  

#if ANSI_COMPILER
#include <stdlib.h>
#else
#if (! IBM_TBC) && (! IBM_MSC)
extern char *malloc();
#endif
#endif 

#if IBM_TBC
#include <alloc.h>
#endif
#if IBM_MSC
#include <malloc.h>
#endif

/****************************************/
/* LOCAL INTERNAL FUNCTION DEFINITIONS */
/****************************************/

#if BLOCK_MEMORY

#if ANSI_COMPILER  
   static int                     InitializeBlockMemory(unsigned int);
   static int                     AllocateBlock(struct blockInfo *,unsigned int);
   static VOID                    AllocateChunk(struct blockInfo *,struct chunkInfo *,unsigned int);
#else
   static int                     InitializeBlockMemory();
   static int                     AllocateBlock();
   static VOID                    AllocateChunk();
#endif

#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static long int                MemoryAmount = 0;
   static long int                MemoryCalls = 0;
   static BOOLEAN                 ConserveMemory = CLIPS_FALSE;
#if ANSI_COMPILER
   static int                   (*OutOfMemoryFunction)(int)
                                       = DefaultOutOfMemoryFunction;
#else
   static int                   (*OutOfMemoryFunction)()
                                       = DefaultOutOfMemoryFunction;
#endif

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle struct memoryPtr       *TempMemoryPtr;
   globle struct memoryPtr       *MemoryTable[MEM_TABLE_SIZE];
   globle int                     TempSize;
  
/***************************************************/
/* genalloc: A generic memory allocation function. */
/***************************************************/
globle VOID *genalloc(size)
  unsigned int size;
  {
   char *memPtr;   
   
#if   BLOCK_MEMORY  
   memPtr = RequestChunk(size);
   if (memPtr == NULL)
     { 
      ReleaseMemory((long) ((size * 5 > 4096) ? size * 5 : 4096),CLIPS_FALSE);
      memPtr = RequestChunk(size);
      if (memPtr == NULL) 
        {
         PeriodicCleanup(CLIPS_FALSE,CLIPS_FALSE);
         ReleaseMemory(-1L,CLIPS_TRUE);
         memPtr = RequestChunk(size);
         while (memPtr == NULL)
           {
            if ((*OutOfMemoryFunction)(size)) ExitCLIPS(1);
            memPtr = RequestChunk(size);
           }
        }
     }
#else
   memPtr = malloc(size);
   if (memPtr == NULL)
     { 
      ReleaseMemory((long) ((size * 5 > 4096) ? size * 5 : 4096),CLIPS_FALSE);
      memPtr = malloc(size);
      if (memPtr == NULL) 
        {
         PeriodicCleanup(CLIPS_FALSE,CLIPS_FALSE);
         ReleaseMemory(-1L,CLIPS_TRUE);
         memPtr = malloc(size);
         while (memPtr == NULL)
           {
            if ((*OutOfMemoryFunction)(size)) ExitCLIPS(1);
            memPtr = malloc(size);
           }
        }
     }
#endif

   MemoryAmount += size;
   MemoryCalls++;

   return((VOID *) memPtr);
  }

/****************************************************/
/* DefaultOutOfMemoryFunction: Function called when */
/*   CLIPS runs out of memory.                      */
/****************************************************/
globle int DefaultOutOfMemoryFunction(size)
  int size;
  {
#if MAC_MPW
#pragma unused(size)
#endif
   PrintCLIPS(WERROR,"ERROR: Out of memory\n");
   return(CLIPS_TRUE);
  }
  
/**********************************************************/
/* SetOutOfMemoryFunction:                                */
/**********************************************************/
globle int (*SetOutOfMemoryFunction(functionPtr))()
  int (*functionPtr)();
  {
   int (*tmpPtr)();
   
   tmpPtr = OutOfMemoryFunction;
   OutOfMemoryFunction = functionPtr;
   return(tmpPtr);
  }

/****************************************************/
/* genfree: A generic memory deallocation function. */
/****************************************************/
globle int genfree(waste,size)
  VOID *waste;
  int size;
  {
   
#if BLOCK_MEMORY
   if (ReturnChunk(waste,size) == -1)
     { 
      PrintCLIPS(WERROR,"Release error in genfree\n");
      return(-1);
     }
#else
   free(waste);
#endif

   MemoryAmount -= size;
   MemoryCalls--;

   return(0);
  }

/*********************************************************/
/* genrealloc: Simple (i.e. dumb) version of realloc. */
/* Should be reimplemented later.                        */
/*********************************************************/
globle VOID *genrealloc(oldaddr,oldsz,newsz)
  VOID *oldaddr;
  unsigned oldsz, newsz;
  {
   char *newaddr;
   int i, limit;

   newaddr = (char *) gm2(newsz);
   
   if (oldaddr != NULL)
     {
      limit = (oldsz < newsz) ? oldsz : newsz;
      for (i = 0 ; i < limit ; i++)
        { newaddr[i] = ((char *) oldaddr)[i]; }
      rm(oldaddr,oldsz);
     }
     
   return((VOID *) newaddr);
  }
  
/*****************************************/
/* genmemcpy: Generic version of memcpy. */
/*****************************************/
globle VOID genmemcpy(dst,src,size)
  VOID *dst, *src;
  unsigned long size;
  {
   unsigned long i;
   
   for (i = 0; i < size ; i++)
     { ((char *) dst)[i] = ((char *) src)[i]; }
  }

/************************************************/
/* genlongalloc: Allocates blocks of memory for */
/*   sizes expressed using long integers.       */
/************************************************/
globle VOID *genlongalloc(size)
  unsigned long size;
  {
#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   unsigned int test;
#endif

   if (sizeof(int) == sizeof(long))
     { return(genalloc((int) size)); }

#if MAC_TC4 || MAC_TC5 || MAC_MPW
   MemoryAmount += size;
   MemoryCalls++;
   return((VOID *) malloc(size));
#endif  

#if IBM_TBC
   MemoryAmount += size;
   MemoryCalls++;
   return((VOID *) farmalloc(size));
#endif

#if IBM_MSC
   MemoryAmount += size;
   MemoryCalls++;
   return((VOID *) halloc(size,1));
#endif
  
#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   test = (unsigned int) size;
   if (test != size) 
     {
      PrintCLIPS(WERROR,"ERROR: Unable to allocate memory block > 32K\n");
      ExitCLIPS(1);
     }
   
   return((VOID *) genalloc(test));
#endif
  }
  
/*********************************************/
/* genlongfree: Returns blocks of memory for */
/*   sizes expressed using long integers.    */
/*********************************************/
globle int genlongfree(ptr,size)
  VOID *ptr;
  unsigned long size;
  {
#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   unsigned int test;
#endif

   if (sizeof(unsigned int) == sizeof(unsigned long))
     { return(genfree(ptr,(int) size)); }

#if MAC_TC4 || MAC_TC5 || MAC_MPW
   MemoryAmount -= size;
   MemoryCalls--;
   free(ptr);
   return(0);
#endif  
  
#if IBM_TBC
   MemoryAmount -= size;
   MemoryCalls--;
   farfree(ptr);
   return(0);
#endif

#if IBM_MSC
   MemoryAmount -= size;
   MemoryCalls--;
   return(hfree(ptr));
#endif

#if (! MAC_TC4) && (! MAC_TC5) && (! MAC_MPW) && (! IBM_TBC) && (! IBM_MSC)
   test = (unsigned int) size;
   if (test != size) return(-1);
   
   return(genfree(ptr,test));
#endif
  }
  
/*******************************************/
/* MemoryUsed:  Returns the amount of memory */
/*   currently allocated by CLIPS.         */
/*******************************************/
globle long int MemoryUsed()
  {
   return(MemoryAmount);
  }

/***********************************************************/
/* MemoryRequests:  Returns the number of outstanding memory */
/*   memory calls made through memory functions.           */
/***********************************************************/
globle long int MemoryRequests()
  {
   return(MemoryCalls);
  }
  
/******************************************/
/* UpdateMemoryUsed:                      */
/******************************************/
globle long int UpdateMemoryUsed(value)
  long int value;
  {
   MemoryAmount += value;
   return(MemoryAmount);
  }
  
/******************************************/
/* UpdateMemoryRequests:                  */
/******************************************/
globle long int UpdateMemoryRequests(value)
  long int value;
  {
   MemoryCalls += value;
   return(MemoryCalls);
  }

/*****************************************************************/
/* ReleaseMemory:  Returns all the lists of stored data structures */
/*   to the pool of free memory.                                 */
/*****************************************************************/
globle long int ReleaseMemory(maximum,printMessage)
  long int maximum;
  int printMessage;
  {
   struct memoryPtr *tmpPtr, *memPtr;
   int i;
   long int amount = 0;
   
   if (printMessage == CLIPS_TRUE)
     { PrintCLIPS(WDIALOG,"\n*** DEALLOCATING MEMORY ***\n"); }
   
   for (i = (MEM_TABLE_SIZE - 1) ; i >= sizeof(char *) ; i--)
     {
      memPtr = MemoryTable[i];
      while (memPtr != NULL)
        {
         tmpPtr = memPtr->next;
         genfree(memPtr,i);
         memPtr = tmpPtr;
         amount += i;
        }
      MemoryTable[i] = NULL;
      if ((amount > maximum) && (maximum > 0)) 
        {
         if (printMessage == CLIPS_TRUE)
           { PrintCLIPS(WDIALOG,"*** MEMORY  DEALLOCATED ***\n"); }
         return(amount);
        }
     }
   
   if (printMessage == CLIPS_TRUE)
     { PrintCLIPS(WDIALOG,"*** MEMORY  DEALLOCATED ***\n"); }
     
   return(amount);
  }
  
/*****************************************************/
/* gm1: Allocates memory and sets all bytes to zero. */
/*****************************************************/
globle VOID *gm1(size)
  int size;
  {
   struct memoryPtr *memPtr;
   char *tmpPtr;
   int i;
   
   if (size < sizeof(char *)) size = sizeof(char *);
   
   if (size >= MEM_TABLE_SIZE) return(genalloc(size));
   
   memPtr = (struct memoryPtr *) MemoryTable[size];
   if (memPtr == NULL) 
     {
      tmpPtr = (char *) genalloc(size); 
      for (i = 0 ; i < size ; i++) 
        { tmpPtr[i] = '\0'; }
      return((VOID *) tmpPtr);
     }
   
   MemoryTable[size] = memPtr->next;
   
   tmpPtr = (char *) memPtr;
   for (i = 0 ; i < size ; i++) 
     { tmpPtr[i] = '\0'; }
   
   return ((VOID *) tmpPtr);
  }

/*****************************************************/
/* gm2: Allocates memory and does not initialize it. */
/*****************************************************/
globle VOID *gm2(size)
  int size;
  {
   struct memoryPtr *memPtr;
   
   if (size < sizeof(char *)) size = sizeof(char *);
   
   if (size >= MEM_TABLE_SIZE) return(genalloc(size));
   
   memPtr = (struct memoryPtr *) MemoryTable[size];
   if (memPtr == NULL) 
     {
      return(genalloc(size)); 
     }
   
   MemoryTable[size] = memPtr->next;

   return ((VOID *) memPtr);
  }
  
/*****************************************************/
/* gm3: Allocates memory and does not initialize it. */
/*****************************************************/
globle VOID *gm3(size)
  long size;
  {
   struct memoryPtr *memPtr;
   
   if (size < sizeof(char *)) size = sizeof(char *);
   
   if (size >= MEM_TABLE_SIZE) return(genlongalloc(size));
   
   memPtr = (struct memoryPtr *) MemoryTable[(int) size];
   if (memPtr == NULL) 
     {
      return(genalloc((unsigned int) size)); 
     }
   
   MemoryTable[(int) size] = memPtr->next;

   return ((VOID *) memPtr);
  }
  
/*****************************************************************/
/* rm: Returns memory.                                          */
/*****************************************************************/
globle int rm(str,size)
  VOID *str;
  int size;
  {
   struct memoryPtr *memPtr;
   
   if (size == 0)
     {   
      CLIPSSystemError("MEMORY",1);
      ExitCLIPS(3);
     }
     
   if (size < sizeof(char *)) size = sizeof(char *);
   
   if (size >= MEM_TABLE_SIZE) return(genfree(str,size));
   
   memPtr = (struct memoryPtr *) str;
   memPtr->next = MemoryTable[size];
   MemoryTable[size] = memPtr;
   return(1);
  }
  
/*****************************************************************/
/* rm3: Returns memory.                                          */
/*****************************************************************/
globle int rm3(str,size)
  VOID *str;
  long size;
  {
   struct memoryPtr *memPtr;
   
   if (size == 0)
     {   
      CLIPSSystemError("MEMORY",1);
      ExitCLIPS(3);
     }
     
   if (size < sizeof(char *)) size = sizeof(char *);
   
   if (size >= MEM_TABLE_SIZE) return(genlongfree(str,size));
   
   memPtr = (struct memoryPtr *) str;
   memPtr->next = MemoryTable[(int) size];
   MemoryTable[(int) size] = memPtr;
   return(1);
  }

/*********************************************************/
/* PoolSize: Returns number of bytes in CLIPS free pool. */
/*********************************************************/
globle unsigned long PoolSize()
  {
   register int i;
   struct memoryPtr *memPtr;
   unsigned long cnt = 0;

   for (i = sizeof(char *) ; i < MEM_TABLE_SIZE ; i++)
     {
      memPtr = MemoryTable[i];
      while (memPtr != NULL)
        {
         cnt += (unsigned long) i;
         memPtr = memPtr->next;
        }
     }
   return(cnt);
  }

/***************************************************************/
/* ActualPoolSize : Returns number of bytes DOS requires to    */
/*   store the CLIPS free pool.  This routine is functionally  */
/*   equivalent to pool_size on anything other than the IBM-PC */
/***************************************************************/
globle unsigned long ActualPoolSize()
  {
#if IBM_TBC
   register int i;
   struct memoryPtr *memPtr;
   unsigned long cnt = 0;

   for (i = sizeof(char *) ; i < MEM_TABLE_SIZE ; i++)
     {
      memPtr = MemoryTable[i];
      while (memPtr != NULL)
        {
         /*==============================================================*/
         /* For a block of size n, the Turbo-C Library routines require  */
         /* a header of size 8 bytes and further require that all memory */
         /* allotments be paragraph (16-bytes) aligned.                  */
         /*==============================================================*/
         
         cnt += (((unsigned long) i) + 19L) & 0xfffffff0L;
         memPtr = memPtr->next;
        }
     }
   return(cnt);
#else
   return(PoolSize());
#endif
  }

/********************************************/
/* SetConserveMemory: Allows the setting of */
/*    the memory conservation flag.         */       
/********************************************/
globle BOOLEAN SetConserveMemory(value)
  BOOLEAN value;
  {
   int ov;
   
   ov = ConserveMemory;
   ConserveMemory = value;
   return(ov);
  }
  
/*******************************************/
/* GetConserveMemory: Returns the value of */
/*    the memory conservation flag.        */       
/*******************************************/
globle BOOLEAN GetConserveMemory()
  {
   return(ConserveMemory);
  }

/*****************************************************************/
/* BLOCK MEMORY FUNCTIONS:                                       */
/*****************************************************************/

#if BLOCK_MEMORY

static struct blockInfo  *TopMemoryBlock;
static int                BlockInfoSize;
static int                ChunkInfoSize;
static int                BlockMemoryInitialized = CLIPS_FALSE;

/**************************/
/* InitializeBlockMemory: */
/**************************/
static int InitializeBlockMemory(requestSize)
  unsigned int requestSize;
  {
   struct chunkInfo *chunkPtr;
   unsigned int initialBlockSize, usableBlockSize;
  
   if (sizeof(char) != 1)
     {
      printf("Size of character data is not 1\n");
      printf("Memory allocation functions may not work\n");
      return(0);
     }
    
   ChunkInfoSize = sizeof(struct chunkInfo);
   ChunkInfoSize = (((ChunkInfoSize - 1) / 4) + 1) * 4;
   
   BlockInfoSize = sizeof(struct blockInfo);
   BlockInfoSize = (((BlockInfoSize - 1) / 4) + 1) * 4;
   
   initialBlockSize = (INITBLOCKSIZE > requestSize ? INITBLOCKSIZE : requestSize);
   initialBlockSize += ChunkInfoSize * 2 + BlockInfoSize;
   initialBlockSize = (((initialBlockSize - 1) / 4) + 1) * 4;
   
   usableBlockSize = initialBlockSize - (2 * ChunkInfoSize) - BlockInfoSize;

   /* make sure we get a buffer big enough to be usable */
   if ((requestSize < INITBLOCKSIZE) &&
       (usableBlockSize <= requestSize + ChunkInfoSize))
     {
      initialBlockSize = requestSize + ChunkInfoSize * 2 + BlockInfoSize;
      initialBlockSize = (((initialBlockSize - 1) / 4) + 1) * 4;
      usableBlockSize = initialBlockSize - (2 * ChunkInfoSize) - BlockInfoSize;
     }
     
   TopMemoryBlock = (struct blockInfo *) malloc(initialBlockSize);

   if (TopMemoryBlock == NULL)
     {
      printf("Unable to allocate initial memory pool\n");
      return(0);
     } 

   TopMemoryBlock->nextBlock = NULL;
   TopMemoryBlock->prevBlock = NULL;
   TopMemoryBlock->nextFree = (struct chunkInfo *) (((char *) TopMemoryBlock) + BlockInfoSize);
   TopMemoryBlock->size = usableBlockSize;

   chunkPtr = (struct chunkInfo *) (((char *) TopMemoryBlock) + BlockInfoSize + ChunkInfoSize + usableBlockSize);
   chunkPtr->nextFree = NULL;   
   chunkPtr->lastFree = NULL;
   chunkPtr->prevChunk = TopMemoryBlock->nextFree;
   chunkPtr->size = 0;

   TopMemoryBlock->nextFree->nextFree = NULL;
   TopMemoryBlock->nextFree->lastFree = NULL;
   TopMemoryBlock->nextFree->prevChunk = NULL;
   TopMemoryBlock->nextFree->size = usableBlockSize;

   BlockMemoryInitialized = CLIPS_TRUE;
   return(1);
  }

/********************************/
/* AllocateBlock:           */
/********************************/
static int AllocateBlock(blockPtr,requestSize)
  struct blockInfo *blockPtr;
  unsigned int requestSize;
  {
   unsigned int blockSize, usableBlockSize;
   struct blockInfo *newBlock;
   struct chunkInfo *newTopChunk;
   
   blockSize = (BLOCKSIZE > requestSize ? BLOCKSIZE : requestSize);
   blockSize += BlockInfoSize + ChunkInfoSize * 2;
   blockSize = (((blockSize - 1) / 4) + 1) * 4;
   
   usableBlockSize = blockSize - BlockInfoSize - (2 * ChunkInfoSize);
   
   /* make sure we get a buffer big enough to be usable */
   if ((requestSize < BLOCKSIZE) &&
       (usableBlockSize <= requestSize + ChunkInfoSize))
     {
      blockSize = requestSize + ChunkInfoSize * 2 + BlockInfoSize;
      blockSize = (((blockSize - 1) / 4) + 1) * 4;
      usableBlockSize = blockSize - (2 * ChunkInfoSize) - BlockInfoSize;
     }

   newBlock = (struct blockInfo *) malloc(blockSize);

   if (newBlock == NULL) return(0);

   newBlock->nextBlock = NULL;
   newBlock->prevBlock = blockPtr;
   newBlock->nextFree = (struct chunkInfo *) (((char *) newBlock) + BlockInfoSize);
   newBlock->size = usableBlockSize;
   blockPtr->nextBlock = newBlock;
   
   newTopChunk = (struct chunkInfo *) (((char *) newBlock) + BlockInfoSize + ChunkInfoSize + usableBlockSize);
   newTopChunk->nextFree = NULL;
   newTopChunk->lastFree = NULL;
   newTopChunk->size = 0;
   newTopChunk->prevChunk = newBlock->nextFree;
   
   newBlock->nextFree->nextFree = NULL;
   newBlock->nextFree->lastFree = NULL;
   newBlock->nextFree->prevChunk = NULL;
   newBlock->nextFree->size = usableBlockSize;
 
   return(1);
  }
  
/********************************/
/* RequestChunk:                */
/********************************/
globle VOID *RequestChunk(requestSize)  
  unsigned int requestSize;
  {
   struct chunkInfo *chunkPtr;
   struct blockInfo *blockPtr;

   /*==================================================*/
   /* Allocate initial memory pool block if it has not */
   /* already been allocated.                          */
   /*==================================================*/
   
   if (BlockMemoryInitialized == CLIPS_FALSE) 
      {
       if (InitializeBlockMemory(requestSize) == 0) return(NULL);
      }
 
   /*====================================================*/
   /* Make sure that the amount of memory requested will */
   /* fall on a word boundary.  Assume a word size of 4. */
   /*====================================================*/
   
   requestSize = (((requestSize - 1) / 4) + 1) * 4;

   /*=====================================================*/
   /* Search through the list of free memory for a block  */
   /* of the appropriate size.  If a block is found, then */
   /* allocate and return a pointer to it.                */
   /*=====================================================*/

   blockPtr = TopMemoryBlock;
   
   while (blockPtr != NULL)
     {
      chunkPtr = blockPtr->nextFree;

      while (chunkPtr != NULL)
        {
         if ((chunkPtr->size == requestSize) ||
             (chunkPtr->size > (requestSize + ChunkInfoSize)))
           {
            AllocateChunk(blockPtr,chunkPtr,requestSize);
   
            return((VOID *) (((char *) chunkPtr) + ChunkInfoSize));
           }
         chunkPtr = chunkPtr->nextFree;
        }
      
      if (blockPtr->nextBlock == NULL)
        {
         if (AllocateBlock(blockPtr,requestSize) == 0)  /* get another block */
           { return(NULL); }
        }
      blockPtr = blockPtr->nextBlock;
     }
     
   CLIPSSystemError("MEMORY",2);
   ExitCLIPS(1);
   return(NULL); /* Unreachable, but prevents warning. */
  } 

/********************************************/
/* AllocateChunk: Allocates a chunk from an */
/*   existing chunk in a block of memory.   */
/********************************************/
static VOID AllocateChunk(parentBlock,chunkPtr,requestSize)
  struct blockInfo *parentBlock;
  struct chunkInfo *chunkPtr;
  unsigned int requestSize;
  {
   struct chunkInfo *splitChunk, *nextChunk;

   /*=============================================================*/
   /* If the size of the memory chunk is an exact match for the   */
   /* requested amount of memory, then the chunk can be allocated */
   /* without splitting it.                                       */
   /*=============================================================*/
   
   if (requestSize == chunkPtr->size)
     {
      chunkPtr->size = - (long int) requestSize;
      if (chunkPtr->lastFree == NULL)
        {
         if (chunkPtr->nextFree != NULL)
           { parentBlock->nextFree = chunkPtr->nextFree; }
         else
           { parentBlock->nextFree = NULL; }
        }
      else
        { chunkPtr->lastFree->nextFree = chunkPtr->nextFree; }

      if (chunkPtr->nextFree != NULL)
        { chunkPtr->nextFree->lastFree = chunkPtr->lastFree; }

      chunkPtr->lastFree = NULL;
      chunkPtr->nextFree = NULL;
      return;
     }

   /*===========================================================*/
   /* If the size of the memory chunk is larger than the memory */
   /* request, then split the chunk into two pieces.            */
   /*===========================================================*/

   nextChunk = (struct chunkInfo *)
              (((char *) chunkPtr) + ChunkInfoSize + chunkPtr->size);

   splitChunk = (struct chunkInfo *) 
                  (((char *) chunkPtr) + (ChunkInfoSize + requestSize));

   splitChunk->size = chunkPtr->size - (requestSize + ChunkInfoSize);
   splitChunk->prevChunk = chunkPtr;

   splitChunk->nextFree = chunkPtr->nextFree;
   splitChunk->lastFree = chunkPtr->lastFree;

   nextChunk->prevChunk = splitChunk;
   
   if (splitChunk->lastFree == NULL)
     { parentBlock->nextFree = splitChunk; }
   else
     { splitChunk->lastFree->nextFree = splitChunk; }

   if (splitChunk->nextFree != NULL)
     { splitChunk->nextFree->lastFree = splitChunk; }

   chunkPtr->size = - (long int) requestSize;
   chunkPtr->lastFree = NULL;
   chunkPtr->nextFree = NULL;
   
   return;
  }

/*************************/
/* ReturnChunk:          */
/*************************/
globle int ReturnChunk(memPtr,size)
  VOID *memPtr;
  unsigned int size;
  {
   struct chunkInfo *chunkPtr, *lastChunk, *nextChunk, *topChunk;
   struct blockInfo *blockPtr;

   /*=====================================================*/
   /* Determine if the expected size of the chunk matches */
   /* the size stored in the chunk's information record.  */
   /*=====================================================*/
         
   size = (((size - 1) / 4) + 1) * 4;

   chunkPtr = (struct chunkInfo *) (((char *) memPtr) - ChunkInfoSize);

   if (chunkPtr == NULL) 
     { return(-1); }

   if (chunkPtr->size >= 0) 
     { return(-1); }
   
   if (chunkPtr->size != - (long int) size) 
     { return(-1); }

   chunkPtr->size = - chunkPtr->size;  
   
   /*=============================================*/
   /* Determine in which block the chunk resides. */
   /*=============================================*/
   
   topChunk = chunkPtr;
   while (topChunk->prevChunk != NULL)
     { topChunk = topChunk->prevChunk; }
   blockPtr = (struct blockInfo *) (((char *) topChunk) - BlockInfoSize);

   /*===========================================*/
   /* Determine the chunks physically preceding */
   /* and following the returned chunk.         */
   /*===========================================*/
   
   lastChunk = chunkPtr->prevChunk;
   nextChunk = (struct chunkInfo *) (((char *) memPtr) + size);

   /*=========================================================*/
   /* Add the chunk to the list of free chunks for the block. */
   /*=========================================================*/
   
   if (blockPtr->nextFree != NULL) 
     { blockPtr->nextFree->lastFree = chunkPtr; } 

   chunkPtr->nextFree = blockPtr->nextFree;
   chunkPtr->lastFree = NULL;
   
   blockPtr->nextFree = chunkPtr;

   /*=====================================================*/
   /* Combine this chunk with previous chunk if possible. */
   /*=====================================================*/

   if (lastChunk != NULL)
     {
      if (lastChunk->size > 0)
        {
         lastChunk->size += (ChunkInfoSize + chunkPtr->size);

         if (nextChunk != NULL)
           { nextChunk->prevChunk = lastChunk; }
         else 
           { return(-1); }

         if (lastChunk->lastFree != NULL)
           { lastChunk->lastFree->nextFree = lastChunk->nextFree; }

         if (lastChunk->nextFree != NULL)
           { lastChunk->nextFree->lastFree = lastChunk->lastFree; }

         lastChunk->nextFree = chunkPtr->nextFree;
         if (chunkPtr->nextFree != NULL)
           { chunkPtr->nextFree->lastFree = lastChunk; }
         lastChunk->lastFree = NULL;
         
         blockPtr->nextFree = lastChunk;
         chunkPtr->lastFree = NULL;
         chunkPtr->nextFree = NULL;
         chunkPtr = lastChunk;
        }
     }

   /*=====================================================*/
   /* Combine this chunk with the next chunk if possible. */
   /*=====================================================*/

   if (nextChunk == NULL) return(-1);
   if (chunkPtr == NULL) return(-1);

   if (nextChunk->size > 0)
     {
      chunkPtr->size += (ChunkInfoSize + nextChunk->size);
      
      topChunk = (struct chunkInfo *) (((char *) nextChunk) + nextChunk->size + ChunkInfoSize);
      if (topChunk != NULL)
        { topChunk->prevChunk = chunkPtr; }
      else 
        { return(-1); }

      if (nextChunk->lastFree != NULL)
        { nextChunk->lastFree->nextFree = nextChunk->nextFree; }

      if (nextChunk->nextFree != NULL)
        { nextChunk->nextFree->lastFree = nextChunk->lastFree; }
        
     }
   
   /* Free the buffer if we can, but */
   /* don't free the first buffer    */
   /* if it's the only one.          */
   
   if ((chunkPtr->prevChunk == NULL) &&
       (chunkPtr->size == blockPtr->size))
     {
      if (blockPtr->prevBlock != NULL)
        { 
         blockPtr->prevBlock->nextBlock = blockPtr->nextBlock;
         if (blockPtr->nextBlock != NULL)
           { blockPtr->nextBlock->prevBlock = blockPtr->prevBlock; }
         free((char *) blockPtr);
        }
      else
        {
         if (blockPtr->nextBlock != NULL)
           {
            blockPtr->nextBlock->prevBlock = NULL;
            TopMemoryBlock = blockPtr->nextBlock;
            free((char *) blockPtr);
           }
        }
     }     

   return(1);
  }

#endif
