   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Message-passing support functions                */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#ifndef _H_msgfun
#define _H_msgfun

#include "object.h"

typedef struct msg_handler_link
  {
   unsigned implicit : 1;
   unsigned get      : 1;
   unsigned type     : 2;
   VOID *hnd;
   struct msg_handler_link *nxt;
  } HANDLER_LINK;

/* =================================================================================
   Message-handler types - don't change these values: a string array depends on them
   ================================================================================= */
#define MAROUND        0
#define MBEFORE        1
#define MPRIMARY       2
#define MAFTER         3
#define MERROR         4

#define LOOKUP_HANDLER_INDEX   0
#define LOOKUP_HANDLER_ADDRESS 1

#define GSM_PREFIX    "get-"
#define PSM_PREFIX    "put-"

#define GSMP_LEN       4
#define PSMP_LEN       4

#ifdef LOCALE
#undef LOCALE
#endif

#ifdef _MSGFUN_SOURCE_
#define LOCALE
#else
#define LOCALE extern
#endif

#if ANSI_COMPILER
LOCALE VOID DirectMessage(char *,SYMBOL_HN *,INSTANCE_TYPE *,
                          DATA_OBJECT *,EXPRESSION *);
LOCALE VOID CLIPSSendMessage(DATA_OBJECT *,char *,char *,DATA_OBJECT *);
VOID DestroyHandlerLinks(HANDLER_LINK *);

#if DEBUGGING_FUNCTIONS
LOCALE VOID DisplayCore(HANDLER_LINK *,int);
LOCALE HANDLER_LINK *FindPreviewApplicableHandlers(CLASS_TYPE *,SYMBOL_HN *);
#endif

#if (! BLOAD_ONLY) && (! RUN_TIME)
LOCALE HANDLER *NewHandler(void);
LOCALE int HandlersExecuting(CLASS_TYPE *);
LOCALE int DeleteHandler(CLASS_TYPE *,SYMBOL_HN *,int,int);
LOCALE VOID DeallocateMarkedHandlers(CLASS_TYPE *);
LOCALE VOID PrintAbbreviatedHandlerRemoval(HANDLER *);
#endif
LOCALE int HandlerType(char *,char *);
LOCALE int CheckCurrentMessage(char *,int);
LOCALE VOID PrintCurrentMessage(char *);
LOCALE VOID PrintHandler(char *,char *,char *,char *);
LOCALE VOID *FindHandler(CLASS_TYPE *,SYMBOL_HN *,unsigned,int);
LOCALE int FindHandlerNameGroup(CLASS_TYPE *,SYMBOL_HN *);
LOCALE VOID HandlerDeleteError(char *);
LOCALE VOID DispatchMessage(DATA_OBJECT *);
LOCALE int NextHandlerAvailable(void);
LOCALE VOID CallNextHandler(DATA_OBJECT *);

#else

LOCALE VOID DirectMessage();
LOCALE VOID CLIPSSendMessage();
LOCALE VOID DestroyHandlerLinks();

#if DEBUGGING_FUNCTIONS
LOCALE VOID DisplayCore();
LOCALE HANDLER_LINK *FindPreviewApplicableHandlers();
#endif

#if (! BLOAD_ONLY) && (! RUN_TIME)
LOCALE HANDLER *NewHandler();
LOCALE int HandlersExecuting();
LOCALE int DeleteHandler();
LOCALE VOID DeallocateMarkedHandlers();
LOCALE VOID PrintAbbreviatedHandlerRemoval();
#endif
LOCALE int HandlerType();
LOCALE int CheckCurrentMessage();
LOCALE VOID PrintCurrentMessage();
LOCALE VOID PrintHandler();
LOCALE VOID *FindHandler();
LOCALE int FindHandlerNameGroup();
LOCALE VOID HandlerDeleteError();
LOCALE VOID DispatchMessage();
LOCALE int NextHandlerAvailable();
LOCALE VOID CallNextHandler();
#endif
      
#ifndef _MSGFUN_SOURCE_
extern char *hndquals[];

#if DEBUGGING_FUNCTIONS
extern int WatchHandlers,WatchMessages;
#endif
extern DATA_OBJECT *CurrentMessageFrame;
extern int CurrentMessageSize;
extern SYMBOL_HN *CurrentMessageName;
#endif

#endif







