   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*               CLIPS Version 5.10  07/17/91          */
   /*                                                     */
   /*                                                     */
   /*******************************************************/

/*************************************************************/
/* Purpose: Binary Load/Save Functions for Classes and their */
/*             message-handlers                              */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Brian L. Donnell                                     */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
   
/* =========================================
   *****************************************
               EXTERNAL DEFINITIONS
   =========================================
   ***************************************** */
#include "setup.h"

#if OBJECT_SYSTEM && (BLOAD || BLOAD_ONLY || BLOAD_AND_BSAVE)

#include "constant.h"
#include "clipsmem.h"
#include "bload.h"
#include "bsave.h"

#define _OBJBIN_SOURCE_
#include "objbin.h"

#include "defins.h"
#include "classfun.h"

#if ! BLOAD_ONLY
#include "msgcom.h"
#endif

/* =========================================
   *****************************************
                   CONSTANTS
   =========================================
   ***************************************** */
   
/* =========================================
   *****************************************
               MACROS AND TYPES
   =========================================
   ***************************************** */
typedef unsigned long UNLN;

#define ClassIndex(p) ((CLASS_TYPE *) (((p) != NULL) ? (p)->bsaveIndex : -1L))
#define SlotIndex(p) ((SLOT_DESC *) (((p) != NULL) ? ((long) (p)->cls) : -1L))
#define DefinstancesIndex(p) ((DEFINSTANCES *) (((p) != NULL) ? (p)->bsaveIndex : -1L))

#define DefinstancesPointer(i) ((((long) (i)) == -1L) ? NULL : \
                                (DEFINSTANCES *) &definstancesArray[(long) (i)])
#define LinkPointer(i) ((((long) (i)) == -1L) ? NULL : \
                        (CLASS_LINK *) &linkArray[(long) (i)])
#define SlotPointer(i) ((((long) (i)) == -1L) ? NULL : \
                        (SLOT_DESC *) &slotArray[(long) (i)])
#define TemplateSlotPointer(i) ((((long) (i)) == -1L) ? NULL : \
                                (SLOT_DESC **) &tmpslotArray[(long) (i)])
#define OrderedSlotPointer(i) ((((long) (i)) == -1L) ? NULL : \
                                (unsigned *) &mapslotArray[(long) (i)])
#define HandlerPointer(i) ((((long) (i)) == -1L) ? NULL : \
                           (HANDLER *) &handlerArray[(long) (i)])
#define OrderedHandlerPointer(i) ((((long) (i)) == -1L) ? NULL : \
                                  (unsigned *) &maphandlerArray[(long) (i)])

/* =========================================
   *****************************************
      INTERNALLY VISIBLE FUNCTION HEADERS
   =========================================
   ***************************************** */
#if ANSI_COMPILER

#if BLOAD_AND_BSAVE

static VOID BsaveObjectsFind(void);
static VOID BsaveObjectsExpressions(FILE *);
static VOID BsaveObjects(FILE *);
#endif

static int BloadObjectsReady(void);
#if ! BLOAD_ONLY
static VOID BloadObjectsAbort(void);
#endif
static VOID BloadObjects(void);
static VOID RefreshObjects(void);
static VOID ClearBloadObjects(void);

#else

#if BLOAD_AND_BSAVE

static VOID BsaveObjectsFind();
static VOID BsaveObjectsExpressions();
static VOID BsaveObjects();

#endif

static int BloadObjectsReady();
#if ! BLOAD_ONLY
static VOID BloadObjectsAbort();
#endif
static VOID BloadObjects();
static VOID RefreshObjects();
static VOID ClearBloadObjects();

#endif
      
/* =========================================
   *****************************************
      EXTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */
globle CLASS_TYPE HUGE_ADDR *defclassArray = NULL;

/* =========================================
   *****************************************
      INTERNALLY VISIBLE GLOBAL VARIABLES
   =========================================
   ***************************************** */

static long ClassCount = 0L,
            LinkCount = 0L,
            SlotCount = 0L,
            TemplateSlotCount = 0L,
            HandlerCount = 0L,
            DefinstancesCount = 0L;

static CLASS_LINK HUGE_ADDR *linkArray = NULL;
static SLOT_DESC HUGE_ADDR *slotArray = NULL;
static SLOT_DESC * HUGE_ADDR *tmpslotArray = NULL;
static unsigned HUGE_ADDR *mapslotArray = NULL;
static HANDLER HUGE_ADDR *handlerArray = NULL;
static unsigned HUGE_ADDR *maphandlerArray = NULL;
static DEFINSTANCES HUGE_ADDR *definstancesArray = NULL;
static char HUGE_ADDR *tvidsArray = NULL;

/* =========================================
   *****************************************
          EXTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */
   
/***********************************************************
  NAME         : SetupObjectsBload
  DESCRIPTION  : Initializes data structures and
                   routines for binary loads of
                   generic function constructs
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Routines defined and structures initialized
  NOTES        : None
 ***********************************************************/
globle VOID SetupObjectsBload()
  {
#if ! BLOAD_ONLY
   AddBeforeBloadFunction("object system",ObjectSystemPurge,-750);
   AddAbortBloadFunction("object system",BloadObjectsAbort,0);
#endif
   AddBloadReadyFunction("object system",BloadObjectsReady,0);
   AddClearBloadReadyFunction("object system",BloadObjectsReady,0);
#if BLOAD_AND_BSAVE
   AddBinaryItem("object system",0,BsaveObjectsFind,BsaveObjectsExpressions,
                             BsaveObjects,BloadObjects,RefreshObjects,
                             ClearBloadObjects);
#endif
#if BLOAD || BLOAD_ONLY
   AddBinaryItem("object system",0,NULL,NULL,NULL,BloadObjects,RefreshObjects,
                             ClearBloadObjects);
#endif
  }
   
/* =========================================
   *****************************************
          INTERNALLY VISIBLE FUNCTIONS
   =========================================
   ***************************************** */

#if BLOAD_AND_BSAVE

/***************************************************************************
  NAME         : BsaveObjectsFind
  DESCRIPTION  : For all classes and their message-handlers, this routine
                   marks all the needed symbols and system functions.
                 Also, it also counts the number of expression structures
                   needed.
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : ExpressionCount (a global from BSAVE.C) is incremented
                   for every expression needed
                 Symbols are marked in their structures
  NOTES        : Also sets bsaveIndex for each class (assumes classes
                   will be bsaved in order of binary list)
 ***************************************************************************/
static VOID BsaveObjectsFind()
  {
   CLASS_TYPE *cls;
   CLASS_LINK *sptr;
   register unsigned i;
   DEFINSTANCES *dfiptr;

   ClassCount = 0L;
   LinkCount = 0L;
   SlotCount = 0L;
   TemplateSlotCount = 0L;
   HandlerCount = 0L;
   DefinstancesCount = 0L;
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      cls->bsaveIndex = ClassCount;
      cls->name->neededSymbol = CLIPS_TRUE;
      for (sptr = cls->superlink ; sptr != NULL ; sptr = sptr->nxt)
        LinkCount++;
      for (sptr = cls->sublink ; sptr != NULL ; sptr = sptr->nxt)
        LinkCount++;
      for (sptr = cls->precedence ; sptr != NULL ; sptr = sptr->nxt)
        LinkCount++;
      for (i = 0 ; i < cls->slot_cnt ; i++)
        {
         cls->slots[i].name->neededSymbol = CLIPS_TRUE;
         cls->slots[i].gethandler->neededSymbol = CLIPS_TRUE;
         cls->slots[i].puthandler->neededSymbol = CLIPS_TRUE;
         ExpressionCount += ExpressionSize(cls->slots[i].defexp);
         MarkNeededItems(cls->slots[i].defexp);
        }
      SlotCount += (long) cls->slot_cnt;
      TemplateSlotCount += (long) cls->islot_cnt;
      for (i = 0 ; i < cls->handler_cnt ; i++)
        {
         cls->handlers[i].name->neededSymbol = CLIPS_TRUE;
         ExpressionCount += ExpressionSize(cls->handlers[i].actions);
         MarkNeededItems(cls->handlers[i].actions);
        }
      HandlerCount += (long) cls->handler_cnt;
      ClassCount++;
     }
   for (dfiptr = DefinstancesList ; dfiptr != NULL ; dfiptr = dfiptr->nxt)
     {
      dfiptr->bsaveIndex = DefinstancesCount;
      dfiptr->name->neededSymbol = CLIPS_TRUE;
      ExpressionCount += ExpressionSize(dfiptr->mkinstance);
      MarkNeededItems(dfiptr->mkinstance);
      DefinstancesCount++;
     }
  }

/***************************************************
  NAME         : BsaveObjectsExpressions
  DESCRIPTION  : Writes out all expressions needed
                   by classes and handlers
  INPUTS       : The file pointer of the binary file
  RETURNS      : Nothing useful
  SIDE EFFECTS : File updated
  NOTES        : None
 ***************************************************/
static VOID BsaveObjectsExpressions(fp)
  FILE *fp;
  {
   CLASS_TYPE *cls;
   register unsigned i;
   DEFINSTANCES *dfiptr;
   
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->slot_cnt ; i++)
        BsaveExpression(cls->slots[i].defexp,fp);
     }
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->handler_cnt ; i++)
        BsaveExpression(cls->handlers[i].actions,fp);
     }
   for (dfiptr = DefinstancesList ; dfiptr != NULL ; dfiptr = dfiptr->nxt)
     BsaveExpression(dfiptr->mkinstance,fp);
  }

/*************************************************************************************
  NAME         : BsaveObjects
  DESCRIPTION  : Writes out classes and message-handlers in binary format
                 Space required (unsigned long)
                 Number of classes (long)
                 Number of links to classes (long)
                 Number of slots (long)
                 Number of instance template slots (long)
                 Number of handlers (long)
                 Followed by the data structures in order
  INPUTS       : File pointer of binary file
  RETURNS      : Nothing useful
  SIDE EFFECTS : Binary file adjusted
  NOTES        : None
 *************************************************************************************/
static VOID BsaveObjects(fp)
  FILE *fp;
  {
   UNLN space;
   CLASS_TYPE *cls,dummy_class;
   CLASS_LINK *sptr,dummy_link;
   register unsigned i;
   SLOT_DESC *sp,dummy_slot,*tsp;
   HANDLER *hnd,dummy_handler;
   DEFINSTANCES dummy_dfi,*dfiptr;

   space = ((UNLN) ClassCount * (UNLN) sizeof(CLASS_TYPE)) +
           ((UNLN) LinkCount * (UNLN) sizeof(CLASS_LINK)) +
           ((UNLN) SlotCount * (UNLN) sizeof(SLOT_DESC)) +
           ((UNLN) TemplateSlotCount * (UNLN) sizeof(SLOT_DESC *)) +
           ((UNLN) TemplateSlotCount * (UNLN) sizeof(unsigned)) +
           ((UNLN) HandlerCount * (UNLN) sizeof(HANDLER)) +
           ((UNLN) HandlerCount * (UNLN) sizeof(unsigned)) +
           ((UNLN) DefinstancesCount * (UNLN) sizeof(DEFINSTANCES));
   if (space == 0L)
     {
      GenWrite((VOID *) &space,(UNLN) sizeof(UNLN),fp);
      return;
     }
   space += ((UNLN) 6 * (UNLN) sizeof(long));
   GenWrite((VOID *) &space,(UNLN) sizeof(UNLN),fp);
   GenWrite((VOID *) &ClassCount,(UNLN) sizeof(long),fp);
   GenWrite((VOID *) &LinkCount,(UNLN) sizeof(long),fp);
   GenWrite((VOID *) &SlotCount,(UNLN) sizeof(long),fp);
   GenWrite((VOID *) &TemplateSlotCount,(UNLN) sizeof(long),fp);
   GenWrite((VOID *) &HandlerCount,(UNLN) sizeof(long),fp);
   GenWrite((VOID *) &DefinstancesCount,(UNLN) sizeof(long),fp);
   
   LinkCount = 0L;
   SlotCount = 0L;
   TemplateSlotCount = 0L;
   HandlerCount = 0L;
   
   /* =====================
      Write out the classes 
      ===================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      CopyMemory(CLASS_TYPE,1,&dummy_class,cls);
      dummy_class.busy = 0;
      dummy_class.predecessors = 0;
      dummy_class.name = (SYMBOL_HN *) cls->name->bucket;
      if (cls->superlink != NULL)
        {
         dummy_class.superlink = (CLASS_LINK *) LinkCount;
         for (sptr = cls->superlink ; sptr != NULL ; sptr = sptr->nxt)
           LinkCount++;
        }
      else
        dummy_class.superlink = (CLASS_LINK *) -1L;
      if (cls->sublink != NULL)
        {
         dummy_class.sublink = (CLASS_LINK *) LinkCount;
         for (sptr = cls->sublink ; sptr != NULL ; sptr = sptr->nxt)
           LinkCount++;
        }
      else
        dummy_class.sublink = (CLASS_LINK *) -1L;
      if (cls->precedence != NULL)
        {
         dummy_class.precedence = (CLASS_LINK *) LinkCount;
         for (sptr = cls->precedence ; sptr != NULL ; sptr = sptr->nxt)
           LinkCount++;
        }
      else
        dummy_class.precedence = (CLASS_LINK *) -1L;
      if (cls->slots != NULL)
        {
         dummy_class.slots = (SLOT_DESC *) SlotCount;
         SlotCount += cls->slot_cnt;
        }
      else
        dummy_class.slots = (SLOT_DESC *) -1L;
      if (cls->itemplate != NULL)
        {
         dummy_class.itemplate = (SLOT_DESC **) TemplateSlotCount;
         dummy_class.sordered = (unsigned *) TemplateSlotCount;
         TemplateSlotCount += cls->islot_cnt;
        }
      else
        {
         dummy_class.itemplate = (SLOT_DESC **) -1L;
         dummy_class.sordered = (unsigned *) -1L;
        }
      dummy_class.instances = (INSTANCE_TYPE *) -1L;
      if (cls->handlers != NULL)
        {
         dummy_class.handlers = (HANDLER *) HandlerCount;
         dummy_class.hordered = (unsigned *) HandlerCount;
         HandlerCount += cls->handler_cnt;
        }
      else
        {
         dummy_class.handlers = (HANDLER *) -1L;
         dummy_class.hordered = (unsigned *) -1L;
        }
      dummy_class.prv_hsh = (CLASS_TYPE *) -1L;
      dummy_class.nxt_hsh = (CLASS_TYPE *) -1L;
      dummy_class.prv_lst = ClassIndex(cls->prv_lst);
      dummy_class.nxt_lst = ClassIndex(cls->nxt_lst);
      dummy_class.pp_form = (char *) -1L;
      dummy_class.tvids = (char *) -1L;
      GenWrite((VOID *) &dummy_class,(UNLN) sizeof(CLASS_TYPE),fp);
     }

   /* =========================
      Write out the class links
      ========================= */
   LinkCount = 0L;
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (sptr = cls->superlink ; sptr != NULL ; sptr = sptr->nxt)
        {
         dummy_link.cls = ClassIndex(sptr->cls);
         if (sptr->nxt != NULL)
           dummy_link.nxt = (CLASS_LINK *) (LinkCount+1L);
         else
           dummy_link.nxt = (CLASS_LINK *) -1L;
         LinkCount++;
         GenWrite((VOID *) &dummy_link,(UNLN) sizeof(CLASS_LINK),fp);
        }
      for (sptr = cls->sublink ; sptr != NULL ; sptr = sptr->nxt)
        {
         dummy_link.cls = ClassIndex(sptr->cls);
         if (sptr->nxt != NULL)
           dummy_link.nxt = (CLASS_LINK *) (LinkCount+1L);
         else
           dummy_link.nxt = (CLASS_LINK *) -1L;
         LinkCount++;
         GenWrite((VOID *) &dummy_link,(UNLN) sizeof(CLASS_LINK),fp);
        }
      for (sptr = cls->precedence ; sptr != NULL ; sptr = sptr->nxt)
        {
         dummy_link.cls = ClassIndex(sptr->cls);
         if (sptr->nxt != NULL)
           dummy_link.nxt = (CLASS_LINK *) (LinkCount+1L);
         else
           dummy_link.nxt = (CLASS_LINK *) -1L;
         LinkCount++;
         GenWrite((VOID *) &dummy_link,(UNLN) sizeof(CLASS_LINK),fp);
        }
     }

   /* ===================
      Write out the slots
      =================== */
   SlotCount = 0L;
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->slot_cnt ; i++)
        {
         sp = &cls->slots[i];
         CopyMemory(SLOT_DESC,1,&dummy_slot,sp);
         dummy_slot.cls = ClassIndex(sp->cls);
         dummy_slot.name = (SYMBOL_HN *) sp->name->bucket;
         dummy_slot.gethandler = (SYMBOL_HN *) sp->gethandler->bucket;
         dummy_slot.puthandler = (SYMBOL_HN *) sp->puthandler->bucket;
         if (sp->defexp != NULL)
           {
            dummy_slot.defexp = (EXPRESSION *) ExpressionCount;
            ExpressionCount += ExpressionSize(sp->defexp);
           }
         else
           dummy_slot.defexp = (EXPRESSION *) -1L;
         dummy_slot.override = 0;
         dummy_slot.share_cnt = 0;
         dummy_slot.share_val = (DATA_OBJECT *) -1L;
         GenWrite((VOID *) &dummy_slot,(UNLN) sizeof(SLOT_DESC),fp);

        /* ========================================================
           Temporarily store the slot indices in the class field so
           the template instance slots can have a reference
           ======================================================== */
         sp->cls = (CLASS_TYPE *) SlotCount;
         SlotCount++;
        }
     }

   /* =====================================
      Write out the template instance slots
      ===================================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->islot_cnt ; i++)
        {
         tsp = SlotIndex(cls->itemplate[i]);
         GenWrite((VOID *) &tsp,(UNLN) sizeof(SLOT_DESC *),fp);
        }
     }

   /* ========================================
      Write out the ordered instance slot maps
      ======================================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->islot_cnt ; i++)
        GenWrite((VOID *) &cls->sordered[i],(UNLN) sizeof(unsigned),fp);
     }

   /* ==============================================================
      Write out the message-handlers and restore slot class pointers
      ============================================================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->slot_cnt ; i++)
        cls->slots[i].cls = cls;
      for (i = 0 ; i < cls->handler_cnt ; i++)
        {
         hnd = &cls->handlers[i];
         CopyMemory(HANDLER,1,&dummy_handler,hnd);
         dummy_handler.mark = 0;
         dummy_handler.busy = 0;
         dummy_handler.name = (SYMBOL_HN *) hnd->name->bucket;
         dummy_handler.cls = ClassIndex(hnd->cls);
         if (hnd->actions != NULL)
           {
            dummy_handler.actions = (EXPRESSION *) ExpressionCount;
            ExpressionCount += ExpressionSize(hnd->actions);
           }
         else
           dummy_handler.actions = (EXPRESSION *) -1L;
         dummy_handler.pp_form = (char *) -1L;
         GenWrite((VOID *) &dummy_handler,(UNLN) sizeof(HANDLER),fp);
        }
     }
   
   /* ==========================================
      Write out the ordered message-handler maps
      ========================================== */
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     {
      for (i = 0 ; i < cls->handler_cnt ; i++)
        GenWrite((VOID *) &cls->hordered[i],(UNLN) sizeof(unsigned),fp);
     }

   /* ==========================
      Write out the definstances
      ========================== */
   for (dfiptr = DefinstancesList ; dfiptr != NULL ; dfiptr = dfiptr->nxt)
     {
      dummy_dfi.busy = 0;
      dummy_dfi.name = (SYMBOL_HN *) dfiptr->name->bucket;
      if (dfiptr->mkinstance != NULL)
        {
         dummy_dfi.mkinstance = (EXPRESSION *) ExpressionCount;
         ExpressionCount += ExpressionSize(dfiptr->mkinstance);
        }
      else
        dummy_dfi.mkinstance = (EXPRESSION *) -1L;
      dummy_dfi.pp_form = (char *) -1L;
      dummy_dfi.prv = DefinstancesIndex(dfiptr->prv);
      dummy_dfi.nxt = DefinstancesIndex(dfiptr->nxt);
      GenWrite((VOID *) &dummy_dfi,(UNLN) sizeof(DEFINSTANCES),fp);
     }
   ClassCount = DefinstancesCount = 0L;
  }

#endif

/***************************************************
  NAME         : BloadObjectsReady
  DESCRIPTION  : Determines if it's safe to bload
                  a new binary iamge for objects
  INPUTS       : None
  RETURNS      : TRUE if OK, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 ***************************************************/
static int BloadObjectsReady()
  {
   CLASS_TYPE *cls;
   DEFINSTANCES *dptr;
   
   InstancesPurge();
   for (cls = ClassList ; cls != NULL ; cls = cls->nxt_lst)
     if (cls->busy > 0)
        return(FALSE);
   for (dptr = DefinstancesList ; dptr != NULL ; dptr = dptr->nxt)
     if (dptr->busy > 0)
        return(FALSE);
   return(TRUE);
  }

#if ! BLOAD_ONLY

/*****************************************************
  NAME         : BloadObjectsAbort
  DESCRIPTION  : Reinstalls system classes and
                   message-handlers after bload abort
  INPUTS       : None
  RETURNS      : TRUE if OK, FALSE otherwise
  SIDE EFFECTS : None
  NOTES        : None
 *****************************************************/
static VOID BloadObjectsAbort()
  {
   ReinitializeClasses();
   AddSystemHandlers();
  }

#endif

/***********************************************************************
  NAME         : BloadObjects
  DESCRIPTION  : This routine reads class and handler information from
                 a binary file in five chunks:
                 Class count
                 Handler count
                 Class array
                 Handler array
  INPUTS       : Notthing
  RETURNS      : Nothing useful
  SIDE EFFECTS : Arrays allocated and set
  NOTES        : This routine makes no attempt to reset any pointers
                   within the structures
                 Bload fails if there are still classes in the system!!
 ***********************************************************************/
static VOID BloadObjects()
  {
   UNLN space;
   long counts[6];
   
   GenRead((VOID *) &space,(UNLN) sizeof(UNLN));
   if (space == 0L)
     {
      ClassCount = DefinstancesCount = 0L;
      return;
     }
   GenRead((VOID *) counts,(UNLN) (sizeof(long) * 6));
   ClassCount = counts[0];
   LinkCount = counts[1];
   SlotCount = counts[2];
   TemplateSlotCount = counts[3];
   HandlerCount = counts[4];
   DefinstancesCount = counts[5];
   if (ClassCount != 0L)
     {
      space = (UNLN) (sizeof(CLASS_TYPE) * ClassCount);
      defclassArray = (CLASS_TYPE HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) defclassArray,space);
      space = (UNLN) (sizeof(char) * TRAVERSAL_BYTES * ClassCount);
      tvidsArray = (char HUGE_ADDR *) genlongalloc(space);
     }
   if (LinkCount != 0L)
     {
      space = (UNLN) (sizeof(CLASS_LINK) * LinkCount);
      linkArray = (CLASS_LINK HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) linkArray,space);
     } 
   if (SlotCount != 0L)
     {
      space = (UNLN) (sizeof(SLOT_DESC) * SlotCount);
      slotArray = (SLOT_DESC HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) slotArray,space);
     } 
   if (TemplateSlotCount != 0L)
     {
      space = (UNLN) (sizeof(SLOT_DESC *) * TemplateSlotCount);
      tmpslotArray = (SLOT_DESC * HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) tmpslotArray,space);
      space = (UNLN) (sizeof(unsigned) * TemplateSlotCount);
      mapslotArray = (unsigned HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) mapslotArray,space);
     } 
   if (HandlerCount != 0L)
     {
      space = (UNLN) (sizeof(HANDLER) * HandlerCount);
      handlerArray = (HANDLER HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) handlerArray,space);
      space = (UNLN) (sizeof(unsigned) * HandlerCount);
      maphandlerArray = (unsigned HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) maphandlerArray,space);
     } 
   if (DefinstancesCount != 0L)
     {
      space = (UNLN) (sizeof(DEFINSTANCES) * DefinstancesCount);
      definstancesArray = (DEFINSTANCES HUGE_ADDR *) genlongalloc(space);
      GenRead((VOID *) definstancesArray,space);
     } 
  }

/***************************************************************
  NAME         : RefreshObjects
  DESCRIPTION  : This routine moves through the class and handler
                   binary arrays updating pointers
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Pointers reset from array indices
  NOTES        : Assumes all loading is finished
 **************************************************************/
static VOID RefreshObjects()
  {
   register long i;
   CLASS_TYPE *cls;
   SLOT_DESC *sp;
   HANDLER *hnd;
   DEFINSTANCES *dfiptr;
   
   if (ClassCount != 0L)
     {
      for (i = 0L ; i < LinkCount ; i++)
        {
         linkArray[i].cls = ClassPointer(linkArray[i].cls);
         linkArray[i].nxt = LinkPointer(linkArray[i].nxt);
        }  
      for (i = 0L ; i < SlotCount ; i++)
        {
         sp = (SLOT_DESC *) &slotArray[i];
         sp->cls = ClassPointer(sp->cls);
         sp->name = SymbolPointer(sp->name);
         IncrementCount(sp->name);
         sp->gethandler = SymbolPointer(sp->gethandler);
         sp->gethandler->relatedSymbol = sp->name;
         IncrementCount(sp->gethandler);
         sp->puthandler = SymbolPointer(sp->puthandler);
         sp->puthandler->relatedSymbol = sp->name;
         IncrementCount(sp->puthandler);
         sp->defexp = ExpressionPointer(sp->defexp);
         sp->share_val = NULL;
        }
      for (i = 0L ; i < TemplateSlotCount ; i++)
        tmpslotArray[i] = SlotPointer(tmpslotArray[i]);
      for (i = 0L ; i < HandlerCount ; i++)
        {
         hnd = (HANDLER *) &handlerArray[i];
         hnd->name = SymbolPointer(hnd->name);
         IncrementCount(hnd->name);
         hnd->cls = ClassPointer(hnd->cls);
         hnd->actions = ExpressionPointer(hnd->actions);
         hnd->pp_form = NULL;
        }
      PrimitiveClassMap[FLOAT] = (CLASS_TYPE *) &defclassArray[FLOAT];
      PrimitiveClassMap[INTEGER] = (CLASS_TYPE *) &defclassArray[INTEGER];
      PrimitiveClassMap[STRING] = (CLASS_TYPE *) &defclassArray[STRING];
      PrimitiveClassMap[SYMBOL] = (CLASS_TYPE *) &defclassArray[SYMBOL];
      PrimitiveClassMap[MULTIFIELD] = (CLASS_TYPE *) &defclassArray[MULTIFIELD];
      PrimitiveClassMap[EXTERNAL_ADDRESS] = 
                                (CLASS_TYPE *) &defclassArray[EXTERNAL_ADDRESS];
      for (i = 0L ; i < ClassCount ; i++)
        {
         cls = (CLASS_TYPE *) &defclassArray[i];
         cls->name = SymbolPointer(cls->name);
         IncrementCount(cls->name);
         cls->superlink = LinkPointer(cls->superlink);
         cls->sublink = LinkPointer(cls->sublink);
         cls->precedence = LinkPointer(cls->precedence);
         cls->slots = SlotPointer(cls->slots);
         cls->itemplate = TemplateSlotPointer(cls->itemplate);
         cls->sordered = OrderedSlotPointer(cls->sordered);
         cls->instances = NULL;
         cls->handlers = HandlerPointer(cls->handlers);
         cls->hordered = OrderedHandlerPointer(cls->hordered);
         cls->prv_hsh = NULL;
         cls->nxt_hsh = NULL;
         cls->prv_lst = ClassPointer(cls->prv_lst);
         cls->nxt_lst = ClassPointer(cls->nxt_lst);
         cls->pp_form = NULL;
         cls->tvids = (char *) (tvidsArray + (i * TRAVERSAL_BYTES));
         PutClassInTable(cls);
        }
      SetClassList((CLASS_TYPE *) &defclassArray[0]);
     }
   if (DefinstancesCount != 0L)
     {
      for (i = 0L ; i < DefinstancesCount ; i++)
        {
         dfiptr = (DEFINSTANCES *) &definstancesArray[i];
         dfiptr->name = SymbolPointer(dfiptr->name);
         IncrementCount(dfiptr->name);
         dfiptr->mkinstance = ExpressionPointer(dfiptr->mkinstance);
         dfiptr->pp_form = NULL;
         dfiptr->prv = DefinstancesPointer(dfiptr->prv);
         dfiptr->nxt = DefinstancesPointer(dfiptr->nxt);
        }
      SetDefinstancesList((DEFINSTANCES *) &definstancesArray[0]);
     }
  }

/***************************************************************
  NAME         : ClearBloadObjects
  DESCRIPTION  : Release all binary-loaded class and handler
                   structure arrays (and others)
                 Resets systemclasses and handlers
  INPUTS       : None
  RETURNS      : Nothing useful
  SIDE EFFECTS : Memory cleared
  NOTES        : None
 ***************************************************************/
static VOID ClearBloadObjects()
  {
   register long i;
   UNLN space;
   
   if (ClassCount != 0L)
     {
      for (i = 0L ; i < ClassCount ; i++)
        DecrementSymbolCount(defclassArray[i].name);
      for (i = 0L ; i < SlotCount ; i++)
        {
         DecrementSymbolCount(slotArray[i].name);
         DecrementSymbolCount(slotArray[i].gethandler);
         DecrementSymbolCount(slotArray[i].puthandler);
        }
      for (i = 0L ; i < HandlerCount ; i++)
        DecrementSymbolCount(handlerArray[i].name);
     
      space = (UNLN) (sizeof(char) * TRAVERSAL_BYTES * ClassCount);
      if (space != 0L)
        {
         genlongfree((VOID *) tvidsArray,space);
         tvidsArray = NULL;
        }

      space = (UNLN) (sizeof(CLASS_TYPE) * ClassCount);
      if (space != 0L)
        {
         genlongfree((VOID *) defclassArray,space);
         defclassArray = NULL;
         ClassCount = 0L;
        }
   
      space = (UNLN) (sizeof(char) * TRAVERSAL_BYTES * ClassCount);
      if (space != 0L)
        {
         genlongfree((VOID *) tvidsArray,space);
         tvidsArray = NULL;
        }
   
      space = (UNLN) (sizeof(CLASS_LINK) * LinkCount);
      if (space != 0L)
        {
         genlongfree((VOID *) linkArray,space);
         linkArray = NULL;
         LinkCount = 0L;
        }
   
      space = (UNLN) (sizeof(SLOT_DESC) * SlotCount);
      if (space != 0L)
        {
         genlongfree((VOID *) slotArray,space);
         slotArray = NULL;
         SlotCount = 0L;
        }
   
      space = (UNLN) (sizeof(SLOT_DESC *) * TemplateSlotCount);
      if (space != 0L)
        {
         genlongfree((VOID *) tmpslotArray,space);
         tmpslotArray = NULL;
         space = (UNLN) (sizeof(unsigned) * TemplateSlotCount);
         genlongfree((VOID *) mapslotArray,space);
         mapslotArray = NULL;
         TemplateSlotCount = 0L;
        }
   
      space = (UNLN) (sizeof(HANDLER) * HandlerCount);
      if (space != 0L)
        {
         genlongfree((VOID *) handlerArray,space);
         handlerArray = NULL;
         space = (UNLN) (sizeof(unsigned) * HandlerCount);
         genlongfree((VOID *) maphandlerArray,space);
         maphandlerArray = NULL;
         HandlerCount = 0L;
        }
     }
   if (DefinstancesCount != 0L)
     {
      for (i = 0L ; i < DefinstancesCount ; i++)
        DecrementSymbolCount(definstancesArray[i].name);
      space = (UNLN) (sizeof(DEFINSTANCES) * DefinstancesCount);
      genlongfree((VOID *) definstancesArray,space);
      definstancesArray = NULL;
      DefinstancesCount = 0L;
     }
#if BLOAD_ONLY
   SetClassList(NULL);
#else
   ReinitializeClasses();
   AddSystemHandlers();
#endif
   SetDefinstancesList(NULL);
  }

#endif

/***************************************************
  NAME         : 
  DESCRIPTION  : 
  INPUTS       : 
  RETURNS      : 
  SIDE EFFECTS : 
  NOTES        : 
 ***************************************************/
