   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                    REORDER MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/
  
#define _REORDER_SOURCE_

#include "setup.h"

#if (! RUN_TIME) && (! BLOAD_ONLY) && DEFRULE_CONSTRUCT
 
#include <stdio.h>
#define _CLIPS_STDIO_

#include "constant.h"
#include "clipsmem.h"
#include "expressn.h"
#include "router.h"
#include "reorder.h"
         
#if ANSI_COMPILER  
   static struct node            *ReverseOR(struct node *,struct node *,int);
   static struct node            *AdjacentReduction(struct node *);
#else
   static struct node            *ReverseOR();
   static struct node            *AdjacentReduction();
#endif

/**************************************************/
/* ReorderPatterns                               */
/**************************************************/
globle struct node *ReorderPatterns(ao_list)
  struct node *ao_list;
  {
   struct node *args;
   struct node *before, *save;
   int count;
   
   ao_list = AdjacentReduction(ao_list);

   if (ao_list->type == PAT_AND)
     {
      count = 1;
      args = ao_list->right;
      while (args != NULL)
        {
         if (args->type == PAT_OR)
           {
            ao_list = ReverseOR(ao_list,args->right,count);
            args = NULL;
           }
         else
           {
            count++;
            args = args->bottom;
           }
        }
     }

   ao_list = AdjacentReduction(ao_list);

   before = NULL;
   args = ao_list->right;

   while (args != NULL)
     {
      save = args->bottom;
      if ((args->type == PAT_AND) || 
          (args->type == PAT_OR))
        {
         if (before == NULL)
           {
            args->bottom = NULL;
            ao_list->right = ReorderPatterns(args);
            ao_list->right->bottom = save;
            before = ao_list->right;
           }
         else
           {
            args->bottom = NULL;
            before->bottom = ReorderPatterns(args);
            before->bottom->bottom = save;
            before = before->bottom;
           }
        }
      args = save;
     }

   ao_list = AdjacentReduction(ao_list);

   return(ao_list);
  }

/**************************************************/
/* ReverseOR: Switches and/or constructs into     */
/*   or/and constructs.                           */ 
/*   For example:                                 */
/*     (and (or a b) (or c d))                    */
/*   would be converted to                        */
/*     (or (and a (or c d)) (and b (or c d))),    */
/*   if the or pointer were pointing at (or a b). */
/**************************************************/
static struct node *ReverseOR(ao_list,or_ptr,or_num)
  struct node *ao_list;
  struct node *or_ptr;
  int or_num;
  {
   int count;
   struct node *my_list = NULL;
   struct node *last_add = NULL;
   struct node *new_list, *replace_slot;

   while (or_ptr != NULL)
     {
      new_list = CopyNodes(ao_list);
      count = 1;
      replace_slot = new_list->right;
      while (count != or_num)
        {
         replace_slot = replace_slot->bottom;
         count++;
        }
      ReturnNodes(replace_slot->right);
      replace_slot->type = or_ptr->type;
      replace_slot->negated = or_ptr->negated;
      replace_slot->logical = or_ptr->logical;
      replace_slot->value = or_ptr->value;
      replace_slot->right = CopyNodes(or_ptr->right);
      replace_slot->expression = CopyExpression(or_ptr->expression);

      if (last_add == NULL)
        { 
         my_list = new_list;
         new_list->bottom = NULL;
         last_add = new_list;
        }
      else
        {
         last_add->bottom = new_list;
         new_list->bottom = NULL;
         last_add = new_list;
        }

      or_ptr = or_ptr->bottom;
     }

   ReturnNodes(ao_list); 
   
   new_list = get_struct(node);
   new_list->type = PAT_OR;
   new_list->value = NULL;
   new_list->expression = NULL;
   new_list->bottom = NULL;
   new_list->right = my_list;
   
   return(new_list);
  }
 
/***********************************************************/
/* AdjacentReduction:  Given the represention of the lhs   */
/*   of a rule, this routine remove redundant information. */
/*   Logicial Pattern And's within and's and or's within   */
/*   or's are simplified.                                  */
/*                                                         */
/*   For example:                                          */
/*     (or (and (and (a) (b)) (and (c) (d))))              */
/*   would be converted to                                 */
/*     (or (and (a) (b)) (and (c) (d)))                    */
/*                                                         */
/*   The and's and or's to be simplified must be adjacent. */
/*   For example, the following lhs would not be           */
/*   simplified by this routine:                           */
/*     (or (and (or (a))))                                 */
/***********************************************************/
static struct node *AdjacentReduction(ao_list)
  struct node *ao_list;
  {
   struct node *node_ptr, *arg_ptr;
   struct node *last_arg, *next_arg, *temp_arg;

   if (ao_list == NULL) 
     {  return(NULL); }

   node_ptr = ao_list;
   while (node_ptr != NULL)
     {
      if ((node_ptr->type == PAT_AND) ||
          (node_ptr->type == PAT_OR))
        { node_ptr->right = AdjacentReduction(node_ptr->right); } 
      node_ptr = node_ptr->bottom;
     }

   node_ptr = ao_list;
   while (node_ptr != NULL)
     {

      arg_ptr = node_ptr->right;
      last_arg = NULL;

      while (arg_ptr != NULL)
        {
         next_arg = arg_ptr->bottom;

         if ((arg_ptr->type == node_ptr->type) && 
             ((node_ptr->type == PAT_AND) || 
              (node_ptr->type == PAT_OR)))
           {  
            temp_arg = arg_ptr->right;
            arg_ptr->right = NULL;
            arg_ptr->bottom = NULL;
            ReturnNodes(arg_ptr);

            if (last_arg == NULL)
              { node_ptr->right = temp_arg; }
            else
              { last_arg->bottom = temp_arg; }

            while (temp_arg->bottom != NULL)
              { temp_arg = temp_arg->bottom; }
            
            temp_arg->bottom = next_arg;
            last_arg = temp_arg;
           }
         else
           { last_arg = arg_ptr; }

         arg_ptr = next_arg;
        }

      node_ptr = node_ptr->bottom;
     }

   return(ao_list);
  }

/**************************************************/
/* CopyNodes                                     */
/**************************************************/
globle struct node *CopyNodes(node_list)
  struct node *node_list;
  {
   struct node *head;

   if (node_list == NULL)
     { return(NULL); }

   head = get_struct(node);
   head->type = node_list->type;
   head->negated = node_list->negated;
   head->logical = node_list->logical;
   head->value = node_list->value;
   head->expression = CopyExpression(node_list->expression);
   head->right = CopyNodes(node_list->right);
   head->bottom = CopyNodes(node_list->bottom);

   return(head);
  }
  
/***********************************************************/
/* GetNode                                                */
/***********************************************************/
globle struct node *GetNode()
  {
   struct node *new_node;
   
   new_node = get_struct(node);
   new_node->type = KUNKNOWN;
   new_node->negated = CLIPS_FALSE;
   new_node->logical = CLIPS_FALSE;
   new_node->value = NULL;
   new_node->expression = NULL;
   new_node->right = NULL;
   new_node->bottom = NULL;
   
   return(new_node);
  }
   
/*******************************************************************/
/* ReturnNodes:  Returns a multiply linked list of node structures */
/*   to the list of free nodes.                                    */
/*******************************************************************/
globle VOID ReturnNodes(waste)
  struct node *waste;
  {
   if (waste != NULL)
     {
      ReturnNodes(waste->right);
      ReturnNodes(waste->bottom);
      if (waste->expression != NULL) ReturnExpression(waste->expression); 
      rtn_struct(node,waste);
     } 
  }
  
#endif

