   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*                  LHS PARSE MODULE                   */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _RULEPRSR_SOURCE_

#include "setup.h"

#if (! RUN_TIME) && (! BLOAD_ONLY) &&  DEFRULE_CONSTRUCT

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "constant.h"
#include "symbol.h"
#include "clipsmem.h"
#include "expressn.h"
#include "evaluatn.h"
#include "scanner.h"
#include "reorder.h"
#include "ruleprsr.h"
#include "router.h"
#include "utility.h"
#include "defrule.h"
#include "constrct.h"
#include "parsutil.h"

#if DEFTEMPLATE_CONSTRUCT
#include "deftempl.h"
#include "deftmlhs.h"
#endif

#ifdef _RULEPRSR_SOURCE_
#if ANSI_COMPILER  
   static struct node            *RuleBodyParse(char *,struct token *);
   static VOID                    DeclarationParse(char *);
   static struct node            *LHSPattern(char *,int);
   static struct node            *ConnectedPatternParse(char *,struct token *);
   static struct node            *GroupPatterns(char *,int,char *);
   static struct node            *NotPatternParse(char *);
   static struct node            *TestPattern(char *);
   static struct node            *AssignmentParse(char *,SYMBOL_HN *);
   static struct node            *SimplePatternParse(char *,struct token *);
   static struct node            *SequenceRestrictionParse(char *,struct token *);
   static struct node            *ConjuctiveRestrictionParse(char *,struct token *);
   static struct node            *LiteralRestrictionParse(char *,struct token *);
   static struct node            *CreateInitialPattern(void);
   static VOID                    TagLHSLogicalNodes(struct node *);
#else
   static struct node            *RuleBodyParse();
   static VOID                    DeclarationParse();
   static struct node            *LHSPattern();
   static struct node            *ConnectedPatternParse();
   static struct node            *GroupPatterns();
   static struct node            *NotPatternParse();
   static struct node            *TestPattern();
   static struct node            *AssignmentParse();
   static struct node            *SimplePatternParse();
   static struct node            *SequenceRestrictionParse();
   static struct node            *ConjuctiveRestrictionParse();
   static struct node            *LiteralRestrictionParse();
   static struct node            *CreateInitialPattern();
   static VOID                    TagLHSLogicalNodes();
#endif
#endif

/****************************/
/* LOCAL INTERNAL VARIABLES */
/****************************/

   static int                LHSError = CLIPS_FALSE;

/****************************************/
/* GLOBAL INTERNAL VARIABLE DEFINITIONS */
/****************************************/

   globle int                GlobalSalience;
   globle struct expr       *SalienceExpression = NULL;

/*******************************************************************/
/* ParseRuleLHS: Coordinates all the actions necessary for parsing */
/*   the LHS of a rule including the reordering of pattern         */
/*   conditional elements to conform with the CLIPS Rete topology. */
/*******************************************************************/
globle struct node *ParseRuleLHS(readSource,inp_tkn)
  char *readSource;
  struct token *inp_tkn;
  {
   struct node *logicalrepresentation, *convert_rep;

   /*==========================================================*/
   /* Create the logical representation for the left hand side */
   /* of the rule.                                             */
   /*==========================================================*/

   GlobalSalience = 0;

   SetIndentDepth(3);        
   logicalrepresentation = RuleBodyParse(readSource,inp_tkn);

   if (logicalrepresentation == NULL) return(NULL);

   /*======================================================*/
   /* Convert logical representation to new representation */
   /* with only a single top level or.                     */
   /*======================================================*/

   convert_rep = GetNode();
   convert_rep->type = PAT_AND;
   convert_rep->right = CopyNodes(logicalrepresentation);

   convert_rep = ReorderPatterns(convert_rep);

   ReturnNodes(logicalrepresentation);

   return(convert_rep);
  }
  
/*********************************************************/
/* RuleBodyParse: Parses the LHS of a rule, but does not */
/*   reorder any of the LHS patterns to conform with the */
/*   CLIPS Rete Topology.                                */
/*                                                       */
/* <rule-body> ::= [<declaration>]                       */
/*                 <conditional-element>*                */
/*                 =>                                    */
/*********************************************************/
static struct node *RuleBodyParse(readSource,lhs_tkn)
  char *readSource;
  struct token *lhs_tkn;
  {
   struct node *data_spec, *other_stuff;
   
   LHSError = CLIPS_FALSE;
   
   /*============================================*/
   /* Parse the first pattern as a special case. */
   /*============================================*/
   
   if ((lhs_tkn->type == SYMBOL) ? (strcmp(ValueToString(lhs_tkn->value),"=>") == 0) : CLIPS_FALSE)
     { return(CreateInitialPattern()); }
   else if (lhs_tkn->type == LPAREN)
     { 
      GetToken(readSource,lhs_tkn);
      if (lhs_tkn->type == SYMBOL)
        {
         if (strcmp(ValueToString(lhs_tkn->value),"declare") == 0)
           { 
            DeclarationParse(readSource);
            data_spec = NULL;
           }
         else if (strcmp(ValueToString(lhs_tkn->value),"test") == 0)
           { data_spec = TestPattern(readSource); }
         else if (strcmp(ValueToString(lhs_tkn->value),"not") == 0)
           { data_spec = NotPatternParse(readSource); }
         else if ((strcmp(ValueToString(lhs_tkn->value),"and") == 0) || 
                  (strcmp(ValueToString(lhs_tkn->value),"logical") == 0) || 
                  (strcmp(ValueToString(lhs_tkn->value),"or") == 0))
           { data_spec = ConnectedPatternParse(readSource,lhs_tkn); }         
         else
           { data_spec = SimplePatternParse(readSource,lhs_tkn); }
        }
      else
        { data_spec = SimplePatternParse(readSource,lhs_tkn); }
     }
   else if (lhs_tkn->type == BWORD)
     { data_spec = AssignmentParse(readSource,lhs_tkn->value); }
   else 
     { 
      SyntaxErrorMessage("defrule");
      LHSError = CLIPS_TRUE;
      return(NULL);
     }
     
   if (LHSError == CLIPS_TRUE)
     {
      ReturnNodes(data_spec);
      return(NULL);
     }
   
   PPCRAndIndent();
   
   /*======================================*/
   /* Parse the other patterns in the LHS. */
   /*======================================*/
   
   other_stuff = GroupPatterns(readSource,SYMBOL,"=>");
   
   if (LHSError == CLIPS_TRUE)
     {
      ReturnNodes(data_spec);
      return(NULL);
     }
     
   if (data_spec == NULL)
     { data_spec = other_stuff; }
   else
     { data_spec->bottom = other_stuff; }
     
   if (data_spec == NULL) return(CreateInitialPattern());
   
   return(data_spec);
  }

/********************************************************/
/* DeclarationParse: Parses a defrule declaration. Only */
/*   salience declarations are currently allowed.       */
/*                                                      */
/* <declaration> ::= (declare <rule-property>+)         */
/*                                                      */
/* <rule-property> ::= (salience <integer-expression>)  */
/********************************************************/
static VOID DeclarationParse(readSource)
  char *readSource;
  {  
   struct token sal_tkn; 
   int salience;
   DATA_OBJECT salienceValue;
   struct expr *packPtr;

   /*===========================*/
   /* Next token must be a '('. */
   /*===========================*/

   SavePPBuffer(" ");   

   GetToken(readSource,&sal_tkn);                     
   if (sal_tkn.type != LPAREN)
     {  
      SyntaxErrorMessage("declare statement");
      LHSError = CLIPS_TRUE;
      return;
     }
   
   /*=========================================*/
   /* Next token must be the symbol "salience". */
   /*=========================================*/
   
   GetToken(readSource,&sal_tkn);
   if (strcmp(ValueToString(sal_tkn.value),"salience") != 0)
     {
      PrintCLIPS(WERROR,"\nOnly salience declarations are allowed\n");
      LHSError = CLIPS_TRUE;
      return;
     }

   /*==============================*/
   /* Get the salience expression. */
   /*==============================*/

   SavePPBuffer(" ");     

   SalienceExpression = ParseAtomOrExpression(readSource);
   if (SalienceExpression == NULL)
     {
      LHSError = CLIPS_TRUE;
      return;
     }
     
   /*============================================================*/
   /* Evaluate the expression and determine if it is an integer. */
   /*============================================================*/
   
   SetEvaluationError(CLIPS_FALSE);
   if (EvaluateExpression(SalienceExpression,&salienceValue))
     {
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
      SalienceInformationError(NULL);
      LHSError = CLIPS_TRUE;
      return;
     }
     
   if (salienceValue.type != INTEGER)
     {
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
      PrintCLIPS(WERROR,"\nSalience value must be an integer value.\n");
      LHSError = CLIPS_TRUE;
      return;
     }

   /*=======================================================*/
   /* Salience number must be in the range -10000 to 10000. */
   /*=======================================================*/

   salience = (int) ValueToLong(salienceValue.value);

   if ((salience > MAX_SALIENCE) || (salience < MIN_SALIENCE))
     {  
      PrintCLIPS(WERROR,"\nSalience value out of range ");
      PrintLongInteger(WERROR,(long int) MIN_SALIENCE);
      PrintCLIPS(WERROR," to ");
      PrintLongInteger(WERROR,(long int) MAX_SALIENCE);
      PrintCLIPS(WERROR,"\n");
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
      LHSError = CLIPS_TRUE;
      return;
     }

   /*============================================*/
   /* Salience declaration is closed with a ')'. */
   /*============================================*/   

   GetToken(readSource,&sal_tkn);
   if (sal_tkn.type != RPAREN)
     {
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
      SyntaxErrorMessage("declare statement");
      LHSError = CLIPS_TRUE;
      return;
     }

   /*===================================*/
   /* Declaration is closed with a ')'. */
   /*===================================*/
 
   GetToken(readSource,&sal_tkn);
   if (sal_tkn.type != RPAREN)
     {
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
      SyntaxErrorMessage("declare statement");
      LHSError = CLIPS_TRUE;
      return;
     }

   /*==========================================*/
   /* If the expression is a constant integer, */
   /* don't bother storing the expression.     */
   /*==========================================*/
   
   if (SalienceExpression->type == INTEGER)
     {
      ReturnExpression(SalienceExpression);
      SalienceExpression = NULL;
     }
      
   /*==============================================*/
   /* Return the value of the salience through the */
   /* global variable salience_value.              */
   /*==============================================*/

   packPtr = PackExpression(SalienceExpression);
   ReturnExpression(SalienceExpression);
   SalienceExpression = packPtr; 
   GlobalSalience = salience;
   return;
  }

/*****************************************************************/
/* LHSPattern: Parses a single conditional element found on the  */
/*   LHS of a rule. Conditonal element types include pattern CEs */ 
/*   (which may be assigned to a variable), test CEs, not CEs,   */
/*   logical CEs, and CEs, and or CEs.                           */
/*                                                               */
/* <conditional-element> ::= <pattern-CE> |                      */
/*                           <assigned-pattern-CE> |             */
/*                           <not-CE> | <and-CE> | <or-CE> |     */
/*                           <logical-CE> | <test-CE>            */
/*****************************************************************/
static struct node *LHSPattern(readSource,terminator)
  char *readSource;
  int terminator;
  {
   struct token lhs_tkn;
   struct node *data_spec;
   
   GetToken(readSource,&lhs_tkn);
   if (lhs_tkn.type == LPAREN)
     { 
      GetToken(readSource,&lhs_tkn);
      if (lhs_tkn.type == SYMBOL)
        {
         if (strcmp(ValueToString(lhs_tkn.value),"test") == 0)
           { data_spec = TestPattern(readSource); }
         else if (strcmp(ValueToString(lhs_tkn.value),"not") == 0)
           { data_spec = NotPatternParse(readSource); }
         else if ((strcmp(ValueToString(lhs_tkn.value),"and") == 0) || 
                  (strcmp(ValueToString(lhs_tkn.value),"logical") == 0) || 
                  (strcmp(ValueToString(lhs_tkn.value),"or") == 0))
           { data_spec = ConnectedPatternParse(readSource,&lhs_tkn); }         
         else
           { data_spec = SimplePatternParse(readSource,&lhs_tkn); }
        }
      else
        { data_spec = SimplePatternParse(readSource,&lhs_tkn); }
     }
   else if (lhs_tkn.type == BWORD)
     { data_spec = AssignmentParse(readSource,lhs_tkn.value); }
   else if (lhs_tkn.type == terminator)
     { return(NULL);  }
   else 
     { 
      SyntaxErrorMessage("defrule");
      LHSError = CLIPS_TRUE;
      return(NULL);
     }

   if (LHSError == CLIPS_TRUE) 
     {
      ReturnNodes(data_spec);
      return(NULL);
     }
     
   return(data_spec);
  }
  
/*********************************************************************/
/* ConnectedPatternParse:  Handles parsing of the connected          */
/*   conditional elements (i.e. those conditional elements that may  */
/*   contain one or more other conditional elements).  The connected */
/*   conditional elements include the and CE, the or CE, and the     */
/*   logical CE.  This routine is entered with the parsing pointing  */
/*   to the name of the connected CE. It is exited with the parser   */
/*   pointing to the closing right parenthesis of the connected CE.  */ 
/*                                                                   */
/* <and-CE>      ::= (and <conditional-element>+)                    */
/*                                                                   */
/* <or-CE>       ::= (or <conditional-element>+)                     */
/*                                                                   */
/* <logical-CE>  ::= (logical <conditional-element>+)                */
/*********************************************************************/
static struct node *ConnectedPatternParse(readSource,con_tkn)
  char *readSource;
  struct token *con_tkn;
  {
   int connector_value = 0;
   struct node *data_spec;
   int logical = CLIPS_FALSE;

   /*==========================================================*/
   /* Use appropriate spacing for pretty printing of the rule. */
   /*==========================================================*/

   IncrementIndentDepth(5);
   if (strcmp(ValueToString(con_tkn->value),"or") == 0)  
     { 
      connector_value = PAT_OR;
      SavePPBuffer("  "); 
     }
   else if (strcmp(ValueToString(con_tkn->value),"and") == 0)  
     {
      connector_value = PAT_AND;
      SavePPBuffer(" ");
     }
   else if (strcmp(ValueToString(con_tkn->value),"logical") == 0)  
     {
      logical = CLIPS_TRUE;
      connector_value = PAT_AND;
      PPCRAndIndent();
     }

   /*========================================================*/
   /* Advance the scanner to the first point of interest in  */
   /* pattern that follows the logical connector: Either the */
   /* fact binder for the pattern or the pattern's first     */
   /* element slot.                                          */
   /*========================================================*/

   data_spec = GetNode();
   data_spec->type = connector_value;
   data_spec->logical = logical;
   data_spec->right = GroupPatterns(readSource,RPAREN,")");
   
   DecrementIndentDepth(5);
         
   if (data_spec->right == NULL)
     {
      PrintCLIPS(WERROR,"\nNo patterns within and/or/logical conditional element\n");
      ReturnNodes(data_spec);
      LHSError = CLIPS_TRUE;
      return(NULL);
     }
     
   if (logical) TagLHSLogicalNodes(data_spec->right);
   return(data_spec);
  }
  
/***********************************************************/
/* GroupPatterns: Groups a series of connected conditional */
/*   elements together.                                    */
/*                                                         */
/* <and-CE>      ::= (and <conditional-element>+)          */
/*                                                         */
/* <or-CE>       ::= (or <conditional-element>+)           */
/*                                                         */
/* <logical-CE>  ::= (logical <conditional-element>+)      */
/***********************************************************/
static struct node *GroupPatterns(readSource,terminator,term_str)
  int terminator;
  char *readSource, *term_str;
  {
   struct node *last_node, *new_node, *data_spec;
  
   last_node = data_spec = NULL;
   
   while (CLIPS_TRUE)
     {
      new_node = LHSPattern(readSource,terminator);
      
      if (LHSError == CLIPS_TRUE) 
        {
         ReturnNodes(data_spec);
         return(NULL);
        }
        
      if (new_node == NULL)
        {
         PPBackup();
         PPBackup();
         if (terminator == RPAREN)
           { SavePPBuffer(term_str); }
         else
           {
            PPCRAndIndent();
            SavePPBuffer(term_str);
           }
           
         return(data_spec);
        }
        
      if (last_node == NULL)
        { data_spec = new_node; }
      else
        { last_node->bottom = new_node; }
        
      last_node = new_node;
      PPCRAndIndent();
     }
  }
  
/*****************************************************************/
/* NotPatternParse: Handles parsing of not conditional elements. */
/*                                                               */
/* <not-CE> ::= (not <pattern-CE>)                               */
/*****************************************************************/
static struct node *NotPatternParse(readSource)
  char *readSource;
  {
   struct token not_tkn;
   struct node *data_spec;

   /*==========================================================*/
   /* Use appropriate spacing for pretty printing of the rule. */
   /*==========================================================*/

   SavePPBuffer(" ");

   /*=======================================================*/
   /* Get the opening left parenthesis for the not pattern. */
   /*=======================================================*/
   
   GetToken(readSource,&not_tkn);
   if (not_tkn.type == BWORD)
     { 
      PrintCLIPS(WERROR,"\nAttempt to bind a not'ed pattern to ");
      PrintCLIPS(WERROR,"a fact-address\n"); 
      LHSError = CLIPS_TRUE;
      return(NULL);
     }
   else if (not_tkn.type != LPAREN)
     { 
      SyntaxErrorMessage("defrule");
      LHSError = CLIPS_TRUE;
      return(NULL);
     }
     
   /*========================================================*/
   /* Get the template name for the not pattern. Ensure that */
   /* a logical AND or OR was not used.                      */
   /*========================================================*/
   
   GetToken(readSource,&not_tkn);
   if ((not_tkn.type == SYMBOL) &&
       ((strcmp(ValueToString(not_tkn.value),"and") == 0) ||
        (strcmp(ValueToString(not_tkn.value),"logical") == 0) || 
        (strcmp(ValueToString(not_tkn.value),"or") == 0) ||
        (strcmp(ValueToString(not_tkn.value),"test") == 0) ||
        (strcmp(ValueToString(not_tkn.value),"not") == 0)))
     { 
      PrintCLIPS(WERROR,"\nOnly a pattern may be used within a not conditional element\n");
      LHSError = CLIPS_TRUE;
      return(NULL);
     }

   /*======================================================*/
   /* Parse the patterns within the logical not construct. */
   /*======================================================*/
   
   data_spec = SimplePatternParse(readSource,&not_tkn);
   
   if (LHSError == CLIPS_TRUE)
     {
      ReturnNodes(data_spec); 
      return(NULL);
     }
     
   data_spec->negated = CLIPS_TRUE;
   data_spec->value = NULL;
  
   /*================================*/
   /* Return the data specification. */
   /*================================*/
     
   GetToken(readSource,&not_tkn);
   if (not_tkn.type != RPAREN)
     {
      SyntaxErrorMessage("defrule");
      LHSError = CLIPS_TRUE;
      ReturnNodes(data_spec);
      return(NULL);
     }

   return(data_spec);
  }
  
/**************************************************************/
/* TestPattern: Handles parsing of test conditional elements. */
/*                                                            */
/* <test-CE> ::= (test <function-call>)                       */
/**************************************************************/
static struct node *TestPattern(readSource)
  char *readSource;
  {
   struct node *data_spec;
   struct token tst_tkn;
       
   /*================================================*/
   /* Create the data specification for the test CE. */
   /*================================================*/

   SavePPBuffer(" ");
   data_spec = GetNode();
   data_spec->type = PAT_TEST;
   data_spec->expression = Function0Parse(readSource);

   if (data_spec->expression == NULL)
     {
      LHSError = CLIPS_TRUE;
      ReturnNodes(data_spec);
      return(NULL);
     }

   /*=========================================================*/
   /* Check for the closing right parenthesis of the test CE. */
   /*=========================================================*/

   GetToken(readSource,&tst_tkn);
   if (tst_tkn.type != RPAREN)
     { 
      SyntaxErrorMessage("test conditional element");
      LHSError = CLIPS_TRUE;
      ReturnNodes(data_spec);
      return(NULL);
     }
  
   return(data_spec);
  }
  
/****************************************************************/
/* AssignmentParse: Finishes the parsing of pattern conditional */
/*   elements that have been bound to a  variable.              */
/*                                                              */
/* <assigned-pattern-CE> ::= ?<variable-symbol> <- <pattern-CE> */
/****************************************************************/
static struct node *AssignmentParse(readSource,fact_address)
  char *readSource;
  SYMBOL_HN *fact_address;
  {
   struct node *data_spec;
   struct token lhs_tkn;

   /*===================================*/
   /* Store the name of the variable to */
   /* which the pattern is bound.       */
   /*===================================*/

   data_spec = GetNode();
   data_spec->type = PATTERN;
   data_spec->negated = CLIPS_FALSE;
   data_spec->value = (VOID *) fact_address;
   data_spec->expression = NULL;
   data_spec->right = NULL;
   data_spec->bottom = NULL;

   SavePPBuffer(" ");         

   /*===============================================*/
   /* Check for binder token, "<-", after variable. */
   /*===============================================*/

   GetToken(readSource,&lhs_tkn);
   
   if ((lhs_tkn.type == SYMBOL) ? (strcmp(ValueToString(lhs_tkn.value),"<-") != 0) : 
                                   CLIPS_TRUE)
     {
      SyntaxErrorMessage("binding patterns");
      LHSError = CLIPS_TRUE;
      ReturnNodes(data_spec);
      return(NULL);
     }

   SavePPBuffer(" ");          

   /*================================================*/
   /* Check for opening left parenthesis of pattern. */
   /*================================================*/ 
        
   GetToken(readSource,&lhs_tkn);
   if (lhs_tkn.type != LPAREN)
     {
      SyntaxErrorMessage("binding patterns");
      LHSError = CLIPS_TRUE;
      ReturnNodes(data_spec);
      return(NULL);
     }

   /*======================================================*/
   /* Parse the pattern and return the data specification. */
   /*======================================================*/    

   GetToken(readSource,&lhs_tkn);

   data_spec->right = SequenceRestrictionParse(readSource,&lhs_tkn);
   if (LHSError == CLIPS_TRUE) 
     {
      ReturnNodes(data_spec);
      return(NULL);
     }

   return(data_spec);
  }
  
/***********************************************************/
/* SimplePatternParse: Parses a simple pattern (an opening */
/*   parenthesis followed by one or more fields followed   */
/*   by a closing parenthesis).                            */
/*                                                         */
/* <pattern-CE> ::= <ordered-pattern-CE> |                 */
/*                  <template-pattern-CE>                  */
/***********************************************************/
static struct node *SimplePatternParse(readSource,rln_tkn)
  char *readSource;
  struct token *rln_tkn;
  {
   struct node *data_spec;

   data_spec = GetNode();
   data_spec->type = PATTERN;
   data_spec->negated = CLIPS_FALSE;
   data_spec->value = NULL;
   data_spec->expression = NULL;
   data_spec->bottom = NULL;
   data_spec->right = SequenceRestrictionParse(readSource,rln_tkn);

   if (LHSError == CLIPS_TRUE)
     {
      ReturnNodes(data_spec); 
      return(NULL);
     }

   return(data_spec);
  }
  
/********************************************************************/
/* SequenceRestrictionParse: Parses a sequence of constraint fields */
/*   found within a pattern.  This function recognizes deftemplate  */
/*   patterns and will call the appropriate routines to parse these */
/*   types of patterns.                                             */
/*                                                                  */
/* <pattern-CE> ::= <ordered-pattern-CE> | <template-pattern-CE>    */
/*                                                                  */
/* <ordered-pattern-CE> ::= (<constraint>+)                         */
/********************************************************************/
static struct node *SequenceRestrictionParse(readSource,lhs_tkn)
  char *readSource;
  struct token *lhs_tkn;
  {
   struct node *head;
   struct node *next_element, *last_element = NULL;

   head = NULL;
 
   /*========================*/
   /* Check for deftemplate. */
   /*========================*/
 
#if DEFTEMPLATE_CONSTRUCT
   if (lhs_tkn->type == SYMBOL)
     {
      if (FindDeftemplate(ValueToString(lhs_tkn->value)) != NULL)
        {
         head = DeftemplateLHSParse(readSource,lhs_tkn->value);
         if (head == NULL)
           { 
            LHSError = CLIPS_TRUE;
            return(NULL);
           }
         return(head);
        }
     }
#endif
   
   /*==============================*/
   /* Parse other simple patterns. */
   /*==============================*/
   
   while (lhs_tkn->type != RPAREN)
     {
      next_element = RestrictionParse(readSource,lhs_tkn);
  
      if (LHSError == CLIPS_TRUE)
        {
         ReturnNodes(head);
         return(NULL);
        }

      if (last_element == NULL)
        { head = next_element; }
      else
        { last_element->right = next_element; }
      last_element = next_element;

      if (lhs_tkn->type != RPAREN)   
        {
         PPBackup(); 
         SavePPBuffer(" ");
         SavePPBuffer(lhs_tkn->print_rep);
        }      
     }

   if (head == NULL)
     {
      PrintCLIPS(WERROR,"\nA pattern must have one or more fields\n");
      LHSError = CLIPS_TRUE;
     }

   return(head);   
  }
  
/******************************************************************/
/* RestrictionParse: Parses a single field within a pattern. This */
/*    field may either be a single field wildcard, a multifield   */
/*    wildcard, a single field variable, a multifield variable,   */
/*    or a series of connected constraints.                       */
/*                                                                */
/* <constraint> ::= ? |                                           */
/*                  $? |                                          */
/*                  <multifield-variable> |                       */
/*                  <connected-constraint>                        */
/******************************************************************/
struct node *RestrictionParse(readSource,rst_tkn)
  char *readSource;
  struct token *rst_tkn;
  {      
   struct node *head;

   if ((rst_tkn->type == SINGLE) || 
       (rst_tkn->type == BWORDS) || 
       (rst_tkn->type == MULTIFIELD))
     {
      head = GetNode();
      head->type = rst_tkn->type;
      head->negated = CLIPS_FALSE;
      if (rst_tkn->type == BWORDS)
        { head->value = rst_tkn->value; }
      GetToken(readSource,rst_tkn);
     }
   else
     { head = ConjuctiveRestrictionParse(readSource,rst_tkn); }
   
   return(head);   
  }
  
/*******************************************************************/
/* ConjuctiveRestrictionParse: Parses a single constraint field in */
/*   a pattern that is not a single field wildcard, multifield     */
/*   wildcard, or multifield variable. The field may consist of a  */
/*   number of subfields tied together using the & connective      */
/*   constraint and/or the | connective constraint.                */
/*                                                                 */
/* <connected-constraint>                                          */
/*            ::= <single-constraint> |                            */
/*                <single-constraint> & <connected-constraint> |   */
/*                <single-constraint> | <connected-constraint>     */
/*******************************************************************/
static struct node *ConjuctiveRestrictionParse(readSource,cnj_tkn)
  char *readSource;
  struct token *cnj_tkn;
  {
   struct node *bind_node;
   struct node *new_node, *next_or, *next_and;
   int conn_type;
   
   /*=====================================*/
   /* Get the first node and determine if */
   /* it is a binding variable.           */
   /*=====================================*/
   
   new_node = LiteralRestrictionParse(readSource,cnj_tkn);
   
   if (LHSError == CLIPS_TRUE)
     { return(NULL); }
   
   GetToken(readSource,cnj_tkn);
   
   if ((new_node->type == BWORD) && 
       (new_node->negated == CLIPS_FALSE) &&
       (cnj_tkn->type != LOR))
     { 
      bind_node = new_node; 
      next_or = NULL;
      next_and = NULL;
     }
   else
     {
      bind_node = GetNode();
      bind_node->type = SINGLE;
      bind_node->negated = CLIPS_FALSE;
      bind_node->bottom = new_node;
      next_or = new_node;
      next_and = new_node;
     }
     
   /*=====================================*/
   /* Get the first node and determine if */
   /* it is a binding variable.           */
   /*=====================================*/
      
   while ((cnj_tkn->type == LOR) || (cnj_tkn->type == LAND))
     {
      conn_type = cnj_tkn->type;

      GetToken(readSource,cnj_tkn);
      new_node = LiteralRestrictionParse(readSource,cnj_tkn);      

      if (LHSError == CLIPS_TRUE)
        { 
         ReturnNodes(bind_node);
         return(NULL);
        }
        
      if (conn_type == LOR)
        {
         if (next_or == NULL)
           { bind_node->bottom = new_node; }
         else
           { next_or->bottom = new_node; }
         next_or = new_node;
         next_and = new_node;
        }
      else if (conn_type == LAND)
        {
         if (next_and == NULL)
           { 
            bind_node->bottom = new_node;
            next_or = new_node;
           }
         else
           { next_and->right = new_node; }
         next_and = new_node;
        }
      else
        {
         CLIPSSystemError("LHSPARSE",1);
         ExitCLIPS(4);
        }

      /*==================================================*/
      /* Determine if any more restrictions are connected */
      /* to the current list of restrictions.             */
      /*==================================================*/

      GetToken(readSource,cnj_tkn);
     }
     
   return(bind_node);
  }

/***********************************************************/
/* LiteralRestrictionParse: Parses a subfield of a field.  */
/*   The subfield may be a literal constraint, a predicate */
/*   constraint, a return value constraint, or a variable  */
/*   constraint. The constraints may also be negated using */
/*   the ~ connective constraint.                          */
/*                                                         */
/* <single-constraint>     ::= <term> | ~<term>            */
/*                                                         */  
/*  <term>                 ::= <constant> |                */
/*                             <single-field-variable> |   */ 
/*                             :<function-call> |          */
/*                             =<function-call>            */
/***********************************************************/
static struct node *LiteralRestrictionParse(readSource,ltr_tkn)
  char *readSource;
  struct token *ltr_tkn;
  {     
   struct node *head;
   
   head = GetNode();

   /*====================================================*/
   /* Determine if the field has a '~' preceding it. If  */
   /* it  does, then the field has negative state logic. */
   /* Otherwise the field has positive state logic.      */
   /*====================================================*/

   if (ltr_tkn->type == LNOT)
     {
      GetToken(readSource,ltr_tkn);
      head->negated = CLIPS_TRUE;
     }
   else
     { head->negated = CLIPS_FALSE; }
   
   /*================================================================*/ 
   /* Determine if the field is valid.  Valid fields are ?variables, */
   /* words, strings, numbers, :(expression), and =(expression).     */
   /*================================================================*/ 
   
   head->type = ltr_tkn->type;

   if (ltr_tkn->type == SYMBOL)
     {
      if (strcmp(ValueToString(ltr_tkn->value),"=") == 0)
        {
         head->expression = Function0Parse(readSource);
         if (head->expression == NULL) 
           {
            LHSError = CLIPS_TRUE;
            ReturnNodes(head);
            return(NULL);
           }
         head->type = head->expression->type;   
        }
      else if (strcmp(ValueToString(ltr_tkn->value),":") == 0)
        {
         head->type = COAMP;
         head->expression = Function0Parse(readSource);
         if (head->expression == NULL) 
           {
            LHSError = CLIPS_TRUE;
            ReturnNodes(head);
            return(NULL);
           }       
        }
      else
        { head->value = ltr_tkn->value; }
     }
   else if ((ltr_tkn->type == BWORD)  || 
            (ltr_tkn->type == FLOAT) ||
            (ltr_tkn->type == INTEGER) ||
            (ltr_tkn->type == STRING))
     { head->value = ltr_tkn->value; }
   else
     {  
      SyntaxErrorMessage("defrule");
      LHSError = CLIPS_TRUE;
      ReturnNodes(head);
      return(NULL);
     }

   return(head);
  }

/************************************************************/
/* CreateInitialPattern: Creates the pattern (initial-fact) */
/*   for use in rules which have no LHS patterns.           */
/************************************************************/
static struct node *CreateInitialPattern()
  {
   struct node *head;
   
   head = GetNode();
   head->type = PATTERN;
   head->negated = CLIPS_FALSE;
   head->right = GetNode();
   head->right->type = SINGLE;
   head->right->negated = CLIPS_FALSE;
   head->right->bottom = GetNode();
   head->right->bottom->negated = CLIPS_FALSE;
   head->right->bottom->type = SYMBOL;
   head->right->bottom->value = (VOID *) AddSymbol("initial-fact");
   return(head);
  }
  
/***********************************************************/
/* ParseRuleRHS: Coordinates all the actions necessary for */
/*    parsing the RHS of a rule.                           */
/***********************************************************/
globle struct expr *ParseRuleRHS(readSource)
  char *readSource;
  {
   struct expr *actions;
   struct token end_tkn;

   /*=========================================================*/
   /* Process the actions on the right hand side of the rule. */ 
   /*=========================================================*/

   SavePPBuffer("\n   ");
   SetIndentDepth(3);

   actions = GroupActions(readSource,&end_tkn,CLIPS_TRUE,NULL);

   if (actions == NULL) return(NULL);
     
   /*=============================*/
   /* Reformat the closing token. */
   /*=============================*/
   
   PPBackup();
   PPBackup();
   SavePPBuffer(end_tkn.print_rep);

   /*======================================================*/
   /* Check for the closing right parenthesis of the rule. */
   /*======================================================*/

   if (end_tkn.type != RPAREN)
     {
      SyntaxErrorMessage("defrule");
      ReturnExpression(actions);
      return(NULL);
     }
     
   return(actions);
  }
  
/*****************************************************************/
/* TagLHSLogicalNodes: Marks all and and or conditional elements */
/*   contained within a logical conditional element as having    */
/*   the properties associated with a logical CE.                */
/*****************************************************************/
static VOID TagLHSLogicalNodes(nodePtr)
  struct node *nodePtr;
  {
   while (nodePtr != NULL)
     {
      nodePtr->logical = CLIPS_TRUE;
      if ((nodePtr->type == PAT_AND) || (nodePtr->type == PAT_OR))
        { TagLHSLogicalNodes(nodePtr->right); }
      nodePtr = nodePtr->bottom;
     }
  }

#endif


