   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*               SYSTEM PRIMARY MODULE                 */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*      Brian L. Donnell                                     */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _SYSPRIME_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_

#include "setup.h"
#include "match.h"
#include "constant.h"
#include "utility.h"
#include "constrct.h"
#include "evaluatn.h"
#include "clipsmem.h"
#include "expressn.h"
#include "router.h"
#include "sysprime.h"

#if DEFRULE_CONSTRUCT
#include "engine.h"
#include "reteutil.h"
#endif

#if DEFGLOBAL_CONSTRUCT
#include "defglobl.h"
#endif

#if ANSI_COMPILER  
   VOID                           PrimaryFunctionDefinitions(void);
   VOID                           WatchCommand(void);
   VOID                           UnwatchCommand(void);
   VOID                           ResetCommand(void);
   VOID                           ClearCommand(void);
   VOID                           ExitCommand(void);
   VOID                           WhileFunction(DATA_OBJECT_PTR);
   VOID                           IfFunction(DATA_OBJECT_PTR);
   VOID                           BindFunction(DATA_OBJECT_PTR);
   VOID                           PrognFunction(DATA_OBJECT_PTR);
   VOID                           ReturnFunction(DATA_OBJECT_PTR);
   VOID                           BreakFunction(void);
   
   BOOLEAN                        NopSysFunction(void);
   VOID                           GetBindSysFunction(DATA_OBJECT_PTR);
   VOID                           GetPointerSysFunction(DATA_OBJECT_PTR);
   VOID                           GetEndSysFunction(DATA_OBJECT_PTR); 
#else
   VOID                           PrimaryFunctionDefinitions();
   VOID                           WatchCommand();
   VOID                           UnwatchCommand();
   VOID                           ResetCommand();
   VOID                           ClearCommand();
   VOID                           ExitCommand();
   VOID                           WhileFunction();
   VOID                           IfFunction();
   VOID                           BindFunction();
   VOID                           PrognFunction();
   VOID                           ReturnFunction();
   VOID                           BreakFunction();
   
   BOOLEAN                        NopSysFunction();
   VOID                           GetBindSysFunction();
   VOID                           GetPointerSysFunction();
   VOID                           GetEndSysFunction(); 
#endif

globle int ReturnFlag = CLIPS_FALSE; 
globle int ReturnContext = CLIPS_FALSE;
globle int BreakContext = CLIPS_FALSE;

static int BreakFlag = CLIPS_FALSE;

#if ! RUN_TIME
/*******************************************/
/* PrimaryFunctionDefinitions        */
/*******************************************/
globle VOID PrimaryFunctionDefinitions()
  {
   /*===============================*/
   /* Primary command line actions. */
   /*===============================*/
   
   DefineFunction("watch",   'v', PTIF WatchCommand,   "WatchCommand");
   DefineFunction("unwatch", 'v', PTIF UnwatchCommand, "UnwatchCommand");
   DefineFunction("clear",   'v', PTIF ClearCommand,   "ClearCommand");
   DefineFunction("reset",   'v', PTIF ResetCommand,   "ResetCommand");
   DefineFunction("exit",    'v', PTIF ExitCommand,    "ExitCommand");
  
   /*======================*/
   /* Primary RHS actions. */
   /*======================*/

   DefineFunction("if",           'u', PTIF IfFunction,    "IfFunction");
   DefineFunction("while",        'u', PTIF WhileFunction, "WhileFunction");
   DefineFunction("bind",         'u', PTIF BindFunction,  "BindFunction");
   DefineFunction("progn",        'u', PTIF PrognFunction, "PrognFunction");
   DefineFunction("return",       'u', PTIF ReturnFunction,"ReturnFunction");
   DefineFunction("break",        'v', PTIF BreakFunction, "BreakFunction");

   /*========================*/
   /* System Only functions. */
   /*========================*/

#if DEFRULE_CONSTRUCT
   DefineFunction("(nop)",        'b', NopSysFunction,        "NopSysFunction");
   DefineFunction("(constant)",   'b', NopSysFunction,        "NopSysFunction");
   DefineFunction("(notconstant)",'b', NopSysFunction,        "NopSysFunction");
   DefineFunction("(neq_field)",  'b', NopSysFunction,        "NopSysFunction");
   DefineFunction("(eq_field)",   'b', NopSysFunction,        "NopSysFunction");
   DefineFunction("(get_bind)",   'u', PTIF GetBindSysFunction,    "GetBindSysFunction");
   DefineFunction("(pointer)",    'u', PTIF GetPointerSysFunction, "GetPointerSysFunction");
   DefineFunction("(get_field)",  'u', PTIF GetFieldSysFunction,   "GetFieldSysFunction");
   DefineFunction("(get_end)",    'u', PTIF GetEndSysFunction,     "GetEndSysFunction");
#endif
  }
#endif

/************************************************************/
/* WatchCommand: parses the watch command and sets the      */
/*   appropriate global variable to the value of 1. Fact    */
/*   assertions and retractions, and rule activations,      */
/*   deactivations, and firings are respectively monitored  */
/*   if facts or rules are watched.                         */
/************************************************************/
globle VOID WatchCommand()
  {
   DATA_OBJECT arg_ptr;
   char *argument;

   if (ArgCountCheck("watch",EXACTLY,1) == -1) return;
   
   if (ArgTypeCheck("watch",1,SYMBOL,&arg_ptr) == CLIPS_FALSE) return;
   argument = DOToString(arg_ptr);
   
   SetWatchItem(argument,ON);
  }

/************************************************************/
/* UnwatchCommand: parses the unwatch command and sets the  */
/*   appropriate global variable to the value of 0. Fact    */
/*   assertions and retractions, and rule activations,      */
/*   deactivations, and firings are respectively not        */
/*   monitored if facts or rules are unwatched.             */
/************************************************************/
globle VOID UnwatchCommand()
  {
   DATA_OBJECT arg_ptr;
   char *argument;

   if (ArgCountCheck("unwatch",EXACTLY,1) == -1) return;
 
   if (ArgTypeCheck("unwatch",1,SYMBOL,&arg_ptr) == CLIPS_FALSE) return;
   
   argument = DOToString(arg_ptr);
   
   SetWatchItem(argument,OFF);
  }

/************************************************/
/* ClearCommand: Clears the CLIPS environment. */
/*   Syntax:  (clear)                           */
/************************************************/
globle VOID ClearCommand()
  {
   if (ArgCountCheck("clear",EXACTLY,0) == -1) return;
   ClearCLIPS();
   return;
  }
  
/************************************************/
/* ResetCommand: Resets the CLIPS environment. */
/*   Syntax:  (reset)                           */
/************************************************/
globle VOID ResetCommand()
  { 
   if (ArgCountCheck("reset",EXACTLY,0) == -1) return;  
   ResetCLIPS();
   return;
  }
  
/**********************************************/
/* ExitCommand: Exits the CLIPS environment. */
/*   Syntax:  (exit)                          */
/**********************************************/
globle VOID ExitCommand()
  {
   if (ArgCountCheck("exit",EXACTLY,0) == -1) return;  
#if VMS
   ExitCLIPS(-2);  /* Fix for weird VMS com file problem. */
#else
   ExitCLIPS(-1);
#endif
   return;
  }

/********************************************************************/
/* WhileFunction:                                                        */
/********************************************************************/
globle VOID WhileFunction(while_result)
  DATA_OBJECT_PTR while_result;
  {
   DATA_OBJECT arg_ptr;
   int svBreakContext;
   
   svBreakContext = BreakContext;
   BreakContext = CLIPS_TRUE;
   RtnUnknown(1,&arg_ptr);
   while (((arg_ptr.value != (VOID *) CLIPSFalseSymbol) ||
           (arg_ptr.type != SYMBOL)) && 
           (GetHaltExecution() != CLIPS_TRUE))
     { 
      if ((BreakFlag == CLIPS_TRUE) || (ReturnFlag == CLIPS_TRUE))
        break;
      CurrentEvaluationDepth++;
      RtnUnknown(2,&arg_ptr);
      CurrentEvaluationDepth--;
      PeriodicCleanup(CLIPS_FALSE,CLIPS_TRUE);
      if ((BreakFlag == CLIPS_TRUE) || (ReturnFlag == CLIPS_TRUE))
        break;
      RtnUnknown(1,&arg_ptr); 
     }
   
   BreakContext = svBreakContext;
   BreakFlag = CLIPS_FALSE;
   if (ReturnFlag == CLIPS_TRUE)
     {
      while_result->type = arg_ptr.type;
      while_result->value = arg_ptr.value;
     }
   else
     {
      while_result->type = SYMBOL;
      while_result->value = (VOID *) CLIPSFalseSymbol;
     }
  }

/********************************************************************/
/* IfFunction:                                                       */
/********************************************************************/
globle VOID IfFunction(if_result)
  DATA_OBJECT_PTR if_result;
  {
   int num_a;

   if ((num_a = ArgRangeCheck("if",2,3)) == -1)
     {
      if_result->type = SYMBOL;
      if_result->value = (VOID *) CLIPSFalseSymbol;
      return;
     }

   RtnUnknown(1,if_result);
   if ((BreakFlag == CLIPS_TRUE) || (ReturnFlag == CLIPS_TRUE))
     return;

   /* IMPORTANT: Check neccessary tests */
   
   if ((if_result->value == (VOID *) CLIPSFalseSymbol) &&
       (if_result->type == SYMBOL) && 
       (num_a == 3))
     {
      RtnUnknown(3,if_result); 
      return;
     }
   else if ((if_result->value != (VOID *) CLIPSFalseSymbol) ||
            (if_result->type != SYMBOL)) 
     {
      RtnUnknown(2,if_result); 
      return;
     }


   if_result->type = SYMBOL;
   if_result->value = (VOID *) CLIPSFalseSymbol;
   return;   
  }

/*******************************************/
/* BindFunction:  Performs bind operation. */
/*******************************************/
globle VOID BindFunction(bind_result)
  DATA_OBJECT_PTR bind_result;
  {
   DATA_OBJECT arg_ptr, *bind_ptr, *last_bind;
   int found = CLIPS_FALSE;
   SYMBOL_HN *var_name = NULL;
#if DEFGLOBAL_CONSTRUCT
   int pos = -1;
#endif

   /*===============================================*/
   /* Determine the name of the variable to be set. */
   /*===============================================*/

   RtnUnknown(1,&arg_ptr);
#if DEFGLOBAL_CONSTRUCT
   if (arg_ptr.type == INTEGER) pos = (int) ValueToLong(arg_ptr.value);
   else var_name = (SYMBOL_HN *) arg_ptr.value;
#else
   var_name = (SYMBOL_HN *) arg_ptr.value;
#endif
 
   /*===========================================*/
   /* Determine the new value for the variable. */
   /*===========================================*/

   RtnUnknown(2,&arg_ptr);

   bind_result->type = arg_ptr.type;
   bind_result->value = arg_ptr.value;
   bind_result->begin = arg_ptr.begin;
   bind_result->end = arg_ptr.end; 

   /*==================================*/
   /* Bind a defglobal if appropriate. */
   /*==================================*/
   
#if DEFGLOBAL_CONSTRUCT
   if (pos >= 0)
     { 
      QSetDefglobalValue(pos,bind_result);
      return;
     }
#endif
   
   /*===============================================*/
   /* Search for the variable in the list of binds. */
   /*===============================================*/

   bind_ptr = BindList;
   last_bind = NULL;

   while ((bind_ptr != NULL) && (found == CLIPS_FALSE))
     {
      if (bind_ptr->name == var_name)
        { found = CLIPS_TRUE; }
      else
        {
         last_bind = bind_ptr;
         bind_ptr = bind_ptr->next;
        }
     }

   /*========================================================*/
   /* If variable was not in the list of binds, then add it. */
   /* Make sure that this operation preserves the bind list  */
   /* as a stack.                                            */
   /*========================================================*/

   if (found == CLIPS_FALSE)
     {
      bind_ptr = get_struct(dataObject);
      bind_ptr->name = var_name;
      bind_ptr->next = NULL;
      if (last_bind == NULL)
        { BindList = bind_ptr; }
      else
        { last_bind->next = bind_ptr; }
     }
   else
     { ValueDeinstall(bind_ptr); }

   /*================================*/
   /* Set the value of the variable. */
   /*================================*/

   bind_ptr->type = arg_ptr.type;
   bind_ptr->value = arg_ptr.value;
   bind_ptr->begin = arg_ptr.begin;
   bind_ptr->end = arg_ptr.end; 
   ValueInstall(&arg_ptr);
   
   return;
  }
  
/****************************************/
/* PrognFunction                                */
/****************************************/
globle VOID PrognFunction(prognValuePtr)
  DATA_OBJECT_PTR prognValuePtr;
  {
   int numa, i;

   numa = RtnArgCount();

   if (numa == 0)
     {
      prognValuePtr->type = SYMBOL;
      prognValuePtr->value = (VOID *) CLIPSFalseSymbol; 
      return;
     }

   i = 1;
   while ((i <= numa) && (GetHaltExecution() != CLIPS_TRUE))
     {
      RtnUnknown(i,prognValuePtr);
      if ((ReturnFlag == CLIPS_TRUE) || (BreakFlag == CLIPS_TRUE))
        break;
      i++;
     }
   
   if (GetHaltExecution() == CLIPS_TRUE)
     {
      prognValuePtr->type = SYMBOL; 
      prognValuePtr->value = (VOID *) CLIPSFalseSymbol;
      return;
     }
     
   return;
  }

/****************************************/
/* ReturnFunction                       */
/****************************************/
globle VOID ReturnFunction(result)
  DATA_OBJECT_PTR result;
  {
   int numa;

   if ((ReturnContext == CLIPS_FALSE) ? CLIPS_TRUE :
       ((numa = ArgCountCheck("return",NO_MORE_THAN,1)) == -1))
     {
      result->type = RVOID;
      result->value = CLIPSFalseSymbol;
      return;
     }
   if (numa == 0)
     {
      result->type = RVOID;
      result->value = (VOID *) CLIPSFalseSymbol; 
     }
   else
     RtnUnknown(1,result);
   ReturnFlag = CLIPS_TRUE;
  }

/****************************************/
/* BreakFunction                        */
/****************************************/
globle VOID BreakFunction()
  {
   if ((BreakContext == CLIPS_FALSE) ? CLIPS_TRUE :
       (ArgCountCheck("break",EXACTLY,0) == -1))
     return;
   BreakFlag = CLIPS_TRUE;
  }

#if DEFRULE_CONSTRUCT

/**************************************************************************/
/* NopSysFunction:  This function does nothing. It is used as a stub for  */
/*   the constant, notconstant, eqvars, neqvars, eqfield and neqfield.    */
/**************************************************************************/
globle BOOLEAN NopSysFunction()
  { return(CLIPS_TRUE); }

/*******************************************************************************/
/* GetBindSysFunction:  Performs bind operation as the actions of a construct. */
/*******************************************************************************/
globle VOID GetBindSysFunction(bind_result)
  DATA_OBJECT_PTR bind_result;
  {
   DATA_OBJECT_PTR bind_ptr;
   SYMBOL_HN *var_name;
   struct expr *test_ptr;
   
   /*============================================*/
   /* Get the name of the variable being sought. */
   /*============================================*/

   test_ptr = GetFirstArgument();
   var_name = (SYMBOL_HN *) test_ptr->value;

   /*===============================================*/
   /* Search for the variable in the list of binds. */
   /*===============================================*/

   bind_ptr = BindList;
   while (bind_ptr != NULL)
     {
      if (bind_ptr->name == var_name)
        {
         bind_result->type = bind_ptr->type;
         bind_result->value = bind_ptr->value;
         bind_result->begin = bind_ptr->begin;
         bind_result->end = bind_ptr->end; 
         return; 
        }

      bind_ptr = bind_ptr->next;
     }

   /*=====================================================*/
   /* If the variable was not found in the list of binds, */
   /* determine if the variable is only bound on the rhs  */
   /* of the rule.                                        */
   /*=====================================================*/
  
   test_ptr = test_ptr->next_arg;
   if (test_ptr == NULL)
     {    
      SetHaltExecution(CLIPS_TRUE);
      SetEvaluationError(CLIPS_TRUE);
      PrintCLIPS(WERROR,"Variable ");
      PrintCLIPS(WERROR,ValueToString(var_name));
      PrintCLIPS(WERROR," is unbound\n");
      bind_result->type = SYMBOL;
      bind_result->value = (VOID *) AddSymbol("");
      return;
     }
     
   /*=====================================================*/
   /* Variable was bound on the LHS of the rule.  Extract */
   /* the value from one of the LHS patterns.             */
   /*=====================================================*/
   
   GetVarSysFunction(bind_result,test_ptr->value);

   return;
  }
  
/************************************************************************/
/* GetPointerSysFunction:  Returns a pointer to the fact which matches  */
/*   one of the patterns in a rule.  This function is called by         */
/*   retract to get the value of the fact binders.                      */
/*   The protocol for using this argument type has not yet been defined */
/*   for user functions.                                                */
/************************************************************************/
globle VOID GetPointerSysFunction(bound_var)
  DATA_OBJECT_PTR bound_var;
  {
   int pattern;
   struct fact *fact_ptr = NULL;
   struct expr *arg_ptr;
   int lcount, rcount;
   
   /*==================================*/
   /* Extract the pointer to the fact. */
   /*==================================*/

   arg_ptr = GetFirstArgument();
   pattern = (int) ValueToLong(arg_ptr->value) - 1;
   
   if (GlobalLHSBinds == NULL) lcount = 0; 
   else lcount = GlobalLHSBinds->bcount;
   
   if (GlobalRHSBinds == NULL) rcount = 0; 
   else rcount = GlobalRHSBinds->bcount;
   
   if (pattern < lcount)
     { fact_ptr = GlobalLHSBinds->binds[pattern]->origin; }
   else if (pattern <  lcount +  rcount)
     { fact_ptr = GlobalRHSBinds->binds[pattern - lcount]->origin; }
   else
     { 
      CLIPSSystemError("SYSPRIME",1);
      ExitCLIPS(5);
     }

   /*======================================================*/
   /* Return the unknown structure with the pointer value. */
   /*======================================================*/

   bound_var->type = FACT_ADDRESS;
   bound_var->value = (VOID *) fact_ptr;
   return;
  }
  
/*************************************************************/
/* GetVarSysFunction:  Extracts the nth element from the mth */ 
/*   pattern of a rule.                                      */
/*************************************************************/
globle VOID GetVarSysFunction(bound_var,theValue)
  DATA_OBJECT_PTR bound_var;
  VOID *theValue;
  {
   int pattern, element;
   struct fact *fact_ptr = NULL;
   struct element *elem_ptr;
   struct multifieldMarker *marks = NULL;
   int extent;
   int lcount, rcount;
   union
     {
      void *theValue;
      struct 
         {
          unsigned int flag : 1;
          unsigned int int1 : 8;
          unsigned int int2 : 8;
         } theInts;
     } hack;

   hack.theValue = theValue;
   pattern = (int) hack.theInts.int1 - 1;
   element = (int) hack.theInts.int2;
   
   if (GlobalLHSBinds == NULL) lcount = 0; 
   else lcount = GlobalLHSBinds->bcount;
   
   if (GlobalRHSBinds == NULL) rcount = 0; 
   else rcount = GlobalRHSBinds->bcount;
   
   if (pattern < lcount)
     { 
      fact_ptr = GlobalLHSBinds->binds[pattern]->origin;
      marks = GlobalLHSBinds->binds[pattern]->marker;
     }
   else if (pattern <  lcount +  rcount)
     { 
      fact_ptr = GlobalRHSBinds->binds[pattern - lcount]->origin; 
      marks = GlobalRHSBinds->binds[pattern - lcount]->marker; 
     }
   else
     { 
      CLIPSSystemError("SYSPRIME",2);
      ExitCLIPS(5);
     }

   /*===============================================================*/
   /* If the element was less than zero, then all that was required */
   /* is a pointer to the fact that matched the pattern.            */
   /*===============================================================*/
   
   if (hack.theInts.flag)
     {
      bound_var->type = FACT_ADDRESS;
      bound_var->value = (VOID *) fact_ptr;
      return;
     }
   
   /*==================================================================*/
   /* Extract the needed value from the fact that matched the pattern. */
   /*==================================================================*/
   
   extent = -1;

   if (marks != NULL)
     { 
      if (marks->element < element)
        { element = AdjustFieldPosition(marks,element,&extent); }
      else if (marks->element == element)
        { extent = (marks->end - marks->start) + 1; }
     }

   if (extent != -1)
     {
      bound_var->type = MULTIFIELD;
      bound_var->value = (VOID *) fact_ptr;
      bound_var->begin = --element;
      bound_var->end = element + extent - 1;
      return;
     }
   else
     { extent = 1; }  

   element--;
   elem_ptr = &fact_ptr->atoms[element];

   bound_var->type = elem_ptr->type;
   bound_var->value = elem_ptr->value;
   
   return;
  }
  
/*****************************************************************/
/* GetEndSysFunction: Used for modify function.  */
/*****************************************************************/
globle VOID GetEndSysFunction(bound_var)
  DATA_OBJECT_PTR bound_var;
  {
   int pattern, element;
   struct fact *fact_ptr = NULL;
   struct expr *test_ptr;
   int lcount, rcount;

   test_ptr = GetFirstArgument();
   pattern = DOPToInteger(test_ptr) - 1;
   element = DOPToInteger(GetNextArgument(test_ptr));
   
   if (GlobalLHSBinds == NULL) lcount = 0; 
   else lcount = GlobalLHSBinds->bcount;
   
   if (GlobalRHSBinds == NULL) rcount = 0; 
   else rcount = GlobalRHSBinds->bcount;
   
   if (pattern < lcount)
     { fact_ptr = GlobalLHSBinds->binds[pattern]->origin; }
   else if (pattern <  lcount +  rcount)
     { fact_ptr = GlobalRHSBinds->binds[pattern - lcount]->origin; }
   else
     { 
      CLIPSSystemError("SYSPRIME",3);
      ExitCLIPS(5);
     }

   bound_var->type = MULTIFIELD;
   bound_var->value = (VOID *) fact_ptr;
   bound_var->begin = --element;
   bound_var->end = (int) fact_ptr->factLength - 1;
   return;
  }

#endif
  
