   /*******************************************************/
   /*      "C" Language Integrated Production System      */
   /*                                                     */
   /*                  A Product Of The                   */
   /*             Software Technology Branch              */
   /*             NASA - Johnson Space Center             */
   /*                                                     */
   /*             CLIPS Version 5.10  07/17/91            */
   /*                                                     */
   /*              SYSTEM SECONDARY MODULE                */
   /*******************************************************/

/*************************************************************/
/* Purpose:                                                  */
/*                                                           */
/* Principal Programmer(s):                                  */
/*      Gary D. Riley                                        */
/*                                                           */
/* Contributing Programmer(s):                               */
/*                                                           */
/* Revision History:                                         */
/*                                                           */
/*************************************************************/

#define _SYSSECND_SOURCE_

#include <stdio.h>
#define _CLIPS_STDIO_
#include <string.h>

#include "setup.h"

#include "clipsmem.h"
#include "sysdep.h"
#include "facts.h"
#include "utility.h"
#include "router.h"

#if ANSI_COMPILER  
   VOID                           SecondaryFunctionDefinitions(void);
   long int                       IntegerFunction(void);
   double                         FloatFunction(void);
   VOID                           AbsFunction(DATA_OBJECT_PTR);
   VOID                           MinFunction(DATA_OBJECT_PTR);
   VOID                           MaxFunction(DATA_OBJECT_PTR);
   VOID                           StrAssertFunction(void);
   long int                       SetgenFunction(void);
   VOID                          *GensymFunction(void);
   VOID                          *GensymStarFunction(void);
   long                           RandomFunction(void);
   VOID                           SeedFunction(void);
   long int                       LengthFunction(void);
   VOID                           ConserveMemCommand(void);
   long int                       ReleaseMemCommand(void);
   long int                       MemUsedCommand(void);
   long int                       MemRequestsCommand(void);
   VOID                           OptionsCommand(void);
#else
   VOID                           SecondaryFunctionDefinitions();
   long int                       IntegerFunction();
   double                         FloatFunction();
   VOID                           AbsFunction();
   VOID                           MinFunction();
   VOID                           MaxFunction();
   VOID                           StrAssertFunction();
   long int                       SetgenFunction();
   VOID                          *GensymFunction();
   VOID                          *GensymStarFunction();
   long                           RandomFunction();
   VOID                           SeedFunction();
   long int                       LengthFunction();
   VOID                           ConserveMemCommand();
   long int                       ReleaseMemCommand();
   long int                       MemUsedCommand();
   long int                       MemRequestsCommand();
   VOID                           OptionsCommand();
#endif

/***************************************/
/* LOCAL INTERNAL VARIABLE DEFINITIONS */
/***************************************/

   static long int         GensymNumber = 1;

#if ! RUN_TIME
/*******************************************/
/* SecondaryFunctionDefinitions:           */
/*******************************************/
globle VOID SecondaryFunctionDefinitions()
  {
   DefineFunction("trunc",         'l', PTIF IntegerFunction,    "IntegerFunction");
   DefineFunction("integer",       'l', PTIF IntegerFunction,    "IntegerFunction");
   DefineFunction("float",         'd', PTIF FloatFunction,      "FloatFunction");
   DefineFunction("abs",           'u', PTIF AbsFunction,        "AbsFunction");
   DefineFunction("min",           'u', PTIF MinFunction,        "MinFunction");
   DefineFunction("max",           'u', PTIF MaxFunction,        "MaxFunction");
   DefineFunction("gensym",        'w', PTIF GensymFunction,     "GensymFunction");
   DefineFunction("gensym*",       'w', PTIF GensymStarFunction, "GensymStarFunction");
   DefineFunction("setgen",        'l', PTIF SetgenFunction,     "SetgenFunction");
   DefineFunction("system",        'v', PTIF gensystem,          "gensystem");
   DefineFunction("str-assert",    'v', PTIF StrAssertFunction,  "StrAssertFunction");
   DefineFunction("length",        'l', PTIF LengthFunction,     "LengthFunction");
   DefineFunction("length$",       'l', PTIF LengthFunction,     "LengthFunction");
   DefineFunction("time",          'd', PTIF gentime,            "gentime");
   DefineFunction("random",        'l', PTIF RandomFunction,     "RandomFunction");
   DefineFunction("seed",          'v', PTIF SeedFunction,       "SeedFunction");
   DefineFunction("conserve-mem",  'v', PTIF ConserveMemCommand, "ConserveMemCommand");
   DefineFunction("release-mem",   'l', PTIF ReleaseMemCommand,  "ReleaseMemCommand");
#if DEBUGGING_FUNCTIONS
   DefineFunction("mem-used",      'l', PTIF MemUsedCommand,     "MemUsedCommand");
   DefineFunction("mem-requests",  'l', PTIF MemRequestsCommand, "MemRequestsCommand");
#endif
   DefineFunction("options",       'v', PTIF OptionsCommand,     "OptionsCommand");
  }
#endif

/*************************************/
/* IntegerFunction: */
/*************************************/
globle long int IntegerFunction()
  { 
   DATA_OBJECT valstruct;
   
   if (ArgCountCheck("integer",EXACTLY,1) == -1) return(0L);
   if (ArgTypeCheck("integer",1,INTEGER,&valstruct) == CLIPS_FALSE) return(0L);
   
   return(ValueToLong(valstruct.value));
  } 
  
/*************************************/
/* FloatFunction: */
/*************************************/
globle double FloatFunction()
  { 
   DATA_OBJECT valstruct;
   
   if (ArgCountCheck("float",EXACTLY,1) == -1) return(0.0);
   if (ArgTypeCheck("float",1,FLOAT,&valstruct) == CLIPS_FALSE) return(0.0);
   
   return(ValueToDouble(valstruct.value));
  } 

/****************************************/
/* AbsFunction:                               */
/****************************************/
globle VOID AbsFunction(result)  
  DATA_OBJECT_PTR result;                               
  {
   if (ArgCountCheck("abs",EXACTLY,1) == -1) 
     { 
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }
     
   if (ArgTypeCheck("abs",1,INTEGER_OR_FLOAT,result) == CLIPS_FALSE) 
     { 
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }

   if (result->type == INTEGER)
     { 
      if (ValueToLong(result->value) < 0L) 
        { result->value = (VOID *) AddLong(- ValueToLong(result->value)); }
     }
   else if (ValueToDouble(result->value) < 0.0) 
     { result->value = (VOID *) AddDouble(- ValueToDouble(result->value)); }
  }
  
/****************************************/
/* MinFunction                                  */
/****************************************/
globle VOID MinFunction(result)
  DATA_OBJECT_PTR result;
  {
   DATA_OBJECT arg_value;
   int num_a, i;

   if ((num_a = ArgCountCheck("min",AT_LEAST,1)) == -1)
     {
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }

   if (ArgTypeCheck("min",1,INTEGER_OR_FLOAT,result) == CLIPS_FALSE)
     {
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }
   
   for (i = 2 ; i <= num_a ; i++)
     {
      if (ArgTypeCheck("min",i,INTEGER_OR_FLOAT,&arg_value) == CLIPS_FALSE) return;
          
      if (result->type == INTEGER)
        {
         if (arg_value.type == INTEGER)
           {
            if (ValueToLong(result->value) > ValueToLong(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
         else
           { 
            if ((double) ValueToLong(result->value) > 
                         ValueToDouble(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
        }
      else
        {
         if (arg_value.type == INTEGER)
           {
            if (ValueToDouble(result->value) > 
                (double) ValueToLong(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
         else
           { 
            if (ValueToDouble(result->value) > ValueToDouble(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
        }
     }
   
   return;
  } 

/****************************************/
/* MaxFunction                                  */
/****************************************/
globle VOID MaxFunction(result)
  DATA_OBJECT_PTR result;
  {
   DATA_OBJECT arg_value;
   int num_a, i;

   if ((num_a = ArgCountCheck("max",AT_LEAST,1)) == -1)
     {
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }

   if (ArgTypeCheck("max",1,INTEGER_OR_FLOAT,result) == CLIPS_FALSE)
     {
      result->type = INTEGER;
      result->value = (VOID *) AddLong(0L);
      return;
     }
   
   for (i = 2 ; i <= num_a ; i++)
     {
      if (ArgTypeCheck("max",i,INTEGER_OR_FLOAT,&arg_value) == CLIPS_FALSE) return;
          
      if (result->type == INTEGER)
        {
         if (arg_value.type == INTEGER)
           {
            if (ValueToLong(result->value) < ValueToLong(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
         else
           { 
            if ((double) ValueToLong(result->value) < 
                         ValueToDouble(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
        }
      else
        {
         if (arg_value.type == INTEGER)
           {
            if (ValueToDouble(result->value) < 
                (double) ValueToLong(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
         else
           { 
            if (ValueToDouble(result->value) < ValueToDouble(arg_value.value))
              { 
               result->type = arg_value.type;
               result->value = arg_value.value;
              }
           }
        }
     }
   
   return;
  } 

/*****************************************************************/
/* StrAssertFunction:  Allows a fact to be asserted as a string. */
/*****************************************************************/
globle VOID StrAssertFunction()
  {
   DATA_OBJECT arg_ptr;

   if (ArgCountCheck("str-assert",EXACTLY,1) == -1) return;

   if (ArgTypeCheck("str-assert",1,STRING,&arg_ptr) == CLIPS_FALSE) 
     { return; }

   AssertString(DOToString(arg_ptr));
   return;
  }

/****************************************/
/* SetgenFunction:                      */
/****************************************/
globle long int SetgenFunction()
  {
   long lnum;
   DATA_OBJECT val_ptr;

   if (ArgCountCheck("setgen",EXACTLY,1) == -1) return(GensymNumber);
   if (ArgTypeCheck("setgen",1,INTEGER,&val_ptr) == CLIPS_FALSE) return(GensymNumber);
   
   lnum = CoerceToLongInteger(val_ptr.type,val_ptr.value);
   
   if (lnum < 1L) 
     {
      ExpectedTypeError("setgen",1,"number (greater than or equal to 1)");
      return(GensymNumber);
     }
     
   GensymNumber = lnum;
   return(lnum);
  }

/****************************************/
/* GensymFunction:                      */
/****************************************/
globle VOID *GensymFunction()
  {
   char genstring[15];

   ArgCountCheck("gensym",EXACTLY,0);
   
   sprintf(genstring,"gen%ld",GensymNumber);
   GensymNumber++;
   return(AddSymbol(genstring));
  }
  
/****************************************/
/* GensymStarFunction:                  */
/****************************************/
globle VOID *GensymStarFunction()
  {
   char genstring[15];

   ArgCountCheck("gensym*",EXACTLY,0);
   do
     {
      sprintf(genstring,"gen%ld",GensymNumber);
      GensymNumber++;
     } while (FindSymbol(genstring) != NULL);
     
   return(AddSymbol(genstring));
  }
  
/****************************************/
/* RandomFunction:                           */
/****************************************/
globle long RandomFunction()
  {
   ArgCountCheck("random",EXACTLY,0);
   
   return((long) genrand());
  }
  
/****************************************/
/* SeedFunction:                              */
/****************************************/
globle VOID SeedFunction()
  {
   int fnum;
   DATA_OBJECT val_ptr;

   if (ArgCountCheck("seed",EXACTLY,1) == -1) return;
   if (ArgTypeCheck("seed",1,INTEGER,&val_ptr) == CLIPS_FALSE) return;
   
   fnum = (int) DOToLong(val_ptr);
   
   genseed(fnum);
  }

/****************************************/
/* LengthFunction:                           */
/****************************************/
globle long int LengthFunction()
  {
   DATA_OBJECT item;

   if (ArgCountCheck("length",EXACTLY,1) == -1) return(-1L);

   RtnUnknown(1,&item);

   if ((GetType(item) == STRING) || (GetType(item) == SYMBOL))
     {  return( (long) strlen(DOToString(item))); }
     
   else if (GetType(item) == MULTIFIELD)
     { return ( (long) GetDOLength(item)); }
   
   else
     {
      ExpectedTypeError("length",1,"multifield, string, or symbol");
      return(-1L);
     }
  }

/*******************************************/
/* ReleaseMemCommand:  Returns free memory */
/*   stored by CLIPS to the system.        */ 
/*******************************************/
globle long ReleaseMemCommand()
  {
   if (ArgCountCheck("release-mem",EXACTLY,0) == -1) return(0);
   return(ReleaseMemory(-1L,CLIPS_FALSE));
  }
  
/**********************************************/
/* ConserveMemCommand:  Determines whether pretty print */
/*   representation for defrules and deffacts */
/*   will be stored.                          */
/**********************************************/
globle VOID ConserveMemCommand()
  {
   char *argument;
   DATA_OBJECT arg_ptr;

   if (ArgCountCheck("conserve-mem",EXACTLY,1) == -1) return;
   if (ArgTypeCheck("conserve-mem",1,SYMBOL,&arg_ptr) == CLIPS_FALSE) return;
 
   argument = DOToString(arg_ptr);
 
   if (strcmp(argument,"on") == 0)
     { SetConserveMemory(CLIPS_TRUE); }
   else if (strcmp(argument,"off") == 0)
     { SetConserveMemory(CLIPS_FALSE); }
   else
     {
      ExpectedTypeError("conserve-mem",1,"symbol with value on or off");
      return;
     }
 
   return;
  }
  
#if DEBUGGING_FUNCTIONS 
  
/***************************************/
/* MemUsedCommand:  Returns amount of  */
/*   memory allocated by CLIPS.        */
/***************************************/
globle long int MemUsedCommand()
  {
   if (ArgCountCheck("mem-used",EXACTLY,0) == -1) return(0);
   return(MemoryUsed());
  }
  
/**********************************************/
/* MemRequestsCommand:  Indicates number of memory */
/*   allocations made by CLIPS.               */
/**********************************************/
globle long int MemRequestsCommand()
  {

   if (ArgCountCheck("mem-requests",EXACTLY,0) == -1) return(0);
   return(MemoryRequests());
  }
 
#endif
 
/***************************************************************/
/* OptionsCommand: Lists the compiler flags that have been turned on. */
/***************************************************************/
globle VOID OptionsCommand()
  {
   if (ArgCountCheck("options",EXACTLY,0) == -1) return;
  
   PrintCLIPS(WDISPLAY,"Machine type: ");
   
#if GENERIC
   PrintCLIPS(WDISPLAY,"Generic ");
#endif
#if VMS
   PrintCLIPS(WDISPLAY,"VAX VMS ");
#endif
#if UNIX_V
   PrintCLIPS(WDISPLAY,"UNIX System V or 4.2BSD ");
#endif
#if UNIX_7
   PrintCLIPS(WDISPLAY,"UNIX System III Version 7 or Sun Unix ");
#endif
#if MAC_TC4
   PrintCLIPS(WDISPLAY,"Apple Macintosh with Think C 4.0.5 ");
#endif
#if MAC_TC5
   PrintCLIPS(WDISPLAY,"Apple Macintosh with Think C 5.0 ");
#endif
#if MAC_MPW
   PrintCLIPS(WDISPLAY,"Apple Macintosh with MPW C ");
#endif
#if IBM_MSC
   PrintCLIPS(WDISPLAY,"IBM PC with Microsoft C");
#endif
#if IBM_TBC
   PrintCLIPS(WDISPLAY,"IBM PC with Turbo C");
#endif
PrintCLIPS(WDISPLAY,"\n");

PrintCLIPS(WDISPLAY,"ANSI Compiler is ");
#if ANSI_COMPILER
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Defrule construct is ");
#if DEFRULE_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Deffacts construct is ");
#if DEFFACTS_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Deftemplate construct is ");
#if DEFTEMPLATE_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Defglobal construct is ");
#if DEFGLOBAL_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Deffunction construct is ");
#if DEFFUNCTION_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Defgeneric/Defmethod constructs are ");
#if DEFGENERIC_CONSTRUCT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Object System is ");
#if OBJECT_SYSTEM
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"EMACS Editor is ");
#if CLP_EDIT
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Construct compiler is ");
#if CONSTRUCT_COMPILER
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Text processing package is ");
#if CLP_TEXTPRO
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Help system is ");
#if CLP_HELP
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Extended math package is ");
#if EX_MATH
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"String function package is ");
#if STRING_FUNCTIONS
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Multifield function package is ");
#if MULTIFIELD_FUNCTIONS
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Debugging functions are ");
#if DEBUGGING_FUNCTIONS
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Basic I/O is ");
#if BASIC_IO
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif
  
PrintCLIPS(WDISPLAY,"Extended I/O is ");
#if EXT_IO
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Bload capability is ");
#if BLOAD_ONLY
  PrintCLIPS(WDISPLAY,"BLOAD ONLY");
#endif
#if BLOAD
  PrintCLIPS(WDISPLAY,"BLOAD");
#endif
#if BLOAD_AND_BSAVE
  PrintCLIPS(WDISPLAY,"BLOAD AND BSAVE");
#endif
#if (! BLOAD_ONLY) && (! BLOAD) && (! BLOAD_AND_BSAVE)
  PrintCLIPS(WDISPLAY,"OFF ");
#endif
PrintCLIPS(WDISPLAY,"\n");

PrintCLIPS(WDISPLAY,"Block memory is ");
#if BLOCK_MEMORY
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Run time module is ");
#if RUN_TIME
  PrintCLIPS(WDISPLAY,"ON\n");
#else
  PrintCLIPS(WDISPLAY,"OFF\n");
#endif

PrintCLIPS(WDISPLAY,"Window Interface flag is ");
#if WINDOW_INTERFACE
   PrintCLIPS(WDISPLAY,"ON\n");
#else
   PrintCLIPS(WDISPLAY,"OFF\n");
#endif
  }
