/*********************************** xmenu.c **********************************/
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*        Authors:  BeBe Ly - NASA/Johnson Space Center                       */
/*                  Daniel J. McCoy - University of Houston-Downtown          */
/*                                  & Loral Space Information Systems         */
/*                                                                            */
/******************************************************************************/
/*  This file contains all the functions that create top level menus          */
/******************************************************************************/

#include <stdio.h>
#include <errno.h>
#include "xsetup.h"
#include "xclips.h"
#include "xmenu.h"
#include "setup.h"

#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>

/********** Global variables **********/
extern Widget toplevel,dialog;
extern Pixmap checker, clips_logo;

/********** local functions visible outside this file **********/
void CreatePullDownMenus();
void CancelPopupSelect();
void MenuFunc();

/********** local functions not visible outside this function **********/
void AboutXCLIPS();

static void CreateFileMenu();
/********** callbacks for File menu **********/
extern void EditCallback();
extern void CompletionDialogCallback();
extern void CompletionEditCallback();
extern void LoadRulesCallback();
extern void LoadBatchCallback();
extern void LoadBinaryCallback();
extern void LoadFactsCallback();
extern void DribbleCallback();
extern void CRSVTraceCallback();
extern void SaveRulesCallback();
extern void SaveBinaryCallback();
extern void SaveFactsCallback();
extern void QuitCallback();
extern void LoadRules();
extern void LoadBatch();
extern void DribbleOn();
extern void LoadBinary();
extern void LoadTheFacts();
extern void CRSVOn();
extern void Save();

static void CreateExecutionMenu();
/********** callbacks for Execution menu **********/
extern void ResetCallback();
extern void RunCallback();
extern void StepCallback();
extern void ClearCLIPSCallback();
extern void ClearScreenCallback();

static void CreateWatchMenu();
static void WatchWindow();
void PopdownSelect();
/********** callbacks for Watch menu **********/
extern void WatchAllCallback();
extern void WatchNoneCallback();
extern void OkWatchCallback();
extern void SetWatchFlagsCallback();
static void CreateBrowseMenu();

static void CreateOptionsMenu();
static void OptionsWindow();
/********** callbacks for Options menu **********/
extern void SetSalienceCallback();
extern void SetStrategyCallback();
extern void OkayOptionsCallback();
extern void CancelOptionsCallback();

static void CreateWindowsMenu();
/********** callbacks for Browse menu **********/
extern void DefruleManagerCallback();
extern void DeffactManagerCallback();
extern void DeftemplateManagerCallback();
extern void DeffunctionManagerCallback();
extern void DefgenericManagerCallback();
extern void DefinstancesManagerCallback();
extern void DefclassManagerCallback();
extern void AgendaManagerCallback();

/********* callbacks for Window menu **********/

extern void FactsWindowCallback();
extern void AgendaWindowCallback();
extern void InstancesWindowCallback();
extern void GlobalsWindowCallback();
extern void AllWindowsCallback();
extern void NoWindowsCallback();
extern void CommandLineCLIPSCallback();
extern void ColorUtilityCallback();

/********** from CLIPS code **********/
extern int SetCommandString();
extern int PrintCLIPS();
extern int AddRouter();
extern void DeleteRouter();

/********** from xclips.c **********/
extern int XclipsExit();
extern int XclipsQuery();

/********** from xedit.c **********/
extern void EditNewFile();
extern void EditorRevert();
extern void EditorSaveAs();

/********** from ClipsTextAct.c **********/
extern void MoveEndOfFile();

/********** external variables **********/
extern Arg args[10];
extern char *WatchName[];
Widget facts_window, agenda_window, 
       instances_window, globals_window;
Widget file_dribble, file_crsv_trace;
Widget button_form, button;
Widget file, file_list;

Widget watch_widgets[13];
Widget watchShell = NULL,watchForm;
Widget optionsShell = NULL,optionsForm;
Widget option_widgets[5];
Widget strategy_widgets[7];
Widget sal_opt_widgets[3];
Widget FileItemWidgets[7];
Widget ExecItemWidgets[5];
extern Widget defrule_manager, deffact_manager, deftemplate_manager,
              deffunction_manager, defgeneric_manager, definstances_manager,
              defclass_manager, agenda_manager;
extern Widget agenda_form, agenda_text;
extern Widget dialog_text,facts_form, facts_text;

extern char path[255];
char **filenames;
int number_entries;
int file_item = -1;
extern XEvent event;

/*******************************************************************************
          Name:        CreatePullDownMenus
          Description: Creates all pulldown menus
          arguements:  parent - widget all menu buttons will be in
          Returns:     None
*******************************************************************************/
void CreatePullDownMenus(parent)
  Widget parent;
  {
  button_form = XtCreateManagedWidget("buttonForm",
                                      formWidgetClass,
                                      parent,
                                      NULL, 0);
  XtSetArg(args[0], XtNbitmap, clips_logo);
  XtSetArg(args[1], XtNinternalHeight, 0);
  XtSetArg(args[2], XtNinternalWidth, 0);
  XtSetArg(args[3], XtNshapeStyle, XmuShapeOval);
  XtSetArg(args[4], XtNborderWidth, 0);
  button = XtCreateManagedWidget("button",
                                 commandWidgetClass,
                                 button_form,
                                 args, 5);

  XtAddCallback(button, XtNcallback, AboutXCLIPS, NULL);

  CreateFileMenu(button_form);
  CreateExecutionMenu(button_form);
  CreateBrowseMenu(button_form);
  CreateWindowsMenu(button_form);
  }

/*******************************************************************************
          Name:        CreateFileMenu
          Description: Creates File menu
          arguements:  parent - form widget the menu button will be in
          Returns:     None
*******************************************************************************/
static void CreateFileMenu(parent)
  Widget parent;
  {
  Widget line,menu, entry;
  int i = 0;

  XtSetArg(args[0], XtNfromHoriz, button);
  XtSetArg(args[1], XtNlabel, "File");
  button = XtCreateManagedWidget("button",
                                 menuButtonWidgetClass,
                                 parent,
                                 args, 2);

  menu = XtCreatePopupShell("menu",
                            simpleMenuWidgetClass,
                            button,
                            NULL, 0);

  XtSetArg(args[0], XtNleftMargin, 15);

  FileItemWidgets[i] = XtCreateManagedWidget("Edit...         ^V",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, EditCallback, NULL);

  FileItemWidgets[i] = XtCreateManagedWidget("Complete...     ^C",smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, CompletionDialogCallback,(XtPointer)NULL);
  
  line = XtCreateManagedWidget("line",
                               smeLineObjectClass,
                               menu,
                               NULL, 0);

  FileItemWidgets[i] = XtCreateManagedWidget("Load...         ^L",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, LoadRulesCallback, NULL);

  FileItemWidgets[i] = XtCreateManagedWidget("Load Batch...",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, LoadBatchCallback, NULL);

  FileItemWidgets[i] = XtCreateManagedWidget("Load Binary...",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, LoadBinaryCallback, NULL);

  file_dribble = XtCreateManagedWidget("Dribble...      ^D",
                                       smeBSBObjectClass,
                                       menu,
                                       args, 1);
  XtAddCallback(file_dribble, XtNcallback, DribbleCallback, NULL);

  file_crsv_trace = XtCreateManagedWidget("CRSV Trace...   ^O",
                                          smeBSBObjectClass,
                                          menu,
                                          args, 1);
  XtAddCallback(file_crsv_trace, XtNcallback, CRSVTraceCallback, NULL);

  line = XtCreateManagedWidget("line",
                               smeLineObjectClass,
                               menu,
                               NULL, 0);

  FileItemWidgets[i] = XtCreateManagedWidget("Save Binary...",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i++], XtNcallback, SaveBinaryCallback, NULL);

  line = XtCreateManagedWidget("line",
                               smeLineObjectClass,
                               menu,
                               NULL, 0);

  FileItemWidgets[i] = XtCreateManagedWidget("Quit            ^Q",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(FileItemWidgets[i], XtNcallback, QuitCallback, NULL);
  }

/*******************************************************************************
          Name:        CreateExecutionMenu
          Description: Creates Execution menu
          arguements:  parent - form widget the menu button will be in
          Returns:     None
*******************************************************************************/
static void CreateExecutionMenu(parent)
  Widget parent;
  {
  Widget line,menu, entry;
  int i = 0;

  XtSetArg(args[0], XtNfromHoriz, button);
  XtSetArg(args[1], XtNlabel, "Execution");
  button = XtCreateManagedWidget("button",
                                 menuButtonWidgetClass,
                                 parent,
                                 args, 2);

  menu = XtCreatePopupShell("menu",
                            simpleMenuWidgetClass,
                            button,
                            NULL, 0);

  ExecItemWidgets[i] = XtCreateManagedWidget("Reset        ^E",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(ExecItemWidgets[i], XtNcallback, ResetCallback, NULL);
  i++;
  ExecItemWidgets[i] = XtCreateManagedWidget("Run          ^R",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(ExecItemWidgets[i], XtNcallback, RunCallback, NULL);
  i++;
  ExecItemWidgets[i] = XtCreateManagedWidget("Step         ^T",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(ExecItemWidgets[i], XtNcallback, StepCallback, NULL);
  i++;
  entry = XtCreateManagedWidget("Watch...",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(entry,XtNcallback,WatchWindow,NULL);
  
  entry = XtCreateManagedWidget("Options...",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(entry,XtNcallback,OptionsWindow,NULL);
  line =  XtCreateManagedWidget("line",
                               smeLineObjectClass,
                               menu,
                               NULL, 0);

  ExecItemWidgets[i] = XtCreateManagedWidget("Clear CLIPS  ^K",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(ExecItemWidgets[i], XtNcallback, ClearCLIPSCallback, NULL);
  i++;
  ExecItemWidgets[i] = XtCreateManagedWidget("Clear Window ^N",
                                smeBSBObjectClass,
                                menu,
                                NULL, 0);
  XtAddCallback(ExecItemWidgets[i],XtNcallback,ClearScreenCallback,NULL);
  }

/*******************************************************************************
          Name:        CreateWatchMenu
          Description: Creates Watch menu
          arguements:  parent - form widget the menu button will be in
          Returns:     None
*******************************************************************************/
static void CreateWatchMenu(parent)
  Widget parent;
  {

  int n = 0;

  XtSetArg(args[n], XtNfromHoriz, button);n++;
  XtSetArg(args[n], XtNlabel, "Watch");n++;
  button = XtCreateManagedWidget("watchButton",
                                 commandWidgetClass,
                                 button_form,
                                 args, n);

  XtAddCallback(button, XtNcallback, WatchWindow, NULL);

 }
/*******************************************************************************
          Name:        WatchWindow
          Description: Creates Watch menu
          arguments:  
          Returns:     None
*******************************************************************************/
static void WatchWindow(w,client_data,call_data)
Widget w;
XtPointer client_data,call_data;

{
  int n = 0,i;

  Widget Okay,cancel,All,none;
  static char *WidgetName[13] = {"Compilations","Facts","Instances","Rules","Generic-functions",
                                 "Messages", "Deffunctions","Statistics","Globals","Slots","Activations",
                                 "Methods","Message-handlers"};
  /* ====================================================== */
  /*  If the watch shell have already existed pop it up,    */
  /*  else create it.                                       */
  /* ====================================================== */

  if(watchShell != NULL)
    {
       for(i = 0; i < 13; i++)
         {
           n = 0;
           if (GetWatchItem(WatchName[i]))
             XtSetArg( args[n],XtNstate,True);
           else
             XtSetArg( args[n],XtNstate,False);
           n++;
           XtSetValues(watch_widgets[i],args,n);
         }
       XtPopup(watchShell,XtGrabExclusive);
       return;
    }
  /* ========================================= */
  /*   Create the watch toggle menu            */
  /* ========================================= */

  XtSetArg( args[n], XtNwidth,250);n++;
  XtSetArg( args[n], XtNheight,400);n++;
  watchShell = XtCreatePopupShell("Watch Menu",topLevelShellWidgetClass,XtParent(w),NULL,0);
  n = 0;
  watchForm = XtCreateManagedWidget( "watch_form", formWidgetClass,
                                        watchShell, args,n);
  n = 0;
  XtSetArg(args[n],XtNwidth,150);n++;
  XtSetArg(args[n],XtNhorizDistance,15);n++;
  for(i = 0; i < 7;i++)
   {
     if (GetWatchItem(WatchName[i]))
        XtSetArg( args[n],XtNstate,True);
     else
        XtSetArg( args[n],XtNstate,False);
     n++;
     watch_widgets[i] = XtCreateManagedWidget(WidgetName[i],
                                      toggleWidgetClass,
                                      watchForm,
                                      args, n);
     n =  2;
     XtSetArg(args[n],XtNfromVert,watch_widgets[i]);n++;
  }
 n = 1;
 XtSetArg(args[n],XtNfromHoriz,watch_widgets[0]);n++;
 for(; i < 13 ; i++)
  {
      if (GetWatchItem(WatchName[i]))
        XtSetArg( args[n],XtNstate,True);
      else
        XtSetArg( args[n],XtNstate,False);
      n++;
      watch_widgets[i] = XtCreateManagedWidget(WidgetName[i],
                                      toggleWidgetClass,
                                      watchForm,
                                      args, n);
      n = 1;
      XtSetArg(args[n],XtNfromVert,watch_widgets[i]);n++;
      XtSetArg(args[n],XtNfromHoriz,watch_widgets[i - 7]);n++;
  }
  /* ======================= */
  /* Create the "All" button */
  /* ======================= */

  n = 0;
  XtSetArg(args[n],XtNcornerRoundPercent,40);n++;
  XtSetArg(args[n],XtNshapeStyle,XmuShapeRoundedRectangle);n++;
  XtSetArg(args[n],XtNwidth,150);n++;
  XtSetArg(args[n],XtNfromVert,watch_widgets[6]);n++;
  XtSetArg(args[n],XtNvertDistance,31);n++;
  XtSetArg(args[n],XtNlabel,"All");n++;
  All = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                        watchForm,
                                        args, n);
  XtAddCallback(All, XtNcallback, WatchAllCallback, NULL);

  /* ============================= */
  /* Create the "None" button      */
  /* ============================= */

  n = 3;
  XtSetArg(args[n],XtNfromHoriz,All);n++;
  XtSetArg(args[n],XtNfromVert,watch_widgets[6]);n++;
  XtSetArg(args[n],XtNvertDistance,31);n++;
  XtSetArg(args[n],XtNhorizDistance,30);n++;
  XtSetArg(args[n],XtNlabel,"None");n++;
  none = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                        watchForm,args,n);
  XtAddCallback(none, XtNcallback, WatchNoneCallback, NULL);

  /* ================================ */
  /* Create the "Okay" button         */
  /* ================================ */

  n = 3;
  XtSetArg(args[n],XtNfromVert,All);n++;
  XtSetArg(args[n],XtNvertDistance,15);n++;
  XtSetArg(args[n],XtNlabel,"Okay");n++;
  Okay = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                        watchForm,
                                        args, n);
  XtAddCallback(Okay,XtNcallback,OkWatchCallback,(XtPointer)watchForm);

  /* ================================ */
  /*   Create the "Cancel" button     */
  /* ================================ */

  n = 3;
  XtSetArg(args[n],XtNfromVert,none);n++;
  XtSetArg(args[n],XtNvertDistance,15);n++;
  XtSetArg(args[n],XtNfromHoriz,Okay);n++;
  XtSetArg(args[n],XtNlabel,"Cancel");n++;
  XtSetArg(args[n],XtNhorizDistance,30);n++;
  cancel = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                        watchForm,
                                        args, n);
  XtAddCallback(cancel,XtNcallback,PopdownSelect,(XtPointer)watchForm);
  XtPopup(watchShell,XtGrabExclusive);
  }

/*******************************************************************************
          Name:        CreateBrowseMenu
          Description: Creates Browse menu
          arguements:  parent - form widget the menu button will be in
          Returns:     None
*******************************************************************************/
static void CreateBrowseMenu(parent)
  Widget parent;
  {
  Widget menu;

  XtSetArg(args[0], XtNfromHoriz, button);
  XtSetArg(args[1], XtNlabel, "Browse");
  button = XtCreateManagedWidget("button",
                                 menuButtonWidgetClass,
                                 parent,
                                 args, 2);
  menu = XtCreatePopupShell("menu",
                            simpleMenuWidgetClass,
                            button,
                            NULL, 0);

  XtSetArg(args[0], XtNleftMargin, 15);
  XtSetArg(args[1], XtNsensitive, False);

  defrule_manager = XtCreateManagedWidget("Defrule Manager...",
                                          smeBSBObjectClass,
                                          menu,
                                          args, 2);
  XtAddCallback(defrule_manager, XtNcallback, DefruleManagerCallback, NULL);

  deffact_manager = XtCreateManagedWidget("Deffacts Manager...",
                                          smeBSBObjectClass,
                                          menu,
                                          args, 2);
  XtAddCallback(deffact_manager, XtNcallback, DeffactManagerCallback, NULL);

  deftemplate_manager = XtCreateManagedWidget("Deftemplates Manager...",
                                              smeBSBObjectClass,
                                              menu,
                                              args, 2);
  XtAddCallback(deftemplate_manager, XtNcallback,
                DeftemplateManagerCallback, NULL);

  deffunction_manager = XtCreateManagedWidget("Deffunction Manager...",
                                              smeBSBObjectClass,
                                              menu,
                                              args, 2);
  XtAddCallback(deffunction_manager, XtNcallback,
                DeffunctionManagerCallback, NULL);

  defgeneric_manager = XtCreateManagedWidget("Defgeneric Manager...",
                                             smeBSBObjectClass,
                                             menu,
                                             args, 2);
  XtAddCallback(defgeneric_manager, XtNcallback,
                DefgenericManagerCallback, NULL);

  defclass_manager = XtCreateManagedWidget("Defclass Manager...",
                                           smeBSBObjectClass,
                                           menu,
                                           args, 2);
  XtAddCallback(defclass_manager, XtNcallback, DefclassManagerCallback, NULL);

  definstances_manager = XtCreateManagedWidget("Definstances Manager...",
                                               smeBSBObjectClass,
                                               menu,
                                               args, 2);
  XtAddCallback(definstances_manager, XtNcallback,
                DefinstancesManagerCallback, NULL);

  agenda_manager = XtCreateManagedWidget("Agenda Manager...",
                                         smeBSBObjectClass,
                                         menu,
                                         args, 1);
  XtAddCallback(agenda_manager, XtNcallback, AgendaManagerCallback, NULL);

}


/*******************************************************************************
          Name:        OptionsWindow
          Description: Creates the Options menu
          arguements:  w - Widget that invokes the callbaack function
                       client,data - unused
          Returns:     None
*******************************************************************************/
static void OptionsWindow(w,client,data)
  Widget w;
  XtPointer client,data;
  {
    Widget Cancel,Okay,menu,entry,salienceEval,strategy;
    int n,i;
    static char *optionList1[5] = {"Dynamic Deftemplate Checking","Reset Global Variables",
                                   "Fact Duplication","Incremental Reset","Auto-Float Dividend"};
    static char *optionList2[7] = {"Depth","Breadth","LEX","MEA","Complexity","Simplicity","Random"};
    static char *optionList3[3] = {"When Defined","When Activated","Every Cycle"};
    Pixmap SalienceBM,StrategyBM;

  /* =============================================== */
  /*    If optionsShell exists pop it up             */
  /* =============================================== */

    if(optionsShell != NULL)
     {
       UpdateOptionsMenu();
       XtPopup(optionsShell,XtGrabNonexclusive);
       return;
     }

  /* ================================================ */
  /*  If optionShell not available create one.        */
  /*  The followings are the widget classes used in   */
  /*  building the option menu                        */
  /*     topLevelShellWidgetClass                     */
  /*        formWidgetClass                           */
  /*          menuButtonWidgetClass                   */
  /*            simpleMenuWidgetClass                 */
  /*             smeBSBOjectClass                     */
  /* ================================================ */
    n = 0;
    XtSetArg( args[n], XtNwidth,250);n++;
    XtSetArg( args[n], XtNheight,400);n++;
    optionsShell = XtCreatePopupShell("Execution Options",topLevelShellWidgetClass,XtParent(w),NULL,0);

  /* ================================================ */
  /*  Create the outside form for the option menu     */
  /* ================================================ */

    optionsForm = XtCreateManagedWidget( "watch_form", formWidgetClass,
                                        optionsShell, args,n);
    SalienceBM = XCreateBitmapFromData(XtDisplay(toplevel),
                                  RootWindowOfScreen(XtScreen(toplevel)),
                                  salience_bits,
                                  salience_width,
                                  salience_height);

  /* ================================================ */
  /*   Create the salience Evaluation setting menu    */
  /* ================================================ */
    
    n = 0;
    XtSetArg(args[n],XtNbitmap,SalienceBM);n++;
    XtSetArg(args[n],XtNwidth,180);n++;
    salienceEval = XtCreateManagedWidget("salienceEvaluation",
                                 menuButtonWidgetClass,
                                 optionsForm,
                                 args, n);
    menu = XtCreatePopupShell("menu",
                                simpleMenuWidgetClass,
                                salienceEval,
                                NULL,0);
    n = 0;
    XtSetArg(args[n],XtNleftMargin,15);n++;
    XtSetArg(args[n],XtNleftBitmap,checker);n++;
    for(i = 0; i < 3 ; i++)
      {
       sal_opt_widgets[i] = XtCreateManagedWidget(optionList3[i],
                                smeBSBObjectClass,
                                menu,
                                args, n);
       XtAddCallback(sal_opt_widgets[i],XtNcallback,SetSalienceCallback,(XtPointer)i);
       n = 1;
     }

  /* ================================================= */
  /*   Create the strategy setting menu                */
  /* ================================================= */

    StrategyBM = XCreateBitmapFromData(XtDisplay(toplevel),
                                  RootWindowOfScreen(XtScreen(toplevel)),
                                  strategy_bits,
                                  strategy_width,
                                  strategy_height);
    n = 0;
    XtSetArg(args[n],XtNhorizDistance,10);n++;
    XtSetArg(args[n],XtNfromHoriz,salienceEval);n++;
    XtSetArg(args[n],XtNbitmap,StrategyBM);n++;
    XtSetArg(args[n],XtNwidth,150);n++;
    strategy = XtCreateManagedWidget("strategy",
                                 menuButtonWidgetClass,
                                 optionsForm,
                                 args, n);
    menu = XtCreatePopupShell("menu",
                            simpleMenuWidgetClass,
                            strategy,
                            NULL, 0);
    n = 0;
    XtSetArg(args[n],XtNleftMargin, 15);n++;
    XtSetArg(args[n], XtNleftBitmap, checker);n++;
    for(i = 0; i < 7; i++)
     {
       strategy_widgets[i] = XtCreateManagedWidget(optionList2[i],
                                smeBSBObjectClass,
                                menu,
                                args, n);
       XtAddCallback(strategy_widgets[i],XtNcallback,SetStrategyCallback,(XtPointer)i);
       n = 1;
     }
  /* ========================================== */
  /*   Create the rest of the option menu       */
  /* ========================================== */

    n = 0;
    XtSetArg(args[n],XtNwidth,200);n++;
    XtSetArg(args[n],XtNhorizDistance,80);n++;
    XtSetArg(args[n],XtNfromVert,salienceEval);n++;
    XtSetArg(args[n],XtNvertDistance,30);n++;
    for(i = 0;i < 5; i++)
     {
        if(( i == AUTO_FLOAT_DIV) || (i == INCREMENTAL_RESET) || ( i == RESET_GLOBALS))
          {
            XtSetArg(args[n], XtNstate,True);n++;
          }
        option_widgets[i] = XtCreateManagedWidget(optionList1[i],
                                      toggleWidgetClass,
                                      optionsForm,
                                      args, n);
        n = 2;
        XtSetArg(args[n],XtNfromVert,option_widgets[i]);n++;
     }

   /* ====================================== */
   /*  Create the "Cancel" button            */
   /* ====================================== */

    n = 0;
    XtSetArg(args[n],XtNshapeStyle,XmuShapeRoundedRectangle);n++;
    XtSetArg(args[n],XtNwidth,150);n++;
    XtSetArg(args[n],XtNfromVert,option_widgets[4]);n++;
    XtSetArg(args[n],XtNvertDistance,31);n++;
    XtSetArg(args[n],XtNlabel,"Cancel");n++;
    Cancel = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                       optionsForm,
                                        args, n);
    XtAddCallback(Cancel,XtNcallback,PopdownSelect,(XtPointer)optionsForm);

   /* ====================================== */
   /*  Create the "OKay" button              */
   /* ====================================== */
    n = 4;
    XtSetArg(args[n],XtNfromHoriz,Cancel);n++;
    XtSetArg(args[n],XtNhorizDistance,30);n++;
    XtSetArg(args[n],XtNlabel,"OKay");n++;
    Okay = XtCreateManagedWidget("watchButton",
                                       commandWidgetClass,
                                       optionsForm,
                                        args, n);
    XtAddCallback(Okay,XtNcallback,OkayOptionsCallback,(XtPointer)NULL);
    XtPopup(optionsShell,XtGrabNonexclusive);
  }

/*******************************************************************************
          Name:        CreateWindowsMenu
          Description: Creates the Windows menu
          arguements:  parent - form widget the menu button will be in
          Returns:     None
*******************************************************************************/
static void CreateWindowsMenu(parent)
  Widget parent;
  {
  Widget line,menu, entry,all,none;

  XtSetArg(args[0], XtNfromHoriz, button);
  XtSetArg(args[1], XtNlabel, "Windows");
  button = XtCreateManagedWidget("button",
                                 menuButtonWidgetClass,
                                 parent,
                                 args, 2);

  menu = XtCreatePopupShell("menu",
                            simpleMenuWidgetClass,
                            button,
                            NULL, 0);

  /* =========================================== */
  /*   Create the "Fact Window" button           */
  /* =========================================== */

  XtSetArg(args[0], XtNleftMargin, 15);
  facts_window = XtCreateManagedWidget("Facts Window",
                                       smeBSBObjectClass,
                                       menu,
                                       args, 1);
  XtAddCallback(facts_window, XtNcallback, FactsWindowCallback, NULL);

  /* =========================================== */
  /*   Create the "Agenda Window" button         */
  /* =========================================== */

  agenda_window = XtCreateManagedWidget("Agenda Window",
                                        smeBSBObjectClass,
                                        menu,
                                        args, 1);
  XtAddCallback(agenda_window, XtNcallback, AgendaWindowCallback, NULL);
 
  /* =========================================== */
  /*   Create the "Instances Window" button      */
  /* =========================================== */

  instances_window =  XtCreateManagedWidget("Instances Window",
                                        smeBSBObjectClass,
                                        menu,
                                        args, 1);

  XtAddCallback(instances_window, XtNcallback,InstancesWindowCallback,NULL);

  /* =========================================== */
  /*   Create the "Global Window" button         */
  /* =========================================== */

  globals_window =  XtCreateManagedWidget("Global Window",
                                        smeBSBObjectClass,
                                        menu,
                                        args, 1);
  XtAddCallback(globals_window, XtNcallback,GlobalsWindowCallback,NULL);

  line = XtCreateManagedWidget("line", smeLineObjectClass, menu, NULL, 0);

  /* =========================================== */
  /*   Create the "All Windows" button           */
  /* =========================================== */

  all = XtCreateManagedWidget("All Windows",
                                        smeBSBObjectClass,
                                        menu,
                                        args, 1);
  XtAddCallback(all,XtNcallback,AllWindowsCallback,NULL);

  /* =========================================== */
  /*   Create the "None" button                  */
  /* =========================================== */

  none = XtCreateManagedWidget("None",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(none,XtNcallback,NoWindowsCallback,NULL);
  line = XtCreateManagedWidget("line", smeLineObjectClass, menu, NULL, 0);

  /* =========================================== */
  /*   Create the "Command Line CLIPS" button    */
  /* =========================================== */


  entry = XtCreateManagedWidget("Command Line CLIPS ^Z",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(entry, XtNcallback, CommandLineCLIPSCallback, NULL);

  /* =========================================== */
  /*   Create the "Color Utilities" button       */
  /* =========================================== */

  entry = XtCreateManagedWidget("Color Utility",
                                smeBSBObjectClass,
                                menu,
                                args, 1);
  XtAddCallback(entry, XtNcallback, ColorUtilityCallback, NULL);

  }

/*******************************************************************************
          Name:        AboutXCLIPS
          Description: Called when CLIPS logo is selected form menu form.
                       It displays the general information about CLIPS.

          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void AboutXCLIPS(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget about, about_form, about_list, about_quit;

  about = XtCreatePopupShell("About XCLIPS",
                             topLevelShellWidgetClass,
                             toplevel,
                             NULL, 0);

  XtSetArg(args[0], XtNdefaultDistance, 0);
  about_form = XtCreateManagedWidget("about_form",
                                     formWidgetClass,
                                     about,
                                     args, 1);

  XtSetArg(args[0], XtNborderWidth, 0);
  XtSetArg(args[1], XtNdefaultColumns, 1);
  XtSetArg(args[2], XtNforceColumns, True);
  XtSetArg(args[3], XtNlist, about_info);
  about_list = XtCreateManagedWidget("menu",
                                     listWidgetClass,
                                     about_form,
                                     args, 4);
  XtAddCallback(about_list, XtNcallback, CancelPopupSelect,
                (XtPointer)about_form);

  XtPopup(about, XtGrabNone);
  }

/*******************************************************************************
          Name:        DialogReturn
          Description: This function will be executed when return was pressed
                       while curssor is in the dialog box of the file select
                       window.
          Arguments:  w - Widget that caused action to be called
                       event - Not used
                       params - Dialog widget
                       num_params - Not used
          Returns:     None
*******************************************************************************/
void DialogReturn(w, event, params, num_params)
  Widget w;
  XEvent *event;
  String *params;
  Cardinal *num_params;
  {
  MenuFunc(w, params, (XtPointer)NULL);
  }

/*******************************************************************************
          Name:        MenuFunc
          Description: Simulates callbacks for the dialog box of the file
                       select window. which callback function is executed
                       depends on which menu item from the file menu of 
                       the main CLIPS window is activated.
          Arguments:  w - Dialog widget
                       client_data - Dialog widget
                       call_data - Not Used
          Returns:     None
*******************************************************************************/
void MenuFunc(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  String filename = XawDialogGetValueString(XtParent(w));
  char fullpath[255];
  Widget popup = XtParent(XtParent(XtParent(w)));

  if(filename[0] == 0)
     return;
  strcpy(fullpath, path);
  strcat(fullpath, filename);
  MoveEndOfFile(dialog_text, &event);
  switch(file_item)
    {
    case EDIT:
      XtDestroyWidget(popup);
      EditNewFile(w, client_data, call_data);
    break;

    case LOADBATCH:
      XtDestroyWidget(popup);
      LoadBatch(fullpath);
    break;

    case LOADBINARY:
      XtDestroyWidget(popup);
      LoadBinary(fullpath);
    break;

    case LOADFACTS:
      XtDestroyWidget(popup);
      LoadTheFacts(fullpath);
    break;

    case LOADRULES:
      XtDestroyWidget(popup);
      LoadRules(fullpath);
    break;

    case DRIBBLEON:
      XtDestroyWidget(popup);
      DribbleOn(fullpath);
    break;

    case CRSVTRACE:
      XtDestroyWidget(popup);
      CRSVOn(fullpath);
    break;

    case SAVERULES:
      Save(w, client_data, call_data);
    break;

    case SAVEBINARY:
      Save(w, client_data, call_data);
    break;

    case SAVEFACTS:
      Save(w, client_data, call_data);
    break;

     }
  }

/*******************************************************************************
          Name:        CancelPopupSelect
          Description: Destroys a the parent of the widget sent
          Arguments:  w - Not Used
                       client_data - Child of widget to destroy
                       call_data - Not Used
          Returns:     None
*******************************************************************************/
void CancelPopupSelect(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  XtDestroyWidget(XtParent((Widget) client_data));
  }


/*******************************************************************************
          Name:        PopdownSelect
          Description: Popdown the parent of the widget sent
          Arguments:  w - Not Used
                       client_data - Child of widget to destroy
                       call_data - Not Used
          Returns:     None
*******************************************************************************/
void PopdownSelect(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
    XtPopdown(XtParent((Widget) client_data));
  }
