/******************************** xmenu_opt.c ********************************/
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*        Author:  BeBe Ly - NASA/Johnson Space Center                       */
/*                 Daniel J. McCoy - University of Houston - Downtown        */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include "xsetup.h"
#include "xclips.h"
#include "setup.h"
#include "engine.h"

/********** local functions visible outside this file **********/
void SetStrategyCallback();
void SetSalienceCallback();

/********** from CLIPS code **********/
extern int SetCommandString();
extern int PrintCLIPS();
extern int GetFactDuplication();
extern int SetFactDuplication();
extern int GetDynamicDeftemplateChecking();
extern int SetDynamicDeftemplateChecking();
extern int GetIncrementalReset();
extern int SetIncrementalReset();

/********** Global variables **********/
extern Widget dialog_text;         /* the standard IO (main CLIPS window) */

extern Widget strategy_widgets[];  /* Widgets in the options window */  
extern Widget sal_opt_widgets[];
extern Widget option_widgets[];
extern XEvent event;              /* Current event */
extern Boolean quit_get_event;    /* Flag to break out of the get event loop*/
extern Pixmap checker;
extern Arg args[10];

int optionFlags[2] = {DEPTH_STRATEGY,WHEN_DEFINED};
/*******************************************************************************
          Name:        SetStrategyCallback
          Description: Called when Depth Strategy is selected from Options menu.
                       It marks the new selection in the Depth Strategy menu,
                       but the value is changed only when the Okay button 
                       is pressed.
          Arguments:   w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void SetStrategyCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  int i,n = 0,index = (int)client_data;

  MoveEndOfFile(dialog_text, &event);

  XtSetArg(args[n], XtNleftBitmap, None);n++;
  for(i = 0 ; i <=  RANDOM_STRATEGY;i++)
    XtSetValues(strategy_widgets[i], args, n);
  XtSetArg(args[0], XtNleftBitmap, checker);n++;
  XtSetValues(strategy_widgets[index], args, n);
  optionFlags[STRATEGY_FLAG] = index;
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        SetSalienceCallback
          Description: Called when Evaluate When salience is set
                       It marks the salience evaluation method to
                       the new selected method, but the new value only
                       changes when Okay button is pressed.
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void SetSalienceCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  int i,n = 0,index = (int)client_data;
  MoveEndOfFile(dialog_text, &event);

  n = 0;
  XtSetArg(args[n], XtNleftBitmap, None);n++;
  for(i = 0; i <= EVERY_CYCLE; i++)
   {
     if(i != index)
      XtSetValues(sal_opt_widgets[i],args,n);
   }
  n = 0;
  XtSetArg(args[n], XtNleftBitmap, checker);n++;
  XtSetValues(sal_opt_widgets[index], args, n);
  optionFlags[SALIENCE_FLAG] = index;
  quit_get_event = True;
  }


/**************************************************************************
 OkayOptionsCallback                                         
 Description : This function reset the option flags to the new values
               and remove the window from the screen.
 Arguments   : w - widget that where the event happened
 Return      : None
**************************************************************************/
void OkayOptionsCallback(w,client,call)
Widget w;
XtPointer client,call;
{
  Boolean OnOff;
  int i,n,flag;
 
   if((GetStrategy()) != optionFlags[STRATEGY_FLAG])
     SetStrategy(optionFlags[STRATEGY_FLAG]);
   if((optionFlags[SALIENCE_FLAG]) != GetSalienceEvaluation())
     SetSalienceEvaluation(optionFlags[SALIENCE_FLAG]);
   n = 0;
   XtSetArg(args[n],XtNstate,&OnOff);n++;
   XtGetValues(option_widgets[FACT_DUPLICATION],args,n);
   if((OnOff)&&(!GetFactDuplication()))
    {
      PrintCLIPS("wclips","(set-fact-duplication TRUE)\n");
      SetCommandString("(set-fact-duplication TRUE)\n");
    }
   else if ((!OnOff)&&(GetFactDuplication()))
    {
      PrintCLIPS("wclips","(set-fact-duplication FALSE)\n");
      SetCommandString("(set-fact-duplication FALSE)\n");
    }
   XtGetValues(option_widgets[DYN_DEFTEMP_CHK],args,n);
   if((OnOff)&&(!GetDynamicDeftemplateChecking()))
    {
      PrintCLIPS("wclips","(set-dynamic-deftemplate-checking TRUE)\n");
      SetCommandString("(set-dynamic-deftemplate-checking TRUE)\n");
    }
   else if((!OnOff) &&(GetDynamicDeftemplateChecking()))
    {
      PrintCLIPS("wclips","(set-dynamic-deftemplate-checking FALSE)\n");
      SetCommandString("(set-dynamic-deftemplate-checking FALSE)\n");
    }
   XtGetValues(option_widgets[AUTO_FLOAT_DIV],args,n);
   if((OnOff)&&(!GetAutoFloatDividend()))
    {
      PrintCLIPS("wclips","(set-auto-float-dividend TRUE)\n");
      SetCommandString("(set-auto-float-dividend TRUE)\n");
    }
   else if((!OnOff) &&(GetAutoFloatDividend()))
    {
      PrintCLIPS("wclips","(set-auto-float-dividend FALSE)\n");
      SetCommandString("(set-auto-float-dividend FALSE)\n");
    }
   XtGetValues(option_widgets[INCREMENTAL_RESET],args,n);
   if((OnOff)&&(!GetIncrementalReset()))
    {
      PrintCLIPS("wclips","(set-incremental-reset TRUE)\n");
      SetCommandString("(set-incremental-reset TRUE)\n");
    }
   else if((!OnOff) &&(GetIncrementalReset()))
    {
      PrintCLIPS("wclips","(set-incremental-reset FALSE)\n");
      SetCommandString("(set-incremental-reset FALSE)\n");
    }
   XtGetValues(option_widgets[RESET_GLOBALS],args,n);
   if((OnOff)&&(!GetResetGlobals()))
    {
      PrintCLIPS("wclips","(set-reset-globals TRUE)\n");
      SetCommandString("(set-reset-globals TRUE)\n");
    }
   else if((!OnOff) &&(GetResetGlobals()))
    {
      PrintCLIPS("wclips","(set-reset-globals FALSE)\n");
      SetCommandString("(set-reset-globals FALSE)\n");
    }
   XtPopdown(XtParent(XtParent(w)));
  quit_get_event = True;
}

/**************************************************************************/

/**************************************************************************/

void CancelOptionsCallback(w,client,call)
Widget w;
XtPointer client,call;
{
   XtPopdown(XtParent(XtParent(w)));
}



/**************************************************************************/

/**************************************************************************/



