/********************************* xmenu_watch.c ******************************/
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*        Author:  BeBe Ly - NASA/Johnson Space Center                        */
/*                 Daniel J. McCoy - University of Houston - Downtown         */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include "xsetup.h"
#include "xclips.h"
#include "setup.h"

/********** local functions visible outside this file **********/
void OkWatchCallback();
void WatchAllCallback();
void WatchNoneCallback();

/********** external variables **********/
extern Widget dialog_text;
extern Widget watch_widgets[];
extern Arg args[10];
extern Pixmap checker;
extern Boolean quit_get_event;
extern XEvent event;      /* Current event */


/******* Global variables ***********/
char *WatchName[MAX_WATCH] = {"compilations","facts","instances","rules","generic-functions",
                              "messages","deffunctions","statistics","globals","slots","activations",
                              "methods","message-handlers"};


/**************************************************************************
   OkWatchCallback
   Description:  This function will reset the watch flags to the new
                 values and remove the watch window from the screen.
   Arguments:    w - widget that event was activated
                 client_data - NULL
                 call_data - Unused
   Return:       None
 **************************************************************************/
void OkWatchCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
{
   int i,n;
   Boolean OnOff;
   
   for(i = 0; i< MAX_WATCH; i++)
    {
       n = 0;
       XtSetArg(args[n],XtNstate,&OnOff);n++;
       XtGetValues(watch_widgets[i],args,n);
     /*----------------------------------------------------------------*/
     /* I have to do this because I am not sure if True and False in X */
     /* are defined the same as CLIPS_TRUE and CLIPS_FALSE             */
     /*----------------------------------------------------------------*/
       if((OnOff)&&(!GetWatchItem(WatchName[i])))
        {
          SetWatchItem(WatchName[i], !GetWatchItem(WatchName[i]));
        }
       else if((!OnOff) && (GetWatchItem(WatchName[i])))
        {
          SetWatchItem(WatchName[i], !GetWatchItem(WatchName[i]));
        }
    }
   XtPopdown(XtParent(XtParent(w)));
   quit_get_event = True;
}

/*******************************************************************************
          Name:        WatchAllCallback
          Description: Called when Watch All button is activated.
                       It turns all the watch toggle buttons to ON, and
                       the watch flags' values will be reset only when 
                       the Okay button is pressed.
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void WatchAllCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
     int i, n = 0;

     XtSetArg(args[n],XtNstate,True);n++;
     for(i = 0;i < MAX_WATCH; i++)
      {
         XtSetValues(watch_widgets[i],args,n);
      }
   quit_get_event = True;
  }

/*******************************************************************************
          Name:        WatchNoneCallback
          Description: Called when Watch None is selected from Watch menu.
                       It turns all the watch toggle buttons to OFF, and
                       the values of the watch flags are resset only when 
                       the Okay button is pressed.
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void WatchNoneCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
     int i, n = 0;

     XtSetArg(args[n],XtNstate,False);n++;
     for(i = 0;i < MAX_WATCH ; i++)
      {
         XtSetValues(watch_widgets[i],args,n);
      }
    quit_get_event = True;
  }

