
/******************************** xmenu _wind.c ********************************/
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*        Author:   BeBe Ly - NASA/Johnson Space Center                       */
/*                  Daniel J. McCoy - University of Houston-Downtown          */
/*                                   & Loral Space Information Systems        */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include "xsetup.h"
#include "xmenu_wind.h"
#include "xclips.h"
#include "setup.h"
#include "defrule.h"
#include "deffacts.h"
#include "deffnctn.h"
#include "genrcfun.h"
#include "defins.h"
#include "object.h"
#include "engine.h"
#include "deftempl.h"

/********** local functions visible outside this file **********/
void DefruleManagerCallback();
void DeffactManagerCallback();
void DeftemplateManagerCallback();
void DeffunctionManagerCallback();
void DefgenericManagerCallback();
void DefinstancesManagerCallback();
void DefclassManagerCallback();
void AgendaManagerCallback();
void FactsWindowCallback();
void AgendaWindowCallback();
void InstancesWindowCallback();
void GlobalsWindowCallback();
void CommandLineCLIPSCallback();
void ColorUtilityCallback();
void CancelSelectPrimary();

/********** local functions not visible outside this file **********/
static void AgendaFire();
static void AgendaManagerSelect();
static void AgendaRemove();
static void CancelSelectSecondary();
static void CancelRulemngrSelect();
static void DeffactsManagerSelect();
static void DeffactsPprint();
static void DeffactsRemove();
static void DeftemplateManagerSelect();
static void DeftemplatePprint();
static void DeftemplateRemove();
static void Restart();
static void RuleManagerSelect();
static void DefruleMatches();
static void DefrulePprint();
static void DefruleRemove();
static void DefruleRemoveBreak();
static void DefruleSetBreak();
static void DefruleRefresh();
static void DeffunctionRemove();
static void DeffunctionPprint();
static void DefgenericRemove();
static void DefgenericPprint();
static void DefgenericMethod();
static void RemoveDefmethod();
static void PprintDefmethod();
static void DefinstancesRemove();
static void DefinstancesPprint();
static void DefclassRemove();
static void DefclassDescribe();
static void DefclassBrowse();
static void DefclassPprint();
static void DefclassMessageHandlers();
static void PprintMessageHandler();
static void RemoveMessageHandler();
/********** from CLIPS code **********/
extern int SetCommandString();
extern int PrintCLIPS();
extern int XclipsExit();
extern int XclipsQuery();
extern int XclipsPrint();
extern VOID *FindDefgeneric();
extern VOID *FindDefclass();
extern unsigned GetNextDefmethod();
extern VOID GetDefmethodDescription();
extern char *GetDefmessageHandlerName();
extern char *GetDefmessageHandlerType();
extern unsigned GetNextDefmessageHandler();
extern char *itoa();

/********** Global variables **********/
extern Widget toplevel, dialog, facts, agenda,instances, globals;
extern Arg args[10];
extern XEvent event;     /* Current event */
extern Boolean quit_get_event;
extern Pixmap checker;
extern Boolean Browse_status[];
extern Widget dialog_text;
extern Widget agenda_form, agenda_text,
              facts_form, facts_text,
              instances_form, instances_text,
              globals_form, globals_text;

extern int EvaluatingTopLevelCommand;

/********** local variables **********/
Boolean list_change = False;
Boolean list1_change = False;
static Widget manager_list, manager_list1;
static Boolean defrulemanager_flag = False;
static Boolean deffactsmanager_flag = False;
static Boolean deftemplatemanager_flag = False;
static Boolean deffunctionmanager_flag = False;
static Boolean defgenericmanager_flag = False;
static Boolean defmethodmanager_flag = False;
static Boolean definstancesmanager_flag = False;
static Boolean defclassmanager_flag = False;
static Boolean agendamanager_flag = False;
static Boolean defmessagehandler_flag = False;
Widget facts_window, agenda_window,instances_window, globals_window;
static String curr_def_name;
String *item_list;
String *item_list1;


/*******************************************************************************
          Name:        DefruleManagerCallback
          Description: Pop up the Rule Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DefruleManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
   static Widget defrulemanager = NULL;
   Widget defrulemanager_form,defrulemanager_viewport,
          remove = NULL, refresh = NULL, matches = NULL, pprint = NULL, 
          set_break = NULL, remove_break = NULL, cancel = NULL;
   

  GetDefruleList();

  if(item_list == NULL)
    {
        ClearParameters();
        defrulemanager_flag = False;
        return;
     }

  defrulemanager_flag = True;
 
  defrulemanager = XtCreatePopupShell("Defrule Manager",
                                      topLevelShellWidgetClass,
                                      toplevel,
                                      NULL, 0);

  defrulemanager_form = XtCreateManagedWidget("manager_form",
                                              formWidgetClass,
                                              defrulemanager,
                                              NULL, 0);

  defrulemanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                  viewportWidgetClass,
                                                  defrulemanager_form,
                                                  NULL, 0);
  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       defrulemanager_viewport,
                                       args, 1);

  XtSetArg(args[0], XtNfromHoriz, defrulemanager_viewport);

  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defrulemanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DefruleRemove, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Refresh");
  refresh = XtCreateManagedWidget("managerButton",
                                  commandWidgetClass,
                                  defrulemanager_form,
                                  args, 3);
  XtAddCallback(refresh, XtNcallback, DefruleRefresh, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, refresh);
  XtSetArg(args[2], XtNlabel, "Matches");
  matches = XtCreateManagedWidget("managerButton",
                                  commandWidgetClass,
                                  defrulemanager_form,
                                  args, 3);
  XtAddCallback(matches, XtNcallback, DefruleMatches, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, matches);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defrulemanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DefrulePprint, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Set break");
  set_break = XtCreateManagedWidget("managerButton",
                                    commandWidgetClass,
                                    defrulemanager_form,
                                    args, 3);
  XtAddCallback(set_break, XtNcallback, DefruleSetBreak,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, set_break);
  XtSetArg(args[2], XtNlabel, "Remove break");
  remove_break = XtCreateManagedWidget("managerButton",
                                       commandWidgetClass,
                                       defrulemanager_form,
                                       args, 3);
  XtAddCallback(remove_break, XtNcallback, DefruleRemoveBreak,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove_break);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 defrulemanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary, (XtPointer) defrulemanager);

  XtPopup(defrulemanager, XtGrabNonexclusive);

  }

/*******************************************************************************
          Name:        DeffactManagerCallback
          Description: Pop up the Deffacts Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DeffactManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget deffactsmanager,deffactsmanager_form, deffactsmanager_viewport,
         remove, pprint, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetFactList();

  if(item_list == NULL)
    {
     ClearParameters();
     deffactsmanager_flag = False;
     return;
    }

  deffactsmanager_flag = True;


  deffactsmanager = XtCreatePopupShell("Deffacts Manager",
                                       topLevelShellWidgetClass,
                                       toplevel,
                                       NULL, 0);

  deffactsmanager_form = XtCreateManagedWidget("manager_form",
                                               formWidgetClass,
                                               deffactsmanager,
                                               NULL, 0);

  deffactsmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                         viewportWidgetClass,
                                                         deffactsmanager_form,
                                                         NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                        listWidgetClass,
                                        deffactsmanager_viewport,
                                        args, 1);
  XtSetArg(args[0], XtNfromHoriz, deffactsmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deffactsmanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DeffactsRemove, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deffactsmanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DeffactsPprint, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 deffactsmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary, (XtPointer) deffactsmanager);

  XtPopup(deffactsmanager, XtGrabNonexclusive);
 
  }

/*******************************************************************************
          Name:        DeftemplateManagerCallback
          Description: Pop up the Deftemplate Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DeftemplateManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget deftemplatemanager, deftemplatemanager_form,
         deftemplatemanager_viewport, remove, pprint, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetDeftemplateList();

  if(item_list == NULL)
    {
    deftemplatemanager_flag = False;
    return;
    }

  deftemplatemanager_flag = True;

  deftemplatemanager = XtCreatePopupShell("Deftemplate Manager",
                                          topLevelShellWidgetClass,
                                          toplevel,
                                          NULL, 0);

  deftemplatemanager_form = XtCreateManagedWidget("manager_form",
                                                  formWidgetClass,
                                                  deftemplatemanager,
                                                  NULL, 0);

  deftemplatemanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                      viewportWidgetClass,
                                                      deftemplatemanager_form,
                                                      NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       deftemplatemanager_viewport,
                                       args, 1);
  XtSetArg(args[0], XtNfromHoriz, deftemplatemanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deftemplatemanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DeftemplateRemove,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deftemplatemanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DeftemplatePprint,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 deftemplatemanager_form,
                                 args, 3);
  XtAddCallback(cancel,XtNcallback,CancelSelectPrimary, (XtPointer)deftemplatemanager);

  XtPopup(deftemplatemanager, XtGrabNonexclusive);
  }

/*******************************************************************************
          Name:        DeffunctionManagerCallback
          Description: Pops up the Deffunction Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DeffunctionManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget deffunctionmanager, deffunctionmanager_form,
         deffunctionmanager_viewport, remove, pprint, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetDeffunctionList();

  if(item_list == NULL)
    {
      ClearParameters();
      deffunctionmanager_flag = False;
      return;
    }

  deffunctionmanager_flag = True;

  deffunctionmanager = XtCreatePopupShell("Deffunction Manager",
                                          topLevelShellWidgetClass,
                                          toplevel,
                                          NULL, 0);

  deffunctionmanager_form = XtCreateManagedWidget("manager_form",
                                                  formWidgetClass,
                                                  deffunctionmanager,
                                                  NULL, 0);

  deffunctionmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                      viewportWidgetClass,
                                                      deffunctionmanager_form,
                                                      NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       deffunctionmanager_viewport,
                                       args, 1);
  XtSetArg(args[0], XtNfromHoriz, deffunctionmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deffunctionmanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DeffunctionRemove,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 deffunctionmanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DeffunctionPprint,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 deffunctionmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary,
                (XtPointer)deffunctionmanager);

  XtPopup(deffunctionmanager, XtGrabNonexclusive);
  }

/*******************************************************************************
          Name:        DefgenericManagerCallback
          Description: Pops up the Defgeneric Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DefgenericManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget defgenericmanager, defgenericmanager_form, defgenericmanager_viewport,
         remove, pprint, methods, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetDefgenericList();

  if(item_list == NULL)
    {
    ClearParameters();
    defgenericmanager_flag = False;
    release(curr_def_name);
    return;
    }

  defgenericmanager_flag = True;

  defgenericmanager = XtCreatePopupShell("Defgeneric Manager",
                                         topLevelShellWidgetClass,
                                         toplevel,
                                         NULL, 0);

  defgenericmanager_form = XtCreateManagedWidget("manager_form",
                                                 formWidgetClass,
                                                 defgenericmanager,
                                                 NULL, 0);

  defgenericmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                     viewportWidgetClass,
                                                     defgenericmanager_form,
                                                     NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       defgenericmanager_viewport, 
                                       args, 1);
  XtSetArg(args[0], XtNfromHoriz, defgenericmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defgenericmanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DefgenericRemove, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defgenericmanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DefgenericPprint, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Methods...");
  methods = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defgenericmanager_form,
                                 args, 3);
  XtAddCallback(methods,XtNcallback,DefgenericMethod, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, methods);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 defgenericmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary,(XtPointer)defgenericmanager);

  XtPopup(defgenericmanager, XtGrabNonexclusive);
  }


/*******************************************************************************
          Name:        DefinstancesManagerCallback
          Description: Pops up the Definstances Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DefinstancesManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget definstancesmanager, definstancesmanager_form,
         definstancesmanager_viewport, remove, pprint, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetDefinstancesList();

  if(item_list == NULL)
    {
    ClearParameters();
    definstancesmanager_flag = False;
    return;
    }

  definstancesmanager_flag = True;

  definstancesmanager = XtCreatePopupShell("Definstances Manager",
                                           topLevelShellWidgetClass,
                                           toplevel,
                                           NULL, 0);

  definstancesmanager_form = XtCreateManagedWidget("manager_form",
                                                   formWidgetClass,
                                                   definstancesmanager,
                                                   NULL, 0);

  definstancesmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                       viewportWidgetClass,
                                                       definstancesmanager_form,
                                                       NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       definstancesmanager_viewport,
                                       args, 1);
  XtSetArg(args[0], XtNfromHoriz, definstancesmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 definstancesmanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DefinstancesRemove,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 definstancesmanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DefinstancesPprint,
                (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 definstancesmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary,
                (XtPointer)definstancesmanager);

  XtPopup(definstancesmanager, XtGrabNonexclusive);
  }

/*******************************************************************************
          Name:        DefclassManagerCallback
          Description: Pops up the Defclass Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void DefclassManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget defclassmanager, defclassmanager_form, defclassmanager_viewport,
         remove,  describe, browse, pprint, message_handlers, cancel;
 
/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetDefclassList();

  if(item_list == NULL)
    {
    ClearParameters();
    release(curr_def_name);
    defclassmanager_flag = False;
    return;
    }

  
  defclassmanager_flag = True;

  defclassmanager = XtCreatePopupShell("Defclass Manager",
                                       topLevelShellWidgetClass,
                                       toplevel,
                                       NULL, 0);

  defclassmanager_form = XtCreateManagedWidget("manager_form",
                                               formWidgetClass,
                                               defclassmanager,
                                               NULL, 0);

  defclassmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                   viewportWidgetClass,
                                                   defclassmanager_form,
                                                   NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       defclassmanager_viewport,
                                       args, 1);
  XtSetArg(args[0], XtNfromHoriz, defclassmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defclassmanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, DefclassRemove, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Describe");
  describe = XtCreateManagedWidget("managerButton",
                                   commandWidgetClass,
                                   defclassmanager_form,
                                   args, 3);
  XtAddCallback(describe, XtNcallback, DefclassDescribe, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, describe);
  XtSetArg(args[2], XtNlabel, "Browse");
  browse  = XtCreateManagedWidget("managerButton",
                                   commandWidgetClass,
                                   defclassmanager_form,
                                   args, 3);
  XtAddCallback(browse, XtNcallback, DefclassBrowse, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, browse);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defclassmanager_form,
                                 args, 3);
  XtAddCallback(pprint, XtNcallback, DefclassPprint, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Message Handlers...");
  message_handlers = XtCreateManagedWidget("managerButton",
                                           commandWidgetClass,
                                           defclassmanager_form,
                                           args, 3);
  XtAddCallback(message_handlers,XtNcallback, DefclassMessageHandlers,(XtPointer)manager_list);
  XtSetArg(args[1], XtNfromVert, message_handlers);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 defclassmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary, (XtPointer)defclassmanager);

  XtPopup(defclassmanager, XtGrabNonexclusive);
  }


/*******************************************************************************
          Name:        AgendaManagerCallback
          Description: Pop up the Agenda Manager
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void AgendaManagerCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget agendamanager, agendamanager_form, agendamanager_viewport, remove,
         fire, cancel;

/*  if(EvaluatingTopLevelCommand)
   return;
*/
  GetAgendaList();

  if(item_list == NULL)
    {
    ClearParameters();
    agendamanager_flag = False;
    return;
    }

  agendamanager_flag = True;

  agendamanager = XtCreatePopupShell("Agenda Manager",
                                     topLevelShellWidgetClass,
                                     toplevel,
                                     NULL, 0);

  agendamanager_form = XtCreateManagedWidget("manager_form",
                                             formWidgetClass,
                                             agendamanager,
                                             NULL, 0);

  agendamanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                 viewportWidgetClass,
                                                 agendamanager_form,
                                                 NULL, 0);

  XtSetArg(args[0], XtNlist, item_list);
  manager_list = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                       agendamanager_viewport,
                                       args, 1);

  XtSetArg(args[0], XtNfromHoriz, agendamanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 agendamanager_form,
                                 args, 2);
  XtAddCallback(remove, XtNcallback, AgendaRemove, (XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Fire");
  fire = XtCreateManagedWidget("managerButton",
                               commandWidgetClass,
                               agendamanager_form,
                               args, 3);
  XtAddCallback(fire, XtNcallback, AgendaFire,(XtPointer)manager_list);

  XtSetArg(args[1], XtNfromVert, fire);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 agendamanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectPrimary, (XtPointer) agendamanager);

  XtPopup(agendamanager,  XtGrabNonexclusive);
  }

/*******************************************************************************
          Name:        GetAgendaList
          Description: Gets agenda list
          Arguments:  None
          Returns:
*******************************************************************************/
GetAgendaList()
  {
  int maxItems = 20,itemCount;
  struct activation *act_ptr;

  if((act_ptr = (struct activation *) GetNextActivation(NULL)) == NULL)
    {
      item_list = NULL;
      return(0);
    }
  item_list = (String *)calloc(maxItems,sizeof(String));
  for(itemCount = 0; act_ptr != NULL;act_ptr = act_ptr->next)
      {
      item_list[itemCount] = balloc(strlen(act_ptr->theRule->name->contents) + 1, char);
      strcpy(item_list[itemCount++],act_ptr->theRule->name->contents);
      if(itemCount == (maxItems - 1))
        {
          maxItems = maxItems * 2;
          item_list = (String *)realloc(item_list,maxItems*sizeof(String));
        }
      }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  return(itemCount);
  }

/******************************************************************************
          Name:        AgendaRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void AgendaRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  register int i;
  struct activation *act_ptr;
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  act_ptr = (struct activation *) GetNextActivation(NULL);
  for (i = 0; i < current->list_index ; i++)
    act_ptr = act_ptr->next;
  DeleteActivation((VOID*)act_ptr);
  quit_get_event = True;
  list_change = True;
  }

/******************************************************************************
          Name:        AgendaFire
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void AgendaFire(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  register int i;
  struct activation *act_ptr;
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  act_ptr = (struct activation *) GetNextActivation(NULL);
  for (i = 0; i < current->list_index ; i++)
    act_ptr = act_ptr->next;
  MoveActivationToTop((VOID*)act_ptr);
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(run 1)\n");
  PrintCLIPS("wclips","(run 1)\n");
  quit_get_event = True;
  list_change = True;
  }


/*******************************************************************************
          Name:        FactsWindowCallback
          Description: Called when Facts Window is selected from the Window menu
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void FactsWindowCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {

  if (Browse_status[FACT_WIN])
    {
    XtSetArg(args[0], XtNleftBitmap, None);
    XtPopdown(facts);
    }
  else if(facts_text != NULL)
   {
    XtPopup(facts,XtGrabNone);
    SetFactListChanged(CLIPS_FALSE);
    PrintChangedFacts();
    XtSetArg(args[0], XtNleftBitmap, checker);
   }
  else
   {
    CreateFactWindow();
    SetFactListChanged(CLIPS_FALSE);
    PrintChangedFacts();
    XtSetArg(args[0], XtNleftBitmap, checker);
    }

  XtSetValues(facts_window, args, 1);
  Browse_status[FACT_WIN] = !Browse_status[FACT_WIN];
  }

/**********************************************************************
 *    CreateFactWindow
 *
 **********************************************************************/

CreateFactWindow()
{
    Dimension height;
    int n = 0;

    XtSetArg(args[n], XtNheight, &height);n++;
    XtGetValues(dialog, args, n);
    height = (height + 150) / 3;
    n = 0;
    XtSetArg(args[n], XtNheight, height);n++;
    facts = XtCreatePopupShell("Facts Window",
                               topLevelShellWidgetClass,
                               toplevel,
                               args, n);
    n = 0;
    XtSetArg(args[n], XtNdefaultDistance, 0);n++;
    facts_form = XtCreateManagedWidget("facts_form",
                                       formWidgetClass,
                                       facts,
                                       args, n);
    n = 0;
    XtSetArg(args[n], XtNwidth, 250);n++;
    XtSetArg(args[n], XtNeditType, XawtextAppend);n++;
    XtSetArg(args[n], XtNscrollHorizontal, XawtextScrollAlways);n++;
    XtSetArg(args[n], XtNscrollVertical, XawtextScrollAlways);n++;
    facts_text = XtCreateManagedWidget("facts_text",
                                       asciiTextWidgetClass,
                                       facts_form,
                                       args, n);

    XtOverrideTranslations(facts_text,
                           XtParseTranslationTable(xclips_translation2));

    XtPopup(facts, XtGrabNone);

    if(!AddRouter("xfacts", 10, XclipsQuery, XclipsPrint, NULL, NULL, XclipsExit))
      {
      PrintCLIPS("werror", "Could not allocate xclips router!\n");
      XclipsExit(0);
      ExitCLIPS(0);
      }

}
/*******************************************************************************
          Name:        AgendaWindowCallback
          Description: Called when Agenda Window is selected from the Window
                       menu
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void AgendaWindowCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
    Dimension width;
    extern int PrintChangedAgenda();

  if (Browse_status[AGENDA_WIN])
    {
     XtSetArg(args[0], XtNleftBitmap, None);
     XtPopdown(agenda);
    }
  else if(agenda != NULL)
   {
     XtPopup(agenda,XtGrabNone);
     SetAgendaChanged(CLIPS_FALSE);
     PrintChangedAgenda();
     XtSetArg(args[0], XtNleftBitmap,checker);
   }
  else
  {
    CreateAgendaWindow();
    SetAgendaChanged(CLIPS_FALSE);
    PrintChangedAgenda();
    XtSetArg(args[0], XtNleftBitmap, checker);
    }

  XtSetValues(agenda_window, args, 1);

  Browse_status[AGENDA_WIN] = !Browse_status[AGENDA_WIN];
  }
/*******************************************************************************
          Name:
          Description:
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
CreateAgendaWindow()
{
    Dimension width;

    XtSetArg(args[0], XtNwidth, &width);
    XtGetValues(dialog, args, 1);
    XtSetArg(args[0], XtNwidth, width);
    agenda = XtCreatePopupShell("Agenda Window",
                                topLevelShellWidgetClass,
                                toplevel,
                                args, 1);

    XtSetArg(args[0], XtNdefaultDistance, 0);
    agenda_form = XtCreateManagedWidget("agenda_form",
                                        formWidgetClass,
                                        agenda,
                                        args, 1);

    XtSetArg(args[0], XtNheight, 150);
    XtSetArg(args[1], XtNeditType, XawtextAppend);
    XtSetArg(args[2], XtNscrollHorizontal, XawtextScrollAlways);
    XtSetArg(args[3], XtNscrollVertical, XawtextScrollAlways);
    agenda_text = XtCreateManagedWidget("agenda_text",
                                        asciiTextWidgetClass,
                                        agenda_form,
                                        args, 4);
    XtOverrideTranslations(agenda_text,
                           XtParseTranslationTable(xclips_translation2));

    XtPopup(agenda, XtGrabNone);

    if(!AddRouter("xagenda", 10, XclipsQuery, XclipsPrint,NULL,NULL,XclipsExit))
      {
      PrintCLIPS("werror", "Could not allocate xclips router!\n");
      XclipsExit(0);
      ExitCLIPS(0);
      }
}

/*******************************************************************************
          Name:        
          Description: 
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void InstancesWindowCallback(w,client_data,call_data)
Widget w;
XtPointer call_data, client_data;
{
  if (Browse_status[INSTANCE_WIN])
    {
    XtSetArg(args[0], XtNleftBitmap, None);
    XtPopdown(instances);
    }
  else if(instances != NULL)
   {
     XtPopup(instances,XtGrabNone);
     SetInstancesChanged(CLIPS_FALSE);
     PrintChangedInstances();
     XtSetArg(args[0], XtNleftBitmap,checker);
   }
  else
   {
    CreateInstanceWindow();
    SetInstancesChanged(CLIPS_FALSE);
    PrintChangedInstances();
    XtSetArg(args[0], XtNleftBitmap, checker);
    }
  XtSetValues(instances_window, args, 1);
  Browse_status[INSTANCE_WIN] = !Browse_status[INSTANCE_WIN];
}

/**********************************************************************************
 *    CreateInstanceWindow
 **********************************************************************************/
CreateInstanceWindow()
{
   Dimension height;

    XtSetArg(args[0], XtNheight, &height);
    XtGetValues(dialog, args, 1);
    height = (height + 150)/3;
    XtSetArg(args[0], XtNheight, height);
    instances = XtCreatePopupShell("Instances Window",
                               topLevelShellWidgetClass,
                               toplevel,
                               args, 1);

    XtSetArg(args[0], XtNdefaultDistance, 0);
    instances_form = XtCreateManagedWidget("instances_form",
                                       formWidgetClass,
                                       instances,
                                       args, 1);

    XtSetArg(args[0], XtNwidth, 250);
    XtSetArg(args[1], XtNeditType, XawtextAppend);
    XtSetArg(args[2], XtNscrollHorizontal,XawtextScrollAlways);
    XtSetArg(args[3], XtNscrollVertical, XawtextScrollAlways);
    instances_text = XtCreateManagedWidget("instances_text",
                                       asciiTextWidgetClass,
                                       instances_form,
                                       args, 4);

    XtOverrideTranslations(instances_text,
                           XtParseTranslationTable(xclips_translation2));

    XtPopup(instances, XtGrabNone);

    if(!AddRouter("xinstances", 10, XclipsQuery, XclipsPrint, NULL, NULL, XclipsExit))
      {
      PrintCLIPS("werror", "Could not allocate xclips router!\n");
      XclipsExit(0);
      ExitCLIPS(0);
      }
}

/*******************************************************************************
          Name:
          Description:
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void GlobalsWindowCallback(w,client_data,call_data)
Widget w;
XtPointer call_data, client_data;
{

  if (Browse_status[GLOBAL_WIN])
    {
    XtPopdown(globals);
    XtSetArg(args[0], XtNleftBitmap, None);
    }
  else if(globals != NULL)
   {
     XtPopup(globals,XtGrabNone);
     SetGlobalsChanged(CLIPS_FALSE);
     PrintChangedGlobals();
     XtSetArg(args[0], XtNleftBitmap,checker);
   }
  else
    {
     CreateGlobalWindow();
     SetGlobalsChanged(CLIPS_FALSE);
     PrintChangedGlobals();
     XtSetArg(args[0], XtNleftBitmap, checker);
    }

  XtSetValues(globals_window, args, 1);

  Browse_status[GLOBAL_WIN] = !Browse_status[GLOBAL_WIN];

}
/**********************************************************************************
 *    CreateGlobalWindow
 **********************************************************************************/

CreateGlobalWindow()
{
    
    Dimension height;

    XtSetArg(args[0], XtNheight, &height);
    XtGetValues(dialog, args, 1);
    height = (height + 150)/3;
    XtSetArg(args[0], XtNheight, height);
    globals = XtCreatePopupShell("Globals Window",
                               topLevelShellWidgetClass,
                               toplevel,
                               args, 1);

    XtSetArg(args[0], XtNdefaultDistance, 0);
    globals_form = XtCreateManagedWidget("globals_form",
                                       formWidgetClass,
                                       globals,
                                       args, 1);

    XtSetArg(args[0], XtNwidth, 250);
    XtSetArg(args[1], XtNeditType, XawtextAppend);
    XtSetArg(args[2], XtNscrollHorizontal,XawtextScrollAlways);
    XtSetArg(args[3], XtNscrollVertical, XawtextScrollAlways);
    globals_text = XtCreateManagedWidget("globals_text",
                                       asciiTextWidgetClass,
                                       globals_form,
                                       args, 4);

    XtOverrideTranslations(globals_text,
                           XtParseTranslationTable(xclips_translation2));

    XtPopup(globals, XtGrabNone);

    if(!AddRouter("xglobals", 10, XclipsQuery, XclipsPrint, NULL, NULL, XclipsExit))
      {
      PrintCLIPS("werror", "Could not allocate xclips router!\n");
      XclipsExit(0);
      ExitCLIPS(0);
      }
}

/*******************************************************************************
          Name:
          Description:
          Arguments:
          Returns:
*******************************************************************************/
void AllWindowsCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
    int n = 0;
    if(!Browse_status[FACT_WIN])
     {
       if(facts != NULL)
         XtPopup(facts,XtGrabNone);
       else
         CreateFactWindow();
       XtSetArg(args[n],XtNleftBitmap,checker);n++;
       XtSetValues(facts_window,args,n);       
       Browse_status[FACT_WIN] = !Browse_status[FACT_WIN];
     }
    if(!Browse_status[AGENDA_WIN])
     {
       if(agenda != NULL)
        XtPopup(agenda,XtGrabNone);
       else
        CreateAgendaWindow();
       XtSetArg(args[n],XtNleftBitmap,checker);n++;
       XtSetValues(agenda_window,args,n);
       Browse_status[AGENDA_WIN] = !Browse_status[AGENDA_WIN];
     }
    if(!Browse_status[INSTANCE_WIN])
     {
       
       if(instances != NULL)
        XtPopup(instances,XtGrabNone);
       else
        CreateInstanceWindow();
       XtSetArg(args[n],XtNleftBitmap,checker);n++;
       XtSetValues(instances_window,args,n);
       Browse_status[INSTANCE_WIN] = !Browse_status[INSTANCE_WIN];
     }
    if(!Browse_status[GLOBAL_WIN])
     {
       if(globals != NULL)
        XtPopup(globals,XtGrabNone);
       else
        CreateGlobalWindow();
       XtSetArg(args[n],XtNleftBitmap,checker);n++;
       XtSetValues(globals_window,args,n);
       Browse_status[GLOBAL_WIN]  = !Browse_status[GLOBAL_WIN];
    }
  }

/*******************************************************************************
          Name:
          Description:
          Arguments:
          Returns:
*******************************************************************************/
void NoWindowsCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
    int n = 0;

    XtSetArg(args[n], XtNleftBitmap, None);    n++;

    if (Browse_status[GLOBAL_WIN])
     {
       XtPopdown(globals);
       XtSetValues(globals_window,args,n);
       Browse_status[GLOBAL_WIN] = !Browse_status[GLOBAL_WIN];
     }
    if (Browse_status[INSTANCE_WIN])
     {
       XtPopdown(instances);
       XtSetValues(instances_window,args,n);
       Browse_status[INSTANCE_WIN] = !Browse_status[INSTANCE_WIN];
     }     
    if(Browse_status[FACT_WIN])
     {
       XtPopdown(facts);
       XtSetValues(facts_window,args,n);
       Browse_status[FACT_WIN] = !Browse_status[FACT_WIN]; 
     }
    if(Browse_status[AGENDA_WIN])
     {
       XtPopdown(agenda);
       XtSetValues(agenda_window,args,n);
       Browse_status[AGENDA_WIN] = !Browse_status[AGENDA_WIN];
     }
  }


/*******************************************************************************
          Name:        CommandLineCLIPSCallback
          Description: Called when Command Line CLIPS is selected from the
                       Window menu
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void CommandLineCLIPSCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  system("xterm -e clips &");
  }

/*******************************************************************************
          Name:        ColorUtilityCallback
          Description: Called when Color is selected from the Window menu
          Arguments:  w - menu item that was selected
                       client_data - entry number on menu
                       call_data - not used
          Returns:     None
*******************************************************************************/
void ColorUtilityCallback(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  system("color&");
  }

/*******************************************************************************
          Name:        GetDefruleList
          Description: Gets the list of rules
          Arguments:  None
          Returns:
*******************************************************************************/
GetDefruleList()
{
  struct defrule *rule_ptr;
  int maxItems = 20,itemCount;

  if((rule_ptr = (struct defrule *) GetNextDefrule(NULL)) == NULL)
   {
     item_list = NULL;
     return(0);
   }
  item_list = (String *)calloc(maxItems,sizeof(String));
  for(itemCount = 0;rule_ptr != NULL;rule_ptr = rule_ptr->next)
      {
         item_list[itemCount] = balloc(strlen(rule_ptr->name->contents) + 1,char);
         strcpy(item_list[itemCount], rule_ptr->name->contents);
         itemCount++;
         if(itemCount == (maxItems -1))
          {
            maxItems = 2*maxItems;
            item_list = (String *)realloc(item_list,maxItems * sizeof(String));
          } 
      }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  }

/******************************************************************************
          Name:        DefruleRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data - not used
          Returns:     None
*******************************************************************************/
static void DefruleRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);

  if(current->list_index == XAW_LIST_NONE)
    return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(excise ");
  PrintCLIPS("wclips","(excise ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }

/******************************************************************************
          Name:        DefruleMatches
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefruleMatches(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(matches ");
  PrintCLIPS("wclips","(matches ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefrulePprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefrulePprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(pprule ");
  PrintCLIPS("wclips","(pprule ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefruleSetBreak
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefruleSetBreak(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(set-break ");
  PrintCLIPS("wclips","(set-break ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefruleRemoveBreak
          Description: Calls CLIPS remove-break command
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefruleRemoveBreak(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(remove-break ");
  PrintCLIPS("wclips","(remove-break ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefruleRefresh
          Description: Calls CLIPS refresh command
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefruleRefresh(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(refresh ");
  PrintCLIPS("wclips","(refresh ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        GetFactList
          Description: Gets list of facts
          Arguments:  None
          Returns:
*******************************************************************************/
GetFactList()
{
  struct deffacts *fact_ptr;
  int maxItems = 20,itemCount;


  if((fact_ptr = (struct deffacts *) GetNextDeffacts(NULL)) == NULL)
    {
      item_list = NULL;
      return(0);
    }
  item_list = (String*)calloc(maxItems,sizeof(String));
  for(itemCount = 0; fact_ptr != NULL;fact_ptr = fact_ptr->next)
     {
      item_list[itemCount] = balloc(strlen(fact_ptr->name->contents) + 1,char);
      strcpy(item_list[itemCount],fact_ptr->name->contents);
      itemCount += 1;
      if(itemCount == (maxItems - 1))
       {
         maxItems = 2 * (maxItems);
         item_list = (String*)realloc(item_list,maxItems * sizeof(String));
       }
      }
    item_list[itemCount] = NULL;
    sortList(item_list,itemCount);
    return(itemCount);
  }

/******************************************************************************
          Name:        DeffactsRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeffactsRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undeffacts ");
  PrintCLIPS("wclips","(undeffacts ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }

/******************************************************************************
          Name:        DeffactsPprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeffactsPprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdeffacts ");
  PrintCLIPS("wclips","(ppdeffacts ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        GetDeftemplateList
          Description: Gets list of deftemplates
          Arguments:  None
          Returns:
*******************************************************************************/
GetDeftemplateList()
  {

  struct deftemplate *dtmpl_ptr;
  int itemCount,maxItems = 20;

  if((dtmpl_ptr = (struct deftemplate *) GetNextDeftemplate(NULL)) == NULL)
   {
     item_list = NULL;
     return(0);
   }
  item_list = (String *)calloc(maxItems,sizeof(String));
  for(itemCount = 0; dtmpl_ptr != NULL;dtmpl_ptr = dtmpl_ptr->next)
    {
      item_list[itemCount] = balloc(strlen(dtmpl_ptr->name->contents) + 1, char);
      strcpy(item_list[itemCount++],dtmpl_ptr->name->contents);
      if(itemCount == (maxItems - 1))
       {
         maxItems = 2 * maxItems;
         item_list = (String*)realloc(item_list,maxItems * sizeof(String));
       }
    }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  return(itemCount); 
  }

/******************************************************************************
          Name:        DeftemplateRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeftemplateRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undeftemplate ");
  PrintCLIPS("wclips","(undeftemplate ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }


/******************************************************************************
          Name:        DeftemplatePprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeftemplatePprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdeftemplate ");
  PrintCLIPS("wclips","(ppdeftemplate ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        GetDeffunctionList
          Description: Gets list of deffunctions
          Arguments:  None
          Returns:
*******************************************************************************/
GetDeffunctionList()
  {

  struct dfunc *dfunc_ptr;
  int itemCount,maxItems = 20;

  if((dfunc_ptr = (struct dfunc *) GetNextDeffunction(NULL)) == NULL)
    {
      item_list = NULL;
      return(0);
    }
  item_list = (String *)calloc(maxItems,sizeof(String));
  for(itemCount = 0; dfunc_ptr != NULL;dfunc_ptr = dfunc_ptr->next)
    {
      item_list[itemCount] = balloc(strlen(dfunc_ptr->name->contents) + 1, char);
      strcpy(item_list[itemCount++],dfunc_ptr->name->contents);
      if(itemCount == (maxItems - 1))
        {
          maxItems = maxItems * 2;
          item_list = (String *)realloc(item_list,maxItems * sizeof(String));
        }
     }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  return(itemCount);
  }

/******************************************************************************
          Name:        DeffunctionRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeffunctionRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undeffunction ");
  PrintCLIPS("wclips","(undeffunction ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }


/******************************************************************************
          Name:        DeffunctionPprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DeffunctionPprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);

  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdeffunction");
  PrintCLIPS("wclips","(ppdeffunction ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        GetDefgenericList
          Description: Gets list of defgenerics
          Arguments:  None
          Returns:
*******************************************************************************/
GetDefgenericList()
  {
  int maxItems = 20,itemCount;
  struct generic_func *generic_func_ptr;

  if((generic_func_ptr = (struct generic_func *) GetNextDefgeneric(NULL)) == NULL)
    {
       item_list  =  NULL;
       return(0);
    }
  item_list = (String *)calloc(maxItems,sizeof(String));
  for(itemCount = 0; generic_func_ptr != NULL;generic_func_ptr = generic_func_ptr->nxt)
   {
      item_list[itemCount] = balloc(strlen(generic_func_ptr->name->contents) + 1, char);
      strcpy(item_list[itemCount++],generic_func_ptr->name->contents);
      if(itemCount == (maxItems - 1))
       {
         maxItems = maxItems * 2;
         item_list = (String *)realloc(item_list,maxItems*sizeof(String));
       }
   }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  return(itemCount);
  }

/******************************************************************************
          Name:        DefgenericRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefgenericRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undefgeneric ");
  PrintCLIPS("wclips","(undefgeneric ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }


/******************************************************************************
          Name:        DefgenericPprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefgenericPprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdefgeneric ");
  PrintCLIPS("wclips","(ppdefgeneric ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefgenericMethod
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefgenericMethod(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  
  
  Widget defmethodmanager,  defmethodmanager_form,  defmethodmanager_viewport,
         remove, pprint, cancel;
  char title[274];
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);
   
 
  if(current->list_index == XAW_LIST_NONE)
      return;
  GetDefmethodList(current->string);
  if(item_list1 == NULL)
    {
      defmethodmanager_flag = False;
      return;
    }
   curr_def_name = balloc(strlen(current->string) +  1,char);
   strcpy(curr_def_name,current->string);
   defmethodmanager_flag = True;
   title[0] = 0;
   strcpy(title,curr_def_name);
   strcat(title," Defmethod Manager");  
   defmethodmanager = XtCreatePopupShell(title,
                                         topLevelShellWidgetClass,
                                         toplevel,
                                         NULL, 0);

   defmethodmanager_form = XtCreateManagedWidget("manager_form",
                                                 formWidgetClass,
                                                  defmethodmanager,
                                                 NULL, 0);

   defmethodmanager_viewport = XtCreateManagedWidget("manager_viewport",
                                                     viewportWidgetClass,
                                                      defmethodmanager_form,
                                                     NULL, 0);

   XtSetArg(args[0], XtNlist, item_list1);
   manager_list1 = XtCreateManagedWidget("manager_list",
                                       listWidgetClass,
                                        defmethodmanager_viewport,
                                       args, 1);

  XtSetArg(args[0], XtNfromHoriz,  defmethodmanager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                  defmethodmanager_form,
                                 args, 2);
  XtAddCallback(remove,XtNcallback,RemoveDefmethod,(XtPointer)manager_list1);  

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                  defmethodmanager_form,
                                 args, 3);
  XtAddCallback(pprint,XtNcallback,PprintDefmethod,(XtPointer)manager_list1);
 
  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                  defmethodmanager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectSecondary,
                (XtPointer) defmethodmanager);

  XtPopup(defmethodmanager, XtGrabNonexclusive);

  }

/******************************************************************************
          Name:        RemoveDefmethod
          Description: removes aa method from the method list
          Arguments:  w - not used
                       client_data - not used
                       call_data - not used
          Returns:     None
*******************************************************************************/
static void RemoveDefmethod(w,client_data,call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
    char index[5];
    int i;
    Widget aList = (Widget)client_data;
    XawListReturnStruct *current = XawListShowCurrent(aList);

    if(current->list_index == XAW_LIST_NONE)
      return;
    for (i = 0; ('0' <= current->string[i]) && (current->string[i] <= '9');i++)
      index[i] = current->string[i];

    index[i] = 0;
    MoveEndOfFile(dialog_text, &event);
    SetCommandString("(undefmethod  ");
    PrintCLIPS("wclips","(undefmethod ");
    AppendCommandString(curr_def_name);
    PrintCLIPS("wclips",curr_def_name);
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(index);
    PrintCLIPS("wclips",index);
    AppendCommandString(")\n");
    PrintCLIPS("wclips",")\n");
    quit_get_event = True;
    list1_change = True;

  }
/******************************************************************************
          Name:        PprintDefmethod
          Description: Print the method
          Arguments:  w - not used
                       client_data - the list widget
                       call_data - not used
          Returns:     None
*******************************************************************************/
static void PprintDefmethod(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;

{
    char index[5];
    int i;
    Widget aList = (Widget)client_data;
    XawListReturnStruct *current = XawListShowCurrent(aList);

    if(current->list_index == XAW_LIST_NONE)
      return;
    for (i = 0; ('0' <= current->string[i]) && (current->string[i] <= '9');i++)
      index[i] = current->string[i];

    index[i] = 0;
    MoveEndOfFile(dialog_text, &event);
    AppendCommandString("(ppdefmethod ");
    PrintCLIPS("wclips","(ppdefmethod ");
    AppendCommandString(curr_def_name);
    PrintCLIPS("wclips",curr_def_name);
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(index);
    PrintCLIPS("wclips",index);
    AppendCommandString(")\n");
    PrintCLIPS("wclips",")\n");
    quit_get_event = True;

}


/*******************************************************************************
          Name:        GetDefmethodList
          Description: Gets list of defmethods
          Arguments:  None
          Returns:
*******************************************************************************/

GetDefmethodList(Aname)
char *Aname;
  {
  VOID *genrc_ptr;
  unsigned index;
  char buf[61];
  register int itemCount = 0;
  int maxItems = 20;

  genrc_ptr = FindDefgeneric(Aname);
  if(item_list1 != NULL)
   free(item_list1);
  if((index = GetNextDefmethod(genrc_ptr,0)) == 0)
   {
     item_list1 = NULL;
     return(0);
   }
  item_list1 = (String*) calloc(maxItems,sizeof(String));
  while (index != 0)
   {
     GetDefmethodDescription(buf,60,genrc_ptr,index);
     item_list1[itemCount] = balloc(strlen(buf) + 1,char);
     strcpy(item_list1[itemCount++],buf);
     if(itemCount == (maxItems - 1))
      {
        maxItems = 2 * maxItems;
        item_list1 = (String*)realloc(item_list1,maxItems*sizeof(String));
      }
     index = GetNextDefmethod(genrc_ptr,index);
   }
  item_list1[itemCount] = NULL;
  sortList(item_list1,itemCount);
  return(itemCount);
  }

/*******************************************************************************
          Name:        GetDefinstancesList
          Description: Gets list of definstances
          Arguments:  None
          Returns:
*******************************************************************************/
GetDefinstancesList()
{
  int maxItems = 20, itemCount;
  struct definstance *definstance_ptr;


  if((definstance_ptr = (struct definstance *)  GetNextDefinstances(NULL)) == NULL)
    {
      item_list = NULL;
      return(0);
    }
  item_list = (String*)calloc(maxItems,sizeof(String));
  for( itemCount = 0; definstance_ptr != NULL;definstance_ptr = definstance_ptr->nxt)
   {
      item_list[itemCount] = balloc(strlen(definstance_ptr->name->contents) + 1, char);
      strcpy(item_list[itemCount++],definstance_ptr->name->contents);
      if(itemCount == (maxItems - 1))
       {
         maxItems = maxItems * 2;
         item_list = (String*)realloc(item_list,maxItems * sizeof(String));
       }
   }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);
  return(itemCount);
}

/******************************************************************************
          Name:        DefinstancesRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefinstancesRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undefinstances ");
  PrintCLIPS("wclips","(undefinstances ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;
  }


/******************************************************************************
          Name:        DefinstancesPprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefinstancesPprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdefinstances ");
  PrintCLIPS("wclips","(ppdefinstances ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/*******************************************************************************
          Name:        GetDefclassList
          Description: Gets list of defclasses
          Arguments:  None
          Returns:
*******************************************************************************/
GetDefclassList()
{
  int maxItems = 20, itemCount;
  struct cls *cls_ptr;

  if((cls_ptr = (struct cls *) GetNextDefclass(NULL)) == NULL)
    {
      item_list = NULL;
      return(0);
    }
  item_list = (String*)calloc(maxItems,sizeof(String));
  for( itemCount = 0; cls_ptr != NULL;cls_ptr = cls_ptr->nxt_lst)
    {
      item_list[itemCount] = balloc(strlen(cls_ptr->name->contents) + 1, char);
      strcpy(item_list[itemCount++], cls_ptr->name->contents);
      if(itemCount == (maxItems - 1))
       {
         maxItems = maxItems * 2;
         item_list = (String*)realloc(item_list,maxItems * sizeof(String));
       }
    }
  item_list[itemCount] = NULL;
  sortList(item_list,itemCount);  
}

/******************************************************************************
          Name:        DefclassRemove
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefclassRemove(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(undefclass ");
  PrintCLIPS("wclips","(undefclass ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  list_change = True;

  }

/******************************************************************************
          Name:        DefclassDescribe
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefclassDescribe(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(describe-class ");
  PrintCLIPS("wclips","(describe-class ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefclassBrowse
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefclassBrowse(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(browse-classes ");
  PrintCLIPS("wclips","(browse-classes ");
  AppendCommandString(current->string);
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }


/******************************************************************************
          Name:        DefclassPprint
          Description: Calls CLIPS
          Arguments:  w - not used
                       client_data - not used
                       call_data
          Returns:     None
*******************************************************************************/
static void DefclassPprint(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);


  if(current->list_index == XAW_LIST_NONE)
      return;
  MoveEndOfFile(dialog_text, &event);
  SetCommandString("(ppdefclass ");
  PrintCLIPS("wclips","(ppdefclass ");
  AppendCommandString(current->string);	
  PrintCLIPS("wclips",current->string);
  AppendCommandString(")\n");
  PrintCLIPS("wclips",")\n");
  quit_get_event = True;
  }

/******************************************************************************
          Name:        DefclassMessageHandlers
          Description: Displays a dialog box allowing the message handlers for 
                       the currently selected class to  be browsed.
          Arguments:  w - not used
                       client_data - not used
                       call_data - the list widget containing the list of defclass
          Returns:     None
*******************************************************************************/
static void DefclassMessageHandlers(w,client_data, call_data)
 Widget w;
  XtPointer client_data, call_data;
  {
  char title[284];
  Widget defmessHdlrManager, defmessHdlrManager_form, defmessHdlrManager_viewport,
         remove, pprint, cancel;
 
  Widget list_widget = (Widget)client_data;
  XawListReturnStruct *current = XawListShowCurrent(list_widget);
 

  if(current->list_index == XAW_LIST_NONE)
      return;
  GetDefmessageHandlers(current->string);
  if(item_list1 == NULL)
   {
     defmessagehandler_flag = False;
     return;
   }
  curr_def_name = balloc(strlen(current->string) + 1,char);
  strcpy(curr_def_name,current->string);
  title[0] = 0;
  strcpy(title,curr_def_name);
  strcat(title," Defmessage-handler manager");
  defmessagehandler_flag = True;
  defmessHdlrManager =  XtCreatePopupShell(title,
                                         topLevelShellWidgetClass,
                                         toplevel,
                                         NULL, 0);
  defmessHdlrManager_form =  XtCreateManagedWidget("manager_form",
                                                 formWidgetClass,
                                                 defmessHdlrManager,
                                                 NULL, 0);
  defmessHdlrManager_viewport = XtCreateManagedWidget("manager_viewport",
                                                     viewportWidgetClass,
                                                     defmessHdlrManager_form,
                                                     NULL, 0);
   XtSetArg(args[0], XtNlist, item_list1);
   manager_list1 = XtCreateManagedWidget("manager_list",
                                         listWidgetClass,
                                         defmessHdlrManager_viewport,
                                         args, 1);
  XtSetArg(args[0], XtNfromHoriz,  defmessHdlrManager_viewport);
  XtSetArg(args[1], XtNlabel, "Remove");
  remove = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defmessHdlrManager_form,
                                 args, 2);
  XtAddCallback(remove,XtNcallback,RemoveMessageHandler,(XtPointer)manager_list1); 

  XtSetArg(args[1], XtNfromVert, remove);
  XtSetArg(args[2], XtNlabel, "Pprint");
  pprint = XtCreateManagedWidget("managerButton",
                                 commandWidgetClass,
                                 defmessHdlrManager_form,
                                 args, 3);
  XtAddCallback(pprint,XtNcallback,PprintMessageHandler,(XtPointer)manager_list1); 

  XtSetArg(args[1], XtNfromVert, pprint);
  XtSetArg(args[2], XtNlabel, "Done");
  cancel = XtCreateManagedWidget("managerCancel",
                                 commandWidgetClass,
                                 defmessHdlrManager_form,
                                 args, 3);
  XtAddCallback(cancel, XtNcallback, CancelSelectSecondary,
                (XtPointer) defmessHdlrManager);

  XtPopup(defmessHdlrManager, XtGrabNonexclusive);

  }

/*******************************************************************************
          Name:        GetDefmessageHandlers
          Description: Gets the defmessage-handlers list
          Arguments:  name of the defclass
          Returns:
*******************************************************************************/
GetDefmessageHandlers(name)
char *name;
{
   VOID *defclass_ptr;
   unsigned index;
   char *buf1,*buf2;
   int maxItems = 20,itemCount = 0;
 
   defclass_ptr = FindDefclass(name);
   if(item_list1 != NULL)
     free(item_list1);
   if((index = GetNextDefmessageHandler(defclass_ptr,0)) == 0)
    {
      item_list1 = NULL;
      return(0);
    }
   item_list1 = (String*)calloc(maxItems,sizeof(String));
   while(index != 0)
    {
      buf1 = GetDefmessageHandlerName(defclass_ptr,index);
      buf2 = GetDefmessageHandlerType(defclass_ptr,index);
      item_list1[itemCount] = balloc(strlen(buf1) + strlen(buf2) + 2,char);
      strcpy(item_list1[itemCount],buf1);
      strcat(item_list1[itemCount]," ");
      strcat(item_list1[itemCount++],buf2);
      if(itemCount == (maxItems - 1))
       {
          maxItems = 2 * maxItems;
          item_list1 = (String*)realloc(item_list1,maxItems * sizeof(String));
       }
      index = GetNextDefmessageHandler(defclass_ptr,index);
    }
   item_list1[itemCount] = NULL;
   sortList(item_list1,itemCount);
   return(itemCount);
}
/*******************************************************************************
          Name:        RemoveMessageHandler
          Description: Take the message-handler out of the list
          Arguments:   w - widget that initiate this call back
                       call_data - not used
                       client_data - the list widget which contains the list
                                     of the defmessage-handlers
          Returns:
*******************************************************************************/
static void RemoveMessageHandler(w,client_data,call_data)
Widget w;
XtPointer call_data,client_data;
{
    char buf[256];
    int i;
    Widget aList = (Widget)client_data;
    XawListReturnStruct *current = XawListShowCurrent(aList);

    if(current->list_index == XAW_LIST_NONE)
      return;
    for (i = 0; current->string[i] != ' '; i++)
      buf[i] = current->string[i];

    buf[i] = 0;
    MoveEndOfFile(dialog_text, &event);
    AppendCommandString("(undefmessage-handler  ");
    PrintCLIPS("wclips","(undefmessage-handler  ");
    AppendCommandString(curr_def_name);
    PrintCLIPS("wclips",curr_def_name);
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(buf);
    PrintCLIPS("wclips",buf);
    strcpy(buf,&(current->string[i]));
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(buf);
    PrintCLIPS("wclips",buf);
    AppendCommandString(")\n");
    PrintCLIPS("wclips",")\n");
    quit_get_event = True;
    list1_change = True;    

}
/*******************************************************************************
          Name:        RemoveMessageHandler
          Description: Take the message-handler out of the list
          Arguments:   w - widget that initiate this call back
                       call_data - not used
                       client_data - the list widget which contains the list
                                     of the defmessage-handlers
          Returns:
*******************************************************************************/
static void PprintMessageHandler(w,client_data,call_data)
Widget w;
XtPointer call_data,client_data;
{
    char buf[256];
    int i;
    Widget aList = (Widget)client_data;
    XawListReturnStruct *current = XawListShowCurrent(aList);

    if(current->list_index == XAW_LIST_NONE)
      return;
    for (i = 0; current->string[i] != ' '; i++)
      buf[i] = current->string[i];

    buf[i] = 0;
    MoveEndOfFile(dialog_text, &event);
    AppendCommandString("(ppdefmessage-handler  ");
    PrintCLIPS("wclips","(ppdefmessage-handler  ");
    AppendCommandString(curr_def_name);
    PrintCLIPS("wclips",curr_def_name);
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(buf);
    PrintCLIPS("wclips",buf);
    strcpy(buf,&(current->string[i]));
    AppendCommandString(" ");
    PrintCLIPS("wclips"," ");
    AppendCommandString(buf);
    PrintCLIPS("wclips",buf);
    AppendCommandString(")\n");
    PrintCLIPS("wclips",")\n");
    quit_get_event = True;

}
/*******************************************************************************
          Name:        Initialize
          Description:
          Arguents:  list -
          Returns:
*******************************************************************************/
InitializeList(list)
  String list[1000];
  {
  register int i  = 0;

  while(list[i] != NULL)
    {
    release(list[i]);
    list[i++] = NULL;
    }
  }

/*******************************************************************************
          Name:        SetManagerList
          Description:
          Arguments:  widget
          Returns:
*******************************************************************************/
SetManagerList(widget)
Widget widget;
  {
  manager_list  = widget;
  }

/*******************************************************************************
          Name:        GetManagerList
          Description:
          Arguments:  None
          Returns:     manager_list
*******************************************************************************/
Widget GetManagerList()
  {
  return(manager_list);
  }

/*******************************************************************************
          Name:        RefreshMngrList
          Description:
          Arguments:  None
          Returns:
          Notes:       manager_list and manager_list1 are the global variables
                       which store the list widget(s) on the current manager window(s)

*******************************************************************************/
RefreshMngrList()
  {

  list_change = False;
  
  if(defrulemanager_flag)
    {
       GetDefruleList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           defrulemanager_flag = False;
         }
       else
        {
           defrulemanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
    }

  else if (deffactsmanager_flag)
    {
       GetFactList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           deffactsmanager_flag = False;
         }
       else
        {
           deffactsmanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }

    }
  else if (deftemplatemanager_flag)
   {
       GetDeftemplateList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           deftemplatemanager_flag = False;
         }
       else
        {
           deftemplatemanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
   }
  else if (deffunctionmanager_flag)
   {
       GetDeffunctionList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           deffunctionmanager_flag = False;
         }
       else
        {
           deffunctionmanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
   }


  else if (defgenericmanager_flag)
   {
       GetDefgenericList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           defgenericmanager_flag = False;
         }
       else
        {
           defgenericmanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
   }

  

  else if (definstancesmanager_flag)
  {
       GetDefinstancesList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           definstancesmanager_flag = False;
         }
       else
        {
           definstancesmanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
   }

  else if (defclassmanager_flag)
   {
       GetDefclassList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           defclassmanager_flag = False;
         }
       else
        {
           defclassmanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
   }

  else if (agendamanager_flag)
    {
       GetAgendaList();
       if(item_list == NULL)
         {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list))));
           ClearParameters();
           agendamanager_flag = False;
         }
       else
        {
           agendamanager_flag = True;
           XawListChange(manager_list,item_list,0,0,False);
        }
     }
  
  if ((defmethodmanager_flag)&&(list1_change))
    {
          
       list1_change = False;
       GetDefmethodList(curr_def_name);
       if(item_list1 == NULL)
        {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list1))));
           defmethodmanager_flag = False;
           release(curr_def_name);
        }
       else
        {
          XawListChange(manager_list1,item_list1,0,0,False);
        }
    
    }
  if((defmessagehandler_flag)&&(list1_change))
   {

       list1_change = False;
       GetDefmessageHandlers(curr_def_name);
       if(item_list1 == NULL)
        {
           XtDestroyWidget(XtParent(XtParent(XtParent(manager_list1))));
           defmessagehandler_flag = False;
           release(curr_def_name);
        }
       else
        {
          XawListChange(manager_list1,item_list1,0,0,False);
        }
   }
  
  }

/*******************************************************************************
          Name:        ClearParameters
          Description:
          Arguments:  None
          Returns:
*******************************************************************************/
ClearParameters()
  {
  list_change = False;
  SetManagerList((Widget)NULL);
  }

/******************************************************************************
          Name:        CancelSelectPrimary
          Description: Destroys popup window for managers
          Arguments:  w - not used
                       client_data - widget to destroy
                       call_data - not used
          Returns:     None
*******************************************************************************/
void CancelSelectPrimary(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget widget = (Widget) client_data;

  XtDestroyWidget(widget);
  ClearParameters();
  free(item_list);
  item_list = NULL;
  defrulemanager_flag = False;
  deffactsmanager_flag = False;
  deftemplatemanager_flag = False;
  deffunctionmanager_flag = False;
  defgenericmanager_flag = False;
  definstancesmanager_flag = False;
  defclassmanager_flag = False;
  agendamanager_flag = False;
  }

/******************************************************************************
          Name:        CancelSelectSecondary
          Description: Destroys popup window for managers
          Arguments:  w - not used
                       client_data - widget to destroy
                       call_data - not used
          Returns:     None
*******************************************************************************/
static void CancelSelectSecondary(w, client_data, call_data)
  Widget w;
  XtPointer client_data, call_data;
  {
  Widget widget = (Widget) client_data;

  XtDestroyWidget(widget);
  list1_change = False;
  manager_list1 = NULL;
  release(curr_def_name);
  free(item_list1);
  item_list1 = NULL;
  }

